Package: lidR
Type: Package
Title: Airborne LiDAR Data Manipulation and Visualization for Forestry
        Applications
Version: 4.3.2
Authors@R: c(
    person("Jean-Romain", "Roussel", email = "info@r-lidar.com", role = c("aut", "cre", "cph")),
    person("David", "Auty", email = "", role = c("aut", "ctb"), comment = "Reviews the documentation"),
    person("Florian", "De Boissieu", email = "", role = ("ctb"), comment = "Fixed bugs and improved catalog features"),
    person("Andrew", "Sánchez Meador", email = "", role = ("ctb"), comment = "Implemented wing2015() for segment_snags()"),
    person("Bourdon", "Jean-François", email = "", role = ("ctb"), comment = "Contributed to Roussel2020() for track_sensor()"),
    person("Gatziolis", "Demetrios", email = "", role = ("ctb"), comment = "Implemented Gatziolis2019() for track_sensor()"),
    person("Leon", "Steinmeier", email = "", role = ("ctb"), comment = "Contributed to parallelization management"),
    person("Stanislaw", "Adaszewski", email = "", role = ("cph"), comment = "Author of the C++ concaveman code"),
    person("Benoît", "St-Onge", email = "", role = "cph", comment = "Author of the 'chm_prep' function"))
Description: Airborne LiDAR (Light Detection and Ranging) interface for data
    manipulation and visualization. Read/write 'las' and 'laz' files, computation
    of metrics in area based approach, point filtering, artificial point reduction,
    classification from geographic data, normalization, individual tree segmentation
    and other manipulations.
URL: https://github.com/r-lidar/lidR
BugReports: https://github.com/r-lidar/lidR/issues
License: GPL-3
Depends: R (>= 3.5.0), methods
Imports: classInt, data.table (>= 1.12.0), glue, grDevices, lazyeval,
        Rcpp (>= 1.0.3), rgl, rlas (>= 1.8.2), sf, stats, stars, terra
        (>= 1.5-17), tools, utils, parallelly
Suggests: EBImage, future, geometry, gstat, raster, RCSF, RMCC, rjson,
        mapview, mapedit, progress, sp, testthat (>= 2.1.0), knitr,
        rmarkdown
RoxygenNote: 7.3.3
LinkingTo: BH (>= 1.72.0),Rcpp,RcppArmadillo
Encoding: UTF-8
ByteCompile: true
VignetteBuilder: knitr
Biarch: true
Collate: 'Class-LAS.R' 'Class-LAScatalog.R' 'Class-LAScluster.R'
        'RcppExports.R' 'add_attribute.R' 'algorithm-dec.R'
        'algorithm-dsm.R' 'algorithm-dtm.R' 'algorithm-gnd.R'
        'algorithm-itd.R' 'algorithm-its.R' 'algorithm-noi.R'
        'algorithm-out.R' 'algorithm-shp.R' 'algorithm-snag.R'
        'algorithm-trk.R' 'backward_compatibility.R' 'catalog_apply.R'
        'catalog_boundaries.R' 'catalog_fakerun.R'
        'catalog_intersect.R' 'catalog_laxindex.R' 'catalog_overlaps.R'
        'catalog_retile.R' 'catalog_select.R' 'classify.R'
        'classify_ground.R' 'classify_noise.R' 'classify_poi.R'
        'clip_roi.R' 'connected_components.R' 'decimate_points.R'
        'io_readLAScatalog.R' 'io_readXLAS.R' 'deprecated.R'
        'doc-drivers.R' 'doc-lidR.R' 'doc-parallelism.R'
        'doc-spatialindex.R' 'engine.R' 'engine_apply.R'
        'engine_chunks.R' 'engine_crop.R' 'engine_index.R'
        'engine_merge.R' 'engine_options.R' 'engine_write.R'
        'fasterize.R' 'filters.R' 'fit_shapes.R' 'fullwaveform.R'
        'generate_las.R' 'io_readLAS.R' 'io_writeLAS.R' 'knn.R'
        'las_check.R' 'las_compression.R' 'las_tools.R'
        'locate_localmaxima.R' 'locate_trees.R' 'merge_spatial.R'
        'methods-LAS.R' 'methods-LAScatalog.R' 'methods-LAScluster.R'
        'methods-LASheader.R' 'metrics_cloud.R' 'metrics_crowns.R'
        'metrics_hexagons.R' 'metrics_pixels.R' 'metrics_plot.R'
        'metrics_point.R' 'metrics_polygon.R' 'metrics_stdmetrics.R'
        'metrics_template.R' 'metrics_voxels.R' 'normalize.R'
        'normalize_height.R' 'normalize_intensity.R' 'plot.R'
        'plot.s3.R' 'plugins.R' 'print.R' 'rasterize.R'
        'rasterize_canopy.R' 'rasterize_density.R'
        'rasterize_terrain.R' 'retrieve_info.R' 'segment.R'
        'segment_shapes.R' 'segment_snags.R' 'segment_trees.R'
        'smooth_height.R' 'st_area.R' 'st_as_sf.R' 'st_bbox.R'
        'st_coordinates.R' 'st_crs.R' 'st_hull.R' 'st_misc.R'
        'st_transform.R' 'track_sensor.R' 'utils_assertive.R'
        'utils_base.R' 'utils_chm.R' 'utils_colors.R'
        'utils_constant.R' 'utils_delaunay.R' 'utils_geometry.R'
        'utils_is.R' 'utils_misc.R' 'utils_raster.R'
        'utils_spatial_index.R' 'utils_threads.R' 'voxelize_points.R'
        'zzz.R'
NeedsCompilation: yes
Packaged: 2026-04-21 11:21:52 UTC; jr
Author: Jean-Romain Roussel [aut, cre, cph],
  David Auty [aut, ctb] (Reviews the documentation),
  Florian De Boissieu [ctb] (Fixed bugs and improved catalog features),
  Andrew Sánchez Meador [ctb] (Implemented wing2015() for
    segment_snags()),
  Bourdon Jean-François [ctb] (Contributed to Roussel2020() for
    track_sensor()),
  Gatziolis Demetrios [ctb] (Implemented Gatziolis2019() for
    track_sensor()),
  Leon Steinmeier [ctb] (Contributed to parallelization management),
  Stanislaw Adaszewski [cph] (Author of the C++ concaveman code),
  Benoît St-Onge [cph] (Author of the 'chm_prep' function)
Maintainer: Jean-Romain Roussel <info@r-lidar.com>
Repository: CRAN
Date/Publication: 2026-04-21 12:32:08 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-04-21 19:20:09 UTC; unix
Archs: lidR.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method("st_crs<-",LAS)
S3method("st_crs<-",LAScatalog)
S3method("st_crs<-",LASheader)
S3method(as.list,LASheader)
S3method(as.spatial,LAS)
S3method(as.spatial,LAScatalog)
S3method(classify_ground,LAS)
S3method(classify_ground,LAScatalog)
S3method(classify_noise,LAS)
S3method(classify_noise,LAScatalog)
S3method(classify_poi,LAS)
S3method(classify_poi,LAScatalog)
S3method(clip_circle,LAS)
S3method(clip_circle,LAScatalog)
S3method(clip_rectangle,LAS)
S3method(clip_rectangle,LAScatalog)
S3method(cloud_metrics,LAS)
S3method(cloud_metrics,LAScluster)
S3method(crown_metrics,LAS)
S3method(crown_metrics,LAScatalog)
S3method(crown_metrics,LAScluster)
S3method(decimate_points,LAS)
S3method(decimate_points,LAScatalog)
S3method(dim,LAS)
S3method(dim,LAScatalog)
S3method(engine_crop,LAS)
S3method(engine_crop,Raster)
S3method(engine_crop,SpatRaster)
S3method(engine_crop,Spatial)
S3method(engine_crop,default)
S3method(engine_crop,sf)
S3method(engine_crop,sfc)
S3method(engine_crop,stars)
S3method(filter_duplicates,LAS)
S3method(filter_duplicates,LAScatalog)
S3method(filter_surfacepoints,LAS)
S3method(filter_surfacepoints,LAScatalog)
S3method(hexagon_metrics,LAS)
S3method(hexagon_metrics,LAScluster)
S3method(las_check,LAS)
S3method(las_check,LAScatalog)
S3method(las_check,LAScluster)
S3method(las_check,LASheader)
S3method(locate_trees,LAS)
S3method(locate_trees,LAScatalog)
S3method(locate_trees,RasterLayer)
S3method(locate_trees,SpatRaster)
S3method(locate_trees,stars)
S3method(merge_spatial,LAS)
S3method(names,LAS)
S3method(names,LAScatalog)
S3method(names,LASheader)
S3method(normalize_height,LAS)
S3method(normalize_height,LAScatalog)
S3method(normalize_intensity,LAS)
S3method(normalize_intensity,LAScatalog)
S3method(npoints,LAS)
S3method(npoints,LAScatalog)
S3method(npoints,LASheader)
S3method(pixel_metrics,LAS)
S3method(pixel_metrics,LAScatalog)
S3method(pixel_metrics,LAScluster)
S3method(plot,lasmetrics3d)
S3method(point_metrics,LAS)
S3method(polygon_metrics,LAS)
S3method(print,LAS)
S3method(print,LAScatalog)
S3method(print,lidRAlgorithm)
S3method(print,raster_template)
S3method(rasterize_canopy,LAS)
S3method(rasterize_canopy,LAScatalog)
S3method(rasterize_canopy,LAScluster)
S3method(rasterize_density,LAS)
S3method(rasterize_density,LAScatalog)
S3method(rasterize_terrain,LAS)
S3method(rasterize_terrain,LAScatalog)
S3method(rasterize_terrain,LAScluster)
S3method(rbind,LAS)
S3method(rbind,LAScatalog)
S3method(readLAS,LAScatalog)
S3method(readLAS,LAScluster)
S3method(readLAS,character)
S3method(rumple_index,RasterLayer)
S3method(rumple_index,SpatRaster)
S3method(rumple_index,matrix)
S3method(rumple_index,numeric)
S3method(rumple_index,stars)
S3method(segment_shapes,LAS)
S3method(segment_shapes,LAScatalog)
S3method(segment_snags,LAS)
S3method(segment_snags,LAScatalog)
S3method(segment_trees,LAS)
S3method(segment_trees,LAScatalog)
S3method(st_area,LAS)
S3method(st_area,LAScatalog)
S3method(st_area,LASheader)
S3method(st_as_sf,LAS)
S3method(st_as_sf,LAScatalog)
S3method(st_bbox,LAS)
S3method(st_bbox,LAScatalog)
S3method(st_bbox,LAScluster)
S3method(st_bbox,LASheader)
S3method(st_concave_hull,LAS)
S3method(st_convex_hull,LAS)
S3method(st_coordinates,LAS)
S3method(st_coordinates,LAScatalog)
S3method(st_crs,LAS)
S3method(st_crs,LAScatalog)
S3method(st_crs,LAScluster)
S3method(st_crs,LASheader)
S3method(st_transform,LAS)
S3method(summary,LAS)
S3method(summary,LAScatalog)
S3method(template_metrics,LAS)
S3method(track_sensor,LAS)
S3method(track_sensor,LAScatalog)
S3method(track_sensor,LAScluster)
S3method(voxelize_points,LAS)
S3method(voxelize_points,LAScatalog)
S3method(voxelize_points,LAScluster)
export("epsg<-")
export("index<-")
export("opt_chunk_alignment<-")
export("opt_chunk_buffer<-")
export("opt_chunk_size<-")
export("opt_filter<-")
export("opt_independent_files<-")
export("opt_laz_compression<-")
export("opt_merge<-")
export("opt_output_files<-")
export("opt_progress<-")
export("opt_restart<-")
export("opt_select<-")
export("opt_stop_early<-")
export("opt_wall_to_wall<-")
export("projection<-")
export("sensor<-")
export("st_crs<-")
export("wkt<-")
export(.stdmetrics)
export(.stdmetrics_ctrl)
export(.stdmetrics_i)
export(.stdmetrics_pulse)
export(.stdmetrics_rn)
export(.stdmetrics_z)
export(.stdshapemetrics)
export(.stdtreemetrics)
export(Gatziolis2019)
export(LAD)
export(LAS)
export(LASBRIGDE)
export(LASBUILDING)
export(LASGROUND)
export(LASHIGHVEGETATION)
export(LASKEYPOINT)
export(LASLOWPOINT)
export(LASLOWVEGETATION)
export(LASMEDIUMVEGETATION)
export(LASNOISE)
export(LASNONCLASSIFIED)
export(LASRAIL)
export(LASROADSURFACE)
export(LASTRANSMISSIONTOWER)
export(LASUNCLASSIFIED)
export(LASWATER)
export(LASWIRECONDUCTOR)
export(LASWIREGUARD)
export(LASheader)
export(Roussel2020)
export(VCI)
export(add_attribute)
export(add_circle3d)
export(add_dtm3d)
export(add_flightlines3d)
export(add_lasattribute)
export(add_lasattribute_manual)
export(add_lasnir)
export(add_lasrgb)
export(add_treetops3d)
export(as.spatial)
export(barycenter_per_voxel)
export(catalog)
export(catalog_apply)
export(catalog_boundaries)
export(catalog_intersect)
export(catalog_map)
export(catalog_retile)
export(catalog_sapply)
export(catalog_select)
export(catalog_split_clusters)
export(classify_ground)
export(classify_noise)
export(classify_poi)
export(clip_circle)
export(clip_polygon)
export(clip_rectangle)
export(clip_roi)
export(clip_transect)
export(cloud_metrics)
export(concaveman)
export(connected_components)
export(count_not_quantized)
export(crown_metrics)
export(csf)
export(dalponte2016)
export(decimate_points)
export(delineate_crowns)
export(density)
export(dsmtin)
export(engine_apply)
export(engine_chunks)
export(engine_crop)
export(engine_merge)
export(engine_write)
export(entropy)
export(epsg)
export(evlr)
export(filter_duplicates)
export(filter_first)
export(filter_firstlast)
export(filter_firstofmany)
export(filter_ground)
export(filter_last)
export(filter_nth)
export(filter_poi)
export(filter_single)
export(filter_surfacepoints)
export(find_trees)
export(fit_circle)
export(forest.colors)
export(gap_fraction_profile)
export(get_lidr_threads)
export(get_range)
export(grid_canopy)
export(grid_density)
export(grid_metrics)
export(grid_terrain)
export(header)
export(height.colors)
export(height_above_ground)
export(hexagon_metrics)
export(highest)
export(highest_attribute_per_voxel)
export(homogenize)
export(index)
export(interpret_waveform)
export(is)
export(is.algorithm)
export(is.indexed)
export(is.overlapping)
export(is.parallelised)
export(is.quantized)
export(ivf)
export(knn)
export(knn_distance)
export(knnidw)
export(knnx)
export(kriging)
export(las_check)
export(las_is_compressed)
export(las_quantize)
export(las_reoffset)
export(las_rescale)
export(las_size)
export(las_update)
export(li2012)
export(lmf)
export(locate_trees)
export(lowest)
export(lowest_attribute_per_voxel)
export(lspikefree)
export(manual)
export(mcc)
export(merge_spatial)
export(ncol.LAS)
export(normalize_height)
export(normalize_intensity)
export(npoints)
export(nrow.LAScatalog)
export(opt_chunk_alignment)
export(opt_chunk_buffer)
export(opt_chunk_size)
export(opt_filter)
export(opt_independent_files)
export(opt_laz_compression)
export(opt_merge)
export(opt_output_files)
export(opt_progress)
export(opt_select)
export(opt_stop_early)
export(opt_wall_to_wall)
export(p2r)
export(pastel.colors)
export(payload)
export(phb)
export(pitfill_stonge2008)
export(pitfree)
export(pixel_metrics)
export(plot)
export(plot_dtm3d)
export(plot_metrics)
export(plugin_decimate)
export(plugin_dsm)
export(plugin_dtm)
export(plugin_gnd)
export(plugin_itd)
export(plugin_its)
export(plugin_nintensity)
export(plugin_outliers)
export(plugin_shape)
export(plugin_snag)
export(plugin_track)
export(pmf)
export(point_eigenvalues)
export(point_metrics)
export(polygon_metrics)
export(projection)
export(ptd)
export(quantize)
export(random)
export(random.colors)
export(random_per_voxel)
export(range_correction)
export(rasterize_canopy)
export(rasterize_density)
export(rasterize_terrain)
export(readALS)
export(readALSLAS)
export(readALSLAScatalog)
export(readALScatalog)
export(readDAPLAS)
export(readDAPLAScatalog)
export(readLAS)
export(readLAScatalog)
export(readLASheader)
export(readMSLAS)
export(readTLS)
export(readTLSLAS)
export(readTLSLAScatalog)
export(readTLScatalog)
export(readUAVLAS)
export(readUAVLAScatalog)
export(remove_ground)
export(remove_lasattribute)
export(remove_noise)
export(remove_water)
export(retrieve_flightlines)
export(retrieve_pulses)
export(retrieve_scanlines)
export(rumple_index)
export(segment_shapes)
export(segment_snags)
export(segment_trees)
export(sensor)
export(set_lidr_threads)
export(shp_hline)
export(shp_hplane)
export(shp_line)
export(shp_plane)
export(shp_vline)
export(silva2016)
export(smooth_height)
export(sor)
export(spikefree)
export(st_area)
export(st_bbox)
export(st_concave_hull)
export(st_crs)
export(stdmetrics)
export(stdmetrics_ctrl)
export(stdmetrics_i)
export(stdmetrics_pulse)
export(stdmetrics_rn)
export(stdmetrics_z)
export(stdshapemetrics)
export(stdtreemetrics)
export(storable_coordinate_range)
export(template_metrics)
export(tin)
export(track_sensor)
export(tree_metrics)
export(unnormalize_height)
export(unsmooth_height)
export(util_makeZhangParam)
export(vlr)
export(voxel_metrics)
export(voxelize_points)
export(watershed)
export(wing2015)
export(wkt)
export(writeLAS)
exportClasses(LAS)
exportClasses(LAScatalog)
exportClasses(LASheader)
exportMethods("$")
exportMethods("$<-")
exportMethods("-")
exportMethods("[")
exportMethods("[[")
exportMethods("[[<-")
exportMethods("crs<-")
exportMethods("epsg<-")
exportMethods("wkt<-")
exportMethods(area)
exportMethods(crs)
exportMethods(density)
exportMethods(epsg)
exportMethods(ext)
exportMethods(is.empty)
exportMethods(plot)
exportMethods(wkt)
import(Rcpp)
import(data.table)
import(methods)
importFrom(data.table,":=")
importFrom(sf,"st_crs<-")
importFrom(sf,st_area)
importFrom(sf,st_as_sf)
importFrom(sf,st_bbox)
importFrom(sf,st_convex_hull)
importFrom(sf,st_coordinates)
importFrom(sf,st_crs)
importFrom(sf,st_transform)
importFrom(stats,na.omit)
importFrom(terra,area)
importFrom(terra,ext)
importFrom(terra,is.empty)
importMethodsFrom(terra,"crs<-")
importMethodsFrom(terra,crs)
useDynLib(lidR, .registration = TRUE)

lidR <img src="https://raw.githubusercontent.com/r-lidar/lidR/master/man/figures/logo200x231.png" align="right"/>
======================================================================================================
![license](https://img.shields.io/badge/Licence-GPL--3-blue.svg) 
[![R build status](https://github.com/r-lidar/lidR/workflows/R-CMD-check/badge.svg)](https://github.com/r-lidar/lidR/actions)
[![Codecov test coverage](https://codecov.io/gh/r-lidar/lidR/branch/master/graph/badge.svg)](https://app.codecov.io/gh/r-lidar/lidR?branch=master)

**R package for Airborne LiDAR Data Manipulation and Visualization for Forestry Applications**

The lidR package provides functions to read and write `.las` and `.laz` files, plot point clouds, compute metrics using an area-based approach, compute digital canopy models, thin LiDAR data, manage a collection of LAS/LAZ files, automatically extract ground inventories, process a collection of tiles using multicore processing, segment individual trees, classify points from geographic data, and provides other tools to manipulate LiDAR data in a **research and development context.**

- 📖 Read [the book](https://r-lidar.github.io/lidRbook/index.html) to get started with the lidR package.
- 💻 Install `lidR` from R with: `install.packages("lidR")`
- 💵 [Sponsor `lidR`](https://github.com/sponsors/Jean-Romain). It is free and open source, but requires time and effort to develop and maintain.

`lidR` has been cited by more than 1,000 scientific papers. To cite the package use `citation()` from within R:

```r
citation("lidR")
#> Roussel, J.R., Auty, D., Coops, N. C., Tompalski, P., Goodbody, T. R. H., Sánchez Meador, A., Bourdon, J.F., De Boissieu, F., Achim, A. (2021). lidR : An R package for analysis of Airborne Laser Scanning (ALS) data. Remote Sensing of Environment, 251 (August), 112061. <doi:10.1016/j.rse.2020.112061>.
#> Jean-Romain Roussel and David Auty (2023). Airborne LiDAR Data Manipulation and Visualization for Forestry Applications. R package version 3.1.0. https://cran.r-project.org/package=lidR
```   

You may also be interested by our new [`lasR`](https://github.com/r-lidar/lasR) package.

# Key features

<img align="right" src="https://raw.githubusercontent.com/Jean-Romain/storage/master/README/point-cloud-rotating.gif">

### Read and display a las file

In R-fashion style the function `plot`, based on `rgl`, enables the user to display, rotate and zoom a point cloud.

```r
las <- readLAS("<file.las>")
plot(las)
```

### Compute a canopy height model

<img align="left" src="https://raw.githubusercontent.com/Jean-Romain/storage/master/README/chm-Khosravipour.png">

`lidR` has several algorithms from the literature to compute canopy height models either point-to-raster based or triangulation based. This allows testing and comparison of some methods that rely on a CHM, such as individual tree segmentation or the computation of a canopy roughness index.

```r
las <- readLAS("<file.las>")

# Khosravipour et al. pitfree algorithm
thr <- c(0,2,5,10,15)
edg <- c(0, 1.5)
chm <- rasterize_canopy(las, 1, pitfree(thr, edg))

plot(chm)
```

<br>

### Read and display a catalog of las files

<img align="right" src="https://raw.githubusercontent.com/Jean-Romain/storage/master/README/catalog-plot_interactive.gif">

`lidR` enables the user to manage, use and process a collection of `las` files. The function `readLAScatalog` builds a `LAScatalog` object from a folder. The function `plot` displays this collection on an interactive map using the `mapview` package (if installed).

```r
ctg <- readLAScatalog("<folder/>")
plot(ctg, map = TRUE)
```

From a `LAScatalog` object the user can (for example) extract some regions of interest (ROI) with `clip_roi()`. Using a catalog for the extraction of the ROI guarantees fast and memory-efficient clipping. `LAScatalog` objects allow many other manipulations that can be done with multicore processing.

<br>

### Individual tree segmentation

<img align="left" src="https://raw.githubusercontent.com/Jean-Romain/storage/master/README/its-rotating-tree-segmented.gif">

The `segment_trees()` function has several algorithms from the literature for individual tree segmentation, based either on the digital canopy model or on the point-cloud. Each algorithm has been coded from the source article to be as close as possible to what was written in the peer-reviewed papers. Our goal is to make published algorithms usable, testable and comparable.

```r
las <- readLAS("<file.las>")

las <- segment_trees(las, li2012())
col <- random.colors(200)
plot(las, color = "treeID", colorPalette = col)
```

### Wall-to-wall dataset processing

<img align="right" src="https://raw.githubusercontent.com/Jean-Romain/storage/master/README/catalog-processing.gif">

Most of the lidR functions can seamlessly process a set of tiles and return a continuous output. Users can create their own methods using the `LAScatalog` processing engine via the `catalog_apply()` function. Among other features the engine takes advantage of point indexation with lax files, takes care of processing tiles with a buffer and allows for processing big files that do not fit in memory.

```r
# Load a LAScatalog instead of a LAS file
ctg <- readLAScatalog("<path/to/folder/>")

# Process it like a LAS file
chm <- rasterize_canopy(ctg, 2, p2r())
col <- random.colors(50)
plot(chm, col = col)
```

### Full waveform

<img align="left" src="https://raw.githubusercontent.com/Jean-Romain/storage/master/FWF/fwf.gif">

lidR can read full waveform data from LAS files and provides interpreter functions to convert the raw data into something easier to manage and display in R. The support of FWF is still in the early stages of development.

```r
fwf <- readLAS("<fullwaveform.las>")

# Interpret the waveform into something easier to manage
las <- interpret_waveform(fwf)

# Display discrete points and waveforms
x <- plot(fwf, colorPalette = "red", bg = "white")
plot(las, color = "Amplitude", add = x)
```

# About

`lidR` is developed openly by [r-lidar](https://www.r-lidar.com/).

The development of `lidR` was made possible through the financial support of [Laval University](https://www.ulaval.ca/en), the [AWARE project](https://awareproject.ca/) and [Ministry of Natural Ressources and Forests](https://www.quebec.ca/en/government/ministere/ressources-naturelles-forets) of Québec. To continue the development of this free software, we now offer consulting, programming, and training services. For more information, please visit [our website](https://www.r-lidar.com/).

# Install dependencies on GNU/Linux

```
# Ubuntu
sudo add-apt-repository ppa:ubuntugis/ubuntugis-unstable
sudo apt-get update
sudo apt-get install libgdal-dev libgeos++-dev libudunits2-dev libproj-dev libx11-dev libgl1-mesa-dev libglu1-mesa-dev libfreetype6-dev libxt-dev libfftw3-dev

# Fedora
sudo dnf install gdal-devel geos-devel udunits2-devel proj-devel mesa-libGL-devel mesa-libGLU-devel freetype-devel libjpeg-turbo-devel
```


  

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 2.5,
  fig.height = 2.5,
  dev.args = list(pointsize = 9)
)
knitr::knit_hooks$set(time_it = local({
  now <- NULL
  function(before, options) {
    if (before) {
      # record the current time before each chunk
      now <<- Sys.time()
    } else {
      # calculate the time difference after a chunk
      res <- difftime(Sys.time(), now)
      # return a character string to show the time
      #if (res > 0.1)
      #paste("<br/>========================<br/>Time for this code chunk ", options$label, " to run:", round(res,2), "<br/>========================<br/>")
    }
  }
}))
knitr::opts_chunk$set(time_it = TRUE)
#rgl::setupKnitr()
options(rmarkdown.html_vignette.check_title = FALSE)

library(lidR)

## ----data, echo = FALSE-------------------------------------------------------
data = structure(list(Max.X = c(332099.99, 333600, 335099.99, 336217.52, 
332099.99, 333599.99, 335099.99, 336368.67, 332099.99, 333599.99, 
335100, 336217.52), Min.X = c(331016.91, 332100.01, 333600.01, 
335100, 331016.91, 332100, 333600, 335100, 331016.92, 332100.01, 
333600.01, 335100.01), Max.Y = c(5529993.99, 5529993.99, 5529993.99, 
5529993.99, 5528399.99, 5528399.99, 5528399.99, 5528399.99, 5526399.98, 
5526399.96, 5526399.99, 5526399.99), Min.Y = c(5528400, 5528400, 
5528400, 5528400, 5526400, 5526400, 5526400, 5526400, 5524793.5, 
5524793.5, 5524800.38, 5524793.5), Max.Z = c(53.53, 47.59, 48.66, 
49.36, 46.13, 48.16, 50.51, 50.86, 45, 74.18, 52.56, 49.33), 
    Min.Z = c(-15.95, -7.87, -3.55, -14.96, -5.94, -11.15, -5.11, 
    -4.12, -9.63, -8.27, -35.88, -20.59), filename = c("folder/file1.las", 
    "folder/file2.las", "folder/file3.las", "folder/file4.las", 
    "folder/file5.las", "folder/file6.las", "folder/file7.las", 
    "folder/file8.las", "folder/file9.las", "folder/file10.las", 
    "folder/file11.las", "folder/file12.las")), row.names = c(NA, 
-12L), class = "data.frame")

geom <- lapply(1:nrow(data), function(i)
{
  mtx <- matrix(c(data$Min.X[i], data$Max.X[i], data$Min.Y[i], data$Max.Y[i])[c(1, 1, 2, 2, 1, 3, 4, 4, 3, 3)], ncol = 2)
  sf::st_polygon(list(mtx))
})

geom <-sf::st_sfc(geom)
sf::st_crs(geom) <- 26917
data <- sf::st_set_geometry(data, geom)

ctg       <- new("LAScatalog")
ctg@data  <- data

## ----setbuffer2, echo = FALSE-------------------------------------------------
opt_chunk_buffer(ctg) <- 0

## ----plotctg, fig.show='hold'-------------------------------------------------
opt_chunk_size(ctg) <- 0 # Processing by files
plot(ctg, chunk = TRUE)

opt_chunk_size(ctg) <- 1000 # Processing chunks of 1000 x 1000
plot(ctg, chunk = TRUE)

## ----setbuffer1, echo = FALSE-------------------------------------------------
opt_chunk_size(ctg) <- 0

## ----plotbuffer, fig.show='hold'----------------------------------------------
opt_chunk_buffer(ctg) <- 0 # No buffer
plot(ctg, chunk = TRUE)

opt_chunk_buffer(ctg) <- 200 # 200 m buffer
plot(ctg, chunk = TRUE)

## ----dtmnobuffer, error=TRUE--------------------------------------------------
try({
opt_chunk_buffer(ctg) <- 0
rasterize_terrain(ctg, 1, tin())
})

## ----alignment, fig.show='hold'-----------------------------------------------
opt_chunk_size(ctg) <- 2000
opt_chunk_buffer(ctg) <- 0
plot(ctg, chunk = TRUE)

opt_chunk_size(ctg) <- 2000
opt_chunk_buffer(ctg) <- 0
opt_chunk_alignment(ctg) <- c(1000, 1000)
plot(ctg, chunk = TRUE)

## ----void, echo = FALSE, rgl=TRUE, dev='png'----------------------------------
#LASfile <- system.file("extdata", "Topography.laz", package="lidR")
#ctg = readLAScatalog(LASfile)
#opt_progress(ctg) <- FALSE
#opt_filter(ctg) <- "-keep_class 2 9"
#las = clip_circle(ctg, 273500, 5274500, 40)
#m = structure(c(0.921, -0.146, 0.362, 0, 0.386, 0.482, -0.787, 0, 
#-0.06, 0.864, 0.5, 0, 0, 0, 0, 1), .Dim = c(4L, 4L))
#plot(las)
#rgl::view3d(fov = 50, userMatrix = m)

## ----writeondisk, echo = FALSE, eval = FALSE----------------------------------
# LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
# ctg2 <- readLAScatalog(LASfile)
# opt_progress(ctg2) <- FALSE
# opt_chunk_size(ctg2) <- 100

## ----template, eval = FALSE---------------------------------------------------
# # Force the results to be written on disk
# opt_output_files(ctg2) <- paste0(tempdir(), "/tree_coordinate_{XLEFT}_{YBOTTOM}")
# trees <- locate_trees(ctg2, lmf(3))
# 
# # The output has been modified by these options and it now gives
# # the paths to the written files (here shapefiles)
# trees
# #> "/tmp/RtmpJQHPNz/tree_coordinate_481200_3812900.shp" "/tmp/RtmpJQHPNz/tree_coordinate_481300_3812900.shp" "/tmp/RtmpJQHPNz/tree_coordinate_481200_3813000.shp"
# #> [4] "/tmp/RtmpJQHPNz/tree_coordinate_481300_3813000.shp"

## ----writechm, eval = FALSE---------------------------------------------------
# # Force the results to be written on disk
# opt_output_files(ctg2) <- paste0(tempdir(), "/tree_coordinate_{XLEFT}_{YBOTTOM}")
# chm <- rasterize_canopy(ctg2, 1, p2r())
# 
# # Many rasters have been written on disk
# # but a light raster has been returned anyway
# chm
# #> class      : RasterLayer
# #> dimensions : 90, 90, 8100  (nrow, ncol, ncell)
# #> resolution : 1, 1  (x, y)
# #> extent     : 481260, 481350, 3812921, 3813011  (xmin, xmax, ymin, ymax)
# #> crs        : +proj=utm +zone=12 +datum=NAD83 +units=m +no_defs
# #> source     : /tmp/RtmpZVJ2hy/rasterize_canopy.vrt
# #> names      : tree_coordinate_481260_3812921
# #> values     : 0, 32.07  (min, max)

## ----clip, fig.show='hold', eval=FALSE----------------------------------------
# opt_output_files(ctg2) <- "{tempdir()}/plot_{ID}"
# new_ctg <- clip_circle(ctg2, x, y, 20)
# new_ctg
# #> class       : LAScatalog (v1.2 format 0)
# #> extent      : 32.372, 163.136, 38.494, 198.636 (xmin, xmax, ymin, ymax)
# #> coord. ref. : NAD83 / UTM zone 17N
# #> area        : 3895.031 m²
# #> points      : 44 points
# #> density     : 8 points/m²
# #> num. files  : 4

## ----echo=FALSE---------------------------------------------------------------
opt_chunk_size(ctg) <- 0
opt_output_files(ctg) <- ""
opt_wall_to_wall(ctg) <- FALSE
opt_progress(ctg) <- TRUE
cl <- engine_chunks(ctg)
for (i in 1:5){
  bbox <- cl[[i]]@bbox
  graphics::rect(bbox[1], bbox[2], bbox[3], bbox[4], border = "black", col = "green3")
}

bbox <- cl[[6]]@bbox
graphics::rect(bbox[1], bbox[2], bbox[3], bbox[4], border = "black", col = "cornflowerblue")


## ----echo=FALSE---------------------------------------------------------------
opt_chunk_size(ctg) <- 0
opt_output_files(ctg) <- ""
opt_wall_to_wall(ctg) <- FALSE
opt_progress(ctg) <- TRUE
cl <- engine_chunks(ctg)
for (i in 1:6){
  bbox <- cl[[i]]@bbox
  graphics::rect(bbox[1], bbox[2], bbox[3], bbox[4], border = "black", col = "green3")
}

bbox <- cl[[7]]@bbox
graphics::rect(bbox[1], bbox[2], bbox[3], bbox[4], border = "black", col = "orange")

bbox <- cl[[8]]@bbox
graphics::rect(bbox[1], bbox[2], bbox[3], bbox[4], border = "black", col = "cornflowerblue")

## ----echo=FALSE---------------------------------------------------------------
opt_chunk_size(ctg) <- 0
opt_output_files(ctg) <- ""
opt_wall_to_wall(ctg) <- FALSE
opt_progress(ctg) <- TRUE
cl <- engine_chunks(ctg)
for (i in 1:8){
  bbox <- cl[[i]]@bbox
  graphics::rect(bbox[1], bbox[2], bbox[3], bbox[4], border = "black", col = "green3")
}

bbox <- cl[[7]]@bbox
graphics::rect(bbox[1], bbox[2], bbox[3], bbox[4], border = "black", col = "orange")

bbox <- cl[[9]]@bbox
graphics::rect(bbox[1], bbox[2], bbox[3], bbox[4], border = "black", col = "red")


## ----echo=FALSE---------------------------------------------------------------
opt_chunk_size(ctg) <- 0
opt_output_files(ctg) <- ""
opt_wall_to_wall(ctg) <- FALSE
opt_progress(ctg) <- TRUE
opt_restart(ctg) <- 9
cl <- engine_chunks(ctg)
for (i in 1:4){
  bbox <- cl[[i]]@bbox
  graphics::rect(bbox[1], bbox[2], bbox[3], bbox[4], border = "black", col = "green3")
}

## ----echo=FALSE---------------------------------------------------------------
opt_chunk_size(ctg) <- 0
opt_output_files(ctg) <- ""
opt_wall_to_wall(ctg) <- FALSE
opt_progress(ctg) <- TRUE
opt_restart(ctg) <- 1
cl <- engine_chunks(ctg)
for (i in 1:length(cl)){
  bbox <- cl[[i]]@bbox
  graphics::rect(bbox[1], bbox[2], bbox[3], bbox[4], border = "black", col = "green3")
}

bbox <- cl[[7]]@bbox
graphics::rect(bbox[1], bbox[2], bbox[3], bbox[4], border = "black", col = "orange")

bbox <- cl[[9]]@bbox
graphics::rect(bbox[1], bbox[2], bbox[3], bbox[4], border = "black", col = "red")


## ----echo=FALSE---------------------------------------------------------------
opt_chunk_size(ctg) <- 400
opt_output_files(ctg) <- ""
opt_wall_to_wall(ctg) <- FALSE
opt_progress(ctg) <- TRUE
cl <- engine_chunks(ctg)
for (i in 1:50){
  bbox <- cl[[i]]@bbox
  graphics::rect(bbox[1], bbox[2], bbox[3], bbox[4], border = "black", col = "green3")
}

bbox <- cl[[1]]@bbox
graphics::rect(bbox[1], bbox[2], bbox[3], bbox[4], border = "black", col = "gray")

bbox <- cl[[2]]@bbox
graphics::rect(bbox[1], bbox[2], bbox[3], bbox[4], border = "black", col = "gray")

bbox <- cl[[3]]@bbox
graphics::rect(bbox[1], bbox[2], bbox[3], bbox[4], border = "black", col = "gray")

bbox <- cl[[14]]@bbox
graphics::rect(bbox[1], bbox[2], bbox[3], bbox[4], border = "black", col = "gray")

bbox <- cl[[15]]@bbox
graphics::rect(bbox[1], bbox[2], bbox[3], bbox[4], border = "black", col = "gray")

bbox <- cl[[16]]@bbox
graphics::rect(bbox[1], bbox[2], bbox[3], bbox[4], border = "black", col = "gray")

bbox <- cl[[29]]@bbox
graphics::rect(bbox[1], bbox[2], bbox[3], bbox[4], border = "black", col = "gray")

## ----echo=FALSE---------------------------------------------------------------
opt_chunk_size(ctg) <- 0
opt_output_files(ctg) <- ""
opt_wall_to_wall(ctg) <- FALSE
opt_progress(ctg) <- TRUE
cl <- engine_chunks(ctg)
for (i in 1:6) {
  bbox <- cl[[i]]@bbox
  graphics::rect(bbox[1], bbox[2], bbox[3], bbox[4], border = "black", col = "green3")
}

for (i in 7:11) {
  bbox <- cl[[i]]@bbox
  graphics::rect(bbox[1], bbox[2], bbox[3], bbox[4], border = "black", col = "cornflowerblue")
}

## -----------------------------------------------------------------------------
ctg$processed <- FALSE
ctg$processed[6:7] <- TRUE
plot(ctg)

## ----echo = FALSE-------------------------------------------------------------
opt_wall_to_wall(ctg) <- TRUE
opt_progress(ctg) <- FALSE

## ----error = TRUE-------------------------------------------------------------
try({
routine <- function(chunk){ 
  las <- readLAS(chunk)
}

catalog_apply(ctg, routine)
})

## ----getachunk, eval=FALSE,echo=FALSE,warning=FALSE,message=FALSE,error=FALSE,results='hide',fig.keep='none'----
# LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
# test = readLAScatalog(LASfile)
# 
# opt_chunk_size(test) <- 150
# opt_chunk_alignment(test) <- c(50,10)
# opt_progress(ctg) <- FALSE
# chunks = engine_chunks(test)
# chunk = chunks[[5]]

## ----rglbuffer, rgl = TRUE, eval = FALSE--------------------------------------
# las <- readLAS(chunk)
# plot(las, color = "buffer")

## ----eval = FALSE-------------------------------------------------------------
# print(chunk)
# #> class       : LAScluster
# #> features    : 1
# #> extent      : 684800, 684950, 5017810, 5017960  (xmin, xmax, ymin, ymax)
# #> crs         : +proj=utm +zone=17 +datum=NAD83 +units=m +no_defs

## ----warning = FALSE, eval = FALSE--------------------------------------------
# raster::extent(chunk)
# #> class      : Extent
# #> xmin       : 684800
# #> xmax       : 684950
# #> ymin       : 5017810
# #> ymax       : 5017960
# sf::st_bbox(chunk)
# #>    xmin    ymin    xmax    ymax
# #>  684800 5017810  684950 5017960

## ----bufferror, error = TRUE--------------------------------------------------
try({
opt_chunk_buffer(ctg) <- 0
rasterize_terrain(ctg, 1, tin())
})

## ----routineerror, error = TRUE-----------------------------------------------
try({
routine <- function(chunk){ 
   las <- readLAS(chunk)
   if (is.empty(las)) return(NULL)
}

options = list(need_buffer = TRUE)
catalog_apply(ctg, routine, .options = options)
})

## ----preparectg, echo=FALSE,warning=FALSE,message=FALSE,error=FALSE,results='hide',fig.keep='none'----
LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
ctg = readLAScatalog(LASfile)

opt_chunk_buffer(ctg) <- 10
opt_chunk_size(ctg) <- 100
opt_chunk_alignment(ctg) <- c(50,50)
opt_progress(ctg) <- FALSE

## ----applyroutine, eval = FALSE-----------------------------------------------
# routine <- function(chunk){
#    las <- readLAS(chunk)               # read the chunk
#    if (is.empty(las)) return(NULL)     # exit if empty
#    ttop <- locate_trees(las, lmf(3))     # make any computation
#    ttop <- sf::st_crop(ttop, st_bbox(chunk))   # remove the buffer
#    return(ttop)
# }
# 
# out <- catalog_apply(ctg, routine)
# class(out)
# #> [1] "list"
# print(out[[1]])
# #> Simple feature collection with 178 features and 2 fields
# #> Attribute-geometry relationship: 2 constant, 0 aggregate, 0 identity
# #> Geometry type: POINT
# #> Dimension:     XYZ
# #> Bounding box:  xmin: 481260.8 ymin: 3812980 xmax: 483299.6 ymax: 3816011
# #> Projected CRS: NAD83 / UTM zone 12N

## ----eval = FALSE-------------------------------------------------------------
# out <- do.call(rbind, out)
# print(out)
# #> Simple feature collection with 17865 features and 2 fields
# #> Attribute-geometry relationship: 2 constant, 0 aggregate, 0 identity
# #> Geometry type: POINT
# #> Dimension:     XYZ
# #> Bounding box:  xmin: 481260.8 ymin: 3812980 xmax: 483299.6 ymax: 3816011
# #> Projected CRS: NAD83 / UTM zone 12N

## ----automerge, eval = FALSE--------------------------------------------------
# options <- list(automerge = TRUE)
# out <- catalog_apply(ctg, routine, .options = options)
# print(out)
# #> Simple feature collection with 17865 features and 2 fields
# #> Attribute-geometry relationship: 2 constant, 0 aggregate, 0 identity
# #> Geometry type: POINT
# #> Dimension:     XYZ
# #> Bounding box:  xmin: 481260.8 ymin: 3812980 xmax: 483299.6 ymax: 3816011
# #> Projected CRS: NAD83 / UTM zone 12N

#' LAScatalog drivers
#'
#' This document explains how objects are written on disk when processing a LAScatalog. As mentioned
#' in \link{LAScatalog-class}, users can set a templated filename to store the outputs on disk instead
#' of in R memory. By defaut `LAS` objects are stored in .las files with \link{writeLAS},
#' raster objects are stored in .tif files using native write function for `raster`, `stars` or `terra`,
#' `Spatial*` objects are stored in .shp files with after coercion to `sf` with \link[sf:st_write]{st_write},
#' `sf` object are stored to `.gpkg` files with \link[sf:st_write]{st_write}.`data.frame` objects are
#' stored in .csv files with \link[data.table:fwrite]{fwrite}, and other objects are not supported.
#' However, users can modify all these default settings and even add new drivers.
#' This manual page explain how. One may also refer to some unofficial documentation
#' \href{https://github.com/r-lidar/lidR/wiki/Modify-the-LAScatalog-drivers}{here} or
#' \href{https://gis.stackexchange.com/questions/325367/configuring-lidr-catalog-to-overwrite-raster-output-with-different-extension}{here}.
#'
#' @section Generic form of a driver:
#' A driver is stored in the  slot `@output_options` of a `LAScatalog`. It is a list that contains:
#' \describe{
#' \item{write}{A function that receives an object and a path, and writes the object into a file using
#' the path. The function can also have extra options.}
#' \item{extension}{A string that gives the file extension.}
#' \item{object}{A string that gives the name of the argument used to pass the object to write in the
#' function used to write the object.}
#' \item{path}{A string that gives the name of the argument used to pass the path of the file to write
#' in the function used to write the object.}
#' \item{param}{A labelled list of extra parameters for the function used to write the object}
#' }
#' For example, the driver to write a  `Raster*` is
#' \preformatted{
#' list(
#'  write = raster::writeRaster,
#'  extension = ".tif",
#'  object = "x",
#'  path = "filename",
#'  param = list(format = "GTiff"))
#' }
#' And the driver to write a `LAS` is
#' \preformatted{
#' list(
#'  write = lidR::writeLAS,
#'  extension = ".las",
#'  object = "las",
#'  path = "file",
#'  param = list())
#' }
#'
#' @section Modify a driver (1/2):
#' Users can modify the drivers to write different file types than the default. For example, to write in
#' shapefile instead of a GeoPackage, one must change the `sf` driver:
#' \preformatted{
#' ctg@output_options$drivers$sf$extension <- ".shp"
#' }
#' To write a `Raster*` in .grd files instead of .tif files one must change the `Raster` driver:
#' \preformatted{
#' ctg@output_options$drivers$Raster$extension <- ".grd"
#' ctg@output_options$drivers$Raster$param$format <- "raster"
#' }
#' To write in .laz files instead of .las files one must change the `LAS` driver:
#' \preformatted{
#' ctg@output_options$drivers$LAS$extension <- ".laz"
#' }
#' @section Add a new driver:
#' The drivers allow `LAS`, `Spatial`,`sf,` `Raster`, `stars`, `SpatRaster` and `data.frame` objects
#' to be written. When using the engine (\link{catalog_apply}) to build new tools, users may need to
#' be able to write other objects such as a `list`. To do that users need to add a `list` element
#' into `@output_options`:
#' \preformatted{
#' ctg@output_options$drivers$list = list(
#'  write = base::saveRDS,
#'  object = "object",
#'  path = "file",
#'  extension = ".rds",
#'  param = list(compress = TRUE))
#' }
#' The `LAScatalog` now has a new driver capable of writing a `list`.
#'
#' @section Modify a driver (2/2):
#' It is also possible to completely overwrite an existing driver. By default `sf` objects are written
#' into GeoPackage with \link[sf:st_write]{st_write}. `st_write` can also wite in GeoJSON and even in
#' SQLlite database objects. But it cannot add data into an existing SQLlite database. Let's create
#' our own driver for a `sf`. First we need a function able to write and append a `sf` into a `SQLlite`
#' database from the object and the path.
#' \preformatted{
#' dbWrite_sf = function(x, path, name)
#' {
#'  x <- sf::st_drop_geometry(x)
#'  con <- RSQLite::dbConnect(RSQLite::SQLite(), path)
#'  RSQLite::dbWriteTable(con, name, x, append = TRUE)
#'  RSQLite::dbDisconnect(con)
#' }}
#' Then we create the driver. User-defined drivers supersede default drivers:
#' \preformatted{
#' ctg@output_options$drivers$sf = list(
#'  write = dbWrite_sf,
#'  extension = ".sqlite",
#'  object = "x",
#'  path = "path",
#'  param = list(name = "layername"))
#' }
#' Then to be sure that we do not write several .sqlite files, we don't use templated filename.
#' \preformatted{
#' opt_output_files(ctg) <- paste0(tempdir(), "/mysqlitefile")}
#' And all the `sf` will be appended in a single database. To preserve the geometry one can
#' @name lidR-LAScatalog-drivers
#' @md
NULL

#' Retrieve individual pulses, flightlines or scanlines
#'
#' Retrieve each individual pulse, individual flightline or individual scanline and assigns a number
#' to each point. The LAS object must be properly populated according to LAS specifications otherwise
#' users could find unexpected outputs.
#'
#' \describe{
#' \item{\code{retrieve_pulses}}{Retrieves each individual pulse. It uses GPS time. An attribute
#' \code{pulseID} is added in the \code{LAS} object}
#' \item{\code{retrieve_scanlines}}{Retrieves each individual scanline. When data are sampled according to a
#' saw-tooth pattern (oscillating mirror), a scanline is one line, or row of data. The function relies
#' on the GPS field time to order the data. Then, the \code{ScanDirectionFlag} attribute is used to
#' retrieve each scanline. An attribute \code{scanlineID} is added in the \code{LAS} object}
#' \item{\code{retrieve_flightlines}}{Retrieves each individual flightline. It uses GPS time. In a
#' continuous dataset, once points are ordered by GPS time, the time between two consecutive points
#' does not exceed a few milliseconds. If the time between two consecutive points is too long it means
#' that the second point is from a different flightline. The default threshold is 30 seconds.
#' An attribute \code{flightlineID} is added in the \code{LAS} object.}
#' }
#'
#' @param las A LAS object
#' @param dt numeric. The threshold time-lag used to retrieve flightlines
#'
#' @return An object of class \code{LAS}
#'
#' @export
#' @rdname identify
#'
#' @examples
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' las <- readLAS(LASfile)
#'
#' las <- retrieve_pulses(las)
#' las
#'
#' las <- retrieve_flightlines(las)
#' #plot(las, color = "flightlineID")
retrieve_pulses = function(las)
{
  stopifnotlas(las)

  if (!"gpstime" %in% names(las))
    stop("No 'gpstime' attribute found. Pulse IDs cannot be computed from this object", call. = FALSE)

  if (all(las@data[["gpstime"]] == 0))
  {
    warning("'gpstime' attribute is populated with 0 only. A single ID has been returned for each point.", call. = FALSE)
    las@data[["pulseID"]] <- 1L
    return(las)
  }

  gpstime <- NULL
  data.table::setorder(las@data, gpstime)
  pulseID <- .lagisdiff(las@data[["gpstime"]])
  las@data[["pulseID"]] <- pulseID
  return(las)
}

#' @export
#' @rdname identify
retrieve_flightlines = function(las, dt = 30)
{
  stopifnotlas(las)
  assert_is_a_number(dt)
  assert_all_are_non_negative(dt)

  if (!"gpstime" %in% names(las))
    stop("No 'gpstime' attribute found. Flightlines cannot be computed from this object", call. = FALSE)

  if (all(las@data[["gpstime"]] == 0))
  {
    warning("'gpstime' attribute is populated with 0 only. A single ID has been returned for each point.", call. = FALSE)
    las@data[["flightlineID"]] <- 1L
    return(las)
  }

  gpstime <- NULL
  data.table::setorder(las@data, gpstime)
  flightlineID <- .lagissup(las@data[["gpstime"]], dt)
  las@data[["flightlineID"]] <- flightlineID
  return(las)
}

#' @export
#' @rdname identify
retrieve_scanlines = function(las)
{
  stopifnotlas(las)

  if (!"gpstime" %in% names(las))
    stop("No 'gpstime' attribute found. Scanlines IDs cannot be computed from this object", call. = FALSE)

  if (!"ScanDirectionFlag" %in% names(las))
    stop("No 'ScanDirectionFlag' attribute found. Scanlines cannot be computed from this object", call. = FALSE)

  values <- unique(las@data[["ScanDirectionFlag"]])

  if (!all(sort(values) == c(0L,1L)))
    stop("'ScanDirectionFlag' attribute is not properly populated according to LAS specifications. Scanlines IDs cannot be computed from this object'", call. = FALSE)

  gpstime <- NULL
  data.table::setorder(las@data, gpstime)
  scanlineID <- .lagisdiff(las@data[["ScanDirectionFlag"]])
  las@data[["scanlineID"]] <- scanlineID
  return(las)
}

.lagissup = function(x, dx)
{
  boo = (x - data.table::shift(x)) > dx
  boo[1] = TRUE
  return(cumsum(boo))
}

.lagisdiff = function(x)
{
  boo = x != data.table::shift(x)
  boo[1] = TRUE
  return(cumsum(boo))
}

context("retrieve")

las = clip_rectangle(megaplot, 684766, 5017870, 684866, 5017970)
las@data = las@data[,c(1:4, 6:7)]
# retrieve_pulses

test_that("retrieve_pulses works", {
  las = retrieve_pulses(las)
  n = names(las)
  expect_true("pulseID" %in% n)
  expect_is(las$pulseID, "integer")
})

test_that("retrieve_pulses fails if no gpstime", {
  las@data[["gpstime"]] <- NULL
  expect_error(retrieve_pulses(las), "No 'gpstime' attribute found")
})

test_that("retrieve_pulses warn if gpstime zeroed", {
  las@data$gpstime <- 0
  expect_warning(retrieve_pulses(las), "populated with 0")

  las <- suppressWarnings(retrieve_pulses(las))
  expect_true(all(las$pulseID == 1L))
})

# retrieve_flightlines

test_that("retrieve_flightlines works", {
  las = retrieve_flightlines(las)

  n = names(las)
  expect_true("flightlineID" %in% n)
  expect_equal(tabulate(las$flightlineID), c(12565,4496))
})

test_that("retrieve_flightlines fails if no gpstime", {
  las@data[["gpstime"]] <- NULL
  expect_error(retrieve_flightlines(las), "No 'gpstime' attribute found")
})

test_that("retrieve_flightlines warn if gpstime zeroed", {
  las@data$gpstime <- 0
  expect_warning(retrieve_flightlines(las), "populated with 0")

  las <- suppressWarnings(retrieve_pulses(las))
  expect_true(all(las$flightlineID == 1L))
})

# retrieve_scanlines

# Connot be tested because no data with a valid scan direction flag
#test_that("retrieve_scanlines works", {
#  las = retrieve_scanlines(las)
#
#  n = names(las)
#  expect_true("retrieve_scanlinesID" %in% n)
#  expect_equal(tabulate(las$flightlineID), c(12565,4496))
#})

test_that("retrieve_scanlines fails if no gpstime", {
  las@data[["gpstime"]] <- NULL
  expect_error(retrieve_scanlines(las), "No 'gpstime' attribute found")
})

#' @rdname aggregate
#' @name aggregate
#' @export
cloud_metrics = function(las, func, ...)
{
  UseMethod("cloud_metrics", las)
}

#' @export
cloud_metrics.LAS = function(las, func, ...)
{
  metrics <- template_metrics(las, func, st_bbox(las), ...)
  if (length(metrics) == 1L) return(metrics[[1]])
  return(metrics)
}

#' @export
cloud_metrics.LAScluster = function(las, func, ...)
{
  las <- readLAS(las)
  if (is.empty(las)) return(NULL)
  is_formula <- tryCatch(lazyeval::is_formula(func), error = function(e) FALSE)
  if (!is_formula) func <- lazyeval::f_capture(func)
  return(cloud_metrics(las, func,...))
}

LASfile <- system.file("extdata", "example.laz", package = "rlas")
las     <- example
ctg     <- suppressMessages(readLAScatalog(c(LASfile, LASfile)))
opt_progress(ctg) <- FALSE

test_that("cloud_metrics works with LAS", {
  x = cloud_metrics(las, mean(Z))
  expect_equal(x, 975.9, tol = 0.1)

  x = cloud_metrics(las, ~mean(Z))
  expect_equal(x, 975.9, tol = 0.1)

  x = cloud_metrics(las, ~list(mean(Z), max(Z)))
  expect_is(x, "list")
  expect_equal(length(x), 2L)
  expect_equal(x[[1]], 975.9, tol = 0.1)
  expect_equal(x[[2]], 978.3, tol = 0.1)
})

test_that("cloud_metrics works with catalog_apply", {
  x = catalog_apply(ctg, cloud_metrics, func = ~mean(Z))
  expect_is(x, "list")
  expect_equal(length(x), 2L)
  expect_equal(x[[1]], 975.9, tol = 0.1)
  expect_equal(x[[2]], 975.9, tol = 0.1)

  x = catalog_apply(ctg, cloud_metrics, func = ~list(mean(Z), max(Z)))
  expect_is(x, "list")
  expect_equal(length(x), 2L)
  expect_is(x[[1]], "list")
  expect_is(x[[2]], "list")
})

#' Algorithms for shape detection of the local point neighbourhood
#'
#' These functions are made to be used in \link{segment_shapes}. They implement algorithms for local
#' neighbourhood shape estimation.
#'
#' In the following, \eqn{a_1, a_2, a_3}{a1, a2, a3} denote the eigenvalues of the covariance matrix
#' of the neighbouring points in ascending order. \eqn{|Z_1|, |Z_2|, |Z_3|}{|Z1|, |Z2|, |Z3|} denote
#' the norm of the Z component of first, second and third axis of the decomposition.
#' \eqn{th_1, th_2, th_3}{th1, th2, th3}  denote a set of threshold values. Points are labelled
#' \code{TRUE} if they meet the following criteria. \code{FALSE} otherwise.\cr
#' \describe{
#' \item{shp_plane}{Detection of plans based on criteria defined by Limberger & Oliveira (2015) (see
#' references). A point is labelled TRUE if the neighborhood is approximately planar, that is:
#' \deqn{a_2 > (th_1 \times a_1) \& (th_2 \times a_2) > a_3}{a2 > (th1*a1) and (th2*a2) > a3}}
#' \item{shp_hplane}{The same as 'plane' but with an extra test on the orientation of the Z vector
#' of the principal components to test the horizontality of the surface.
#' \deqn{a_2 > (th_1 \times a_1) \& (th_2 \times a_2) > a_3 \& |Z_3| > th_3}{a2 > (th1*a1) and (th2*a2) > a3 and |Z3| > th3}
#' In theory  \eqn{|Z_3|}{|Z3|} should be exactly equal to 1. In practice 0.98 or 0.99 should be fine}
#' \item{shp_line}{Detection of lines inspired by the Limberger & Oliveira (2015) criterion. A point is
#' labelled TRUE if the neighbourhood is approximately linear, that is:
#' \deqn{th_1 \times a_2 < a_3 \& th_1 \times a_1 < a_3}{th1*a2 < a3 and th1*a1 < a3}}
#' \item{shp_hline}{Detection of horizontal lines inspired by the Limberger & Oliveira (2015) criterion.
#' A point is labelled TRUE if the neighbourhood is approximately linear and horizontal, that is:
#' \deqn{th_1 \times a_2 < a_3 \& th_1 \times a_1 < a_3 \& |Z_1| < th_2}{th1*a2 < a3 and th1*a1 < a3 and |Z1| < th2}
#' In theory \eqn{|Z_1|}{|Z1|} should be exactly equal to 0. In practice 0.02 or 0.01 should be fine}
#' \item{shp_vline}{Detection of vertical lines inspired by the Limberger & Oliveira (2015) criterion.
#' A point is labelled TRUE if the neighbourhood is approximately linear and vertical, that is:
#' \deqn{th_1 \times a_2 < a_3 \& th_1 \times a_1 < a_3 \& |Z_1| > th_2}{th1*a2 < a3 and th1*a1 < a3 and |Z1| > th2}
#' In theory \eqn{|Z_1|}{|Z1|} should be exactly equal to 1. In practice 0.98 or 0.99 should be fine}
#' }
#'
#' @param th1,th2,th3 numeric. Threshold values (see details)
#' @param k integer. Number of neighbours used to estimate the neighborhood.
#'
#' @references
#' Limberger, F. A., & Oliveira, M. M. (2015). Real-time detection of planar regions in unorganized
#' point clouds. Pattern Recognition, 48(6), 2043–2053. https://doi.org/10.1016/j.patcog.2014.12.020\cr\cr
#'
#' @examples
#' # Generating some data
#' n = 400
#' xplane = runif(n,0,6)
#' yplane = runif(n,0,6)
#' zplane = xplane + 0.8 * yplane + runif(n, 0, 0.1)
#' plane = data.frame(X = xplane, Y = yplane, Z = zplane)
#'
#' xhplane = runif(n,5,15)
#' yhplane = runif(n,0,10)
#' zhplane = 5 + runif(n, 0, 0.)
#' hplane = data.frame(X = xhplane, Y = yhplane, Z = zhplane)
#'
#' tline = 1:n
#' xline = 0.05*tline
#' yline = 0.01*tline
#' zline = 0.02*tline + runif(n, 0, 0.1)
#' line = data.frame(X = xline, Y = yline, Z = zline)
#'
#' thline = 1:n
#' xhline = 0.05*thline + runif(n, 0, 0.05)
#' yhline = 10 - 0.01*thline + runif(n, 0, 0.05)
#' zhline = 3 + runif(n, 0, 0.05)
#' hline = data.frame(X = xhline, Y = yhline, Z = zhline)
#'
#' tvline = 1:n
#' xvline = 5 + runif(n, 0, 0.05)
#' yvline = 5 + runif(n, 0, 0.05)
#' zvline = 0.02*tline
#' vline = data.frame(X = xvline, Y = yvline, Z = zvline)
#'
#' las <- rbind(plane, line, hplane, hline, vline)
#' las <- LAS(las)
#'
#' las <- segment_shapes(las, shp_plane(k = 20), "plane")
#' las <- segment_shapes(las, shp_hplane(k = 20), "hplane")
#' las <- segment_shapes(las, shp_line(k = 20), "line")
#' las <- segment_shapes(las, shp_hline(k = 20), "hline")
#' las <- segment_shapes(las, shp_vline(k = 20), "vline")
#'
#' #plot(las)
#' #plot(las, color = "plane")
#' #plot(las, color = "hplane")
#' #plot(las, color = "line")
#' #plot(las, color = "hline")
#' #plot(las, color = "vline")
#' @name shape_detection
#' @rdname shape_detection
NULL

#' @export
#' @rdname shape_detection
shp_plane = function(th1 = 25, th2 = 6, k = 8)
{
  assert_is_a_number(th1)
  assert_is_a_number(th2)
  assert_is_a_number(k)

  th1 <- lazyeval::uq(th1)
  th2 <- lazyeval::uq(th2)
  k   <- lazyeval::uq(k)

  f = function(las, filter)
  {
    assert_is_valid_context(LIDRCONTEXTSHP, "shp_plane")
    return(C_lasdetectshape(las, 1L , c(th1, th2), k, filter, getThread()))
  }

  f <- plugin_shape(f, TRUE)
  return(f)
}

#' @export
#' @rdname shape_detection
shp_hplane = function(th1 = 25, th2 = 6, th3 = 0.98, k = 8)
{
  assert_is_a_number(th1)
  assert_is_a_number(th2)
  assert_is_a_number(th3)
  assert_is_a_number(k)

  th1 <- lazyeval::uq(th1)
  th2 <- lazyeval::uq(th2)
  th3 <- lazyeval::uq(th3)
  k   <- lazyeval::uq(k)

  f = function(las, filter)
  {
    assert_is_valid_context(LIDRCONTEXTSHP, "shp_hplane")
    return(C_lasdetectshape(las, 2L , c(th1, th2, th3), k, filter, getThread()))
  }

  f <- plugin_shape(f, TRUE)
  return(f)
}

#' @export
#' @rdname shape_detection
shp_line = function(th1 = 10, k = 8)
{
  assert_is_a_number(th1)
  assert_is_a_number(k)

  th1 <- lazyeval::uq(th1)
  k   <- lazyeval::uq(k)

  f = function(las, filter)
  {
    assert_is_valid_context(LIDRCONTEXTSHP, "shp_line")
    return(C_lasdetectshape(las, 3L , th1, k, filter, getThread()))
  }

  f <- plugin_shape(f, TRUE)
  return(f)
}

#' @export
#' @rdname shape_detection
shp_hline = function(th1 = 10, th2 = 0.02, k = 8)
{
  assert_is_a_number(th1)
  assert_is_a_number(k)

  th1 <- lazyeval::uq(th1)
  th2 <- lazyeval::uq(th2)
  k   <- lazyeval::uq(k)

  f = function(las, filter)
  {
    assert_is_valid_context(LIDRCONTEXTSHP, "shp_line")
    return(C_lasdetectshape(las, 4L, c(th1, th2), k, filter, getThread()))
  }

  f <- plugin_shape(f, TRUE)
  return(f)
}

#' @export
#' @rdname shape_detection
shp_vline = function(th1 = 10, th2 = 0.98, k = 8)
{
  assert_is_a_number(th1)
  assert_is_a_number(k)

  th1 <- lazyeval::uq(th1)
  th2 <- lazyeval::uq(th2)
  k   <- lazyeval::uq(k)

  f = function(las, filter)
  {
    assert_is_valid_context(LIDRCONTEXTSHP, "shp_line")
    return(C_lasdetectshape(las, 5L , c(th1, th2), k, filter, getThread()))
  }

  f <- plugin_shape(f, TRUE)
  return(f)
}

context("algorithm dispatch")

las <- lidR:::generate_las(10)

test_that("An error is thrown if algorithms are called in wrong context", {
  f <- lmf(2)
  g <- shp_plane()
  h <- li2012()
  k <- tin()

  expect_error(f(), "The 'lmf' algorithm has not been called in the correct context.")
  expect_error(g(), "The 'shp_plane' algorithm has not been called in the correct context.")
  expect_error(h(), "The 'li2012' algorithm has not been called in the correct context.")
  expect_error(k(), "The 'tin' algorithm has not been called in the correct context.")

  expect_error(locate_trees(las, h), "The algorithm used is not an algorithm for individual tree detection.")
  expect_error(grid_terrain(las, 1, g), "The algorithm used is not an algorithm for spatial interpolation.")
  expect_error(segment_trees(las, k), "The algorithm used is not an algorithm for individual tree segmentation.")
  expect_error(segment_shapes(las, h), "The algorithm used is not an algorithm for shape detection.")
})

test_that("An error is thrown if the input is not an algorithm", {
  expect_error(locate_trees(las, mean), "Invalid function provided as algorithm.")
  expect_error(locate_trees(las, 3), "Invalid function provided as algorithm.")
  expect_error(segment_trees(las, quantile), "Invalid function provided as algorithm.")
  expect_error(segment_shapes(las, 12), "Invalid function provided as algorithm.")
})

context("print")

las     <- example

LASfile <- system.file("extdata", "extra_byte.laz", package = "rlas")
las2     <- readLAS(LASfile)
las2@header@EVLR = las2@header@VLR

las3 <- random_10_points
las3@data$X <- las3@data$X * 1000
las3@data$Y <- las3@data$Y * 1000
las3 = las_update(las3)

folder = system.file("extdata", "", package = "lidR")
ctg = readLAScatalog(folder)

test_that("print works with LAS", {
  sink(tempfile())
  expect_error(print(las), NA)
  sink(NULL)
})

test_that("summary works with LAS", {
  sink(tempfile())
  expect_error(summary(las), NA)
  sink(NULL)
})

test_that("print works with LAS and extra bytes", {
  sink(tempfile())
  expect_error(print(las2), NA)
  sink(NULL)
})

test_that("print works with LAS and extra bytes", {
  sink(tempfile())
  expect_error(print(las3), NA)
  sink(NULL)
})

test_that("summary works with LAS and extra bytes", {
  sink(tempfile())
  expect_error(summary(las2), NA)
  sink(NULL)
})

test_that("print works with LAScatalog", {
  sink(tempfile())
  expect_error(show(ctg), NA)
  expect_error(summary(ctg), NA)
  sink(NULL)
})

test_that("print works with LAScluster", {
  sink(tempfile())
  cl = engine_chunks(ctg)[[1]]
  expect_error(show(cl), NA)
  sink(NULL)
})

test_that("print works with lidR algorithms ", {

  f <- lmf(2)
  g <- shp_plane()
  h <- li2012()
  k <- tin()
  j <- p2r(2)
  l <- random(2)
  m <- pmf(3,4)
  #o <- wing2015()
  p <- Roussel2020()

  sink(tempfile())
  expect_error(print(f), NA)
  expect_error(print(g), NA)
  expect_error(print(h), NA)
  expect_error(print(k), NA)
  expect_error(print(j), NA)
  expect_error(print(l), NA)
  expect_error(print(m), NA)
  #expect_error(print(o), NA)
  expect_error(print(p), NA)
  sink(NULL)

})

test_that("print raster_template works", {
  sink(tempfile())
  tpl <- lidR:::raster_layout(megaplot, 10)
  expect_error(print(tpl), NA)
  sink(NULL)
})

context("segment_shapes")

set.seed(1)
las <- decimate_points(megaplot, random(0.1))
las@data = las@data[,c(1:4, 10)]

test_that("segment_shapes works with shp_coplanar", {
  las <- segment_shapes(las, shp_plane(k = 15), "Coplanar")

  cn <- names(las)

  expect_true("Coplanar" %in% cn)
  expect_true(is.logical(las@data$Coplanar))
  expect_equivalent(as.numeric(table(las$Coplanar)), c(4845, 312))
})

test_that("segment_shapes works with shp_hcoplanar", {
  las <- segment_shapes(las, shp_hplane(k = 15), "Hcoplanar")

  cn <- names(las)

  expect_true("Hcoplanar" %in% cn)
  expect_true(is.logical(las@data$Hcoplanar))
  expect_equivalent(as.numeric(table(las$Hcoplanar)), c(4851, 306))
})


test_that("segment_shapes works with shp_hline", {
  las <- segment_shapes(las, shp_line(k = 8), "line")

  cn <- names(las)

  expect_true("line" %in% cn)
  expect_true(is.logical(las@data$line))
  expect_equivalent(as.numeric(table(las$line)), c(5117, 40))
})

test_that("filter argument works", {
  las <- segment_shapes(las, shp_plane(k = 15), "Coplanar", filter = ~Classification != 2L)

  cn <- names(las)

  expect_true("Coplanar" %in% cn)
  expect_true(is.logical(las@data$Coplanar))
  expect_equivalent(as.numeric(table(las$Coplanar)), c(5125, 32))
  expect_error(segment_shapes(las, shp_plane(k = 15), "Coplanar", filter = ~Intensity > 200), "'Intensity'")
})


test_that("point_eigenvalue works", {
  skip_if(lidR:::is_disable_point_metrics(), "Point metrics disable")

  res1 <- point_eigenvalues(las, k = 5, metrics = TRUE)
  res2 <- point_metrics(las, .stdshapemetrics, k = 5)
  res2[["horizontality"]] = NULL

  expect_equal(res1, res2)

  res1 <- point_eigenvalues(las, k = 5, r = 5, metrics = TRUE)
  res2 <- point_metrics(las, .stdshapemetrics, k = 5, r = 5)
  res2[["horizontality"]] = NULL

  expect_equal(res1, res2)

  res1 <- point_eigenvalues(las, r = 5, metrics = TRUE)
  res2 <- point_metrics(las, .stdshapemetrics, r = 5)
  res2[["horizontality"]] = NULL

  expect_equal(res1, res2)

  res1 <- point_eigenvalues(las, r = 5, metrics = TRUE, filter = ~Classification == 2L)
  res2 <- point_metrics(las, .stdshapemetrics, r = 5,  filter = ~Classification == 2L)
  res2[["horizontality"]] = NULL

  expect_equal(nrow(res1), 477L)
  expect_equal(res1, res2)
})

#' Create a \code{LASheader} object
#'
#' Creates a  \code{LASheader} object either from a raw \code{list} containing all the
#' elements named according to the \code{rlas} package or creates a header from a \code{data.frame}
#' or \code{data.table} containing a point cloud. In the latter case it will generate a header
#' according to the data using \link[rlas:header_create]{rlas::header_create()}. It will
#' guess the LAS file format, the point data format, and initialize the scale factors and offsets,
#' but these may not suit a user's needs. Users are advised to
#' manually modify the results to fit their specific needs.
#'
#' @param data a list containing the data from the header of a LAS file. Can also be
#' a \code{data.frame} or \code{data.table}
#' @return An object of class \code{LASheader}
#' @examples
#' data = data.frame(X = c(339002.889, 339002.983, 339002.918),
#'                   Y = c(5248000.515, 5248000.478, 5248000.318),
#'                   Z = c(975.589, 974.778, 974.471),
#'                   gpstime = c(269347.28141, 269347.28142, 269347.28143),
#'                   Intensity = c(82L, 54L, 27L),
#'                   ReturnNumber = c(1L, 1L, 2L),
#'                   NumberOfReturns = c(1L, 1L, 2L),
#'                   ScanDirectionFlag = c(1L, 1L, 1L),
#'                   EdgeOfFlightline = c(1L, 0L, 0L),
#'                   Classification = c(1L, 1L, 1L),
#'                   ScanAngleRank = c(-21L, -21L, -21L),
#'                   UserData = c(32L, 32L, 32L),
#'                   PointSourceID = c(17L, 17L, 17L))
#'
#' header = LASheader(data)
#' header
#'
#' # Record an EPSG code
#' epsg(header) <- 32618
#' header
#'
#' las <- LAS(data, header)
#' las
#'
#' # The function inferred a LAS 1.2 format 1 which is correct
#' # Upgrade to LAS 1.4 for the example
#' header@VLR <- list() # Erase VLR previously written
#' header@PHB[["Global Encoding"]][["WKT"]] <- TRUE
#' header@PHB[["Version Minor"]] <- 4L
#' header@PHB[["Header Size"]] <- 375L
#' header@PHB[["Offset to point data"]] <- 375L
#' wkt(header) <- sf::st_crs("EPSG:32618")$wkt
#' header
#' las1.4 <- LAS(data, header)
#' las1.4
#' @export
LASheader <- function(data = list()) {return(new("LASheader", data))}


#' Transform to a list
#'
#' Functions to construct, coerce and check for both kinds of R lists.
#'
#' @param x A LASheader object
#' @param ... unused
#' @method as.list LASheader
#' @name as
#' @export
as.list.LASheader <- function(x, ...)
{
  PHB  <- x@PHB
  VLR  <- list(`Variable Length Records` = x@VLR)
  EVLR <- list(`Extended Variable Length Records` = x@EVLR)
  return(c(PHB, VLR, EVLR))
}

#' @export
#' @rdname Extract
setMethod("$", "LASheader", function(x, name) { return(x[[name]]) })

#' @export
#' @rdname Extract
setMethod("$<-", "LASheader", function(x, name, value) { x[[name]] <- value ; return(x) })

#' @export
#' @rdname Extract
setMethod("[[", c("LASheader", "ANY", "missing"), function(x, i, j, ...) {

  assert_is_a_string(i)

  if (i %in% names(x@PHB))
    return(x@PHB[[i]])

  if (i %in% names(x@VLR))
    return(x@VLR[[i]])

  if (i %in% names(x@EVLR))
    return(x@EVLR[[i]])

  return(NULL)
})

#' @export
#' @rdname Extract
setMethod("[[<-", c("LASheader", "character", "missing"), function(x, i, value) {

  assert_is_a_string(i)

  if (i %in% names(x@PHB))
    x@PHB[[i]] <- value

  if (i %in% names(x@VLR))
    x@VLR[[i]] <- value

  if (i %in% names(x@EVLR))
    x@EVLR[[i]] <- value

  return(x)
})

#' Spatial index
#'
#' This document explains how to process point-clouds taking advantage of different spatial
#' indexes available in the lidR package. lidR can use several types of spatial indexes to
#' apply algorithms (that need a spatial indexing) as fast as possible. The choice of the spatial
#' index depends on the type of point-cloud that is processed and the algorithm that is performed.
#' lidR can use a grid partition, a voxel partition, a quadtree or an octree. See details.
#'
#' From lidR (>= 3.1.0), a \link[=LAS-class]{LAS} object records the sensor used to sample
#' the point-cloud (ALS, TLS, UAV, DAP) as well as the spatial index that must be used
#' for processing the point cloud. This can be set manually by the user but the simplest is
#' to use one of the \link[=readLAS]{read*LAS()} functions. By default a point-cloud is associated
#' to a sensor and the best spatial index is chosen on-the-fly depending on the algorithm
#' applied. It is possible to force the use of a specific spatial index.
#'
#' Information relative to the spatial indexing is stored in slot `@index` that contains
#' a `list` with two elements:
#'
#' - `sensor`: an integer that records the sensor type
#' - `index`: an integer that records the spatial index to be used
#'
#' By default the spatial index code is 0 ("automatic") meaning that each function is free
#' to choose a different spatial index depending on the recorded sensor. If the code is not
#' 0 then each function will be forced to used the spatial index that is imposed. This,
#' obviously, applies only to functions that use spatial indexing.
#'
#' \link[=LAScatalog-class]{LAScatalog} objects also record such information that is automatically
#' propagated to the LAS objects when processing.
#'
#' Note: before version 3.1.0, point-clouds were all considered as ALS because lidR was originally
#' designed for ALS. Consequently, for legacy and backwards-compatibility reasons, `readLAS()`
#' and `readALSLAS()` are actually equivalent. `readLAS()` tags the point cloud with "unknown"
#' sensor while `readALSLAS()` tags it with 'ALS'. Both behave the same and this is
#' especially true compared with versions < 3.1. As a consequence, using `readLAS()` provides
#' the same performance (no degradation) than in previous versions, while using one of the `read*LAS()`
#' functions may improve the performance.
#'
#' @template param-las
#' @param h boolean. Human readable. Everything is stored as integers that are understood
#' internally. Use `h = TRUE` for user readable output.
#' @param value integer or character. A code for referring to a sensor type or a spatial
#' index type. Use one of `"unknown"`, `"als"`, `"tls"`, `"uav"`, `"dap"`, `"multispectral"`
#' for sensor type and one of `"auto"`, `"gridpartition"`, `"voxelpartition"`, `"quadtree"`, `"octree"`
#' for spatial index type.
#'
#' @examples
#' LASfile <- system.file("extdata", "example.laz", package="rlas")
#' las <- readLAS(LASfile)
#'
#' # By default the sensor and spatial index codes are 0
#' sensor(las)
#' index(las)
#'
#' # Codes are used internally and not intended to be known by users
#' # Use h option for human readable output
#' sensor(las, h = TRUE)
#' index(las, h = TRUE)
#'
#' # Modification of the sensor enables users to select a better spatial index
#' # when processing the point-cloud.
#' sensor(las) <- "tls"
#' sensor(las, h = TRUE)
#' index(las, h = TRUE)
#'
#' # Modification of the spatial index forces users to choose one of the available
#' # spatial indexes.
#' index(las) <- "quadtree"
#' sensor(las, h = TRUE)
#' index(las, h = TRUE)
#'
#' # The simplest way to take advantage of appropriate spatial indexing is
#' # to use one of the read*LAS() functions.
#' las <- readTLSLAS(LASfile)
#' sensor(las, h = TRUE)
#' index(las, h = TRUE)
#'
#' # But for some specific point-clouds / algorithms it might be advisable to force
#' # the use of a specific spatial index to perform the computation faster
#' index(las) <- "voxelpartition"
#' index(las, h = TRUE)
#'
#' # With a LAScatalog, spatial indexing information is propagated to the
#' # different chunks
#' ctg = readTLSLAScatalog(LASfile)
#' index(ctg) <- "voxelpartition"
#' sensor(ctg, h = TRUE)
#' index(ctg, h = TRUE)
#'
#' # ==================
#' # PERFORMANCE TESTS
#' # ==================
#'
#' \dontrun{
#' # Performance tests on TLS
#' # ------------------------
#'
#' # The package does not include TLS data
#' # so we can generate something that looks TLS-ish
#' # >>>>>>>>>>>
#' X <- runif(50, -25, 25)
#' Y <- runif(50, -25, 25)
#' X <- as.numeric(sapply(X, function(x) rnorm(2000, x, 2)))
#' Y <- as.numeric(sapply(Y, function(x) rnorm(2000, x, 2)))
#' Z <- abs(rnorm(length(Y), 10, 5))
#' veg <- data.frame(X,Y,Z)
#' X <- runif(5000, -30, 30)
#' Y <- runif(5000, -30, 30)
#' Z <- runif(5000, 0, 1)
#' ground <- data.frame(X,Y,Z)
#' X <- runif(30, -30, 30)
#' Y <- runif(30, -30, 30)
#' Z <- runif(30, 0, 30)
#' noise <- data.frame(X,Y,Z)
#' las <- LAS(rbind(ground, veg, noise))
#' # <<<<<<<<<<<<<
#'
#' plot(las)
#'
#' # If read with readALSLAS()
#' sensor(las) <- "als"
#' system.time(classify_noise(las, sor(20, 8)))
#' #> 1.5 sec
#'
#' # If read with readTLSLAS()
#' sensor(las) <- "tls"
#' system.time(classify_noise(las, sor(20, 8)))
#' #> 0.6 sec
#'
#' # Performance tests on ALS
#' # ------------------------
#'
#' # The package does not include large ALS data
#' # so we can generate something that looks ALS-ish
#' # >>>>>>>>>>>
#' X <- runif(4e5, 0, 1000)
#' Y <- runif(4e5, 0, 1000)
#' Z <- 40*sin(0.01*X) + 50*cos(0.005*Y) + abs(rnorm(length(Y), 10, 5))
#' veg <- data.frame(X,Y,Z)
#' X <- runif(100, 0, 1000)
#' Y <- runif(100, 0, 1000)
#' Z <- 40*sin(0.01*X) + 50*cos(0.005*Y) + abs(rnorm(length(Y), 10, 5)) + runif(100, 30, 70)
#' noise <- data.frame(X,Y,Z)
#' las <- LAS(rbind(veg, noise))
#' # <<<<<<<<<<<<<
#'
#' plot(las)
#'
#' # If read with readALSLAS()
#' sensor(las) <- "als"
#' system.time(classify_noise(las, sor(15, 8)))
#' #> 3 sec
#'
#' # If read with readTLSLAS()
#' sensor(las) <- "tls"
#' system.time(classify_noise(las, sor(15, 8)))
#' #> 4.3 sec
#' }
#' @name lidR-spatial-index
#' @md
NULL

#' Classify points
#'
#' Classify points that meet some criterion and/or that belong in a region of interest. The
#' functions updates the attribute `Classification` of the LAS object according to
#' \href{https://community.asprs.org/leadership-restricted/leadership-content/public-documents/standards}{las specifications}
#'
#' \describe{
#' \item{classify_noise}{Classify points as 'noise' (outliers) with several possible algorithms.
#' lidR has: \link{sor}, \link{ivf}. The points classified as 'noise' are assigned a value of 18.}
#' \item{classify_ground}{Classify points as 'ground' with several possible algorithms.
#' lidR has \link{ptd}, \link{pmf}, \link{csf} and \link{mcc}. The points classified as 'ground'
#' are assigned a value of 2. Using `ptd` is strongly recommended.}
#' \item{classify_poi}{Classify points that meet some logical criterion and/or that belong in a
#' region of interest with class of choice.}
#' }
#'
#' @section Non-supported LAScatalog options:
#' The option `select` is not supported and not respected because it always preserves the file format
#' and all the attributes. `select = "*"` is imposed internally.
#'
#' @template param-las
#' @param algorithm An algorithm for classification. lidR has has: \link{sor}, \link{ivf} for noise
#' classification, and \link{pmf}, \link{csf}, \link{mcc} for ground classification (see respective
#' documentation).
#' @param class The ASPRS class to attribute to the points that meet the criterion.
#' @param poi a formula of logical predicates. The points that are `TRUE` will be classified `class`.
#' @param roi A `SpatialPolygons*`, from `sp` or a `sf/sfc_POLYGON` from `sf`.
#' The points that are in the region of interest delimited by the polygon(s) are classified
#' `class`.
#' @param inverse_roi bool. Inverses the `roi`. The points that are outside the polygon(s)
#' are classified `class`.
#' @param by_reference bool. Updates the classification in place (LAS only).
#' @param last_returns logical. The algorithm will use only the last returns (including the first returns
#' in cases of a single return) to run the algorithm. If FALSE all the returns are used. If the attributes
#' \code{'ReturnNumber'} or \code{'NumberOfReturns'} are absent, \code{'last_returns'} is turned
#' to \code{FALSE} automatically.
#'
#' @name classify
#' @rdname classify
#' @md
#' @examples
#' # ===============
#' # Classify ground
#' # ===============
#'
#' if (require(RCSF, quietly = TRUE))
#' {
#' LASfile <- system.file("extdata", "Topography.laz", package="lidR")
#' las <- readLAS(LASfile, select = "xyzrn", filter = "-inside 273450 5274350 273550 5274450")
#'
#' # (Parameters chosen mainly for speed)
#' mycsf <- csf(TRUE, 1, 1, time_step = 1)
#' las <- classify_ground(las, mycsf)
#' #plot(las, color = "Classification")
#' }
#'
#' # ===============
#' # Classify noise
#' # ===============
#'
#' LASfile <- system.file("extdata", "Topography.laz", package="lidR")
#' las <- readLAS(LASfile, filter = "-inside 273450 5274350 273550 5274450")
#'
#' # Add 20 artificial outliers
#' set.seed(314)
#' id = round(runif(20, 0, npoints(las)))
#' set.seed(42)
#' err = runif(20, -50, 50)
#' las$Z[id] = las$Z[id] + err
#'
#' # Using IVF
#' las <- classify_noise(las, ivf(5,2))
#' #plot(las, color = "Classification")
#'
#' # Remove outliers using filter_poi()
#' las_denoise <- filter_poi(las, Classification != LASNOISE)
#'
#' # ===============
#' # Classify POI
#' # ===============
#'
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' shp <- system.file("extdata", "lake_polygons_UTM17.shp", package = "lidR")
#'
#' las  <- readLAS(LASfile, filter = "-keep_random_fraction 0.1")
#' lake <- sf::st_read(shp, quiet = TRUE)
#'
#' # Classifies the points that are NOT in the lake and that are NOT ground points as class 5
#' poi <- ~Classification != LASGROUND
#' las <- classify_poi(las, LASHIGHVEGETATION, poi = poi, roi = lake, inverse = TRUE)
#'
#' # Classifies the points that are in the lake as class 9
#' las <- classify_poi(las, LASWATER, roi = lake, inverse = FALSE)
#'
#' #plot(las, color = "Classification")
NULL

#' Get or set the projection of a LAS* object
#'
#' Get or set the projection of a `LAS*` object.  `st_crs()` extends `sf:st_crs()`, `projection()` and
#' `crs()` extend `raster:projection()` and `raster:crs()`. `projection()` and `crs()` are provided
#' for backward compatibility. For `epsg()` and `wkt()`, see details.
#'
#' There are two ways to store the CRS of a point cloud in a LAS file:
#'
#'    - Store an EPSG code (for LAS 1.0 to 1.3)
#'    - Store a WTK string (for LAS 1.4)
#'
#' On the other hand, R spatial packages use a `crs` object to store the CRS. This is why the CRS is duplicated
#' in a LAS object. The information belongs within the header in a format that can be written in a
#' LAS file and in the slot `crs`, and also in a format that can be understood by other R packages.
#'
#'    - `st_crs` return the CRS in `sf` format.
#'    - `st_crs<-`: assigns a CRS from a `CRS` (`sp`), a `crs` (`sf`), a WKT
#'    string, a proj4string or an epsg code. It updates the header of the LAS
#'    object either with the EPSG code for LAS formats < 1.4 or with a WKT string
#'    - `epsg`: reads the epsg code from the header.
#'    - `wkt`: reads the WKT string from the header.
#'
#' @param x,object,obj An object of class LAS*
#' @param ... Unused.
#' @param asText logical. If TRUE, the projection is returned as text. Otherwise a CRS object is returned.
#' @param value A `CRS` or a `crs` or a `proj4string` string or WKT string or an EPSG code.
#'
#' @return A `st_crs()` returns a `sf::crs`. `projection()` and `crs()` return a `sp::CRS` and should
#' no longer be used.
#'
#' @export
#' @examples
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' las <- readLAS(LASfile)
#'
#' # Get the EPSG code stored in the header (returns 0 if not recorded)
#' epsg(las)
#'
#' # Get the WKT string stored in the header (LAS >= 1.4)
#' wkt(las)
#'
#' # Overwrite the CRS (but does not reproject)
#' st_crs(las) <- 26918
#' las
#'
#' @importFrom sf st_crs
#' @importFrom sf st_crs<-
#' @name st_crs
#' @md
NULL

#' @export
#' @rdname st_crs
st_crs.LAS = function(x, ...) {
  # Workaround to minimize backward incompatibilities
  if (is_las_v3(x))
    return(sf::st_crs(x@proj4string)) #nocov

  return(x@crs)
}

#' @export
#' @rdname st_crs
st_crs.LAScatalog = function(x, ...)
{
  # Workaround to minimize backward incompatibilities
  if (is_lascatalog_v3(x))
    return(sf::st_crs(x@proj4string)) #nocov

  return(sf::st_crs(x@data))
}

#' @export
#' @rdname st_crs
st_crs.LASheader = function(x, ...)
{
  if (use_epsg(x) && epsg(x) != 0L)
    return(epsg2crs(epsg(x)))

  if (use_wktcs(x) && wkt(x) != "")
    return(wkt2crs(wkt(x)))

  return(sf::NA_crs_)
}

#' @export
#' @rdname st_crs
st_crs.LAScluster = function(x, ...) { return(x@crs) }

#' @export st_crs<-
#' @rdname st_crs
#' @name st_crs<-
NULL

#' @export
#' @rdname st_crs
`st_crs<-.LAS` = function(x, value)
{
  if (is(value, "crs") && value == sf::NA_crs_)
  {
    x@crs <- sf::NA_crs_
    return(x)
  }

  # The input is either:
  # - a string > proj4 or WKT
  # - a CRS > from sp
  # - a crs > from sf
  # - a number > EPSG code
  # In all case we need to get a WKT or an EPSG code to update the
  # header depending on the LAS format (1.4 or above)

  # Workaround to repair LAS v3 and minimize backward incompatibilities with v4
  x <- las_v3_repair(x)

  all   <- all_crs_formats(value)
  #CRS   <- all[["CRS"]]
  crs   <- all[["crs"]]
  wkt   <- all[["wkt"]]
  epsg  <- all[["epsg"]]
  #proj4 <- all[["proj4"]]

  if (is.na(crs)) stop("Invalid CRS as input", call. = FALSE)

  if (is.null(epsg) | is.na(epsg)) epsg = 0

  if (use_wktcs(x))
  {
    if (is.null(wkt) || wkt == "" || is.na(wkt))
      x@header@VLR[["WKT OGC CS"]] <- NULL
    else
      wkt(x@header) <- wkt

    x@crs <- crs
  }
  else
  {
    if (epsg == 0L)
    {
      warning(paste0("EPSG code not found: header not updated. Try to provide the ESPG code instead of a ", class(value)[1]), call. = FALSE)
      pos <- rlas::find_epsg_position(as.list(x@header))
      if (pos != 0)  x@header@VLR[["GeoKeyDirectoryTag"]][["tags"]][[pos]] <- NULL
    }
    else
    {
      epsg(x@header) <- epsg
    }

    x@crs <- crs
  }

  return(x)
}

#' @export
#' @rdname st_crs
`st_crs<-.LASheader` = function(x, value)
{
  crs <- all_crs_formats(value)

  if (use_epsg(x))
    epsg(x) <- crs$epsg
  else
    wkt(x) <- crs$wkt

  return(x)
}

#' @export
#' @rdname st_crs
`st_crs<-.LAScatalog` = function(x, value)
{
  # Workaround to repair LAScatalog v3 and minimize backward incompatibilities with v4
  x <- lascatalog_v3_repair(x)

  x@data <- sf::st_set_crs(x@data, sf::NA_crs_)
  x@data <- sf::st_set_crs(x@data, sf::st_crs(value))
  return(x)
}

# ==== PROJECTION ====

#' @export
#' @rdname st_crs
projection <- function(x, asText = TRUE)
{
  CRS <- as(sf::st_crs(x), "CRS")
  if (asText) return(CRS@projargs) else return(CRS)
}

#' @export
#' @rdname st_crs
"projection<-" <- function(x, value)
{
  if (inherits(x, "BasicRaster"))
    raster::projection(x) <- value
  else
    sf::st_crs(x) <- value

  return(x)
}

# ==== CRS ====

#' @importMethodsFrom terra crs
#' @importMethodsFrom terra crs<-
#' @rdname st_crs
setMethod("crs", "LASheader", function(x)
{
  return(projection(x, FALSE))
})

#' @export
#' @rdname st_crs
setMethod("crs", "LAS", function(x)
{
  return(projection(x, FALSE))
})

#' @export
#' @rdname st_crs
setMethod("crs<-", "LAS", function(x, ..., value)
{
  projection(x) <- value
  return(x)
})

#' @export
#' @rdname st_crs
setMethod("crs", "LAScatalog", function(x, asText = FALSE)
{
  return(projection(x, asText))
})

#' @export
#' @rdname st_crs
setMethod("crs", "LAScluster", function(x, asText = FALSE)
{
  return(projection(x, asText))
})

#' @export
#' @rdname st_crs
setMethod("crs<-", "LAScatalog", function(x, ..., value)
{
  projection(x) <- value
  return(x)
})

#' @export
#' @rdname st_crs
setMethod("crs<-", "LASheader", function(x, ..., value)
{
  projection(x) <- value
  return(x)
})

# ==== EPSG ====

#' @export
#' @rdname st_crs
setGeneric("epsg", function(object, ...) standardGeneric("epsg"))

#' @export
#' @rdname st_crs
setGeneric("epsg<-", function(object, value) standardGeneric("epsg<-"))

#' @export
#' @rdname st_crs
setMethod("epsg", "LASheader", function(object, ...)
{
  if (use_wktcs(object))  warning("This LAS object stores the CRS as WKT. CRS field might not be correctly populated, yielding uncertain results; use 'wkt()' instead.", call. = FALSE)
  return(rlas::header_get_epsg(as.list(object)))
})

#' @export
#' @rdname st_crs
setMethod("epsg<-", "LASheader", function(object, value)
{
  if (use_wktcs(object)) stop("This object is not in a format that supports EPSG code. Use a WKT string.", call. = FALSE)
  header <- as.list(object)
  header <- rlas::header_set_epsg(header, value)
  return(LASheader(header))
})

#' @export
#' @rdname st_crs
setMethod("epsg", "LAS", function(object)
{
  return(epsg(object@header))
})

#' @export
#' @rdname st_crs
setMethod("epsg<-", "LAS", function(object, value)
{
  assert_is_a_number(value)
  crs <- epsg2crs(value, fail = TRUE)
  epsg(object@header) <- value
  object@crs <- crs
  return(object)
})

# ==== WKT =====

#' @export
#' @rdname st_crs
setGeneric("wkt", function(obj) standardGeneric("wkt"))

#' @export
#' @rdname st_crs
setGeneric("wkt<-", function(obj, value) standardGeneric("wkt<-"))

#' @export
#' @rdname st_crs
setMethod("wkt", "LASheader", function(obj)
{
  return(rlas::header_get_wktcs(as.list(obj)))
})

#' @export
#' @rdname st_crs
setMethod("wkt<-", "LASheader", function(obj, value)
{
  if (!use_wktcs(obj)) stop("This object is not in a format that supports WKT CRS. Use an EPSG code.", call. = FALSE)
  header <- as.list(obj)
  header <- rlas::header_set_wktcs(header, value)
  return(LASheader(header))
})

#' @export
#' @rdname st_crs
setMethod("wkt", "LAS", function(obj)
{
  return(wkt(obj@header))
})

#' @export
#' @rdname st_crs
setMethod("wkt<-", "LAS", function(obj, value)
{
  assert_is_a_string(value)
  crs <- wkt2crs(value, fail = TRUE)
  wkt(obj@header) <- value
  obj@crs <- crs
  return(obj)
})

# ===== INTERNAL TOOLS =======

use_wktcs <- function(x) {
  UseMethod("use_wktcs", x)
}

use_wktcs.LAS <- function(x) {
  return(x[["Global Encoding"]][["WKT"]])
}

use_wktcs.LASheader <- function(x) {
  return(x@PHB[["Global Encoding"]][["WKT"]])
}

use_epsg <- function(x) {
  UseMethod("use_epsg", x)
}

use_epsg.LAS <- function(x) {
  return(!x[["Global Encoding"]][["WKT"]])
}

use_epsg.LASheader <- function(x) {
  return(!x@PHB[["Global Encoding"]][["WKT"]])
}

all_crs_formats = function(value)
{
  crs <- sf::st_crs(value)

  if (is(value, "CRS"))
  {
    #proj4 <- value@projargs
    #CRS <- as(crs, "CRS")
    wkt <- crs$wkt
    epsg <- crs$epsg
  }
  else if (is(value, "crs"))
  {
    wkt <- value$wkt
    #CRS <- as(value, "CRS")
    #proj4 <- CRS@projargs
    epsg <- crs$epsg
  }
  else if (is.character(value))
  {
    #CRS <- as(crs, "CRS")
    #proj4 <- CRS@projargs
    wkt <- crs$wkt
    epsg <- crs$epsg
  }
  else if (is.numeric(value))
  {
    epsg <- value
    #CRS <- as(crs, "CRS")
    #proj4 <- CRS@projargs
    wkt <- crs$wkt
  }
  else
    stop("'value' is not a CRS, a crs or a string or a number.", call. = FALSE)

  return(list(epsg = epsg, crs = crs, wkt = wkt))
  #return(list(epsg = epsg, CRS = CRS, crs = crs, proj4 = proj4, wkt = wkt))
}

epsg2crs <- function(epsg, fail = FALSE)
{
  crs <- tryCatch(
    {
      crs <- suppressMessages(suppressWarnings(sf::st_crs(epsg)))
      if(is.na(crs)) stop("Error")
      return(crs)
    },
    error = function(e)
    {
      if (!fail)
        return(sf::NA_crs_)
      else
        stop(paste("Invalid epsg code", epsg), call. = FALSE)
    })

  return(crs)
}

wkt2crs <- function(wkt, fail = FALSE)
{
  crs <- tryCatch(
    {
      sf::st_crs(wkt)
    },
    error = function(e)
    {
      if (!fail)
        return(sf::NA_crs_)
      else
        stop("Invalid WKT string", call. = FALSE)
    })

  return(crs)
}

context("projection")

las = random_10_points

test_that("st_crs works", {

  expect_error(st_crs(las), NA)
  expect_error(st_crs(header(las)), NA)
})

test_that("Internal projection conversion works", {

  wkt <- "PROJCS[\"NAD27(CGQ77) / SCoPQ zone 2\",GEOGCS[\"NAD27(CGQ77)\",DATUM[\"North_American_Datum_1927_CGQ77\",SPHEROID[\"Clarke 1866\",6378206.4,294.978698213898,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199433,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4609\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\",-55.5],PARAMETER[\"scale_factor\",0.9999],PARAMETER[\"false_easting\",304800],PARAMETER[\"false_northing\",0],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]"
  expected <- sf::st_crs(2008)

  expect_equal(lidR:::epsg2crs(2008), expected)
  expect_equal(lidR:::epsg2crs(200800), sf::NA_crs_)
  expect_error(lidR:::epsg2crs(200800, fail = TRUE), "Invalid epsg code")
  expect_equal(lidR:::wkt2crs("INVALID"), sf::NA_crs_)
  expect_error(lidR:::wkt2crs("INVALID", fail = TRUE), "Invalid WKT")
})

test_that("st_crs<- with EPSG works", {
  st_crs(las) <- 26919
  expect_equal(epsg(las), 26919)

  st_crs(las) <- 26918
  expect_equal(epsg(las), 26918)
})

test_that("st_crs<- with crs works", {
  st_crs(las) <- sf::st_crs(26919)
  expect_equal(epsg(las), 26919)

  st_crs(las) <- sf::st_crs(26918)
  expect_equal(epsg(las), 26918)
})

test_that("st_crs<- with wkt works", {

  las@header@PHB[["Global Encoding"]][["WKT"]] <- TRUE

  st_crs(las) <- 26919
  expect_match(wkt(las), "NAD83 / UTM zone 19N")

  st_crs(las) <- 26918
  expect_match(wkt(las), "NAD83 / UTM zone 18N")
})

test_that("epsg<- works", {

  epsg(las) <- 26917
  expect_equal(epsg(las), 26917)
})

test_that("Set an invalid code or WKT fails", {
  expect_error({ epsg(las) <- 200800 }, "Invalid epsg")

  las@header@PHB[["Global Encoding"]][["WKT"]] <- TRUE

  expect_error({ wkt(las) <- "INVALID" }, "Invalid WKT")
})

test_that("#323 do not segfault", {
  wkt = "PROJCS[\"NAD83 (2011) / Conus Albers\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101,AUTHORITY[\"EPSG\",7019]],AUTHORITY[\"EPSG\",1116]],PRIMEM[\"Greenwich\",0.0,AUTHORITY[\"EPSG\",8901]],UNIT[\"Degree\",0.0174532925199433,AUTHORITY[\"EPSG\",9102]],AUTHORITY[\"EPSG\",6318]],PROJECTION[\"Albers\",AUTHORITY[\"Esri\",43007]],PARAMETER[\"False_Easting\",0.0,AUTHORITY[\"Esri\",100001]],PARAMETER[\"False_Northing\",0.0,AUTHORITY[\"Esri\",100002]],PARAMETER[\"Central_Meridian\",-96.0,AUTHORITY[\"Esri\",100010]],PARAMETER[\"Standard_Parallel_1\",29.5,AUTHORITY[\"Esri\",100025]],PARAMETER[\"Standard_Parallel_2\",45.5,AUTHORITY[\"Esri\",100026]],PARAMETER[\"Latitude_Of_Origin\",23.0,AUTHORITY[\"Esri\",100021]],UNIT[\"Meter\",1.0,AUTHORITY[\"EPSG\",9001]]],VERTCS[\"NAVD_1988\",VDATUM[\"North_American_Vertical_Datum_1988\",AUTHORITY[\"EPSG\",5103]],PARAMETER[\"Vertical_Shift\",0.0,AUTHORITY[\"Esri\",100006]],PARAMETER[\"Direction\",1.0,AUTHORITY[\"Esri\",100007]],UNIT[\"Meter\",1.0,AUTHORITY[\"EPSG\",9001]],AUTHORITY[\"EPSG\",5703]]"
  expect_error(lidR:::wkt2crs(wkt), NA)
})

test_that("legacy crs works", {
   skip_if_not_installed("sp")
   expect_is(crs(las), "CRS")
})

test_that("legacy projection works", {
  skip_if_not_installed("sp")
  expect_is(projection(las), "character")
})


test_that("legacy crs<- works", {
  skip_if_not_installed("sp")
  skip_on_cran() # For solaris old GDAL/PROJ

  las@header@PHB[["Global Encoding"]][["WKT"]] <- TRUE

  crs(las) <- as(sf::st_crs(26919), "CRS")
  expect_match(wkt(las), "NAD83 / UTM zone 19N")

  las@header@PHB[["Global Encoding"]][["WKT"]] <- FALSE

  crs(las) <- as(sf::st_crs(26918), "CRS")
  expect_equal(epsg(las), 26918)

  las@header@PHB[["Global Encoding"]][["WKT"]] <- TRUE

  crs(las) <- sf::st_crs(26919)
  expect_match(wkt(las), "NAD83 / UTM zone 19N")

  las@header@PHB[["Global Encoding"]][["WKT"]] <- FALSE

  crs(las) <- sf::st_crs(26918)
  expect_equal(epsg(las), 26918)
})

test_that("legacy projection<- works with CRS and crs", {
  skip_if_not_installed("sp")
  skip_on_cran() # For solaris old GDAL/PROJ

  projection(las) <- as(sf::st_crs(26917), "CRS")
  expect_equal(epsg(las), 26917)

  projection(las) <- as(sf::st_crs(26918), "CRS")
  expect_equal(epsg(las), 26918)

  projection(las) <- sf::st_crs(26917)
  expect_equal(epsg(las), 26917)

  projection(las) <- sf::st_crs(26918)
  expect_equal(epsg(las), 26918)
})

fasterize = function(las, layout, buffer = 0, method = c("max", "min", "count"))
{
  methods <- c("max", "min", "count")
  method <- match.arg(method, methods)
  id <- which(method == methods)
  layout <- raster_template(layout)
  C_rasterize(las, layout, buffer, id)
}

rasterize_fast = function(las, res, buffer = 0, method = c("max", "min", "count"), pkg = getOption("lidR.raster.default"))
{
  layout <- raster_layout(las, res, buffer = buffer)
  val    <- fasterize(las, layout, buffer, method)
  layout <- raster_materialize(layout, pkg)
  layout <- raster_set_values(layout, val)
  return(layout)
}

LIDRCIRCLE            <- 0L
LIDRRECTANGLE         <- 1L

LIDRNOBUFFER          <- 0L
LIDRBUFFER            <- 1L
LIDRBOTTOMBUFFER      <- 1L
LIDRLEFTBUFFER        <- 2L
LIDRTOPBUFFER         <- 3L
LIDRRIGHTBUFFER       <- 4L


#' ASPRS LAS Classification
#'
#' A set of global variables corresponding to the point classification defined by the ASPRS for the
#' LAS format. Instead of remembering the classification table of the specification it is possible
#' to use one of these global variables.
#'
#' @examples
#' LASfile <- system.file("extdata", "example.laz", package="rlas")
#' las = readLAS(LASfile)
#' las2 = filter_poi(las, Classification %in% c(LASGROUND, LASWATER))
#'
#' print(LASGROUND)
#' @rdname asprs
#' @name asprs
#' @export
LASNONCLASSIFIED      <- 0L
#' @export
#' @rdname asprs
LASUNCLASSIFIED       <- 1L
#' @export
#' @rdname asprs
LASGROUND             <- 2L
#' @export
#' @rdname asprs
LASLOWVEGETATION      <- 3L
#' @export
#' @rdname asprs
LASMEDIUMVEGETATION   <- 4L
#' @export
#' @rdname asprs
LASHIGHVEGETATION     <- 5L
#' @export
#' @rdname asprs
LASBUILDING           <- 6L
#' @export
#' @rdname asprs
LASLOWPOINT           <- 7L
#' @export
#' @rdname asprs
LASKEYPOINT           <- 8L
#' @export
#' @rdname asprs
LASWATER              <- 9L
#' @export
#' @rdname asprs
LASRAIL               <- 10L
#' @export
#' @rdname asprs
LASROADSURFACE        <- 11L
#' @export
#' @rdname asprs
LASWIREGUARD          <- 13L
#' @export
#' @rdname asprs
LASWIRECONDUCTOR      <- 14L
#' @export
#' @rdname asprs
LASTRANSMISSIONTOWER  <- 15L
#' @export
#' @rdname asprs
LASBRIGDE             <- 17L
#' @export
#' @rdname asprs
LASNOISE              <- 18L

LASALLECHOS <- 0L
LASFIRST <- 1L
LASINTERMEDIATE <- 2L
LASLASTOFMANY <- 3L
LASSINGLE <- 4L
LASMULTIPLE <- 5L
LASLAST <- 6L

CHUNK_WAINTING        <- 0L
CHUNK_OK              <- 1L
CHUNK_NULL            <- 2L
CHUNK_ERROR           <- 3L
CHUNK_WARNING         <- 4L
CHUNK_PROCESSING      <- 5L

LASATTRIBUTES         <- c("X", "Y", "Z", "Intensity",
                           "ReturnNumber", "NumberOfReturns",
                           "ScanDirectionFlag", "EdgeOfFlightline",
                           "Classification",
                           "Synthetic_flag", "Keypoint_flag",
                           "Withheld_flag", "Overlap_flag",
                           "ScanAngle", "ScanAngleRank",
                           "ScannerChannel", "NIR",
                           "UserData", "gpstime", "PointSourceID",
                           "R", "G", "B")

LASCATALOGATTRIBUTES <- c("File.Signature", "File.Source.ID", "GUID", "Version.Major",
                          "Version.Minor", "System.Identifier", "Generating.Software",
                          "File.Creation.Day.of.Year", "File.Creation.Year", "Header.Size",
                          "Offset.to.point.data", "Number.of.variable.length.records",
                          "Point.Data.Format.ID", "Point.Data.Record.Length", "Number.of.point.records",
                          "X.scale.factor", "Y.scale.factor", "Z.scale.factor", "X.offset",
                          "Y.offset", "Z.offset", "Max.X", "Min.X", "Max.Y", "Min.Y", "Max.Z",
                          "Min.Z", "EPSG", "Number.of.1st.return", "Number.of.2nd.return",
                          "Number.of.3rd.return", "Number.of.4th.return", "Number.of.5th.return")

#' @rdname engine
#' @export
#' @param x LAS, Raster, stars, SpatRaster,sf, sfc, Spatial
#' @param bbox bbox
engine_crop <- function(x, bbox)
{
  UseMethod("engine_crop", x)
}

#' @export
engine_crop.LAS <- function(x, bbox)
{
  buffer <- NULL
  x <- filter_poi(x, buffer == LIDRNOBUFFER)
  if (is.empty(x)) return(NULL)
  return(x)
}

#' @export
engine_crop.Raster <- function(x, bbox)
{
  return(raster_crop(x, bbox))
}

#' @export
engine_crop.stars <- function(x, bbox)
{
  return(raster_crop(x, bbox))
}

#' @export
engine_crop.SpatRaster <- function(x, bbox)
{
  return(raster_crop(x, bbox))
}

#' @export
engine_crop.sf <- function(x, bbox)
{
  return(sf::st_crop(x, bbox))
}

#' @export
engine_crop.sfc <- function(x, bbox)
{
  return(sf::st_crop(x, bbox))
}

#' @export
engine_crop.Spatial <- function(x, bbox)
{
  x <- sf::st_as_sf(x)
  x <- sf::st_crop(x, bbox)
  return(sf::as_Spatial(x))
}

#' @export
engine_crop.default <- function(x, bbox)
{
  stop(glue::glue("No method for cropping '{class(x)}'. Use catalog_apply instead of catalog_map."), call. = FALSE)
}

#' @export
#' @rdname readLAS
readALS = function(files,  select = "*", filter = "")
{
  do_no_read_lascluster(files)
  las <- readLAS(files, select, filter)
  las@index <- LIDRALSINDEX
  return(las)
}


#' @export
#' @rdname readLAS
#' @param sort boolean. To optimize even more the computation speed the point cloud is spatially sorted
readTLS = function(files,  select = "*", filter = "", sort = TRUE)
{
  do_no_read_lascluster(files)
  las <- readLAS(files, select, filter)
  las@index <- LIDRTLSINDEX
  if (sort) tls_spatial_sort(las)
  return(las)
}

#' @section Multispectral data:
#' Multispectral laser data are often stored in 3 different files. If this is the case
#' \code{readMSLAS} reads the .las or .laz files of each channel and merges them into
#' an object of class \link[=LAS-class]{LAS} and takes care of attributing an ID to each
#' channel. If the multisprectral point cloud is already stored in a single file
#' leave \code{file2} and \code{file3} missing.
#'
#' @param files1,files2,files3 characters. Path(s) to one or several a file(s). Each argument being
#' one channel (see section 'Multispectral data'). `files2` and `files3` can be missing.
#'
#' @export
#' @rdname readLAS
readMSLAS = function(files1, files2, files3, select = "*", filter = "")
{
  if (missing(files2) && missing(files3))
  {
    las <- readLAS(files1, select, filter)
    las@index <- LIDRMLSINDEX
    return(las)
  }

  if (missing(files3)) files3 <- NULL

  las <- readLAS(c(files1, files2, files3), select, filter)

  if ("ScanAngleRank" %in% names(las))
  {
    data.table::setnames(las@data, "ScanAngleRank", "ScanAngle")
    las@data[["ScanAngle"]] <- as.numeric(las@data[["ScanAngle"]])
  }

  if (!"ScannerChannel" %in% names(las))
  {
    tmp <- readLAS(files1, "", filter)
    n1 <- npoints(tmp)
    tmp <- readLAS(files2, "", filter)
    n2 <- npoints(tmp)

    las@data[["ScannerChannel"]] <- 1L
    las@data[["ScannerChannel"]][(n1+1):(n1+n2)] <- 2L
    if (!is.null(files3))
      las@data[["ScannerChannel"]][(n1+n2+1):npoints(las)] <- 3L

    las@header@PHB[["Point Data Format ID"]] <- 6L
    las@header@PHB[["Version Minor"]] <- 4L
    las@header@PHB[["Header Size"]] <- 375L
    message("Multispectral data read from point format < 6 with no ScannerChannel attribute. The LAS object has been upgraded to LAS 1.4 prf 6 with a ScannerChannel.")
  }
  else
  {
    n <- fast_countequal(las@data[["ScannerChannel"]], 0L)

    if (n > 0 && n < npoints(las))
      warning("Some points have a ScannerChannel of 0 meaning they come from single source sensor.", call. = FALSE)

    if (n == npoints(las)) {
      tmp <- readLAS(files1, "", filter)
      n1 <- npoints(tmp)
      tmp <- readLAS(files2, "", filter)
      n2 <- npoints(tmp)

      las@data[["ScannerChannel"]] <- 1L
      las@data[["ScannerChannel"]][(n1+1):(n1+n2)] <- 2L
      if (!is.null(files3))
        las@data[["ScannerChannel"]][(n1+n2+1):npoints(las)] <- 3L

      message("ScannerChannel was not populated. The ScannerChannel of the LAS object has been automatically populated.")
    }
  }

  las@index <- LIDRMLSINDEX
  return(las)
}

tls_spatial_sort = function(las)
{
  gpstime <- NULL

  las@data[["order"]] = C_voxel_id(las, 0.1)

  if ("gpstime" %in% names(las))
    data.table::setorder(las@data, order, gpstime)
  else
    data.table::setorder(las@data, order)

  las@data[, order := NULL]
  return(invisible())
}

do_no_read_lascluster = function(x)
{
  if (is(x, "LAScluster"))
    stop("Use readLAS() to read a LAScluster object not one of readXXXLAS(). A LAScluster object already contains metadata about the point cloud type.", call. = FALSE)
}

#' @rdname engine
#' @export
#' @param any_list list of LAS, Raster, stars, SpatRaster, sf, sfc, Spatial, data.frame
#' @param ctg LAScatalog
#' @family LAScatalog processing engine
engine_merge = function(ctg, any_list, ...)
{
  object_are_in_files <- opt_output_files(ctg) != ""

  if (!object_are_in_files & length(any_list) == 1)
    return(any_list[[1]])

  # Determine the merging type to apply
  data_types <- sapply(any_list, function(x) { class(x)[1] })
  data_type  <- unique(data_types)
  any_type   <- "unknown"
  elmtone    <- any_list[[1]]

  if (length(data_type) > 1)
  {
    warning("The list returned by 'catalog_apply' contains heterogeneous objects. Merging is impossible. A list has been returned.", call. = FALSE)
    return(any_list)
  }

  # The case with strings is special. Strings should be returned 'as is' but in some cases we can be
  # clever and merge the strings into a LAScatalog or a VRT
  if (object_are_in_files)
  {
    any  <- unlist(any_list)
    ext  <- unique(tools::file_ext(any))

    if (length(ext) > 1)
    {
      warning("The list returned by 'catalog_apply' contains heterogeneous objects. Merging is impossible. A list has been returned.", call. = FALSE)
      return(any_list)
    }

    # LAS and raster are the two special cases supported. In other case we return the list
    # for raster it can be Raster or stars default is stars.
    if (ext %in% c("las", "laz"))
      any_type = "las"
    else if (ext %in% c("grd", "asc", "sdat", "rst", "nc", "tif", "tiff", "envi", "bil", "img"))
    {
      any_type <- attr(elmtone, "rasterpkg")
      if (any_type == "terra") any_type = "rterra"
    }
    else
      return(unlist(any_list))
  }
  else
  {
    x <- elmtone

    if (inherits(x, "Raster")) {
      any_type <- "raster"
    } else if (inherits(x, "stars")) {
      any_type <- "stars"
    } else if (inherits(x, "SpatRaster")) {
      any_type <- "rterra"
    } else if (inherits(x, "SpatVector")) {
      any_type <- "vterra"
    } else if (inherits(x, "LAS")) {
      any_type <- "las"
    } else if (inherits(x, "SpatialPolygons") | inherits(x, "SpatialPoints") | inherits(x, "SpatialLines")) {
      any_type <- "sp"
    } else if (inherits(x, "sf")) {
      any_type <- "sf"
    } else if (inherits(x, "sfc")) {
      any_type <- "sfc"
    } else if  (inherits(x, "data.frame")) {
      any_type <- "dataframe"
    } else {
      any_type <- "unknown"
    }
  }

  # The data_type has been guessed. Merge:
  tryCatch(
  {
    if (any_type == "raster")
    {
      if (object_are_in_files)
      {
        names <- attr(elmtone, "layernames")
        vrt <- raster_build_vrt(any_list, ...)
        raster <- raster::brick(vrt)
        raster_names(raster) <- names

        if (dim(raster)[3] == 1)
          return(raster[[1]])
        else
          return(raster)
      }
      else
      {
        names <- raster_names(any_list[[1]])
        raster <- do.call(raster::mosaic, c(any_list, fun = mean))
        raster_names(raster) <- names
        return(raster)
      }
    }
    if (any_type == "stars")
    {
      if (object_are_in_files)
      {
        names <- attr(elmtone, "layernames")
        vrt <- raster_build_vrt(any_list, ...)
        raster <- stars::read_stars(vrt, proxy = TRUE)
        raster_names(raster) <- names
        return(raster)
      }
      else
      {
        names <- raster_names(any_list[[1]])
        raster <- do.call(stars::st_mosaic, any_list)
        raster_names(raster) <- names
        return(raster)
      }
    }
    if (any_type == "rterra")
    {
      if (object_are_in_files)
      {
        names <- attr(elmtone, "layernames")
        vrt <- raster_build_vrt(any_list, ...)
        raster <- terra::rast(vrt)
        raster_names(raster) <- names
        return(raster)
      }
      else
      {
        names <- raster_names(any_list[[1]])
        raster <- do.call(terra::mosaic, any_list)
        raster_names(raster) <- names
        return(raster)
      }
    }
    else if (any_type == "vterra")
    {
      if (object_are_in_files)
      {
        return(unlist(any_list))
      }
      else
      {
        output <- do.call(rbind, any_list)
        return(output)
      }
    }
    else if (any_type == "las")
    {
      if (object_are_in_files)
      {
        output <- unlist(any_list)
        output <- suppressMessages(suppressWarnings(readLAScatalog(output)))
        opt_copy(output) <- ctg
        return(output)
      }
      else
      {
        return(do.call(rbind, any_list))
      }
    }
    else if (any_type == "sp")
    {
      if (object_are_in_files)
      {
        return(unlist(output))
      }
      else
      {
        output <- do.call(rbind, any_list)
        output@proj4string <- as(st_crs(ctg), "CRS")
        return(output)
      }
    }
    else if (any_type == "sf")
    {
      if (object_are_in_files)
      {
        return(unlist(output))
      }
      else
      {
        output <- data.table::rbindlist(any_list)
        output <- sf::st_as_sf(output)
        sf::st_crs(output) <- st_crs(ctg)
        bbox <- st_bbox(ctg)
        attributes(bbox)$crs <- NULL
        attributes(sf::st_geometry(output))$bbox  <- bbox
        return(output)
      }
    }
    else if (any_type == "sfc")
    {
      if (object_are_in_files)
      {
        return(unlist(output))
      }
      else
      {
        output <- do.call(c, any_list)
        sf::st_crs(output) <- st_crs(ctg)
        return(output)
      }
    }
    else if (any_type == "dataframe")
    {
      if (object_are_in_files)
      {
        return(unlist(output))
      }
      else
      {
        output <- data.table::rbindlist(any_list)
        return(output)
      }
    }
    else
    {
      warning("The list returned by 'catalog_apply' contains unsupported objects. Merging is impossible. A list has been returned.", call. = FALSE)
      return(any_list)
    }
  },
  error = function(e)
  {
    warning("An error occured during the automatic merge of 'catalog_apply'. Merging is impossible. A list has been returned.", call. = FALSE)
    message(e)
    return(any_list)
  })
}

#' LAScatalog processing engine
#'
#' This function gives users access to the \link[=LAScatalog-class]{LAScatalog} processing engine.
#' It allows the application of a user-defined routine over a collection of LAS/LAZ files. The
#' LAScatalog processing engine is explained in the \link[=LAScatalog-class]{LAScatalog class}\cr\cr
#' `catalog_apply()` is the core of the lidR package. It drives every single function that can process a
#' \code{LAScatalog}. It is flexible and powerful but also complex. `catalog_map()` is a simplified version
#' of `catalog_apply()` that suits for 90% of use cases.\cr\cr
#' `catalog_sapply()` is a previous attempt to provide simplified version of `catalog_apply()`. Use
#' `catalog_map()` instead.
#'
#' @param ctg A \link[=LAScatalog-class]{LAScatalog} object.
#' @param FUN A user-defined function that respects a given template (see section function template)
#' @param ... Optional arguments to FUN.
#' @param .options See dedicated section and examples.
#'
#' @section Edge artifacts:
#'
#' It is important to take precautions to avoid 'edge artifacts' when processing wall-to-wall
#' tiles. If the points from neighbouring tiles are not included during certain processes,
#' this could create 'edge artifacts' at the tile boundaries. For example, empty or incomplete
#' pixels in a rasterization process, or dummy elevations in a ground interpolation. The LAScatalog
#' processing engine provides internal tools to load buffered data 'on-the-fly'. `catalog_map()` takes
#' care of removing automatically the results computed in the buffered area to avoid unexpected output
#' with duplicated entries or conflict between values computed twice. It does that in predefined way
#' that may not suit all cases. `catalog_apply()` does not remove the buffer and leave users free
#' to handle this in a custom way. This is why `catalog_apply()` is more complex but gives more freedom
#' to build new applications.
#'
#' @section Buffered data:
#'
#' The LAS objects loaded in theses functions have a special attribute called 'buffer' that indicates,
#' for each point, if it comes from a buffered area or not. Points from non-buffered areas have a
#' 'buffer' value of 0, while points from buffered areas have a 'buffer' value of 1, 2, 3 or 4, where
#' 1 is the bottom buffer and 2, 3 and 4 are the left, top and right buffers, respectively. This allows
#' for filtering of buffer points if required.
#'
#' @section Function template:
#'
#' The parameter `FUN` of `catalog_apply` expects a function with a first argument that will be
#' supplied automatically by the `LAScatalog` processing engine. This first argument is a `LAScluster`.
#' A `LAScluster` is an internal undocumented class but the user needs to know only three things about
#' this class:
#'
#' - It represents a chunk of the file collection
#' - The function \link{readLAS} can be used with a `LAScluster`
#' - The function \link[raster:extent]{extent} or \link[sf:st_bbox]{st_bbox}
#' can be used with a `LAScluster` and they return the bounding box of the chunk without the buffer.
#' It must be used to clip the output and remove the buffered region (see examples).
#'
#' A user-defined function must be templated like this:
#'
#' ```
#' myfun <- function(chunk, ...) {
#'    # Load the chunk + buffer
#'    las <- readLAS(chunk)
#'    if (is.empty(las)) return(NULL)
#'
#'    # do something
#'    output <- do_something(las, ...)
#'
#'    # remove the buffer of the output
#'    bbox <- bbox(chunk)
#'    output <- remove_buffer(output, bbox)
#'    return(output)
#' }
#' ```
#'
#' The line `if (is.empty(las)) return(NULL)` is important because some clusters (chunks) may contain
#' 0 points (we can't know this before reading the file). In this case an empty point cloud with 0 points
#' is returned by `readLAS()` and this may fail in subsequent code. Thus, exiting early from the user-defined
#' function by returning `NULL` indicates to the internal engine that the chunk was empty.
#'
#' `catalog_map` is much simpler (but less versatile). It automatically takes care of reading the chunk
#' and checks if the point cloud is empty. It also automatically crop the buffer. The way it crops the
#' buffer suits for most cases but for some special cases it may be advised to handle this in a more
#' specific way i.e. using `catalog_apply()`. For `catalog_map()` the first argument is a `LAS` and the
#' template is:
#'
#' ```
#' myfun <- function(las, ...) {
#'    # do something
#'    output <- do_something(las, ...)
#'    return(output)
#' }
#' ```
#'
#' @section .options:
#' Users may have noticed that some lidR functions throw an error when the processing options are
#' inappropriate. For example, some functions need a buffer and thus `buffer = 0` is forbidden.
#' Users can add the same constraints to protect against inappropriate options. The `.options`
#' argument is a `list` that allows users to tune the behaviour of the processing engine.
#'
#' - `drop_null = FALSE`: not intended to be used by regular users. The engine does not remove
#' NULL outputs
#' - `need_buffer = TRUE`: the function complains if the buffer is 0.
#' - `need_output_file = TRUE` the function complains if no output file template is provided.
#' - `raster_alignment = ...` the function checks the alignment of the chunks. This option is
#' important if the output is a raster. See below for more details.
#' - `automerge = TRUE` by default the engine returns a `list` with one item per chunk. If
#' `automerge = TRUE`, it tries to merge the outputs into a single object: a `Raster*`, a
#' `Spatial*`, a `LAS*`, an `sf`, a ` stars` similarly to other functions of the package. This is a
#'  fail-safe option so in the worst case, if the merging fails, the `list` is returned. This is
#'  activated by `catalog_map` making it simpler.
#' - `autoread = TRUE`. Introduced in v3.0.0 this option enables to get rid of the first steps of the
#' function i.e `readLAS()` and `if (is.empty())`. In this case the function must take two
#' objects as input, first a `LAS` object and second a `bbox` from `sf`. This is
#' activated by `catalog_map` making it simpler.
#' - `autocrop = TRUE`. Introduced in v4.0.0 this option enables to get rid of the buffer crop steps
#' of the function i.e `something <- remove_buffer(something, bbox)`. In this case the function must
#' take one `LAS` object as input. This is activated by `catalog_map` making it simpler.
#'
#' When the function `FUN` returns a raster it is important to ensure that the chunks are aligned
#' with the raster to avoid edge artifacts. Indeed, if the edge of a chunk does not correspond to the edge
#' of the pixels, the output will not be strictly continuous and will have edge artifacts (that might
#' not be visible). Users can check this with the options `raster_alignment`, that can take the
#' resolution of the raster as input, as well as the starting point if needed. The following are accepted:
#'
#' ```
#' # check if chunks are aligned with a raster of resolution 20
#' raster_alignment = 20
#' raster_alignment = list(res = 20)
#'
#' # check if chunks are aligned with a raster of resolution 20
#' # that starts at (0,10)
#' raster_alignment = list(res = 20, start = c(0,10))
#' ```
#'
#' @examples
#' # More examples might be avaible in the official lidR vignettes or
#' # on the github book <https://jean-romain.github.io/lidRbook/>
#'
#' ## =========================================================================
#' ## Example 1: detect all the tree tops over an entire catalog
#' ## (this is basically a reproduction of the existing function 'locate_trees')
#' ## =========================================================================
#'
#' # 1. Build the user-defined function that analyzes each chunk of the catalog.
#' # The function's first argument is a LAScluster object. The other arguments can be freely
#' # chosen by the users.
#' my_tree_detection_method <- function(chunk, ws)
#' {
#'   # The chunk argument is a LAScluster object. The users do not need to know how it works.
#'   # readLAS will load the region of interest (chunk) with a buffer around it, taking advantage of
#'   # point cloud indexation if possible. The filter and select options are propagated automatically
#'   las <- readLAS(chunk)
#'   if (is.empty(las)) return(NULL)
#'
#'   # Find the tree tops using a user-developed method
#'   # (here simply a LMF for the example).
#'   ttops <- locate_trees(las, lmf(ws))
#'
#'   # ttops is an sf object that contains the tree tops in our region of interest
#'   # plus the trees tops in the buffered area. We need to remove the buffer otherwise we will get
#'   # some trees more than once.
#'   bbox  <- st_bbox(chunk)
#'   ttops <- sf::st_crop(ttops, bbox)
#'   return(ttops)
#' }
#'
#' # 2. Build a collection of file
#' # (here, a single file LAScatalog for the purposes of this simple example).
#' LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
#' ctg <- readLAScatalog(LASfile)
#' plot(ctg)
#'
#' # 3. Set some processing options.
#' # For this small single file example, the chunk size is 100 m + 10 m of buffer
#' opt_chunk_buffer(ctg) <- 10
#' opt_chunk_size(ctg)   <- 100            # Small because this is a dummy example.
#' opt_chunk_alignment(ctg) <- c(-50, -35) # Align such as it creates 2 chunks only.
#' opt_select(ctg)       <- "xyz"          # Read only the coordinates.
#' opt_filter(ctg)       <- "-keep_first"  # Read only first returns.
#'
#' # 4. Apply a user-defined function to take advantage of the internal engine
#' opt    <- list(need_buffer = TRUE,   # catalog_apply will throw an error if buffer = 0
#'                automerge   = TRUE)   # catalog_apply will merge the outputs into a single object
#' output <- catalog_apply(ctg, my_tree_detection_method, ws = 5, .options = opt)
#'
#' plot(output)
#'
#' \donttest{
#' ## =========================================================================
#' ## Example 1: simplified. There is nothing that requires special data
#' ## manipulation in the previous example. Everything can be handled automatically
#' ##=========================================================================
#'
#' # 1. Build the user-defined function that analyzes a point cloud.
#' my_tree_detection_method <- function(las, ws)
#' {
#'   # Find the tree tops using a user-developed method
#'   # (here simply a LMF for the example).
#'   ttops <- locate_trees(las, lmf(ws))
#'   return(ttops)
#' }
#'
#' # 2. Build a project
#' LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
#' ctg <- readLAScatalog(LASfile)
#' plot(ctg)
#'
#' # 3. Set some processing options.
#' # For this dummy example, the chunk size is 100 m and the buffer is 10 m
#' opt_chunk_buffer(ctg) <- 10
#' opt_chunk_size(ctg)   <- 100            # small because this is a dummy example.
#' opt_chunk_alignment(ctg) <- c(-50, -35) # Align such as it creates 2 chunks only.
#' opt_select(ctg)       <- "xyz"          # Read only the coordinates.
#' opt_filter(ctg)       <- "-keep_first"  # Read only first returns.
#'
#' # 4. Apply a user-defined function to take advantage of the internal engine
#' opt    <- list(need_buffer = TRUE)   # catalog_apply will throw an error if buffer = 0
#' output <- catalog_map(ctg, my_tree_detection_method, ws = 5, .options = opt)
#' }
#'
#' \dontrun{
#' ## ===================================================
#' ## Example 2: compute a rumple index on surface points
#' ## ===================================================
#'
#' rumple_index_surface = function(las, res)
#' {
#'   las    <- filter_surfacepoints(las, 1)
#'   rumple <- pixel_metrics(las, ~rumple_index(X,Y,Z), res)
#'   return(rumple)
#' }
#'
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' ctg <- readLAScatalog(LASfile)
#'
#' opt_chunk_buffer(ctg) <- 1
#' opt_chunk_size(ctg)   <- 140     # small because this is a dummy example.
#' opt_select(ctg)       <- "xyz"   # read only the coordinates.
#'
#' opt    <- list(raster_alignment = 20) # catalog_apply will adjust the chunks if required
#' output <- catalog_map(ctg, rumple_index_surface, res = 20, .options = opt)
#'
#' plot(output, col = height.colors(25))
#' }
#' @export
#' @md
catalog_apply <- function(ctg, FUN, ..., .options = NULL)
{
  # Assert correctness of inputs
  assert_is_all_of(ctg, "LAScatalog")
  assert_is_function(FUN)

  # Parse options to check validity or initialize some missing
  opt <- engine_parse_options(.options)

  # Assert correctness and check alignment
  if (opt[["autoread"]] == FALSE) assert_fun_is_null_with_empty_cluster(ctg, FUN, ...)
  assert_processing_constraints_are_repected(ctg, opt[["need_buffer"]], opt[["need_output_file"]])

  # Realignment of the chunk if needed
  realigment <- if (opt[["check_alignment"]]) realigment = opt[["raster_alignment"]] else FALSE

  # Produce the chunks
  clusters <- engine_chunks(ctg, realigment)

  # Disable the progress bar of the functions and ensure user options are restored
  oldstate <- getOption("lidR.progress")
  options(lidR.progress = FALSE)
  on.exit(options(lidR.progress = oldstate), add = TRUE)

  # Process with the catalog processing engine
  output <- engine_apply(clusters, FUN, ctg@processing_options, ctg@output_options, opt[["globals"]], opt[["autoread"]], opt[["autocrop"]], ...)

  # Filter NULLs and return
  if (isTRUE(opt[["drop_null"]]))
    output <- Filter(Negate(is.null), output)

  # Automerge
  if (!isFALSE(opt[["automerge"]]) && opt_merge(ctg))
    output <- engine_merge(ctg, output, as.character(substitute(FUN)))

  return(output)
}

#' @export
#' @rdname catalog_apply
catalog_sapply <- function(ctg, FUN, ..., .options = NULL)
{
  if (is.null(.options))
    .options <- list(automerge = TRUE)
  else
  {
    .options$automerge <- TRUE
  }

  return(catalog_apply(ctg, FUN, ..., .options = .options))
}

#' @export
#' @rdname catalog_apply
catalog_map <- function(ctg, FUN, ..., .options = NULL)
{
  if (is.null(.options))
    .options <- list(automerge = TRUE,
                     autocrop  = TRUE,
                     autoread  = TRUE)
  else
  {
    .options$automerge <- TRUE
    .options$autocrop  <- TRUE
    .options$autoread  <- TRUE
  }

  return(catalog_apply(ctg, FUN, ..., .options = .options))
}

assert_fun_is_null_with_empty_cluster = function(ctg, FUN, ...)
{
  if (opt_wall_to_wall(ctg) == TRUE)
  {
    cl <- LAScluster(list(x = 0, y = 0), 0, 0, 0, LIDRRECTANGLE, system.file("extdata", "example.laz", package = "rlas"), "noname")
    cl@select <- "*"

    u <- tryCatch(FUN(cl, ...), error = function(e)
    {
      stop(paste0("The function is tested before starting processing the collection and failed with following error:\n", e), call. = FALSE)
    })

    if (!is.null(u))
      stop("User's function does not return NULL for empty chunks. Please see the documentation of catalog_apply.", call. = FALSE)
  }
}

assert_processing_constraints_are_repected <- function(ctg, need_buffer, need_output_file)
{
  # The function expects a buffer to guarantee a strict wall-to-wall output
  if (need_buffer & opt_chunk_buffer(ctg) <= 0 && opt_wall_to_wall(ctg))
    stop("A buffer greater than 0 is required to process the catalog.", call. = FALSE)

  # The function requires the output file template to be written in files
  # (because the output is likely to be too big to be returned in R)
  if (need_output_file & opt_output_files(ctg) == "")
    stop("This function requires that the LAScatalog provides an output file template.", call. = FALSE)
}

engine_parse_options = function(.option)
{
  output <- list()

  # Alignment

  raster_alignment <- .option$raster_alignment

  if (is.null(raster_alignment))
  {
    raster_alignment <- NULL
    check_alignment  <- FALSE
  }
  else if (is.numeric(raster_alignment))
  {
    assert_is_a_number(raster_alignment)
    assert_all_are_non_negative(raster_alignment)
    raster_alignment <- list(res = raster_alignment, start = c(0,0))
    check_alignment <- TRUE
  }
  else if (is.list(raster_alignment))
  {
    assert_is_a_number(raster_alignment$res)
    assert_all_are_non_negative(raster_alignment$res)
    check_alignment <- TRUE

    if (!is.null(raster_alignment$start))
      assert_is_numeric(raster_alignment$start)
  }
  else
    stop("Invalid parameter 'raster_alignment")

  output$raster_alignment <- raster_alignment
  output$check_alignment  <- check_alignment

  # output file

  if (is.null(.option$need_output_file))
  {
    need_output_file <- FALSE
  }
  else
  {
    need_output_file <- .option$need_output_file
    assert_is_a_bool(need_output_file)
  }

  output$need_output_file <- need_output_file

  # drop null

  if (is.null(.option$drop_null))
  {
    drop_null <- TRUE
  }
  else
  {
    drop_null <- .option$drop_null
    assert_is_a_bool(drop_null)
  }

  output$drop_null <- drop_null

  # automerge

  if (is.null(.option$automerge))
  {
    automerge <- FALSE
  }
  else
  {
    automerge <- .option$automerge

    if (isTRUE(automerge))
      automerge <- "auto"
  }

  output$automerge <- automerge

  # autoread

  if (is.null(.option$autoread))
  {
    autoread <- FALSE
  }
  else
  {
    autoread <- .option$autoread

    if (isTRUE(autoread))
      autoread <- TRUE
  }

  output$autoread <- autoread

  # autocrop

  if (is.null(.option$autocrop))
  {
    autocrop <- FALSE
  }
  else
  {
    autocrop <- .option$autocrop

    if (isTRUE(autocrop))
      autocrop <- TRUE
  }

  output$autocrop <- autocrop

  # need buffer

  if (is.null(.option$need_buffer))
  {
    need_buffer <- FALSE
  }
  else
  {
    need_buffer <- .option$need_buffer
    assert_is_a_bool(need_buffer)
  }

  output$need_buffer <- need_buffer

  # export globals

  if (is.null(.option$globals))
    output$globals <- NULL
  else
    output$globals <- .option$globals

  return(output)
}

test_that("catalog_apply autoread works", {

  skip_on_cran()

  ctg <- random_2files_250points
  opt_progress(ctg) = FALSE

  test <- function(las, bbox, layers = 1L) {
    sf <- sf::st_as_sf(las[1:10])
    sf::st_agr(sf) <- "constant"
    sp <- sf::st_crop(sf, bbox)
    return(sp)
  }

  # automerge option
  option <- list(autoread = TRUE, automerge = TRUE)
  req2 <- catalog_apply(ctg, test, .options = option)

  expect_equal(nrow(req2), 20)
})

#' Pits and spikes filling
#'
#' Pits and spikes filling for raster. Typically used for post-processing CHM. This algorithm
#' is from St-Onge 2008 (see reference).
#'
#' @param x raster. SpatRaster, RasterLayer, stars.
#' @param lap_size integer. Size of the Laplacian filter kernel (integer value, in pixels).
#' @param thr_lap numeric. Threshold Laplacian value for detecting a cavity (all values above this
#' value will be considered a cavity). A positive value.
#' @param thr_spk numeric. Threshold Laplacian value for detecting a spike (all values below this
#' value will be considered a spike). A negative value.
#' @param med_size integer. Size of the median filter kernel (integer value, in pixels).
#' @param dil_radius integer. Dilation radius (integer value, in pixels).
#'
#' @references
#' St-Onge, B., 2008. Methods for improving the quality of a true orthomosaic of Vexcel UltraCam
#' images created using alidar digital surface model, Proceedings of the Silvilaser 2008, Edinburgh,
#' 555-562. https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=81365288221f3ac34b51a82e2cfed8d58defb10e
#'
#' @examples
#' LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
#' las <- readLAS(LASfile)
#' chm <- rasterize_canopy(las, 0.5, dsmtin())
#' sto <- pitfill_stonge2008(chm)
#'
#' #terra::plot(c(chm, sto), col = lidR::height.colors(25))
#' @export
pitfill_stonge2008 = function(x, lap_size = 3L, thr_lap = 0.1, thr_spk = -0.1, med_size = 3L, dil_radius = 0L)
{
  res <- raster_res(x)
  z   <- raster_as_matrix(x)$z
  dim = dim(z)
  z = as.numeric(t(apply(z,1,rev)))
  z[is.na(z)] = -99999
  y = C_chm_prep(z, dim[2], dim[1], lap_size, thr_lap, thr_spk, med_size, dil_radius, -99999)
  y[y == -99999] = NA
  y = matrix(y, nrow = dim[1], ncol = dim[2])
  x[] = t(y)
  return(x)
}

#' Decimate a LAS object
#'
#' Reduce the number of points using several possible algorithms.
#'
#' @template param-las
#' @param algorithm function. An algorithm of point decimation. \code{lidR} have: \link{random},
#' \link{homogenize}, \link{highest}, \link{lowest}, \link{random_per_voxel} and \link{barycenter_per_voxel}.
#'
#' @section Non-supported LAScatalog options:
#' The option `select` is not supported and not respected because it always preserves the file format
#' and all the attributes. `select = "*"` is imposed internally.\cr
#' The options `chunk buffer` is not supported and not respected because it is not needed.
#'
#' @template return-lasfilter-las-lascatalog
#'
#' @export
#'
#' @examples
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' las <- readLAS(LASfile, select = "xyz")
#'
#' # Select points randomly to reach an overall density of 1
#' thinned1 <- decimate_points(las, random(1))
#' #plot(rasterize_density(las))
#' #plot(rasterize_density(thinned1))
#'
#' # Select points randomly to reach an homogeneous density of 1
#' thinned2 <- decimate_points(las, homogenize(1,5))
#' #plot(rasterize_density(thinned2))
#'
#' # Select the highest point within each pixel of an overlayed grid
#' thinned3 = decimate_points(las, highest(5))
#' #plot(thinned3)
decimate_points = function(las, algorithm)
{
  UseMethod("decimate_points", las)
}

#' @export
decimate_points.LAS = function(las, algorithm)
{
  assert_is_algorithm(algorithm)
  assert_is_algorithm_dec(algorithm)
  lidR.context <- "decimate_points"
  selected <- algorithm(las)
  return(las[selected])
}

#' @export
decimate_points.LAScatalog = function(las, algorithm)
{
  # Defensive programming
  assert_is_algorithm(algorithm)
  assert_is_algorithm_dec(algorithm)

  # Enforce some options
  opt_select(las) <- "*"
  opt_chunk_buffer(las) <- 0

  e <- environment(algorithm)
  if (!is.null(e[["res"]])) opt_chunk_buffer(las) <- e[["res"]]

  # Processing
  options <- list(need_buffer = FALSE, drop_null = TRUE, need_output_file = TRUE)
  output  <- catalog_map(las, decimate_points, algorithm = algorithm, .options = options)
  return(output)
}

context("merge_spatial")

las     <- decimate_points(megaplot, highest(7))
sflakes <- sf::st_read(system.file("extdata", "", package="lidR"), "lake_polygons_UTM17", quiet = TRUE)
sf::st_agr(sflakes) <- "constant"
sflakes <- sf::st_crop(sflakes, st_bbox(las))
sflakes$ID <- 1L

zmean   <- pixel_metrics(las, ~mean(Z))
rgb     <- stars::st_as_stars(st_bbox(las), dx = 10, dy = 10, nz = 3L, values = runif(300, 0, 255))
rgbi    <- stars::st_as_stars(st_bbox(las), dx = 20, dy = 20, nz = 4L, values = 10)

test_that("merge_spatial works with SpatialPolygonsDataFrame", {

  skip_if_not_installed("sp")
  splakes <- sf::as_Spatial(sflakes)

  las <- merge_spatial(las, splakes, "inlakes")
  cn <- names(las)

  expect_true("inlakes" %in% cn)
  expect_true(is.logical(las$inlakes))
  expect_equivalent(as.numeric(table(las$inlakes)), c(966, 216))

  las <- merge_spatial(las, splakes, "LAKENAME_1")
  cn <- names(las)

  expect_true("LAKENAME_1" %in% cn)
  expect_equal(typeof(las$LAKENAME_1), typeof(splakes$LAKENAME_1))
  expect_equivalent(as.numeric(table(las$LAKENAME_1)), c(216))

  las <- merge_spatial(las, splakes)
  cn <- names(las)

  expect_true("id" %in% cn)
  expect_true(is.integer(las$id))
  expect_equivalent(as.numeric(table(las$id)), c(216))
})

test_that("merge_spatial works with sf", {

  las <- merge_spatial(las, sflakes, "inlakes")
  cn  <- names(las)

  expect_true("inlakes" %in% cn)
  expect_true(is.logical(las$inlakes))
  expect_equivalent(as.numeric(table(las$inlakes)), c(966, 216))

  las <- merge_spatial(las, sflakes, "LAKENAME_1")
  cn <- names(las)

  expect_true("LAKENAME_1" %in% cn)
  expect_equal(typeof(las$LAKENAME_1), typeof(sflakes$LAKENAME_1))
  expect_equivalent(as.numeric(table(las$LAKENAME_1)), c(216))

  las <- merge_spatial(las, sflakes)
  cn <- names(las)

  expect_true("id" %in% cn)
  expect_true(is.integer(las$id))
  expect_equivalent(as.numeric(table(las$id)), c(216))
})

test_that("merge_spatial preserve storage mode", {

  las <- merge_spatial(las, sflakes, "ID")
  cn  <- names(las)

  expect_equal(storage.mode(las$ID), storage.mode(sflakes$ID))
})


# test_that("merge_spatial works with SpatialPolygons", {
#   skip_if_not_installed("sp")
#   lakes <- as(sflakes, "SpatialPolygons")
#
#   las <- merge_spatial(las, lakes)
#   cn <- names(las)
#
#   expect_true("id" %in% cn)
#   expect_true(is.integer(las$id))
#   expect_equivalent(as.numeric(table(las$id)), c(216))
# })

test_that("merge_spatial works with sfc", {

  las <- merge_spatial(las, sf::st_geometry(sflakes), "inlakes")
  cn <- names(las)

  expect_true("inlakes" %in% cn)
  expect_true(is.logical(las$inlakes))
  expect_equivalent(as.numeric(table(las$inlakes)), c(966, 216))

  las <- merge_spatial(las, sf::st_geometry(sflakes))
  cn <- names(las)

  expect_true("id" %in% cn)
  expect_true(is.integer(las$id))
  expect_equivalent(as.numeric(table(las$id)), c(216))
})

test_that("merge_spatial does not fail if no polygon encompass the points", {

  shift <- sf::st_sfc(sf::st_point(c(2000,200)), crs = sf::st_crs(sflakes))
  lakes <- sf::st_geometry(sflakes) + shift

  las <- merge_spatial(las, lakes)
  cn  <- names(las)

  expect_true("id" %in% cn)
  expect_true(is.integer(las$id))
  expect_true(all(is.na(las$id)))
})

test_that("merge_spatial works with raster", {
  skip_if_not_installed("raster")
  las <- merge_spatial(las, as(zmean, "Raster"), "Zmean")
  cn  <- names(las)

  expect_true("Zmean" %in% cn)
  expect_true(is.numeric(las$Zmean))
  expect_equal(mean(las$Zmean), 17.9, tol = 0.01)
})

test_that("merge_spatial works with stars", {

  las <- merge_spatial(las, zmean, "Zmean")
  cn  <- names(las)

  expect_true("Zmean" %in% cn)
  expect_true(is.numeric(las$Zmean))
  expect_equal(mean(las$Zmean), 17.9, tol = 0.01)
})

test_that("merge_spatial works a RGB RasterBrick", {
  skip_if_not_installed("raster")
  las <- merge_spatial(las, as(rgb, "Raster"))
  cn  <- names(las)

  expect_true(all(c("R", "G", "B") %in% cn))
  expect_true(is.integer(las$R))
  expect_equal(las[["Point Data Format ID"]], 3L)
})

test_that("merge_spatial works a RGB stars", {

  las <- merge_spatial(las, rgb)
  cn  <- names(las)

  expect_true(all(c("R", "G", "B") %in% cn))
  expect_true(is.integer(las$R))
  expect_equal(las[["Point Data Format ID"]], 3L)
})

test_that("merge_spatial works a RGB SpatRaster", {

  las <- merge_spatial(las, as(rgb, "SpatRaster"))
  cn  <- names(las)

  expect_true(all(c("R", "G", "B") %in% cn))
  expect_true(is.integer(las$R))
  expect_equal(las[["Point Data Format ID"]], 3L)
})


test_that("merge_spatial fails with too much bands", {

  expect_error(merge_spatial(las, rgbi), "rasters must have 1 or 3 bands")
})

test_that("merge_spatial fails with unknown input type", {

  expect_error(merge_spatial(las, 4), "No method for this source format")
})

test_that("merge_spatial do not fail with 1 point (#347)", {

  one_in <- filter_poi(las, Intensity == 105)

  res1 <- merge_spatial(one_in, sflakes)
  #res2 <- merge_spatial(one_in, splakes)
  res3 <- merge_spatial(one_in, sf::st_geometry(sflakes))

  expect_equal(res1$id, 1L)
  #expect_equal(res2$id, 1L)
  expect_equal(res3$id, 1L)
})

#' @rdname engine
#' @export
#' @param path strings
#' @param drivers list. Drivers of a LAScatalog
engine_write = function(x, path, drivers)
{
  dir <- dirname(path)
  if (!dir.exists(dir)) dir.create(dir, recursive = TRUE)

  if (class(x)[1] %in% names(drivers))
    driver <- drivers[[class(x)[1]]]
  else if (inherits(x, "LAS"))
    driver <- drivers$LAS # nocov
  else if (inherits(x, "stars"))
    driver <- drivers$stars
  else if (inherits(x, "Raster"))
    driver <- drivers$Raster
  else if (inherits(x, "Spatial"))
    driver <- drivers$Spatial
  else if (inherits(x, "sf"))
    driver <- drivers$sf
  else if (inherits(x, "data.frame"))
    driver <- drivers$data.frame
  else if (is(x, "lidr_internal_skip_write"))
    return(x)
  else
    stop(glue::glue("Trying to write an object of class {class(x)} but this type is not supported."))

  path <- paste0(path, driver$extension)
  if (is_raster(x)) {
    attr(path, "rasterpkg") <- raster_pkg(x)
    attr(path, "layernames") <- raster_names(x)
  }

  driver$param[[driver$object]] <- x
  driver$param[[driver$path]]   <- path
  do.call(driver$write, driver$param)
  return(path)
}

writeSpatial = function(x, filename, overwrite = FALSE, ...)
{
  filename <- normalizePath(filename, winslash = "/", mustWork = FALSE)
  x <- sf::st_as_sf(x)

  if (isTRUE(overwrite))
    append = FALSE
  else
    append = NA

  sf::st_write(x, filename, append = append, quiet = TRUE)
}

#' An S4 class to represent the header of .las or .laz files
#'
#' An S4 class to represent the header of .las or .laz files according to the
#' \href{https://community.asprs.org/leadership-restricted/leadership-content/public-documents/standards}{LAS file format specifications}.
#' A `LASheader` object contains a `list` in the slot `@PHB` with
#' the data read from the Public Header Block, a `list` in the slot `@VLR` with
#' the data read from the Variable Length Records and a `list` in the slot `EVLR` with the data read
#' from the Extended Variable Lenght Records.
#'
#' @slot PHB list. Represents the Public Header Block
#' @slot VLR list. Represents the Variable Length Records
#' @slot EVLR list. Represents the Extended Variable Length Records
#'
#' @exportClass LASheader
#' @md
setClass(
  Class = "LASheader",
  representation(PHB = "list", VLR = "list", EVLR = "list")
)

setMethod("initialize", "LASheader", function(.Object, data = list())
{
  if (is.data.frame(data))
  {
    data <- rlas::header_create(data)
    if (data[["X scale factor"]] == 1e-8)
      warning("A scale factor of 1e-8 was infered from the data when creating a LASheader. Data is likely to be over precise", call. = FALSE)
  }

  stopifnot(is.list(data))

  vlr <- list()
  if (!is.null(data[["Variable Length Records"]]))
    vlr <- data[["Variable Length Records"]]

  evlr <- list()
  if (!is.null(data[["Extended Variable Length Records"]]))
    evlr <- data[["Extended Variable Length Records"]]

  .Object@PHB <- data
  .Object@PHB[["Variable Length Records"]] <- NULL
  .Object@PHB[["Extended Variable Length Records"]] <- NULL
  .Object@VLR <- vlr
  .Object@EVLR <- evlr

  return(.Object)
})


#' An S4 class to represent a .las or .laz file
#'
#' Class LAS is the representation of a las/laz file according to the
#' \href{https://community.asprs.org/leadership-restricted/leadership-content/public-documents/standards}{LAS file format specifications}.
#'
#' A `LAS` object contains a `data.table` with the data read from a `las/laz` file and
#' a \link[=LASheader-class]{LASheader} (see the ASPRS documentation for the LAS file format for more information).
#' Because las files are standardized the table of attributes read from the las/laz file is also standardized.
#' Columns are named:
#' \itemize{
#' \item{`X`, `Y`, `Z` (numeric)}
#' \item{`gpstime` (numeric)}
#' \item{`Intensity` (integer)}
#' \item{`ReturnNumber`, `NumberOfReturns` (integer)}
#' \item{`ScanDirectionFlag` (integer)}
#' \item{`EdgeOfFlightline` (integer)}
#' \item{`Classification` (integer)}
#' \item{`Synthetic_flag`,`Keypoint_flag`, `Withheld_flag`  (logical)}
#' \item{`ScanAngleRank`/`ScanAngle`  (integer/numeric)}
#' \item{`UserData` (integer)}
#' \item{`PointSourceID` (integer)}
#' \item{`R`,`G`,`B`, `NIR` (integer)}
#' }
#'
#' @slot crs Object of class \link[sf:st_crs]{crs} from sf.
#' @slot data Object of class \link[data.table:data.table]{data.table}. Point cloud data according to the LAS file format.
#' @slot header Object of class \link[=LASheader-class]{LASheader}. LAS file header according to the LAS file format.
#' @slot index list. See \link[=lidR-spatial-index]{spatial indexing}.
#'
#' @export
#' @md
#' @examples
#' # Read a las/laz file
#' LASfile <- system.file("extdata", "example.laz", package="rlas")
#' las <- readLAS(LASfile)
#' las
#'
#' # Creation of a LAS object out of external data
#' data <- data.frame(X = runif(100, 0, 100),
#'                    Y = runif(100, 0, 100),
#'                    Z = runif(100, 0, 20))
#'
#' # 'data' has many decimal digits
#' data
#'
#' # Create a default header and quantize *by reference*
#' # the coordinates to fit with offset and scale factors
#' cloud <- LAS(data)
#'
#' # 'data' has been updated and coordinates were quantized
#' data
#' cloud
#'
#' # Be careful when providing a header the function assumes that
#' # it corresponds to the data and won't quantize the coordinates
#' data <- data.frame(X = runif(100, 0, 100),
#'                    Y = runif(100, 0, 100),
#'                    Z = runif(100, 0, 20))
#' header <- header(las)
#'
#' # This works but triggers warnings and creates an invalid LAS object
#' cloud <- LAS(data, header)
#'
#' las_check(cloud)
#' @seealso
#' \link{readLAS}
setClass(
  Class = "LAS",
  representation(data = "data.table", header = "LASheader", crs = "crs", index = "list")
)

setMethod("initialize", "LAS", function(.Object)
{
  x      <- numeric(0)
  data   <- data.table::data.table(X = x, Y = x, Z = x)
  header <- suppressWarnings(rlas::header_create(data))
  header$`System Identifier` <- "lidR R package"
  header$`Generating Software` <- "lidR R package"
  header$`Min X` <- 0
  header$`Max X` <- 0
  header$`Min Y` <- 0
  header$`Max Y` <- 0
  header$`Min Z` <- 0
  header$`Max Z` <- 0
  header$`X offset` <- 0
  header$`Y offset` <- 0
  header$`Z offset` <- 0

  .Object@crs         <- sf::NA_crs_
  .Object@header      <- LASheader(header)
  .Object@data        <- data
  .Object@index       <- LIDRDEFAULTINDEX

  return(.Object)
})

#' @export
#' @rdname segment
segment_trees = function(las, algorithm, attribute = "treeID", uniqueness = 'incremental')
{
  UseMethod("segment_trees", las)
}

#' @export
segment_trees.LAS = function(las, algorithm, attribute = "treeID", uniqueness = 'incremental')
{
  stopif_forbidden_name(attribute)
  assert_is_algorithm(algorithm)
  assert_is_algorithm_its(algorithm)
  match.arg(uniqueness, c('incremental', 'gpstime', 'bitmerge'))
  lidR.context <- "segment_trees"

  if (uniqueness == 'gpstime' && !"gpstime" %in% names(las))
    stop("Impossible to compute unique IDs using gpstime: no gpstime found.", call. = FALSE)

  if (uniqueness == 'gpstime' &&  fast_countequal(las[["gpstime"]], 0L) == npoints(las))
    stop("Impossible to compute unique IDs using gpstime: gpstime is not populated.", call. = FALSE)

  if (is(algorithm, "RasterBased"))
    output <- algorithm(st_bbox(las))
  else if (is(algorithm, "PointCloudBased"))
    output <- algorithm(las)
  else
    stop("Invalid algorithm provided in segment_trees. The algorithm must have a class 'RasterBased' or 'PointCloudBased'")

  if (is_raster(output))
    las <- merge_spatial(las, output, attribute)
  else if (is.integer(output))
    las <- add_attribute(las, output, attribute)
  else
    stop(glue::glue("Wrong output type for the algorithm used. Expected 'RasterLayer' or 'integer', received {class(output)}"))

  if (all(is.na(las@data[[attribute]])))
    message("No tree found. Maybe use different parameters.")

  # If uniqueness is incremental we are done
  if (uniqueness == 'incremental')
  {
    type <- if (is.integer(las[[attribute]])) "int" else "double"
    NA_value <- if (is.integer(las[[attribute]])) .Machine$integer.max else .Machine$double.xmin
    las <- add_lasattribute_manual(las, name = attribute, desc = "An ID for each segmented tree", type = type, NA_value = NA_value)
    return(las)
  }

  # Otherwise we must compute an ID that is guaranteed to be unique

  tapex <- function(z,t) {
    zmax <- max(z)
    j <- which(z == zmax) # allows to retrieve potentially multiple points that are all max
    if (length(j) > 1) j <- j[which.min(t[j])] # arbitrarily takes the lowest gpstime
    return(list(t.pos.t = t[j]))
  }

  xyapex <- function(x,y,z) {
    zmax <- max(z)
    j <- which(z == zmax) # allows to retrieve potentially multiple points that are all max
    if (length(j) > 1) j <- j[which.min(x[j])] # arbitrarily takes the lowest x coordinate
    return(list(x.pos.t = x[j], y.pos.t = y[j]))
  }

  X <- Y <- Z <- gpstime <- NULL
  ids <- las@data[[attribute]]

  if (uniqueness == 'gpstime')
  {
    identifyers <- las@data[, if (!is.na(.BY)) tapex(Z, gpstime), by = ids]

    ids <- data.frame(ids)
    data.table::setDT(ids)
    matching <- identifyers[ids, on = 'ids']

    las@data[[attribute]] <- matching[["t.pos.t"]]

    las <- add_lasattribute_manual(las, name = attribute, desc = "An ID for each segmented tree", type = "double", NA_value = .Machine$double.xmin)
    return(las)
  }

  if (uniqueness == 'bitmerge')
  {
    identifyers <- las@data[, if (!is.na(.BY)) xyapex(X, Y, Z), by = ids]

    xoffset <- las[["X offset"]]
    yoffset <- las[["Y offset"]]
    zoffset <- las[["Z offset"]]

    xscale  <- las[["X scale factor"]]
    yscale  <- las[["Y scale factor"]]
    zscale  <- las[["Z scale factor"]]

    xscaled <- as.integer((identifyers[["x.pos.t"]] - xoffset)/xscale)
    yscaled <- as.integer((identifyers[["y.pos.t"]] - yoffset)/yscale)

    identifyers[["IDs"]] <- bitmerge(xscaled, yscaled)

    identifyers <- identifyers[, c(1,4)]
    ids <- data.frame(ids)
    data.table::setDT(ids)
    matching <- identifyers[ids, on = 'ids']

    las@data[[attribute]] <- matching[["IDs"]]

    las <- add_lasattribute_manual(las, name = attribute, desc = "An ID for each segmented tree", type = "double", NA_value = .Machine$double.xmin)
    return(las)
  }
}

#' @export
segment_trees.LAScatalog = function(las, algorithm, attribute = "treeID", uniqueness = 'incremental')
{
  # Defensive programming
  assert_is_algorithm(algorithm)
  assert_is_algorithm_its(algorithm)

  # Enforce some options
  opt_select(las) <- "*"

  # Processing
  options <- list(need_buffer = TRUE, drop_null = TRUE, need_output_file = TRUE)
  output  <- catalog_map(las, segment_trees, algorithm = algorithm, attribute = attribute, uniqueness = uniqueness, .options = options)
  return(output)
}

point_in_polygons <- function(las, sfc, by_poly = FALSE)
{
  if (!all(sf::st_geometry_type(sfc) %in% c("POLYGON", "MULTIPOLYGON")))
    stop("Only spatial polygons are supported", call. = FALSE)

  bbox <- sf::st_as_sfc(st_bbox(las))
  id  <- unlist(sf::st_intersects(bbox, sfc))
  id  <- as.integer(id)
  sfc2 <- sfc[id]

  rings = extract_polygons(sfc2)

  res <- C_in_polygon(las, rings, by_poly)
  return(res)
}

extract_polygons = function(sfc)
{
  # Fix #763. Complex code to avoid using boost
  coordinates = lapply(sfc, sf::st_coordinates)
  rings = lapply(coordinates, extract_rings)
  return(rings)
}

extract_rings = function(coordinates)
{
  if (ncol(coordinates) == 5L)
    features = lapply( split( coordinates[,1:5], coordinates[,5] ), matrix, ncol = 5)
  else
    features = list(coordinates)

  polygons = lapply(features,  function (x) lapply( split( x[,1:3], x[,4] ), matrix, ncol = 3))
  rings = lapply(polygons,  function (x) lapply(x, function(x) lapply( split( x[,1:3], x[,3] ), matrix, ncol = 3)))
  rings = lapply(rings, function(x) Reduce(c, x))

  return(rings[[1]])
}

#' LAS utilities
#'
#' Tools to manipulate LAS objects maintaining compliance with
#' \href{https://community.asprs.org/leadership-restricted/leadership-content/public-documents/standards}{ASPRS specification}
#'
#' In the specification of the LAS format the coordinates are expected to be given
#' with a certain precision e.g. 0.01 for a millimeter precision (or millifeet), meaning
#' that a file records e.g. 123.46 not 123.45678. Also, coordinates are stored as
#' integers. This is made possible with a scale and offset factor. For example,
#' 123.46 with an offset of 100 and a scale factor of 0.01 is actually stored as
#' (123.46 - 100)/0.01 = 2346. Storing 123.45678 with a scale factor of 0.01 and an offset
#' of 100 is invalid because it does not convert to an integer: (123.45678-100)/0.01
#' = 2345.678. Having an invalid LAS object may be critical in some lidR applications.
#' When writing into a LAS file, users will loose the extra precision without
#' warning and some algorithms in lidR use the integer conversion to make integer-based
#' computation and thus speed-up some algorithms and use less memory. Creation of an
#' invalid LAS object may cause problems and incorrect outputs.
#'
#' @param las An object of class LAS
#' @param x numeric. Coordinates vector
#' @param scale,offset scalar. scale and offset
#' @param xscale,yscale,zscale scalar. Can be missing if not relevant.
#' @param xoffset,yoffset,zoffset scalar. Can be missing if not relevant.
#' @param by_reference bool. Update the data in place without allocating new memory.
#' @param ... Unused.
#'
#' @examples
#' LASfile <- system.file("extdata", "example.laz", package="rlas")
#' las = readLAS(LASfile)
#'
#' # Manual modification of the coordinates (e.g. rotation, re-alignment, ...)
#' las@data$X <- las@data$X + 2/3
#' las@data$Y <- las@data$Y - 5/3
#'
#' # The point cloud is no longer valid
#' las_check(las)
#'
#' # It is important to fix that
#' las_quantize(las)
#'
#' # Now the file is almost valid
#' las_check(las)
#'
#' # Update the object to set up-to-date header data
#' las <- las_update(las)
#' las_check(las)
#'
#' # In practice the above code is not useful for regular users because the operators
#' # $<- already perform such operations on-the-fly. Thus the following
#' # syntax must be preferred and returns valid objects. Previous tools
#' # were only intended to be used in very specific cases.
#' las$X <- las$X + 2/3
#' las$Y <- las$Y - 5/3
#'
#' # Rescale and reoffset recompute the coordinates with
#' # new scales and offsets according to LAS specification
#' las <- las_rescale(las, xscale = 0.01, yscale = 0.01)
#' las <- las_reoffset(las, xoffset = 300000, yoffset = 5248000)
#' @rdname las_utilities
#' @name las_utilities
#' @family las utilities
NULL


#' @export
#' @rdname las_utilities
las_rescale = function(las, xscale, yscale, zscale)
{
  xoffset <- las[["X offset"]]
  yoffset <- las[["Y offset"]]
  zoffset <- las[["Z offset"]]

  if (!missing(xscale))
  {
    assert_is_a_number(xscale)
    newX <- round((las@data[["X"]] - xoffset)/xscale) * xscale + xoffset
    diff <- round(mean(abs(las@data[["X"]] - newX)), 4)
    las@data[["X"]] <- newX
    las@header@PHB[["X scale factor"]] <- xscale
    message(glue::glue("X coordinates were moved by {diff} on average"))
  }

  if (!missing(yscale))
  {
    assert_is_a_number(yscale)
    newY <- round((las@data[["Y"]] - yoffset)/yscale) * yscale + yoffset
    diff <- round(mean(abs(las@data[["Y"]] - newY)), 4)
    las@data[["Y"]] <- newY
    las@header@PHB[["Y scale factor"]] <- yscale
    message(glue::glue("Y coordinates were moved by {diff} on average"))
  }

  if (!missing(zscale))
  {
    assert_is_a_number(zscale)
    newZ <- round((las@data[["Z"]] - zoffset)/zscale) * zscale + zoffset
    diff <- round(mean(abs(las@data[["Z"]] - newZ)), 4)
    las@data[["Z"]] <- newZ
    las@header@PHB[["Z scale factor"]] <- zscale
    message(glue::glue("Z coordinates were moved by {diff} on average"))
  }

  las <- lasupdateheader(las)
  return(las)
}

#' @rdname las_utilities
#' @export
las_reoffset = function(las, xoffset, yoffset, zoffset)
{
  xscale  <- las[["X scale factor"]]
  yscale  <- las[["Y scale factor"]]
  zscale  <- las[["Z scale factor"]]

  xrange  <- c(las[["Min X"]], las[["Max X"]])
  yrange  <- c(las[["Min Y"]], las[["Max Y"]])
  zrange  <- c(las[["Min Z"]], las[["Max Z"]])

  if (!missing(xoffset))
  {
    assert_is_a_number(xoffset)

    newX <- suppressWarnings(as.integer(round((xrange - xoffset)/xscale)) * xscale + xoffset)
    if (anyNA(newX)) stop("Incorrect xoffset: integer overflow.", call. = FALSE)

    newX <- round((las@data[["X"]] - xoffset)/xscale) * xscale + xoffset
    diff <- round(mean(abs(las@data[["X"]] - newX)), 4)
    las@data[["X"]] <- newX
    las@header@PHB[["X offset"]] <- xoffset
    message(glue::glue("X coordinates were moved by {diff} on average"))
  }

  if (!missing(yoffset))
  {
    assert_is_a_number(yoffset)

    newY <- suppressWarnings(as.integer(round((yrange - yoffset)/yscale)) * yscale + yoffset)
    if (anyNA(newY)) stop("Incorrect yoffset: integer overflow.", call. = FALSE)

    newY <- round((las@data[["Y"]] - yoffset)/yscale) * yscale + yoffset
    diff <- round(mean(abs(las@data[["Y"]] - newY)), 4)
    las@data[["Y"]] <- newY
    las@header@PHB[["Y offset"]] <- yoffset
    message(glue::glue("Y coordinates were moved by {diff} on average"))
  }

  if (!missing(zoffset))
  {
    assert_is_a_number(zoffset)

    newZ <- suppressWarnings(as.integer(round((zrange - zoffset)/zscale)) * zscale + zoffset)
    if (anyNA(newZ)) stop("Incorrect zoffset: integer overflow.", call. = FALSE)

    newZ <- round((las@data[["Z"]] - zoffset)/zscale) * zscale + zoffset
    diff <- round(mean(abs(las@data[["Z"]] - newZ)), 4)
    las@data[["Z"]] <- newZ
    las@header@PHB[["Z offset"]] <- zoffset
    message(glue::glue("Z coordinates were moved by {diff} on average"))
  }

  las <- lasupdateheader(las)
  return(las)
}

#' @export
#' @rdname las_utilities
las_quantize = function(las, by_reference = TRUE)
{
  xscale <- las[["X scale factor"]]
  yscale <- las[["Y scale factor"]]
  zscale <- las[["Z scale factor"]]
  xoffset <- las[["X offset"]]
  yoffset <- las[["Y offset"]]
  zoffset <- las[["Z offset"]]

  if (isTRUE(by_reference))
  {
    quantize(las$X, xscale, xoffset)
    quantize(las$Y, yscale, yoffset)
    quantize(las$Z, zscale, zoffset)
    return(invisible(las))
  }
  else
  {
    las@data[["X"]] <- quantize(las$X, xscale, xoffset, FALSE)
    las@data[["Y"]] <- quantize(las$Y, yscale, yoffset, FALSE)
    las@data[["Z"]] <- quantize(las$Z, zscale, zoffset, FALSE)
    return(las)
  }
}

#' @export
#' @rdname las_utilities
las_update = function(las)
{
  stopifnotlas(las)

  header     <- as.list(las@header)
  new_header <- rlas::header_update(header, las@data)
  new_header <- LASheader(new_header)
  las@header <- new_header
  return(las)
}

#' @export
#' @rdname las_utilities
quantize = function(x, scale, offset, by_reference = TRUE, ...)
{
  umin <- min(x)
  umax <- max(x)
  urange <- storable_coordinate_range(scale, offset)

  if (umax > urange[2] | umin < urange[1])
    stop("'x' contains unquantizable values out of the storable range.", call. = FALSE)

  if (isTRUE(by_reference))
  {
    fast_quantization(x, scale, offset)
    return(invisible(x))
  }
  else
  {
    y <- data.table::copy(x)
    fast_quantization(y, scale, offset)
    return(y)
  }
}

#' @export
#' @rdname las_utilities
is.quantized = function(x, scale, offset, ...)
{
  p <- list(...)
  if (!is.null(p$sample))
  {
    n <- min(100L, length(x))
    s <- as.integer(seq(1L, length(x), length.out = n))
    x <- x[s]
  }

  return(fast_countunquantized(x, scale, offset) == 0L)
}

#' @export
#' @rdname las_utilities
count_not_quantized = fast_countunquantized


#' @export
#' @rdname las_utilities
storable_coordinate_range <- function(scale, offset) {

  assert_is_a_number(scale)
  assert_is_a_number(offset)

  storable_min <- -2147483647 * scale + offset
  storable_max <-  2147483647 * scale + offset

  return(c("min" = storable_min, "max" = storable_max))
}

#' @export
#' @rdname las_utilities
header <- function(las)
{
  return(las@header)
}

#' @export
#' @rdname las_utilities
payload <- function(las)
{
  return(las@data)
}
#' @export
#' @rdname las_utilities
phb <- function(las)
{
  if (!is(las, "LASheader"))
    las <- header(las)

  return(las@PHB)
}

#' @export
#' @rdname las_utilities
vlr <- function(las)
{
  if (!is(las, "LASheader"))
    las <- header(las)

  return(las@VLR)
}

#' @export
#' @rdname las_utilities
evlr <- function(las)
{
  if (!is(las, "LASheader"))
    las <- header(las)

  if (!methods::.hasSlot(las, "EVLR"))
    return(NULL)

  return(las@EVLR)
}

lasupdateheader = las_update

test_that("$ and [[ works", {
  expect_equal(megaplot$`File Signature`, "LASF")
  expect_equal(megaplot$`X scale factor`, 0.01)
})

test_that("$<- and [[<-] works", {
  megaplot$`File Signature` <- "ABC"
  megaplot$`X scale factor` <- 0.1
  expect_equal(megaplot[["File Signature"]], "ABC")
  expect_equal(megaplot[["X scale factor"]], 0.1)
})

test_that("vlr works", {
  VLR <- vlr(megaplot)
  EVLR <- evlr(megaplot)
  expect_is(VLR, "list")
  expect_is(EVLR, "list")

  VLR <- vlr(header(megaplot))
  EVLR <- evlr(header(megaplot))
  expect_is(VLR, "list")
  expect_is(EVLR, "list")
})

context("las_merge")

las1 = random_10_points
las2 = random_10_points
las2 = las_rescale(las2, 0.1)
las3 = random_10_points
las4 = random_10_points
epsg(las4) <- 32619

test_that("rbind merge 2 las", {

  o <- rbind(las1, las3)

  expect_equal(npoints(o), 20)
  expect_equal(o@header@PHB[["X scale factor"]], 0.001)
  expect_equal(o@header@PHB[["Y scale factor"]], 0.001)
})

test_that("rbind warns about incompatibilities", {

  expect_warning(rbind(las1, las2), "different scales")

  o <- suppressWarnings(rbind(las1, las2))

  expect_equal(npoints(o), 20)
  expect_equal(o@header@PHB[["X scale factor"]], 0.001)
  expect_equal(o@header@PHB[["Y scale factor"]], 0.001)

  expect_warning(rbind(las2, las1), "different scales")

  o <- suppressWarnings(rbind(las2, las1))

  expect_equal(npoints(o), 20)
  expect_equal(o@header@PHB[["X scale factor"]], 0.1)
  expect_equal(o@header@PHB[["Y scale factor"]], 0.001)
})


test_that("rbind fails with different CRS", {

  expect_error(rbind(las1, las4), "Different CRS")
})

test_that("rbind works with LAScatalog", {

  expect_error( rbind(megaplot_ctg, mixedconifer_ctg), "Different CRS")

  st_crs(megaplot_ctg) <- sf::NA_crs_
  st_crs(topography_ctg) <- sf::NA_crs_

  expect_warning(rbind(megaplot_ctg, topography_ctg))
})

context("las_rescale")

las <- example

test_that("las_rescale works", {
  las2 = suppressMessages(las_rescale(las, xscale = 1, yscale = 1, zscale = 1))
  expect_equal(las2@data[,1:2], round(las@data[,1:2]))
})

test_that("lasroffset works", {
  las2 = suppressMessages(las_reoffset(las, xoffset = 100000, yoffset = 6000000, zoffset = 2))
  expect_equal(las2@data, las@data)
})

context("quantization")

x = runif(100, 1000, 2000)
y = data.table::copy(x)
scale = 0.001
offset = 500

las = megaplot

test_that("is.quantized works", {
  expect_true(!is.quantized(x, scale, offset))
})

test_that("Quantization works", {

  quantize(x, scale, offset)

  expect_equal(x, round(y, 3), tolerance = 1e-10)
  expect_equal(count_not_quantized(x, scale, offset), 0L)
  expect_equal(count_not_quantized(y, scale, offset), 100L)
})

test_that("Quantization works on las data", {

  xscale = las@header@PHB$`X scale factor`
  xoffset = las@header@PHB$`X offset`
  yscale = las@header@PHB$`Y scale factor`
  yoffset = las@header@PHB$`Y offset`
  zscale = las@header@PHB$`Z scale factor`
  zoffset = las@header@PHB$`Z offset`

  expect_equal(count_not_quantized(las$X, xscale, xoffset), 0L)
  expect_equal(count_not_quantized(las$Y, yscale, yoffset), 0L)
  expect_equal(count_not_quantized(las$Z, zscale, zoffset), 0L)

  x = las@data[["X"]][1]
  las@data[["X"]][1] = x + 0.0001

  expect_equal(count_not_quantized(las$X, xscale, xoffset), 1L)

  las_quantize(las)

  expect_equal(las@data[["X"]][1], x,  tolerance = 0)
})

test_that("Quantization works with out of 32 bits integer range values", {

  expect_error(quantize(1234567890.1, 0.01, 0), "'x' contains unquantizable values out of the storable range.")
  expect_equal(count_not_quantized(1234567890.1, 0.01, 0), 1L)
})

context("LAS")

las  <- megaplot
data <- data.frame(X = runif(10), Y = runif(10), Z = runif(10))
outs <- data.frame(x = runif(10), Y = runif(10), z = runif(10), pointsourceid = 1:10)

test_that("Print a LAS object works", {

  sink(tempfile())
  expect_error(print(las), NA)
  expect_error(summary(las), NA)
  sink(NULL)
})

test_that("LAS builds a LAS 1.2 prf 0 object from basic data", {

  las2 <- LAS(data)

  expect_is(las2, "LAS")
  expect_equal(las2@header@PHB$`Version Minor`, 2L)
  expect_equal(las2@header@PHB$`Point Data Format ID`, 0L)
  expect_equal(las2@header@PHB$`Min X`, min(data$X))
  expect_equal(las2@header@PHB$`Min Y`, min(data$Y))
  expect_equal(las2@header@PHB$`Min Z`, min(data$Z))
  expect_equal(las2@header@PHB$`Max X`, max(data$X))
  expect_equal(las2@header@PHB$`Max Y`, max(data$Y))
  expect_equal(las2@header@PHB$`Max Z`, max(data$Z))
})

test_that("LAS builds a LAS 1.2 prf 1 object from time data", {

  data$gpstime   <- runif(10, 0, 100)
  data$Intensity <- as.integer(runif(10, 0, 100))
  las2 <- LAS(data)

  expect_is(las2, "LAS")
  expect_equal(las2@header@PHB$`Version Minor`, 2L)
  expect_equal(las2@header@PHB$`Point Data Format ID`, 1L)
  expect_equal(las2@header@PHB$`Min X`, min(data$X))
  expect_equal(las2@header@PHB$`Min Y`, min(data$Y))
  expect_equal(las2@header@PHB$`Min Z`, min(data$Z))
  expect_equal(las2@header@PHB$`Max X`, max(data$X))
  expect_equal(las2@header@PHB$`Max Y`, max(data$Y))
  expect_equal(las2@header@PHB$`Max Z`, max(data$Z))
})

test_that("LAS builds a LAS 1.2 prf 3 object from RGB data", {

  data$gpstime   <- runif(10, 0, 100)
  data$Intensity <- as.integer(runif(10, 0, 100))
  data$R         <- as.integer(runif(10, 0, 2^16))
  data$G         <- as.integer(runif(10, 0, 2^16))
  data$B         <- as.integer(runif(10, 0, 2^16))
  las2 <- LAS(data)

  expect_is(las2, "LAS")
  expect_equal(las2@header@PHB$`Version Minor`, 2L)
  expect_equal(las2@header@PHB$`Point Data Format ID`, 3L)
})

test_that("LAS builds a LAS object respecting a header", {

  las2 <- LAS(data, las@header, check = FALSE)

  expect_is(las2, "LAS")
  expect_equal(las2@header@PHB$`Version Minor`, las@header@PHB$`Version Minor`)
  expect_equal(las2@header@PHB$`Point Data Format ID`, las@header@PHB$`Point Data Format ID`)
  expect_equal(las2@header@PHB$`Min X`, min(data$X))
  expect_equal(las2@header@PHB$`Min Y`, min(data$Y))
  expect_equal(las2@header@PHB$`Min Z`, min(data$Z))
  expect_equal(las2@header@PHB$`Max X`, max(data$X))
  expect_equal(las2@header@PHB$`Max Y`, max(data$Y))
  expect_equal(las2@header@PHB$`Max Z`, max(data$Z))
})

test_that("LAS builds a LAS object with a CRS", {

  las2 <- LAS(data)

  expect_true(is.na(las2@crs))
  expect_equal(epsg(las2), 0)

  las2 <- LAS(data, las@header, check = FALSE)

  expect_equal(st_crs(las2), st_crs(las))
  expect_equal(epsg(las2), 26917)

  las2 <- LAS(data, crs = las@crs)

  expect_equal(st_crs(las2), st_crs(las))
  #expect_equal(epsg(las2), 26917)
})

test_that("LAS throws a warning/error if building an invalid LAS", {

  # Invalid point format

  data$R <- as.integer(runif(10, 0, 2^16))
  data$G <- as.integer(runif(10, 0, 2^16))
  data$B <- as.integer(runif(10, 0, 2^16))

  expect_warning(LAS(data, las@header), "point data format")

  # Invalid scale factor

  header <- las@header
  header@PHB$`Point Data Format ID` <- 3L
  header@PHB$`X scale factor` <- 0.412

  expect_warning(LAS(data, header), "X scale factor")

  # Invalid RGB

  data$R <- as.integer(runif(10, 0, 2^8))
  data$G <- as.integer(runif(10, 0, 2^8))
  data$B <- as.integer(runif(10, 0, 2^8))

  header <- las@header
  header@PHB$`Point Data Format ID` <- 3L

  expect_warning(LAS(data, header), "8 bits")

  # Invalid coordinates

  data$X <- as.integer(runif(10, 1, 100))
  expect_error(LAS(data), "double")

  data$X <- NULL
  expect_error(LAS(data), "coordinates")

  # Invalid ReturnNumber

  data$X <- runif(10)
  data$R <- NULL
  data$G <- NULL
  data$B <- NULL
  data$ReturnNumber <- as.integer(rep(1, 10L))
  data$ReturnNumber[5] <- 0L

  expect_warning(LAS(data), "return number")

  # Invalid NumberOfReturns

  set.seed(0)
  data$ReturnNumber <- as.integer(sample(1:3, 10, replace = TRUE))
  set.seed(1)
  data$NumberOfReturns <- as.integer(sample(1:3, 10, replace = TRUE))

  expect_warning(LAS(data), "number of returns")

  # Invalid Version

  header <- las@header
  header@PHB$`Version Minor` <- NULL
  data <- data.frame(X = runif(10), Y = runif(10), Z = runif(10))

  expect_error(LAS(data, header), "Version")

  # Non quantized coordinates

  header@PHB$`Version Minor` <- 1L
  expect_warning(LAS(data, header), "quantization errors")
})

test_that("LAS redefined behavior of $, [, and [[", {

  las <- random_10_points

  expect_true(is.numeric(las$Z))
  expect_equal(length(las$Z), 10)

  expect_true(is.null(las$U))

  expect_error(las$Z <- 1:10, "replace data of type double")
  expect_error(las$U <- 1:10, "Addition of a new column")

  expect_true(is.numeric(las[["Z"]]))
  expect_equal(length(las[["Z"]]), 10)

  expect_true(is.null(las[["U"]]))

  expect_error(las[["Z"]] <- 1:10, "replace data of type double")
  expect_error(las[["U"]] <- 1:10, "Addition of a new column")
})

test_that("LAS operator $ quantize on the fly and update header", {

  las <- random_10_points

  x = runif(10)
  y = runif(10)
  z = runif(10)

  las$X <- x
  las$Y <- y
  las$Z <- z

  expect_equal(las$X, quantize(x, 0.001, 0, FALSE))
  expect_equal(las$Y, quantize(y, 0.001, 0, FALSE))
  expect_equal(las$Z, quantize(z, 0.001, 0, FALSE))

  xbbox = range(quantize(x, 0.001, 0, FALSE))
  ybbox = range(quantize(y, 0.001, 0, FALSE))

  expect_equal(las@header@PHB[["Min X"]], xbbox[1])
  expect_equal(las@header@PHB[["Min Y"]], ybbox[1])
})

test_that("LAS operator [[ quantize on the fly and update header", {

  las <- random_10_points

  x = runif(10)
  y = runif(10)
  z = runif(10)

  las[["X"]] <- x
  las[["Y"]] <- y
  las[["Z"]] <- z

  expect_equal(las$X, quantize(x, 0.001, 0, FALSE))
  expect_equal(las$Y, quantize(y, 0.001, 0, FALSE))
  expect_equal(las$Z, quantize(z, 0.001, 0, FALSE))

  xbbox = range(quantize(x, 0.001, 0, FALSE))
  ybbox = range(quantize(y, 0.001, 0, FALSE))

  expect_equal(las@header@PHB[["Min X"]], xbbox[1])
  expect_equal(las@header@PHB[["Min Y"]], ybbox[1])
})

test_that("LAS operator[[ and $ throw error for not storable coordinates", {
  las <- random_10_points
  x <- round(runif(10),2)
  x[5] <- 21474836.47

  expect_error(las$X <- x, "Trying to store values ranging in")
  expect_error(las$Y <- x, "Trying to store values ranging in")
  expect_error(las[["X"]] <- x, "Trying to store values ranging in")
  expect_error(las[["Y"]] <- x, "Trying to store values ranging in")
})


test_that("LAS conversion to SpatialPointsDataFrame works", {
  skip_if_not_installed("sp")
  las <- random_10_points
  splas <- as.spatial(las)

  expect_true(is(splas, "SpatialPointsDataFrame"))
})

test_that("LAS build an empty point cloud with no header (#314)", {
  las = LAS(data.frame(X = numeric(0), Y = numeric(0), Z = numeric(0)))
  expect_equal(npoints(las), 0L)
  expect_equal(names(las), c("X", "Y", "Z"))
})

test_that("LAS rename the attribute", {
  x <- outs[["x"]]
  las = LAS(outs)
  expect_equal(names(las), c("X", "Y", "Z", "PointSourceID"))
  expect_reference(las@data[["X"]], x)
})

#' @rdname tools
#' @export
print.LAS <- function(x, ...)
{
  show(x)
  return(invisible(x))
}

#' @rdname tools
#' @export
print.LAScatalog <- function(x, ...)
{
  show(x)
  return(invisible(x))
}

#' @export
#' @rdname tools
print.lidRAlgorithm = function(x, ...)
{
  e <- environment(x)
  params <- ls(e)
  params <- Filter(function(i) !is.algorithm(get(i,e)), params)
  omp <- if (is.parallelised(x)) "yes" else "no"

  if (is(x, LIDRDSM))
  { use = "digital surface model" ; with = LIDRCONTEXTDSM }
  else if (is (x, LIDRDEC))
  { use = "point cloud thinning" ; with = LIDRCONTEXTDEC }
  else if (is (x, LIDRGND))
  { use = "ground classification" ; with = LIDRCONTEXTGND }
  else if (is (x, LIDRITD))
  { use = "individual tree detection" ; with = LIDRCONTEXTITD }
  else if (is (x, LIDRITS))
  { use = "individual tree segmentation" ; with = LIDRCONTEXTITS }
  else if (is (x, LIDRNIT))
  { use = "intensity normalisation" ; with = LIDRCONTEXTNIT }
  else if (is (x, LIDRSNG))
  { use = "snag segmentation" ; with = LIDRCONTEXTSNG }
  else if (is (x, LIDRTRK))
  { use = "sensor tracking" ; with = LIDRCONTEXTTRK }
  else if (is (x, LIDRSHP))
  { use = "shape segmentation" ; with = LIDRCONTEXTSHP }
  else if (is (x, LIDRSPI))
  { use = "spatial interpolation" ; with = LIDRCONTEXTSPI }
  else if (is (x, LIDROUT))
  { use = "noise classification" ; with = LIDRCONTEXTOUT }
  else
  { use = "unknown" ; with = "unknown" } # nocov

  with = paste(with, collapse = " or ")

  cat("Object of class lidR algorithm\n")
  cat("Algorithm for:", use, "\n")
  cat("Designed to be used with:", with, "\n")
  cat("Native C++ parallelization:", omp, "\n")
  cat("Parameters: ")

  if (length(params) == 0L)
  {
    cat("none\n")
    return(invisible(x))
  }
  else
    cat("\n")

  for (param in params)
  {
    v = get(param, e)
    if (is.numeric(v) | is.complex(v) | is.logical(v) | is.character(v))
      cat(" - ", param, " = " , v, " <", class(v), ">\n", sep = "")
    else if (is.null(v))
      cat(" - ", param, " = NULL <NULL>\n", sep = "")
    else
      cat(" - ", param, " <", class(v), ">\n", sep = "")
  }

  return(invisible(x))
}

#' @export
#' @rdname tools
print.raster_template <- function(x, ...)
{
  cat("Object of class raster_template\n")
  cat("extent     : ", x$xmin, ", ", x$xmax, ", ",  x$ymin, ", ", x$ymax, " (xmin, xmax, ymin, ymax)\n", sep = "")
  cat("resolution :", x$xres, x$yres, "\n")
  cat("size       :", x$ncol, x$nrow, "\n")
  cat("crs        :", x$crs$Name, "\n")
}

#' @rdname tools
#' @param object A \code{LAS*} object or other lidR related objects.
#' @export
summary.LAS <- function(object, ...)
{
  print(object)
  print(object@header)
  return(invisible(object))
}

#' @rdname tools
#' @export
summary.LAScatalog <- function(object, ...)
{
  inmemory <- if (opt_output_files(object) == "") "in memory" else "on disk"
  w2w  <- if (opt_wall_to_wall(object)) "guaranteed" else "not guaranteed"
  merging <- if (opt_merge(object)) "enabled" else "disable"

  show(object)
  cat("proc. opt.  : buffer: ", opt_chunk_buffer(object), " | chunk: ", opt_chunk_size(object), "\n", sep = "")
  cat("input opt.  : select: ", opt_select(object), " | filter: ", opt_filter(object), "\n", sep = "")
  cat("output opt. : ", inmemory, " | w2w ", w2w, " | merging ", merging, "\n", sep = "")
  cat("drivers     :\n")

  drivers <- object@output_options$drivers
  dnames <- names(drivers)
  for (i in 1:length(drivers))
  {
    driver <- drivers[[i]]
    params <- driver$param
    pnames <- names(params)
    cat(" -", dnames[i], ": ")
    if (length(params) > 0)
    {
      for (j in 1:length(params))
        cat(pnames[j], "=", params[[j]], " ")
    }
    else
    {
      cat("no parameter")
    }
    cat("\n")

  }
  return(invisible(object))
}

setMethod("show", "LAS", function(object)
{
  size      <- format(las_size(object), units = "auto")
  area      <- as.numeric(st_area(object))
  area.h    <- area
  npoints   <- nrow(object@data)
  npulses   <- object@header[["Number of points by return"]][1]
  npoints.h <- npoints
  dpts      <- if (area > 0) npoints/area else 0
  dpulse    <- if (area > 0) npulses/area else 0
  ext       <- st_bbox(object)
  phb       <- object@header@PHB
  major     <- phb[["Version Major"]]
  minor     <- phb[["Version Minor"]]
  version   <- paste(major, minor, sep = ".")
  format    <- phb[["Point Data Format ID"]]
  units     <- st_crs(object)$units
  units     <- if (is.null(units) || is.na(units)) "units" else units
  type      <- sensor(object)
  if (type == TLSLAS) type = "terrestrial"
  else if (type == ALSLAS) type = "airborne"
  else if (type == UKNLAS) type = "unknown"
  else type = "unknown"

  areaprefix  <- ""
  pointprefix <- ""

  if (area > 1000*1000/2)
  {
    areaprefix <- if (length(units) == 0) "thousand " else "k"
    area.h     <- round(area/(1000*1000), 2)
  }

  if (npoints > 1000 & npoints < 1000^2)
  {
    pointprefix <- "thousand"
    npoints.h   <- round(npoints/1000, 1)
  }
  else if (npoints >= 1000^2 & npoints < 1000^3)
  {
    pointprefix <- "million" #nocov
    npoints.h   <- round(npoints/(1000^2), 2) #nocov
  }
  else if (npoints >= 1000^3)
  {
    pointprefix <- "billion" #nocov
    npoints.h   <- round(npoints/(1000^3), 2) #nocov
  }

  cat("class        : ", class(object), " (v", version, " format ", format, ")\n", sep = "")
  cat("memory       :", size, "\n")
  cat("extent       : ", ext[1], ", ", ext[3], ", ", ext[2], ", ", ext[4], " (xmin, xmax, ymin, ymax)\n", sep = "")
  cat("coord. ref.  :", st_crs(object)$Name, "\n")
  cat("area         : ", area.h, " ", areaprefix, units, "\u00B2\n", sep = "")
  cat("points       : ", npoints.h, " ", pointprefix, " points\n", sep = "")
  cat("type         : ", type, "\n", sep = "")
  cat("density      : ", round(dpts, 2), " points/", units, "\u00B2\n", sep = "")
  if (dpulse > 0)
    cat("density      : ", round(dpulse, 2), " pulses/", units, "\u00B2\n", sep = "")
  #cat("names        :", attr, "\n")

  return(invisible(object))
})

setMethod("show", "LAScatalog", function(object)
{
  area        <- as.numeric(st_area(object))
  area.h      <- area
  npoints     <- sum(object@data$Number.of.point.records)
  npulse      <- sum(object@data$Number.of.1st.return)
  npoints.h   <- npoints
  ext         <- st_bbox(object)
  units       <- st_crs(object)$units
  units       <- if (is.na(units)) "units" else units
  areaprefix  <- ""
  pointprefix <- ""
  major       <- sort(unique(object[["Version.Major"]]))
  minor       <- sort(unique(object[["Version.Minor"]]))
  if (is.null(major)) major <- "?"
  if (is.null(minor)) minor <- "?"
  version     <- paste(major, minor, sep = ".", collapse = " and ")
  format      <- paste(sort(unique(object[["Point.Data.Format.ID"]])), collapse = " and ")
  if (format == "") format <- "?"
  density     <- round(npoints/area, 1)
  if (is.nan(density)) density <- 0
  dpulse      <- round(npulse/area, 1)
  type      <- sensor(object)
  if (type == TLSLAS) type = "terrestrial"
  else if (type == ALSLAS) type = "airborne"
  else if (type == UKNLAS) type = "unknown"
  else type = "unknown"

  if (area > 1000*1000/2)
  {
    areaprefix <- if (length(units) == 0) "thousand " else "k"
    area.h     <- round(area/(1000*1000), 2)
  }

  if (npoints <= 1000)
  {
    pointprefix <- ""
    npoints.h   <- npoints
  }
  else if (npoints > 1000 & npoints < 1000^2)
  {
    pointprefix <- "thousand"
    npoints.h   <- round(npoints/1000, 1)
  }
  else if (npoints >= 1000^2 & npoints < 1000^3)
  {
    pointprefix <- "million" #nocov
    npoints.h   <- round(npoints/(1000^2), 2) #nocov
  }
  else if (npoints >= 1000^3)
  {
    pointprefix <- "billion" #nocov
    npoints.h   <- round(npoints/(1000^3), 2) #nocov
  }

  cat("class       : ", class(object), " (v", version, " format ", format, ")\n", sep = "")
  cat("extent      : ", ext[1], ", ", ext[3], ", ", ext[2], ", ", ext[4], " (xmin, xmax, ymin, ymax)\n", sep = "")
  cat("coord. ref. :", st_crs(object)$Name, "\n")
  cat("area        : ", area.h, " ", areaprefix, units, "\u00B2\n", sep = "")
  cat("points      : ", npoints.h, " ", pointprefix, " points\n", sep = "")
  cat("type        : ", type, "\n", sep = "")
  cat("density     : ", density, " points/", units, "\u00B2\n", sep = "")
  if (dpulse > 0)
    cat("density     : ", round(dpulse, 2), " pulses/", units, "\u00B2\n", sep = "")
  cat("num. files  :", dim(object@data)[1], "\n")
  return(invisible(object))
})

setMethod("show", "LASheader",  function(object)
{
  x = object@PHB

  gpstype   = if (x[["Global Encoding"]][["GPS Time Type"]]) "Standard GPS Time" else "GPS Week Time"
  synthetic = if (x[["Global Encoding"]][["Synthetic Return Numbers"]]) "true" else "no"
  WKT       = if (x[["Global Encoding"]][["WKT"]]) "CRS is WKT" else "CRS is GeoTIFF"
  aggregate = if (x[["Global Encoding"]][["Aggregate Model"]]) "true" else "false"

  cat("File signature:          ", x$`File Signature`, "\n")
  cat("File source ID:          ", x$`File Source ID`, "\n")
  cat("Global encoding:\n")
  cat(" - GPS Time Type:" , gpstype, "\n")
  cat(" - Synthetic Return Numbers:" , synthetic, "\n")
  cat(" - Well Know Text:" , WKT, "\n")
  cat(" - Aggregate Model:" , aggregate, "\n")
  cat("Project ID - GUID:       ", x$`Project ID - GUID`, "\n")
  cat("Version:                  ", x$`Version Major`, ".", x$`Version Minor`, "\n", sep = "")
  cat("System identifier:       ", x$`System Identifier`, "\n")
  cat("Generating software:     ", x$`Generating Software`, "\n")
  cat("File creation d/y:        ", x$`File Creation Day of Year`, "/", x$`File Creation Year`, "\n", sep = "")
  cat("header size:             ", x$`Header Size`, "\n")
  cat("Offset to point data:    ", x$`Offset to point data`, "\n")
  cat("Num. var. length record: ", x$`Number of variable length records`, "\n")
  cat("Point data format:       ", x$`Point Data Format ID`, "\n")
  cat("Point data record length:", x$`Point Data Record Length`, "\n")
  cat("Num. of point records:   ", x$`Number of point records`, "\n")
  cat("Num. of points by return:", x$`Number of points by return`, "\n")
  cat("Scale factor X Y Z:      ", x$`X scale factor`, x$`Y scale factor`, x$`Z scale factor`, "\n")
  cat("Offset X Y Z:            ", x$`X offset`, x$`Y offset`, x$`Z offset`, "\n")
  cat("min X Y Z:               ", x$`Min X`, x$`Min Y`, x$`Min Z`, "\n")
  cat("max X Y Z:               ", x$`Max X`, x$`Max Y`, x$`Max Z`, "\n")

  nvlr  <- length(object@VLR)
  nevlr <- length(object@EVLR)

  if (nvlr == 0)
  {
    cat("Variable Length Records (VLR):  void\n")
  }
  else
  {
    cat("Variable Length Records (VLR):\n")

    for (i in 1:nvlr)
    {
      vlr = object@VLR[[i]]

      cat("   Variable Length Record", i, "of", nvlr, "\n")
      #cat("       Reserve:            ",  vlr$reserved, "\n")
      #cat("       User ID:             ", vlr$`user ID`, "\n")
      #cat("       record ID:           ", vlr$`record ID`, "\n")
      #cat("       Length after header: ", vlr$`length after header`, "\n")
      cat("       Description:", vlr$description, "\n")

      if (vlr$`record ID` == 34735)
      {
        cat("       Tags:\n")
        lapply(vlr[[6]], function(xx)
        {
          cat("          Key", xx$key, "value", xx$`value offset`, "\n")
        })
      }
      else if (vlr$`record ID` == 34736)
      {
        cat("       data:                ", vlr[[6]], "\n")
      }
      else if (vlr$`record ID` == 34737)
      {
        cat("       data:                ", vlr[[6]], "\n")
      }
      else if (vlr$`record ID` == 4)
      {
        cat("       Extra Bytes Description:\n")
        lapply(vlr$`Extra Bytes Description`, function(xx)
        {
          cat("          ", xx$name, ": ", xx$description, "\n", sep = "")
        })
      }
      else if (vlr$`record ID` == 2112)
      {
        cat("       WKT OGC COORDINATE SYSTEM: ", strtrim(vlr$`WKT OGC COORDINATE SYSTEM`, 70), " [...] (truncated)\n", sep = "")
      }
    }
  }

  if (nevlr == 0)
  {
    cat("Extended Variable Length Records (EVLR):  void\n")
  }
  else
  {
    cat("Extended Variable Length Records (EVLR):\n")

    for (i in 1:nevlr)
    {
      vlr = object@EVLR[[i]]

      cat("   Extended Variable length record", i, "of", nevlr, "\n")
      #cat("       Reserve:            ",  vlr$reserved, "\n")
      #cat("       User ID:             ", vlr$`user ID`, "\n")
      #cat("       record ID:           ", vlr$`record ID`, "\n")
      #cat("       Length after header: ", vlr$`length after header`, "\n")
      cat("       Description:", vlr$description, "\n")

      if (vlr$`record ID` == 34735)
      {
        cat("       Tags:\n")
        lapply(vlr[[6]], function(xx)
        {
          cat("          Key", xx$key, "value", xx$`value offset`, "\n")
        })
      }
      else if (vlr$`record ID` == 34736)
      {
        cat("       data:                ", vlr[[6]], "\n")
      }
      else if (vlr$`record ID` == 34737)
      {
        cat("       data:                ", vlr[[6]], "\n")
      }
      else if (vlr$`record ID` == 4)
      {
        cat("       Extra Bytes Description:\n")
        lapply(vlr$`Extra Bytes Description`, function(xx)
        {
          cat("          ", xx$name, ": ", xx$description, "\n", sep = "")
        })
      }
      else if (vlr$`record ID` == 2112)
      {
        cat("       WKT OGC COORDINATE SYSTEM: ", strtrim(vlr$`WKT OGC COORDINATE SYSTEM`, 70), " [...] (truncated)\n", sep = "")
      }
    }
  }

  return(invisible(object))
})

setMethod("show", "LAScluster", function(object)
{
  cat("class   : LAScluster\n")
  cat("name    :", object@name, "\n")
  cat("center  :", object@center$x, ",", object@center$y, "\n")
  cat("extent  :", object@bbox[1], ",", object@bbox[3], ",", object@bbox[2], ",", object@bbox[4], "(xmin, xmax, ymin, ymax)\n")
  cat("extent+ :", object@bbbox[1], ",", object@bbbox[3], ",", object@bbbox[2], ",", object@bbbox[4], "(xmin, xmax, ymin, ymax)\n")
  cat("size    :", object@width, "x", object@height, "\n")
  cat("files   :", basename(object@files), "\n")
  cat("filter  :", object@filter, "\n")
  return(invisible(object))
})

#' @export
#' @rdname normalize
normalize_intensity <- function(las, algorithm)
{
  UseMethod("normalize_intensity", las)
}

#' @export
normalize_intensity.LAS <- function(las, algorithm)
{
  assert_is_algorithm(algorithm)
  assert_is_algorithm_nit(algorithm)
  lidR.context <- "normalize_intensity"
  intensity <- algorithm(las)
  invalid   <- fast_countequal(intensity, 65535)

  if (invalid > 0)
    warning(glue::glue("{invalid} points have a normalized intensity greater than 65535. Intensity replaced by 65535"), call. = FALSE)

  las@data[["RawIntensity"]] <- las@data[["Intensity"]]
  las@data[["Intensity"]]    <- intensity

  return(las)
}

#' @export
normalize_intensity.LAScatalog = function(las, algorithm)
{
  opt_select(las) <- "*"
  opt_chunk_buffer(las) <- 0

  options <- list(need_buffer = FALSE, drop_null = TRUE, need_output_file = TRUE)
  output  <- catalog_map(las, normalize_intensity, algorithm = algorithm, .options = options)
  return(output)
}

context("normalize_intensity")

las <- topography
sensor <- track_sensor(las, Roussel2020(pmin = 15))

test_that("normalize_intensity works", {

  expect_error(las <- normalize_intensity(las, range_correction(sensor, Rs = 2000)), NA)
  expect_true("RawIntensity" %in% names(las))
  expect_true(is.integer(las$Intensity))
  expect_equal(mean(las$Intensity), 1183, tolerance = 0.5)
})

test_that("normalize_intensity fails with invalid LAS data", {

  las@data[["Intensity"]] <- NULL
  expect_error(normalize_intensity(las, range_correction(sensor, Rs = 2000)), "No 'Intensity' attribute found")

  las@data[["gpstime"]] <- NULL
  expect_error(normalize_intensity(las, range_correction(sensor, Rs = 2000)), "No 'gpstime' attribute found in las")
})

test_that("normalize_intensity fails with invalid sensor data", {

  sink(tempfile())

  p = sf::st_sfc(sf::st_point(c(0, -10000, 0)))
  geom = sf::st_geometry(sensor) + p
  sf::st_geometry(sensor) = geom
  sf::st_crs(sensor) <- st_crs(las)

  expect_error(normalize_intensity(las, range_correction(sensor, Rs = 2000)), "Unrealistic range")

  sensor[["gpstime"]] <- NULL
  expect_error(normalize_intensity(las, range_correction(sensor, Rs = 2000)), "No 'gpstime' attribute found in sensor")

  sensor[["gpstime"]] <- 1:8
  expect_error(normalize_intensity(las, range_correction(sensor, Rs = 2000)), "gpstime range from the sensor does not contain gpstime range from the point-cloud")

  sink(NULL)
})


test_that("normalize_intensity works with 2 coordinates points", {

  sensor2 <- sf::st_zm(sensor)
  sensor2$Z <- sf::st_coordinates(sensor)[,3]

  las1 <- normalize_intensity(las, range_correction(sensor2, Rs = 2000))
  las2 <- normalize_intensity(las, range_correction(sensor, Rs = 2000))

  expect_equal(las1, las2)

  expect_error(normalize_intensity(las, range_correction(sensor2, Rs = 2000, elevation = "K")), "No 'K' attribute found in sensor")
  expect_error(normalize_intensity(las, range_correction(sensor2, Rs = 2000, gpstime = "t")), "No 't' attribute found in sensor")
  expect_error(normalize_intensity(las, range_correction(sensor2, Rs = 2000, elevation = NULL)), "There are only two dimensions in the coordinates of the sensor")
})


test_that("normalize_intensity is able to detect unrealistic range corrections", {

  sink(tempfile())
  sensor$geometry[[5]][2] <- 8000
  expect_error(normalize_intensity(las, range_correction(sensor, Rs = 2000)), "Unrealistic range")
  sink(NULL)
})

test_that("normalize_intensity clamps values to 65535 with warning", {

  las$Intensity[1000] <- 60000L
  expect_warning(normalize_intensity(las, range_correction(sensor, Rs = 2000)), "1 points have a normalized intensity greater than 65535")

  las1 <- suppressWarnings(normalize_intensity(las, range_correction(sensor, Rs = 2000)))
  expect_equal(max(las1$Intensity), 65535)
})

#' Point Cloud Decimation Algorithm
#'
#' This function is made to be used in \link{decimate_points}. It implements an algorithm that
#' randomly removes points or pulses to reach the desired density over the area.
#'
#' @param density numeric. The desired output density.
#'
#' @param use_pulse logical. Decimate by removing random pulses instead of random points (requires running
#' \link{retrieve_pulses} first)
#'
#' @export
#'
#' @family point cloud decimation algorithms
#'
#' @examples
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' las = readLAS(LASfile, select = "xyz")
#'
#' # Reach a pulse density of 1 on the overall dataset
#' thinned1 = decimate_points(las, random(1))
#' plot(rasterize_density(las))
#' plot(rasterize_density(thinned1))
#' @name sample_random
random = function(density, use_pulse = FALSE)
{
  assert_is_a_number(density)
  assert_all_are_positive(density)
  assert_is_a_bool(use_pulse)

  density   <- lazyeval::uq(density)
  use_pulse <- lazyeval::uq(use_pulse)

  f = function(las)
  {
    assert_is_valid_context(LIDRCONTEXTDEC, "random")

    if(use_pulse & !"pulseID" %in% names(las))
    {
      warning("No 'pulseID' attribute found. Decimation by points is used.")
      use_pulse <- FALSE
    }

    n <- round(density*area(las))

    if (use_pulse)
      return(.selected_pulses(las@data$pulseID, n))
    else
    {
      if (nrow(las@data) > n)
      {
        idx = sample(1:nrow(las@data), n)
        idx = sort(idx)
        return(idx)
      }
      else
        return(1:nrow(las@data))
    }
  }

  f <- plugin_decimate(f)
  return(f)
}

#' Point Cloud Decimation Algorithm
#'
#' This function is made to be used in \link{decimate_points}. It implements an algorithm that
#' creates a grid with a given resolution and filters the point cloud by randomly selecting some
#' points in each cell. It is designed to produce point clouds that have uniform densities throughout
#' the coverage area. For each cell, the proportion of points or pulses that will be retained is computed
#' using the actual local density and the desired density. If the desired density is greater than the actual
#' density it returns an unchanged set of points (it cannot increase the density). The cell size must be
#' large enough to compute a coherent local density. For example, in a 2 points/m^2 point cloud, 25 square
#' meters would be feasible; however 1 square meter cells would not be feasible because density does
#' not have meaning at this scale.
#'
#' @param density numeric. The desired output density.
#'
#' @param res numeric. The resolution of the grid used to filter the point cloud
#'
#' @param use_pulse logical. Decimate by removing random pulses instead of random points (requires running
#' \link{retrieve_pulses} first)
#'
#' @export
#'
#' @family point cloud decimation algorithms
#'
#' @examples
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' las = readLAS(LASfile, select = "xyz")
#'
#' # Select points randomly to reach an homogeneous density of 1
#' thinned <- decimate_points(las, homogenize(1,5))
#' plot(rasterize_density(thinned, 10))
#' @name sample_homogenize
homogenize = function(density, res = 5, use_pulse = FALSE)
{
  assert_is_a_number(density)
  assert_all_are_positive(density)
  assert_is_a_bool(use_pulse)
  assert_is_a_number(res)
  assert_all_are_positive(res)

  density   <- lazyeval::uq(density)
  res       <- lazyeval::uq(res)
  use_pulse <- lazyeval::uq(use_pulse)

  f = function(las)
  {
    assert_is_valid_context(LIDRCONTEXTDEC, "homogenize")

    if (use_pulse & !"pulseID" %in% names(las))
    {
      warning("No 'pulseID' attribute found. Decimation by points is used.")
      use_pulse <- FALSE
    }

    pulseID <- NULL

    n       <- round(density*res^2)
    layout  <- raster_layout(las, res)
    cells   <- get_group(layout, las)

    if (use_pulse)
      return(las@data[, .I[.selected_pulses(pulseID, n)], by = cells]$V1)
    else
      return(las@data[, .I[.selected_pulses(1:.N, n)], by = cells]$V1)
  }

  f <- plugin_decimate(f)
  return(f)
}

#' Point Cloud Decimation Algorithm
#'
#' These functions are made to be used in \link{decimate_points}. They implement algorithms that
#' create a grid with a given resolution and filters the point cloud by selecting the highest/lowest
#' point within each cell.
#'
#' @param res numeric. The resolution of the grid used to filter the point cloud
#'
#' @export
#'
#' @family point cloud decimation algorithms
#'
#' @examples
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' las = readLAS(LASfile, select = "xyz")
#'
#' # Select the highest point within each cell of an overlayed grid
#' thinned = decimate_points(las, highest(4))
#' #plot(thinned)
#'
#' # Select the lowest point within each cell of an overlayed grid
#' thinned = decimate_points(las, lowest(4))
#' #plot(thinned)
#' @name sample_maxima
highest = function(res = 1)
{
  assert_is_a_number(res)
  assert_all_are_positive(res)

  res <- lazyeval::uq(res)

  f = function(las)
  {
    assert_is_valid_context(LIDRCONTEXTDEC, "highest")
    layout <- raster_layout(las, res)
    return(C_highest(las, layout))
  }

  f <- plugin_decimate(f)
  return(f)
}

#' @family point cloud decimation algorithms
#' @export
#' @name sample_maxima
lowest = function(res = 1)
{
  assert_is_a_number(res)
  assert_all_are_positive(res)

  res <- lazyeval::uq(res)

  f = function(las)
  {
    assert_is_valid_context(LIDRCONTEXTDEC, "lowest")
    layout <- raster_layout(las, res)
    return(C_lowest(las, layout))
  }

  f <- plugin_decimate(f)
  return(f)
}

#' Point Cloud Decimation Algorithm
#'
#' These functions are made to be used in \link{decimate_points}. They implements algorithm that
#' creates a 3D grid with a given resolution and filters the point cloud by selecting
#' points of interest within each voxel. `random_per_voxel()` sample random points. `barycenter_per_voxel()`
#' samples the point that is the closest to the barycenter of the points within a given voxel.
#' `[lowest|highest]_attribute_per_voxel()` sample respectively the point that have the highest/lowest
#' attribute (e.g. Intensity) per voxel.
#'
#' @param res numeric. The resolution of the voxel grid used to filter the point cloud
#' @param n integer. The number of points to select
#' @param attribute string name of an attribute (such as 'intensity')
#'
#' @examples
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' las <- readLAS(LASfile, select = "xyz")
#' thinned <- decimate_points(las, random_per_voxel(8, 1))
#' #plot(thinned)
#' @family point cloud decimation algorithms
#' @export
#' @name sample_per_voxel
random_per_voxel = function(res = 1, n = 1)
{
  assert_all_are_positive(n)
  assert_all_are_positive(res)
  n <- as.integer(n)
  if (length(res) == 1) res <- c(res, res)

  n <- lazyeval::uq(n)
  res <- lazyeval::uq(res)

  f = function(las)
  {
    by <- group_grid_3d(las$X, las$Y, las$Z, res)
    return(las@data[, .I[.selected_pulses(1:.N, n)], by = by]$V1)
  }

  f <- plugin_decimate(f)
  return(f)
}

#' @export
#' @rdname sample_per_voxel
barycenter_per_voxel = function(res = 1)
{
  assert_all_are_positive(res)
  res <- lazyeval::uq(res)

  f = function(las)
  {
    X <- Y <- Z <- NULL

    id = C_voxel_id(las, res)

    which.mean = function(X,Y,Z)
    {
      x = mean(X)
      y = mean(Y)
      z = mean(Z)
      d = (X-x)^2 + (Y-y)^1 + (Z-z)^2
      return(which.min(d))
    }

    return(las@data[, .I[which.mean(X,Y,Z)], by = id]$V1)
  }

  f <- plugin_decimate(f)
  return(f)

}

#' @export
#' @rdname sample_per_voxel
lowest_attribute_per_voxel = function(res, attribute = "Z")
{
  assert_all_are_positive(res)
  assert_is_character(attribute)

  res <- lazyeval::uq(res)
  attribute <- lazyeval::uq(attribute)

  f = function(las)
  {
    if (!attribute %in% names(las))
      stop(paste0(attribute, " is not an attribute in this point cloud"), call. = FALSE)

    tmp <- NULL
    voxelID = C_voxel_id(las, res)
    las@data$tmp = las@data[[attribute]]
    return(las@data[, .I[which.min(tmp)], by = voxelID]$V1)
  }

  f <- plugin_decimate(f)
  return(f)
}

#' @export
#' @rdname sample_per_voxel
highest_attribute_per_voxel = function(res, attribute = "Z")
{
  assert_all_are_positive(res)
  assert_is_character(attribute)
  tmp <- NULL

  res <- lazyeval::uq(res)
  attribute <- lazyeval::uq(attribute)

  f = function(las)
  {
    voxelID = C_voxel_id(las, res)
    las@data$tmp = las@data[[attribute]]
    return(las@data[, .I[which.max(tmp)], by = voxelID]$V1)
  }

  f <- plugin_decimate(f)
  return(f)
}


.selected_pulses = function(pulseID, n)
{
  p <- unique(pulseID)

  if (n > length(p))
    return(rep(TRUE, length(pulseID)))

  selectedPulses <- sample(p, n)
  selectedPulses <- sort(selectedPulses)
  selectedPulses <- pulseID %in% selectedPulses


  return(selectedPulses)
}

context("decimate_points")

las = megaplot
ctg = megaplot_ctg

test_that("decimate_points homogenize algorithm works", {
  lasdec = decimate_points(las, homogenize(0.5,5))
  xdec = rasterize_density(lasdec, res = 5)

  expect_true(data.table::between(median(xdec[], na.rm = TRUE), 0.5-sd(xdec[], na.rm = TRUE), 0.5+sd(xdec[], na.rm = TRUE) ))
})

test_that("decimate_points random algorithm works", {
  lasdec = decimate_points(las, random(0.5))
  xdec = rasterize_density(lasdec, res = 5)

  expect_true(data.table::between(median(xdec[], na.rm = TRUE), 0.5-sd(xdec[], na.rm = TRUE), 0.5+sd(xdec[], na.rm = TRUE) ))
})

test_that("decimate_points highest algorithm works", {
  lasdec = decimate_points(las, highest(2))
  xdec = rasterize_density(lasdec, res = 5)
  zmean = mean(lasdec$Z)

  expect_true(data.table::between(median(xdec[], na.rm = TRUE), 0.23, 0.25))
  expect_equal(zmean, 16.25, tolerance = 0.01)
})

test_that("decimate_points lowest algorithm works", {
  lasdec = decimate_points(las, lowest(2))
  xdec = rasterize_density(lasdec, res = 5)
  zmean = mean(lasdec$Z)

  expect_true(data.table::between(median(xdec[], na.rm = TRUE), 0.23, 0.25))
  expect_equal(zmean, 7.42, tolerance = 0.01)
})

test_that("decimate_points fails with unproper catalog options", {

  expect_error(decimate_points(ctg, homogenize(0.8,5)), "output file")
})

test_that("decimate_points works with a LAScatalog", {

  ctg <- random_2files_250points
  opt_output_files(ctg) <- "{tempdir()}/{ID}"
  ctg2 <- decimate_points(ctg, highest(10))

  expect_is(ctg2, "LAScatalog")
  expect_gt(npoints(ctg), npoints(ctg2))
})

assert_all_are_non_negative = function(x)
{
  x. <- lazyeval::expr_text(x)
  if (!all(x >= 0))
    stop(glue::glue("Values of {x.} are not all positive or null."), call. = FALSE)
}


assert_all_are_positive = function(x)
{
  x. <- lazyeval::expr_text(x)
  if (!all(x > 0))
    stop(glue::glue("Values of {x.} are not all positive."), call. = FALSE)
}


assert_all_are_in_closed_range = function(x, a, b)
{
  x. <- lazyeval::expr_text(x)
  if (!all(x >= a) | !all(x <= b))
    stop(glue::glue("Values of {x.} are not all in range [{a}, {b}]."), call. = FALSE)
}


assert_all_are_in_open_range = function(x, a, b)
{
  x. <- lazyeval::expr_text(x)
  if (!all(x > a) | !all(x < b))
    stop(glue::glue("Values of {x.} are not all in range ]{a}, {b}[."), call. = FALSE)
}

assert_all_are_true = function(x)
{
  x. <- lazyeval::expr_text(x)
  if (!all(x == TRUE))
    stop(glue::glue("Values of {x.} are not all TRUE."), call. = FALSE)
}

assert_are_same_length = function(x, y)
{
  x. <- lazyeval::expr_text(x)
  y. <- lazyeval::expr_text(y)
  if (length(x) != length(y))
    stop(glue::glue("{x.} and {y.} have different lengths."), call. = FALSE)
}

assert_all_are_existing_files = function(x)
{
  x. <- lazyeval::expr_text(x)
  if (!all(file_exists(x)))
    stop(glue::glue("File does not exist."), call. = FALSE)
}

assert_is_a_bool = function(x)
{
  x. <- lazyeval::expr_text(x)
  if (!is.logical(x) | length(x) > 1)
    stop(glue::glue("{x.} is not a boolean."), call. = FALSE)
}

assert_is_a_number = function(x)
{
  x. <- lazyeval::expr_text(x)
  if (!is.numeric(x) | length(x) > 1)
    stop(glue::glue("{x.} is not a number."), call. = FALSE)
}

is_a_number = function(x)
{
  return(is.numeric(x) & length(x) == 1)
}

assert_is_a_string = function(x)
{
  x. <- lazyeval::expr_text(x)
  if (!is.character(x) | length(x) > 1)
    stop(glue::glue("{x.} is not a string."), call. = FALSE)
}


assert_is_numeric = function(x)
{
  x. <- lazyeval::expr_text(x)
  if (!is.numeric(x))
    stop(glue::glue("{x.} is not numeric it is a {class(x)}."), call. = FALSE)
}

assert_is_vector = function(x)
{
  x. <- lazyeval::expr_text(x)
  if (!is.vector(x))
    stop(glue::glue("{x.} is not a vector it is a {class(x)}."), call. = FALSE)
}

assert_is_list = function(x)
{
  x. <- lazyeval::expr_text(x)
  if (!is.list(x))
    stop(glue::glue("{x.} is not a list it is a {class(x)}"), call. = FALSE)
}

assert_is_of_length = function(x, l)
{
  x. <- lazyeval::expr_text(x)
  if (length(x) != l)
    stop(glue::glue("{x.} is not of length {l} it is of length {length(x)}."), call. = FALSE)
}

assert_is_character = function(x)
{
  x. <- lazyeval::expr_text(x)
  if (!is.character(x))
    stop(glue::glue("{x.} is not character it is a {class(x)}"), call. = FALSE)
}


assert_is_all_of = function(x, class)
{
  x. <- lazyeval::expr_text(x)
  if (!is(x, class))
    stop(glue::glue("{x.} is not a {class} it is {class(x)[1]}."), call. = FALSE)
}

assert_is_function = function(x)
{
  x. <- lazyeval::expr_text(x)
  if (!is.function(x))
    stop(glue::glue("{x.} is not a function it is a {class(x)[1]}."), call. = FALSE)
}

assert_all_are_same_crs = function(x)
{
  ref = x[[1]]
  lapply(x, function(x) {
    if (st_crs(x) != st_crs(ref))
      stop("Different CRS.", call. = FALSE)
  })
}

assert_is_algorithm = function(x)
{
  if (!is.algorithm(x))
    stop("Invalid function provided as algorithm.", call. = FALSE)
}

assert_is_algorithm_dsm = function(x)
{
  if (!is(x, LIDRDSM))
    stop("The algorithm used is not an algorithm for digital surface model.", call. = FALSE)
}

assert_is_algorithm_itd = function(x)
{
  if (!is(x, LIDRITD))
    stop("The algorithm used is not an algorithm for individual tree detection.", call. = FALSE)
}

assert_is_algorithm_shp = function(x)
{
  if (!is(x, LIDRSHP))
    stop("The algorithm used is not an algorithm for shape detection.", call. = FALSE)
}

assert_is_algorithm_spi = function(x)
{
  if (!is(x, LIDRSPI))
    stop("The algorithm used is not an algorithm for spatial interpolation.", call. = FALSE)
}

assert_is_algorithm_its = function(x)
{
  if (!is(x, LIDRITS))
    stop("The algorithm used is not an algorithm for individual tree segmentation.", call. = FALSE)
}

assert_is_algorithm_dec = function(x)
{
  if (!is(x, LIDRDEC))
    stop("The algorithm used is not an algorithm for point cloud decimation.", call. = FALSE)
}

assert_is_algorithm_gnd = function(x)
{
  if (!is(x, LIDRGND))
    stop("The algorithm used is not an algorithm for ground segmentation.", call. = FALSE)
}

assert_is_algorithm_sng = function(x)
{
  if (!is(x, LIDRSNG))
    stop("The algorithm used is not an algorithm for snags segmentation.", call. = FALSE)
}

assert_is_algorithm_nit = function(x)
{
  if (!is(x, LIDRNIT))
    stop("The algorithm used is not an algorithm for intensity normalization.", call. = FALSE)
}

assert_is_algorithm_trk = function(x)
{
  if (!is(x, LIDRTRK))
    stop("The algorithm used is not an algorithm for sensor tracking.", call. = FALSE)
}

assert_is_algorithm_out = function(x)
{
  if (!is(x, LIDROUT))
    stop("The algorithm used is not an algorithm for noise segmentation.", call. = FALSE)
}

assert_is_valid_context = function(expected_contexts, name = "", null_allowed = FALSE)
{
  received_context <- tryCatch({get("lidR.context", envir = parent.frame(n = 2L))}, error = function(e) {return(NULL)})

  if (length(received_context) > 1)
    stop("Internal error: the context in which an algorithm has been called is multiple. Please report this error.")

  if (is.null(received_context) && !null_allowed)
    stop(glue::glue("The '{name}' algorithm has not been called in the correct context. Maybe it has been called alone but it should be used within a lidR function."), call. = FALSE)

  if (!null_allowed && !received_context %in% expected_contexts)
    stop(glue::glue("The '{name}' algorithm has not been called in the correct context."), call. = FALSE)

  return(NULL)
}

assert_las_is_not_empty = function(x)
{
  if (is.empty(x))
    stop("The point cloud contains 0 point", call. = FALSE)
}

assert_package_is_installed = function(x)
{
  if (!requireNamespace(x, quietly = TRUE))
    stop(glue::glue("Package '{x}' needed for this function to work. Please install it."), call. = FALSE) # nocov
}


stopifnotlas = function(x)
{
  if (!inherits(x, "LAS"))
    stop("Argument is not a LAS object", call. = FALSE)
}

stopif_forbidden_name = function(name)
{
  if (name %in% LASATTRIBUTES)
    stop(glue::glue("{name} is part of the core attributes and is a forbidden name."), call. = FALSE)
}

context("assertions")

test_that("assertions work", {
  x = c(-1, 1)
  expect_error(lidR:::assert_all_are_non_negative(x))
  expect_error(lidR:::assert_all_are_positive(x))
  expect_error(lidR:::assert_all_are_in_closed_range(x, 0, 1))
  expect_error(lidR:::assert_all_are_in_open_range(x, -1, 1))
  expect_error(lidR:::assert_is_a_number(x))
  expect_error(lidR:::assert_is_function(x))

  y = c(T, F, T)
  expect_error(lidR:::assert_all_are_true(y))
  expect_error(lidR:::assert_is_a_bool(y))
  expect_error(lidR:::assert_is_numeric(y))

  expect_error(lidR:::assert_are_same_length(x, y))

  f = tempfile()
  expect_error(lidR:::assert_all_are_existing_files(f))

  g = function(x) {x}
  expect_error(lidR:::assert_is_vector(g))
  expect_error(lidR:::assert_is_list(g))
  expect_error(lidR:::assert_is_character(g))
  expect_error(lidR:::assert_is_all_of(g, "numeric"))


  expect_error(lidR:::assert_is_algorithm(g))
  expect_error(lidR:::assert_is_algorithm_dsm(g))
  expect_error(lidR:::assert_is_algorithm_itd(g))
  expect_error(lidR:::assert_is_algorithm_its(g))
  expect_error(lidR:::assert_is_algorithm_spi(g))
  expect_error(lidR:::assert_is_algorithm_dec(g))
  expect_error(lidR:::assert_is_algorithm_gnd(g))
  expect_error(lidR:::assert_is_algorithm_sng(g))

  expect_error(lidR:::assert_is_valid_context(tin(), "test"))

  lazfile <- system.file("extdata", "example.laz", package="rlas")
  las <- readLAS(lazfile)
  las <- filter_poi(las, Z > 1000)
  expect_error(lidR:::stopifnotlas(g))
  expect_error(lidR:::assert_las_is_not_empty(las))
  expect_error(lidR:::stopif_forbidden_name("X"))
})

#' @param all_voxels boolean. By default the function returns only voxels that
#' contain 1 or more points. Empty voxels do not exist as the metrics are undefined.
#' If \code{all_voxels = TRUE} all the voxels are returned and metrics are NA for
#' voxels with 0 points.
#' @rdname aggregate
#' @export
voxel_metrics = function(las, func, res = 1, ..., all_voxels = FALSE)
{
  stopifnotlas(las)
  assert_all_are_non_negative(res)
  assert_is_a_bool(all_voxels)

  if (length(res) == 1L)
    res <- c(res,res)
  else if (length(res) > 2L)
    stop("Wrong resolution provided.")

  .I <- .GRP <- GRP <- NULL
  by <- group_grid_3d(las$X, las$Y, las$Z, res, c(0,0,0.5*res[2]))
  by <- data.table::setDT(by)
  by <- by[, GRP := .GRP, by = c("Xgrid", "Ygrid", "Zgrid")]
  grp <- by$GRP
  by <- by[by[, .I[1], by = GRP]$V1]

  M <- template_metrics(las, func, grp, ...)

  if ("Z" %in% names(M))
    warning("Z is reserved for the coordinates of the voxels and cannot be used as metric name. Metric Z discared.", call. = FALSE)

  M[["X"]] <- by[["Xgrid"]][M[["GRPID"]]]
  M[["Y"]] <- by[["Ygrid"]][M[["GRPID"]]]
  M[["Z"]] <- by[["Zgrid"]][M[["GRPID"]]]
  M[["GRPID"]] <- NULL

  if (all_voxels)
  {
    xrange <- range(M$X)
    yrange <- range(M$Y)
    zrange <- range(M$Z)
    X <- f_grid(seq(xrange[1], xrange[2], by = res[1]), res[1], 0)
    Y <- f_grid(seq(yrange[1], yrange[2], by = res[1]), res[1], 0)
    Z <- f_grid(seq(zrange[1], zrange[2], by = res[2]), res[2], 0.5*res[2])
    all <- expand.grid(X = X,Y = Y, Z = Z)
    data.table::setDT(all)
    data.table::setkey(M, X,Y,Z)
    data.table::setkey(all, X,Y,Z)
    M <- merge(M, all, all = TRUE)
  }

  data.table::setcolorder(M, c("X", "Y", "Z"))
  data.table::setattr(M, "class", c("lasmetrics3d", attr(M, "class")))
  data.table::setattr(M, "res", res[1])
  return(M)
}

las <- random_500_points

test_that("voxel_metrics returns a named data.frame", {

  x <- voxel_metrics(las, list(Imean = mean(Intensity)), 5)

  expect_true(is(x, "data.frame"))
  expect_equal(names(x)[4], "Imean")
  expect_equal(nrow(x), 444)
})

test_that("voxel_metrics works with all_voxels", {

  x <- voxel_metrics(las, list(Imean = mean(Intensity)), 5, all_voxels = TRUE)

  expect_true(is(x, "data.frame"))
  expect_equal(names(x)[4], "Imean")
  expect_equal(nrow(x), 2400)
})

test_that("voxel_metrics accepts both an expression or a formula", {

  x <- voxel_metrics(las,  list(Imean = mean(Intensity)), 5)
  y <- voxel_metrics(las, ~list(Imean = mean(Intensity)), 5)

  expect_equal(x, y)
})

#' Automatic colorization
#'
#' Attribute a color to each element of a vector
#'
#' @param x A vector.
#' @param palette function. A color palette function. Default is \code{height.colors} provided by the package lidR.
#' @param trim numeric.
#' @keywords internal
#' @noRd
set.colors = function(x, palette, trim = Inf, value_index = FALSE)
{
  if (all(is.na(x)))
    return()

  # Classification case
  if (value_index)
  {
    x[x >= length(palette)] <- length(palette) - 1
    return(palette[x + 1])
  }

  ncolors <- length(palette)
  if (!is.infinite(trim)) x[x > trim] <- trim
  minx <- min(x, na.rm = T)
  maxx <- max(x, na.rm = T)

  if (maxx - minx == 0) {
    if (!anyNA(x)) {
      colors <- palette[1]
    } else {
      colors <- rep(NA_character_, length(x))
      colors[!is.na(x)] <- palette[1]
    }
  } else {
    idx <- findInterval(x, seq(minx, maxx, length.out = ncolors))
    colors <- palette[idx]
  }

  return(colors)
}

#' @param n The number of colors (> 1) to be in the palette
#' @export
#' @rdname plot
height.colors = function(n)
{
  colfunc <- grDevices::colorRampPalette(c("blue", "cyan2", "yellow", "red"))
  return(colfunc(n))
}

#' @export
#' @rdname plot
forest.colors = function(n)
{
  colfunc <- grDevices::colorRampPalette(c("darkgreen", "lightgreen"))
  return(colfunc(n))
}

amplitude.color = function(n)
{
  colfunc <- grDevices::colorRampPalette(rev(c("#ff0000", "#ff1919", "#ff3333", "#ff4d4d", "#ff6666", "#ff8080", "#ff9999", "#ffb3b3", "#ffcccc", "#ffe6e6")))
  return(colfunc(n))
}

#' @export
#' @rdname plot
random.colors = function(n)
{
  h = stats::runif(n, 0, 1);
  s = stats::runif(n, 0.2, 1);
  l = stats::runif(n, 0.8, 1);

  return(grDevices::hsv(h,s,l))
}

#' @export
#' @rdname plot
pastel.colors = function(n)
{
  h = stats::runif(n, 0, 360);
  c = stats::runif(n, 42, 98);
  l = stats::runif(n, 40, 90);

  return(grDevices::hcl(h,c,l))
}

lasclass.colors = function(...)
{
  pal = c("lightgrey",   # 0 never classifiied
          "lightgray",   # 1 unclassified
          "blue",        # 2 ground
          "limegreen",   # 3 low vegetation
          "forestgreen", # 4 medium vegetation
          "darkgreen",   # 5 high vegetation
          "red",         # 6 building
          "yellow",      # 7 low point (noise)
          "yellow",      # 8 reserved
          "#6495ED",     # 9 water
          "yellow",      # 10 rail
          "gray20",      # 11 Road surface
          "yellow",      # 12 reserved
          "pink",        # 13 wire
          "pink",        # 14 wire
          "purple",      # 15 transmission tower
          "pink",        # 16 Wire connector
          "orange",      # 17 bridge deck
          "yellow")

  hex <- grDevices::rgb(t(grDevices::col2rgb(pal)), maxColorValue = 255)
  return(hex)
}

context("utils_colors")

test_that("set.colors attribute the proper color", {
  x = 1:10
  c = lidR:::set.colors(x, height.colors(50))

  expect_equal(length(unique(c)), length(x))

  x = rep(1,10)
  c = lidR:::set.colors(x, height.colors(50))

  expect_equal(c, "#0000FF")
})

test_that("colors palette works", {
  expect_error(random.colors(10), NA)
  expect_error(forest.colors(10), NA)
  expect_error(pastel.colors(10), NA)
  expect_error(height.colors(10), NA)
})

#' Noise Segmentation Algorithm
#'
#' This function is made to be used in \link{classify_noise}. It implements an
#' algorithm for outliers (noise) segmentation based on Statistical Outliers
#' Removal (SOR) methods first described in the PCL library
#' and also implemented in CloudCompare (see references).
#' For each point, it computes the mean distance to all its k-nearest neighbours.
#' The points that are farther than the average distance plus a number of times
#' (multiplier) the standard deviation are considered noise.
#'
#' @param k numeric. The number of neighbours
#' @param m numeric. Multiplier. The maximum distance will be: `avg distance + m * std deviation`.
#' If `quantile = TRUE`, `m` becomes the quantile threshold.
#' @param quantile boolean. Modification of the original SOR to use a quantile
#' threshold instead of a standard deviation multiplier. In this case the maximum
#' distance will be: `quantile(distances, probs = m)`
#'
#' @export
#'
#' @family noise segmentation algorithms
#'
#' @examples
#' LASfile <- system.file("extdata", "Topography.laz", package="lidR")
#' las <- readLAS(LASfile, filter = "-inside 273450 5274350 273550 5274450")
#'
#' # Add some artificial outliers because the original
#' # dataset is 'clean'
#' set.seed(314)
#' id = round(runif(20, 0, npoints(las)))
#' set.seed(42)
#' err = runif(20, -50, 50)
#' las$Z[id] = las$Z[id] + err
#'
#' las <- classify_noise(las, sor(15,7))
#' @md
#' @references
#' https://pointclouds.org/documentation/tutorials/statistical_outlier.html \cr
#' https://www.cloudcompare.org/doc/wiki/index.php?title=SOR_filter
#' @name noise_sor
sor = function(k = 10, m = 3, quantile = FALSE)
{
  assert_is_a_number(k)
  assert_is_a_number(m)
  assert_all_are_positive(k)
  assert_is_a_bool(quantile)

  if (!quantile)
    assert_all_are_non_negative(m)
  else
    assert_all_are_in_closed_range(m, 0, 1)

  k <- lazyeval::uq(k)
  m <- lazyeval::uq(m)
  quantile <- lazyeval::uq(quantile)

  f = function(las)
  {
    assert_is_valid_context(LIDRCONTEXTOUT, "sor")
    dmean <- C_knn_distance(las, k, getThreads())
    if (quantile) {
      th <- quantile(dmean, probs = m)
    } else {
      avg <- mean(dmean)
      std <- stats::sd(dmean)
      th <- avg + m*std
    }

    return(dmean > th)
  }

  f <- plugin_outliers(f, omp = TRUE)
  return(f)
}

#' Noise Segmentation Algorithm
#'
#' This function is made to be used in \link{classify_noise}. It implements an
#' algorithm for outliers (noise) segmentation based on isolated voxels filter (IVF).
#' The algorithm finds points that have only a few other points in their surrounding
#' 3 x 3 x 3 = 27 voxels.
#'
#' @param res numeric. Resolution of the voxels
#' @param n integer. The maximal number of 'other points' in the 27 voxels
#'
#' @export
#'
#' @family noise segmentation algorithms
#'
#' @examples
#' LASfile <- system.file("extdata", "Topography.laz", package="lidR")
#' las <- readLAS(LASfile, filter = "-inside 273450 5274350 273550 5274450")
#'
#' # Add some artificial outliers
#' set.seed(314)
#' id = round(runif(20, 0, npoints(las)))
#' set.seed(42)
#' err = runif(20, -50, 50)
#' las$Z[id] = las$Z[id] + err
#'
#' las <- classify_noise(las, ivf(5,2))
#' @name noise_ivf
ivf = function(res = 5, n = 6)
{
  assert_is_a_number(res)
  assert_is_a_number(n)
  assert_all_are_positive(res)
  assert_all_are_non_negative(n)

  res <- lazyeval::uq(res)
  n <- lazyeval::uq(n)

  f = function(las)
  {
    assert_is_valid_context(LIDRCONTEXTOUT, "ivf")
    return(C_isolated_voxel(las, res, n))
  }

  f <- plugin_outliers(f)
  return(f)
}

#' @export
#' @rdname classify
classify_poi = function(las, class, poi = NULL, roi = NULL, inverse_roi = FALSE, by_reference = FALSE)
{
  UseMethod("classify_poi", las)
}

#' @export
classify_poi.LAS = function(las, class, poi = NULL, roi = NULL, inverse_roi = FALSE, by_reference = FALSE)
{
  assert_is_a_number(class)
  assert_all_are_non_negative(class)
  assert_is_a_bool(inverse_roi)
  stopifnot(class == as.integer(class))

  Classification <- NULL

  bool1 <- rep(TRUE, npoints(las))
  if (!is.null(poi)) bool1 <- parse_filter(las, poi)

  bool2 <- rep(TRUE, npoints(las))
  if (!is.null(roi))
  {
    las <- merge_spatial(las, roi, "..inpoly..")
    bool2 <- las[["..inpoly.."]]
    las[["..inpoly.."]] <- NULL
    if (inverse_roi) bool2 <- !bool2
  }

  bool <- bool1 & bool2

  if (! "Classification" %in% names(las))
  {
    if (by_reference)
      las@data[, Classification := 0L]
    else
      las@data[["Classification"]] <- 0L
  }

  if (by_reference)
  {
    las@data[bool, Classification := class]
    return(invisible(las))
  }
  else
  {
    las$Classification[bool] <- class
    return(las)
  }
}

#' @export
classify_poi.LAScatalog = function(las, class, poi = NULL, roi = NULL, inverse_roi = FALSE, by_reference = FALSE)
{
  opt_select(las) <- "*"
  opt_chunk_buffer(las) <- 0

  options <- list(need_buffer = FALSE, drop_null = TRUE, need_output_file = TRUE)
  output  <- catalog_map(las, classify_poi, class = class, poi = poi, roi = roi, inverse_roi = inverse_roi, .options = options)
  return(output)
}

options(lidR.progress = FALSE)
options(rgl.useNULL = TRUE)
set_lidr_threads(1)

cat("Tests using raster:", getOption("lidR.raster.default"), "\n")
cat("Tests using future:", lidR:::engine_use_future(), "\n")
cat("Tests using OpenMP thread:", get_lidr_threads(), "\n")
cat("Tests using rlas:", as.character(packageVersion("rlas")), "\n")
#Sys.setenv(NOT_CRAN = "false")

megaplot_laz_path <- system.file("extdata", "Megaplot.laz", package="lidR")
mixedconifer_laz_path <- system.file("extdata", "MixedConifer.laz", package="lidR")
topography_laz_path <- system.file("extdata", "Topography.laz", package="lidR")
example_laz_path <- system.file("extdata", "example.laz", package="rlas")

megaplot_las_path <- tempfile(fileext = ".las")
mixedconifer_las_path <- tempfile(fileext = ".las")
topography_las_path <- tempfile(fileext = ".las")
example_las_path <- tempfile(fileext = ".las")

megaplot <- readLAS(megaplot_laz_path) # 110 ms
mixedconifer <- readLAS(mixedconifer_laz_path)
topography <- readLAS(topography_laz_path) # 110 ms
example <- readLAS(example_laz_path) # 30 ms
random_10_points <- lidR:::generate_las(10)
random_500_points <- lidR:::generate_las(500)

writeLAS(megaplot, megaplot_las_path)
writeLAS(mixedconifer, mixedconifer_las_path)
writeLAS(topography, topography_las_path)
writeLAS(example, example_las_path)

megaplot_ctg <- readLAScatalog(megaplot_las_path) # 50 ms + 30 ms at read time
mixedconifer_ctg <- readLAScatalog(mixedconifer_las_path)
topography_ctg <- readLAScatalog(topography_las_path)
example_ctg <- readLAScatalog(example_las_path)
random_2files_250points <- lidR:::catalog_generator(2, 250) # 180 ms

chunk = lidR:::engine_chunks(random_2files_250points)[[1]]

#' @param res numeric. The resolution of the output. Can optionally be a `RasterLayer` or a `stars` or
#' a `SpatRaster`. In that case the raster is used as the template.
#' @param start vector of x and y coordinates for the reference raster. Default is (0,0) meaning that the
#' grid aligns on (0,0). Not consiered if `res` is a raster
#' @rdname aggregate
#' @export
#' @md
pixel_metrics = function(las, func, res = 20, start = c(0,0), ...)
{
  UseMethod("pixel_metrics", las)
}

#' @export
pixel_metrics.LAS = function(las, func, res = 20, start = c(0,0), ...)
{
  # Defensive programming
  if (is_a_number(res)) assert_all_are_non_negative(res)
  assert_is_numeric(start)

  template <- raster_layout(las, res, start)
  M <- template_metrics(las, func, template, ...)
  return(M)
}

#' @export
pixel_metrics.LAScluster = function(las, func, res = 20, start = c(0,0), ...)
{
  x <- readLAS(las)
  if (is.empty(x)) return(NULL)

  # crop the raster to the extent of the chunk because the raster can be a very large proxy
  # that will be materialized by rasterize_terrain LAS
  if (is_raster(res)) res <- raster_crop(res, st_adjust_bbox(x, raster_res(res)))

  M <- pixel_metrics(x, func, res, start, ...)
  M <- raster_crop(M, st_bbox(las))
  return(M)
}


#' @export
pixel_metrics.LAScatalog = function(las, func, res = 20, start = c(0,0), ...)
{
  if (is_a_number(res))
    assert_all_are_non_negative(res)
  else if (!raster_is_supported(res))
    stop("'res' must be a number of a raster.", call. = FALSE)

  assert_is_numeric(start)

  # Compute the alignment options including the case when res is a RasterLayer/stars/terra
  alignment <- raster_alignment(res, start)

  # subset the collection to the size of the layout (if any)
  if (!is_a_number(res)) las <- catalog_intersect(las, res)

  if (opt_chunk_size(las) > 0 && opt_chunk_size(las) < 2*alignment$res)
    stop("The chunk size is too small. Process aborted.", call. = FALSE)

  # Enforce some options
  if (opt_wall_to_wall(las)) opt_chunk_buffer(las) <- 0.1*alignment[["res"]]

  # Processing
  globals = NULL
  if (engine_use_future()) globals <- future::getGlobalsAndPackages(func)
  options <- list(need_buffer = FALSE, drop_null = TRUE, globals = names(globals$globals), raster_alignment = alignment, automerge = TRUE)
  output  <- catalog_apply(las, pixel_metrics, func = func, res = res, start = start, ..., .options = options)
  return(output)
}

LASfile <- system.file("extdata", "example.las", package="rlas")
las = readLAS(LASfile)
compression_enabled <- utils::packageVersion("rlas") >= "1.6.0"

test_that("readLAS reads compressed data", {
  skip_if_not(compression_enabled, "Compression not supported")
  expect_gte(sum(las_is_compressed(las)), 5)
})

test_that("las_check does not materialize", {
  skip_if_not(compression_enabled, "Compression not supported")
  las_check(las, print = FALSE)
  expect_gte(sum(las_is_compressed(las)), 5)
})

test_that("filer does not materialize", {
  skip_if_not(compression_enabled, "Compression not supported")
  filter_ground(las)
  expect_gte(sum(las_is_compressed(las)), 5)
})

test_that("serialization does not materialize", {
  skip_if_not(compression_enabled, "Compression not supported")
  x <- serialize(las, NULL)
  y <- unserialize(x)
  expect_gte(sum(las_is_compressed(y)), 5)
})

test_that("locate_tree does not materialize", {
  skip_if_not(compression_enabled, "Compression not supported")
  m <- locate_trees(las, lmf(3))
  expect_gte(sum(las_is_compressed(las)), 5)
})

test_that("writeLAS does not materialize", {
  skip_if_not(compression_enabled, "Compression not supported")
  f <- tempfile(fileext = ".las")
  writeLAS(las, f)
  expect_gte(sum(las_is_compressed(las)), 5)
})


test_that("aggregation does not materialize", {
  skip_if_not(compression_enabled, "Compression not supported")
  m <- pixel_metrics(las, ~max(Z), 5)
  expect_gte(sum(las_is_compressed(las)), 5)
})

test_that("aggregation does materializes only requiered columns", {
  skip_if_not(compression_enabled, "Compression not supported")
  m <- pixel_metrics(las, ~sum(UserData), 5)
  expect_gte(sum(las_is_compressed(las)), 4)
})

context("catalog_apply restart")

ctg <- random_2files_250points
ctg@chunk_options$size = 50
ctg@chunk_options$buffer = 0
ctg@processing_options$progress <- FALSE

test_that("catalog drops some chunk", {
  ctg@chunk_options$drop = 1:3
  cls = engine_chunks(ctg)
  expect_equal(length(cls), 5)

  ctg@chunk_options$drop = c(1:3, 8)
  cls = engine_chunks(ctg)
  expect_equal(length(cls), 4)
})

test_that("catalog engine returns a valid output", {

  ctg@chunk_options$drop = 1:3
  m = pixel_metrics(ctg, ~mean(Z), 20)

  expect_equivalent(st_bbox(m), st_bbox(c(xmin = 0, xmax = 100,ymin = 60,ymax = 200)))
  expect_equal(sum(is.na(m[])), 9)
})

tDelaunay = function(P, trim = 0, scales = c(1,1), offsets = c(0,0), option = "QbB")
{
  stopifnot(is.numeric(trim), length(trim) == 1L)
  stopifnot(is.numeric(scales), length(scales) == 2L)
  stopifnot(is.numeric(offsets), length(offsets) == 2L)

  # Use internal triangulation based on Boost voronoi tesselation
  # This methods needs to convert xy to integers with a scale and an offset.
  boosted_triangulation <- TRUE

  # P is a LAS object. Use the scale and offset of the header
  if (inherits(P, "LAS")) {
    xscale  <- P[["X scale factor"]]
    yscale  <- P[["Y scale factor"]]
    xoffset <- P[["X offset"]]
    yoffset <- P[["Y offset"]]
    scales  <- c(xscale, yscale)
    offsets <- c(xoffset, yoffset)
    P <- coordinates3D(P)
  }

  if (!is.data.frame(P) && !is.matrix(P)) {
    stop("Internal error. No method to triangulate this input", call. = FALSE)
  }

  # If the input is a matrix fall back to the old methods using the geometry package
  # for backward compatibility
  if (is.matrix(P)) {
    boosted_triangulation <- FALSE
    Q <- if (ncol(P) > 2) P[,1:2] else P
  }

  # If the xy scale factors are incomptaible fall back to geometry because the
  # geometry of coordinates cannot be preserved when converted to integers
  if (boosted_triangulation) {
    if (scales[1] != scales[2]) {
      message("The Delaunay triangulation reverted to the old slow method because xy scale factors are different, so the fast method cannot be applied.")
      boosted_triangulation <- FALSE
      P <- as.matrix(P)
      Q <- as.matrix(P[,1:2])
    }
  }

  # Rcpp needs a name to acces each column. Ensure that the columns are X and Y
  if (boosted_triangulation) {
    if (!all(c("X", "Y") %in% names(P)))
      stop("Columns are not named XY.", call. = FALSE)
  }

  # Check that the conversion to integers actually gives integers. This may fail if the scale
  # factors were edited by hand, if several incompatible dataset were merged or something else
  # Only the first one is tested to fail early but the C++ code check each conversion.
  if (boosted_triangulation) {
    X <- P$X[1]
    Y <- P$Y[1]
    x <- (X - offsets[1]) / scales[1]
    y <- (Y - offsets[2]) / scales[2]

    if (abs(x - round(x)) > 1e-5 | abs(y - round(y)) > 1e-5) {
      message("The Delaunay triangulation reverted to the old slow method because xy coordinates were not converted to integer. xy scale factors and offsets are likely to be invalid")
      boosted_triangulation <- FALSE
      P <- as.matrix(P)
      Q <- as.matrix(P[,1:2])
    }
  }

  if (boosted_triangulation) {
    D <- C_delaunay(P, scales, offsets, trim)
    if (trim != 0) D <- D[!is.na(D[,1]),]
    return(D)
  }
  else {
    D <- suppressMessages(geometry::delaunayn(Q, options = option))

    if (trim != 0) {
      N <- tInfo(D, P)
      K <- N[,7] < abs(trim)
      if (trim < 0) K <- !K
      D <- D[K, , drop = FALSE]
    }

    return(D)
  }
}

# @rdname tDelaunay
tInterpolate = function(D, P, X, threads = 1L)
{
  stopifnot(is.matrix(D), is.matrix(P), is.matrix(X))

  I <- tSearch(D, P, X, threads)
  N <- tInfo(D, P)
  N <- N[I,1:4]
  Z <- -(X[,1] * N[,1] + X[,2] * N[,2] + N[,4]) / N[,3]
  return(Z)
}

# @rdname tDelaunay
tSearch = C_tsearch

# @rdname tDelaunay
tInfo =  C_tinfo

context("triangulation")

las <- filter_ground(topography)
las <- las[seq(1, npoints(las), 10)]

test_that("Boosted delaunay produces the correct output", {

  ps <- data.frame(
    X = c(0,  1, -1, 1, -1),
    Y = c(0, -1, 1, 1, -1))

  expected = structure(c(3L, 1L, 3L, 4L, 5L, 5L, 1L, 1L, 1L, 2L, 4L, 2L), .Dim = 4:3)

  ts <- lidR:::tDelaunay(ps, scales = c(1,1), offsets = c(0,0))

  expect_is(ts, "matrix")
  expect_equal(ts, expected)
})

test_that("Old delaunay produces the correct output", {

  skip_if_not_installed("geometry")

  ps <- data.frame(
    X = c(0,  1, -1, 1, -1),
    Y = c(0, -1, 1, 1, -1))
  ps <- as.matrix(ps)

  expected <- structure(c(1L, 1L, 5L, 5L, 4L, 4L, 1L, 1L, 3L, 2L, 3L, 2L), .Dim = 4:3)

  ts <- lidR:::tDelaunay(ps)

  expect_is(ts, "matrix")
  expect_equal(ts, expected)
})

test_that("Delaunay works with a LAS", {

  ts <- lidR:::tDelaunay(las)

  expect_is(ts, "matrix")
  expect_equal(dim(ts), c(1617,3))
})

test_that("Boosted Delaunay trimming option works", {

  set.seed(42)
  X <- round(runif(50,0,10),2)
  set.seed(123)
  Y <- round(runif(50,0,10),2)
  ps <- data.frame(X,Y)

  ts <- lidR:::tDelaunay(ps, scale = c(0.01, 0.01), offsets = c(0,0), trim = 0)
  expect_is(ts, "matrix")
  expect_equal(dim(ts), c(89,3))

  ts <- lidR:::tDelaunay(ps, scale = c(0.01, 0.01), offsets = c(0,0), trim = 2)
  expect_is(ts, "matrix")
  expect_equal(dim(ts), c(41,3))

  ts <- lidR:::tDelaunay(ps, scale = c(0.01, 0.01), offsets = c(0,0), trim = -2)
  expect_is(ts, "matrix")
  expect_equal(dim(ts), c(48,3))
})

test_that("Old Delaunay trimming option works", {

  skip_if_not_installed("geometry")

  set.seed(42)
  X <- round(runif(50,0,10),2)
  set.seed(123)
  Y <- round(runif(50,0,10),2)
  ps <- as.matrix(data.frame(X, Y, 0))

  ts <- lidR:::tDelaunay(ps, trim = 0)
  expect_is(ts, "matrix")
  expect_equal(dim(ts), c(89,3))

  ts <- lidR:::tDelaunay(ps, trim = 2)
  expect_is(ts, "matrix")
  expect_equal(dim(ts), c(41,3))

  ts <- lidR:::tDelaunay(ps, trim = -2)
  expect_is(ts, "matrix")
  expect_equal(dim(ts), c(48,3))
})

test_that("Delaunay fails when necessary", {

  ps <- data.frame(X = 1, Y = 2)
  expect_error(lidR:::tDelaunay(ps), "cannot triangulate less than 3 points")

  ps <- data.frame( A = c(0, 1, 1), B = c(0, 0, 1))
  expect_error(lidR:::tDelaunay(ps), "Columns are not named XY")

  ps <- 2
  expect_error(lidR:::tDelaunay(ps), "No method to triangulate this input")
})


test_that("Internal C_delaunay fails where necessary", {

  ps <- data.frame(
    X = c(0, -1),
    Y = c(0, -1))

  expect_error(lidR:::C_delaunay(ps, scales = c(1,1), offsets = c(0,0)), "cannot triangulate less than 3 points")

  ps <- data.frame(
    A = c(0, 1, 1),
    B = c(0, 0, 1))

  expect_error(lidR:::C_delaunay(ps, scales = c(1,1), offsets = c(0,0)), "columns are not named XY")

  ps <- data.frame(
    X = c(0, 1, 1),
    Y = c(0, 0, 1))

  expect_error(lidR:::C_delaunay(ps, scales = c(1,1), offsets = c(0.25,0)), "Scale factors are likely to be invalid")

  ps <- data.frame(
    X = runif(5, 0, 10),
    Y = runif(5, 0, 10))

  expect_error(lidR:::C_delaunay(ps, scales = c(1,1), offsets = c(0,0)), "Scale factors are likely to be invalid")
  expect_error(lidR:::C_delaunay(ps, scales = c(1,2), offsets = c(0,0)), "cannot triangulate points with different xy scale factors")
})

test_that("Internal C_Delaunay works with degenerated points", {

  ps <- data.frame(
    X = c(0, -1, 1, -1, 1, -1, 1, -1, 1, 0),
    Y = c(0, -1, -1, 1, 1, -1, -1, 1, 1, 0))

  expected = structure(c(4L, 1L, 4L, 5L, 2L, 2L, 1L, 1L, 1L, 3L, 5L, 3L), .Dim = 4:3)

  ts <- lidR:::C_delaunay(ps, scales = c(1,1), offsets = c(0,0))

  expect_is(ts, "matrix")
  expect_equal(ts, expected)
})

test_that("delaunay rasterization fall back to geometry", {

  skip_if_not_installed("geometry")

  ps <- data.frame(
    X = c(0,  1, -1, 1, -1),
    Y = c(0, -1, 1, 1, -1),
    Z = c(0, 1, -1, 0, 1))

  rs <- data.frame(
    X = c(0.2, 0.2),
    Y = c(0.8, -0.4))

  expect_message(lidR:::interpolate_delaunay(ps, rs, 0, scales = c(1,2), offsets = c(0,0)),  "reverted to the old slow method")
  ts <- suppressMessages(lidR:::interpolate_delaunay(ps, rs, 0, scales = c(1,2), offsets = c(0,0)))

  expect_equal(ts, c(-0.3, 0.4), tol = 1e-5)

  ps <- data.frame(
    X = c(0, 1, 1),
    Y = c(0, 0, 1))

  expect_message(lidR:::tDelaunay(ps, scales = c(1,1), offsets = c(0.25,0)), "reverted to the old slow method")
})

#' Plot voxelized LiDAR data
#'
#' This function implements a 3D plot method for 'lasmetrics3d' objects
#'
#' @param x An object of the class `lasmetrics3d`
#' @param y Unused (inherited from R base)
#' @param \dots Supplementary parameters for \link[lidR:plot]{plot}. The function internally uses the
#' same plot function than LAS objects.
#' @examples
#' \dontrun{
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' lidar = readLAS(LASfile)
#'
#' voxels = voxel_metrics(lidar, list(Imean = mean(Intensity)), res = 5)
#' plot(voxels, color = "Imean", colorPalette = heat.colors(50), trim=60)
#' }
#' @export
#' @method plot lasmetrics3d
#' @md
plot.lasmetrics3d = function(x, y, ...)
{
  cl <- class(x)
  on.exit(data.table::setattr(x, "class", cl))
  header = rlas::header_create(x)
  las = LAS(x, header, check = FALSE)
  attr(las, "res") <- attr(x, "res")
  plot(las, ...)
}

#' Add a spatial object to a point cloud scene
#'
#' Add a raster (`raster`, `stars` `terra`) object that represents a digital terrain model or a
#' `SpatialPointsDataFrame` or `sf` that represents tree tops to a point cloud scene. To add elements
#' to a scene with a point cloud plotted with the function plot from lidR, the functions `add_*`
#' take as first argument the output of the plot function (see examples), because the plot function
#' does not plot the actual coordinates of the point cloud, but offset values. See function
#' \link[=plot]{plot} and its argument `clear_artifacts` for more details. It works only
#' with `rgl` i.e. `backend = "rgl"` which is the default.
#'
#' @param dtm An object of the class `RasterLayer` or `stars` or `SpatRaster`
#' @param bg The color for the background. Default is black.
#' @param \dots Supplementary parameters for \link[rgl]{surface3d} or
#' \link[rgl:spheres]{spheres3d}.
#' @param x The output of the function plot used with a LAS object.
#' @param ttops A `SpatialPointsDataFrame` or `sf/sfc` that contains tree tops coordinates.
#' @param flightlines A `SpatialPointsDataFrame` or `sf` that contains flightlines coordinates.
#' @param z character. The name of the attribute that contains the height of the tree tops or of the
#' flightlines. Only for XY geometries Ignored if the input have XYZ geometries
#' @param clear_artifacts logical. It is a known and documented issue that 3D visualisation with
#' \code{rgl} displays artifacts. The points and lines are inaccurately positioned in the space and thus
#' the rendering may look false or weird. This is because `rgl` computes with single precision `float`.
#' To fix this, the objects are shifted to (0,0) to reduce the number of digits needed to represent
#' their coordinates. The drawback is that the objects are not plotted at their actual coordinates.
#' @param center_x,center_y,radius,height horizontal circle parameters
#'
#' @name plot_3d
#' @examples
#' \dontrun{
#' LASfile <- system.file("extdata", "Topography.laz", package="lidR")
#' las <- readLAS(LASfile)
#'
#' dtm <- rasterize_terrain(las, algorithm = tin())
#' ttops <- locate_trees(las, lmf(ws = 5))
#'
#' plot_dtm3d(dtm)
#'
#' x <- plot(las)
#' add_dtm3d(x, dtm)
#' add_treetops3d(x, ttops)
#'
#' plot(las) |> add_dtm3d(dtm) |> add_treetops3d(ttops)
#' }
NULL

#' @rdname plot_3d
#' @export
#' @md
plot_dtm3d = function(dtm, bg = "black", clear_artifacts = TRUE, ...)
{
  rgl::open3d()
  rgl::bg3d(color = bg)
  shift = c(0,0)

  if (clear_artifacts)
  {
    bbox  <- raster_bbox(dtm)
    shift <- c(bbox$xmin, bbox$ymin)
  }

  add_dtm3d(shift, dtm, ...)

  .pan3d(2)

  if (clear_artifacts)
    return(invisible(shift))
  else
    return(invisible(c(0,0)))
}

#' @rdname plot_3d
#' @export
add_dtm3d = function(x, dtm, ...)
{
  args <- list(...)

  assert_is_numeric(x)
  assert_is_of_length(x, 2)

  if (!is_raster(dtm))
    stop("'dtm' is not a raster")

  res <- raster_as_matrix(dtm, downsample = TRUE)

  if (is.null(args$front))
    args$front <- "lines"

  if (is.null(args$col))
    args$col <- "white"

  args$x <- res$x - x[1]
  args$y <- res$y - x[2]
  args$z <- res$z

  do.call(rgl::surface3d, args)
  return(invisible(x))
}

#' @rdname plot_3d
#' @export
add_treetops3d = function(x, ttops, z = "Z", ...)
{
  args <- list(...)

  assert_is_numeric(x)
  assert_is_of_length(x, 2)

  if (is(ttops, "SpatialPointsDataFrame"))
    ttops <- sf::st_as_sf(ttops)

  if (!is(ttops, "sf"))
    stop("'ttops' is not a SpatialPointsDataFrame or sf")

  if (is.null(args$size))
    args$size <- 5

  if (is.null(args$col))
    args$col <- "red"

  args$add <- TRUE
  coords   <- sf::st_coordinates(ttops)
  args$x   <- coords[,1] - x[1]
  args$y   <- coords[,2] - x[2]

  if (ncol(coords) == 3)
    args$z <- coords[,3]
  else
    args$z <- ttops[[z]]

  do.call(rgl::spheres3d, args)
  return(invisible(x))
}

#' @rdname plot_3d
#' @export
add_flightlines3d = function(x, flightlines, z = "Z", ...)
{
  return(add_treetops3d(x,flightlines, z = "Z", ...))
}

#' @rdname plot_3d
#' @export
add_circle3d <- function(x, center_x, center_y, radius, height)
{
  theta <- seq(0, 2 * pi, length.out = 50)
  xx <- center_x - x[1] + radius * cos(theta)
  yy <- center_y - x[2] + radius * sin(theta)
  zz <- rep(height, 50)

  # Plot the circle in 3D
  rgl::lines3d(xx, yy, zz, lwd = 5, col = "red")
}

context("plot")

las     <- example
ctg     <- megaplot_ctg

test_that("plot LAS works", {
  skip_on_cran()

  expect_error(plot(las), NA)
  rgl::close3d()
})

test_that("plot LAS works with attributes", {

  skip_on_cran()

  las@data$treeID <- sample(1:3, npoints(las), T)
  las@data$R <- 255
  las@data$G <- 125
  las@data$B <- 125

  expect_error(plot(las, color = "Intensity"), NA)
  rgl::close3d()

  expect_error(plot(las, color = "Classification"), NA)
  rgl::close3d()

  expect_error(plot(las, color = "ScanAngleRank"), NA)
  rgl::close3d()

  expect_error(plot(las, color = "ReturnNumber"), NA)
  rgl::close3d()

  expect_error(plot(las, color = "Synthetic_flag"), NA)
  rgl::close3d()

  expect_error(plot(las, color = "treeID"), NA)
  rgl::close3d()

  expect_error(plot(las, color = "RGB", nbits = 8), NA)
  rgl::close3d()
})

test_that("plot LAS does not work with missing attributes", {
  expect_error(plot(las, color = "Plop"), "color' should refer to an attribute")
})

test_that("plot LAS works with artifact", {
  skip_on_cran()
  expect_error(plot(las, clear_artifact = FALSE), NA)
})

test_that("plot LAS does not works with 0 points", {
  expect_error(plot(filter_poi(las, Z > 1000)), "Cannot display an empty point cloud")
})

test_that("plot LAS checks the arguments", {
  expect_error(plot(las, color = c("Z", "Intensity")), "'color' should contain a single value")

  expect_error(plot(las, color = "RGB"), "No 'RGB' attributes found")
})

test_that("plot LASheader works", {
  expect_error(plot(las@header), NA)
})

test_that("Plot LAScatalog object works", {
  expect_error(plot(ctg), NA)
  expect_error(plot(ctg, chunk = TRUE), NA)
  expect_error(plot(ctg, col = "red"), NA)
  expect_error(plot(ctg@data["Min.Z"]), NA)

  skip_on_cran()
  skip_on_os("windows")
  expect_error(plot(ctg, mapview = TRUE), NA)
})

test_that("plot lasmetrics3d works", {
  skip_on_cran()
  x = voxel_metrics(las, ~length(Z), 5)
  expect_error(plot(x), NA)
  rgl::close3d()
})

test_that("plot dtm3 works", {
  skip_on_cran()
  x = rasterize_terrain(las, 1, knnidw())
  expect_error(plot_dtm3d(x), NA)
  rgl::close3d()
})

test_that("add dtm3d works", {
  skip_on_cran()
  x = rasterize_terrain(las, 1, knnidw())
  expect_error({y = plot(las) ; add_dtm3d(y, x)}, NA)
  rgl::close3d()
})

test_that("add treetop3d works", {
  skip_on_cran()
  x = locate_trees(las, lmf(3))
  expect_error({y = plot(las) ; add_treetops3d(y, x)}, NA)
  rgl::close3d()
})

test_that("add = x overlay a second point cloud", {
  skip_on_cran()
  gnd = filter_poi(las, Classification == LASGROUND)
  ngnd = filter_poi(las, Classification != LASGROUND)

  x = plot(ngnd)
  expect_error(plot(gnd, color = "Classification", add = x), NA)
  rgl::close3d()

})

test_that("plot voxels", {
  skip_on_cran()

  v <- voxel_metrics(example, func = ~length(Z))

  expect_error(plot(v, voxel = TRUE), NA)
  rgl::close3d()

  expect_error(plot(v, voxel = 2), NA)
  rgl::close3d()

  attr(v, "res") = NULL

  expect_error(plot(v, voxel = TRUE), NA)
  rgl::close3d()
})

LIDRTHREADS = new.env()
LIDRTHREADS$n = 1L
LIDRTHREADS$input <- 1L

#' Set or get number of threads that lidR should use
#'
#' Set and get number of threads to be used in lidR functions that are parallelized with OpenMP.
#' 0 means to utilize all CPU available. \code{get_lidr_threads()} returns the number
#' of threads that will be used. This affects \code{lidR} package but also the \code{data.table} package
#' by internally calling \link[data.table:openmp-utils]{setDTthreads} because several functions  of
#' lidR rely on \code{data.table} but it does not change R itself or other packages using OpenMP.
#'
#' @seealso \link{lidR-parallelism}
#'
#' @param threads Positive scalar. Default 0 means use all CPU available. Values > 1 mean
#' using n cores, values in ]0, 1[ mean using a fraction of the cores e.g. 0.5 = half.
#' @export
set_lidr_threads = function(threads)
{
  assert_is_a_number(threads)
  assert_all_are_non_negative(threads)

  max <- R_omp_get_max_threads()

  if (max < 0)
  {
    if (threads > 1) message("This installation of lidR has not been compiled with OpenMP support") # nocov
    LIDRTHREADS$input <- threads
    LIDRTHREADS$n <- 1L # nocov
    data.table::setDTthreads(1L) # nocov
    return(invisible()) # nocov
  }

  if (threads == 0 | threads > max)
  {
    LIDRTHREADS$input <- threads
    LIDRTHREADS$n <- max
    data.table::setDTthreads(0L)
  }
  else if(threads < 1)
  {
    LIDRTHREADS$input <- threads
    LIDRTHREADS$n <- as.integer(max*threads)
    data.table::setDTthreads(as.integer(max*threads))
  }
  else
  {
    LIDRTHREADS$input <- as.integer(threads)
    LIDRTHREADS$n <- as.integer(threads)
    data.table::setDTthreads(threads)
  }

  return(invisible())
}

#' @rdname set_lidr_threads
#' @export
get_lidr_threads = function()
{
  return(LIDRTHREADS$n)
}

#' Counts the number of cores used by simple multi-core future::plan()s
#'
#' lidR employs two different parallelization mechanisms: OpenMP for
#' multi-threaded execution of certain algorithms and the future API for
#' processing LAScatalog tiles in parallel. Since the future API can be used for
#' different kinds of parallelization (including multi-threading but also
#' parallel execution on computer clusters), it is possible for OpenMP and
#' future to have incompatible settings. Example: A user has a single machine
#' with four cores and sets both, the number of OpenMP threads and the number of
#' future processes to four. In this case OpenMP has to be disabled so that it
#' doesn't interfere with the future parallelization.
#'
#' Essentially, this function analyzes the argument to the 'workers' parameter
#' of future::plan(). When this argument is either 1) an integer, 2) a character
#' vector of strings all indicating the "localhost", or 3) a function call which
#' returns 1) or 2), then the integer or the number of localhost strings is
#' returned as the number of used cores. Currently, other cases cannot be
#' evaluated.
#'
#' @return A single integer giving the number of cores or NULL if the current
#'   future::plan() could not be evaluated. The latter might be the case when a
#'   future::plan() involves complex architectures such as remote computers or
#'   multiple nodes on a HPC.
#' @noRd
try_to_get_num_future_cores = function()
{
  if (!"future" %in% rownames(utils::installed.packages()))
    return(1L)

  # nocov start

  # get the current plan
  plan <- future::plan()

  # if the plan is explicitly single-threaded, return 1
  if (is(plan, "uniprocess")) { return(1L) }
  # "uniprocess" covers future::sequential and future::transparent and hopefully
  # also single-threaded plans provided by other packages

  # if there is no argument called "workers", return NULL
  if (!"workers" %in% names(formals(plan))) { return(NULL) }

  # get the value of the "workers" argument
  workers_arg <- formals(plan)$workers

  # These values might be used to reference the current machine and are compared
  # to the "workers" argument in the code below
  localhosts <- c("localhost", "127.0.0.1", Sys.info()[["nodename"]])
  # taken from the documentation of parallelly::makeClusterPSOCK and
  # parallelly::makeNodePSOCK hoping that they will also be valid for other
  # interfaces

  # Now check whether a number of locally used cores can be extracted from the
  # "workers" argument

  # if "workers" is a positive number, return that number
  if (is.numeric(workers_arg) &&
      length(workers_arg) == 1L &&
      !is.na(workers_arg) &&
      workers_arg >= 1)
  {
    return(as.integer(workers_arg))
  }

  # if "workers" is a character vector of "localhosts", return the vector length
  if (is.character(workers_arg) &&
      length(workers_arg) >= 1L &&
      all(!is.na(workers_arg)) &&
      all(workers_arg %in% localhosts))
  {
    return(length(workers_arg))
  }

  # if "workers" is a function call, check whether the return value is one of
  # the above tested options
  if (is.call(workers_arg))
  {
    # try to evaluate the function call
    evaluated_workers_arg <- tryCatch(
      eval(workers_arg),
      error = function(dummy) { return(NULL) } # on error: return NULL
    )

    if (is.null(evaluated_workers_arg)) # if the "normal" evaluation didn't work
    {
      # Try to evaluate the function call in the environment of the future
      # package. This deals with situations where functions cannot be evaluated
      # because the future package has not been loaded (using library(future),
      # library(lidR), or others).
      evaluated_workers_arg <- tryCatch(
        # use an arbitrary future function to get the environment of the future package
        eval(workers_arg, envir = environment(future::plan)),
        error = function(dummy) { return(NULL) }
      )
    }

    # if the call could not be evaluated, return NULL
    if (is.null(evaluated_workers_arg)) { return(NULL) }

    # If the call could be evaluated, check if the result is a number or a
    # vector of localhosts.
    if (is.numeric(evaluated_workers_arg) &&
        length(evaluated_workers_arg) == 1L &&
        !is.na(evaluated_workers_arg) &&
        evaluated_workers_arg >= 1)
    {
      return(as.integer(evaluated_workers_arg))
    }

    if (is.character(evaluated_workers_arg) &&
        length(evaluated_workers_arg) >= 1L &&
        all(!is.na(evaluated_workers_arg)) &&
        all(evaluated_workers_arg %in% localhosts))
    {
      return(length(evaluated_workers_arg))
    }

  } # end if (is.call(workers_arg))

  # -> "workers" is neither a number, nor a character vector of localhosts,
  # nor a function call returning any of these values.
  return(NULL)

  # nocov end
}


# Because I made some typos and I renamed stuff and I did
# not check the code yet
getThread <- get_lidr_threads
getThreads <- get_lidr_threads
getWorkers <- try_to_get_num_future_cores
setThreads <- set_lidr_threads

must_disable_openmp = function()
{
  if (!engine_use_future()) return(FALSE)
  if (isTRUE(getOption("lidR.threads.manual"))) return(FALSE) # Backward compatibility
  if (isFALSE(getOption("lidR.check.nested.parallelism"))) return(FALSE)

  workers    <- getWorkers()
  threads    <- getThreads()
  cores      <- parallelly::availableCores()

  if (is.null(workers))
  {
    warning("The parallel evaluation strategy was not recognized and lidR does not know if OpenMP should be disabled.
      OpenMP has been disabled by security.
      Use options(lidR.check.nested.parallelism = FALSE) and set_lidr_threads() for a fine control of parallelism.", call. = FALSE)
    return(TRUE)
  }

  if (workers * threads > cores)
  {
    # nocov because tested with a single core on CRAN
    verbose(glue::glue("Cannot nest {workers} future threads and {threads} OpenMP threads. Precedence given to future: OpenMP threads set to 1.")) # nocov
    return(TRUE)
  }

  return(FALSE)
}

fit_circle_on_3_points <- function(points_subset)
{
  stopifnot(nrow(points_subset) == 3L, ncol(points_subset) > 2L)

  # Extract the coordinates
  x1 <- points_subset[1, 1]
  y1 <- points_subset[1, 2]
  x2 <- points_subset[2, 1]
  y2 <- points_subset[2, 2]
  x3 <- points_subset[3, 1]
  y3 <- points_subset[3, 2]

  # Calculate the coefficients for the linear system
  A <- 2 * (x2 - x1)
  B <- 2 * (y2 - y1)
  C <- x2^2 + y2^2 - x1^2 - y1^2
  D <- 2 * (x3 - x1)
  E <- 2 * (y3 - y1)
  G <- x3^2 + y3^2 - x1^2 - y1^2

  # Solve for a and b using Cramer's rule
  denominator <- A * E - B * D
  if (denominator == 0)
  {
    return(c(0, 0, 0))
  }
  a <- (C * E - B * G) / denominator
  b <- (A * G - C * D) / denominator

  # Calculate the radius
  r <- sqrt((x1 - a)^2 + (y1 - b)^2)

  # Return the center and radius
  return(c(a, b, r))
}

# fit_circle_on_3_points <- function(points_subset)
# {
#   stopifnot(nrow(points_subset) == 3L, ncol(points_subset) > 2L)
#
#   # Initial guesses
#   x_mean <- mean(points_subset[, 1])
#   y_mean <- mean(points_subset[, 2])
#   r_guess <- sqrt(mean((points_subset[, 1] - x_mean)^2 + (points_subset[, 2] - y_mean)^2))
#   initial_guess <- c(x_mean, y_mean, r_guess)
#
#   # Optimize circle parameters
#   fit <- optim(par = initial_guess, fn = residuals_circle, points = points_subset)
#   return(fit$par)
# }

#' Fits 2D Circles on a Point Cloud
#'
#' Fits a 2D horizontally-aligned circle to a set of 3D points. The method uses RANSAC-based fitting
#' for robust estimation. The function returns a list with the circle parameters and additional data
#' to help determine whether the points form a circular shape. While it is always possible to fit a
#' circle to any dataset, the additional information assists in evaluating if the data genuinely
#' represent a circular pattern. The root mean square error (RMSE) may not always be a reliable metric
#' for assessing the quality of the fit due to non-circular elements (such as branches) that can arbitrarily
#' increase the RMSE of an otherwise well-fitted circle, as shown in the example below. Therefore, the
#' function also returns the angular range of the data, which indicates the spread of the inlier points:
#' 360 degrees suggests a full circle, 180 degrees indicates a half-circle, or a smaller range may
#' suggest a partial arc.
#'
#' @param points A LAS object representing a clustered slice, or an nx3 matrix of point coordinates.
#' @param num_iterations The number of iterations for the RANSAC fitting algorithm.
#' @param inlier_threshold A threshold value; points are considered inliers if their residuals are
#' below this value.
#'
#' @return A list containing the circle parameters and additional information for determining whether
#' the data fit a circular shape:
#' - `center_x`, `center_y`, `radius`: parameters of the fitted circle.
#' - `z`: average elevation of the circle.
#' - `rmse`: root mean square error between the circle and all points.
#' - `covered_arc_degree`: angular sector covered by inlier points.
#' - `percentage_inlier`: percentage of points that are inliers
#' - `percentage_inside`: percentage of points inside the circle
#' - `inliers`: IDs of the inlier points.
#' @md
#' @export
#' @examples
#' LASfile <- system.file("extdata", "dbh.laz", package="lidR")
#' las <- readTLS(LASfile, select = "xyz")
#' xyz = sf::st_coordinates(las)
#' circle = fit_circle(xyz)
#' plot(xyz, asp = 1, pch = 19, cex = 0.1)
#' symbols(circle$center_x, circle$center_y, circles = circle$radius,
#'   add = TRUE, fg = "red", inches = FALSE)
#'
#' trunk = las[circle$inliers]
#'
#' # Fitting a half-circle
#' half = xyz[xyz[,1] > 101.45,]
#' circle = fit_circle(half)
#' plot(half, asp = 1, pch = 19, cex = 0.1)
#' symbols(circle$center_x, circle$center_y, circles = circle$radius,
#'   add = TRUE, fg = "red", inches = FALSE)
#' circle$covered_arc_degree
fit_circle <- function(points, num_iterations = 100, inlier_threshold = 0.01)
{
  best_circle <- NULL
  max_inliers <- 0

  if (is(points, "LAS")) points = sf::st_coordinates(points)

  stopifnot(is.matrix(points), ncol(points) == 3L, nrow(points) > 3L)

  z = points[, 3]

  for (i in 1:num_iterations)
  {
    # Randomly sample points
    sample_indices <- sample(1:nrow(points), 3L)
    points_subset <- points[sample_indices, ]

    params <- fit_circle_on_3_points(points_subset)

    # Compute residuals for all points
    distances <- sqrt((points[, 1] - params[1])^2 + (points[, 2] - params[2])^2)
    residuals <- abs(distances - params[3])

    # Count inliers (points whose residuals are below the threshold)
    inliers <- sum(residuals < inlier_threshold)

    # Update best model if more inliers are found
    if (inliers > max_inliers)
    {
      max_inliers <- inliers
      best_circle <- params
    }
  }

  if (is.null(best_circle))
  {
    center_x = mean(points[,1])
    center_y = mean(points[,2])
    radius = 0.001
    z = mean(z)
    angle_range_degrees = 0
    inlier_ratio = 0
    percentage_inside = 0
    inliers = rep(TRUE, nrow(points))
    cfqi = 0
  }
  else
  {
    center_x <- best_circle[1]
    center_y <- best_circle[2]
    radius   <- best_circle[3]

    # Goodness of fit
    distances <- sqrt((points[, 1] - center_x)^2 + (points[, 2] - center_y)^2)
    residuals <- abs(distances - radius)
    inliers <- residuals < inlier_threshold

    # Angular range
    angle_res = 3
    angles <- atan2(points[inliers, 2] - center_y, points[inliers, 1] - center_x)
    angles <- ifelse(angles < 0, angles + 2 * pi, angles)
    angles <- sort(angles*180/pi)
    rangles <- unique(round(angles/angle_res)*angle_res)

    angle_range_degrees = sum(diff(rangles) <= angle_res)*angle_res
    arc_score <- angle_range_degrees / 360

    # Inlier ratio
    inlier_ratio <- sum(inliers) / nrow(points)

    # Percentage of point inside the circle
    ninside = sum(distances < radius - inlier_threshold)
    percentage_inside = ninside/nrow(points)
    score_inside = 1-percentage_inside
  }

  return(list(center_x = center_x,
              center_y = center_y,
              radius = radius,
              z = mean(z),
              covered_arc_degree = angle_range_degrees,
              percentage_inlier = inlier_ratio,
              percentage_inside = percentage_inside,
              inliers = which(inliers)))
}

#' Write a .las or .laz file
#'
#' Write a \link[=LAS-class]{LAS} object into a binary .las or .laz file (compression
#' specified in filename)
#'
#' @param las an object of class LAS.
#' @param file character. A character string naming an output file.
#' @param index boolean. Also write a lax file to index the points in the files
#'
#' @return Nothing. This function is used for its side-effect of writing a file.
#'
#' @export
#'
#' @examples
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' las = readLAS(LASfile)
#' subset = clip_rectangle(las, 684850, 5017850, 684900, 5017900)
#' writeLAS(subset, tempfile(fileext = ".laz"))
writeLAS = function(las, file, index = FALSE)
{
  stopifnotlas(las)
  assert_is_a_string(file)
  assert_is_a_bool(index)

  if (is.empty(las)) stop("Cannot write a file with 0 point", call. = FALSE)

  file  = path.expand(file)
  islas = tools::file_ext(file) %in% c("las", "laz")

  if (!islas) stop(glue::glue("File(s) {file} not supported"), call. = FALSE)

  rlas::write.las(file, as.list(las@header), las@data)

  if (index) rlas::writelax(file)

  return(invisible(file))
}

context("writeLAS")

i = megaplot
ofile = paste0(tempfile(), ".las")

test_that("Test if I/O are equal", {
  writeLAS(i, ofile)
  o = readLAS(ofile)

  # Because those field are expepected to be different
  i@header@PHB["Generating Software"] <- NULL
  o@header@PHB["Generating Software"] <- NULL

  i@header@PHB["System Identifier"] <- NULL
  o@header@PHB["System Identifier"] <- NULL

  i@header@PHB["Number of variable length records"] <- NULL
  o@header@PHB["Number of variable length records"] <- NULL

  i@header@PHB["Offset to point data"] <- NULL
  o@header@PHB["Offset to point data"] <- NULL

  expect_equal(i@data, o@data)
  expect_equal(i@header@PHB, o@header@PHB)
})

test_that("writeLAS does not write empty point cloud", {
  o <- filter_poi(i, Z > 1000)
  expect_error(writeLAS(o, ofile), "Cannot write a file with 0 point")
})

test_that("writeLAS fails nicely with no LAS", {
  expect_error(writeLAS(1, ofile), "Argument is not a LAS object")
})

LIDRDSM <- "DigitalSurfaceModel"
LIDRSPI <- "SpatialInterpolation"
LIDRGND <- "GroundSegmentation"
LIDRITD <- "IndividualTreeDetection"
LIDRDEC <- "PointCloudDecimation"
LIDRSHP <- "ShapeDetection"
LIDRSNG <- "SnagsSegmentation"
LIDRITS <- "IndividualTreeSegmentation"
LIDRNIT <- "normalize_intensity"
LIDRTRK <- "SensorTracking"
LIDROUT <- "NoiseSegmentation"

LIDRCONTEXTDSM <- "rasterize_canopy"
LIDRCONTEXTSPI <- c("normalize_height", "rasterize_terrain", "p2r", "spatial_interpolation")
LIDRCONTEXTGND <- "classify_ground"
LIDRCONTEXTOUT <- "classify_noise"
LIDRCONTEXTITD <- "locate_trees"
LIDRCONTEXTDEC <- "decimate_points"
LIDRCONTEXTSHP <- "segment_shapes"
LIDRCONTEXTSNG <- "segment_snags"
LIDRCONTEXTITS <- "segment_trees"
LIDRCONTEXTNIT <- "normalize_intensity"
LIDRCONTEXTTRK <- "track_sensor"

LIDRALGORITHM <- "lidRAlgorithm"
LIDRALGORITHMGENERIC = c(LIDRALGORITHM, "ANY", "function")
LIDRALGORITHMOPENMP <- "omp"
LIDRALGORITHMPOINTCLOUDBASED <- "PointCloudBased"
LIDRALGORITHMRASTERBASED <- "RasterBased"
LIDRALGORITHMDEC <- LIDRALGORITHMDSM <- LIDRALGORITHMGND <- LIDRALGORITHMITD <- LIDRALGORITHMITS <- LIDRALGORITHMNIT <- LIDRALGORITHMSNG <- LIDRALGORITHMTRK <- LIDRALGORITHMSHP <- LIDRALGORITHMSPI <- LIDRALGORITHMOUT <- LIDRALGORITHMGENERIC
LIDRALGORITHMDEC[2] <- LIDRDEC
LIDRALGORITHMDSM[2] <- LIDRDSM
LIDRALGORITHMGND[2] <- LIDRGND
LIDRALGORITHMITD[2] <- LIDRITD
LIDRALGORITHMITS[2] <- LIDRITS
LIDRALGORITHMNIT[2] <- LIDRNIT
LIDRALGORITHMSNG[2] <- LIDRSNG
LIDRALGORITHMTRK[2] <- LIDRTRK
LIDRALGORITHMSHP[2] <- LIDRSHP
LIDRALGORITHMSPI[2] <- LIDRSPI
LIDRALGORITHMOUT[2] <- LIDROUT

plugin <- function(f, class_t, omp = FALSE)
{
  if (omp) class_t <- c(class_t, LIDRALGORITHMOPENMP)
  class(f) <- class_t
  return(f)
}


#' Plugin system
#'
#' Tools to build plugin functions for lidR
#'
#' @param f a function
#' @param omp logical is the function natively parallized with OpenMP
#' @param raster_based logical. For ITS and ITD algorithms, is the method raster-based or
#' or point-cloud-based?
#'
#' @name plugins
#' @rdname plugins
#' @examples
#' \dontrun{
#' mba <- function(n = 1, m = 1, h = 8, extend = TRUE) {
#'   f <- function(las, where) {
#'     res <- MBA::mba.points(las@data, where, n, m , h, extend)
#'     return(res$xyz.est[,3])
#'   }
#'
#'   f <- plugin_dtm(f)
#'   return(f)
#' }
#'
#' LASfile <- system.file("extdata", "Topography.laz", package="lidR")
#' las = readLAS(LASfile)
#'
#' dtm = rasterize_terrain(las, algorithm = mba())
#' }
NULL

#' @export
#' @rdname plugins
plugin_dsm      <- function(f, omp = FALSE) return(plugin(f, LIDRALGORITHMDSM, omp))

#' @export
#' @rdname plugins
plugin_dtm      <- function(f, omp = FALSE) return(plugin(f, LIDRALGORITHMSPI, omp))

#' @export
#' @rdname plugins
plugin_gnd      <- function(f, omp = FALSE) return(plugin(f, LIDRALGORITHMGND, omp))

#' @export
#' @rdname plugins
plugin_decimate <- function(f, omp = FALSE) return(plugin(f, LIDRALGORITHMDEC, omp))

#' @export
#' @rdname plugins
plugin_shape    <- function(f, omp = FALSE) return(plugin(f, LIDRALGORITHMSHP, omp))

#' @export
#' @rdname plugins
plugin_snag     <- function(f, omp = FALSE) return(plugin(f, LIDRALGORITHMSNG, omp))

#' @export
#' @rdname plugins
plugin_track    <- function(f, omp = FALSE) return(plugin(f, LIDRALGORITHMTRK, omp))

#' @export
#' @rdname plugins
plugin_nintensity    <- function(f, omp = FALSE) return(plugin(f, LIDRALGORITHMNIT, omp))

#' @export
#' @rdname plugins
plugin_outliers    <- function(f, omp = FALSE) return(plugin(f, LIDRALGORITHMOUT, omp))

#' @export
#' @rdname plugins
plugin_itd    <- function(f, omp = FALSE, raster_based = FALSE)
{
  if (isTRUE(raster_based))
    return(plugin(f, c(LIDRALGORITHMITD, LIDRALGORITHMRASTERBASED), omp))
  else
    return(plugin(f, c(LIDRALGORITHMITD, LIDRALGORITHMPOINTCLOUDBASED), omp))
}

#' @export
#' @rdname plugins
plugin_its    <- function(f, omp = FALSE, raster_based = FALSE)
{
  if (isTRUE(raster_based))
    return(plugin(f, c(LIDRALGORITHMITS, LIDRALGORITHMRASTERBASED), omp))
  else
    return(plugin(f, c(LIDRALGORITHMITS, LIDRALGORITHMPOINTCLOUDBASED), omp))
}

setClass(
  Class = "LAScluster",
  representation(
    center = "list",
    bbox   = "matrix",
    bbbox  = "matrix",
    width  = "numeric",
    height = "numeric",
    buffer = "numeric",
    shape  = "numeric",
    select = "character",
    filter = "character",
    wkt    = "character",
    files  = "character",
    name   = "character",
    save   = "character",
    alt_dir = "character",
    crs     = "crs",
    index   = "list"
  )
)

setMethod("initialize", "LAScluster", function(.Object, center, width, height, buffer, shape, files, name, wkt, crs, index)
{
  hw = width/2
  hh = height/2
  xc = center$x
  yc = center$y

  .Object@center <- center
  .Object@bbox   <- matrix(c(xc - hw, yc - hh, xc + hw, yc + hh), ncol = 2)
  .Object@bbbox  <- .Object@bbox + buffer * matrix(c(-1, -1, +1, +1), ncol = 2)
  .Object@width  <- width  + 2 * buffer
  .Object@height <- height + 2 * buffer
  .Object@buffer <- buffer
  .Object@shape  <- shape
  .Object@name   <- name
  .Object@files  <- files
  .Object@save   <- ""
  .Object@wkt    <- ""
  .Object@crs    <- crs
  .Object@index  <- index

  if (shape == LIDRCIRCLE)
    .Object@filter = paste("-inside_circle", xc, yc, hw + buffer)
  else if (shape == LIDRRECTANGLE)
    .Object@filter = paste("-inside", .Object@bbbox[1], .Object@bbbox[2], .Object@bbbox[3], .Object@bbbox[4])
  else
    stop("Something went wrong internally initializing a cluster. Process aborted.")

  return(.Object)
})

#' Read .las or .laz files
#'
#' Reads .las or .laz files into an object of class \link[=LAS-class]{LAS}. If several files are read at
#' once the returned LAS object is considered as one LAS file. The optional parameters enable the user
#' to save a substantial amount of memory by choosing to load only the attributes or points of interest.
#' LAS formats 1.0 to 1.4 are supported. Point Data Record Format 0 to 10 are supported.\cr\cr
#' `readLAS` is the original function and always works. Using one of the `readALS` or `readTLS` functions
#' adds information to the returned object to register a point-cloud type. Registering the correct point
#' type improves the performance of some functions by enabling users to select an appropriate spatial index.
#' See \link[=lidR-spatial-index]{spatial indexing}. Notice that by legacy and for backwards-compatibility reasons,
#' `readLAS()` and `readALS()` are equivalent because lidR was originally designed for ALS and thus the
#' original function `readLAS()` was (supposedly) used for ALS. Reading a TLS dataset with `readLAS()` instead
#' of `readTLS()` is perfectly valid but less powerful.
#'
#'
#' **Select:** the 'select' argument specifies the data that will actually be loaded. For example,
#' 'xyzia' means that the x, y, and z coordinates, the intensity and the scan angle will be loaded.
#' The supported entries are t - gpstime, a - scan angle, i - intensity, n - number of returns,
#' r - return number, c - classification, s - synthetic flag, k - keypoint flag, w - withheld flag,
#' o - overlap flag (format 6+), u - user data, p - point source ID, e - edge of flight line flag,
#' d - direction of scan flag, R - red channel of RGB color, G - green channel of RGB color,
#' B - blue channel of RGB color, N - near-infrared channel, C - scanner channel (format 6+),
#' W - Full waveform.
#' Also numbers from 1 to 9 for the extra bytes data numbers 1 to 9. 0 enables all extra bytes to be
#' loaded and '*' is the wildcard that enables everything to be loaded from the LAS file. \cr
#' Note that x, y, z are implicit and always loaded. 'xyzia' is equivalent to 'ia'.\cr\cr
#' **Filter:** the 'filter' argument allows filtering of the point cloud while reading files.
#' This is much more efficient than \link{filter_poi} in many ways. If the desired filters are known
#' before reading the file, the internal filters should always be preferred. The available filters are
#' those from `LASlib` and can be found by running the following command: `readLAS(filter = "-help")`.
#' (see also \link[rlas:read.las]{rlas::read.las}). From `rlas` v1.3.6 the transformation commands
#' can also be passed via the argument filter.
#'
#' @template section-fwf
#'
#' @param files characters. Path(s) to one or several a file(s). Can also be a
#' \link[=LAScatalog-class]{LAScatalog} object.
#' @param select character. Read only attributes of interest to save memory (see details).
#' @param filter character. Read only points of interest to save memory (see details).
#'
#' @return A LAS object
#'
#' @export
#'
#' @examples
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' las = readLAS(LASfile)
#' las = readLAS(LASfile, select = "xyz")
#' las = readLAS(LASfile, select = "xyzi", filter = "-keep_first")
#' las = readLAS(LASfile, select = "xyziar", filter = "-keep_first -drop_z_below 0")
#'
#' # Negation of attributes is also possible (all except intensity and angle)
#' las = readLAS(LASfile, select = "* -i -a")
#' @md
readLAS = function(files, select = "*", filter = "")
{
  if (filter == "-h" | filter == "-help")
  {
    rlas::help_filter()
    return(invisible())
  }

  UseMethod("readLAS", files)
}


#' Read a .las or .laz file header
#'
#' Reads a .las or .laz file header into an object of class \link[=LASheader-class]{LASheader}.
#' This function strictly reads the header while the function \link{readLAS} can alter the header to
#' fit the actual data loaded.
#'
#' @param file characters. Path to one file.
#' @return A LASheader object
#' @export
#' @examples
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' header = readLASheader(LASfile)
#'
#' print(header)
#' plot(header)
#'
#' \dontrun{
#' plot(header, mapview = TRUE)}
readLASheader = function(file)
{
  header <- rlas::read.lasheader(file)
  return(LASheader(header))
}

#' @export
readLAS.LAScatalog = function(files, select, filter)
{
  if (missing(select)) select <- opt_select(files)
  if (missing(filter)) filter <- opt_filter(files)

  return(readLAS(files@data$filename, select, filter))
}

#' @export
readLAS.LAScluster = function(files, select = NULL, filter = NULL)
{
  if (!is.null(select)) warning("Argument 'select' is not used with a LAScluster. Use opt_select() with the LAScatalog instead.", call. = FALSE)
  if (!is.null(filter)) warning("Argument 'filter' is not used with a LAScluster. Use opt_filter() with the LAScatalog instead.", call. = FALSE)

  if (!all(file_exists(files@files)) && all(files@alt_dir != "")) {
    for (alt_dir in files@alt_dir) {
      paths <- paste0(alt_dir, basename(files@files))
      if (all(file_exists(paths))) break
    }
    files@files <- paths
  }

  if (!all(file_exists(files@files))) stop("File not found", call. = FALSE)

  buffer <- X <- Y <- NULL

  las <- readLAS(files@files, files@select, files@filter)
  st_crs(las) <- st_crs(files)
  las@index <- files@index

  las@data[["buffer"]] <- rep(LIDRNOBUFFER, npoints(las))

  if (files@buffer > 0)
  {
    ext     <- st_bbox(files)
    ybottom <- ext$ymin
    ytop    <- ext$ymax
    xleft   <- ext$xmin
    xright  <- ext$xmax
    xc      <- files@center$x
    yc      <- files@center$y
    r       <- (files@width - 2*files@buffer)/2

    if (files@shape == LIDRCIRCLE)
    {
      las@data[(X - xc)^2 + (Y - yc)^2 > r^2, buffer := LIDRBUFFER]
    }
    else
    {
      las@data[Y < ybottom, buffer := LIDRBOTTOMBUFFER]
      las@data[X < xleft,   buffer := LIDRLEFTBUFFER]
      las@data[Y >= ytop,    buffer := LIDRTOPBUFFER]
      las@data[X >= xright,  buffer := LIDRRIGHTBUFFER]
      las@data[(X >= xright) & (Y < ybottom), buffer := LIDRBOTTOMBUFFER]
    }

    # We found a region with no actual data. The points all belong in the buffer
    # Return empty point cloud
    if (fast_countequal(las@data[["buffer"]], LIDRNOBUFFER) == 0)
      las <- filter_poi(las, buffer == LIDRNOBUFFER)
  }

  return(las)
}

#' @export
readLAS.character = function(files, select = "*", filter = "")
{
  assert_is_a_string(select)
  assert_is_a_string(filter)
  return(streamLAS(files, ofile = "", select, filter))
}

streamLAS = function(x, ofile, select = "*", filter = "", filter_wkt = "")
{
  UseMethod("streamLAS", x)
}

streamLAS.LAScluster = function(x, ofile, select = "*", filter = "", filter_wkt = "")
{
  las = streamLAS(x@files, ofile, x@select, x@filter, filter_wkt)

  # Automatically remove files that are empty
  if (ofile != "")
  {
    header = rlas::read.lasheader(ofile)

    if (header$`Number of point records` == 0)
    {
      file.remove(ofile)
      return(NULL)
    }
  }

  return(las)
}

streamLAS.character = function(x, ofile, select = "*", filter = "", filter_wkt = "")
{
  # Compatibility with rlas that no longer supports WKT string to filter polygons
  geom = list()
  if (utils::packageVersion("rlas") >= "1.8.0")
  {
    if (is.character(filter_wkt) && filter_wkt[1] != "")
    {
      geom = sf::st_as_sfc(filter_wkt)
      geom = extract_polygons(geom)
    }
  }

  assert_all_are_existing_files(x)

  islas <- tools::file_ext(x) %in% c("las", "laz", "ply", "LAS", "LAZ", "PLY")

  if (any(!islas)) stop("File(s) are not las or laz")

  ifiles <- normalizePath(x, mustWork = FALSE)

  if (ofile != "")
  {
    dir = dirname(ofile)
    if (!dir.exists(dir)) dir.create(dir, recursive = TRUE)
  }

  header <- rlas::read.lasheader(ifiles[1])

  # More than one file we need to check the compatibility
  if (length(ifiles) > 1)
  {
    for (file in ifiles)
    {
      temp.header <- rlas::read.lasheader(file)

      if (temp.header[["Point Data Format ID"]] != header[["Point Data Format ID"]] )
        warning("Different files have different Point Data Format ID and are incompatible. Some attribute may have been discarded or zeroed", call. = FALSE)

      if (temp.header[["X scale factor"]] != header[["X scale factor"]] )
        warning("Different files have different X scale factors and are incompatible. The first file has precedence and data were rescaled.", call. = FALSE)

      if (temp.header[["Y scale factor"]] != header[["Y scale factor"]] )
        warning("Different files have different Y scale factors and are incompatible. The first file has precedence and data were rescaled.", call. = FALSE)

      if (temp.header[["Z scale factor"]] != header[["Z scale factor"]] )
        warning("Different files have different Z scale factors and are incompatible. The first file has precedence and data were rescaled.", call. = FALSE)

      if (rlas::header_get_epsg(header) != rlas::header_get_epsg(temp.header))
        warning("Different files have diferent CRS and are incompatible. The CRS of the first file has been retained.", call. = FALSE)
    }
  }

  if (utils::packageVersion("rlas") >= "1.8.0") {
    data <- rlas::read_and_write.las(ifiles, ofile, select, filter, geom)
  } else {
    data <- rlas::read_and_write.las(ifiles, ofile, select, filter, filter_wkt)
  }

  if (is.null(data))
    return(invisible())

  if (nrow(data) > 0)
  {
    # If the number of files read is > 1 header bbox will not be in accordance with the data. Update the header.
    if (length(ifiles) > 1 || filter != "")
      header <- rlas::header_update(header, data)
  }

  # Remove extrabytes in the header if not loaded

  extrabytes <- names(header[["Variable Length Records"]][["Extra_Bytes"]][["Extra Bytes Description"]])

  if (!is.null(extrabytes))
  {
    for (extrabyte in extrabytes)
    {
      if (!extrabyte %in% names(data))
      {
        header[["Variable Length Records"]][["Extra_Bytes"]][["Extra Bytes Description"]][[extrabyte]] <- NULL
      }
    }

    if (length(header[["Variable Length Records"]][["Extra_Bytes"]][["Extra Bytes Description"]]) == 0)
    {
      header[["Variable Length Records"]][["Extra_Bytes"]] <- NULL
    }
  }

  return(LAS(data, header, check = TRUE, index = NULL, no_attr_name_check = TRUE))
}

is_remote = function(paths)
{
  remote_pattern <- "^https?://|^/vsi(curl|s3|gs|az|adls|oss|swift)/"
  is_remote <- grepl(remote_pattern, paths)
}

file_exists <- function(...)
{
  paths <- c(...)
  is_remote = is_remote(paths)
  result <- ifelse(is_remote, TRUE, file.exists(paths))
  result
}

test_that("interpret waveforn works", {
  skip_on_cran()

  LASfile <- system.file("extdata", "fwf.laz", package="rlas")
  fwf <- readLAS(LASfile)
  expect_error(las <- interpret_waveform(fwf), NA)
  expect_equal(npoints(las), 455168)
})

context("las_check")

las0    <- example
las1 <- las0
las1@data <- data.table::copy(las0@data)

#las1@data[1, X := NA_real_]
las1@data[1, Z := -5]
las1@data[5,] = las1@data[6,]
las1@data[7,1:2] = las1@data[6, 1:2]
las1@data[5:7, Classification := 2L]
las1@data[1, Withheld_flag := TRUE]
las1@data[1, Synthetic_flag := TRUE]
las1@data[1, Keypoint_flag := TRUE]
las1@data[25, Z := 1234567890.1]
las1@data[8:9, Intensity := c(NA_integer_, 12345679L)]
las1@data$gpstime <- 0
las1@crs <- sf::st_crs(26917)
las1@header@PHB[["X scale factor"]] <- 0.123
las1@header@PHB[["Y scale factor"]] <- 0.123
las1@header@PHB[["Z scale factor"]] <- 0.123
las1@header@PHB[["Point Data Format ID"]] <- 25


LASfile <- system.file("extdata", "extra_byte.laz", package = "rlas")
las2    <- readLAS(LASfile, select = "xyz")
las2@header@PHB$`Global Encoding`$WKT = TRUE

wkt(las2) <- "PROJCS[\"unknown\",GEOGCS[\"unknown\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199433,AUTHORITY[\"EPSG\",\"9122\"]]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\",-81],PARAMETER[\"scale_factor\",0.9996],PARAMETER[\"false_easting\",500000],PARAMETER[\"false_northing\",0],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"Easting\",EAST],AXIS[\"Northing\",NORTH]]"

las2@crs <- sf::NA_crs_

las3 <- las1
epsg(las3) <- 2008
las3@crs <- sf::st_crs(26917)

las4 = las2
las4@crs <- sf::st_crs(2008)

ctg0 <- example_ctg

LASfile <- system.file("extdata", "", package = "lidR")
ctg1 <- readLAScatalog(LASfile)

ctg2 <- ctg1
ctg2@data$X.scale.factor <- 0.012
ctg2@data$Y.scale.factor <- 0.012
ctg2@data$Z.scale.factor <- 0.012
ctg2@data$Min.Z <- -1
ctg2@data$Point.Data.Format.ID = 12L

test_that("las_check works without error with LAS", {
  sink(tempfile())
  expect_error(las_check(las0), NA)
  expect_error(las_check(las1), NA)
  expect_error(las_check(las2), NA)
  expect_error(las_check(las3), NA)
  expect_error(las_check(las4), NA)
  sink(NULL)
})

test_that("las_check works without error with LAScatalog", {
  sink(tempfile())
  expect_error(las_check(ctg0), NA)
  expect_error(las_check(ctg1), NA)
  expect_error(las_check(ctg2), NA)
  sink(NULL)
})

test_that("las_check works without error with LAScluster", {
  sink(tempfile())
  expect_error(las_check(chunk), NA)
  sink(NULL)
})

test_that("las_check CRS specific test", {

  sink(tempfile())

  las1 <- las0
  epsg(las1) <- 2008
  las1@header@PHB$`Global Encoding`$WKT <- TRUE

  las2 <- las0
  las2@header@PHB$`Global Encoding`$WKT = TRUE

  wkt(las2) <- "PROJCS[\"RD_New\",GEOGCS[\"GCS_Amersfoort\",DATUM[\"D_Amersfoort\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Double_Stereographic\"],PARAMETER[\"False_Easting\",155000.0],PARAMETER[\"False_Northing\",463000.0],PARAMETER[\"Central_Meridian\",5.38763888888889],PARAMETER[\"Scale_Factor\",0.9999079],PARAMETER[\"Latitude_Of_Origin\",52.1561605555556],UNIT[\"Meter\",1.0]]"

  las2@header@PHB$`Global Encoding`$WKT <- FALSE

  expect_error(las_check(las1), NA)
  expect_error(las_check(las2), NA)

  epsg(las0) <- 2008
  las0@crs <- sf::NA_crs_

  expect_error(las_check(las0), NA)

  las0@header@VLR$GeoKeyDirectoryTag$tags[[1]]$`value offset` <- 200800

  expect_error(las_check(las0), NA)

  las2@header@VLR$`WKT OGC CS`$`WKT OGC COORDINATE SYSTEM` <- "INVALID"
  las2@header@PHB$`Global Encoding`$WKT <- TRUE

  expect_error(las_check(las2), NA)

  sink(NULL)
})

test_that("las_check quantization specific test", {

  sink(tempfile())

  x = las0@data[["X"]][5]
  y = las0@data[["Y"]][5]
  z = las0@data[["Z"]][5]

  las0@header@PHB$`Min X` <- las0@header@PHB$`Min X` + 0.0001
  las0@header@PHB$`Max Y` <- las0@header@PHB$`Max Y` + 0.0001
  las0@header@PHB$`Max Z` <- las0@header@PHB$`Max Z` + 0.0001
  las0@data[["X"]][5] <- x + 0.00123
  las0@data[["Y"]][5] <- y + 0.000123
  las0@data[["Z"]][5] <- z + 0.0000123

  expect_error(las_check(las0), NA)

  sink(NULL)
})

test_that("las_check performs a deep inspection of a LAScatalog", {

  ctg = lidR:::catalog_generator(2, 20)
  o1 = las_check(ctg, print = FALSE, deep = TRUE)

  expect_equal(length(o1), 2L)
})

test_that("las_check returns a list of troubleshooting", {

  report = las_check(las1, FALSE)

  expect_is(report, "list")
  expect_equal(names(report), c("messages", "warnings", "errors"))

  report = las_check(ctg1, FALSE)

  expect_is(report, "list")
  expect_equal(names(report),  c("messages", "warnings", "errors"))
})

context("spTransform")

LASfile <- system.file("extdata", "Megaplot.laz", package = "lidR")
las <- readLAS(LASfile, select = "xyz")

crs  <- sf::st_crs(26918)
las2 <- sf::st_transform(las, crs)

test_that("Datum tranformation works", {

  expect_equal(mean(las$X), 684879,  tol = 1)
  expect_equal(mean(las$Y), 5017900, tol = 1)
  expect_equal(mean(las2$X), 214383,  tol = 1)
  expect_equal(mean(las2$Y), 5021652, tol = 1)
})

test_that("Bounding box is updated", {

  expect_equivalent(as.numeric(st_bbox(las)), c(684766.39, 5017773.08, 684993.29, 5018007.25))
  expect_equivalent(as.numeric(st_bbox(las2)), c(214261.66, 5021517.24, 214504.91, 5021767.46), tol = 0.01)
})

test_that("Bounding header is updated", {

  expect_equivalent(unlist(las@header@PHB[c("Min X", "Min Y", "Max X", "Max Y")]), c(684766.39, 5017773.08, 684993.29, 5018007.25))
  expect_equivalent(unlist(las2@header@PHB[c("Min X", "Min Y", "Max X", "Max Y")]), c(214261.66, 5021517.24, 214504.91, 5021767.46), tol = 0.1)
})

#' @export
#' @rdname classify
classify_ground = function(las, algorithm, last_returns = TRUE)
{
  UseMethod("classify_ground", las)
}

#' @export
classify_ground.LAS = function(las, algorithm, last_returns = TRUE)
{
  assert_is_algorithm(algorithm)
  assert_is_algorithm_gnd(algorithm)

  filter <- TRUE
  if (last_returns) {
    if (!all(c("ReturnNumber", "NumberOfReturns") %in% names(las))) {
      warning("'ReturnNumber' and/or 'NumberOfReturns' not found. Cannot use the option 'last_returns', all the points will be used.", call. = FALSE)
    } else {
      filter <- parse_filter(las, ~ReturnNumber == NumberOfReturns)
      if (sum(filter) == 0) {
        warning("Zero last return found. Cannot use the option 'last_returns', all the points will be used.", call. = FALSE)
        filter <- TRUE
      }
    }
  }

  lidR.context <- "classify_ground"
  idx <- algorithm(las, filter)

  if ("Classification" %in% names(las))
  {
    nground <- fast_countequal(las@data[["Classification"]], 2L)

    if (nground > 0)
    {
      message(glue::glue("Original dataset already contains {nground} ground points. These points were reclassified as 'unclassified' before performing a new ground classification."))
      new_classes <- las@data[["Classification"]]
      new_classes[new_classes == LASGROUND] <- LASUNCLASSIFIED
    }
    else
    {
      new_classes <- las@data[["Classification"]]
    }
  }
  else
    new_classes <- rep(LASUNCLASSIFIED, npoints(las))

  new_classes[idx] <- LASGROUND
  las@data[["Classification"]] <- new_classes
  return(las)
}

#' @export
classify_ground.LAScatalog = function(las, algorithm, last_returns = TRUE)
{
  opt_select(las) <- "*"
  options <- list(need_buffer = TRUE, drop_null = TRUE, need_output_file = TRUE)
  output  <- catalog_map(las, classify_ground, algorithm = algorithm,  last_returns = last_returns, .options = options)
  return(output)
}

#' Subset a LAScatalog
#'
#' Subset a LAScatalog interactively using the mouse. Subset a LAScatalog with a spatial object to
#' keep only the tiles of interest. It can be used to select tiles of interest that encompass spatial
#' objects.
#'
#' @param ctg A \link[=LAScatalog-class]{LAScatalog}
#' @param y  `bbox`, `sf`, `sfc`, `Extent`, `Raster*`, `Spatial*` objects
#' @param ... ignored
#' @param mapview logical. If \code{FALSE}, use R base plot instead of mapview (no pan, no zoom, see
#' also \link[=plot]{plot})
#' @param subset character. By default it subsets the collection It is also possible to flag
#' the files to maintain the collection as a whole but process only a subset of its content.
#' \code{flag_unprocessed} flags the files that will not be processed.
#' \code{flag_processed} flags the files that will be processed.
#'
#' @return A LAScatalog object
#'
#' @export
#'
#' @examples
#' \dontrun{
#' ctg = readLAScatalog("<Path to a folder containing a set of .las files>")
#' new_ctg = catalog_select(ctg)
#' }
#' @name catalog_subset
#' @rdname catalog_subset
# nocov start
catalog_select = function(ctg, mapview = TRUE, subset = c("subset", "flag_unprocessed", "flag_processed"))
{
  assert_is_all_of(ctg, "LAScatalog")
  assert_is_a_bool(mapview)
  method <- match.arg(subset)

  if (is_lascatalog_v3(ctg)) catalog <- lascatalog_v3_repair(ctg)

  Min.X <- Min.Y <- Max.X <- Max.Y <- NULL

  if (mapview & (!requireNamespace("mapview", quietly = TRUE) | !requireNamespace("mapedit", quietly = TRUE)))
  {
    message("Package 'mapview' and 'mapedit' are needed. Function switched to R base plot mode.")
    mapview = FALSE
  }

  if (mapview)
  {
    index <- mapedit::selectFeatures(as.spatial(ctg),index = TRUE)
  }
  else
  {
    plot(ctg, mapview = FALSE)
    index <- with(ctg@data, identify_tile(Min.X, Max.X, Min.Y, Max.Y))
  }

  if (method == "subset") {
    ctg <- ctg[index,]
  } else if (method == "flag_unprocessed") {
    ctg$processed <- TRUE
    ctg$processed[index] <- FALSE
  } else {
    ctg$processed <- FALSE
    ctg$processed[index] <- TRUE
  }

  return(ctg)
}

identify_tile <- function(minx, maxx, miny, maxy, plot = FALSE, ...)
{
  n <- length(minx)
  x <- (minx + maxx)/2
  y <- (miny + maxy)/2

  sel <- rep(FALSE, n)

  while (sum(sel) < n)
  {
    ans <- graphics::identify(x[!sel], y[!sel], n = 1, plot = FALSE, ...)

    if (!length(ans))
      break

    ans <- which(!sel)[ans]

    graphics::rect(minx[ans], miny[ans], maxx[ans], maxy[ans], col = "forestgreen")

    sel[ans] <- TRUE
  }

  return(which(sel))
}
# nocov end

#' @export
#' @rdname classify
classify_noise = function(las, algorithm)
{
  UseMethod("classify_noise", las)
}

#' @export
classify_noise.LAS = function(las, algorithm)
{
  assert_is_algorithm(algorithm)
  assert_is_algorithm_out(algorithm)

  lidR.context <- "classify_noise"
  idx <- algorithm(las)

  if ("Classification" %in% names(las))
  {
    nnoise <- fast_countequal(las@data[["Classification"]], LASNOISE)

    if (nnoise > 0)
    {
      message(glue::glue("Original dataset already contains {nnoise} noise points. These points were reclassified as 'unclassified' before performing a new noise classification."))
      new_classes <- las@data[["Classification"]]
      new_classes[new_classes == LASNOISE] <- LASUNCLASSIFIED
    }
    else
    {
      new_classes <- las@data[["Classification"]]
    }
  }
  else
    new_classes <- rep(LASUNCLASSIFIED, npoints(las))

  new_classes[idx] <- LASNOISE
  las@data[["Classification"]] <- new_classes
  return(las)
}

#' @export
classify_noise.LAScatalog = function(las, algorithm)
{
  opt_select(las) <- "*"
  options <- list(need_buffer = TRUE, drop_null = TRUE, need_output_file = TRUE)
  output  <- catalog_map(las, classify_noise, algorithm = algorithm, .options = options)
  return(output)
}

#' Ground Segmentation Algorithm
#'
#' This function is made to be used in \link{classify_ground}. It implements an
#' algorithm for segmentation of ground points base on a Multiscale Curvature
#' Classification. This method is a strict implementation of the MCC algorithm
#' made by Evans and Hudak. (2007) (see references) that relies on the authors'
#' original source code written and exposed to R via the the \code{RMCC} package.
#'
#' There are two parameters that the user must define, the scale (s) parameter and
#' the curvature threshold (t). The optimal scale parameter is a function of
#' 1) the scale of the objects (e.g., trees) on the ground, and 2) the sampling
#' interval (post spacing) of the lidar data. Current lidar sensors are capable
#' of collecting high density data (e.g., 8 pulses/m2) that translate to a spatial
#' sampling frequency (post spacing) of 0.35 m/pulse (1/sqrt(8 pulses/m2) = 0.35 m/pulse),
#' which is small relative to the size of mature trees and will oversample larger
#' trees (i.e., nominally multiple returns/tree). Sparser lidar data (e.g., 0.25 pulses/m2)
#' translate to a post spacing of 2.0 m/pulse (1/sqrt(0.25 pulses/m2) = 2.0 m/pulse),
#' which would undersample trees and fail to sample some smaller trees (i.e.,
#' nominally <1 return/tree).\cr\cr
#' Therefore, a bit of trial and error is warranted to determine the best scale
#' and curvature parameters to use. Select a las tile containing a good variety
#' of canopy and terrain conditions, as it's likely the parameters that work best
#' there will be applicable to the rest of your project area tiles. As a starting
#' point: if the scale (post spacing) of the lidar survey is 1.5 m, then try 1.5.
#' Try varying it up or down by 0.5 m increments to see if it produces a more desirable
#' digital terrain model (DTM) interpolated from the classified ground returns in
#' the output file. Use units that match the units of the lidar data. For example,
#' if your lidar data were delivered in units of feet with a post spacing of 3 ft,
#' set the scale parameter to 3, then try varying it up or down by 1 ft increments
#' and compare the resulting interpolated DTMs. If the trees are large, then
#' it's likely that a scale parameter of 1 m (3 ft) will produce a cleaner DTM
#' than a scale parameter of 0.3 m (1 ft), even if the pulse density is 0.3 m
#' (1 ft). As for the curvature threshold, a good starting value to try might be
#' 0.3 (if data are in meters; 1 if data are in feet), and then try varying this
#' up or down by 0.1 m increments (if data are in meters; 0.3 if data are in feet).
#'
#' @param s numeric. Scale parameter. The optimal scale parameter is a function of
#' 1) the scale of the objects (e.g., trees) on the ground, and 2) the sampling
#' interval (post spacing) of the lidar data.
#' @param t numeric. Curvature threshold
#'
#' @references Evans, Jeffrey S.; Hudak, Andrew T. 2007. A multiscale curvature
#' algorithm for classifying discrete return LiDAR in forested environments.
#' IEEE Transactions on Geoscience and Remote Sensing. 45(4): 1029-1038.
#'
#' @family ground segmentation algorithms
#'
#' @examples
#' if (require(RMCC, quietly = TRUE))
#' {
#' LASfile <- system.file("extdata", "Topography.laz", package="lidR")
#' las <- readLAS(LASfile, select = "xyzrn", filter = "-inside 273450 5274350 273550 5274450")
#'
#' las <- classify_ground(las, mcc(1.5,0.3))
#' #plot(las, color = "Classification")
#' }
#' @export
#' @name gnd_mcc
mcc <- function(s = 1.5, t = 0.3)
{
  s <- lazyeval::uq(s)
  t <- lazyeval::uq(t)

  assert_is_a_number(s)
  assert_is_a_number(t)
  assert_package_is_installed("RMCC")

  f = function(las, filter)
  {
    . <- X <- Y <- Z <- idx <- NULL
    assert_is_valid_context(LIDRCONTEXTGND, "mcc")
    las@data[["idx"]] <- 1:npoints(las)
    cloud <- las@data
    if (length(filter) > 1) cloud <- cloud[filter, .(X,Y,Z, idx)]
    gnd <- RMCC::MCC(cloud, s, t)
    ids <- cloud$idx[gnd]
    return(ids)
  }

  f <- plugin_gnd(f)
  return(f)
}

#' Ground Segmentation Algorithm
#'
#' This function is made to be used in \link{classify_ground}. It implements an algorithm for segmentation
#' of ground points based on a progressive morphological filter. This method is an implementation of
#' the Zhang et al. (2003) algorithm (see reference). Note that this is not a strict implementation
#' of Zhang et al. This algorithm works at the point cloud level without any rasterization process.
#' The morphological operator is applied on the point cloud, not on a raster. Also, Zhang et al.
#' proposed some formulas (eq. 4, 5 and 7) to compute the sequence of windows sizes and thresholds.
#' Here, these parameters are free and specified by the user. The function \link{util_makeZhangParam}
#' enables computation of the parameters according to the original paper.
#'
#' The progressive morphological filter allows for any sequence of parameters. The `util_makeZhangParam`
#' function enables computation of the sequences using equations (4),
#'  (5) and (7) from Zhang et al. (see reference and details).
#' @details
#' In the original paper the windows size sequence is given by eq. 4 or 5:\cr\cr
#' \eqn{w_k = 2kb + 1} \cr\cr
#' or\cr\cr
#' \eqn{w_k = 2b^k + 1}\cr\cr
#' In the original paper the threshold sequence is given by eq. 7:\cr\cr
#' \eqn{th_k = s*(w_k - w_{k-1})*c + th_0}\cr\cr
#' Because the function \link{classify_ground} applies the morphological operation at the point
#' cloud level the parameter \eqn{c} is set to 1 and cannot be modified.
#'
#' @param b numeric. This is the parameter \eqn{b} in Zhang et al. (2003) (eq. 4 and 5).
#' @param max_ws numeric. Maximum window size to be used in filtering ground returns. This limits
#' the number of windows created.
#' @param dh0 numeric. This is \eqn{dh_0} in Zhang et al. (2003) (eq. 7).
#' @param dhmax numeric. This is \eqn{dh_{max}} in Zhang et al. (2003) (eq. 7).
#' @param s numeric. This is \eqn{s} in Zhang et al. (2003) (eq. 7).
#' @param exp logical. The window size can be increased linearly or exponentially (eq. 4 or 5).
#' @param ws numeric. Sequence of windows sizes to be used in filtering ground returns.
#' The values must be positive and in the same units as the point cloud (usually meters, occasionally
#' feet).
#' @param th numeric. Sequence of threshold heights above the parameterized ground surface to be
#' considered a ground return. The values must be positive and in the same units as the point cloud.
#'
#' @references
#' Zhang, K., Chen, S. C., Whitman, D., Shyu, M. L., Yan, J., & Zhang, C. (2003). A progressive
#' morphological filter for removing nonground measurements from airborne LIDAR data. IEEE
#' Transactions on Geoscience and Remote Sensing, 41(4 PART I), 872–882. http:#doi.org/10.1109/TGRS.2003.810682.
#'
#' @export
#'
#' @family ground segmentation algorithms
#'
#' @examples
#' LASfile <- system.file("extdata", "Topography.laz", package="lidR")
#' las <- readLAS(LASfile, select = "xyzrn", filter = "-inside 273450 5274350 273550 5274450")
#'
#' ws <- seq(3,12, 3)
#' th <- seq(0.1, 1.5, length.out = length(ws))
#'
#' las <- classify_ground(las, pmf(ws, th))
#' #plot(las, color = "Classification")
#' @name gnd_pmf
pmf = function(ws, th)
{
  ws <- lazyeval::uq(ws)
  th <- lazyeval::uq(th)

  f = function(las, filter)
  {
    assert_is_valid_context(LIDRCONTEXTGND, "pmf")
    force_autoindex(las) <- LIDRGRIDPARTITION
    return(C_pmf(las, ws, th, filter))
  }

  f <- plugin_gnd(f)
  return(f)
}

#' Ground Segmentation Algorithm
#'
#' This function is made to be used in \link{classify_ground}. It implements an algorithm for segmentation
#' of ground points base on a Cloth Simulation Filter. This method is a strict implementation of
#' the CSF algorithm made by Zhang et al. (2016) (see references) that relies on the authors' original
#' source code written and exposed to R via the the \code{RCSF} package.
#'
#' @param sloop_smooth logical. When steep slopes exist, set this parameter to TRUE to reduce
#' errors during post-processing.
#' @param class_threshold scalar. The distance to the simulated cloth to classify a point cloud into ground
#' and non-ground. The default is 0.5.
#' @param cloth_resolution scalar. The distance between particles in the cloth. This is usually set to the
#' average distance of the points in the point cloud. The default value is 0.5.
#' @param rigidness integer. The rigidness of the cloth. 1 stands for very soft (to fit rugged
#' terrain), 2 stands for medium, and 3 stands for hard cloth (for flat terrain). The default is 1.
#' @param iterations integer. Maximum iterations for simulating cloth. The default value is 500. Usually,
#' there is no need to change this value.
#' @param time_step scalar. Time step when simulating the cloth under gravity. The default value
#' is 0.65. Usually, there is no need to change this value. It is suitable for most cases.
#'
#' @references
#' W. Zhang, J. Qi*, P. Wan, H. Wang, D. Xie, X. Wang, and G. Yan, “An Easy-to-Use Airborne LiDAR Data
#' Filtering Method Based on Cloth Simulation,” Remote Sens., vol. 8, no. 6, p. 501, 2016.
#' (http://www.mdpi.com/2072-4292/8/6/501/htm)
#'
#' @export
#'
#' @family ground segmentation algorithms
#'
#' @examples
#' if (require(RCSF, quietly = TRUE))
#' {
#' LASfile <- system.file("extdata", "Topography.laz", package="lidR")
#' las <- readLAS(LASfile, select = "xyzrn")
#'
#' mycsf <- csf(TRUE, 1, 1, time_step = 1)
#' las <- classify_ground(las, mycsf)
#' #plot(las, color = "Classification")
#' }
#' @name gnd_csf
csf = function(sloop_smooth = FALSE, class_threshold = 0.5, cloth_resolution = 0.5, rigidness = 1L, iterations = 500L, time_step = 0.65)
{
  sloop_smooth     <- lazyeval::uq(sloop_smooth)
  class_threshold  <- lazyeval::uq(class_threshold)
  cloth_resolution <- lazyeval::uq(cloth_resolution)
  rigidness        <- lazyeval::uq(rigidness)
  iterations       <- lazyeval::uq(iterations)
  time_step        <- lazyeval::uq(time_step)

  assert_is_a_bool(sloop_smooth)
  assert_is_a_number(class_threshold)
  assert_is_a_number(cloth_resolution)
  assert_is_a_number(rigidness)
  assert_is_a_number(iterations)
  assert_is_a_number(time_step)
  assert_package_is_installed("RCSF")

  f = function(las, filter)
  {
    . <- X <- Y <- Z <- NULL
    assert_is_valid_context(LIDRCONTEXTGND, "csf")

    las@data[["idx"]] <- 1:npoints(las)
    cloud <- las@data[filter, .(X,Y,Z, idx)]
    gnd <- RCSF::CSF(cloud, sloop_smooth, class_threshold, cloth_resolution, rigidness, iterations, time_step)
    idx <- cloud$idx[gnd]
    return(idx)
  }

  f <- plugin_gnd(f)
  return(f)
}


#' Progressive TIN Densification
#'
#' Classify ground points using a modified version of the Progressive TIN
#' densification method by Axelsson (2000) (see references). The approach involves iteratively
#' constructing a triangulated surface model (TIN) to classify ground points (see details).
#'
#' The method begins by identifying local low points, assuming at least one ground-level point exists
#' within any X meters area. Initially, the triangles of the TIN model are mostly below the ground
#' surface, only touching lowest points at their vertices. As the iterations progress, additional
#' points are incorporated to more accurately follow the true ground surface.\cr\cr
#' The classification process is controlled by two main parameters:
#' \itemize{
#' \item Iteration angle: This controls the rate at which points are classified based on variations in
#' ground level. Use smaller values (e.g., 20) for flat terrain and larger values (e.g., 40) for
#' mountainous regions.
#' \item Iteration distance: This limits large upward jumps in the model, preventing the misclassification
#'  of points in low vegetation or on small buildings.
#' }
#' The algorithm is robust to low noise points, meaning it can handle some level of noise without
#' needing explicit noise classification. However, it is preferable to have noise classified points
#' to guarantee optimal results.
#' \cr\cr
#' The densification process can also stop if the triangle edges become shorter than a specified length.
#'
#' @param distance Scalar. The threshold for iteration distance, as described in Figure 3 of Axelsson's paper.
#' @param angle Scalar. The threshold for iteration angle, as illustrated in Figure 3 of Axelsson's paper.
#' Use smaller values (close to 20) for flat areas and larger values (close to 40) for mountainous
#' areas.
#' @param res Scalar. The resolution for locating seed points. It takes the lowest point per grid cell.
#' It should be larger than any above ground structure such as building or bridges. In forested lands
#' 10 m is good but larger values are recommended if human made structure can be observed. 50 m is
#' good in urban scene
#' @param spacing Scalar. The spacing of ground points. TIN densification halts if triangles are
#' smaller than this limit avoiding useless over-densification. The default setting is generally suitable
#' for most scenarios.
#' @param verbose bool
#'
#' @references
#' Axelsson, P. (2000). DEM Generation from Laser Scanner Data Using adaptive TIN Models. International
#' Archives of Photogrammetry and Remote Sensing, 33(B4), 110–117.
#' https://www.isprs.org/proceedings/xxxiii/congress/part4/111_xxxiii-part4.pdf
#'
#' @export
#' @family ground segmentation algorithms
#' @md
#'
#' @examples
#' LASfile <- system.file("extdata", "Topography.laz", package="lidR")
#' las <- readLAS(LASfile, select = "xyzrn")
#'
#' myptd <- ptd(res = 20, distance = 1, angle = 20)
#' las <- classify_ground(las, myptd)
#' #plot(las, color = "Classification")
ptd = function(res = 10, angle = 30, distance = 2, spacing = 0.25, verbose = FALSE)
{
  res     <- lazyeval::uq(res)
  angle   <- lazyeval::uq(angle)
  distance <- lazyeval::uq(distance)
  spacing  <- lazyeval::uq(spacing)
  verbose  <- lazyeval::uq(verbose)

  assert_is_a_bool(verbose)
  assert_is_a_number(res)
  assert_is_a_number(angle)
  assert_is_a_number(distance)
  assert_is_a_number(spacing)

  f = function(las, filter)
  {
    . <- X <- Y <- Z <- NULL
    assert_is_valid_context(LIDRCONTEXTGND, "ptd")

    las@data[["idx"]] <- 1:npoints(las)
    cloud <- las@data[filter, .(X,Y,Z, idx)]
    gnd <- C_ptd(cloud, res, angle, distance, spacing, verbose, get_lidr_threads())
    idx <- cloud$idx[gnd+1]
    return(idx)
  }

  f <- plugin_gnd(f)
  return(f)
}

#' @export
#' @rdname gnd_pmf
util_makeZhangParam = function(b = 2, dh0 = 0.5, dhmax = 3.0, s = 1.0,  max_ws = 20, exp = FALSE)
{
  if (exp & b <= 1)
    stop("b cannot be less than 1 with an exponentially growing window")

  if (dh0 >= dhmax)
    stop("dh0 greater than dhmax")

  if (max_ws < 3)
    stop("Minimum windows size is 3. max_ws must be greater than 3")

  if (!is.logical(exp))
    stop("exp should be logical")

  if (!exp & b < 1)
    warning("Due to an incoherence in the original paper when b < 1, the sequences of windows size cannot be computed for a linear increase. The internal routine uses the fact that the increment is constant to bypass this issue.")


  dhtk = c()
  wk = c()
  k = 0
  ws = 0
  th = 0
  c = 1

  while (ws <= max_ws)
  {
    # Determine the initial window size.
    if (exp)
      ws = (2.0*b^k) + 1
    else
      ws = 2.0*(k + 1)*b + 1

    # Calculate the height threshold to be used in the next k.
    if (ws <= 3)
      th = dh0
    else
    {
      if (exp)
        th = s * (ws - wk[k]) * c + dh0
      else
        th = s*2*b*c + dh0
    }

    # Enforce max distance on height threshold
    if (th > dhmax)
      th = dhmax

    if (ws <= max_ws)
    {
      wk = append(wk, ws)
      dhtk = append(dhtk, th)
    }

    k = k + 1
  }

  return(list(ws = wk, th = dhtk))
}

context("classify_ground")

las = clip_rectangle(topography, 273450, 5274350, 273550, 5274450)
ctg = topography_ctg

opt_chunk_size(ctg) <- 300
ctg@chunk_options$alignment = c(50, 200)
opt_chunk_buffer(ctg) <- 0
opt_progress(ctg) <- FALSE

ws = seq(3,21, 5)
th = seq(0.1, 2, length.out = length(ws))

test_that("classify_ground pmf works with LAS", {

  mypmf = pmf(ws, th)

  las <- classify_ground(las, mypmf)

  n = names(las)

  expect_true("Classification" %in% n)
  expect_equal(sort(unique(las@data$Classification)), c(1L, 2L))
  expect_equal(sum(las@data$Classification == 2L), 1933L)

  expect_error(classify_ground(ctg, mypmf), "buffer")
})

test_that("classify_ground pmf works with LAScatalog", {
  opt_chunk_buffer(ctg) <- 30

  mypmf = pmf(ws, th)

  expect_error(classify_ground(ctg, mypmf), "output file")

  opt_output_files(ctg) <- paste0(tempdir(), "/file_{XLEFT}_{YBOTTOM}")

  ctg2 = classify_ground(ctg, mypmf)
  las2 = readLAS(ctg2)

  expect_equal(sum(las2@data$Classification == 2L), 19472L)
})

test_that("classify_ground csf works with LAS", {

  skip_if_not_installed("RCSF")

  mycsf = csf(TRUE, 1, 1, time_step = 1)

  las <- classify_ground(las, mycsf)

  n = names(las)

  expect_true("Classification" %in% n)
  expect_equal(sort(unique(las@data$Classification)), c(1L, 2L))
  expect_equal(sum(las@data$Classification == 2L), 2605L)
})

test_that("classify_ground csf works with LAScatalog", {

  skip_if_not_installed("RCSF")
  skip_on_cran()

  mycsf = csf(TRUE, 1, 1, time_step = 1)

  opt_output_files(ctg) <- paste0(tempdir(), "/file_{XLEFT}_{YBOTTOM}_ground")
  opt_chunk_buffer(ctg) <- 30

  ctg2 = classify_ground(ctg, mycsf)
  las2 = readLAS(ctg2)

  expect_equal(sum(las2@data$Classification == 2L), 26715L-450L)
})

test_that("classify_ground csf works with last_returns = FALSE", {

  skip_if_not_installed("RCSF")

  las <- lidR:::generate_las(500)
  las <- classify_ground(las, csf(), last_returns = FALSE)

  n = names(las)

  expect_true("Classification" %in% n)
  expect_equal(unique(las@data$Classification), c(1L, 2L))
  expect_equal(sum(las@data$Classification == 2L), 105L)
})

test_that("classify_ground ptd works",
{
  skip_on_cran()
  skip_on_os("mac")

  myptd = ptd()
  las <- classify_ground(topography, myptd)
  ans <- as.numeric(table(las$Classification))
  class = c(48967,24383,53)
  expect_equal(ans, class)

  myptd = ptd(angle = 1, distance = 0.5)
  las <- classify_ground(topography, myptd)
  ans <- as.numeric(table(las$Classification))
  class = c(67271,4758,1374)
  expect_equal(ans, class)
})

test_that("classify_ground works with last_returns = TRUE but attribute not properly populated", {

  skip_if_not_installed("RCSF")

  las <- lidR:::generate_las(500)
  las@data$ReturnNumber <- 0
  las@data$Classification <- NULL
  las <- suppressWarnings(classify_ground(las, csf(), last_returns = TRUE))

  n = names(las)

  expect_true("Classification" %in% n)
  expect_equal(unique(las@data$Classification), c(1L, 2L))
  expect_equal(sum(las@data$Classification == 2L), 105L)
})

test_that("makeZhangParam works", {
  expect_error(util_makeZhangParam(), NA)
})

test_that("classify_ground does not erase former classification (but new ground points)", {
  mypmf = pmf(ws, th)
  las <- topography
  las <- filter_poi(las, X < mean(X), Y < mean(Y))
  las$Classification[las$Classification == LASGROUND] <- LASUNCLASSIFIED
  las <- classify_ground(las, mypmf)
  expect_equal(names(table(las$Classification)), c("1", "2", "9"))
})

# ==== STANDARD ====

#' Predefined standard metrics functions
#'
#' Predefined metrics functions intended to me used in `*_metrics` function such as
#' \link{pixel_metrics}, \link{cloud_metrics}, \link{crown_metrics}, \link{voxel_metrics} and
#' so on. Each function comes with a convenient shortcuts for lazy coding. The `lidR` package aims
#' to provide an easy way to compute user-defined metrics rather than to provide them. However, for
#' efficiency and to save time, sets of standard metrics have been predefined (see details). Every
#' function can be computed by every `*_metrics` functions however `stdmetrics*` are
#' more pixel-based metrics, `stdtreemetrics` are more tree-based metrics and `stdshapemetrics` are
#' more point-based metrics. For example the metric `zmean` computed by `stdmetrics_z` makes sense
#' when computed at the pixel level but brings no information at the voxel level.
#'
#' The function names, their parameters and the output names of the metrics rely on a nomenclature
#' chosen for brevity:
#' \itemize{
#' \item{`z`: refers to the elevation}
#' \item{`i`: refers to the intensity}
#' \item{`rn`: refers to the return number}
#' \item{`q`: refers to quantile}
#' \item{`a`: refers to the ScanAngleRank or ScanAngle}
#' \item{`n`: refers to a number (a count)}
#' \item{`p`: refers to a percentage}
#' }
#' For example the metric named `zq60` refers to the elevation, quantile, 60 i.e. the 60th percentile
#' of elevations. The metric `pground` refers to a percentage. It is the percentage of points
#' classified as ground. The function `stdmetric_i` refers to metrics of intensity. A description of
#' each existing metric can be found on the \href{https://github.com/r-lidar/lidR/wiki/stdmetrics}{lidR wiki page}.\cr\cr
#' Some functions have optional parameters. If these parameters are not provided the function
#' computes only a subset of existing metrics. For example, `stdmetrics_i` requires the intensity
#' values, but if the elevation values are also provided it can compute additional metrics such as
#' cumulative intensity at a given percentile of height.\cr\cr
#' Each function has a convenient associated variable. It is the name of the function, with a
#' dot before the name. This enables the function to be used without writing parameters. The cost
#' of such a feature is inflexibility. It corresponds to a predefined behaviour (see examples)\cr
#' \describe{
#' \item{`stdmetrics`}{is a combination of `stdmetrics_ctrl` + `stdmetrics_z` +
#' `stdmetrics_i` +  `stdmetrics_rn`}
#' \item{`stdtreemetrics`}{is a special function that works with \link{crown_metrics}. Actually,
#' it won't fail with other functions but the output makes more sense if computed at the
#' individual tree level.}
#' \item{`stdshapemetrics`}{is a set of eigenvalue based feature described in Lucas et al, 2019
#' (see references).}
#' }
#'
#' @param x,y,z,i Coordinates of the points, Intensity
#' @param rn,class ReturnNumber, Classification
#' @param pulseID The number referencing each pulse
#' @param dz numeric. Layer thickness  metric \link[=entropy]{entropy}
#' @param th numeric. Threshold for metrics pzabovex. Can be a vector to compute with several thresholds.
#' @param zmin numeric. Lower bound of the integral for zpcumx metrics.
#' See \href{https://github.com/r-lidar/lidR/wiki/stdmetrics}{wiki page} and Wood et al.
#' (2008) reference.
#' @examples
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' las <- readLAS(LASfile, select = "*", filter = "-keep_random_fraction 0.5")
#'
#' # All the predefined metrics
#' m1 <- pixel_metrics(las, ~stdmetrics(X,Y,Z,Intensity,ReturnNumber,Classification,dz=1), res = 40)
#'
#' # Convenient shortcut
#' m2 <- pixel_metrics(las, .stdmetrics, res = 40)
#'
#' # Basic metrics from intensities
#' m3 <- pixel_metrics(las, ~stdmetrics_i(Intensity), res = 40)
#'
#' # All the metrics from intensities
#' m4 <- pixel_metrics(las, ~stdmetrics_i(Intensity, Z, Classification, ReturnNumber), res = 40)
#'
#' # Convenient shortcut for the previous example
#' m5 <- pixel_metrics(las, .stdmetrics_i, res = 40)
#'
#' # Combine some predefined function with your own new metrics
#' # Here convenient shortcuts are no longer usable.
#' myMetrics = function(z, i, rn)
#' {
#'   first  <- rn == 1L
#'   zfirst <- z[first]
#'   nfirst <- length(zfirst)
#'   above2 <- sum(z > 2)
#'
#'   x <- above2/nfirst*100
#'
#'   # User's metrics
#'   metrics <- list(
#'      above2aboven1st = x,       # Num of returns above 2 divided by num of 1st returns
#'      zimean  = mean(z*i),       # Mean products of z by intensity
#'      zsqmean = sqrt(mean(z^2))  # Quadratic mean of z
#'    )
#'
#'   # Combined with standard metrics
#'   return( c(metrics, stdmetrics_z(z)) )
#' }
#'
#' m10 <- pixel_metrics(las, ~myMetrics(Z, Intensity, ReturnNumber), res = 40)
#'
#' # Users can write their own convenient shorcuts like this:
#' .myMetrics = ~myMetrics(Z, Intensity, ReturnNumber)
#' m11 <- pixel_metrics(las, .myMetrics, res = 40)
#'
#' @references
#' M. Woods, K. Lim, and P. Treitz. Predicting forest stand variables from LiDAR data in
#' the Great Lakes – St. Lawrence forest of Ontario. The Forestry Chronicle. 84(6): 827-839.
#' https://doi.org/10.5558/tfc84827-6
#' @rdname stdmetrics
#' @export
#' @md
stdmetrics = function(x, y, z, i, rn, class, dz = 1, th = 2, zmin = 0)
{
  C  <- stdmetrics_ctrl(x, y, z)
  Z  <- stdmetrics_z(z, dz, th, zmin)
  I  <- stdmetrics_i(i, z, class, rn)
  RN <- stdmetrics_rn(rn, class)
  #PU = stdmetrics_pulse(pulseID, rn)

  metrics <- c(Z, I, RN, C)
  return(metrics)
}

#' @rdname stdmetrics
#' @export
stdmetrics_z = function(z, dz = 1, th = 2, zmin = 0)
{
  n <- length(z)
  zmax  <- max(z)
  zmean <- mean(z)

  probs <- seq(0.05, 0.95, 0.05)
  zq 	  <- as.list(stats::quantile(z, probs))
  names(zq) <- paste0("zq", probs*100)

  pzabovex <- lapply(th, function(x) { fast_countover(z, x) / n * 100 })
  names(pzabovex) <- paste0("pzabove", th)

  pzabovemean <- fast_countover(z, zmean) / n * 100

  if (zmax <= zmin)
  {
    d <- rep(0, 9)
  }
  else
  {
    breaks <- seq(zmin, zmax, (zmax-zmin)/10)
    d <- findInterval(z[z>zmin], breaks)
    d <- fast_table(d, 10)
    d <- d / sum(d)*100
    d <- cumsum(d)[1:9]
    d <- as.list(d)
  }
  names(d) <- paste0("zpcum", 1:9)

  metrics <- list(
    zmax  = zmax,
    zmean = zmean,
    zsd   = stats::sd(z),
    zskew = (sum((z - zmean)^3)/n)/(sum((z - zmean)^2)/n)^(3/2),
    zkurt = n * sum((z - zmean)^4)/(sum((z - zmean)^2)^2),
    zentropy  = entropy(z, dz),
    pzabovezmean = pzabovemean
  )

  metrics <- c(metrics, pzabovex, zq, d)

  return(metrics)
}

#' @rdname stdmetrics
#' @export
stdmetrics_i = function(i, z = NULL, class = NULL, rn = NULL)
{
  itot <- imax <- imean <- isd <- icv <- iskew <- ikurt <- NULL
  icumzq10 <- icumzq30 <- icumzq50 <- icumzq70 <- icumzq90 <- NULL

  n <- length(i)
  itot <- as.double(sum(i))
  imean <- mean(i)

  probs <- seq(0.05, 0.95, 0.05)
  iq 	  <- as.list(stats::quantile(i, probs))
  names(iq) <- paste("iq", probs*100, sep = "")

  metrics <- list(
    itot = itot,
    imax  = max(i),
    imean = imean,
    isd   = stats::sd(i),
    iskew = (sum((i - imean)^3)/n)/(sum((i - imean)^2)/n)^(3/2),
    ikurt = n * sum((i - imean)^4)/(sum((i - imean)^2)^2)
  )

  if (!is.null(class))
  {
    metrics <- c(metrics, list(ipground = sum(i[class == 2])/itot*100))
  }

  if (!is.null(z))
  {
    zq <- stats::quantile(z, probs = seq(0.1, 0.9, 0.2))

    ipcum <- list(
      ipcumzq10 = sum(i[z <= zq[1]])/itot*100,
      ipcumzq30 = sum(i[z <= zq[2]])/itot*100,
      ipcumzq50 = sum(i[z <= zq[3]])/itot*100,
      ipcumzq70 = sum(i[z <= zq[4]])/itot*100,
      ipcumzq90 = sum(i[z <= zq[5]])/itot*100
    )

    metrics <- c(metrics, ipcum)
  }

  metrics
}

#' @rdname stdmetrics
#' @export
stdmetrics_rn = function(rn, class = NULL)
{
  nground <- NULL

  n <- length(rn)

  prn <- table(factor(rn, levels = c(1:5)))/n*100
  prn <- as.list(prn)
  names(prn) <- paste0("p", names(prn), "th")

  metrics <- prn

  if (!is.null(class))
    metrics = c(metrics, list(pground = sum(class == 2)/n*100))

  return(metrics)
}

#' @rdname stdmetrics
#' @export
stdmetrics_pulse = function(pulseID, rn)
{
  . <- nr <- NULL

  n <- length(rn)

  dt <- data.table::data.table(pulseID, rn)

  np_with_x_return <- dt[, .(nr = max(rn)), by = pulseID][, .N, by = nr]
  data.table::setorder(np_with_x_return, nr)

  np <- numeric(9)
  names(np) <- paste0("ppulse", 1:9, "return")
  np[np_with_x_return$nr] <- np_with_x_return$N/n*100
  np <- as.list(np)

  return(np)
}

#' @rdname stdmetrics
#' @export
stdmetrics_ctrl = function(x, y, z)
{
  xmax <- max(x)
  ymax <- max(y)
  xmin <- min(x)
  ymin <- min(y)
  n    <- length(z)
  area <- (xmax - xmin)*(ymax - ymin)
  metrics <- list(n = n, area = area)

  return(metrics)
}

#' @rdname stdmetrics
#' @export
stdtreemetrics = function(x, y, z)
{
  metrics <- list(
    Z = max(z),
    npoints = length(x),
    convhull_area = round(area_convex_hull(x,y),3)
  )

  return(metrics)
}
#' @rdname stdmetrics
#' @export
#' @references
#' Lucas, C., Bouten, W., Koma, Z., Kissling, W. D., & Seijmonsbergen, A. C. (2019). Identification
#' of Linear Vegetation Elements in a Rural Landscape Using LiDAR Point Clouds. Remote Sensing, 11(3), 292.
stdshapemetrics = function(x,y,z)
{
  xyz     <- cbind(x,y,z)
  eigen   <- fast_eigen_values(xyz)
  eigen_m <- eigen[[1]]
  eigen_m <- as.numeric(eigen_m)

  shapemetrics <- list(
    eigen_largest  = eigen_m[1],
    eigen_medium   = eigen_m[2],
    eigen_smallest = eigen_m[3],
    curvature      = eigen_m[3]/(eigen_m[1] + eigen_m[2] + eigen_m[3]),
    linearity      = (eigen_m[1] - eigen_m[2])/eigen_m[1],
    planarity      = (eigen_m[2] - eigen_m[3])/eigen_m[1],
    sphericity     = eigen_m[3]/eigen_m[1],
    anisotropy     = (eigen_m[1] - eigen_m[3])/eigen_m[1],
    horizontality  = abs(eigen[[2]][3,3])
  )
  return(shapemetrics)
}

stdtreeapex <- function(x,y,z, ...)
{
  if (length(x) < 4)
    return(NULL)

  j <- which.max(z)
  point <- sf::st_point(c(x[j], y[j], z[j]))
  return(list(geom = list(point)))
}

stdtreehullconvex = function(x,y,z, ...)
{
  if (length(x) < 4)
    return(NULL)

  i <- grDevices::chull(x,y)
  i <- c(i, i[1])
  P <- cbind(x[i], y[i])
  poly <- sf::st_polygon(list(P))

  return(list(geom = list(poly)))
}

stdtreehullconcave = function(x,y,z, concavity, length_threshold)
{
  if (length(x) < 4)
    return(NULL)

  P <- concaveman(x, y, concavity, length_threshold)
  P <- as.matrix(P)
  poly <- sf::st_polygon(list(P))

  return(list(geom = list(poly)))
}

stdtreehullbbox = function(x,y,z, ...)
{
  if (length(x) < 4)
    return(NULL)

  xmin <- min(x)
  ymin <- min(y)
  xmax <- max(x)
  ymax <- max(y)

  x <- c(xmin, xmax, xmax, xmin, xmin)
  y <- c(ymin, ymin, ymax, ymax, ymin)
  P <- cbind(x, y)
  poly <- sf::st_polygon(list(P))

  return(list(geom = list(poly)))
}

#' @rdname stdmetrics
#' @export
.stdmetrics = ~stdmetrics(X,Y,Z,Intensity, ReturnNumber, Classification, dz = 1, th = 2)

#' @rdname stdmetrics
#' @export
.stdmetrics_z = ~stdmetrics_z(Z, dz =1)

#' @rdname stdmetrics
#' @export
.stdmetrics_i = ~stdmetrics_i(Intensity, Z, Classification, ReturnNumber)

#' @rdname stdmetrics
#' @export
.stdmetrics_rn = ~stdmetrics_rn(ReturnNumber, Classification)

#' @rdname stdmetrics
#' @export
.stdmetrics_pulse = ~stdmetrics_pulse(pulseID, ReturnNumber)

#' @rdname stdmetrics
#' @export
.stdmetrics_ctrl = ~stdmetrics_ctrl(X, Y, Z)

#' @rdname stdmetrics
#' @export
.stdtreemetrics = ~stdtreemetrics(X, Y, Z)

#' @rdname stdmetrics
#' @export
.stdshapemetrics = ~stdshapemetrics(X, Y, Z)

# ===== OTHER =====

#' Predefined non standard metrics
#'
#' Functions and metrics from the literature. See details and references
#'
#' \describe{
#' \item{rumple_index}{Computes the roughness of a surface as the ratio between its area and its
#' projected area on the ground. If the input is a gridded object (raster) the function computes the
#' surfaces using Jenness's algorithm (see references). If the input is a point cloud the function
#' uses a Delaunay triangulation of the points and computes the area of each triangle.}
#' \item{gap_fraction_profile}{Computes the gap fraction profile using the method of Bouvier et al.
#' (see reference). The function assesses the number of laser points that actually reached the layer
#' z+dz and those that passed through the layer [z, z+dz]. By definition the layer 0
#' will always return 0 because no returns pass through the ground. Therefore, the layer 0 is removed
#' from the returned results.}
#' \item{LAD}{Computes a leaf area density profile based on the method of Bouvier et al. (see reference)
#' The function assesses the number of laser points that actually reached the layer z+dz and those that
#' passed through the layer [z, z+dz] (see \link[=gap_fraction_profile]{gap_fraction_profile}).
#' Then it computes the log of this quantity and divides it by the extinction coefficient k as
#' described in Bouvier et al. By definition the layer 0 will always return infinity because no returns
#'  pass through the ground. Therefore, the layer 0 is removed from the returned results.}
#'  \item{entropy}{A normalized Shannon vertical complexity index. The Shannon diversity index is a
#'  measure for quantifying diversity and is based on the number and frequency of species present. This index,
#' developed by Shannon and Weaver for use in information theory, was successfully transferred
#' to the description of species diversity in biological systems (Shannon 1948). Here it is applied
#' to quantify the diversity and the evenness of an elevational distribution of las points. It
#' makes bins between 0 and the maximum elevation. If there are negative values the function
#' returns NA.}
#' \item{VCI}{Vertical Complexity Index. A fixed normalization of the entropy function from van Ewijk
#' et al. (2011) (see references)}
#' }
#'
#' @name nstdmetrics
#' @rdname nstdmetrics
#' @param by,dz numeric. The thickness of the layers used (height bin)
#' @param zmax numeric. Maximum elevation for an entropy normalized to zmax.
NULL

#' @param x A `RasterLayer`, a `stars` a `SpatRaster` or a vector of x coordinates.
#' @param y numeric. If `x` is a vector of coordinates: the associated y coordinates.
#' @param z numeric. If `x` is a vector of coordinates: the associated z coordinates.
#' @param ... unused
#' @return numeric. The computed Rumple index.
#'
#' @export
#' @examples
#' x <- runif(20, 0, 100)
#' y <- runif(20, 0, 100)
#'
#' if (require(geometry, quietly = TRUE))
#' {
#' # Perfectly flat surface, rumple_index = 1
#' z <- rep(10, 20)
#' rumple_index(x, y, z)
#'
#' # Rough surface, rumple_index > 1
#' z <- runif(20, 0, 10)
#' rumple_index(x, y, z)
#'
#' # Rougher surface, rumple_index increases
#' z <- runif(20, 0, 50)
#' rumple_index(x, y, z)
#'
#' # Measure of roughness is scale-dependent
#' rumple_index(x, y, z)
#' rumple_index(x/10, y/10, z)
#' }
#' @md
#' @references
#' Jenness, J. S. (2004). Calculating landscape surface area from digital elevation
#' models. Wildlife Society Bulletin, 32(3), 829–839.
#' @rdname nstdmetrics
rumple_index = function(x, y = NULL, z = NULL, ...)
{
  UseMethod("rumple_index", x)
}

#' @export
rumple_index.RasterLayer <- function(x, y = NULL, z = NULL, ...)
{
  res <- raster_res(x)
  x   <- raster_as_matrix(x)$z
  return(rumple_index.matrix(x, res[1], res[2]))
}

#' @export
rumple_index.stars <- function(x, y = NULL, z = NULL, ...)
{
  res <- raster_res(x)
  x   <- raster_as_matrix(x)$z
  return(rumple_index.matrix(x, res[1], res[2]))
}

#' @export
rumple_index.SpatRaster <- function(x, y = NULL, z = NULL, ...)
{
  res <- raster_res(x)
  x   <- raster_as_matrix(x)$z
  return(rumple_index.matrix(x, res[1], res[2]))
}

#' @export
rumple_index.matrix <- function(x, y = NULL, z = NULL, ...)
{
  area  <- sp::surfaceArea(x, y, z)
  parea <- sum(!is.na(x))*y*z
  return(area/parea)
}

#' @export
rumple_index.numeric <- function(x, y = NULL, z = NULL, ...)
{
  assert_is_numeric(x)
  assert_is_numeric(y)
  assert_is_numeric(z)
  assert_are_same_length(x,y)
  assert_are_same_length(x,z)
  assert_package_is_installed("geometry")

  if (length(x) <= 3)
    return(NA_real_)

  rumple <- tryCatch(
  {
    X <- cbind(x,y,z)
    dn <- suppressMessages(geometry::delaunayn(X[,1:2], options = "QbB"))
    N <- C_tinfo(dn, X)
    area  <- sum(N[,5])
    parea <- sum(N[,6])
    return(area/parea)
  },
  error = function(e)
  {
    message(paste0(e, "\n'rumple_index' returned NA."))
    if (getOption("lidR.debug")) dput(X[,1:2])
    return(NA_real_)
  })

  return(rumple)
}

#' @param z vector of positive z coordinates
#' @param z0 numeric. The bottom limit of the profile
#' @return A data.frame containing the bin elevations (z) and the gap fraction for each bin (gf)
#' @examples
#' z <- c(rnorm(1e4, 25, 6), rgamma(1e3, 1, 8)*6, rgamma(5e2, 5,5)*10)
#' z <- z[z<45 & z>0]
#'
#' hist(z, n=50)
#'
#' gapFraction = gap_fraction_profile(z)
#'
#' plot(gapFraction, type="l", xlab="Elevation", ylab="Gap fraction")
#' @references Bouvier, M., Durrieu, S., Fournier, R. a, & Renaud, J. (2015).  Generalizing predictive
#' models of forest inventory attributes using an area-based approach with airborne las data. Remote
#' Sensing of Environment, 156, 322-334. http://doi.org/10.1016/j.rse.2014.10.004
#' @export
#' @rdname nstdmetrics
gap_fraction_profile = function(z, dz = 1, z0 = 2)
{
  zrange = range(z)
  if (z0 < zrange[1])
    z0 = floor((zrange[1]- z0)/dz)*dz + z0

  if (z0 >= zrange[2])
    return(data.frame(z = numeric(0), gf = numeric(0)))

  bk <- seq(z0, ceiling((zrange[2] - z0)/dz)*dz + z0, dz)

  histogram <- graphics::hist(z, breaks = c(-Inf, bk), plot = F)
  h <- histogram$mids
  p <- histogram$counts

  cs <- cumsum(p)
  i <- cs[1:(length(cs)-1)]/cs[2:length(cs)]

  i[is.na(i)] = 0

  z = h[-1]

  return(data.frame(z = z, gf = i))
}

#' @param k numeric. is the extinction coefficient
#' @examples
#' z <- c(rnorm(1e4, 25, 6), rgamma(1e3, 1, 8)*6, rgamma(5e2, 5,5)*10)
#' z <- z[z<45 & z>0]
#'
#' lad <- LAD(z)
#'
#' plot(lad, type="l", xlab="Elevation", ylab="Leaf area density")
#' @export
#' @rdname nstdmetrics
LAD = function(z, dz = 1, k = 0.5, z0 = 2) # (Bouvier et al. 2015)
{
  ld <- gap_fraction_profile(z, dz, z0)

  if (nrow(ld) == 0)
    return(data.frame(z = numeric(0), lad = numeric(0)))

  lad <- ld$gf
  lad <- -log(lad)/(k*dz)

  lad[is.infinite(lad)] = NA

  lad <- lad

  return(data.frame(z = ld$z, lad = lad))
}

#' @examples
#' z <- runif(10000, 0, 10)
#'
#' # expected to be close to 1. The highest diversity is given for a uniform distribution
#' entropy(z, by = 1)
#'
#'  z <- runif(10000, 9, 10)
#'
#' # Must be 0. The lowest diversity is given for a unique possibility
#' entropy(z, by = 1)
#'
#' z <- abs(rnorm(10000, 10, 1))
#'
#' # expected to be between 0 and 1.
#' entropy(z, by = 1)
#' @references
#' Pretzsch, H. (2008). Description and Analysis of Stand Structures. Springer Berlin Heidelberg. http://doi.org/10.1007/978-3-540-88307-4 (pages 279-280)
#' Shannon, Claude E. (1948), "A mathematical theory of communication," Bell System Tech. Journal 27, 379-423, 623-656.
#' @return A number between 0 and 1
#' @export
#' @rdname nstdmetrics
entropy = function(z, by = 1, zmax = NULL)
{
  # Fixed entropy (van Ewijk et al. (2011)) or flexible entropy
  if (is.null(zmax))
    zmax = max(z)

  # If zmax < 3 it is meaningless to compute entropy
  if (zmax < 2 * by)
    return(NA_real_)

  if (min(z) < 0)
    return(NA_real_)

  # Define the number of x meters bins from 0 to zmax (rounded to the next integer)
  bk <- seq(0, ceiling(zmax/by)*by, by)

  # Compute the p for each bin
  hist <- findInterval(z, bk)
  hist <- fast_table(hist, length(bk) - 1)
  hist <- hist/sum(hist)

  # Remove bins where there are no points because of log(0)
  p    <- hist[hist > 0]
  pref <- rep(1/length(hist), length(hist))

  # normalized entropy
  S <- -sum(p*log(p)) / -sum(pref*log(pref))

  return(S)
}


#' @return A number between 0 and 1
#' @examples
#' z <- runif(10000, 0, 10)
#'
#' VCI(z, by = 1, zmax = 20)
#'
#' z <- abs(rnorm(10000, 10, 1))
#'
#' # expected to be closer to 0.
#' VCI(z, by = 1, zmax = 20)
#' @references van Ewijk, K. Y., Treitz, P. M., & Scott, N. A. (2011). Characterizing Forest Succession in Central Ontario using LAS-derived Indices. Photogrammetric Engineering and Remote Sensing, 77(3), 261-269. Retrieved from <Go to ISI>://WOS:000288052100009
#' @export
#' @rdname nstdmetrics
VCI = function(z, zmax, by = 1)
{
  z <- z[z < zmax]
  return(entropy(z, by, zmax))
}

las = random_10_points

VCI = function() { 0 }
colorize_points = function() { 0 }
f = function()
{
  entropy <- function(x) x
  cloud_metrics(las, ~entropy(Z))
}


test_that("Ambiguous definitions are handled", {

  expect_error(cloud_metrics(las, ~VCI()), "exists in the package lidR but is also defined in")
  expect_error(cloud_metrics(las, ~colorize_points()), "exists in the package lidR but is also defined in")
  expect_error(f(), "exists in the package lidR but is also defined in")
  #expect_error(cloud_metrics(las, ~lidR::VCI(Z, 10)))
})

test_that("gap fraction returns proper values", {
  set.seed(1)
  Z = runif(50000, 0, 5)
  gf = 1-gap_fraction_profile(Z, 1, 0)$gf

  expect_equal(gf, c(1, 0.5, 0.333, 0.25, 0.2), tolerance = 0.01)
})

test_that("entropy returns proper values", {
  set.seed(1)
  Z = runif(20000, 0, 5)
  S = entropy(Z)

  expect_equal(S, 1, tolerance = 0.0001)

  set.seed(42)
  Z = runif(20000, 0, 1)
  Z = c(Z, 5)
  S = entropy(Z)

  expect_equal(S, 0)
})

test_that("VCI returns the same as entropy values", {
  Z = runif(100, 0, 5)
  S = entropy(Z, zmax = 6)
  V = VCI(Z, 6)

  expect_equal(S, V)
})

context("roughness")

skip_if_not_installed("geometry")

test_that("Flat surfaces have a rumple index of 1. Delaunay method", {
  n = sample(10:20, 1)
  x = runif(n, 0, 100)
  y = runif(n, 0, 100)
  z = rep(10, n)

  expect_equal(rumple_index(x, y, z), 1)
  expect_equal(rumple_index(x/10, y/5, z), 1)
  expect_equal(rumple_index(x, y, z*2), 1)
})

test_that("Flat surfaces have a rumple index of 1. Jenness method", {
  skip_if_not_installed("sp")
  bbox = terra::ext(0,10,0,10)
  r = terra::rast(bbox, res = 0.5)
  r[] = 2.56

  expect_equal(rumple_index(r), 1)
})

test_that("Rought surfaces have the good rumple index. Delaunay", {
  n = sample(10:20, 1)
  x = runif(n, 0, 100)
  y = runif(n, 0, 100)
  z = x

  expect_equal(rumple_index(x, y, z), sqrt(2))

  z = x + y

  expect_equal(rumple_index(x, y, z), sqrt(3))
})

test_that("Error handling. Delaunay", {

  x = runif(3, 0, 100)
  y = runif(3, 0, 100)
  z = x

  expect_equal(rumple_index(x, y, z), NA_real_)

  # x = runif(4, 0, 100)
  # y = x
  # z = x
  #
  # expect_equal(suppressMessages(rumple_index(x, y, z)), NA_real_)
})

las = clip_rectangle(megaplot, 684766.4, 5017773, 684866.4, 5017973)

xall = pixel_metrics(las, .stdmetrics, 20)

test_that("stdmetric returns the same result than .stdmetric", {
  y = pixel_metrics(las, ~stdmetrics(X,Y,Z, Intensity, ReturnNumber, Classification), 20)
  expect_equal(xall, y)
})

test_that("stdmetric_i returns the same result than .stdmetric_i", {
  x = pixel_metrics(las, .stdmetrics_i, 20)
  y = pixel_metrics(las, ~stdmetrics_i(Intensity, Z, Classification), 20)

  expect_equal(x,y)
})

test_that("stdmetric_pulse works", {
  las = retrieve_pulses(las)
  expect_error(pixel_metrics(las, .stdmetrics_pulse, 20), NA)
})

#' Individual tree detection
#'
#' Individual tree detection function that find the position of the trees using several possible
#' algorithms.
#'
#' @param las An object of class `LAS` or `LAScatalog`. Can also be a  raster from `raster`, `stars` or `terra`
#' representing a canopy height model, in which case it is processed like a regularly-spaced point cloud.
#' @param algorithm An algorithm for individual tree detection. lidR has: \link{lmf} and \link{manual}.
#' More experimental algorithms may be found in the package \href{https://github.com/Jean-Romain/lidRplugins}{lidRplugins}.
#' @param uniqueness character. A method to compute a unique ID. Can be 'incremental', 'gpstime' or
#' 'bitmerge'. See section 'Uniqueness'. This feature must be considered as 'experimental'.
#'
#' @template section-uniqueness
#'
#' @return `locate_trees` returns an sf object with POINT Z geometries. The table of attributes
#' contains a column `treeID` with an individual ID for each tree. The height of the trees (`Z`) are
#' also repeated in the table of attribute to be analysed as an attribute and not as a coordinate.
#'
#' @export
#'
#' @examples
#' LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
#' las <- readLAS(LASfile, select = "xyz", filter = "-inside 481250 3812980 481300 3813030")
#'
#' ttops <- locate_trees(las, lmf(ws = 5))
#'
#' #plot(las) |> add_treetops3d(ttops)
#' @md
locate_trees = function(las, algorithm, uniqueness = 'incremental')
{
  UseMethod("locate_trees", las)
}

#' @export
locate_trees.LAS = function(las, algorithm, uniqueness = 'incremental')
{
  assert_is_algorithm(algorithm)
  assert_is_algorithm_itd(algorithm)
  match.arg(uniqueness, c('incremental', 'gpstime', 'bitmerge'))

  if (uniqueness == 'gpstime' && !"gpstime" %in% names(las))
    stop("Impossible to compute unique IDs using gpstime: no gpstime found.", call. = FALSE)

  if (uniqueness == 'gpstime' &&  fast_countequal(las@data[["gpstime"]], 0L) == npoints(las))
    stop("Impossible to compute unique IDs using gpstime: gpstime is not populated.", call. = FALSE)

  lidR.context <- "locate_trees"
  res <- algorithm(las)

  if (is(res, "SpatialPointsDataFrame"))
    res <- sf::st_as_sf(res)

  if (is(res, "sf"))
  {
    sf::st_agr(res) <- "constant"
    return(res)
  }

  if (!is.logical(res) & !is.integer(res))
    stop("The output of the algorithm is incorrect")

  maxima <- las@data[res, c("X", "Y", "Z")]
  no_maxima <- nrow(maxima) == 0

  if (no_maxima)
  {
    data   <- data.frame(x = 0, y = 0, z = 0, treeID = 0L, Z = 0)
    output <- sf::st_as_sf(data, coords = c("x", "y", "z"), crs = st_crs(las))
    output <- output[0,]
  }
  else
  {
    if (uniqueness == "incremental")
    {
      ids <- 1:nrow(maxima)
    }
    else if (uniqueness == "gpstime")
    {
      ids <- las@data[["gpstime"]][res]
    }
    else
    {
      xoffset <- las[["X offset"]]
      yoffset <- las[["Y offset"]]
      zoffset <- las[["Z offset"]]

      xscale  <- las[["X scale factor"]]
      yscale  <- las[["Y scale factor"]]
      zscale  <- las[["Z scale factor"]]

      xscaled <- as.integer((maxima[["X"]] - xoffset)/xscale)
      yscaled <- as.integer((maxima[["Y"]] - yoffset)/yscale)

      ids <- bitmerge(xscaled, yscaled)
    }

    data   <- data.frame(x = maxima[["X"]], y = maxima[["Y"]], z = maxima[["Z"]], treeID = ids, Z = maxima[["Z"]])
    output <- sf::st_as_sf(data, coords = c("x", "y", "z"), crs = st_crs(las))
  }

  sf::st_agr(output) <- "constant"
  return(output)
}

#' @export
locate_trees.RasterLayer = function(las, algorithm, uniqueness = 'incremental')
{
  if (raster_is_proxy(las))
  {
    if (!raster_fits_in_memory(las, n = 2))
      stop("Large on-disk rasters are not supported by locate_tree. Load the raster manually.", call. = FALSE)
    else
      las <- raster_in_memory(las)
  }

  las <- raster_as_las(las)
  return(locate_trees(las, algorithm))
}

#' @export
locate_trees.stars = function(las, algorithm, uniqueness = 'incremental')
{
  if (raster_is_proxy(las))
  {
    if (!raster_fits_in_memory(las, n = 2))
      stop("Large on-disk rasters are not supported by locate_tree. Load the raster manually.", call. = FALSE)
    else
      las <- raster_in_memory(las)
  }

  las <- raster_as_las(las)
  return(locate_trees(las, algorithm))
}

#' @export
locate_trees.SpatRaster = function(las, algorithm, uniqueness = 'incremental')
{
  if (raster_is_proxy(las))
  {
    if (!raster_fits_in_memory(las, n = 2))
      stop("Large on-disk rasters are not supported by locate_tree. Load the raster manually.", call. = FALSE)
    else
      las <- raster_in_memory(las)
  }

  las <- raster_as_las(las)
  return(locate_trees(las, algorithm))
}

#' @export
locate_trees.LAScatalog = function(las, algorithm, uniqueness = 'incremental')
{
  if (uniqueness == "gpstime")
    opt_select(las) <- "xyzt"
  else
    opt_select(las) <- "xyz"

  options <- list(need_buffer = TRUE)
  output  <- catalog_map(las, locate_trees, algorithm = algorithm, uniqueness = uniqueness, .options = options)
  return(output)
}

#' @export
#' @rdname rasterize
rasterize_density = function(las, res = 4, ...)
{
  UseMethod("rasterize_density", las)
}

#' @export
rasterize_density.LAS = function(las, res = 4, ...)
{
  dots <- list(...)
  pkg <- if (is.null(dots$pkg)) getOption("lidR.raster.default") else dots$pkg

  X <- raster_layout(las, res)
  X <- raster_materialize(X, pkg = pkg)

  if (!"pulseID" %in% names(las))
  {
    template <- if (!is_a_number(res)) raster_template(res) else raster_layout(las, res, format = "template")
    count  <- fasterize(las, template, FALSE, "count")

    X <- raster_set_values(X, count)
    raster_names(X) <- "density"
  }
  else
  {
    X <- pixel_metrics(las, ~list(point_density = .N, pulse_density = length(unique(pulseID))), X)
  }

  X <- raster_replace_na(X)
  X <- X/(raster_res(X)[1]^2)
  return(X)
}

#' @export
rasterize_density.LAScatalog = function(las, res = 4, ...)
{
  # Defensive programming
  if (is_a_number(res))
    assert_all_are_non_negative(res)
  else if (!raster_is_supported(res))
    stop("'res' must be a number of a raster.", call. = FALSE)

  # Compute the alignment options including the case when res is a RasterLayer/stars/terra
  alignment <- raster_alignment(res)

  # subset the collection to the size of the layout (if any)
  if (!is_a_number(res)) las <- catalog_intersect(las, res)

  # Enforce some options
  opt_select(las) <- "xyz"
  if (opt_wall_to_wall(las))
    opt_chunk_buffer(las) <- 2

  # Compute the alignment option including the case when res is a RasterLayer
  alignment   <- list(res = res, start = c(0,0))

  if (opt_chunk_size(las) > 0 && opt_chunk_size(las) < 2*alignment$res)
    stop("The chunk size is too small. Process aborted.", call. = FALSE)

  # Processing
  options <- list(need_buffer = TRUE, drop_null = TRUE, raster_alignment = alignment)
  output  <- catalog_map(las, rasterize_density, res = res, ..., .options = options)
  return(output)
}

#' @export
#' @rdname normalize
normalize_height = function(las, algorithm, use_class = c(2L,9L), dtm = NULL, ...)
{
  UseMethod("normalize_height", las)
}

#' @export
normalize_height.LAS = function(las, algorithm, use_class = c(2L,9L), dtm = NULL, ...)
{
  # Non standard evaluation (R CMD check)
  . <- Z <- Zref <- X <- Y <- Classification <- NULL

  lidR.context <- "normalize_height"

  # Ellipsis parsing
  dots <- list(...)
  Wdegenerated <- isTRUE(dots$Wdegenerated)
  add_lasattribute <- isTRUE(dots$add_lasattribute)

  # If algorithm is raster we make a basic substraction
  if (is_raster(algorithm))
  {
    dtm <- algorithm

    if (raster_nlayer(dtm) > 1)
      stop("A DTM must be a single layer raster", call. = FALSE)

    Zg <- raster_value_from_xy(dtm, las$X, las$Y)
    isna <- is.na(Zg)
    nnas <- n_na_not_in_buffer(las, isna)

    # If it remains NAs (not from the buffer) it means that we have points that are not in the DTM
    # (the DTM does not cover the point cloud entirely). We will find a value anyway.
    # Previously the function stopped. It now forces interpolation with NN.
    if (nnas > 0)
    {
      nn <- knnidw(1, rmax = .Machine$double.xmax)
      znn <- nn(raster_as_las(dtm, bbox = st_bbox(las)), las@data[isna, .(X,Y,Z)])
      Zg[isna] <- znn
      warning(glue::glue("{nnas} points do not belong in the raster. Nearest neighbor was used to assign a value."), call. = FALSE)
    }
  }
  # If algorithm is spatial interpolation function
  else if (is.function(algorithm))
  {
    assert_is_algorithm(algorithm)
    assert_is_algorithm_spi(algorithm)

    if (any(as.integer(use_class) != use_class))
      stop("'use_class' is not a vector of integers'", call. = FALSE)

    use_class <- as.integer(use_class)

    if (!"Classification" %in% names(las))
      stop("No attribute 'Classification' found. This attribute is required to interpolate ground points.", call. = FALSE)

    # Select the ground points. If dtm is provided, then the ground point are from the DTM
    if (!is.null(dtm))
    {
      if (raster_nlayer(dtm) > 1)
        stop("A DTM must be a single layer raster", call. = FALSE)

      dtm <- raster_crop(dtm, st_bbox(las))
      ground <- raster_as_las(dtm, st_bbox(las))
    }
    else
    {
      ground  <- las@data[Classification %in% use_class, .(X,Y,Z)]
      if (nrow(ground) == 0) stop("No ground points found. Impossible to compute a DTM.", call. = FALSE)
      ground  <- check_degenerated_points(ground, Wdegenerated)
    }

    # Interpolate the terrain providing what to interpolate (ground) and where
    # to interpolate (the point clouf)
    ground <- LAS(ground, las@header, crs = st_crs(las), check = FALSE, index = las@index)
    Zg <- algorithm(ground, las@data)
    Zg[is.nan(Zg)] <- NA_real_

    # If it remains NAs (not in buffer) it means that we have points very far from ground points
    # and they cannot be interpolated. But we will interpolate them anyway. Previously
    # the function stopped. It now forces interpolation with NN.
    isna <- is.na(Zg)
    nnas <- n_na_not_in_buffer(las, isna)
    if (nnas > 0)
    {
      nn <- knnidw(1, rmax = .Machine$double.xmax)
      znn <- nn(ground, las@data[isna, .(X,Y,Z)])
      Zg[isna] <- znn
      warning(glue::glue("Interpolation of {nnas} points failed because they are too far from ground points. Nearest neighbor was used but interpolation is weak for those points"), call. = FALSE)
    }
  }
  else
  {
    stop(glue::glue("Parameter 'algorithm' is a {class(algorithm)}. Expected type is 'raster' or 'function'"), call. = FALSE)
  }

  las[["Z offset"]] <- 0
  zoffset <- las[["Z offset"]]
  zscale <- las[["Z scale factor"]]

  if (!"Zref" %in% names(las))
    las@data[["Zref"]] <- las[["Z"]]

  las@data[["Z"]] <- las[["Z"]] - Zg

  if (add_lasattribute && is.null(las@header@VLR$Extra_Bytes[["Extra Bytes Description"]][["Zref"]]))
    las <- add_lasattribute_manual(las, name = "Zref", desc = "Elevation above sea level", type = "int", offset = zoffset, scale = zscale)

  fast_quantization(las@data[["Z"]], zscale, zoffset)
  las <- las_update(las)
  las@index$sensor <- las@index$sensor + NLAS
  return(las)
}

#' @export
normalize_height.LAScatalog = function(las, algorithm, use_class = c(2L,9L), dtm = NULL, ...)
{
  opt_select(las) <- "*"

  # Workaround for #580. If normalize height is ran in parallel it will fail with
  # SpatRaster because they are not serializable. SpatRaster are converted to RasterLayer
  # for multicore strategies
  algorithm <- convert_ondisk_spatraster_into_serializable_raster_if_necessary(algorithm)

  options <- list(need_buffer = TRUE, drop_null = TRUE, need_output_file = TRUE)
  output  <- catalog_map(las, normalize_height, algorithm = algorithm, use_class = use_class, dtm = dtm, ..., .options = options)
  return(output)
}

#' @rdname normalize
#' @export
unnormalize_height = function(las)
{
  stopifnotlas(las)

  if ("Zref" %in% names(las))
  {
    las@data[["Z"]] <- las@data[["Zref"]]
    las <- remove_lasattribute(las, "Zref")
    las <- lasupdateheader(las)
  }
  else
    message("No attribute 'Zref' found. Un-normalizisation is impossible.")

  if (las@index$sensor > NLAS) las@index$sensor <- las@index$sensor - NLAS
  return(las)
}

#' @param e1 a LAS object
#' @param e2 A raster representing a digital terrain model in format from `raster`, `stars` or `terra`..
#' @export
#' @md
#' @rdname normalize
setMethod("-", c("LAS"), function(e1, e2)
{
  return(normalize_height(e1,e2))
})

check_degenerated_points = function(points, Wdegenerated = TRUE)
{
  . <- X <- Y <- Z <- NULL

  # test integrity of the data and degenerated points
  dup_xyz  = duplicated(points, by = c("X", "Y", "Z"))
  dup_xy   = duplicated(points, by = c("X", "Y"))
  ndup_xyz = sum(dup_xyz)
  ndup_xy  = sum(dup_xy & !dup_xyz)

  if (ndup_xyz > 0 && Wdegenerated)
    warning(glue::glue("There were {ndup_xyz} degenerated ground points. Some X Y Z coordinates were repeated. They were removed."), call. = FALSE)

  if (ndup_xy > 0 && Wdegenerated)
    warning(glue::glue("There were {ndup_xy} degenerated ground points. Some X Y coordinates were repeated but with different Z coordinates. min Z were retained."), call. = FALSE)

  if (ndup_xy > 0 | ndup_xyz > 0)
    points = points[, .(Z = min(Z)), by = .(X,Y)]

  return(points)
}

n_na_not_in_buffer <- function(las, isna)
{
  nnas <- sum(isna)
  if (nnas == 0) return(0)

  if (!"buffer" %in% names(las))
    return(nnas)

  buffer_na <- las$buffer[isna]
  return(sum(buffer_na == LIDRNOBUFFER))
}

#' @rdname normalize
#' @export
height_above_ground = function(las, ...)
{
  las = normalize_height(las, ...)
  hag = las$Z
  las = unnormalize_height(las)
  las = add_lasattribute(las, hag, "hag", "height above ground")
  las
}

#' @param geom character. Geometry type of the output. Can be 'point', 'convex', 'concave' or 'bbox'.
#' @param concaveman numeric. Only if \code{type = "concave"}. Vector with the two parameters of the
#' function \link{concaveman}.
#' @param attribute character. The column name of the attribute containing tree IDs. Default is
#' \code{"treeID"}
#'
#' @rdname aggregate
#' @export
crown_metrics = function(las, func, geom = "point", concaveman = c(3, 0), attribute = "treeID", ...)
{
  UseMethod("crown_metrics", las)
}

#' @export
crown_metrics.LAS = function(las, func, geom = "point", concaveman = c(3, 0), attribute = "treeID", ...)
{
  assert_is_a_string(attribute)
  if (!attribute %in% names(las)) stop("The trees are not segmented yet. See function 'segment_trees'.", call. = FALSE)
  geom <- match.arg(geom, c("point", "convex", "concave", "bbox"))

  . <- .BY <- X <- Y <- Z <- GRPID <- NULL

  template <- las[[attribute]]

  if (all(is.na(template)))
  {
    warning("No tree found. NULL returned.", call. = FALSE)
    return(NULL)
  }

  if      (geom == "point")   fgeom <- stdtreeapex
  else if (geom == "convex")  fgeom <- stdtreehullconvex
  else if (geom == "concave") fgeom <- stdtreehullconcave
  else if (geom == "bbox")    fgeom <- stdtreehullbbox

  concavity <- concaveman[1]
  length_threshold <- concaveman[2]

  M1 <- las@data[, if (!anyNA(.BY)) fgeom(X,Y,Z, concavity, length_threshold), by = .(GRPID = template)]
  data.table::setorder(M1, GRPID)
  sfgeom <- sf::st_as_sfc(M1[["geom"]])


  ninvalid = 0
  if (is(sfgeom, "sfc_POLYGON"))
  {
    invalid <- !sf::st_is_valid(sfgeom)
    na.invalid <- is.na(invalid)
    invalid[na.invalid] <- TRUE
    ninvalid <- sum(invalid, na.rm = TRUE)
    nna <- sum(na.invalid)

    if (ninvalid > 0)
    {
      if (geom == "concave")
        warning(glue::glue("{ninvalid} invalid polygons created and removed. They likely correspond either to trees with aligned points or to edge cases where the convex hull converged to polygons that are not valid."), call. = FALSE)

      if (geom == "convex")
        warning(glue::glue("{ninvalid} invalid polygons created and removed. They likely correspond to trees with aligned points."), call. = FALSE)
    }
  }

  if (!is.null(func))
  {
    M2 <- template_metrics(las, func, template, ...)
    data.table::setorder(M2, GRPID)
    M2 <- M2[M2$GRPID %in% M1$GRPID]
    data.table::setnames(M2, "GRPID", attribute)
  }
  else
  {
    M2 <- M1[, "GRPID"]
    data.table::setnames(M2, attribute)
  }

  # This is only used for backward compatibility
  # with delineate_crowns() that call crown_metrics()
  xyz <- isTRUE(list(...)$xyz)
  if (xyz)
  {
    f <- function(x,y,z) {
      if (length(x) < 4) { return(NULL) }
      j <- which.max(z)
      return(list(XTOP= x[j], YTOP = y[j], ZTOP = z[j]))
    }

    M3 <- las@data[, if (!anyNA(.BY)) f(X,Y,Z), by = .(GRPID = template)]
    M3[["GRPID"]] <- NULL
    M2 <- cbind(M2, M3)
  }

  output <- sf::st_set_geometry(M2, sfgeom)
  if (ninvalid > 0) output <- output[!invalid,]
  sf::st_crs(output) <- st_crs(las)
  return(output)
}

#' @export
crown_metrics.LAScluster = function(las, func, geom = "point", concaveman = c(3, 0), attribute = "treeID", ...)
{
  x <- readLAS(las)
  if (is.empty(x)) return(NULL)
  metrics <- crown_metrics(x, func, geom, concaveman, attribute, ...)
  if (is.null(metrics)) return(NULL) # fix #733
  bbox <- st_bbox(las)
  sf::st_agr(metrics) <- "constant"
  centroid <- sf::st_centroid(metrics)
  centroid$ID <- 1:nrow(centroid)
  sf::st_agr(centroid) <- "constant"
  centroid <- sf::st_crop(centroid, bbox)
  return(metrics[centroid$ID,])
}

#' @export
crown_metrics.LAScatalog = function(las, func, geom = "point", concaveman = c(3, 0), attribute = "treeID", ...)
{
  assert_is_a_string(attribute)

  is_formula <- tryCatch(lazyeval::is_formula(func), error = function(e) FALSE)
  if (!is_formula && !is.null(func)) func <- lazyeval::f_capture(func)
  globals <- NULL
  if (engine_use_future()) globals <- future::getGlobalsAndPackages(func)

  options <- list(need_buffer = TRUE, drop_null = TRUE, globals = names(globals$globals), automerge = TRUE)
  output  <- catalog_apply(las, crown_metrics, func = func, geom = geom, concaveman = concaveman, attribute = attribute, ..., .options = options)
  return(output)
}

#' An S4 class to represent a collection of .las or .laz files
#'
#' A `LAScatalog` object is a representation of a collection of las/laz files. A `LAScatalog` is
#' a way to manage and batch process a lidar coverage. It allows the user to process a large area, or to
#' selectively clip data from a large area without loading all the data into computer memory.
#' A `LAScatalog` can be built with the function \link{readLAScatalog}.
#'
#' A `LAScatalog` contains an `sf` object to store the geometry and metadata. It is extended with slots
#' that contain processing options. In `lidR`, each function that supports a `LAScatalog` as
#' input will respect these processing options. Internally, processing a catalog is almost always the
#' same and relies on a few steps:\cr
#' 1. Define chunks. A chunk is an arbitrarily-defined region of interest (ROI) of the
#' collection. Altogether, the chunks are a wall-to-wall set of ROIs that encompass the whole dataset.
#' 2. Loop over each chunk (in parallel or not).
#' 3. For each chunk, load the points inside the ROI into R, run some R functions,
#' return the expected output.
#' 4. Merge the outputs of the different chunks once they are all processed to build a continuous
#' (wall-to-wall) output.
#'
#' So basically, a `LAScatalog` is an object that allows for batch processing but with the specificity
#' that `lidR` does not loop through LAS or LAZ files, but loops seamlessly through chunks that do not
#' necessarily match with the file pattern. This way `lidR` can sequentially process tiny ROIs even if
#' each file may be individually too big to fit in memory. This is also why point cloud indexation
#' with `lax` files may significantly speed-up the processing.\cr\cr
#' It is important to note that catalogs with files that overlap each other are not natively supported
#' by `lidR`. When encountering such datasets the user should always filter any overlaps if
#' possible. This is possible if the overlapping points are flagged, for example in the
#' 'withheld' attribute. Otherwise `lidR` will not be able to process the dataset correctly.
#'
#' @slot data sf. An `sf` `data.frame` with the bounding box of each file as well as all the information
#' read from the header of each LAS/LAZ file.
#'
#' @slot processing_options list. A list that contains some settings describing how the collection will be
#' processed (see dedicated section).
#'
#' @slot chunk_options list. A list that contains some settings describing how the collection will be
#' sub-divided into chunks to be processed (see dedicated section).
#'
#' @slot output_options list. A list that contains some settings describing how the collection will return
#' the outputs (see dedicated section).
#'
#' @slot input_options list. A list of parameters to pass to \link{readLAS} (see dedicated section).
#'
#' @slot index list. See \link[=lidR-spatial-index]{spatial indexing}.
#'
#' @section Processing options:
#' The slot `@processing_options` contains a `list` of options that determine how chunks
#' (the sub-areas that are sequentially processed) are processed.
#'
#' - **progress**: boolean. Display a progress bar and a chart of progress. Default is TRUE.
#' Progress estimation can be enhanced by installing the package `progress`. See \link{opt_progress}.
#' - **stop_early**: boolean. Stop the processing if an error occurs in a chunk. If `FALSE`
#' the process can run until the end, removing chunks that failed. Default is TRUE and the user should
#' have no reason to change this. See \link{opt_stop_early}.
#' - **wall.to.wall** logical. The catalog processing engine always guarantees to return a
#' continuous output without edge effects, assuming that the catalog is a wall-to-wall catalog. To do
#' so, some options are checked internally to guard against bad settings, such as `buffer = 0` for an
#' algorithm that requires a buffer. In rare cases it might be useful to disable these controls. If
#' `wall.to.wall = FALSE` controls are disabled and wall-to-wall outputs cannot be guaranteed.
#' See \link{opt_wall_to_wall}
#'
#' @section Chunk options:
#' The slot `@chunk_options` contains a `list` of options that determine how chunks
#' (the sub-areas that are sequentially processed) are made.
#'
#' - **chunk_size**: numeric. The size of the chunks that will be sequentially processed.
#' A small size allows small amounts of data to be loaded at once, saving computer memory.
#' With big chunks the computation is usually faster but uses much more memory. If `chunk_size = 0` the
#' chunk pattern is build using the file pattern. The chunks are expecting to be wall-to-wall coverage,
#' which means that `chunk_size = 0` has meaning only if the files are not overlapping. Default is 0 i.e.
#' by default the processing engine respects the existing tiling pattern. See \link{opt_chunk_size}.
#' - **buffer**: numeric. Each chunk can be read with an extra buffer around it to ensure there are
#' no edge effects between two independent chunks and that the output is continuous. This is mandatory for
#' some algorithms. Default is 30. See \link{opt_chunk_buffer}.
#' - **alignment**: numeric. A vector of size 2 (x and y coordinates, respectively) to align the
#' chunk pattern. By default the alignment is made along (0,0), meaning that the edge of the first chunk
#' will belong on x = 0 and y = 0 and all the the other chunks will be multiples of the chunk size.
#' Not relevant if `chunk_size = 0`. See \link{opt_chunk_alignment}.
#' - **drop**: integers. A vector of integers that specify the IDs of the chunks that should not be
#' created. This is designed to enable users to restart a computation that failed without reprocessing
#' everything. See \link{opt_restart<-}. Technically, this option may be used for partial processing of
#' a collection, but it generally should not be. Partial processing is already a feature of the engine. See
#' [this vignette](https://cran.r-project.org/package=lidR/vignettes/lidR-LAScatalog-engine.html#partial-processing)
#'
#' @section Output options:
#' The slot `@output_options` contains a `list` of options that determine how chunks
#' (the sub-areas that are sequentially processed) are written. By "written" we mean written to files
#' or written in R memory.
#'
#' - **output_files**: string. If `output_files = ""` outputs are returned in R. Otherwise, if
#' `output_files` is a string the outputs will be written to files.
#' This is useful if the output is too big to be returned in R. A path to a filename template
#' without a file extension (the engine guesses it for you) is expected. When several files are going to be
#' written a single string is provided with a template that is automatically filled. For example,
#' the following file names are possible:
#' \preformatted{
#' "/home/user/als/normalized/file_{ID}_segmented"
#' "C:/user/document/als/zone52_{XLEFT}_{YBOTTOM}_confidential"
#' "C:/user/document/als/{ORIGINALFILNAME}_normalized"
#' }
#' This option will generate as many filenames as needed with custom names for each file. The allowed
#' templates are \code{{XLEFT}, {XRIGHT}, {YBOTTOM}, {YTOP}, {ID}, {XCENTER},
#' {YCENTER}, {ORIGINALFILENAME}}. See \link{opt_output_files}.
#' - **drivers**: list. This contains all the drivers required to seamlessly write `Raster*`,
#' `SpatRaster`, `stars`, `Spatial*`, `sf`, and `LAS` objects. It is recommended that only advanced
#' users change this option. A dedicated page describes the drivers in \link{lidR-LAScatalog-drivers}.
#' - **merge**: boolean. Multiple objects are merged into a single object at the end of the processing.
#' See \link{opt_merge}.
#'
#' @section Input options:
#' The slot `@input_options` contains a `list` of options that are passed to the function
#' \link{readLAS}. Indeed, the `readLAS` function is not called directly by the user but by the
#' internal processing engine. Users can propagate these options through the `LAScatalog` settings.
#'
#' - **select**: string. The option `select`. Usually this option is not respected because
#' each function knows which data must be loaded or not. This is documented in each function. See
#' \link{opt_select}.
#' - **filter**: string. The option `filter`. See \link{opt_filter}.
#'
#' @export
#'
#' @examples
#' \dontrun{
#' # Build a catalog
#' ctg <- readLAScatalog("filder/to/las/files/", filter =  "-keep_first")
#'
#'
#' # Summary gives a summary of how the catalog will be processed
#' summary(ctg)
#'
#' # We can seamlessly use lidR functions
#' hmean <- pixel_metrics(ctg, ~~mean(Z), 20)
#' ttops <- tree_detection(ctg, lmf(5))
#'
#' # For low memory config it is probably advisable not to load entire files
#' # and process chunks instead
#' opt_chunk_size(ctg) <- 500
#'
#' # Sometimes the output is likely to be very large
#' # e.g. large coverage and small resolution
#' dtm <- rasterize_terrain(ctg, 1, tin())
#'
#' # In that case it is advisable to write the output(s) to files
#' opt_output_files(ctg) <- "path/to/folder/DTM_chunk_{XLEFT}_{YBOTTOM}"
#'
#' # Raster will be written to disk. The list of written files is returned
#' # or, in this specific case, a virtual raster mosaic.
#' dtm <- rasterize_terrain(ctg, 1, tin())
#'
#' # When chunks are files the original names of the las files can be preserved
#' opt_chunk_size(ctg) <- 0
#' opt_output_files(ctg) <- "path/to/folder/DTM_{ORIGINALFILENAME}"
#' dtm <- rasterize_terrain(ctg, 1, tin())
#'
#' # For some functions, files MUST be written to disk. Indeed, it is certain that R cannot
#' # handle the entire output.
#' opt_chunk_size(ctg) <- 0
#' opt_output_files(ctg) <- "path/to/folder/{ORIGINALFILENAME}_norm"
#' opt_laz_compression(ctg) <- TRUE
#' new_ctg <- normalize_height(ctg, tin())
#'
#' # The user has access to the catalog engine through the functions catalog_apply
#' # and catalog_map
#' output <- catalog_apply(ctg, FUN, ...)
#' }
#' @md
setClass(
  Class = "LAScatalog",
  representation(
    data = "sf",
    chunk_options = "list",
    processing_options = "list",
    output_options = "list",
    input_options = "list",
    index = "list"
  )
)

setMethod("initialize", "LAScatalog", function(.Object)
{
  Raster = NULL
  stars = NULL
  spatial = NULL

  if (system.file(package='raster') != "" && system.file(package='sp') != "")
  {
    Raster = list(
      write = raster::writeRaster,
      extension = ".tif",
      object = "x",
      path = "filename",
      param = list(format = "GTiff", NAflag = -999999))
  }

  if (system.file(package='stars') != "")
  {
    stars = list(
      write = stars::write_stars,
      extension = ".tif",
      object = "obj",
      path = "dsn",
      param = list(NA_value = -999999))
  }

  if (system.file(package='sp') != "")
  {
    spatial = list(
      write = writeSpatial,
      extension = ".shp",
      object = "x",
      path = "filename",
      param = list(overwrite = FALSE))
  }

  drivers = list(
    Raster = Raster,
    stars = stars,
    Spatial = spatial,
    SpatRaster = list(
      write = terra::writeRaster,
      extension = ".tif",
      object = "x",
      path = "filename",
      param = list(overwrite=FALSE,
                   NAflag = -999999)
    ),
    SpatVector = list(
      write = terra::writeVector,
      extension = ".shp",
      object = "x",
      path = "filename",
      param = list(overwrite=FALSE)
    ),
    LAS = list(
      write = lidR::writeLAS,
      extension = ".las",
      object = "las",
      path = "file",
      param = list()
    ),
    sf = list(
      write = sf::st_write,
      extension = ".shp",
      object = "obj",
      path = "dsn",
      param = list(quiet = TRUE)
    ),
    data.frame = list(
      write = data.table::fwrite,
      extension = ".txt",
      object = "x",
      path = "file",
      param = list()
    )
  )

  data <- structure(list(
    File.Signature = "LASF",
    File.Source.ID = 0L,
    GUID = "",
    Version.Major = 1L,
    Version.Minor = 2L,
    System.Identifier = "",
    Generating.Software = "",
    File.Creation.Day.of.Year = 0L,
    File.Creation.Year = 0L,
    Header.Size = 227L,
    Offset.to.point.data = 321,
    Number.of.variable.length.records = 1,
    Point.Data.Format.ID = 1L,
    Point.Data.Record.Length = 28L,
    Number.of.point.records = 0L,
    X.scale.factor = 0.01,
    Y.scale.factor = 0.01,
    Z.scale.factor = 0.01,
    X.offset = 0,
    Y.offset = 0,
    Z.offset = 0,
    Max.X = 0,
    Min.X = 0,
    Max.Y = 0,
    Min.Y = 0,
    Max.Z = 0,
    Min.Z = 0,
    CRS = 0L,
    Number.of.1st.return = 0L,
    Number.of.2nd.return = 0L,
    Number.of.3rd.return = 0L,
    Number.of.4th.return = 0L,
    Number.of.5th.return = 0L,
    filename = ""),
    row.names = c(NA, -1L), class = "data.frame")

  data <- sf::st_sf(data, geometry = sf::st_sfc(lapply(1, function(x) sf::st_polygon() )))
  data <- data[-1,]

  .Object@data <- data

  .Object@chunk_options <- list(
    size = 0,
    buffer = 30,
    alignment = c(0,0),
    drop = NULL
  )

  .Object@processing_options <- list(
    progress = TRUE,
    stop_early = TRUE,
    wall_to_wall = TRUE
  )

  .Object@output_options <- list(
    output_files = "",
    drivers = drivers,
    merge = TRUE
  )

  .Object@input_options <- list(
    select = "*",
    filter = "",
    alt_dir = ""
  )

  .Object@index <- LIDRDEFAULTINDEX

  return(.Object)
})

#' Local Maximum Filter
#'
#' Generic local maximum filter. For individual tree detection use \link{locate_trees} with the
#' \link{lmf} algorithm that is more adequate for ITD. This function is a more generic method for
#' multiple purposes other than tree segmentation. This function is natively parallelized with OpenMP.
#'
#' @param las An object of class LAS
#' @param w numeric. Window shape. 1 number for the diameter of a disc, 2 numbers for a rectangle
#' (width, height), 3 numbers for an oriented rectangle (width, height, angle). The angle must be
#' in radians.
#' @param filter formula. Memory efficient way to work only with a subset of the data without creating a copy
#' of the data.
#' @return \code{sf} with attributes from the corresponding point in the LAS object,
#' @examples
#' LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
#' las <- readLAS(LASfile, select = "xyzi", filter = "-drop_z_below 0 -keep_random_fraction 0.5")
#'
#' # Using a 20x5 rectangle with a 45 degrees angle.
#' # This won't find the tree properly in the general case
#' # but may find some oriented structure.
#' lm = find_localmaxima(las, c(20, 5, pi/4))
#' plot(lm)
#' @noRd
locate_localmaxima = function(las, w, filter = NULL)
{
  assert_all_are_positive(w[1])
  if (!is.na(w[2])) assert_all_are_positive(w[2])
  if (!is.na(w[3])) assert_all_are_in_open_range(w[3], -pi/2, pi/2)
  if (length(w) == 0 || length(w) > 3) stop("Invalid window.", call. = FALSE)

  if (length(w) == 1) verbose("Local maxima search in a disc")
  if (length(w) == 2) verbose("Local maxima search in a rectangle")
  if (length(w) == 3) verbose("Local maxima search in an oriented rectangle")

  filter <- parse_filter(las, filter)
  id <- C_local_maximum(las, w, filter, getThreads())

  maxima <- las@data[id]
  data.table::setDF(maxima)
  no_maxima <- nrow(maxima) == 0

  if (no_maxima) maxima <- las@data[1,]

  output <- sf::st_as_sf(maxima, coords = c("X", "Y"), crs = sf::st_crs(las))

  if (no_maxima) output <- output[0,]

  return(output)
}

test_that("locate local maxima works", {
  skip_on_cran()

  u <- lidR:::locate_localmaxima(example, 3)
  expect_is(u, "sf")
  expect_equal(dim(u), c(3,15))

  u <- lidR:::locate_localmaxima(example, c(15, 3))
  expect_equal(dim(u), c(1,15))

  u <- lidR:::locate_localmaxima(example, c(3, 15))
  expect_equal(dim(u), c(3,15))

  u <- lidR:::locate_localmaxima(example, c(15, 3, pi/4))
  expect_equal(dim(u), c(3,15))
})

#' Deprecated functions in lidR
#'
#' These functions are provided for compatibility with older versions of lidR but are deprecated.
#'
#' @param las,res See the new functions that replace the old ones
#' @param files,select,filter,sort See the new functions that replace the old ones
#' @param folder,... See the new functions that replace the old ones
#'
#' @rdname deprecated
#' @name deprecated
#'
#' @include io_readXLAS.R
#' @include io_readLAScatalog.R
NULL

# nocov start

#' @export
#' @rdname deprecated
readALSLAS = readALS

#' @export
#' @rdname deprecated
readTLSLAS = readTLS

#' @export
#' @rdname deprecated
readUAVLAS = readTLS

#' @export
#' @rdname deprecated
readDAPLAS = readTLS

#' @export
#' @rdname deprecated
readALSLAScatalog = readALScatalog

#' @export
#' @rdname deprecated
readTLSLAScatalog = readTLScatalog

#' @export
#' @rdname deprecated
readUAVLAScatalog = readTLScatalog

#' @export
#' @rdname deprecated
readDAPLAScatalog = readTLScatalog

#' @export
#' @rdname deprecated
filter_surfacepoints = function(las, res)
{
  UseMethod("filter_surfacepoints", las)
}

#' @rdname deprecated
#' @export
filter_surfacepoints.LAS = function(las, res)
{
  return(decimate_points(las, highest(res)))
}

#' @rdname deprecated
#' @export
filter_surfacepoints.LAScatalog = function(las, res)
{
  opt_select(las)       <- "*"
  opt_chunk_buffer(las) <- res
  options <- list(need_buffer = FALSE, drop_null = TRUE, need_output_file = TRUE)
  output  <- catalog_map(las, filter_surfacepoints, res = res, .options = options)
  return(output)
}

.lidr3depreciation <- function(name)
{
  # no depreciation v3.0.0
  #return(invisible())

  # message v3.1.0
  #msg = paste(as.character(sys.call(sys.parent()))[1L], "is deprecated. Use", name, "instead.")
  #message(msg)
  #return(invisible())

  # warning v3.2.0
  #msg = paste(as.character(sys.call(sys.parent()))[1L], "is deprecated. Use", name, "instead.")
  #warning(msg, call. = FALSE)
  #return(invisible())

  # error v3.3.0
  msg = paste(as.character(sys.call(sys.parent()))[1L], "is defunct. Use", name, "instead.")
  stop(msg, call. = FALSE)
  return(invisible())
}

# nocov end

#' Functions for the LAScatalog processing engine not meant to be called directly by users
#'
#' Functions for the LAScatalog processing engine not meant to be called directly by users.
#' They are exported for debugging and to simplify export of internal functions when processing in
#' parallel
#'
#' @param ctg an object of class \code{LAScatalog}
#' @param x LAS, Raster, stars, SpatRaster,sf, sfc, Spatial
#' @name engine
#' @rdname engine
#' @family LAScatalog processing engine
NULL

# ====== POINTS-TO-RASTER =======

#' Digital Surface Model Algorithm
#'
#' This function is made to be used in \link{rasterize_canopy}. It implements an algorithm for digital
#' surface model computation based on a points-to-raster method: for each pixel of the output raster
#' the function attributes the height of the highest point found. The \code{subcircle} tweak replaces
#' each point with 8 points around the original one. This allows for virtual 'emulation' of the fact
#' that a lidar point is not a point as such, but more realistically a disc. This tweak densifies the
#' point cloud and the resulting canopy model is smoother and contains fewer 'pits' and empty pixels.
#'
#' @param subcircle numeric. Radius of the circles. To obtain fewer empty pixels the algorithm
#' can replace each return with a circle composed of 8 points (see details).
#'
#' @param na.fill function. A function that implements an algorithm to compute spatial interpolation
#' to fill the empty pixel often left by points-to-raster methods. \code{lidR} has \link{knnidw},
#' \link{tin}, and \link{kriging} (see also \link{rasterize_terrain} for more details).
#'
#' @export
#'
#' @family digital surface model algorithms
#'
#' @examples
#' LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
#' las <- readLAS(LASfile)
#' col <- height.colors(50)
#'
#' # Points-to-raster algorithm with a resolution of 1 meter
#' chm <- rasterize_canopy(las, res = 1, p2r())
#' plot(chm, col = col)
#'
#' # Points-to-raster algorithm with a resolution of 0.5 meters replacing each
#' # point by a 20 cm radius circle of 8 points
#' chm <- rasterize_canopy(las, res = 0.5, p2r(0.2))
#' plot(chm, col = col)
#'
#' \dontrun{
#' chm <- rasterize_canopy(las, res = 0.5, p2r(0.2, na.fill = tin()))
#' plot(chm, col = col)
#' }
#' @name dsm_point2raster
p2r = function(subcircle = 0, na.fill = NULL)
{
  assert_is_a_number(subcircle)
  assert_all_are_non_negative(subcircle)

  subcircle <- lazyeval::uq(subcircle)
  na.fill   <- lazyeval::uq(na.fill)

  if (!is.null(na.fill) && !is(na.fill, "SpatialInterpolation"))
    stop("'na.fill' is not an algorithm for spatial interpolation")

  f = function(las, layout)
  {
    assert_is_valid_context(LIDRCONTEXTDSM, "p2r")

    dsm <- fasterize(las, layout, subcircle, "max")

    if (!is.null(na.fill))
    {
      verbose("Interpolating empty cells...")

      layout[[1]][] <- NA_real_
      layout[[1]][] <- dsm
      hull <- st_convex_hull(las)
      hull <- sf::st_buffer(hull, dist = raster_res(layout)[1])

      where <- raster_as_dataframe(layout, xy = FALSE, na.rm = FALSE)
      where <- where[is.na(where$Z)]
      where <- sf::st_multipoint(as.matrix(where[,1:2]))
      where <- sf::st_geometry(where)
      where <- sf::st_set_crs(where, sf::st_crs(hull))
      where <- sf::st_intersection(where, hull)
      where <- sf::st_coordinates(where)
      if (ncol(where) == 3L) where = where[,-3L]
      where <- as.data.frame(where)
      data.table::setnames(where, c("X", "Y"))

      grid  <- raster_as_las(layout)

      lidR.context <- "p2r"
      cells <- raster_cell_from_xy(layout, where$X, where$Y)
      layout[[1]][cells] <- na.fill(grid, where)

      dsm <- as.numeric(layout[[1]])
    }

    return(dsm)
  }

  f <- plugin_dsm(f)
  return(f)
}

# ====== STRICT TRIANGULATION =======

#' Digital Surface Model Algorithm
#'
#' This function is made to be used in \link{rasterize_canopy}. It implements an algorithm for digital
#' surface model computation using a Delaunay triangulation of first returns with a linear interpolation
#' within each triangle.
#'
#' @param max_edge numeric. Maximum edge length of a triangle in the Delaunay triangulation.
#' If a triangle has an edge length greater than this value it will be removed to trim dummy interpolation
#' on non-convex areas. If \code{max_edge = 0} no trimming is done (see examples).
#' @param highest bool. By default it keeps only the highest point per pixel before to triangulate to
#' decrease computation time. If highest = FALSE all first returns are used.
#'
#' @export
#'
#' @family digital surface model algorithms
#'
#' @examples
#' LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
#' las <- readLAS(LASfile)
#' col <- height.colors(50)
#'
#' # Basic triangulation and rasterization of first returns
#' chm <- rasterize_canopy(las, res = 1, dsmtin())
#' plot(chm, col = col)
#'
#' \dontrun{
#' # Potentially complex concave subset of point cloud
#' x = c(481340, 481340, 481280, 481300, 481280, 481340)
#' y = c(3812940, 3813000, 3813000, 3812960, 3812940, 3812940)
#' las2 = clip_polygon(las,x,y)
#' plot(las2)
#'
#' # Because the TIN interpolation is done within the convex hull of the point cloud
#' # dummy pixels are interpolated that are correct according to the interpolation method
#' # used, but meaningless in our CHM
#' chm <- rasterize_canopy(las2, res = 0.5, dsmtin())
#' plot(chm, col = col)
#'
#' # Use 'max_edge' to trim dummy triangles
#' chm = rasterize_canopy(las2, res = 0.5, dsmtin(max_edge = 3))
#' plot(chm, col = col)
#' }
#' @name dsm_tin
dsmtin = function(max_edge = 0, highest = TRUE)
{
  max_edge <- lazyeval::uq(max_edge)
  return(pitfree(0, c(max_edge, 0), 0, highest))
}

# ====== PIT-FREE =======

#' Digital Surface Model Algorithm
#'
#' This function is made to be used in \link{rasterize_canopy}. It implements the pit-free algorithm
#' developed by Khosravipour et al. (2014), which is based on the computation of a set of classical
#' triangulations at different heights (see references). The \code{subcircle} tweak replaces each
#' point with 8 points around the original one. This allows for virtual 'emulation' of the fact that
#' a lidar point is not a point as such, but more realistically a disc. This tweak densifies the point
#' cloud and the resulting canopy model is smoother and contains fewer 'pits' and empty pixels.
#'
#' @param subcircle numeric. radius of the circles. To obtain fewer empty pixels the algorithm
#' can replace each return with a circle composed of 8 points (see details).
#' @param thresholds numeric. Set of height thresholds according to the Khosravipour et al. (2014) algorithm
#' description (see references)
#' @param max_edge numeric. Maximum edge length of a triangle in the Delaunay triangulation.
#' If a triangle has an edge length greater than this value it will be removed. The first number is the value
#' for the classical triangulation (threshold = 0, see also \link{dsmtin}), the second number
#' is the value for the pit-free algorithm (for thresholds > 0). If \code{max_edge = 0} no trimming
#' is done (see examples).
#' @param highest bool. By default it keeps only the highest point per pixel before to triangulate to
#' decrease computation time. If highest = FALSE all first returns are used.
#'
#' @references Khosravipour, A., Skidmore, A. K., Isenburg, M., Wang, T., & Hussin, Y. A. (2014).
#' Generating pit-free canopy height models from airborne lidar. Photogrammetric Engineering &
#' Remote Sensing, 80(9), 863-872.
#'
#' @export
#'
#' @family digital surface model algorithms
#'
#' @examples
#' LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
#' poi = "-drop_z_below 0 -inside 481280 3812940 481330 3812990"
#' las <- readLAS(LASfile, filter = poi)
#' col <- height.colors(50)
#'
#' # Basic triangulation and rasterization of first returns
#' chm <- rasterize_canopy(las, res = 0.5, dsmtin())
#' plot(chm, col = col)
#'
#' # Khosravipour et al. pitfree algorithm
#' chm <- rasterize_canopy(las, res = 0.5, pitfree(c(0,2,5,10,15), c(0, 1.5)))
#' plot(chm, col = col)
#'
#' \dontrun{
#' # Potentially complex concave subset of point cloud
#' x = c(481340, 481340, 481280, 481300, 481280, 481340)
#' y = c(3812940, 3813000, 3813000, 3812960, 3812940, 3812940)
#' las2 = clip_polygon(las,x,y)
#' plot(las2)
#'
#' # Because the TIN interpolation is done within the convex hull of the point cloud
#' # dummy pixels are interpolated that are correct according to the interpolation
#' # method used, but meaningless in our CHM
#' chm <- rasterize_canopy(las2, res = 0.5, pitfree(max_edge = c(0, 1.5)))
#' plot(chm, col = col)
#'
#' chm = rasterize_canopy(las2, res = 0.5, pitfree(max_edge = c(3, 1.5)))
#' plot(chm, col = col)
#' }
#' @export
#' @name dsm_pitfree
pitfree <- function(thresholds = c(0, 2, 5, 10, 15), max_edge = c(0, 1), subcircle = 0, highest = TRUE)
{
  assert_is_numeric(thresholds)
  assert_all_are_non_negative(thresholds)
  assert_is_numeric(max_edge)
  assert_all_are_non_negative(max_edge)
  assert_is_a_number(subcircle)
  assert_all_are_non_negative(subcircle)
  assert_is_a_bool(highest)

  if (length(thresholds) > 1L & length(max_edge) < 2L)
  {
    stop("'max_edge' should contain 2 numbers")
  }

  thresholds <- lazyeval::uq(thresholds)
  max_edge <- lazyeval::uq(max_edge)
  subcircle <- lazyeval::uq(subcircle)
  highest <- lazyeval::uq(highest)

  f <- function(las, layout)
  {
    assert_is_valid_context(LIDRCONTEXTDSM, "pitfree")

    if (!"ReturnNumber" %in% names(las)) stop("No attribute 'ReturnNumber' found. This attribute is needed to extract first returns", call. = FALSE)
    if (fast_countequal(las$ReturnNumber, 1L) == 0) stop("No first returns found. Operation aborted.", call. = FALSE)

    # Non standard evaluation (R CMD check)
    . <- .N <- X <- Y <- Z <- ReturnNumber <- NULL

    # Delaunay triangulation with boost requiere to
    # compute back integer coordinates
    xscale  <- las[["X scale factor"]]
    yscale  <- las[["Y scale factor"]]
    xoffset <- las[["X offset"]]
    yoffset <- las[["Y offset"]]
    scales  <- c(xscale, yscale)
    offsets <- c(xoffset, yoffset)

    # Get only first returns and coordinates (nothing else needed)
    verbose("Selecting first returns...")
    cloud <- las@data[ReturnNumber == 1L, .(X, Y, Z)]
    cloud <- LAS(cloud, las@header, st_crs(las), check = FALSE, index = las@index)

    # subcircle the data
    if (subcircle > 0)
    {
      verbose("Subcircling points...")
      cloud <- subcircle(cloud, subcircle, 8L)
    }

    if (highest)
    {
      verbose("Selecting only the highest points within the grid cells...")
      template <- raster_template(layout)
      i <- C_highest(cloud, template)
      cloud <- cloud[i]
      if (nrow(cloud) < 3) stop("There are not enought points to triangulate.", call. = FALSE)
    }

    # Get interpolation grid
    grid <- raster_as_dataframe(layout, na.rm = FALSE)

    # Initialize the interpolated values with NAs
    z <- rep(NA_real_, nrow(grid))

    # Perform the triangulation and the rasterization (1 loop for classical triangulation, several for Khosravipour et al.)
    thresholds <- sort(thresholds)
    for (i in seq_along(thresholds))
    {
      verbose(glue::glue("Triangulation pass {i} of {length(thresholds)}..."))
      th <- thresholds[i]
      edge <- if (th == 0) max_edge[1] else max_edge[2]

      if (fast_countover(cloud$Z, th) > 3)
      {
        b <- cloud$Z >= th
        cloud <- cloud[b]
        Ztemp <- interpolate_delaunay(cloud@data, grid, edge, scales, offsets)

        if (i == 1 && all(is.na(Ztemp)))
        {
          stop("Interpolation failed in the first layer (NAs everywhere). Maybe there are too few points.", call. = FALSE)
        }

        z <- pmax(z, Ztemp, na.rm = T)
      }
    }

    if (all(is.na(z)))
      stop("Interpolation failed (NAs everywhere). Input parameters might be wrong.", call. = FALSE)

    return(z)
  }

  f <- plugin_dsm(f, omp = TRUE)
  return(f)
}

# ====== SPIKE-FREE =======

#' Digital Surface Model Algorithm
#'
#' The spike-free algorithm developed by Khosravipour et al. (2016), which is based on the computation
#' of an incremental triangulation of all returns with triangle freezing criteria (see references).
#'
#' @param freeze_distance freeze distance (see references). Recommended value: 3 times the pulse spacing or
#' a little higher.
#' @param height_buffer buffer distance (see references). Recommended value: 0.5 do not change.
#'
#' @references Khosravipour, Anahita & Skidmore, Andrew & Isenburg, Martin. (2016). Generating spike-free
#' digital surface models using LiDAR raw point clouds: A new approach for forestry applications.
#' International Journal of Applied Earth Observation and Geoinformation. 52. 104-114. 10.1016/j.jag.2016.06.005.
#'
#' @export
#'
#' @family digital surface model algorithms
#'
#' @examples
#' LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
#' poi = "-drop_z_below 0 -inside 481280 3812940 481330 3812990"
#' las <- readLAS(LASfile, filter = poi)
#' col <- height.colors(50)
#'
#' # Khosravipour et al. spikefree algorithm
#' chm <- rasterize_canopy(las, res = 0.25, spikefree(1.5))
#' plot(chm, col = col)
#'
#' # Locally adaptive spikefree algorithm
#' chm <- rasterize_canopy(las, res = 0.25, lspikefree())
#' plot(chm, col = col)
#' @export
#' @name dsm_spikefree
spikefree <- function(freeze_distance, height_buffer = 0.5)
{
  assert_is_numeric(freeze_distance)
  assert_all_are_non_negative(freeze_distance)
  assert_all_are_non_negative(height_buffer)
  assert_is_a_number(height_buffer)

  freeze_distance <- lazyeval::uq(freeze_distance)
  height_buffer <- lazyeval::uq(height_buffer)

  f <- function(las, layout)
  {
    assert_is_valid_context(LIDRCONTEXTDSM, "spikefree")
    grid <- raster_as_dataframe(layout, na.rm = FALSE)
    z <- C_spikefree(las@data, grid, freeze_distance, height_buffer)
    return(z)
  }

  f <- plugin_dsm(f, omp = TRUE)
  return(f)
}

#' Locally adaptative spikefree
#'
#' The locally adaptive spikefree algorithm by Fischer et al. 2024 adjusts the freeze distance
#' automatically and locally depending on point density. Within the typical range of airborne laser
#' scan densities (>1 points/m2), the original 3 parameter formula can be simplified to a
#' linear relationship. This linear formula predicts freeze distance from local point spacing
#' (freeze distance ~ intercept + slope*point_spacing) and can be used to adapt the degree
#' of spike-removal to sampling density, so that data gaps are filled and canopy height estimates
#' are consistent across point densities. By default, the point spacing is computed at a resolution
#' of 5 m. It is recommended to apply this algorithm to first returns only and not at a point
#' spacing > 0.7 m.
#'
#'
#' @param slope,intercept default values selected according to Fisher F. J. (2024). It is recommended
#' to do not change.
#'
#' @references Fischer, F. J., Jackson, T., Vincent, G., & Jucker, T. (2024). Robust characterisation
#' of forest structure from airborne laser scanning—A systematic assessment and sample workflow for
#' ecologists. Methods in Ecology and Evolution, 15, 1873–1888. https://doi.org/10.1111/2041-210X.14416
#'
#' @rdname dsm_spikefree
#' @export
lspikefree <- function(slope = 1.96, intercept = 1.42, height_buffer = 0.5)
{
  assert_all_are_non_negative(height_buffer)
  assert_is_a_number(height_buffer)

  height_buffer <- lazyeval::uq(height_buffer)

  f <- function(las, layout)
  {
    assert_is_valid_context(LIDRCONTEXTDSM, "lspikefree")
    grid <- raster_as_dataframe(layout, na.rm = FALSE)
    z <- C_spikefree(las@data, grid, 0, height_buffer, slope, intercept)
    return(z)
  }

  f <- plugin_dsm(f, omp = TRUE)
  return(f)
}

subcircle = function(las, r, n)
{
  xscale  <- las[["X scale factor"]]
  yscale  <- las[["Y scale factor"]]
  xoffset <- las[["X offset"]]
  yoffset <- las[["Y offset"]]
  bbox    <- st_bbox(las)

  dt <- las@data
  X <- Y <- Z <- NULL

  f = function(x, y, z, px, py)
  {
    x = x + px
    y = y + py
    z = rep(z, length(px))

    list(X = x, Y = y, Z = z)
  }

  n = n + 1

  alpha = seq(0, 2*pi, length.out = n)[-n]
  px = r*cos(alpha)
  py = r*sin(alpha)

  dt <- dt[, f(X, Y, Z, px, py), by = 1:nrow(dt)][, nrow := NULL][]
  dt <- dt[data.table::between(X, bbox$xmin, bbox$xmax) & data.table::between(Y, bbox$ymin, bbox$ymax)]
  dt[1:.N, `:=`(X = round_any(X - xoffset, xscale) + xoffset, Y = round_any(Y - yoffset, yscale) + yoffset)]
  return(LAS(dt, las@header, st_crs(las), check = FALSE, index = las@index))
}

LAScluster <- function(center, width, height, buffer, shape, files, name, crs = sf::NA_crs_, index = LIDRDEFAULTINDEX)
{
  return(new("LAScluster", center, width, height, buffer, shape, files, name, "", crs, index))
}

catalog_overlaps <- function(catalog)
{
  geom <- sf::st_geometry(st_as_sf(catalog))
  ii <- sf::st_intersects(geom)

  if (is_lascatalog_v3(catalog)) catalog <- lascatalog_v3_repair(catalog)

  intersections <- vector("list", length(ii))
  for (i in 1:length(ii))
  {
    k <- ii[[i]]
    k <- k[k != i]

    if (length(k) == 0) next

    p <- geom[k]
    q <- geom[i]
    w <- sf::st_intersection(q, p)

    intersections[[i]] <- w
  }

  intersections <- do.call(c, intersections)
  return(intersections)
}

LIDRSPATIALINDEXES  <- c("auto", "gridpartition", "voxelpartition", "quadtree", "octree", "sparcepartition")
LIDRAUTOINDEX       <- 0L
LIDRGRIDPARTITION   <- 1L
LIDRVOXELPARTITION  <- 2L
LIDRQUADTREE        <- 3L
LIDROCTREE          <- 4L
LIDRSPARSEPARTITION <- 5L

LIDRAQUISITIONDEVICES <- c("unknown", "als", "tls", "uav", "dap", "multispectral")
UKNLAS <- 0L
ALSLAS <- 1L
TLSLAS <- 2L
UAVLAS <- 3L
DAPLAS <- 4L
MLSLAS <- 5L
NLAS   <- 10L

LIDRDEFAULTINDEX <- list(sensor = UKNLAS, index = LIDRAUTOINDEX)
LIDRALSINDEX     <- list(sensor = ALSLAS, index = LIDRAUTOINDEX)
LIDRTLSINDEX     <- list(sensor = TLSLAS, index = LIDRAUTOINDEX)
LIDRUAVINDEX     <- list(sensor = UAVLAS, index = LIDRAUTOINDEX)
LIDRDAPINDEX     <- list(sensor = DAPLAS, index = LIDRAUTOINDEX)
LIDRMLSINDEX     <- list(sensor = MLSLAS, index = LIDRAUTOINDEX)

#' @rdname lidR-spatial-index
#' @export
sensor = function(las, h = FALSE)
{
  if (isTRUE(h))
    index = format_index(las)
  else
    index = las@index

  return(index[["sensor"]])
}

#' @rdname lidR-spatial-index
#' @export
`sensor<-` = function(las, value)
{
  if (is.character(value))
  {
    assert_is_a_string(value)
    value <- tolower(value)
    match.arg(value, LIDRAQUISITIONDEVICES)
    value <- match(value, LIDRAQUISITIONDEVICES) - 1
  }
  else if (is.numeric(value))
  {
    assert_is_a_number(value)
    assert_all_are_in_closed_range(value, 0L, length(LIDRAQUISITIONDEVICES)-1)
    value <- as.integer(value)
  }
  else
  {
    stop("'value' must be an integer or a string.", call. = FALSE)
  }

  las@index[["sensor"]] <- value
  return(las)
}

#' @rdname lidR-spatial-index
#' @export
index = function(las, h = FALSE)
{
  if (isTRUE(h))
    index = format_index(las)
  else
    index = las@index

  return(index[["index"]])
}

#' @rdname lidR-spatial-index
#' @export
`index<-` = function(las, value)
{
  if (is.character(value))
  {
    assert_is_a_string(value)
    value <- tolower(value)
    match.arg(value, LIDRSPATIALINDEXES)
    value <- match(value, LIDRSPATIALINDEXES) - 1L
  }
  else if (is.numeric(value))
  {
    assert_is_a_number(value)
    assert_all_are_in_closed_range(value, 0L, length(LIDRSPATIALINDEXES)-1)
    value <- as.integer(value)
  }
  else
  {
    stop("'value' must be an integer or a string.", call. = FALSE)
  }

  las@index[["index"]] <- value
  return(las)
}

format_index = function(las)
{
  index = las@index
  sensor = index$sensor
  index = index$index

  sensor_names <- c("not registered", "ALS", "TLS", "UAV", "DAP", "mutispectral ALS", NA_character_, NA_character_, NA_character_, NA_character_,
                    "not registered (normalized)", "ALS (normalized)", "TLS (normalized)", "UAV (normalized)", "DAP (normalized)", "mutispectral ALS (normalized)", NA_character_, NA_character_, NA_character_, NA_character_)
  index_names <- c("automatic", "grid partition", "voxel partition", "quadtree", "octree")

  sensor <- sensor_names[sensor + 1]
  if (is.na(sensor)) stop("invalid sensor registred", call. = FALSE)

  index = index_names[index + 1]
  if (is.na(index)) stop("invalid spatial index registered", call. = FALSE)

  return(return(list(sensor = sensor, index = index)))
}

`force_autoindex<-` = function(las, value)
{
  if (index(las) == LIDRAUTOINDEX)
    index(las) <- value

  return(las)
}

#' @export
#' @rdname segment
segment_shapes = function(las, algorithm, attribute = "Shape", filter = NULL)
{
  UseMethod("segment_shapes", las)
}

#' @export
segment_shapes.LAS = function(las, algorithm, attribute = "Shape", filter = NULL)
{
  stopif_forbidden_name(attribute)
  assert_is_a_string(attribute)
  assert_is_algorithm(algorithm)
  assert_is_algorithm_shp(algorithm)
  lidR.context <- "segment_shapes"

  filter    <- parse_filter(las, filter)
  output    <- algorithm(las, filter)
  las@data[[attribute]] <- output
  return(las)
}

#' @export
segment_shapes.LAScatalog = function(las, algorithm, attribute = "Shape", filter = NULL)
{
  stopif_forbidden_name(attribute)
  assert_is_algorithm(algorithm)
  assert_is_algorithm_shp(algorithm)

  opt_select(las) <- "*"

  options <- list(need_buffer = TRUE, drop_null = TRUE, need_output_file = TRUE)
  output  <- catalog_map(las, segment_shapes,  algorithm = algorithm, attribute = attribute, filter = filter, .options = options)
  return(output)
}

#' @export
#' @rdname catalog_subset
catalog_intersect = function(ctg, y, ..., subset = c("subset", "flag_unprocessed", "flag_processed"))
{
  assert_is_all_of(ctg, "LAScatalog")
  subset <- match.arg(subset)

  if (is_lascatalog_v3(ctg)) ctg <- lascatalog_v3_repair(ctg)

  i <- NULL

  if (is(y, "Extent"))
  {
    y <- sf::st_bbox(y)
    sf::st_crs(y) <- st_crs(ctg)
  }

  if (is(y, "SpatExtent"))
  {
    y <- sf::st_bbox(y)
    sf::st_crs(y) <- st_crs(ctg)
  }

  if (is_raster(y))
    y <- raster_bbox(y)

  if (is(y, "bbox"))
    y <- sf::st_as_sfc(y)

  if (is(y, "Spatial"))
    y <- sf::st_as_sf(y)

  if (is(y, "sf"))
    y <- sf::st_geometry(y)

  if (is(y, "sfc"))
  {
    sfctg <- ctg@data
    i <- if (is(y, "sfc_POINT")) sf::st_within(y, sfctg, sparse = T) else sf::st_intersects(y, sfctg, sparse = T)
    i <- Filter(function(x) {length(x) > 0}, i)
    i <- Reduce(c, i)
    i <- unique(i)
  }

  if (is.null(i))
    stop("Not supported input geometry", call. = FALSE)

  if (subset == "subset") {
    ctg <- ctg[i,]
  } else if (subset == "flag_unprocessed") {
    ctg$processed <- TRUE
    ctg$processed[i] <- FALSE
  } else {
    ctg$processed <- FALSE
    ctg$processed[i] <- TRUE
  }

  return(ctg)
}

# ======== GENERIC =========

#' Clip points in regions of interest
#'
#' Clip points within a given region of interest (ROI) from a point cloud (LAS object) or a collection
#' of files (LAScatalog object).
#'
#' @template param-las
#' @param geometry a geometric object. spatial points, spatial polygons in sp or sf/sfc format, Extent,
#' bbox, 2x2 matrix
#' @param xleft,ybottom,xright,ytop numeric. coordinates of one or several rectangles.
#' @param xpoly,ypoly numeric. x coordinates of a polygon.
#' @param xcenter,ycenter numeric. x coordinates of on or several disc centres.
#' @param radius numeric. disc radius or radii.
#' @param ... in `clip_roi`: optional supplementary options (see supported geometries). Unused in
#' other functions
#'
#' @section Non-supported LAScatalog options:
#' The option `chunk size`, `buffer`, `chunk alignment` and `select` are not supported by `clip_*`
#' because they are meaningless in this context.
#'
#' @return If the input is a LAS object: an object of class LAS, or a `list` of LAS objects if the
#' query implies several regions of interest.\cr\cr
#' If the input is a LAScatalog object: an object of class LAS, or a `list` of LAS
#' objects if the query implies several regions of interest, or a LAScatalog if the
#' queries are immediately written into files without loading anything in R.
#'
#' @examples
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#'
#' # Load the file and clip the region of interest
#' las = readLAS(LASfile, select = "xyz", filter = "-keep_first")
#' subset1 = clip_rectangle(las, 684850, 5017850, 684900, 5017900)
#'
#' # Do not load the file(s), extract only the region of interest
#' # from a bigger dataset
#' ctg = readLAScatalog(LASfile, progress = FALSE, filter = "-keep_first")
#' subset2 = clip_rectangle(ctg, 684850, 5017850, 684900, 5017900)
#'
#' # Extract all the polygons from a shapefile
#' f <- system.file("extdata", "lake_polygons_UTM17.shp", package = "lidR")
#' lakes <- sf::st_read(f, quiet = TRUE)
#' subset3 <- clip_roi(las, lakes)
#'
#' # Extract the polygons for a catalog, write them in files named
#' # after the lake names, do not load anything in R
#' opt_output_files(ctg) <- paste0(tempfile(), "_{LAKENAME_1}")
#' new_ctg = clip_roi(ctg, lakes)
#' plot(new_ctg)
#'
#' # Extract a transect
#' p1 <- c(684800, y = 5017800)
#' p2 <- c(684900, y = 5017900)
#' tr1 <- clip_transect(las, p1, p2, width = 4)
#'
#' \dontrun{
#' plot(subset1)
#' plot(subset2)
#' plot(subset3)
#'
#' plot(tr1, axis = TRUE, clear_artifacts = FALSE)
#' }
#' @name clip
#' @md
#' @export
clip_roi = function(las, geometry, ...)
{
  if (is.character(geometry))
    geometry <- sf::st_as_sfc(geometry, crs = sf::st_crs(las))

  if (is(geometry, "Polygon"))
    geometry <- sf::st_sfc(sf::st_polygon(list(geometry@coords)), crs = st_crs(las))

  if (inherits(geometry, "Spatial"))
    geometry <- sf::st_as_sf(geometry)

  if (is(geometry, "sfg"))
    geometry <- sf::st_sfc(geometry, crs = st_crs(las))

  if (is(geometry, "sf") | is(geometry, "sfc"))
  {
    if (all(sf::st_is(geometry, "POLYGON") | sf::st_is(geometry, "MULTIPOLYGON")))
    {
      return(clip_sf(las, geometry))
    }
    else if (all(sf::st_is(geometry, "POINT")))
    {
      p <- list(...)
      if (is.null(p$radius))
        stop("Clipping using sfc_POINT or SpatialPoints* requires addition of parameter 'radius'.", call. = FALSE)

      centers <- sf::st_coordinates(geometry)
      ycenter <- centers[,2]
      xcenter <- centers[,1]
      radius  <- p$radius
      return(clip_circle(las, xcenter, ycenter, radius, data = geometry))
    }
    else
      stop("Incorrect geometry type. POINT, POLYGON and MULTIPOLYGON are supported.", call. = FALSE)
  }
  else if (is(geometry, "Extent"))
  {
    xmin <- geometry@xmin
    xmax <- geometry@xmax
    ymin <- geometry@ymin
    ymax <- geometry@ymax
    return(clip_rectangle(las, xmin, ymin, xmax, ymax))
  }
  else if (is.matrix(geometry))
  {
    if (!all(dim(geometry) == 2))
      stop("Matrix must have a size 2 x 2")

    xmin <- geometry[1]
    xmax <- geometry[3]
    ymin <- geometry[2]
    ymax <- geometry[4]
    return(clip_rectangle(las, xmin, ymin, xmax, ymax))
  }
  else if (is(geometry, "bbox"))
  {
    xmin <- geometry$xmin
    xmax <- geometry$xmax
    ymin <- geometry$ymin
    ymax <- geometry$ymax
    return(clip_rectangle(las, xmin, ymin, xmax, ymax))
  }
  else
  {
    stop(paste0("Geometry type ", paste0(class(geometry), collapse = " "), " not supported"))
  }
}

# ========= RECTANGLE =========

#' @export
#' @rdname clip
clip_rectangle = function(las, xleft, ybottom, xright, ytop, ...)
{
  assert_is_numeric(xleft)
  assert_is_numeric(ybottom)
  assert_is_numeric(xright)
  assert_is_numeric(ytop)
  assert_are_same_length(xleft, ybottom)
  assert_are_same_length(xleft, xright)
  assert_are_same_length(xleft, ytop)

  UseMethod("clip_rectangle", las)
}

#' @export
clip_rectangle.LAS = function(las, xleft, ybottom, xright, ytop, ...)
{
  X <- Y <- NULL

  output <- vector(mode = "list", length(xleft))
  for (i in 1:length(xleft))
  {
    roi <- filter_poi(las, X >= xleft[i] & X < xright[i] & Y >= ybottom[i] & Y < ytop[i])
    if (is.empty(roi)) warning(glue::glue("No point found for within rectangle ({xleft[i]}, {ybottom[i]}, {xright[i]}, {ytop[i]})."), call. = FALSE)
    output[[i]] = roi
  }

  if (length(output) == 1)
    return(output[[1]])
  else
    return(output)
}

#' @export
clip_rectangle.LAScatalog = function(las, xleft, ybottom, xright, ytop, ...)
{
  bboxes  <- st_make_bboxes(xleft, xright, ybottom, ytop)
  output  <- catalog_extract(las, bboxes, LIDRRECTANGLE)

  if (length(output) == 1)
    return(output[[1]])
  else
    return(output)
}

# ========  POLYGON ========

#' @export clip_polygon
#' @rdname clip
clip_polygon = function(las, xpoly, ypoly, ...)
{
  assert_is_numeric(xpoly)
  assert_is_numeric(ypoly)
  assert_are_same_length(xpoly, ypoly)

  poly <- sf::st_polygon(list(cbind(xpoly, ypoly)))
  return(clip_roi(las, poly))
}

# ======== CIRCLE ========

#' @export clip_circle
#' @rdname clip
clip_circle = function(las, xcenter, ycenter, radius, ...)
{
  assert_is_numeric(xcenter)
  assert_is_numeric(ycenter)
  assert_is_numeric(radius)
  assert_are_same_length(xcenter, ycenter)
  UseMethod("clip_circle", las)
}

#' @export
clip_circle.LAS = function(las, xcenter, ycenter, radius, ...)
{
  if (length(radius) > 1)
    assert_are_same_length(xcenter, radius)
  else
    radius <- rep(radius, length(xcenter))

  X <- Y <- NULL

  output <- vector(mode = "list", length(xcenter))
  for (i in 1:length(xcenter))
  {
    roi <- filter_poi(las, (X - xcenter[i])^2 + (Y - ycenter[i])^2 <= radius[i]^2)
    if (is.empty(roi)) warning(glue::glue("No point found for within disc ({xcenter[i]}, {ycenter[i]}, {radius[i]})."), call. = FALSE)
    output[[i]] <- roi
  }

  if (length(output) == 1)
    return(output[[1]])
  else
    return(output)
}

#' @export
clip_circle.LAScatalog = function(las, xcenter, ycenter, radius, ...)
{
  if (length(radius) > 1)
    assert_are_same_length(xcenter, radius)
  else
    radius <- rep(radius, length(xcenter))

  xmin   <- xcenter - radius
  xmax   <- xcenter + radius
  ymin   <- ycenter - radius
  ymax   <- ycenter + radius

  bboxes <- st_make_bboxes(xmin, xmax, ymin, ymax)
  output <- catalog_extract(las, bboxes, LIDRCIRCLE, ...)

  if (length(output) == 1)
    return(output[[1]])
  else
    return(output)
}

# ======= TRANSECT ========

#' @export
#' @rdname clip
#' @param p1,p2 numeric vectors of length 2 that gives the coordinates of two points that define a
#' transect
#' @param width numeric. width of the transect.
#' @param xz bool. If \code{TRUE} the point cloud is reoriented to fit on XZ coordinates
clip_transect = function(las, p1, p2, width, xz = FALSE, ...)
{
  assert_is_a_bool(xz)

  # Not documented but if p1 and p2 are missing it switches
  # to interactive selection (personal use only)
  if (!missing(p1) && !missing(p2))
  {
    assert_is_numeric(p1)
    assert_is_numeric(p2)
    assert_are_same_length(p1, p2)
    assert_is_of_length(p1, 2L)
  }
  else
  {
    # nocov start
    p <- list(...)
    plot <- isTRUE(p$plot)
    if (plot)
    {
      dsm <- rasterize_canopy(las, 2, p2r())
      plot(las@header)
      plot(dsm, col = height.colors(15))
    }
    tr <- graphics::locator(2L, type = "o")
    p1 <- c(tr$x[1L], tr$y[1L])
    p2 <- c(tr$x[2L], tr$y[2L])
    # nocov end
  }

  if (is(las, "LAScatalog"))
  {
    if (xz && opt_output_files(las) != "")
      stop("Reorientation is not available yet with a LAScatalog", call. = FALSE)
  }

  dx <- p1[1] - p2[1]
  dy <- p1[2] - p2[2]
  a  <- atan(dy/dx)
  rot <- matrix(c(cos(a), sin(a), -sin(a), cos(a)), ncol = 2)
  coords <- rbind(p1, p2)
  line <- sf::st_linestring(coords)
  line <- sf::st_sfc(line)
  sf::st_crs(line) <- sf::st_crs(las)
  poly <- sf::st_buffer(line, dist = width/2, endCapStyle = "FLAT")
  las <- clip_roi(las, poly)

  if (!xz) { return(las) }

  bbox <- st_bbox(las)
  coords <- as.matrix(coordinates(las))
  coords[,1] <- coords[,1] - bbox$xmin
  coords[,2] <- coords[,2] - bbox$ymin
  coords <- coords %*% rot
  X <- coords[,1]
  Y <- coords[,2]
  fast_quantization(X, las@header@PHB[["X scale factor"]], 0)
  fast_quantization(Y, las@header@PHB[["Y scale factor"]], 0)
  las@data[["X"]] <- X
  las@data[["Y"]] <- Y
  las@header@PHB[["X offset"]] <- 0
  las@header@PHB[["Y offset"]] <- 0
  las <- las_update(las)
  data.table::setattr(las, "rotation", rot)
  data.table::setattr(las, "offset", c(bbox$xmin, bbox$ymin))
  return(las)
}

# ======== WKT ========

clip_sf = function(las, sf)
{
  UseMethod("clip_sf", las)
}

clip_sf.LAS = function(las, sf)
{
  sfc <- sf::st_geometry(sf)
  idx <- point_in_polygons(las, sfc, TRUE)
  output <- vector(mode = "list", length(sfc))
  uid <- unique(idx)
  uid <- uid[!is.na(uid)]
  for (i in seq_along(idx))
  {
    # memory optimization
    if(length(idx[[i]])== npoints(las))
    {
      output[[i]] = las
      next
    }

    roi = suppressWarnings(las[idx[[i]]])
    if (is.empty(roi)) warning(glue::glue("No point found for within {sf::st_as_text(sfc[i])}."), call. = FALSE)
    output[[i]] = roi
  }

  if (length(output) == 0)
    return(NULL)
  else if (length(output) == 1)
    return(output[[1]])
  else
    return(output)
}

clip_sf.LAScatalog = function(las, sf)
{
  wkt  <- sf::st_as_text(sf::st_geometry(sf), digits = 10)

  # We need the bounding box of each geometry to be able to leverage automatically spatial
  # indexing of LAS files with LAX files
  bboxes <- lapply(sf::st_geometry(sf), sf::st_bbox)
  output <- catalog_extract(las, bboxes, LIDRRECTANGLE, sf = sf)

  if (length(output) == 1)
    return(output[[1]])
  else
    return(output)
}

# ============= GENERIC QUERY  =============

#' @param ctg LAScatalog
#' @param bboxes a list of bbox
#' @param shape shape of the query can be a rectangle or a disc
#' @param sf an object of class sf that is used for extracting polygons carrying WKT strings
#' @param data a data.frame carrying some attributes used to create fill the {TEMPLATE}
#' @noRd
catalog_extract = function(ctg, bboxes, shape = LIDRRECTANGLE, sf = NULL, data = NULL)
{
  stopifnot(shape == LIDRRECTANGLE | shape == LIDRCIRCLE)

  if (opt_progress(ctg)) plot.LAScatalog(ctg, mapview = FALSE) # nocov

  # Define a function to be passed in engine_apply
  extract_query = function(cluster)
  {
    if (cluster@files[1] == "")
      return(NULL)

    x <- suppressMessages(suppressWarnings(streamLAS(cluster, ofile = cluster@save, filter_wkt = cluster@wkt)))

    if (is.null(x))
    {
      x <- 0
      class(x) <- "lidr_internal_skip_write"
      return(x)
    }

    index(x) <- index(cluster)
    sensor(x) <- sensor(cluster)
    return(x)
  }

  # Find the ROIs in the catalog and return LASclusters. If a ROI fall outside the catalog
  # its associated LAScluster is NULL and must receive a special treatment in following code
  clusters <- engine_index(ctg, bboxes, shape, 0, outside_catalog_is_null = FALSE)

  # Add some information in the clusters to correctly extract polygons and to write correct file names
  for (i in 1:length(clusters))
  {
    # skip NULL clusters
    if (clusters[[i]]@files[1] == "")
      next

    # If a simple feature is provided we want to extract a polygon. Insert WKT string
    if (!is.null(sf))
      clusters[[i]]@wkt = sf::st_as_text(sf::st_geometry(sf)[i], digits = 10)

    # If the user wants to write the ROIs in files. Generate a filename.
    if (opt_output_files(ctg) != "")
    {
      if (!is.null(sf))
      {
        if (ncol(sf) > 1)
        {
          X <- sf[i,]
        }
        else
        {
          X <- sf[i,]
          X <- as.list(X)
          names(X) <- names(sf)
        }
      }
      else if (!is.null(data))
      {
        if (ncol(data) > 1)
        {
          X <- data[i,]
        }
        else
        {
          X <- data[i,]
          X <- as.list(X)
          names(X) <- names(data)
        }
      }
      else
        X <- list()

      X$ID      <- i
      X$XCENTER <- format(clusters[[i]]@center$x, scientific = F)
      X$YCENTER <- format(clusters[[i]]@center$y, scientific = F)
      X$XLEFT   <- format(clusters[[i]]@bbox[1], scientific = F)
      X$XRIGHT  <- format(clusters[[i]]@bbox[3], scientific = F)
      X$YBOTTOM <- format(clusters[[i]]@bbox[2], scientific = F)
      X$YTOP    <- format(clusters[[i]]@bbox[4], scientific = F)
      format    <- if (opt_laz_compression(ctg)) ".laz" else ".las"

      usefilename <- grepl("\\{ORIGINALFILENAME\\}",  opt_output_files(ctg))
      if (usefilename)
        stop("The template {ORIGINALFILENAME} makes sense only when processing by file. It is undefined in clip functions.", call. = FALSE)

      filepath  <- paste0(glue::glue_data(X, opt_output_files(ctg)), format)
      n         <- length(filepath)

      if (n > 1)
        stop(glue::glue("Ill-formed template string in the catalog: {n} filenames were generated for each region of interest"), call. = FALSE)

      clusters[[i]]@save <- filepath
    }
  }

  # Process the cluster using LAScatalog internal engine
  output <- engine_apply(clusters, extract_query, ctg@processing_options, ctg@output_options)

  # output should contain nothing because everything has been streamed into files
  if (opt_output_files(ctg) != "")
  {
    written_path = character(0)
    for (i in seq_along(clusters))
    {
      if (clusters[[i]]@files[1] == "")
      {
        message(glue::glue("No point found for within region of interest {i}."))
        next
      }

      if (file.exists(clusters[[i]]@save))
        written_path = append(written_path, clusters[[i]]@save)
      else
        message(glue::glue("No point found for within region of interest {i}."))
    }

    if (length(written_path) > 0)
    {
      new_ctg <- suppressMessages(readLAScatalog(written_path))
      opt_copy(new_ctg) <- ctg
    }
    else
    {
      # Empty LAScatalog
      new_ctg <- new("LAScatalog")
      st_crs(new_ctg) <- st_crs(ctg)
    }

    return(list(new_ctg))
  }
  # output should contain LAS objects returned at the R level
  else
  {
    for (i in 1:length(output))
    {
      if (!is.null(output[[i]]))
      {
        # Transfer the CRS of the catalog.
        st_crs(output[[i]]) <- st_crs(ctg)
      }
      else
      {
        # For consistency with LAS dispatched functions, LAScatalog must return empty LAS objects that respect
        # the select option. The following is definitively a twist to get a consistent output but happened
        # only for dummy queries outide the catalog
        emptylas <- readLAS(ctg@data$filename[1], ctg@input_options$select, filter = "-inside 0 0 0 0")
        output[[i]] <- emptylas
        message(glue::glue("No point found for within region of interest {i}."))
      }
    }

    return(output)
  }
}

context("clip_roi")

las <- readLAS(example_las_path)
ctg <- example_ctg

opt_progress(ctg) = FALSE

test_that("clip_rectangle clips a rectangle both on a LAS and a LAScatalog", {

  rect1 <- clip_rectangle(las, 339006, 5248000, 339012, 5248002)
  rect2 <- clip_rectangle(ctg, 339006, 5248000, 339012, 5248002)

  #expect_true(extent(rect1) <= raster::extent(339006, 5248000, 339012, 5248002))
  expect_equal(rect1@crs, las@crs)
  expect_equal(npoints(rect1), 17L)
  expect_equal(rect1, rect2)
})

test_that("clip_circle clips a circle both on a LAS and a LAScatalog", {

  circ1 <- clip_circle(las, 339008, 5248001, 2)
  circ2 <- clip_circle(ctg, 339008, 5248001, 2)

  expect_equal(st_crs(circ1), st_crs(las))
  expect_equal(npoints(circ1), 15L)
  expect_equal(st_crs(circ2), st_crs(ctg))
  expect_equal(circ1, circ2)
})

test_that("clip_polygon clips a polygon both on a LAS and a LAScatalog", {

  tri1 <- clip_polygon(las, c(339008, 339010, 339010, 339008), c(5248000, 5248000, 5248002, 5248000))
  tri2 <- clip_polygon(ctg, c(339008, 339010, 339010, 339008), c(5248000, 5248000, 5248002, 5248000))

  expect_equal(st_crs(tri1), st_crs(las))
  expect_equal(npoints(tri1), 15L)
  expect_equal(st_crs(tri2), st_crs(ctg))
  expect_equal(tri1, tri2)
})

test_that("clip_roi memory optimization works", {

  las2 <- clip_polygon(las, c(0, 8e6, 8e6, 0), c(0, 5e8, 0, 0))
  las3 <- clip_circle(las, 339008, 5248001, 30)

  expect_reference(las@data, las2@data)
  expect_reference(las@data, las3@data)
})

test_that("clip_roi clips polygon works from WTK both on a LAS and LAScatalog", {

  wkt <- "POLYGON ((339008 5248000, 339010 5248000, 339010 5248002, 339008 5248000))"

  poly1   <- clip_roi(las, wkt)
  poly2   <- clip_roi(ctg, wkt)

  expect_is(poly1, "LAS")
  expect_equal(npoints(poly1), 15L)
  expect_equal(poly1, poly2)
  expect_equal(poly1@crs, las@crs)
})

test_that("clip_roi clips polygon works sfc", {

  wkt <- "POLYGON ((339008 5248000, 339010 5248000, 339010 5248002, 339008 5248000))"
  p <- sf::st_as_sfc(wkt)
  sf::st_crs(p) <- st_crs(las)

  poly1   <- las[p]
  poly2   <- clip_roi(las, p)

  expect_is(poly2, "LAS")
  expect_equal(npoints(poly2), 15L)
  expect_equal(poly1, poly2)
  expect_equal(poly1@crs, las@crs)
})

test_that("clip_roi clips multipolygon with hole", {

  wkt <- "POLYGON ((339008 5248000, 339010 5248000, 339010 5248002, 339008 5248000))"
  p1 <- sf::st_point(c(684850, 5017850))
  p2 <- sf::st_point(c(684900, 5017900))
  poly1e <- sf::st_buffer(p1, 20)
  poly1i <- sf::st_buffer(p1, 10)
  poly1 <- sf::st_difference(poly1e, poly1i)
  poly2 <- sf::st_buffer(p2, 20)
  poly <- c(poly1, poly2)

  donut <- clip_roi(megaplot, poly)

  expect_equal(npoints(donut), 3905)
})

test_that("clip_roi clips polygon works from sp polygons both on a LAS and LAScatalog", {

  skip_if_not_installed("sp")

  wkt1 <- "MULTIPOLYGON (((339010.5 5248000, 339012 5248000, 339010.5 5248002, 339010.5 5248000)), ((339008 5248000, 339010 5248000, 339010 5248002, 339008 5248000), (339008.5 5248000.2, 339009.5 5248000.2, 339009.5 5248001, 339008.5 5248000.2)))"
  wkt2 <- "POLYGON ((339008 5248000, 339010 5248000, 339010 5248002, 339008 5248000))"

  # SpatialPolygons
  spatialpolygons1 <- sf::as_Spatial(sf::st_as_sfc(wkt1))
  spatialpolygons2 <- sf::as_Spatial(sf::st_as_sfc(wkt2))

  polygon1 <- spatialpolygons1@polygons[[1]]@Polygons[[1]]
  poly1    <- clip_roi(las, polygon1)
  poly2    <- clip_roi(ctg, polygon1)

  expect_is(poly1, "LAS")
  expect_equal(npoints(poly1), 2L)
  expect_equal(st_crs(poly1), st_crs(las))
  expect_equal(st_crs(poly2), st_crs(ctg))
  expect_equal(poly1, poly2)
})

test_that("clip_roi clips point with SpatialPoints and sfc on LAS and LAScatalog", {

  xc <- c(339008, 339009)
  yc <- c(5248001, 5248001)
  xy <- data.frame(X = xc, Y = yc)
  r  <- 2

  p <- sf::st_as_sf(xy, coords = c("X", "Y"))
  p <- sf::st_geometry(p)

  #discs1 <- clip_roi(las, sf::as_Spatial(p), radius = r)
  discs2 <- clip_roi(las, p, radius = r)
  discs3 <- clip_roi(ctg, p, radius = r)
  discs4 <- clip_roi(ctg, p, radius = c(r,2))

  expect_is(discs2, "list")
  expect_equal(length(discs2), 2L)
  #expect_equal(discs1, discs2)
  expect_equal(discs2, discs3)
  expect_equal(discs2, discs4)
})

test_that("clip_transect clips a transect on LAS and LAScatalog", {

  p1  <- st_bbox(las)[1:2]
  p2  <- st_bbox(las)[3:4]
  tr1 <- clip_transect(las, p1, p2, 2, xz = FALSE)
  tr2 <- clip_transect(ctg, p1, p2, 2, xz = FALSE)

  expect_equal(npoints(tr1), 29L)
  #expect_equal(tr1, tr2)
})

test_that("clip_transect clips reorients the point-cloud on LAS and LAScatalog", {

  p1  <- st_bbox(las)[1:2]
  p2  <- st_bbox(las)[3:4]
  tr1 <- clip_transect(las, p1, p2, 2, xz = TRUE)
  tr2 <- clip_transect(ctg, p1, p2, 2, xz = TRUE)

  expect_equal(npoints(tr1), 29L)
  #expect_equal(tr1, tr2)
  expect_equal(mean(tr1$Y), 0, tol = 0.5)

  opt_output_files(ctg) <- tempfile()
  expect_error(clip_transect(ctg, p1, p2, 2, xz = TRUE), " not available yet")
})

test_that("clip_roi throw error with points and no radius", {

  xc <- c(684800, 684850)
  yc <- c(5017850, 5017900)
  xy <- data.frame(X = xc, Y = yc)
  r  <- 2
  p <- sf::st_as_sf(xy, coords = c("X", "Y"))
  p <- sf::st_geometry(p)

  expect_error(clip_roi(las, p), "requires addition of parameter 'radius'")
})

test_that("clip_roi throw error with lines", {

  l1 = cbind(c(1,2,3),c(3,2,2))
  l2 = cbind(c(1,2,3),c(1,1.5,1))
  Sl = sf::st_geometry(sf::st_linestring(l1))

  expect_error(clip_roi(las, Sl), "Incorrect geometry type")
})

test_that("clip_roi clips a rectangle from a bounding box both on a LAS and LAScatalog", {

  # Extent
  bbox <- sf::st_bbox(c(xmin = 339005.9, xmax = 339012.1, ymin = 5248000, ymax = 5248001))

  rect1 <- clip_roi(las, bbox)
  rect2 <- clip_roi(ctg, bbox)

  #expect_true(extent(rect1) <= bbox)
  expect_equal(npoints(rect1), 14L)
  expect_equal(rect1, rect2)
})


test_that("clip_* returns an empty point cloud if no point found in the query", {

  circ1 <- suppressWarnings(clip_circle(las, 68480, 5017850, 10))
  circ2 <- suppressWarnings(clip_circle(ctg, 68480, 5017850, 10))

  # Unit test for #400
  opt_output_files(ctg) <- tempfile()
  circ3 <- suppressWarnings(clip_circle(ctg, 68480, 5017850, 10))

  expect_true(is.empty(circ1))
  expect_true(is.empty(circ2))
  expect_equal(dim(circ3), c(0,35))
})

test_that("clip_* supports multiple queries", {

  # Multiple disc
  xc <- c(339008, 339009)
  yc <- c(5248001, 5248001)
  r  <- 2

  circ1 <- clip_circle(las, xc, yc, r)
  circ2 <- clip_circle(ctg, xc, yc, r)

  expect_is(circ1, "list")
  expect_equal(length(circ1), 2L)
  expect_equal(circ1, circ2)

  # Multiple rectangle
  xmin <- 339008 + c(0,0)
  ymin <- 5248000 + c(0,0)
  xmax <- 339008 + c(2,3)
  ymax <- 5248001 + c(2,3)

  rect1 <- clip_rectangle(las, xmin, ymin, xmax, ymax)
  rect2 <- clip_rectangle(ctg, xmin, ymin, xmax, ymax)

  expect_is(rect1, "list")
  expect_equal(length(rect1), 2L)
  expect_equal(rect1, rect2)
})

test_that("clip_* throw error for invalid multiple queries", {

  # Multiple disc
  xc <- c(684800, 684850)
  yc <- c(5017850, 5017900)
  r  <- 10:13

  expect_error(clip_circle(las, xc, yc, r), "xc and r have different lengths")
})

test_that("clip_* return an empty point-cloud for empty multiple queries", {

  # Multiple disc
  xc <- c(339008, 68480)
  yc <- c(5248001, 5017900)
  r  <- 10

  circ1 <- suppressWarnings(clip_circle(las, xc, yc, r))
  circ2 <- suppressWarnings(clip_circle(ctg, xc, yc, r))

  expect_is(circ1, "list")
  expect_equal(length(circ1), 2L)
  expect_equal(npoints(circ1[[1]]), 30)
  expect_true(is.empty(circ1[[2]]))
  expect_equal(circ1, circ2)

  # Multiple rectangle
  xmin <- 339008 + c(0,-2000)
  ymin <- 5248000 + c(0,-2000)
  xmax <- 3390010 + c(0,-2000)
  ymax <- 5248002 + c(0,-2000)

  rect1 <- suppressWarnings(clip_rectangle(las, xmin, ymin, xmax, ymax))
  rect2 <- suppressWarnings(clip_rectangle(ctg, xmin, ymin, xmax, ymax))

  expect_is(rect1, "list")
  expect_equal(length(rect1), 2L)
  expect_equal(npoints(rect1[[1]]), 20)
  expect_true(is.empty(rect1[[2]]))
  expect_equal(rect1, rect2)
})

test_that("clip_* throw errors with invalid queries", {

  # Invalid WKT
  wkt <- "POLGON ((684975.7 5«017899, 685007.3 5017873, 684994.3 5017816, 684936.1 5017812, 684918.8 5017845, 684975.7 5017899))"

  expect_error(clip_roi(las, wkt), "OGR error")
  expect_error(clip_roi(ctg, wkt), "OGR error")

  # Different number of coordinates
  xc <- c(684800)
  yc <- c(5017850, 5017900)
  r  <- 10

  expect_error(clip_circle(las, xc, yc, r), "different lengths")
  expect_error(clip_circle(ctg, xc, yc, r), "different lengths")

  xmin <- 684850
  ymin <- 5017850 + c(0,1)
  xmax <- 684900 + c(0,1)
  ymax <- 5017900 + c(0,1)

  expect_error(clip_rectangle(las, xmin, ymin, xmax, ymax), "different lengths")
  expect_error(clip_rectangle(ctg, xmin, ymin, xmax, ymax), "different lengths")

  # Wrong matrix dimension
  m <- matrix(0, 3, 2)

  expect_error(clip_roi(las, m), "2 x 2")
  expect_error(clip_roi(ctg, m), "2 x 2")

  # Non supported object
  geom <- 1
  class(geom) <- c("A",  "B")
  expect_error(clip_roi(las, geom), "Geometry type A B not supported")
})

test_that("clip_* write file following LAScatalog options", {

  tmp  <- tempdir()

  ctg2 <- ctg
  opt_output_files(ctg2)    <- paste0(tmp, "/file_{XLEFT}")
  opt_laz_compression(ctg2) <- TRUE

  ctg3 <- clip_rectangle(ctg2, 339006, 5248000, 339012, 5248002)

  expect_true(is(ctg3, "LAScatalog"))
  expect_equal(normalizePath(ctg3@data$filename), normalizePath(paste0(tmp, "/file_339006.laz")))

  file.remove(paste0(tmp, "/file_339006.laz"))

  #opt_output_files(ctg2)    <- paste0(tmp, "/file_{LAKENAME_1}")
  opt_laz_compression(ctg2) <- FALSE
  #ctg3 <- clip_roi(ctg2, lakes)

  #expect_equal(normalizePath(ctg3@data$filename), normalizePath(paste0(tmp, "/file_Havelock Lake.las")))

  #file.remove(paste0(tmp, "/file_Havelock Lake.las"))

  xc <- c(339008, 339009)
  yc <- c(5248001, 5248001)
  D = data.frame(X = xc, Y = yc, PlotID = paste0("plot", 1:2))
  P = sf::st_as_sf(D, coords = c("X", "Y"))

  opt_output_files(ctg2) <- paste0(tmp, "/{PlotID}")
  ctg3 = clip_roi(ctg2, P, radius = 2)
  expect_equal(normalizePath(ctg3@data$filename), normalizePath(paste0(tmp, "/plot", 1:2, ".las")))
})

test_that("clip_* throw an error with invalid template", {

  tmp  <- tempdir()

  ctg2 <- ctg
  opt_output_files(ctg2)    <- paste0(tmp, "/file_{1:3}")
  opt_laz_compression(ctg2) <- TRUE

  expect_error(clip_rectangle(ctg2,  339006, 5248000, 339012, 5248002), "Ill-formed template string in the catalog")

  opt_output_files(ctg2)    <- paste0(tmp, "/*")

  expect_error(clip_rectangle(ctg2,  339006, 5248000, 339012, 5248002), "undefined in clip functions")
})

test_that("clip_* respect spatial index metadata in LAScatalog", {
  xc <- 339008
  yc <- 5248001
  r  <- 2
  sensor(ctg) <- "tls"
  index(ctg) <- "octree"
  las = clip_circle(ctg, xc, yc, r)
  expect_equal(index(las), index(ctg))
  expect_equal(sensor(las), sensor(ctg))
})

#' @export
#' @rdname Extract
setMethod("$", "LAScatalog", function(x, name) { x@data[[name]] })

#' @export
#' @rdname Extract
setMethod("[[", c("LAScatalog", "ANY", "missing"), function(x, i, j, ...) { x@data[[i]] })


#' @param ... Unused
#' @param drop Unused
#' @export
#' @rdname Extract
setMethod("[", "LAScatalog", function(x, i, j, ..., drop = FALSE) {

  ctgname <- deparse(substitute(x))
  iname   <- deparse(substitute(i))
  nargs   <- nargs()

  if (!missing(i) & !missing(j))
    stop(glue::glue("This action is not allowed for a {class(x)}. j must be missing. Maybe you meant: {ctgname}[{iname}, ]."), call. = FALSE)

  if (missing(i) & !missing(j))
    stop(glue::glue("This action is not allowed for a {class(x)}. i cannot be missing."), call. = FALSE)

  if (!missing(i) & missing(j) & nargs == 2L)
    return(x@data[i])

  x@data <- x@data[i, ]
  return(x)
})

#' @export
#' @rdname Extract
setMethod("[", c("LAScatalog", "logical"),  function(x, i)
{
  x@data <- x@data[i,]
  return(x)
})

#' @export
#' @rdname Extract
setMethod("[", c("LAScatalog", "sf"),  function(x, i)
{
  return(x[sf::st_geometry(i)])
})

#' @export
#' @rdname Extract
setMethod("[", c("LAScatalog", "sfc"),  function(x, i)
{
  return(catalog_intersect(x, i))
})

#' @export
#' @rdname Extract
setMethod("[[<-", "LAScatalog",  function(x, i, j, value)
{
  if (i %in% LASCATALOGATTRIBUTES)
    stop("LAScatalog data read from standard files cannot be modified", call. = FALSE)

  x@data[[i]] = value
  return(x)
})

#' @export
#' @rdname Extract
setMethod("$<-", "LAScatalog", function(x, name, value)
{
  if (name %in% LASCATALOGATTRIBUTES)
    stop("LAScatalog data read from standard files cannot be modified", call. = FALSE)

  x@data[[name]] = value
  return(x)
})

#' Bounding box of a LAS* object
#'
#' Bounding box of a `LAS*` object. `st_bbox()` extends `sf`, and `ext()` extends `terra`. The values returned are similar to their
#' parent functions.
#'
#' @return A `bbox` from sf, or a `SpatExtent` from `terra`.
#'
#' @param obj,x An object of class \code{LAS*}.
#' @param ... unused
#'
#' @export
#' @name st_bbox
#' @importFrom sf st_bbox
#' @importFrom terra ext
#' @md
NULL

#' @export
#' @rdname st_bbox
#' @examples
#' f <- system.file("extdata", "example.las", package="rlas")
#' las <- readLAS(f)
#'
#' st_bbox(las)
#' ext(las)
st_bbox.LAS = function(obj, ...)
{
  xr <- range(obj$X)
  yr <- range(obj$Y)
  bbox <- st_bbox(c(xmin = xr[1], xmax = xr[2], ymin = yr[1], ymax = yr[2]))
  sf::st_crs(bbox) <- st_crs(obj)
  return(bbox)
}

#' @export
#' @rdname st_bbox
st_bbox.LASheader = function(obj, ...)
{
  if (obj[["Number of point records"]] == 0L)
    return(sf::NA_bbox_)

  bbox <- c(obj[["Min X"]], obj[["Min Y"]], obj[["Max X"]], obj[["Max Y"]])
  names(bbox) <- c("xmin", "ymin", "xmax", "ymax")
  crs <- tryCatch(crs <- st_crs(obj), error = function(e) sf::NA_crs_)
  return(sf::st_bbox(bbox, crs = crs))
}

#' @export
#' @rdname st_bbox
st_bbox.LAScatalog = function(obj, ...)
{
  # Workaround to repair LAScatalog v3 and minimize backward incompatibilities with v4
  obj <- lascatalog_v3_repair(obj)
  return(sf::st_bbox(obj@data))
}

#' @export
#' @rdname st_bbox
st_bbox.LAScluster = function(obj, ...)
{
  bb <- as.numeric(obj@bbox)
  names(bb) <- c("xmin", "ymin", "xmax", "ymax")
  bb <- sf::st_bbox(bb)
  sf::st_crs(bb) <- st_crs(obj)
  return(bb)
}

st_bbox.raster_template = function(obj, ...)
{
  bb <- c(obj$xmin, obj$ymin, obj$xmax, obj$ymax)
  names(bb) <- c("xmin", "ymin", "xmax", "ymax")
  bb <- sf::st_bbox(bb)
  sf::st_crs(bb) <- obj$crs
  return(bb)
}

# ==== SPATEXTENT =====

#' @export
#' @rdname st_bbox
setMethod("ext", "LAS", function(x, ...) { .ext(x) })

#' @export
#' @rdname st_bbox
setMethod("ext", "LASheader", function(x, ...) { .ext(x) })

#' @export
#' @rdname st_bbox
setMethod("ext", "LAScatalog", function(x, ...) { .ext(x) })

#' @export
#' @rdname st_bbox
setMethod("ext", "LAScluster", function(x, ...) { .ext(x) })

.ext <- function(x)
{
  bb <- st_bbox(x)
  return(terra::ext(bb$xmin, bb$xmax, bb$ymin, bb$ymax))
}

#' Individual Tree Segmentation Algorithm
#'
#' This function is made to be used in \link{segment_trees}. It implements an algorithm for tree
#' segmentation based on Dalponte and Coomes (2016) algorithm (see reference).
#' This is a seeds + growing region algorithm. This algorithm exists in the package \code{itcSegment}.
#' This version has been written from the paper in C++. Consequently it is hundreds to millions times
#' faster than the original version. Note that this algorithm strictly performs a segmentation, while the
#' original method as implemented in \code{itcSegment} and described in the manuscript also performs
#' pre- and post-processing tasks. Here these tasks are expected to be done by the user in separate functions.
#'
#' Because this algorithm works on a CHM only there is no actual need for a point cloud. Sometimes the
#' user does not even have the point cloud that generated the CHM. \code{lidR} is a point cloud-oriented
#' library, which is why this algorithm must be used in \link{segment_trees} to merge the result with the point
#' cloud. However the user can use this as a stand-alone function like this:
#' \preformatted{
#'  chm <- raster("chm.tif")
#'  ttops <- locate_trees(chm, lmf(3))
#'  crowns <- dalponte2016(chm, ttops)()
#' }
#'
#' @template param-chm-lastrees
#' @template param-treetops
#' @param th_tree numeric. Threshold below which a pixel cannot be a tree. Default is 2.
#' @param th_seed numeric. Growing threshold 1. See reference in Dalponte et al. 2016. A pixel
#' is added to a region if its height is greater than the tree height multiplied by this value.
#' It should be between 0 and 1. Default is 0.45.
#' @param th_cr numeric. Growing threshold 2. See reference in Dalponte et al. 2016. A pixel
#' is added to a region if its height is greater than the current mean height of the region
#' multiplied by this value. It should be between 0 and 1. Default is 0.55.
#' @param max_cr numeric. Maximum value of the crown diameter of a detected tree (in pixels).
#' Default is 10.
#' @param ID character. If `treetops` contains an attribute with the ID for
#' each tree, the name of this attribute. This way, original IDs will be preserved.
#'
#' @references
#' Dalponte, M. and Coomes, D. A. (2016), Tree-centric mapping of forest carbon density from
#' airborne laser scanning and hyperspectral data. Methods Ecol Evol, 7: 1236–1245. doi:10.1111/2041-210X.12575.
#'
#' @export
#'
#' @family individual tree segmentation algorithms
#' @family raster based tree segmentation algorithms
#'
#' @examples
#' LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
#' poi <- "-drop_z_below 0 -inside 481280 3812940 481320 3812980"
#' las <- readLAS(LASfile, select = "xyz", filter = poi)
#' col <- pastel.colors(200)
#'
#' chm <- rasterize_canopy(las, 0.5, p2r(0.3))
#' ker <- matrix(1,3,3)
#' chm <- terra::focal(chm, w = ker, fun = mean, na.rm = TRUE)
#'
#' ttops <- locate_trees(chm, lmf(4, 2))
#' las   <- segment_trees(las, dalponte2016(chm, ttops))
#' #plot(las, color = "treeID", colorPalette = col)
#' @name its_dalponte2016
#' @md
dalponte2016 = function(chm, treetops, th_tree = 2, th_seed = 0.45, th_cr = 0.55, max_cr = 10, ID = "treeID")
{
  assert_is_a_number(th_tree)
  assert_is_a_number(th_seed)
  assert_is_a_number(th_cr)
  assert_is_a_number(max_cr)
  assert_all_are_in_closed_range(th_seed, 0, 1)
  assert_all_are_in_closed_range(th_cr, 0, 1)
  stopifnot(raster_is_supported(chm))

  # Workaround for #622 If segment_tree is ran in parallel it will fail with
  # SpatRaster because they are not serializable. SpatRaster are converted to RasterLayer
  # for multicore strategies
  chm <- convert_ondisk_spatraster_into_serializable_raster_if_necessary(chm)

  treetops <- check_tree_tops(treetops, ID)
  chm      <- lazyeval::uq(chm)
  treetops <- lazyeval::uq(treetops)
  th_tree  <- lazyeval::uq(th_tree)
  th_seed  <- lazyeval::uq(th_seed)
  th_cr    <- lazyeval::uq(th_cr)
  max_cr   <- lazyeval::uq(max_cr)
  ID       <- lazyeval::uq(ID)

  f = function(bbox)
  {
    assert_is_valid_context(LIDRCONTEXTITS, "dalponte2016", null_allowed = TRUE)

    # If no seed it means we received an sf with 0 data
    if (nrow(treetops) == 0L)
    {
      crown <- chm
      crown <- raster_set_values(crown, NA_integer_)
      return(crown)
    }

    if (raster_is_proxy(chm) & missing(bbox))
      stop("Cannot segment the trees from a raster stored on disk. Use segment_trees() or load the raster in memory", call. = FALSE)

    # If a bbox is given we crop the CHM and the seed to this extent to reduce processing time
    # Otherwise we could get a chm much bigger than the LAS (e.g. LAScatalog processing) and
    # process would never ends
    res <- crop_special_its(treetops, chm, bbox)
    treetops <- res$treetops
    chm <- res$chm

    st_chm <- chm
    if (!is(chm, "stars"))
      st_chm <- stars::st_as_stars(chm)

    # If no seed, exit with with warning
    if (nrow(treetops) == 0L)
    {
      warning("No tree can be used as seed", call. = FALSE)
      crown <- chm
      crown <- raster_set_values(crown, NA_integer_)
      return(crown)
    }

    template <- stars::st_as_stars(st_chm)
    template[[1]][] <- 0L
    storage.mode(template[[1]]) <- "integer"

    treetops[["SEQUENTIALIDTREE"]] <- 1:nrow(treetops)
    rtreetops <- stars::st_rasterize(treetops["SEQUENTIALIDTREE"], template)
    storage.mode(rtreetops[[1]]) <- "integer"

    Maxima <- raster_as_matrix(rtreetops)$z
    Canopy <- raster_as_matrix(chm)$z

    Canopy[is.na(Canopy)] <- -Inf
    Crowns <- C_dalponte2016(Canopy, Maxima, th_seed, th_cr, th_tree, max_cr)

    Maxima[Maxima == 0L] <- NA_integer_
    Crowns[Crowns == 0L] <- NA_integer_
    Crowns[] <- treetops[[ID]][Crowns]
    val <- as.numeric(t(apply(Crowns, 1, rev)))
    storage.mode(val) <- storage.mode(treetops[[ID]])

    output <- chm
    output <- raster_set_values(output, NA_integer_)
    output <- raster_set_values(output, val)
    return(output)
  }

  class(f) <- c(LIDRALGORITHMITS, LIDRALGORITHMRASTERBASED)
  return(f)
}

#' Individual Tree Segmentation Algorithm
#'
#' This functions is made to be used in \link{segment_trees}. It implements an algorithm for tree
#' segmentation based on Silva et al. (2016) (see reference). This is a simple method
#' based on seed + voronoi tesselation (equivalent to nearest neighbour). This algorithm is implemented
#' in the package `rLiDAR`. This version is not the version from `rLiDAR`. It is
#' code written from the original article by the lidR authors and is considerably (between 250
#' and 1000 times) faster.
#'
#' Because this algorithm works on a CHM only there is no actual need for a point cloud. Sometimes the
#' user does not even have the point cloud that generated the CHM. `lidR` is a point cloud-oriented
#' library, which is why this algorithm must be used in \link{segment_trees} to merge the result into the point
#' cloud. However, the user can use this as a stand-alone function like this:
#' \preformatted{
#'  chm <- raster("chm.tif")
#'  ttops <- locate_trees(chm, lmf(3))
#'  crowns <- silva2016(chm, ttops)()
#' }
#'
#' @template param-chm-lastrees
#' @template param-treetops
#'
#' @param max_cr_factor numeric. Maximum value of a crown diameter given as a proportion of the
#' tree height. Default is 0.6, meaning 60% of the tree height.
#' @param exclusion numeric. For each tree, pixels with an elevation lower than `exclusion`
#' multiplied by the tree height will be removed. Thus, this number belongs between 0 and 1.
#' @param ID character. If `treetops` contains an attribute with the ID for
#' each tree, the name of this attribute. This way, original IDs will be preserved.
#'
#' @references
#' Silva, C. A., Hudak, A. T., Vierling, L. A., Loudermilk, E. L., O’Brien, J. J., Hiers,
#' J. K., Khosravipour, A. (2016). Imputation of Individual Longleaf Pine (Pinus palustris Mill.)
#' Tree Attributes from Field and LiDAR Data. Canadian Journal of Remote Sensing, 42(5), 554–573.
#' https://doi.org/10.1080/07038992.2016.1196582.
#'
#' @export
#'
#' @family individual tree segmentation algorithms
#' @family raster based tree segmentation algorithms
#'
#' @examples
#' LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
#' poi <- "-drop_z_below 0 -inside 481280 3812940 481320 3812980"
#' las <- readLAS(LASfile, select = "xyz", filter = poi)
#' col <- pastel.colors(200)
#'
#' chm <- rasterize_canopy(las, res = 0.5, p2r(0.3))
#' ker <- matrix(1,3,3)
#' chm <- terra::focal(chm, w = ker, fun = mean, na.rm = TRUE)
#'
#' ttops <- locate_trees(chm, lmf(4, 2))
#' las   <- segment_trees(las, silva2016(chm, ttops))
#' #plot(las, color = "treeID", colorPalette = col)
#' @name its_silva2016
#' @md
silva2016 = function(chm, treetops, max_cr_factor = 0.6, exclusion = 0.3, ID = "treeID")
{
  assert_is_a_number(max_cr_factor)
  assert_is_a_number(exclusion)
  assert_all_are_positive(max_cr_factor)
  assert_all_are_in_open_range(exclusion, 0, 1)

  # Workaround for #622 If segment_tree is ran in parallel it will fail with
  # SpatRaster because they are not serializable. SpatRaster are converted to RasterLayer
  # for multicore strategies
  chm <- convert_ondisk_spatraster_into_serializable_raster_if_necessary(chm)

  treetops       <- check_tree_tops(treetops, ID)
  chm            <- lazyeval::uq(chm)
  treetops       <- lazyeval::uq(treetops)
  max_cr_factor  <- lazyeval::uq(max_cr_factor)
  exclusion      <- lazyeval::uq(exclusion)
  ID             <- lazyeval::uq(ID)

  f = function(bbox)
  {
    assert_is_valid_context(LIDRCONTEXTITS, "silva2016", null_allowed = TRUE)

    # If no seed it means we received an sf with 0 data
    if (nrow(treetops) == 0L)
    {
      crown <- chm
      crown <- raster_set_values(crown, NA_integer_)
      return(crown)
    }

    if (raster_is_proxy(chm) & missing(bbox))
      stop("Cannot segment the trees from a raster stored on disk. Use segment_trees() or load the raster in memory", call. = FALSE)

    # If a bbox is given we crop the CHM and the seed to this extent to reduce processing time
    # Otherwise we could get a chm much bigger than the LAS (e.g. LAScatalog processing) and
    # process would never ends
    res <- crop_special_its(treetops, chm, bbox)
    treetops <- res$treetops
    chm <- res$chm

    st_chm <- chm
    if (!is(chm, "stars"))
      st_chm <- stars::st_as_stars(chm)

    # If no seed, exit with with warning
    if (nrow(treetops) == 0L)
    {
      warning("No tree can be used as seed", call. = FALSE)
      crown <- chm
      crown <- raster_set_values(crown, NA_integer_)
      return(crown)
    }

    chmdt <- raster_as_dataframe(chm, xy = FALSE, na.rm = TRUE)
    data.table::setDT(chmdt)

    ids <- treetops[[ID]]

    # Voronoi tesselation is nothing else but the nearest neigbour
    coords <- sf::st_coordinates(treetops)
    u <- C_knn(coords[,1], coords[,2], chmdt$X, chmdt$Y, 1L, getThread())

    id <- d <- hmax <- Z <- . <- X <- Y <- NULL
    chmdt[, id := u$nn.idx[,1]]
    chmdt[, id := ids[id]]
    chmdt[, d := u$nn.dist[,1]]
    chmdt[, hmax := max(Z), by = id]

    chmdt <- chmdt[Z >= exclusion*hmax & d <= max_cr_factor*hmax, .(X,Y, id)]

    crown <- chm
    crown <- raster_set_values(crown, NA_integer_)
    cells <- raster_cell_from_xy(crown, chmdt$X, chmdt$Y)
    crown <- raster_set_values(crown, chmdt[["id"]], cells)
    return(crown)
  }

  f <- plugin_its(f, omp = TRUE, raster_based = TRUE)
  return(f)
}

#' Individual Tree Segmentation Algorithm
#'
#' This function is made to be used in \link{segment_trees}. It implements an algorithm for tree
#' segmentation based on a watershed. It is based on the bioconductor package \code{EBIimage}. You
#' need to install this package to run this method (see its \href{https://github.com/aoles/EBImage}{github page}).
#' Internally, the function EBImage::watershed is called.
#'
#' Because this algorithm works on a CHM only there is no actual need for a point cloud. Sometimes the
#' user does not even have the point cloud that generated the CHM. \code{lidR} is a point cloud-oriented
#' library, which is why this algorithm must be used in \link{segment_trees} to merge the result into the point
#' cloud. However, the user can use this as a stand-alone function like this:
#' \preformatted{
#'  chm <- raster("chm.tif")
#'  crowns <- watershed(chm)()
#' }
#'
#' @template param-chm-lastrees
#' @param th_tree numeric. Threshold below which a pixel cannot be a tree. Default is 2.
#' @param tol numeric. Tolerance see ?EBImage::watershed.
#' @param ext numeric. see ?EBImage::watershed.
#'
#' @export
#'
#' @family individual tree segmentation algorithms
#' @family raster based tree segmentation algorithms
#'
#' @examples
#' \dontrun{
#' LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
#' poi <- "-drop_z_below 0 -inside 481280 3812940 481320 3812980"
#' las <- readLAS(LASfile, select = "xyz", filter = poi)
#' col <- pastel.colors(250)
#'
#' # Using raster because focal does not exist in stars
#' chm <- rasterize_canopy(las, res = 0.5, p2r(0.3), pkg = "raster")
#' ker <- matrix(1,3,3)
#' chm <- raster::focal(chm, w = ker, fun = mean, na.rm = TRUE)
#' las <- segment_trees(las, watershed(chm))
#'
#' plot(las, color = "treeID", colorPalette = col)
#' }
#' @md
#' @name its_watershed
watershed = function(chm, th_tree = 2, tol = 1, ext = 1)
{
  assert_is_a_number(th_tree)
  assert_is_a_number(tol)
  assert_is_a_number(ext)
  assert_package_is_installed("EBImage")

  # Workaround for #622 If segment_tree is ran in parallel it will fail with
  # SpatRaster because they are not serializable. SpatRaster are converted to RasterLayer
  # for multicore strategies
  chm <- convert_ondisk_spatraster_into_serializable_raster_if_necessary(chm)

  chm     <- lazyeval::uq(chm)
  th_tree <- lazyeval::uq(th_tree)
  tol     <- lazyeval::uq(tol)
  ext     <- lazyeval::uq(ext)

  f = function(bbox)
  {
    assert_is_valid_context(LIDRCONTEXTITS, "watershed", null_allowed = TRUE)

    if (raster_is_proxy(chm) & missing(bbox))
      stop("Cannot segment the trees from a raster stored on disk. Use segment_trees() or load the raster in memory", call. = FALSE)


    # If a bbox is given we crop the CHM and the seed to this extent to reduce processing time
    # Otherwise we could get a chm much bigger than the LAS (e.g. LAScatalog processing) and
    # process would never ends
    if (!missing(bbox)) chm <- raster_crop(chm, bbox)

    # Convert the CHM to a matrix
    Canopy <- raster_as_matrix(chm)$z
    mask   <- Canopy < th_tree | is.na(Canopy)
    Canopy[mask] <- 0
    Crowns <- EBImage::watershed(Canopy, tol, ext)

    Crowns[mask] <- NA_integer_

    val <- as.numeric(t(apply(Crowns, 1, rev)))
    storage.mode(val) <- "integer"

    output <- chm
    output <- raster_set_values(output, NA_integer_)
    output <- raster_set_values(output, val)
    return(output)
  }

  f <- plugin_its(f, omp = TRUE, raster_based = TRUE)
  return(f)
}
#' Individual Tree Segmentation Algorithm
#'
#' This functions is made to be used in \link{segment_trees}. It implements an algorithm for tree
#' segmentation based on Li et al. (2012) (see reference). This method is a growing region
#' method working at the point cloud level. It is an implementation by lidR authors, from the original
#' paper, as close as possible from the original description. However we added a parameter \code{hmin}
#' to prevent over-segmentation for objects that are too low. This algorithm is known to be slow because
#' it has an algorithmic complexity worst that O(n^2).
#'
#' @param dt1 numeric. Threshold number 1. See reference page 79 in Li et al. (2012). Default is 1.5.
#' @param dt2 numeric. Threshold number 2. See reference page 79 in Li et al. (2012). Default is 2.
#' @param R numeric. Search radius. See page 79 in Li et al. (2012). Default is 2. If \code{R = 0}
#' all the points are automatically considered as local maxima and the search step is skipped (much
#' faster).
#' @param hmin numeric. Minimum height of a detected tree. Default is 2.
#' @param Zu numeric. If point elevation is greater than Zu, \code{dt2} is used, otherwise \code{dt1} is
#' used. See page 79 in Li et al. (2012). Default is 15.
#' @param speed_up numeric. Maximum radius of a crown. Any value greater than a crown is
#' good because this parameter does not affect the result. However, it greatly affects the
#' computation speed by restricting the number of comparisons to perform.
#' The lower the value, the faster the method. Default is 10.
#'
#' @export
#'
#' @family individual tree segmentation algorithms
#' @family point-cloud based tree segmentation algorithms
#'
#' @references
#' Li, W., Guo, Q., Jakubowski, M. K., & Kelly, M. (2012). A new method for segmenting individual
#' trees from the lidar point cloud. Photogrammetric Engineering & Remote Sensing, 78(1), 75-84.
#'
#' @examples
#' LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
#' poi <- "-drop_z_below 0 -inside 481280 3812940 481320 3812980"
#' las <- readLAS(LASfile, select = "xyz", filter = poi)
#' col <- pastel.colors(200)
#'
#' las <- segment_trees(las, li2012(dt1 = 1.4))
#' #plot(las, color = "treeID", colorPalette = col)
#' @name its_li2012
#' @md
li2012 = function(dt1 = 1.5, dt2 = 2, R = 2, Zu = 15, hmin = 2, speed_up = 10)
{
  assert_is_a_number(dt1)
  assert_is_a_number(dt2)
  assert_is_a_number(R)
  assert_is_a_number(Zu)
  assert_is_a_number(hmin)
  assert_is_a_number(speed_up)
  assert_all_are_positive(dt1)
  assert_all_are_positive(dt2)
  assert_all_are_non_negative(R)
  assert_all_are_positive(Zu)
  assert_all_are_positive(hmin)
  assert_all_are_positive(speed_up)

  dt1      <- lazyeval::uq(dt1)
  dt2      <- lazyeval::uq(dt2)
  R        <- lazyeval::uq(R)
  Zu       <- lazyeval::uq(Zu)
  hmin     <- lazyeval::uq(hmin)
  speed_up <- lazyeval::uq(speed_up)

  f = function(las)
  {
    assert_is_valid_context(LIDRCONTEXTITS, "li2012")

    if (las[["Max Z"]] < hmin)
    {
      warning("'hmin' is higher than the highest point. No tree segmented.", call. = FALSE)
      return(rep(NA_integer_, nrow(las@data)))
    }
    else
    {
      return(C_li2012(las, dt1, dt2, Zu, R, hmin, speed_up))
    }
  }

  f <- plugin_its(f, omp = FALSE, raster_based = FALSE)
  return(f)
}

check_tree_tops <- function(treetops, ID)
{
  if (inherits(treetops, "Spatial"))
    treetops <- sf::st_as_sf(treetops)

  if (!is(treetops, "sf") && !is(treetops, "sfc"))
    stop("treetops must be a spatial object", call. = FALSE)

  # Special case if 0 apex is given it must work and never fail
  if (is(sf::st_geometry(treetops), "sfc_GEOMETRY") && nrow(treetops) == 0L)
    return(treetops)

  if (!is(sf::st_geometry(treetops), "sfc_POINT"))
    stop("treetops must be a spatial object of points", call. = FALSE)

  if (is(treetops, 'sfc'))
  {
    n <- length(treetops)
    treetops <- sf::st_sf(treetops)
    treetops[[ID]] <- 1:n
  }

  if (is(treetops, "sf"))
  {
    ids <- treetops[[ID]]

    if (!is.numeric(ids) )
      stop("Tree IDs much be of a numeric type",  call. = FALSE)

    if (length(unique(ids)) < length(ids))
      stop("Duplicated tree IDs found.", call. = FALSE)
  }

  return(treetops)
}

crop_special_its <- function (treetops, chm, bbox)
{
  if (!missing(bbox))
  {
    assert_is_all_of(bbox, "bbox")
    chm <- raster_crop(chm, bbox)
    if (is(chm, "stars_proxy"))
      chm <- stars::st_as_stars(chm)
    sf::st_agr(treetops) <- "constant"
    treetops <- sf::st_crop(treetops, bbox)
  }
  else
  {
    sf::st_agr(treetops) <- "constant"
    bbox = sf::st_bbox(chm)
    sf::st_crs(bbox) <- sf::st_crs(chm)
    treetops <- sf::st_crop(treetops, bbox)
  }

  return(list(treetops = treetops, chm = chm))
}

#' Get or set LAScatalog processing engine options
#'
#' The names of the options and their roles are documented in \link[=LAScatalog-class]{LAScatalog}.
#' The options are used by all the functions that support a `LAScatalog` as input. Most options are
#' easy to understand and to link to the documentation of \link[=LAScatalog-class]{LAScatalog} but some
#' need more details. See section 'Details'.
#'
#' - **opt_restart()** automatically sets the chunk option named "drop" in such a way that
#' the engine will restart at a given chunk and skip all previous chunks. Useful for restarting after a crash.
#' - **opt_independent_file()** automatically sets the chunk size, chunk buffer and wall-to-wall options
#' to process files that are not spatially related to each other, such as plot inventories.
#' - **opt_laz_compression()** automatically modifies the drivers to write LAZ files instead of LAS files.
#'
#' @param ctg An object of class \link[=LAScatalog-class]{LAScatalog}
#' @param value An appropriate value depending on the expected input.
#'
#' @name engine_options
#' @rdname engine_options
#' @family LAScatalog processing engine
#'
#' @examples
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' ctg = readLAScatalog(LASfile)
#'
#' plot(ctg, chunk_pattern = TRUE)
#'
#' opt_chunk_size(ctg) <- 150
#' plot(ctg, chunk_pattern = TRUE)
#'
#' opt_chunk_buffer(ctg) <- 10
#' plot(ctg, chunk_pattern = TRUE)
#'
#' opt_chunk_alignment(ctg) <- c(270,250)
#' plot(ctg, chunk_pattern = TRUE)
#'
#' summary(ctg)
#'
#' opt_output_files(ctg) <- "/path/to/folder/templated_filename_{XBOTTOM}_{ID}"
#' summary(ctg)
#' @md
NULL

# ========= Chunk Options ===============

#' @rdname engine_options
#' @export
opt_chunk_buffer = function(ctg)
{
  return(ctg@chunk_options$buffer)
}

#' @rdname engine_options
#' @export
`opt_chunk_buffer<-` = function(ctg, value)
{
  assert_is_a_number(value)
  if (value < 0) message("Negative buffers are allowed in lidR but use them cautiously!")
  ctg@chunk_options$buffer <- value
  return(ctg)
}

#' @rdname engine_options
#' @export
opt_chunk_size = function(ctg)
{
  return(ctg@chunk_options$size)
}

#' @rdname engine_options
#' @export
`opt_chunk_size<-` = function(ctg, value)
{
  assert_is_a_number(value)
  assert_all_are_non_negative(value)

  if (value > 0) {
     if (value < 250) {
       message("Be careful, a chunk size smaller than 250 is likely to be irrelevant.")
    }
  }

  ctg@chunk_options$size <- value
  return(ctg)
}

#' @rdname engine_options
#' @export
opt_chunk_alignment = function(ctg)
{
  return(ctg@chunk_options$alignment)
}

#' @rdname engine_options
#' @export
`opt_chunk_alignment<-` = function(ctg, value)
{
  assert_is_numeric(value)
  assert_is_of_length(value, 2)
  ctg@chunk_options$alignment <- value
  return(ctg)
}

#' @rdname engine_options
#' @export
`opt_restart<-` = function(ctg, value)
{
  assert_is_numeric(value)
  assert_is_of_length(value, 1)
  value <- as.integer(value)
  stopifnot(value >= 1)

  if (value == 1)
    ctg@chunk_options$drop <- NULL
  else
    ctg@chunk_options$drop <- 1:(value-1)

  return(ctg)
}

opt_chunk_is_file = function(ctg)
{
  return(ctg@chunk_options$size == 0)
}

# ========= Processing Options ===============

#' @rdname engine_options
#' @export
opt_progress = function(ctg)
{
  return(ctg@processing_options$progress)
}

#' @rdname engine_options
#' @export
`opt_progress<-` = function(ctg, value)
{
  assert_is_a_bool(value)
  ctg@processing_options$progress <- value
  return(ctg)
}

#' @rdname engine_options
#' @export
opt_stop_early = function(ctg)
{
  return(ctg@processing_options$stop_early)
}

#' @rdname engine_options
#' @export
`opt_stop_early<-` = function(ctg, value)
{
  assert_is_a_bool(value)
  ctg@processing_options$stop_early <- value
  return(ctg)
}

#' @rdname engine_options
#' @export
opt_wall_to_wall = function(ctg)
{
  return(ctg@processing_options$wall_to_wall)
}

#' @rdname engine_options
#' @export
`opt_wall_to_wall<-` = function(ctg, value)
{
  assert_is_a_bool(value)
  ctg@processing_options$wall_to_wall <- value
  return(ctg)
}

#' @rdname engine_options
#' @export
opt_independent_files = function(ctg)
{
  return(!opt_wall_to_wall(ctg) & opt_chunk_buffer(ctg) == 0 & opt_chunk_size(ctg) == 0)
}

#' @rdname engine_options
#' @export
`opt_independent_files<-` = function(ctg, value)
{
  assert_is_a_bool(value)

  if (isTRUE(value))
  {
    opt_wall_to_wall(ctg) <- FALSE
    opt_chunk_buffer(ctg) <- 0
    opt_chunk_size(ctg) <- 0
  }
  else
  {
    opt_wall_to_wall(ctg) <- TRUE
    opt_chunk_buffer(ctg) <- 30
    opt_chunk_size(ctg) <- 0
  }

  return(ctg)
}

# =========   Output Options   ===============

#' @rdname engine_options
#' @export
opt_output_files = function(ctg)
{
  return(ctg@output_options$output_files)
}

#' @rdname engine_options
#' @export
`opt_output_files<-` = function(ctg, value)
{
  assert_is_a_string(value)
  ext = tools::file_ext(value)

  if (ext != "")
    warning(glue::glue("{value} contains a file extension. Users don't need to provide a file extension. It will be added automaticaly as a function of the output."))

  value <- gsub("\\{?\\*\\}?", "{ORIGINALFILENAME}", value)
  usefilename <- grepl("\\{ORIGINALFILENAME\\}", value)

  if (usefilename & opt_chunk_size(ctg) > 0)
    message("ORIGINALFILENAME template has been used but the chunk size is not 0. This template makes sense only when processing by file.")

  value <- path.expand(value)
  ctg@output_options$output_files <- value
  return(ctg)
}

#' @rdname engine_options
#' @export
opt_laz_compression = function(ctg)
{
  return(ctg@output_options$drivers$LAS$extension == ".laz")
}

#' @rdname engine_options
#' @export
`opt_laz_compression<-` = function(ctg, value)
{
  assert_is_a_bool(value)

  if (value)
    ctg@output_options$drivers$LAS$extension <- ".laz"
  else
    ctg@output_options$drivers$LAS$extension <- ".las"

  return(ctg)
}

#' @rdname engine_options
#' @export
opt_merge = function(ctg)
{
  if (is.null(ctg@output_options$merge))
    return(TRUE)

  return(ctg@output_options$merge)
}

#' @rdname engine_options
#' @export
`opt_merge<-` = function(ctg, value)
{
  assert_is_a_bool(value)
  ctg@output_options$merge <- value
  return(ctg)
}

# =========   Input Options    ===============

#' @rdname engine_options
#' @export
opt_select = function(ctg)
{
  return(ctg@input_options$select)
}

#' @rdname engine_options
#' @export
`opt_select<-` = function(ctg, value)
{
  assert_is_a_string(value)
  ctg@input_options$select <- value
  return(ctg)
}

#' @rdname engine_options
#' @export
opt_filter = function(ctg)
{
  return(ctg@input_options$filter)
}

#' @rdname engine_options
#' @export
`opt_filter<-` = function(ctg, value)
{
  assert_is_a_string(value)
  ctg@input_options$filter <- value
  return(ctg)
}


# ========= Unexported ===============

`opt_copy<-` = function(ctg, value)
{
  assert_is_all_of(value, "LAScatalog")
  ctg@chunk_options <- value@chunk_options
  ctg@processing_options <- value@processing_options
  ctg@output_options <- value@output_options
  ctg@input_options <- value@input_options
  st_crs(ctg) <- st_crs(value)
  return(ctg)
}

context("classify_noise")

las = clip_rectangle(topography, 273450, 5274350, 273550, 5274450)

set.seed(314)
id = round(runif(20, 0, npoints(las)))
set.seed(42)
err = runif(20, -50, 50)
las$Z[id] = las$Z[id] + err

f = tempfile(fileext = ".las")
writeLAS(las, f)

ctg = readLAScatalog(f)
opt_chunk_size(ctg) <- 300
ctg@chunk_options$alignment = c(50, 200)
opt_chunk_buffer(ctg) <- 0
opt_progress(ctg) <- FALSE

mysor = sor(15,7)
myivf = ivf(5,2)

test_that("classify_noise sor works", {

  las <- classify_noise(las, mysor)

  n = names(las)

  expect_true("Classification" %in% n)
  expect_equal(sort(unique(las@data$Classification)), c(1L, 2L, 9L, 18L))
  expect_equal(sum(las@data$Classification == LASNOISE), 13L)

  expect_error(classify_noise(ctg, mysor), "buffer")

  opt_chunk_buffer(ctg) <- 30

  expect_error(classify_noise(ctg, mysor), "output file")

  opt_output_files(ctg) <- paste0(tempdir(), "/file_{XLEFT}_{YBOTTOM}")

  ctg2 = classify_noise(ctg, mysor)
  las2 = readLAS(ctg2)

  expect_equal(sum(las2@data$Classification == 18L), 13)
  expect_equal(nrow(las2@data), nrow(las@data))
})

test_that("classify_noise sor with quantiles", {

  las <- classify_noise(las, sor(15,0.999,TRUE))

  expect_equal(sum(las@data$Classification == 18L), 11L)
})

test_that("classify_noise ivf works", {

  las <- classify_noise(las, myivf)

  n = names(las)

  expect_true("Classification" %in% n)
  expect_equal(sort(unique(las@data$Classification)), c(1L, 2L, 9L, 18L))
  expect_equal(sum(las@data$Classification == LASNOISE), 12L)
})

data <- data.table::data.table(
  Max.X   = c(885228.88, 886993.96, 885260.93, 887025.96, 885292.94, 887056.88,
              892199.94, 893265.54, 892229.99, 893295.15, 888759.96, 890524.95,
              892259.98, 894025.98, 892289.96, 894055.93, 888790.91, 890554.98,
              888820.95, 890585.99, 892319.96, 894084.97, 892349.89, 894114.29,
              895250.23, 895094.78, 895044.96, 895053.55, 885323.96, 887087.95),
  Min.X   = c(885022.37, 885204.73, 885027.52, 885229.03, 885040.86, 885261.03,
              891503.09, 892198.69, 891501.42, 892200.07, 886970.07, 888735.55,
              891499.96, 892230.05, 890521.99, 892260.01, 886994.05, 888760.09,
              887026.07, 888791.01, 890525.05, 892290.04, 890555.01, 892320.12,
              894002.98, 894026.02, 894056.02, 894085.03, 885051.45, 885293.03),
  Max.Y   = c(630219.48, 630214.96, 631609.95, 631604.97, 633001.65, 632995.99,
              625898.35, 625882.94, 627289.82, 627273.89, 630174.88, 630134.94,
              628681.66, 628664.99, 630094.95, 630057.95, 631564.98, 631524.94,
              632955.82, 632915.99, 631486.90, 631447.96, 632876.93, 632838.96,
              628627.89, 630019.93, 631410.97, 631740.88, 634393.05, 634386.96),
  Min.Y   = c(629157.18, 629099.31, 630215.04, 630175.05, 631605.02, 631565.05,
              625816.52, 625793.60, 625883.01, 625860.81, 629036.82, 629017.72,
              627274.01, 627251.36, 628665.04, 628628.01, 630135.08, 630095.02,
              631525.01, 631487.19, 630058.02, 630020.05, 631448.08, 631411.03,
              627506.32, 628612.41, 629999.84, 631390.38, 632996.06, 632956.04),
  X.scale.factor = 0.01,
  Y.scale.factor = 0.01,
  filename = paste0("abc", 1:30)
)


geom <- lapply(1:nrow(data), function(i)
{
  mtx <- matrix(c(data$Min.X[i], data$Max.X[i], data$Min.Y[i], data$Max.Y[i])[c(1, 1, 2, 2, 1, 3, 4, 4, 3, 3)], ncol = 2)
  sf::st_polygon(list(mtx))
})

geom <-sf::st_sfc(geom)
sf::st_crs(geom) <- 26917
data <- sf::st_set_geometry(data, geom)

ctg       <- new("LAScatalog")
ctg@data  <- data

opt_progress(ctg) <- FALSE

test_that("catalog_makecluster makes correct clusters", {

  opt_chunk_size(ctg)   <-  800
  opt_chunk_buffer(ctg) <- 0

  cl <- engine_chunks(ctg)

  width   <- sapply(cl, function(x) x@width)
  buffer  <- sapply(cl, function(x) x@buffer)
  xwidth  <- sapply(cl, function(x) x@bbox[3] - x@bbox[1])
  ywidth  <- sapply(cl, function(x) x@bbox[4] - x@bbox[2])
  xbwidth <- sapply(cl, function(x) x@bbbox[3] - x@bbbox[1])
  ybwidth <- sapply(cl, function(x) x@bbbox[4] - x@bbbox[2])
  xbuffer <- sapply(cl, function(x) x@bbbox[3] - x@bbox[3])
  ybuffer <- sapply(cl, function(x) x@bbbox[4] - x@bbox[4])

  expect_equal(length(cl), 98)
  expect_true(all(width == 800))
  expect_true(all(xwidth == 800))
  expect_true(all(ywidth == 800))
  expect_true(all(xbwidth == 800))
  expect_true(all(ybwidth == 800))
  expect_true(all(buffer == 0))
  expect_true(all(xbuffer == 0))
  expect_true(all(ybuffer == 0))
})

test_that("catalog_makecluster makes correct clusters with buffer", {

  opt_chunk_size(ctg)   <- 800
  opt_chunk_buffer(ctg) <- 50

  cl <- engine_chunks(ctg)

  width   <- sapply(cl, function(x) x@width)
  buffer  <- sapply(cl, function(x) x@buffer)
  xwidth  <- sapply(cl, function(x) x@bbox[3] - x@bbox[1])
  ywidth  <- sapply(cl, function(x) x@bbox[4] - x@bbox[2])
  xbwidth <- sapply(cl, function(x) x@bbbox[3] - x@bbbox[1])
  ybwidth <- sapply(cl, function(x) x@bbbox[4] - x@bbbox[2])
  xbuffer <- sapply(cl, function(x) x@bbbox[3] - x@bbox[3])
  ybuffer <- sapply(cl, function(x) x@bbbox[4] - x@bbox[4])

  expect_equal(length(cl), 98)
  expect_true(all(width == 900))
  expect_true(all(xwidth == 800))
  expect_true(all(ywidth == 800))
  expect_true(all(xbwidth == 900))
  expect_true(all(ybwidth == 900))
  expect_true(all(buffer == 50))
  expect_true(all(xbuffer == 50))
  expect_true(all(ybuffer == 50))
})

test_that("catalog_makecluster makes correct clusters with negative buffer", {

  opt_chunk_size(ctg)   <- 800
  opt_chunk_buffer(ctg) <- -100

  cl <- engine_chunks(ctg)

  width   <- sapply(cl, function(x) x@width)
  buffer  <- sapply(cl, function(x) x@buffer)
  xwidth  <- sapply(cl, function(x) x@bbox[3] - x@bbox[1])
  ywidth  <- sapply(cl, function(x) x@bbox[4] - x@bbox[2])
  xbwidth <- sapply(cl, function(x) x@bbbox[3] - x@bbbox[1])
  ybwidth <- sapply(cl, function(x) x@bbbox[4] - x@bbbox[2])
  xbuffer <- sapply(cl, function(x) x@bbbox[3] - x@bbox[3])
  ybuffer <- sapply(cl, function(x) x@bbbox[4] - x@bbox[4])

  expect_equal(length(cl), 92)
  expect_true(all(width == 600))
})

test_that("catalog_makecluster makes correct clusters by file", {

  opt_chunk_size(ctg)   <- 0
  opt_chunk_buffer(ctg) <- 0

  cl <- engine_chunks(ctg)

  width   <- unname(sapply(cl, function(x) x@width))
  buffer  <- unname(sapply(cl, function(x) x@buffer))
  xwidth  <- unname(sapply(cl, function(x) x@bbox[3] - x@bbox[1]))
  ywidth  <- unname(sapply(cl, function(x) x@bbox[4] - x@bbox[2]))
  xbwidth <- unname(sapply(cl, function(x) x@bbbox[3] - x@bbbox[1]))
  ybwidth <- unname(sapply(cl, function(x) x@bbbox[4] - x@bbbox[2]))
  xbuffer <- unname(sapply(cl, function(x) x@bbbox[3] - x@bbox[3]))
  ybuffer <- unname(sapply(cl, function(x) x@bbbox[4] - x@bbox[4]))
  nfiles  <- unname(sapply(cl, function(x) length(x@files)))
  mainf   <- unname(sapply(cl, function(x) x@files[1]))

  expect_equal(length(cl), nrow(ctg@data))
  expect_equal(width, ctg@data$Max.X - ctg@data$Min.X)
  expect_equal(xwidth, ctg@data$Max.X - ctg@data$Min.X)
  expect_equal(ywidth, ctg@data$Max.Y - ctg@data$Min.Y)
  expect_equal(xbwidth, ctg@data$Max.X - ctg@data$Min.X)
  expect_equal(ybwidth, ctg@data$Max.Y - ctg@data$Min.Y)
  expect_true(all(buffer == 0))
  expect_true(all(xbuffer == 0))
  expect_true(all(ybuffer == 0))
  expect_true(all(nfiles == 1))
  expect_equal(mainf, ctg$filename)
})

test_that("catalog_makecluster makes correct clusters by file with buffer", {

  opt_chunk_size(ctg)   <- 0
  opt_chunk_buffer(ctg) <- 30

  cl  <- engine_chunks(ctg)

  width   <- sapply(cl, function(x) x@width)
  buffer  <- sapply(cl, function(x) x@buffer)
  xwidth  <- sapply(cl, function(x) x@bbox[3] - x@bbox[1])
  ywidth  <- sapply(cl, function(x) x@bbox[4] - x@bbox[2])
  xbwidth <- sapply(cl, function(x) x@bbbox[3] - x@bbbox[1])
  ybwidth <- sapply(cl, function(x) x@bbbox[4] - x@bbbox[2])
  xbuffer <- sapply(cl, function(x) x@bbbox[3] - x@bbox[3])
  ybuffer <- sapply(cl, function(x) x@bbbox[4] - x@bbox[4])
  nfiles  <- sapply(cl, function(x) length(x@files))
  mainf   <- sapply(cl, function(x) x@files[1])

  # Test the number of chunk
  expect_equal(length(cl), nrow(ctg@data))

  # Test that the main file is the processed file and not the buffer files
  expect_equal(mainf, ctg$filename)

  # Test the bounding box
  expect_equivalent(width, (ctg@data$Max.X - ctg@data$Min.X) + 60)
  expect_equivalent(xwidth, ctg@data$Max.X - ctg@data$Min.X)
  expect_equivalent(ywidth, ctg@data$Max.Y - ctg@data$Min.Y)
  expect_equivalent(xbwidth, (ctg@data$Max.X - ctg@data$Min.X) + 60)
  expect_equivalent(ybwidth, (ctg@data$Max.Y - ctg@data$Min.Y) + 60)

  # Test the buffer
  expect_true(all(buffer == 30))
  expect_true(all(xbuffer == 30))
  expect_true(all(ybuffer == 30))
  expect_true(all(nfiles > 1))
})

test_that("catalog_makecluster makes correct clusters by file with negative buffer", {

  opt_chunk_size(ctg)   <- 0
  opt_chunk_buffer(ctg) <- -30

  cl <- engine_chunks(ctg)

  width   <- sapply(cl, function(x) x@width)
  buffer  <- sapply(cl, function(x) x@buffer)
  xwidth  <- sapply(cl, function(x) x@bbox[3] - x@bbox[1])
  ywidth  <- sapply(cl, function(x) x@bbox[4] - x@bbox[2])
  xbwidth <- sapply(cl, function(x) x@bbbox[3] - x@bbbox[1])
  ybwidth <- sapply(cl, function(x) x@bbbox[4] - x@bbbox[2])
  xbuffer <- sapply(cl, function(x) x@bbbox[3] - x@bbox[3])
  ybuffer <- sapply(cl, function(x) x@bbbox[4] - x@bbox[4])
  nfiles  <- sapply(cl, function(x) length(x@files))

  expect_equal(length(cl), nrow(ctg@data))
  expect_equivalent(width, (ctg@data$Max.X - ctg@data$Min.X) - 60)
  expect_equivalent(xwidth, ctg@data$Max.X - ctg@data$Min.X)
  expect_equivalent(ywidth, ctg@data$Max.Y - ctg@data$Min.Y)
  expect_equivalent(xbwidth, (ctg@data$Max.X - ctg@data$Min.X) - 60)
  expect_equivalent(ybwidth, (ctg@data$Max.Y - ctg@data$Min.Y) - 60)
  expect_true(all(buffer == -30))
  expect_true(all(xbuffer == -30))
  expect_true(all(ybuffer == -30))
  expect_true(all(nfiles == 1))
})

test_that("catalog_makecluster realign the chunks with a raster", {
  opt_chunk_size(ctg)   <- 1200
  opt_chunk_buffer(ctg) <- 0

  cl <- engine_chunks(ctg, realignment = list(res = 55, start = c(0,0)))

  width   <- sapply(cl, function(x) x@width)
  buffer  <- sapply(cl, function(x) x@buffer)
  xwidth  <- sapply(cl, function(x) x@bbox[3] - x@bbox[1])
  ywidth  <- sapply(cl, function(x) x@bbox[4] - x@bbox[2])
  xbwidth <- sapply(cl, function(x) x@bbbox[3] - x@bbbox[1])
  ybwidth <- sapply(cl, function(x) x@bbbox[4] - x@bbbox[2])

  expect_true(all(width == 1210))
  expect_true(all(xwidth == 1210))
  expect_true(all(ywidth == 1210))
  expect_true(all(buffer == 0))
  expect_equal(length(cl), 54)
})

test_that("catalog_makecluster extend the chunks with a raster by file", {
  opt_chunk_size(ctg)   <- 0
  opt_chunk_buffer(ctg) <- 0

  cl <- engine_chunks(ctg, realignment = list(res = 55, start = c(0,0)))

  width   <- unname(sapply(cl, function(x) x@width))
  buffer  <- unname(sapply(cl, function(x) x@buffer))
  xwidth  <- unname(sapply(cl, function(x) x@bbox[3] - x@bbox[1]))
  ywidth  <- unname(sapply(cl, function(x) x@bbox[4] - x@bbox[2]))
  xbwidth <- unname(sapply(cl, function(x) x@bbbox[3] - x@bbbox[1]))
  ybwidth <- unname(sapply(cl, function(x) x@bbbox[4] - x@bbbox[2]))

  expect_equal(length(cl), nrow(ctg@data))
  expect_equal(width, lidR:::round_any(width, 55))
  expect_equal(xwidth, lidR:::round_any(xwidth, 55))
  expect_equal(ywidth, lidR:::round_any(ywidth, 55))
  expect_true(all(xwidth > ctg@data$Max.X - ctg@data$Min.X))
  expect_true(all(ywidth > ctg@data$Max.Y - ctg@data$Min.Y))
})

test_that("catalog_makecluster works with partial processing", {
  ctg$processed <- FALSE
  ctg$processed[c(12,13,15,18,21)] <- TRUE
  opt_chunk_size(ctg)   <- 0
  opt_chunk_buffer(ctg) <- 0

  cl <- engine_chunks(ctg)

  nfiles  <- sapply(cl, function(x) length(x@files))
  mainf   <- sapply(cl, function(x) x@files[1])

  expect_equal(length(cl), 5)
  expect_equal(mainf, ctg$filename[c(12,13,15,18,21)])

  ctg$processed <- FALSE
  ctg$processed[c(12,13,15,18,21)] <- TRUE
  opt_chunk_size(ctg)   <- 0
  opt_chunk_buffer(ctg) <- 30

  cl <- engine_chunks(ctg)

  nfiles  <- sapply(cl, function(x) length(x@files))
  mainf   <- sapply(cl, function(x) x@files[1])

  expect_equal(length(cl), 5)
  expect_equal(mainf, ctg$filename[c(12,13,15,18,21)])

  ctg$processed <- FALSE
  ctg$processed[c(12,13,15,18,21)] <- TRUE
  opt_chunk_size(ctg)   <- 1000
  opt_chunk_buffer(ctg) <- 30

  cl <- engine_chunks(ctg)

  nfiles  <- sapply(cl, function(x) length(x@files))

  expect_equal(length(cl), 20)
})

test_that("catalog_makecluster makes no cluster that belong on a tile only with buffer", {

  opt_chunk_size(ctg)   <- 800
  opt_chunk_buffer(ctg) <- 150
  opt_chunk_alignment(ctg) <- c(-100, -150)

  cl <- engine_chunks(ctg)

  expect_equal(length(cl), 100L)
})

project <- megaplot_ctg

test_that("catalog_makecluster makes correct clusters that do not overlap", {

  opt_chunk_buffer(project) <- 15
  opt_chunk_size(project)   <- 120

  cluster <- engine_chunks(project)

  x <- unlist(lapply(cluster, function(cl) {c(cl@bbox[1], cl@bbox[3])}))
  y <- unlist(lapply(cluster, function(cl) {c(cl@bbox[2], cl@bbox[4])}))

  expect_equal(length(unique(x)), 4L)
  expect_equal(length(unique(y)), 4L)
})

test_that("catalog_makecluster throw error when using ORIGINALFILENAME", {

  opt_output_files(ctg) <- "{*}"
  opt_chunk_size(ctg) <- 1000

  expect_error(engine_chunks(ctg), "makes sense only when processing by file")
})

test_that("plot overlaps works", {
  expect_error(plot(ctg, overlaps = TRUE), NA)
})

context("rasterize_density")

ctg  <- megaplot_ctg
las  <- megaplot

opt_chunk_buffer(ctg)    <- 0
opt_chunk_size(ctg)      <- 160
opt_chunk_alignment(ctg) <- c(50, 92)
opt_progress(ctg)        <- FALSE

slr = lidR:::raster_class()
mlr = lidR:::raster_multilayer_class()

test_that("rasterize_density returns the same both with LAScatalog and LAS", {

  d1 <- rasterize_density(ctg)
  d2 <- rasterize_density(las)

  expect_is(d1, slr)
  expect_equal(mean(lidR:::raster_values(d1)), 1.49, tolerance = 0.001)
  expect_equal(lidR:::raster_names(d2), "density")
  expect_equivalent(d1, d2)
})

test_that("rasterize_density returns pulse density", {

  las <- retrieve_pulses(las)
  d1  <- rasterize_density(las)

  expect_is(d1, mlr)
  expect_equal(lidR:::raster_names(d1), c("point_density", "pulse_density"))
})

test_that("grid_density is backward compatible", {
  skip_if_not_installed("raster")
  las <- retrieve_pulses(las)
  d1 <- grid_density(las)

  expect_is(d1, "RasterBrick")
  expect_equal(lidR:::raster_names(d1), c("point_density", "pulse_density"))
})

if (Sys.getenv("LIDR_EXTENSIVE_TESTS") == "TRUE")
{
  gcol = gray(seq(0,1, length.out = 30))

  haliburton_bbox <- sf::st_bbox(c(xmin = 685000, xmax = 695000, ymin = 5010000, ymax = 5015000))
  haliburton <- "/media/jr/Seagate Expansion Drive/Ontario/Haliburton/Landscape LiDAR"
  haliburton <- readLAScatalog(haliburton)
  haliburton <- catalog_intersect(haliburton, haliburton_bbox, subset = "flag_processed")
  plot(haliburton)

  montmorency_bbox <- sf::st_bbox(c(xmin = 325000, xmax = 333000, ymin = 521000, ymax = 5229000))
  montmorency <- "/media/jr/Seagate Expansion Drive/Quebec/Montmorency/las/"
  montmorency <- readLAScatalog(montmorency)
  montmorency <- catalog_intersect(montmorency, montmorency_bbox, subset = "flag_processed")
  montmorency_tile33 <- readLAS(montmorency$filename[33])
  plot(montmorency)

  lidRbook <- "/home/jr/Documents/Ulaval/lidR/lidRbook/data/ENGINE/catalog/"
  lidRbook <- readLAScatalog(lidRbook)

  library(future)
  plan(multisession(workers = 2))

  test_that("las_check works",
  {
    res <- las_check(haliburton)
    res <- lapply(res, length)
    res <- unname(unlist(res))
    expect_equal(res, c(0,0,0))

    res <- las_check(montmorency)
    res <- lapply(res, length)
    res <- unname(unlist(res))
    expect_equal(res, c(0,2,0))
  })

  test_that("rasterize_canopy works with haliburton with small chunks with stars",
  {
    # Small chunk to trigger potential errors with empty chunks
    # Approx 7 minutes
    options(lidR.raster.default = "stars")
    opt_chunk_size(haliburton) <- 250
    opt_output_files(haliburton) <- "{tempdir()}/haliburton-canopy-{ID}"
    chm <- rasterize_canopy(haliburton, 1, p2r(0.15))
    expect_is(chm, "stars_proxy")

    plot(haliburton)
    plot(chm, col = height.colors(25), breaks = "equal", add = T)
  })

  test_that("rasterize_canopy works with haliburton by files with stars",
  {
    # Approx 40 seconds
    options(lidR.raster.default = "stars")
    opt_chunk_size(haliburton) <- 0
    opt_output_files(haliburton) <- ""
    chm = rasterize_canopy(haliburton, 2, p2r())
    expect_is(chm, "stars")

    plot(chm, col = height.colors(25), breaks = "equal")
  })

  test_that("rasterize_canopy works with haliburton by files with raster",
  {
    # Approx 40 seconds
    options(lidR.raster.default = "raster")
    opt_chunk_size(haliburton) <- 0
    opt_output_files(haliburton) <- ""
    chm = rasterize_canopy(haliburton, 2, p2r())
    expect_is(chm, "RasterLayer")
    expect_false(raster::inMemory(chm))

    plot(chm, col = height.colors(25))
  })

  test_that("rasterize_canopy works with haliburton by files with terra",
  {
    # Approx 40 seconds
    options(lidR.raster.default = "terra")
    opt_chunk_size(haliburton) <- 0
    opt_output_files(haliburton) <- ""
    chm = rasterize_canopy(haliburton, 2, p2r())
    expect_is(chm, "SpatRaster")

    plot(chm, col = height.colors(25))
  })

  test_that("rasterize_terrain works with montmorency with small chunks",
  {
    options(lidR.raster.default = "stars")

    # Small chunk to trigger potential errors with empty chunks
    # Approx 10 minutes
    # Need to check if Nearest neighbour was used but interpolation is weak for those points is
    # by passed for the buffer
    opt_chunk_size(montmorency) <- 250
    opt_output_files(montmorency) <- "{tempdir()}/montmorency-terrain-{ID}"
    dtm <- rasterize_terrain(montmorency, 2, tin())
    expect_is(dtm, "stars_proxy")
    expect_equal(basename(dtm[[1]]), "rasterize_terrain.vrt")

    plot(dtm, breaks = "equal", nbreaks = 50)

    ntile33_1 <- normalize_height(montmorency_tile33, dtm)
    plot(ntile33_1)

    ntile33_2 <- normalize_height(montmorency_tile33, tin(), dtm = dtm)
    plot(ntile33_2)
  })

  test_that("rasterize_terrain works with montmorency with by files", {

    options(lidR.raster.default = "stars")

    # Small chunk to trigger potential errors with empty chunks
    # Approx 4 minutes
    opt_chunk_size(montmorency) <- 0
    opt_output_files(montmorency) <- "{tempdir()}/montmorency-terrain-{ID}"
    dtm <- rasterize_terrain(montmorency, 2, tin())
    expect_is(dtm, "stars_proxy")

    plot(dtm, breaks = "equal", nbreaks = 50)

    ntile33_1 <- normalize_height(montmorency_tile33, dtm)
    plot(ntile33_1)

    ntile33_2 <- normalize_height(montmorency_tile33, tin(), dtm = dtm)
    plot(ntile33_2)
  })

 test_that("rasterize_canopy works with montmorency with in memory and stars",
 {
   options(lidR.raster.default = "stars")

    # Approx 4 minutes
    opt_chunk_size(montmorency) <- 0
    opt_output_files(montmorency) <- ""
    dtm = rasterize_canopy(montmorency, 2, p2r())
    expect_is(dtm, "stars")
    expect_false(is(dtm, "stars_proxy"))

    plot(dtm, breaks = "equal")
 })

 test_that("rasterize_terrain works with montmorency with in memory and raster",
 {
    # Approx 5 minutes
    options(lidR.raster.default = "raster")
    opt_chunk_size(montmorency) <- 0
    opt_output_files(montmorency) <- ""
    dtm = rasterize_terrain(montmorency, 2)
    expect_is(dtm, "RasterLayer")
    expect_false(raster::inMemory(dtm))

    plot(dtm, col = gcol)
    ntile33_2 <- normalize_height(montmorency_tile33, tin(), dtm = dtm)
    plot(ntile33_2)
  })

 test_that("rasterize_terrain works with montmorency with in memory and terra",
 {
    # Approx ?? minutes
    options(lidR.raster.default = "terra")
    opt_chunk_size(montmorency) <- 0
    opt_output_files(montmorency) <- ""
    dtm = rasterize_terrain(montmorency, 2) # terra::mosaic run out of memory
    expect_is(dtm, "SpatRaster")
    expect_false(raster::inMemory(dtm))

    plot(dtm, col = gcol)

    ntile33_1 <- normalize_height(montmorency_tile33, dtm)
    plot(ntile33_1)

    ntile33_2 <- normalize_height(montmorency_tile33, tin(), dtm = dtm)
    expect_true(min(ntile33_2$Z) > -1)
    plot(ntile33_2)
  })

 test_that("locate_trees works with haliburton with small chunks",
 {
   # Small chunks to trigger potential errors chunks with 0 trees
   # Approx 20 minutes
   opt_chunk_size(haliburton) <- 250
   opt_output_files(haliburton) <- ""
   ttops <- locate_trees(haliburton, lmf(3)) # never end to merge
   expect_is(ttops, "sf")

   plot(haliburton)
   plot(ttops, col = height.colors(25), breaks = "equal", add = T)
  })

  test_that("lidRbook example with stars and dalponte",
  {
    ctg <- lidRbook

    opt_output_files(ctg) <- paste0(tempdir(), "/{*}_dtm_stars")
    dtm <- rasterize_terrain(ctg, 1, tin(), pkg = "stars")
    dtm
    plot(dtm)

    opt_output_files(ctg) <- paste0(tempdir(), "/{*}_norm")
    ctg_norm <- normalize_height(ctg, dtm)
    plot(ctg_norm)

    opt_output_files(ctg_norm) <- ""
    ttops <- locate_trees(ctg_norm, lmf(4), uniqueness = "bitmerge")
    plot(ttops["treeID"], cex = 0.01, pch = 19)

    opt_output_files(ctg_norm) <- paste0(tempdir(), "/chm_stars_{*}")
    chm <- rasterize_canopy(ctg_norm, 1, p2r(0.15), , pkg = "stars")
    chm
    plot(chm, col = height.colors(50), breaks = "equal")

    expect_error(dalponte2016(chm, ttops)(), "stored on disk")

    opt_output_files(ctg_norm) <- paste0(tempdir(), "/{*}_segmented")
    algo <- dalponte2016(chm, ttops)
    ctg_segmented <- segment_trees(ctg_norm, algo)

    opt_output_files(ctg_segmented) <- ""
    lasplot <- clip_circle(ctg_segmented, 338500, 5238600, 40)
    plot(lasplot, color = "treeID", bg = "white", size = 4)
    pol = crown_metrics(lasplot, NULL, geom = "convex")
    plot(pol, col = pastel.colors(50))
  })

  test_that("lidRbook example with terra and silva",
  {
    ctg <- lidRbook

    opt_output_files(ctg) <- paste0(tempdir(), "/{*}_dtm_terra")
    dtm <- rasterize_terrain(ctg, 1, tin(), pkg = "terra")
    dtm
    plot(dtm)

    opt_output_files(ctg) <- paste0(tempdir(), "/{*}_norm")
    ctg_norm <- normalize_height(ctg, dtm)
    plot(ctg_norm)

    opt_output_files(ctg_norm) <- ""
    ttops <- locate_trees(ctg_norm, lmf(4), uniqueness = "bitmerge")
    plot(ttops["treeID"], cex = 0.01, pch = 19)

    opt_output_files(ctg_norm) <- paste0(tempdir(), "/chm_terra_{*}")
    chm <- rasterize_canopy(ctg_norm, 1, p2r(0.15), pkg = "terra")
    chm
    plot(chm, col = height.colors(50))

    expect_error(silva2016(chm, ttops)(), "stored on disk")

    opt_output_files(ctg_norm) <- paste0(tempdir(), "/{*}_segmented")
    algo <- silva2016(chm, ttops)
    ctg_segmented <- segment_trees(ctg_norm, algo)

    opt_output_files(ctg_segmented) <- ""
    lasplot <- clip_circle(ctg_segmented, 338500, 5238600, 40)
    plot(lasplot, color = "treeID", bg = "white", size = 4)
    pol = crown_metrics(lasplot, NULL, geom = "convex")
    plot(pol, col = pastel.colors(50))
  })

  test_that("rasterize supports proxy",
  {
    ctg <- lidRbook

    opt_output_files(ctg) <- paste0(tempdir(), "/{*}_metrics")
    m <- pixel_metrics(ctg, ~mean(Z), 13)
    vrt <- terra::sources(m)

    # terra
    opt_output_files(ctg) <- ""
    dtm = rasterize_terrain(ctg, m)
    dtm
    plot(dtm)

    chm = rasterize_canopy(ctg, m)
    chm
    plot(chm)
    plot(chm-dtm)

    # stars
    m <- stars::read_stars(vrt, proxy = T)

    opt_output_files(ctg) <- ""
    dtm = rasterize_terrain(ctg, m)
    dtm
    plot(dtm)

    chm = rasterize_canopy(ctg, m)
    chm
    plot(chm)
    plot(chm-dtm)

    # raster
    m <- raster::raster(terra::sources(vrt))
    opt_output_files(ctg) <- ""
    dtm = rasterize_terrain(ctg, m)
    dtm
    plot(dtm)

    chm = rasterize_canopy(ctg, m)
    chm
    plot(chm-dtm)
  })

  test_that("Spat* object are serialized",
  {
    LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
    ctg <- readLAScatalog(LASfile, select = "xyz", chunk_size = 140, chunk_buffer = 0, progress = FALSE)
    opt_chunk_alignment(ctg) <- c(0,20)

    library(future)
    plan(multisession, workers = 2)
    res <- pixel_metrics(ctg, ~max(Z), 20)

    expect_is(res, "SpatRaster")
  })

}

context("voxelize_points")

las     <-  megaplot
ctg     <-  megaplot_ctg

test_that("filter_surfacepoints works", {

  vox  <- voxelize_points(las, 4)

  expect_equal(npoints(vox), 13822L)
})

test_that("filter_surfacepoints works with a LAScatalog", {

  skip_on_cran()

  opt_progress(ctg) <- FALSE
  opt_chunk_size(ctg) <- 300
  opt_chunk_alignment(ctg)<- c(120, 100)
  opt_output_files(ctg) <- paste0(tempdir(), "/file_{ID}")

  ctgdec <-  voxelize_points(ctg, 4)
  expect_equal(npoints(ctgdec), 13822)
})

test_that("filter_surfacepoints fails with unproper catalog options", {

  expect_error(voxelize_points(ctg, 1), "output file")
})

ctg = random_2files_250points

test_that("opt_independent_files set adequate option for plot inventories", {

  opt_independent_files(ctg) <- TRUE

  expect_equal(opt_chunk_buffer(ctg), 0)
  expect_equal(opt_chunk_size(ctg), 0)
  expect_equal(opt_wall_to_wall(ctg), FALSE)
  expect_equal(opt_independent_files(ctg), TRUE)
})

test_that("opt_independent_files built several DTMs without error", {

  opt_independent_files(ctg) <- TRUE
  opt_merge(ctg) <- FALSE
  opt_output_files(ctg) <- ""

  dtms = rasterize_terrain(ctg, 1, tin())

  expect_true(is.list(dtms))
  expect_true(is(dtms[[1]], "SpatRaster"))
})

ctg  <- megaplot_ctg
opt_progress(ctg) <- FALSE

opt_chunk_size(ctg) <- 160
ctg@chunk_options$alignment <- c(684750, 5017760)
opt_chunk_buffer(ctg) <- 0

test_that("functions throw error were needed", {
  expect_error(normalize_height(ctg, tin()), "buffer")
  expect_error(classify_ground(ctg, csf()), "buffer")

  opt_chunk_buffer(ctg) <- 10

  expect_error(normalize_height(ctg, tin()), "output file")
  expect_error(classify_ground(ctg, csf()), "output file")
  expect_error(decimate_points(ctg, homogenize(5)), "output file")
})

test_that("User cannot use one of readXXXLAS in catalog_apply", {
  myfun <- function(cluster, ...) {
    las <- readTLSLAS(cluster)
    if (is.empty(las)) return(NULL)
    return(0)
  }

  expect_error(catalog_apply(megaplot_ctg, myfun), "Use readLAS")
})

ctg <- random_2files_250points
opt_chunk_size(ctg) <- 151
opt_progress(ctg) <- FALSE

test_that("catalog_apply makes strict non-overlaping chunks", {

  # Count the points
  test <- function(cluster)
  {
    las <- readLAS(cluster)
    if (is.empty(las)) return(NULL)
    return(nrow(las@data))
  }

  req <- catalog_apply(ctg, test)
  s1  <- do.call(sum, req)
  s2  <- sum(ctg@data$Number.of.point.records)

  expect_equal(length(req), 2L)
  expect_equal(s1, s2)

  skip_on_cran()

  # Count the first return
  test <- function(cluster)
  {
    las <- readLAS(cluster)
    if (is.empty(las)) return(NULL)
    return(sum(las@data$ReturnNumber == 1))
  }

  req <- catalog_apply(ctg, test)
  s1  <- do.call(sum, req)
  s2  <- sum(ctg@data$Number.of.1st.return)

  expect_equal(length(req), 2L)
  expect_equal(s1, s2)
})

test_that("catalog_apply fixes chunk alignment", {

  test <- function(cluster, res)
  {
    las <- readLAS(cluster)
    if (is.empty(las)) return(NULL)
    return(npoints(las))
  }

  res = 8

  # Expected
  R0 <- list(381, 119)

  # Without option
  R1 <- catalog_apply(ctg, test, res = res)

  # With option
  option <- list(raster_alignment = res)
  expect_message(R2 <- catalog_apply(ctg, test, res = res, .options = option), "Chunk size changed to 152")

  expect_equal(R0, R2)
  expect_true(!identical(R0, R1))
})

test_that("catalog_apply fixes chunk alignment even by file", {

  test <- function(cluster, res, align)
  {
    las <- readLAS(cluster)
    if (is.empty(las)) return(NULL)
    r = pixel_metrics(las, ~max(Z), res, align)
    return(r)
  }

  res = 8
  sta = c(8,8)

  opt_chunk_size(ctg) <- 0
  las = readLAS(ctg)

  # Reference
  R0 = pixel_metrics(las, ~max(Z), res = res, start = sta)

  # Without option
  R1 <- catalog_sapply(ctg, test, res = res, align = sta)

  # With option
  option <- list(raster_alignment = list(res = res, start = sta))
  R2 <- catalog_sapply(ctg, test, res = res, align = sta, .options = option)

  expect_equal(R0, R2)
  #expect_message(catalog_apply(ctg, test, res = 8, .options = option), "Chunk size changed")
})

test_that("catalog_apply generates errors if function does not return NULL for empty chunks", {

  test <- function(cluster)
  {
    las <- readLAS(cluster)
    return(0)
  }

  expect_error(catalog_apply(ctg, test), "not return NULL for empty chunks")
})

test_that("catalog_apply use alternative directories", {

  skip_on_cran()

  test <- function(cluster)
  {
    las <- readLAS(cluster)
    if (is.empty(las)) return(NULL)
    return(0)
  }

  realdir <- paste0(dirname(ctg$filename[1]), "/")
  falsedir <- "/home/user/data/"

  ctg@data$filename <- paste0(falsedir,basename(ctg$filename))

  expect_error(catalog_apply(ctg, test))

  ctg@input_options$alt_dir = c(falsedir,realdir)

  expect_error(catalog_apply(ctg, test), NA)
})

test_that("catalog_apply return a partial ouptut and generates logs", {

  test <- function(cluster) {
    if (st_bbox(cluster)[2] > 80) stop("Test error")
    return(data.frame(X = 1:3))
  }

  opt_chunk_size(ctg) <- 0
  opt_wall_to_wall(ctg) <- FALSE

  option <- list(automerge = TRUE)
  expect_message(req1 <- catalog_apply(ctg, test), "chunk2.rds")
  expect_message(req2 <- catalog_apply(ctg, test, .options = option), "Test error")

  expect_is(req1, "list")
  expect_equal(length(req1), 1L)
  expect_is(req2, "data.frame")
  expect_equal(nrow(req2), 3L)
})

test_that("catalog_apply can bypass errors", {

  test <- function(cluster) {
    if (raster::extent(cluster)@ymin > 80) stop("Test error")
    return(data.frame(X = 1:3))
  }

  opt_chunk_size(ctg) <- 0
  opt_wall_to_wall(ctg) <- FALSE
  opt_stop_early(ctg) <- FALSE

  expect_message(catalog_apply(ctg, test), NA)
})

test_that("catalog_apply works with no future option", {

  options(lidR.no.future = TRUE)

  opt_chunk_size(ctg) <- 0
  opt_wall_to_wall(ctg) <- FALSE
  opt_stop_early(ctg) <- TRUE

  test <- function(cluster) {
    return(0)
  }

  expect_error(catalog_apply(ctg, test), NA)

  options(lidR.no.future = FALSE)
})

test_that("User get a warning/error when using ORIGINALFILENAME", {

  expect_message({opt_output_files(ctg) <- "{*}"}, "makes sense only when processing by file")
  expect_message({opt_output_files(ctg) <- "{ID}"}, NA)
})

test_that("User get throw error if function does not return NULL for empty chun", {

  test <- function(cluster) {
    las <- readLAS(cluster)
    return(nrow(las@data))
  }

  test2 <- function(cluster) {
    stop("dummy error", call. = FALSE)
  }

  expect_error(catalog_apply(ctg, test), "User's function does not return NULL")
  expect_error(catalog_apply(ctg, test2), "dummy error")
})

las <- lidR:::generate_las(2000)
f1  <- ~list(Zmean = mean(Z))
f2  <- ~list(meanZ = mean(Z), maxZ = max(Z))

expected_bbox  <- sf::st_bbox(c(xmin = 0, xmax = 100, ymin = 0, ymax = 100), crs = st_crs(las))
expected_bbox2 <- sf::st_bbox(c(xmin = -10, xmax = 110, ymin = -10, ymax = 110), crs = st_crs(las))

slr = lidR:::raster_class()
mlr = lidR:::raster_multilayer_class()

test_that("pixel_metrics returns a named raster", {

  x <- pixel_metrics(las, f1)

  expect_true(is(x, slr))
  expect_equal(lidR:::raster_res(x), c(20,20))
  expect_equivalent(lidR:::raster_size(x), c(5,5,1))
  expect_equivalent(sf::st_bbox(x), expected_bbox)
  expect_equal(lidR:::raster_names(x), "Zmean")
})

test_that("pixel_metrics returns a named multilayers raster", {

  x <- pixel_metrics(las, f2)
  y <- pixel_metrics(las, f1)

  expect_true(is(x, mlr))
  expect_equal(lidR:::raster_res(x), c(20,20))
  expect_equivalent(lidR:::raster_size(x), c(5,5,2))
  expect_equivalent(sf::st_bbox(x), expected_bbox)
  expect_equal(lidR:::raster_names(x), c("meanZ", "maxZ"))
})

test_that("pixel_metrics return raster terra or stars", {

  #x <- pixel_metrics(las, f2, pkg = "raster")
  #expect_true(is(x, "RasterBrick"))

  x <- pixel_metrics(las, f2, pkg = "terra")
  expect_true(is(x, "SpatRaster"))

  x <- pixel_metrics(las, f2, pkg = "stars")
  expect_true(is(x, "stars"))
})

test_that("pixel_metrics returns a raster aligned with the start option", {

  x <- pixel_metrics(las, f1, start = c(10,10))

  expect_true(is(x, slr))
  expect_equal(lidR:::raster_res(x), c(20,20))
  expect_equivalent(lidR:::raster_size(x), c(6,6, 1))
  expect_equivalent(sf::st_bbox(x), expected_bbox2)
  expect_equal(lidR:::raster_names(x), "Zmean")
})

test_that("pixel_metrics returns a named multilayers raster", {

  x <- pixel_metrics(las, f2)

  expect_true(is(x, mlr))
  expect_equal(lidR:::raster_res(x), c(20,20))
  expect_equivalent(lidR:::raster_size(x), c(5,5,2))
  expect_equivalent(sf::st_bbox(x), expected_bbox)
  expect_equal(lidR:::raster_names(x), c("meanZ", "maxZ"))
})

test_that("pixel_metrics returns a named multilayers raster aligned with the start option", {

  x <- pixel_metrics(las, f2, start = c(10,10))

  expect_true(is(x, mlr))
  expect_equal(lidR:::raster_res(x), c(20,20))
  expect_equivalent(lidR:::raster_size(x), c(6,6,2))
  expect_equivalent(sf::st_bbox(x), expected_bbox2)
  expect_equal(lidR:::raster_names(x), c("meanZ", "maxZ"))
})

test_that("pixel_metrics returns a raster  -- tricky case", {

  las2 <- filter_poi(las, X < 20 | X > 70)
  out  <- pixel_metrics(las2, f1)

  expect_equivalent(lidR:::raster_size(out), c(5,5,1))
  expect_equal(lidR:::raster_res(out), c(20, 20))

  las2 <- filter_poi(las, (X < 20 | X > 70) & (Y < 20 | Y > 70))
  out  <- pixel_metrics(las2, f1, 10)

  expect_equivalent(lidR:::raster_size(out), c(10,10,1))
  expect_equal(lidR:::raster_res(out), c(10, 10))
})

test_that("pixel_metrics return a raster -- tricky case", {

  las2 <- filter_poi(las, (X < 20 | X > 80) & (Y < 20 | Y > 80))
  out  <- pixel_metrics(las2, f2, 10)

  expect_equivalent(lidR:::raster_size(out), c(10,10,2))
  expect_equal(lidR:::raster_res(out), c(10, 10))
})

test_that("pixel_metrics splits by echo type", {

  x <- pixel_metrics(las, f1, by_echo = c("first"))

  expect_equal(lidR:::raster_res(x), c(20,20))
  expect_equivalent(lidR:::raster_size(x), c(5,5,1))
  expect_equivalent(sf::st_bbox(x), expected_bbox)
  expect_equal(lidR:::raster_names(x), "Zmean.first")

  x <- pixel_metrics(las, f1, by_echo = c("first", "lastofmany"))

  expect_equal(lidR:::raster_res(x), c(20,20))
  expect_equivalent(lidR:::raster_size(x), c(5,5,2))
  expect_equivalent(sf::st_bbox(x), expected_bbox)
  expect_equal(lidR:::raster_names(x), c("Zmean.first", "Zmean.lastofmany"))

  x <- pixel_metrics(las, f2, by_echo = c("all", "first", "lastofmany"))

  expect_equal(lidR:::raster_res(x), c(20,20))
  expect_equivalent(lidR:::raster_size(x), c(5,5,6))
  expect_equivalent(sf::st_bbox(x), expected_bbox)
  expect_equal(lidR:::raster_names(x), c("meanZ", "maxZ", "meanZ.first", "maxZ.first", "meanZ.lastofmany", "maxZ.lastofmany"))
})


test_that("3 way to compute with first returns are giving the same", {

  x1 <- pixel_metrics(filter_last(megaplot), f1, 20)
  x2 <- pixel_metrics(megaplot, f1, 20, filter = ~ReturnNumber == NumberOfReturns)

  expect_equal(x1, x2)
})

las <- filter_first(megaplot)
las@data$Intensity <- NULL
ctg <- megaplot_ctg

opt_chunk_size(ctg)      <- 260
opt_chunk_alignment(ctg) <- c(160, 160)
opt_chunk_buffer(ctg)    <- 0
opt_progress(ctg)        <- FALSE
opt_select(ctg)          <- "xyz"
opt_filter(ctg)          <- "-keep_first"

bb <- as.list(round(st_bbox(las)))
bb$xmin <- bb$xmin - 160
bb$xmax <- bb$xmax - 160
bb <- sf::st_bbox(unlist(bb), crs <- st_crs(las))
stars1  <- stars::st_as_stars(bb, dx = 15, inside = TRUE)

bb <- as.list(st_bbox(las))
bb$xmin <- bb$xmin - 360
bb$xmax <- bb$xmax - 360
bb <- sf::st_bbox(unlist(bb), crs <- st_crs(las))
stars2  <- stars::st_as_stars(bb, dx = 15, inside = TRUE)

test_that("pixel_metric returns the same both with LAScatalog and LAS", {

  # We need to add a tolerance because stars::st_mosaic writes to float32 files and
  # some accuracy is lost.
  m1 <- pixel_metrics(ctg, f1, 20)
  m2 <- pixel_metrics(las, f1, 20)
  expect_equivalent(m1, m2, tolerance = 3e-8)

  m1 <- pixel_metrics(ctg, f2, 20)
  m2 <- pixel_metrics(las, f2, 20)
  expect_equivalent(m1, m2, tolerance = 3e-8)

  m1 <- pixel_metrics(ctg, f1, 20, start = c(10,10))
  m2 <- pixel_metrics(las, f1, 20, start = c(10,10))
  expect_equivalent(m1, m2, tolerance = 3e-8)
})

test_that("pixel_metric works with a RasterLayer as input instead of a resolution", {

  skip_if_not_installed("raster")
  raster1 <- as(stars1, "Raster")
  # --- partially matching bbox

  m <- pixel_metrics(las, f1, raster1)

  expect_true(is(m, "RasterLayer"))
  expect_equivalent(sf::st_bbox(m), sf::st_bbox(raster1))
  expect_equal(sum(!is.na(m[])), 75L)
  expect_equal(sum(is.na(m[])), 150L)

  # --- no matching bbox
  raster2 <- as(stars2, "Raster")

  expect_warning(m <- pixel_metrics(las, f1, raster2), "Bounding boxes are not intersecting")
  expect_true(is(m, "RasterLayer"))
  expect_equivalent(sf::st_bbox(m), sf::st_bbox(raster2))
  expect_equal(sum(is.na(m[])), 225)
})

test_that("pixel_metric works with a stars as input instead of a resolution", {

  # --- partially matching bbox

  m <- pixel_metrics(las, f1, stars1)

  expect_true(is(m, "stars"))
  #expect_equal(sf::st_bbox(m), sf::st_bbox(raster1))
  expect_equal(sum(!is.na(m[[1]])), 75L)
  expect_equal(sum(is.na(m[[1]])), 150L)

  # --- no matching bbox

  expect_warning(m <- pixel_metrics(las, f1, stars2), "Bounding boxes are not intersecting")
  #expect_equal(sf::st_bbox(m), sf::st_bbox(raster2))
  expect_equal(sum(is.na(m[[1]])), 225)
})

test_that("predefined metric set work both with a LAS and LAScatalog", {

  las <- random_500_points

  expect_error(pixel_metrics(las, .stdmetrics_z), NA)
  expect_error(pixel_metrics(las, .stdmetrics_i), NA)
  expect_error(pixel_metrics(las, .stdmetrics_rn), NA)
  expect_error(pixel_metrics(las, .stdmetrics_ctrl), NA)
  expect_error(pixel_metrics(las, .stdshapemetrics), NA)
  expect_error(pixel_metrics(ctg, .stdmetrics_z), NA)
})

test_that("Using a non empty layout return correct output (#318)", {

  ldr <- filter_poi(megaplot, ScanAngleRank >= -3, ScanAngleRank <= 3 )
  ref <- lidR:::raster_layout(ldr, 20, format = "stars")
  ref[[1]][] <- 10

  m <- pixel_metrics(ldr, ~mean(Z), ref)

  expect_equal(sum(is.na(m[[1]])), 52L)
})

test_that("grid_metrics is backward compatible", {
  skip_if_not_installed("raster")
  x <- grid_metrics(las, f1)

  expect_true(is(x, "RasterLayer"))
  expect_equal(lidR:::raster_res(x), c(20,20))
  expect_equal(dim(x), c(13,12,1))
  expect_true(sf::st_crs(x) == st_crs(las))
  expect_equal(names(x), "Zmean")
  expect_equal(mean(x[], na.rm = T), 13.459, tolerance = 0.001)

  x <- grid_metrics(las, f2)

  expect_true(is(x, "RasterBrick"))
  expect_equal(lidR:::raster_res(x), c(20,20))
  expect_equal(dim(x), c(13,12,2))
  expect_true(sf::st_crs(x) == st_crs(las))
  expect_equal(names(x), c("meanZ", "maxZ"))
  expect_equal(mean(x$meanZ[], na.rm = T), 13.459, tolerance = 0.001)

  x <- grid_metrics(las, f1, by_echo = c("first", "lastofmany"))

  expect_true(is(x, "RasterBrick"))
  expect_equal(lidR:::raster_res(x), c(20,20))
  expect_equal(dim(x), c(13,12,2))
  expect_true(sf::st_crs(x) == st_crs(las))
  expect_equal(names(x), c("Zmean.first", "Zmean.lastofmany"))
  expect_equal(mean(x$Zmean.first[], na.rm = T), 13.459, tolerance = 0.001)
  expect_true(all(is.na(x$Zmean.lastofmany[])))


  m1 <- grid_metrics(ctg, f1, 20)
  m2 <- grid_metrics(las, f1, 20)
  expect_equivalent(m1, m2)
  expect_equal(names(m2), "Zmean")

  m1 <- grid_metrics(ctg, f2, 20)
  m2 <- grid_metrics(las, f2, 20)
  expect_equivalent(m1, m2)
  expect_equal(names(m2), c("meanZ", "maxZ"))
})

ctg <- random_2files_250points
ctg@output_options$drivers$Raster$param$overwrite = TRUE
ctg@output_options$drivers$Spatial$param$overwrite = TRUE
ctg@output_options$drivers$SimpleFeature$param$delete_dsn = TRUE

rtest <- function(cluster, layers = 1L) {
  las <- readLAS(cluster)
  if (is.empty(las)) return(NULL)
  r = lidR:::raster_layout(las, 4, format = "raster")
  r[] <- 1
  r[55:56] <- NA
  if (layers > 1) {
    r <- raster::brick(r,r)
    raster::crs(r) <- as(st_crs(las), "CRS")
    layers <- 2L
  }
  lidR:::raster_names(r) <- paste0("layername", 1:layers)
  return(r)
}

starstest <- function(cluster, layers = 1L) {
  las <- readLAS(cluster)
  if (is.empty(las)) return(NULL)
  r = lidR:::raster_layout(las, 4, format = "stars")
  r[[1]][] <- 1
  r[[1]][55:56] <- NA
  if (layers > 1) {
    r <- c(r,r, along = 3)
    sf::st_crs(r) <- st_crs(las)
    layers <- 2L
  }

  lidR:::raster_names(r) <- paste0("layername", 1:layers)
  return(r)
}

sftest <- function(cluster, DataFrame = TRUE) {
  las <- readLAS(cluster)
  if (is.empty(las)) return(NULL)
  D <- head(lidR:::coordinates(las))
  D <- sf::st_as_sf(D, coords = c("X", "Y"), crs = cluster@crs)
  D$u <- 1:6

  if (!DataFrame )
    return(sf::st_geometry(D))
  else
    return(D)
}

sptest <- function(cluster, DataFrame = FALSE) {
  x = sftest(cluster, DataFrame)
  if (is.null(x)) return(NULL)
  return(sf::as_Spatial(x))
}

lastest <- function(cluster) {
  las <- readLAS(cluster)
  if (is.empty(las)) return(NULL)
  las[1:6]
}

dftest <- function(cluster) {
  las <- readLAS(cluster)
  if (is.empty(las)) return(NULL)
  head(lidR:::coordinates3D(las))
}

terratest <- function(cluster, DataFrame = FALSE) {
  x <- sftest(cluster, DataFrame)
  if (is.null(x)) return(NULL)
  v <- terra::vect(x)
  terra::crs(v) <- sf::st_crs(x)$wkt
  return(v)
}

expected_bbox <- sf::st_bbox(c(xmin = 0,xmax = 100, ymin =  0, ymax = 200))

test_that("catalog_apply automerge works with in memory RasterLayer", {
  skip_if_not_installed("raster")

  # No automerge option
  req1 <- catalog_apply(ctg, rtest)

  # automerge option
  option <- list(automerge = TRUE)
  req2 <- catalog_apply(ctg, rtest, .options = option)

  expect_is(req1, "list")
  expect_is(req2, "RasterLayer")
  expect_true(raster::inMemory(req2))
  expect_equal(lidR:::raster_names(req2), "layername1")
  #expect_equivalent(crs(req2), crs(ctg))
  expect_equivalent(sf::st_bbox(req2), expected_bbox)
  expect_equal(sum(is.na(req2[])), 4L)
})

test_that("catalog_apply automerge works with in memory RastersBrick", {
  skip_if_not_installed("raster")
  skip_on_cran()

  # No automerge option
  req1 <- catalog_apply(ctg, rtest, layers = 2L)

  # automerge option
  option <- list(automerge = TRUE)
  req2 <- catalog_apply(ctg, rtest, layers = 2, .options = option)

  expect_is(req1, "list")
  expect_is(req2, "RasterBrick")
  expect_true(raster::inMemory(req2))
  expect_equal(lidR:::raster_names(req2), c("layername1", "layername2"))
  expect_equivalent(sf::st_bbox(req2), expected_bbox)
  expect_equal(sum(is.na(req2[])), 8L)
})

test_that("catalog_apply automerge works with on disk rasters as Raster* (VRT)", {
    skip_if_not_installed("raster")

    opt_output_files(ctg) <- paste0(tempdir(), "/{ORIGINALFILENAME}_{lidR:::uuid()}")
    option <- list(automerge = T)

    req1 <- catalog_apply(ctg, rtest, .options = option)

    expect_is(req1, "RasterLayer")
    expect_true(!raster::inMemory(req1))
    expect_equivalent(sf::st_bbox(req1), expected_bbox)
    expect_equal(sum(is.na(req1[])), 4L)
    expect_equal(lidR:::raster_names(req1), "layername1")

    req1 <- catalog_apply(ctg, rtest, layers = 2, .options = option)

    expect_is(req1, "RasterBrick")
    expect_true(!raster::inMemory(req1))
    expect_equivalent(sf::st_bbox(req1), expected_bbox)
    expect_equal(sum(is.na(req1[])), 8L)
    expect_equal(lidR:::raster_names(req1), c("layername1", "layername2"))
})

test_that("catalog_apply automerge works with on disk rasters as stars (VRT)", {

  opt_output_files(ctg) <- paste0(tempdir(), "/{ORIGINALFILENAME}_{lidR:::uuid()}")
  option <- list(automerge = TRUE)

  req1 <- catalog_apply(ctg, starstest, .options = option)
  res <- stars::read_stars(req1[[1]])

  expect_is(req1, "stars_proxy")
  expect_equivalent(sf::st_bbox(req1), expected_bbox)
  expect_equal(sum(is.na(res[[1]])), 4L)
  expect_equal(lidR:::raster_names(req1), "layername1")

  req1 <- catalog_apply(ctg, starstest, layers = 2, .options = option)
  res  <- stars::read_stars(req1[[1]])

  expect_is(req1, "stars_proxy")
  expect_equivalent(sf::st_bbox(req1), expected_bbox)
  expect_equal(sum(is.na(res[[1]])), 8L)
  expect_equal(lidR:::raster_names(req1), c("layername1", "layername2"))
})

test_that("catalog_apply automerge works with in memory SpatialPoints*", {
  skip_if_not_installed("sp")
  skip_on_cran()

  option <- list(automerge = TRUE)
  req2 <- catalog_apply(ctg, sptest, .options = option)

  expect_is(req2, "SpatialPoints")
  #expect_equal(req2@proj4string, crs(ctg))
  expect_equal(length(req2), 12L)

  option <- list(automerge = TRUE)
  req2 <- catalog_apply(ctg, sptest, DataFrame = TRUE, .options = option)

  expect_is(req2, "SpatialPointsDataFrame")
  #expect_equal(req2@proj4string, crs(ctg))
  expect_equal(dim(req2), c(12L,1L))
})

test_that("catalog_apply automerge works with on disk SpatialPoints*", {

  skip_if_not_installed("sp")
  skip_on_cran()

  opt_output_files(ctg) <- paste0(tempdir(), "/{ORIGINALFILENAME}_{lidR:::uuid()}")

  option <- list(automerge = TRUE)
  req3 <- catalog_apply(ctg, sptest, .options = option)

  expect_true(is.character(req3))
  expect_true(all(tools::file_ext(req3) == "shp"))
})

test_that("catalog_apply automerge works with on disk SpatVector", {

  skip_on_cran()

  opt_output_files(ctg) <- ""

  option <- list(automerge = T)
  req3 <- catalog_apply(ctg, terratest, .options = option)

  expect_is(req3, "SpatVector")
  expect_false(terra::crs(req3) == "")

  opt_output_files(ctg) <- paste0(tempdir(), "/{ORIGINALFILENAME}_{lidR:::uuid()}")

  option <- list(automerge = TRUE)
  req3 <- catalog_apply(ctg, terratest, .options = option)

  expect_true(is.character(req3))
  expect_true(all(tools::file_ext(req3) == "shp"))
})

test_that("catalog_apply automerge works with in memory POINTS", {

  skip_on_cran()

  option <- list(automerge = TRUE)
  req2 <- catalog_apply(ctg, sftest, .options = option)

  expect_is(req2, "sf")
  #expect_equal(projection(req2), "+proj=tmerc +lat_0=0 +lon_0=-55.5 +k=0.9999 +x_0=304800 +y_0=0 +ellps=clrk66 +units=m +no_defs")
  expect_equal(nrow(req2), 12L)
})

test_that("catalog_apply automerge works with on disk POINTS*", {

  skip_on_cran()

  opt_output_files(ctg) <- paste0(tempdir(), "/{ORIGINALFILENAME}_{lidR:::uuid()}")

  option <- list(automerge = TRUE)
  req3 <- catalog_apply(ctg, sftest, .options = option)

  expect_true(is.character(unlist(req3)))
  expect_true(all(tools::file_ext(req3) == "shp"))
})

test_that("catalog_apply automerge works with in memory LAS", {

  option <- list(automerge = TRUE)
  req2 <- catalog_apply(ctg, lastest, .options = option)

  expect_is(req2, "LAS")
  expect_equal(st_crs(req2), st_crs(ctg))
  expect_equal(npoints(req2), 12L)
})

test_that("catalog_apply automerge works with on disk LAS (LAScatalog)", {

  opt_output_files(ctg) <- paste0(tempdir(), "/{ORIGINALFILENAME}_{lidR:::uuid()}")

  option <- list(automerge = TRUE)
  req3 <- catalog_apply(ctg, lastest, .options = option)

  expect_is(req3, "LAScatalog")
  expect_equal(st_crs(req3), st_crs(ctg))
})

test_that("catalog_apply automerge works with in memory data.frame", {

  skip_on_cran()

  option <- list(automerge = TRUE)
  req2 <- catalog_apply(ctg, dftest, .options = option)

  expect_is(req2, "data.frame")
  expect_equal(nrow(req2), 12L)
})

test_that("catalog_apply automerge works with on disk data.frame", {

  skip_on_cran()

  opt_output_files(ctg) <- paste0(tempdir(), "/{ORIGINALFILENAME}_{lidR:::uuid()}")

  option <- list(automerge = TRUE)
  req3 <- catalog_apply(ctg, dftest, .options = option)

  expect_true(is.character(req3))
  expect_true(all(tools::file_ext(req3) == "txt"))
})

test_that("catalog_apply automerge does not fail with heterogeneous outputs", {

  test <- function(cluster) {
    if (st_bbox(cluster)[2] > 80) return(list(0))
    return(data.frame(X = 1:3))
  }

  opt_wall_to_wall(ctg) <- FALSE
  option <- list(automerge = FALSE)

  expect_warning(req <- catalog_sapply(ctg, test, .options = option), "heterogeneous objects")
  expect_is(req, "list")
  expect_is(req[[1]], "data.frame")
  expect_is(req[[2]], "list")
})

test_that("catalog_apply automerge does not fail with unsupported objects outputs", {

  test <- function(cluster) {
    x = runif(5)
    y = runif(5)
    stats::lm(y~x)
  }

  opt_wall_to_wall(ctg) <- FALSE
  option <- list(automerge = FALSE)

  expect_warning(req <- catalog_sapply(ctg, test, .options = option), "unsupported objects")
  expect_is(req, "list")
  expect_is(req[[1]], "lm")
})

test_that("catalog_sapply is the same than apply with automerge", {

  skip_if_not_installed("raster")
  skip_on_cran()

  option <- list(automerge = FALSE)
  req1 <- catalog_sapply(ctg, rtest)
  req2 <- catalog_sapply(ctg, rtest, .options = option)

  expect_is(req1, "RasterLayer")
  expect_true(raster::inMemory(req1))
  expect_equal(lidR:::raster_names(req1), "layername1")
  expect_equal(req1, req2)
})

test_that("catalog_apply automerge disabled with opt_merge = FALSE", {

  skip_if_not_installed("raster")
  skip_on_cran()

  opt_merge(ctg) <- FALSE
  option <- list(automerge = TRUE)
  req1 <- catalog_apply(ctg, rtest, .options = option)

  expect_true(is.list(req1))
  expect_true(is(req1[[1]], "RasterLayer"))

  opt_output_files(ctg) <- paste0(tempdir(), "/{ORIGINALFILENAME}_{lidR:::uuid()}")
  req2 <- catalog_apply(ctg, rtest, .options = option)

  expect_true(is.list(req2))
  expect_true(is.character(req2[[1]]))
})

context("normalize_height")

las = clip_rectangle(topography, 273450, 5274350, 273550, 5274450)
ctg = topography_ctg
dtm = rasterize_terrain(las, 1, tin())

opt_chunk_size(ctg) <- 300
ctg@chunk_options$alignment = c(50, 200)
opt_chunk_buffer(ctg) <- 0
opt_progress(ctg) <- FALSE

test_that("Each ground point is at 0 with knnidw", {
  las <- normalize_height(las, knnidw(k = 10L, p = 1))
  Z0 = las@data[Classification == 2]$Z
  expect_true(all(Z0 == 0))

  unnormalize_height(las)
})

test_that("Each ground point is at 0 with delaunay", {
  las <- normalize_height(las, tin())
  Z0 = las@data[Classification == 2]$Z
  expect_true(all(Z0 == 0))

  unnormalize_height(las)
})

test_that("Each ground point is at 0 with kriging", {

  skip_if_not_installed("gstat")

  las <- normalize_height(las, kriging(k = 3L))
  Z0 = las@data[Classification == 2]$Z
  expect_true(all(Z0 == 0))
  unnormalize_height(las)
})

test_that("Each ground point is at 0 with a RasterLayer", {
  las <- normalize_height(las, dtm)
  Z0 = las@data[Classification == 2]$Z
  expect_equal(mean(abs(Z0)), 0.06, tol = 0.002)
})

test_that("Absolute elevation is extrabytes(ed)", {

  las <- normalize_height(las, dtm)

  expect_true(is.null(las@header@VLR$Extra_Bytes))

  las <- normalize_height(las, dtm, add_lasattribute = TRUE)

  expect_equal(las@header@VLR$Extra_Bytes$`Extra Bytes Description`$Zref$data_type, 6L)
  expect_equal(las@header@VLR$Extra_Bytes$`Extra Bytes Description`$Zref$scale, las@header@PHB$`Z scale factor`)
  expect_equal(las@header@VLR$Extra_Bytes$`Extra Bytes Description`$Zref$offset, las@header@PHB$`Z offset`)
})

test_that("Error if NAs in DTM", {
  dtm[5458] <- NA

  expect_warning(nlas <- normalize_height(las, dtm), "1 points do not belong in the raster")
  expect_false(anyNA(nlas$Z))
})

test_that("normalize_height works with a LAScatalog", {
  expect_error(normalize_height(ctg, tin()), "buffer")

  opt_chunk_buffer(ctg) <- 30

  expect_error(normalize_height(ctg, tin()), "output file")

  opt_output_files(ctg) <- paste0(tempdir(), "/file_{XLEFT}_{YBOTTOM}")

  ctg2 = normalize_height(ctg, tin())
  las2 = readLAS(ctg2)

  expect_equal(nrow(las2@data), 73403L)
})

test_that("normalize_height works with minus symbol", {
  dtm = rasterize_terrain(las, 2, tin())
  expect_error(las-dtm, NA)
  expect_error(las-tin(), NA)
})

context("track_sensor")

las = topography
ctg <- topography_ctg
opt_chunk_size(ctg) <- 300
opt_chunk_alignment(ctg) <- c(50,180)
opt_chunk_buffer(ctg) <- 40
opt_progress(ctg) <- FALSE

test_that("track_sensor works with Roussel2020", {

  flightlines <- track_sensor(las, Roussel2020(pmin = 10))

  expect_is(flightlines, "sf")
  expect_is(sf::st_geometry(flightlines), "sfc_POINT")
  expect_equal(dim(flightlines), c(8,4))
  expect_equal(names(flightlines)[1:3], c("gpstime", "PointSourceID", "SCORE"))
  expect_equal(mean(sf::st_coordinates(flightlines)[,3]), 3100, tol = 3)
  expect_equal(sf::st_crs(flightlines), st_crs(las))
})

test_that("track_sensor works with Gatziolis2019", {

  flightlines <- track_sensor(las, Gatziolis2019(deltaT = 0.5))

  expect_is(flightlines, "sf")
  expect_is(sf::st_geometry(flightlines), "sfc_POINT")
  expect_equal(dim(flightlines), c(9,4))
  expect_equal(names(flightlines)[1:3], c("gpstime", "PointSourceID", "SCORE"))
  expect_equal(mean(sf::st_coordinates(flightlines)[,3]), 3100, tol = 3)
  expect_equal(sf::st_crs(flightlines), st_crs(las))
})

test_that("track_sensor returns an empty sf", {

  flightlines <- track_sensor(las, Roussel2020(pmin = 50))

  expect_is(flightlines, "sf")
  expect_equal(dim(flightlines), c(0,4))
  expect_equal(names(flightlines)[1:3], c("gpstime", "PointSourceID", "SCORE"))
})

test_that("track_sensor works with a LAScatalog", {

  skip_on_cran()

  flightlines <- track_sensor(ctg, Roussel2020(pmin = 10))

  expect_is(flightlines, "sf")
  expect_is(sf::st_geometry(flightlines), "sfc_POINT")
  expect_equal(dim(flightlines), c(8,4))
  expect_equal(names(flightlines)[1:3], c("gpstime", "PointSourceID", "SCORE"))
  expect_equal(mean(sf::st_coordinates(flightlines)[,3]), 3100, tol = 3)
  expect_equal(sf::st_crs(flightlines), st_crs(las))
})

test_that("track_sensor handle errors", {

  las2 = las
  las2@data = las@data[, !"PointSourceID"]
  expect_error(track_sensor(las2, Roussel2020(pmin = 15)), "No 'PointSourceID' attribute found")

  las2@data$PointSourceID <- 0L
  expect_error(track_sensor(las2, Roussel2020(pmin = 15)), "'PointSourceID' attribute is not populated")

  las2$gpstime <- 0
  expect_error(track_sensor(las2, Roussel2020(pmin = 15)), "'gpstime' attribute is not populated")

  las2 = las
  las2$PointSourceID[15] <- 5L
  expect_warning(track_sensor(las2, Roussel2020(pmin = 15), thin_pulse_with_time = 0), "points with same gpstime")
})

test_that("track_sensor filter duplicates", {

  las = rbind(las, las[1:25])

  expect_warning(flightlines <- track_sensor(las, Roussel2020(pmin = 10)), "duplicated")

  expect_is(flightlines, "sf")
  expect_is(sf::st_geometry(flightlines), "sfc_POINT")
  expect_equal(dim(flightlines), c(8,4))
  expect_equal(names(flightlines)[1:3], c("gpstime", "PointSourceID", "SCORE"))
  expect_equal(mean(sf::st_coordinates(flightlines)[,3]), 3100, tol = 3)
  expect_equal(sf::st_crs(flightlines), st_crs(las))
})

context("locate_trees")

las = filter_poi(mixedconifer, Z >= 0)
ctg = mixedconifer_ctg

opt_progress(ctg) <- FALSE
opt_chunk_alignment(ctg) <- c(50,60)
opt_chunk_size(ctg) <- 100
opt_chunk_buffer(ctg) <- 15

test_that("locate_trees LMF works with a LAS", {

  ttops = locate_trees(las, lmf(5))

  expect_is(ttops, "sf")
  expect_true(is.integer(ttops$treeID))
  expect_equal(dim(ttops), c(177,3))
  expect_true(sf::st_crs(ttops) == st_crs(las))

  f = function(x) { x * 0.07 + 3}
  ttops = locate_trees(las, lmf(f))

  expect_is(ttops, "sf")
  expect_true(is.integer(ttops$treeID))
  expect_equal(dim(ttops), c(205,3))
  expect_true(sf::st_crs(ttops) == st_crs(las))
})

test_that("find_trees is backward compatible", {
  skip_if_not_installed("sp")

  ttops = find_trees(las, lmf(5))

  expect_is(ttops, "SpatialPointsDataFrame")
  expect_true(is.integer(ttops$treeID))
  expect_equal(dim(ttops), c(177,2))
  expect_true(sf::st_crs(ttops) == st_crs(las))

  f = function(x) { x * 0.07 + 3}
  ttops = find_trees(las, lmf(f))

  expect_is(ttops, "SpatialPointsDataFrame")
  expect_true(is.integer(ttops$treeID))
  expect_equal(dim(ttops), c(205,2))
  expect_true(sf::st_crs(ttops) == st_crs(las))
})


test_that("locate_trees LMF works with a RasterLayer ", {
  skip_if_not_installed("raster")
  chm = grid_canopy(las, 1, p2r(0.15))

  ttops = locate_trees(chm, lmf(5))

  expect_is(ttops, "sf")
  expect_equal(dim(ttops), c(167,3))
  expect_true(sf::st_crs(ttops) == sf::st_crs(chm))

  # variable windows size
  f = function(x) { x * 0.07 + 3 }
  ttops = locate_trees(chm, lmf(f))

  expect_is(ttops, "sf")
  expect_true(is.integer(ttops$treeID))
  expect_equal(dim(ttops), c(204,3))
  expect_true(sf::st_crs(ttops) == sf::st_crs(chm))
})

test_that("locate_trees LMF works with a stars ", {

  chm = rasterize_canopy(las, 1, p2r(0.15), pkg = "stars")

  ttops = locate_trees(chm, lmf(5))

  expect_is(ttops, "sf")
  expect_equal(dim(ttops), c(167,3))
  expect_true(sf::st_crs(ttops) == sf::st_crs(chm))

  # variable windows size
  f = function(x) { x * 0.07 + 3 }
  ttops = locate_trees(chm, lmf(f))

  expect_is(ttops, "sf")
  expect_true(is.integer(ttops$treeID))
  expect_equal(dim(ttops), c(204,3))
  expect_true(sf::st_crs(ttops) == sf::st_crs(chm))
})

test_that("locate_trees LMF works with a LAScatalog", {

  ttops = locate_trees(ctg, lmf(5))

  expect_is(ttops, "sf")
  expect_true(is.integer(ttops$treeID))
  expect_equal(dim(ttops), c(177,3))
  expect_true(sf::st_crs(ttops) == st_crs(ctg))
})

test_that("locate_trees supports different unicity strategies", {

  ttops = locate_trees(las, lmf(5), uniqueness = "gpstime")

  expect_true(is.double(ttops$treeID))
  expect_equal(mean(ttops$treeID), 151436.2, tol = 0.01)

  ttops = locate_trees(las, lmf(5), uniqueness = "bitmerge")

  expect_true(is.double(ttops$treeID))
  expect_equal(mean(ttops$treeID), 2.067186e17, tolerance = 1e-5)
})

shp <- system.file("extdata", "lake_polygons_UTM17.shp", package = "lidR")
lake <- sf::st_read(shp, quiet = TRUE)

test_that("classify_poi works with LAS", {
  megaplot <- classify_poi(megaplot, LASHIGHVEGETATION, poi = ~Classification != 2, roi = lake, inverse = TRUE)
  cl <- megaplot$Classification
  expect_equal(sum(megaplot$Classification == LASHIGHVEGETATION), 71953L)
  expect_equal(sum(megaplot$Classification == LASGROUND), 7389L)

  megaplot <- classify_poi(megaplot, LASWATER, roi = lake, inverse = FALSE, by_reference = TRUE)

  expect_equal(sum(megaplot$Classification == LASHIGHVEGETATION), 71953L)
  expect_equal(sum(megaplot$Classification == LASGROUND), 2599L)
  expect_equal(sum(megaplot$Classification == LASWATER), 7038L)
  expect_reference(megaplot$Classification, cl)
})


test_that("classify_poi works with LAScatalog", {

  skip_on_cran()

  opt_chunk_size(megaplot_ctg) = 250
  opt_chunk_alignment(megaplot_ctg) = c(0, 100)
  opt_output_files(megaplot_ctg) = "{tempdir()}/{ID}_classified"
  opt_progress(megaplot_ctg) = FALSE

  megaplot_ctg <- classify_poi(megaplot_ctg, LASHIGHVEGETATION, poi = ~Classification != 2, roi = lake, inverse = TRUE)

  megaplot = readLAS(megaplot_ctg)
  expect_equal(sum(megaplot$Classification == LASHIGHVEGETATION), 71953L)
  expect_equal(sum(megaplot$Classification == LASGROUND), 7389L)

  opt_output_files(megaplot_ctg) = "{tempdir()}/{ID}_classified2"
  opt_progress(megaplot_ctg) = FALSE
  megaplot_ctg <- classify_poi(megaplot_ctg, LASWATER, roi = lake, inverse = FALSE, by_reference = TRUE)

  megaplot = readLAS(megaplot_ctg)
  expect_equal(sum(megaplot$Classification == LASHIGHVEGETATION), 71953L)
  expect_equal(sum(megaplot$Classification == LASGROUND), 2599L)
  expect_equal(sum(megaplot$Classification == LASWATER), 7038L)
})

context("catalog_apply writemode")

ctg <- random_2files_250points

test <- function(cluster) {
  las <- readLAS(cluster)
  if (is.empty(las)) return(NULL)
  head(lidR:::coordinates3D(las))
}

test_that("catalog_apply respects ORIGINALFILENAME template", {

  opt_output_files(ctg) <- paste0(tempdir(), "/{ORIGINALFILENAME}_{ID}")
  o <- catalog_apply(ctg, test)
  o <- unlist(o)

  ifiles <- paste0(tools::file_path_sans_ext(basename(ctg$filename)), "_", 1:2)
  ofiles <- tools::file_path_sans_ext(basename(o))

  expect_equal(ifiles, ofiles)
})

test_that("catalog_apply fails with ORIGINALFILENAME template if chunks != files", {

  opt_chunk_size(ctg) <- 75
  opt_output_files(ctg) <- paste0(tempdir(), "/{ORIGINALFILENAME}_{lidR:::uuid()}")

  expect_error(catalog_apply(ctg, test))
})

test_that("catalog_apply respects COORDINATES template", {

  opt_chunk_size(ctg) <- 75
  opt_output_files(ctg) <- paste0(tempdir(), "/{XLEFT}_{YBOTTOM}")
  o <- catalog_apply(ctg, test)
  o <- unlist(o)

  i <- c("0_0", "75_0", "0_75", "75_75", "0_150", "75_150")
  o <- tools::file_path_sans_ext(basename(o))

  expect_equal(i, o)
})

test_that("catalog_apply can write with custom drivers", {

  test <- function(cluster) {
    las <- readLAS(cluster)
    if (is.empty(las)) return(NULL)
    return(list(0))
  }

  opt_output_files(ctg) <- paste0(tempdir(), "/{ID}")

  expect_error(catalog_apply(ctg, test), "write an object of class list")

  ctg@output_options$drivers$list <- list(
    write     = base::saveRDS,
    object    = "object",
    path      = "file",
    extension = ".rds",
    param     = list(compress = TRUE))

  req <- catalog_apply(ctg, test)
  req <- unlist(req)

  expect_equal(basename(req), paste0(1:2, ".rds"))
  expect_true(all(file.exists(req)))
})

test_that("catalog_apply can create missing directories", {

 opt_output_files(ctg) <- paste0(tempdir(), "/subfolder/{ID}")

 req <- catalog_apply(ctg, test)

 expect_true(dir.exists(paste0(tempdir(), "/subfolder/")))
})

context("filter_surfacepoints")

LASfile <- system.file("extdata", "Megaplot.laz", package = "lidR")
las     <-  readLAS(LASfile)
ctg     <-  catalog(LASfile)

test_that("filter_surfacepoints works", {

  opt_progress(ctg) <- FALSE
  opt_chunk_size(ctg) <- 180
  opt_output_files(ctg) <- paste0(tempdir(), "/file_{ID}")
  ctgdec <-  filter_surfacepoints(ctg, 1)

  lasdec  <- filter_surfacepoints(las, 1)
  lasdec2 <- readLAS(ctgdec)

  data.table::setorder(lasdec@data, gpstime, ReturnNumber)
  data.table::setorder(lasdec2@data, gpstime, ReturnNumber)

  expect_equal(nrow(lasdec@data), 44401L)
  expect_equal(lasdec@data, lasdec2@data)
})

test_that("filter_surfacepoints fails with unproper catalog options", {

  expect_error(filter_surfacepoints(ctg, 1), "output file")
})

context("segment_snags")

las = clip_rectangle(mixedconifer, 481260, 3812921, 481300, 3812961)
las = filter_poi(las, Z >= 0)
ctg <- megaplot_ctg
opt_progress(ctg) <- FALSE
opt_chunk_size(ctg) <- 250
opt_chunk_alignment(ctg) <- c(10, 130)
opt_chunk_buffer(ctg) <- 10
opt_output_files(ctg) <- "{tempfile()}/{ID}"
opt_select(ctg) <- "2"

BBPRthrsh_mat <- matrix(c(0.80, 0.80, 0.70,
                         0.85, 0.85, 0.60,
                         0.80, 0.80, 0.60,
                         0.90, 0.90, 0.55),
                         nrow =3, ncol = 4)

test_that("Wing's method works", {
  las <- segment_snags(las, wing2015(neigh_radii = c(1.5, 1, 2), BBPRthrsh_mat = BBPRthrsh_mat))
  expect_true("snagCls" %in% names(las))
  expect_equal(as.numeric(table(las@data$snagCls)), c(7223, 93, 55, 15))
})

test_that("segment_snags works with a LAScatalog", {

 skip_on_cran()

 ctg2 <- segment_snags(ctg, wing2015(neigh_radii = c(1.5, 1, 2), BBPRthrsh_mat = BBPRthrsh_mat))
 las2 <- readLAS(ctg2)

 expect_is(ctg, "LAScatalog")
 expect_equal(sum(ctg2$Number.of.point.records), ctg$Number.of.point.records)
 expect_true("snagCls" %in% names(las2@data))
 expect_true(is.integer(las2$snagCls))
})

context("LAScatalog")

ctg <- megaplot_ctg

opt_chunk_size(ctg)      <- 140
opt_chunk_alignment(ctg) <- c(684760, 5017760)
opt_chunk_buffer(ctg)    <- 3

test_that("LAScatalog redefined behavior of $, [, and [[", {

  expect_true(is.character(ctg$File.Signature))
  expect_equal(length(ctg$File.Signature), 1)

  expect_true(is.null(ctg$U))

  expect_error(ctg$File.Signature <- 1, "cannot be modified")

  expect_true(is.character(ctg[["File.Signature"]]))
  expect_equal(length(ctg[["File.Signature"]]), 1)

  expect_true(is.null(ctg[["U"]]))

  expect_error(ctg[1:2,1:2], "not allowed")
})

test_that("LAScatalog conversion to SpatialPolygonsDataFrame works", {
  skip_if_not_installed("sp")
  spctg <- as.spatial(ctg)

  expect_true(is(spctg, "SpatialPolygonsDataFrame"))
  expect_reference(ctg@data$UserData, spctg@data$UserData)
})


test_that("LAScatalog subset works", {

  ctg <- random_2files_250points

  sf1 <- ctg[2]
  expect_is(sf1, "sf")
  expect_equal(dim(sf1), c(2,2))
  expect_equal(names(sf1)[1], "File.Source.ID")

  ctg1 <- ctg[2,]
  expect_is(ctg1, "LAScatalog")
  expect_equal(dim(ctg1), c(1,35))

  ctg1 <- ctg[c(FALSE,TRUE)]
  expect_is(ctg1, "LAScatalog")
  expect_equal(dim(ctg1), c(1,35))

  ctg1 <- ctg[c(FALSE,FALSE)]
  expect_is(ctg1, "LAScatalog")
  expect_equal(dim(ctg1), c(0,35))

  p <- sf::st_point(c(50,50))
  p <- sf::st_sfc(p)
  p <- sf::st_set_crs(p, st_crs(ctg))
  p <- sf::st_as_sf(p)
  ctg1 <- ctg[p]
  expect_is(ctg1, "LAScatalog")
  expect_equal(dim(ctg1), c(1,35))
})

las <- clip_rectangle(mixedconifer, 481270, 3812930, 481310, 3812970)
las <- filter_poi(las, Z >= 0)
ctg <- mixedconifer_ctg

n_expected <- sum(table(las$treeID) > 3)

opt_progress(ctg) <- FALSE
opt_chunk_size(ctg) <- 120
opt_chunk_alignment(ctg) <- c(50, -30)
opt_chunk_buffer(ctg) <- 20
opt_progress(ctg) <- FALSE

test_that("crown_metrics works", {
  hulls = crown_metrics(las, func = NULL)

  expect_is(hulls, "sf")
  expect_is(sf::st_geometry(hulls), "sfc_POINT")
  expect_equal(dim(hulls), c(n_expected,2))
})

test_that("crown_metrics works with bbox hulls", {
  hulls = crown_metrics(las, func = NULL, geom = "bbox")

  expect_is(hulls, "sf")
  expect_is(sf::st_geometry(hulls), "sfc_POLYGON")
  expect_equal(dim(hulls), c(n_expected,2))
})

test_that("crown_metrics works with concave hulls", {

  skip_on_cran()

  hulls = crown_metrics(las, func = NULL, geom = "concave")

  expect_is(hulls, "sf")
  expect_is(sf::st_geometry(hulls), "sfc_POLYGON")
  expect_equal(dim(hulls), c(n_expected,2))
})

test_that("crown_metrics returns NULL for no trees", {

  las@data$treeID <- NA_integer_

  expect_warning(res <- crown_metrics(las, func = NULL), "NULL returned")
  expect_true(is.null(res))
})

test_that("crown_metrics supports custom metrics", {

  hulls = crown_metrics(las, func = ~max(Z))

  expect_is(hulls, "sf")
  expect_equal(dim(hulls), c(n_expected,3))
  expect_equal(names(hulls)[1:2], c("treeID", "V1"))
  expect_true(!anyNA(hulls$V1))
})

test_that("crown_metrics works with a custom metrics", {

  hulls = crown_metrics(ctg, func = ~max(Z))

  expect_is(hulls, "sf")
  expect_is(sf::st_geometry(hulls), "sfc_POINT")
  expect_equal(dim(hulls), c(200, 3))

  hulls = crown_metrics(ctg, func = ~max(Z), geom = "bbox")

  expect_is(hulls, "sf")
  expect_is(sf::st_geometry(hulls), "sfc_POLYGON")
  expect_equal(dim(hulls), c(200,3))
})

test_that("crown_metrics throw an error if no treeID", {

  expect_error(tree_metrics(megaplot, list(`Mean Z` = mean(Z))), "not segmented")
})


test_that("delineate_crowns is backward compatible", {
  skip_if_not_installed("sp")
  hulls <- delineate_crowns(las)

  expect_is(hulls, "SpatialPolygonsDataFrame")
  expect_equal(dim(hulls), c(n_expected,4))
})

test_that("tree_metrics is backward compatible", {
  skip_if_not_installed("sp")
  metrics = tree_metrics(las, ~list(Z = max(Z), MeanZ = mean(Z), MaxI = max(Intensity)))

  expect_is(metrics, "SpatialPointsDataFrame")
  expect_equal(names(metrics), c("treeID", "Z", "MeanZ", "MaxI"))
  expect_equal(dim(metrics), c(n_expected,4))
  expect_true(sf::st_crs(metrics) == st_crs(las))

  metrics = tree_metrics(las, .stdtreemetrics)

  expect_is(metrics, "SpatialPointsDataFrame")
  expect_equal(nrow(metrics), n_expected)
  expect_true(sf::st_crs(metrics) == st_crs(las))
})

test_that("tree_metrics is backward compatible and works with a LAScatalog", {
  skip_if_not_installed("sp")
  metrics = tree_metrics(ctg, ~list(Z = max(Z), MeanZ = mean(Z), MaxI = max(Intensity)))

  expect_is(metrics, "SpatialPointsDataFrame")
  expect_equal(names(metrics), c("treeID", "Z", "MeanZ", "MaxI"))
  expect_equal(dim(metrics), c(200,4))
  expect_true(sf::st_crs(metrics) == st_crs(las))

  skip_on_cran()

  metrics = tree_metrics(ctg, .stdtreemetrics)

  expect_is(metrics, "SpatialPointsDataFrame")
  expect_equal(nrow(metrics), 200)
  expect_true(sf::st_crs(metrics) == st_crs(las))
})

test_that("crown_metrics preserve tree IDs #554", {
  skip_on_cran()

  LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
  las <- readLAS(LASfile, select = "xyz0")
  crowns <- crown_metrics(las, func = .stdmetrics_z, attribute = "treeID", geom = "concave")
  tree <- filter_poi(las, treeID == 80)
  tree <- st_as_sf(tree)
  crown <- crowns[crowns$treeID == 80,]
  res = sf::st_contains(crown, tree)

  expect_true(length(res[[1]]) > 0)


  plot(header(las))
  plot(sf::st_geometry(crown), add = T)
  plot(tree, add = T)
})

context("catalog_retile")

options(lidR.interactive = FALSE)

ctg <- megaplot_ctg

opt_chunk_buffer(ctg)    <- 0
opt_chunk_size(ctg)      <- 120
opt_progress(ctg)        <- FALSE
opt_chunk_alignment(ctg) <- c(40, 90)
opt_output_files(ctg)    <- paste0(tempfile(), "/test_{XLEFT}_{YBOTTOM}")

test_that("catalog_retile preserves the number of points", {
  ctg2 <- catalog_retile(ctg)

  unlink(dirname(opt_output_files(ctg)), recursive = T)

  expect_equal(sum(ctg@data$`Number of point records`), sum(ctg2@data$`Number of point records`))
  expect_equal(nrow(ctg2@data), 4L)
})

LASfile <- example_las_path
ctg <- readLAScatalog(example_las_path, chunk_size = 100, select = "xyzrn")
cls <- engine_chunks(ctg)

test_that("readLAS reads files", {
  las <- readLAS(LASfile, select = "xyzrn")

  expect_is(las, "LAS")
  expect_equal(npoints(las), 30L)
  expect_equal(names(las), c("X", "Y", "Z", "ReturnNumber", "NumberOfReturns"))
})

test_that("readLAS reads LAScatalog", {
  las <- readLAS(ctg)

  expect_is(las, "LAS")
  expect_equal(npoints(las), 30L)
  expect_equal(names(las), c("X", "Y", "Z", "ReturnNumber", "NumberOfReturns"))
})

test_that("readLAS reads LAScluster", {
  las <- readLAS(cls[[1]])

  expect_is(las, "LAS")
  expect_equal(npoints(las), 30L)
  expect_equal(names(las), c("X", "Y", "Z", "ReturnNumber", "NumberOfReturns", "buffer"))
  expect_true((all(las$buffer == 0)))
})

test_that("readLAS returns a error for non-existing file", {
  expect_error(readLAS("unexistingfile.las"), regexp = "File does not exist")
})

test_that("readLAS reads multiple files", {
  las2 = readLAS(rep(LASfile, 3))
  expect_equal(npoints(las2), 90)
})

test_that("readLAS reads multiple files and warn about incompatibilities", {

  LASfile1 <- system.file("extdata", "Megaplot.laz", package = "lidR")
  LASfile2 <- system.file("extdata", "Topography.laz", package = "lidR")

  expect_warning(readLAS(c(LASfile1, LASfile2)), " scale factors and are incompatible")
})

test_that("readLAS throws warning for invalid files", {
  las <- readLAS(LASfile, select = "xyzrn")
  las@header@PHB$`X scale factor` <- 0.01234
  las@header@PHB$`Y scale factor` <- 0.01234
  las@data$ReturnNumber[1] <- 2L
  f <- tempfile(fileext = ".las")
  rlas:::C_writer(f, as.list(las@header), las@data)
  las  <- suppressWarnings(readLAS(f))

  expect_warning(readLAS(f), "X scale factors")
  expect_warning(readLAS(f), "Y scale factors")
  expect_warning(readLAS(f), "return number")
  expect_equal(las@header@PHB$`X scale factor`, 0.01234)
  expect_equal(las@header@PHB$`Y scale factor`, 0.01234)
})

test_that("readLAS can read a file with invalid CRS", {
  example@header@VLR$GeoKeyDirectoryTag$tags[[2]]$`value offset` = 22022
  f <- tempfile(fileext = ".las")
  rlas:::C_writer(f, as.list(example@header), example@data)
  las  <- suppressWarnings(readLAS(f))

  expect_equal(epsg(las), 22022)
  expect_equal(st_crs(las), sf::NA_crs_)
})

test_that("readLAS LAScluster throw warning when select/filter is used", {

  expect_warning(readLAS(cls[[1]], select = "i"), "Argument 'select' is not used")
  expect_warning(readLAS(cls[[1]], filter = "-drop_z_below 0"), "Argument 'filter' is not used")
})

test_that("readLAS LAScatalog gives precedence to argument select and filter", {

  opt_select(ctg) <- "xyz"
  opt_filter(ctg) <- "-keep_class 2"

  las = readLAS(ctg)

  expect_equal(names(las), c("X", "Y", "Z"))
  expect_equal(npoints(las), 3)

  las = readLAS(ctg, select = "i",  filter = "-drop_z_below 975")

  expect_equal(names(las), c("X", "Y", "Z", "Intensity"))
  expect_equal(npoints(las), 19)
})

test_that("readMSLAS reads multispectral data", {

  las1 = lidR:::generate_las(10, 1)
  las2 = lidR:::generate_las(10, 2)
  las3 = lidR:::generate_las(10, 3)
  f1 = tempfile(fileext = ".las")
  f2 = tempfile(fileext = ".las")
  f3 = tempfile(fileext = ".las")
  writeLAS(las1, f1)
  writeLAS(las2, f2)
  writeLAS(las3, f3)

  las = readMSLAS(f1, f2, f3)

  expect_equal(dim(las@data), c(30, 16))
  expect_true("ScannerChannel" %in% names(las))
  expect_equal(las@header@PHB$`Version Minor`, 4L)
  expect_equal(las@header@PHB$`Point Data Format ID`, 6L)
})

context("segment_trees")

las <- clip_rectangle(mixedconifer, 481250, 3812980, 481300, 3813030)
ctg <- mixedconifer_ctg
opt_progress(ctg) = FALSE
opt_chunk_size(ctg) = 100
opt_chunk_buffer(ctg) = 20
opt_chunk_alignment(ctg) <- c(0, 20)

chm   <- rasterize_canopy(las, 0.5, pitfree())
ker   <- matrix(1,3,3)
chm   <- terra::focal(chm, w = ker, fun = mean)
chm   <- stars::st_as_stars(chm)
ttops <- locate_trees(chm, lmf(3, 2))
n     <- nrow(ttops)

ttops_shifted500 <- ttops
sf::st_geometry(ttops_shifted500) <- sf::st_geometry(ttops_shifted500) + sf::st_sfc(sf::st_point(c(0, 500, 0)))
sf::st_crs(ttops_shifted500) <- sf::st_crs(chm)

test_that("segment_tree propagates tree IDs and preserves storage mode", {

  ttops1 <- ttops
  ttops1$treeID <- ttops1$treeID*2L
  las1 <- segment_trees(las, dalponte2016(chm, ttops1))

  ttops2 <- ttops
  ttops2$treeID <- ttops2$treeID + runif(n, -0.1, 0.1)
  las2 <- segment_trees(las, dalponte2016(chm, ttops2))

  expect_equal(sort(unique(las1$treeID)), ttops1$treeID)
  expect_equal(sort(unique(las2$treeID)), ttops2$treeID)
  expect_equal(storage.mode(las1$treeID), storage.mode(ttops1$treeID))
  expect_equal(storage.mode(las2$treeID), storage.mode(ttops2$treeID))
})

test_that("segment_trees can store in a user defined column", {

  las <- segment_trees(las, li2012(speed_up = 5), attribute = "plop")

  expect_true("plop" %in% names(las))
})

test_that("segment_trees supports different unicity strategies", {

  las <- segment_trees(las, li2012(speed_up = 5), uniqueness = "incremental", attribute = "ID1")
  las <- segment_trees(las, li2012(speed_up = 5), uniqueness = "gpstime", attribute = "ID2")
  las <- segment_trees(las, li2012(speed_up = 5), uniqueness = "bitmerge", attribute = "ID3")

  expect_equal(length(na.omit(unique(las$ID1))), 48L)
  expect_equal(length(na.omit(unique(las$ID2))), 48L)
  expect_equal(length(na.omit(unique(las$ID3))), 48L)

  u <- las@data[, if (!anyNA(.BY)) .I[which(Z == max(Z))], by = "ID1"]
  s <- las@data[u$V1]

  expectedID2 <- s$gpstime
  expectedID3 <- lidR:::bitmerge(s$X*100, s$Y*100)

  id42 = s$ID1 == 42

  expect_equal(s$ID2[!id42], expectedID2[!id42])
  expect_equal(s$ID3[!id42], expectedID3[!id42])

  expect_equal(s$ID2[id42][1], min(s$gpstime[id42]))
  expect_equal(s$ID3[id42][1], expectedID3[id42][which.min(s$X[id42])])
})

test_that("segment_trees supports a LAScatalog", {

  # Skipping this test temporarily to be able to have a build on MacOS. Then I will be
  # able to test that on somebody' computer.
  skip_on_cran()

  opt_output_files(ctg) <- "{tempdir()}/{ID}"
  new_ctg <- segment_trees(ctg, li2012(speed_up = 5), uniqueness = 'gpstime')

  las <- readLAS(new_ctg)
  las <- segment_trees(las, li2012(speed_up = 5), uniqueness = 'gpstime', attribute = "treeID2")

  expect_equal(length(na.omit(unique(las$treeID))), 274)  # 272 on Fedora and MacOS. Impossible to reproduce.
  expect_equal(length(na.omit(unique(las$treeID2))), 274) # 272 on Fedora and MacOS. Impossible to reproduce.

  skip("Need investigation")

  expect_equal(las$treeID, las$treeID2)
})

test_that("Dalponte algorithm works", {

  las <- segment_trees(las, dalponte2016(chm, ttops))

  expected_table <- c(46, 80, 143, 98, 175, 15, 175, 196, 222, 132, 91, 124, 77, 147, 100, 52, 35, 81, 102, 81,
                      49, 95, 170, 89, 72, 60, 123, 91, 98, 125, 101, 69, 159, 118, 134, 128, 136, 99, 47, 58)
  expected_table <- table(las$treeID)

  expect_true("treeID" %in% names(las))
  expect_equal(sort(unique(las$treeID)), 1:n)
  expect_equivalent(expected_table, expected_table)
})

test_that("Silva algorithm works", {

  las <- segment_trees(las, silva2016(chm, ttops))

  expected_table <- c(96, 64, 71, 131, 134, 126, 141, 110, 163, 93, 134, 104, 145, 92, 68, 81, 183, 93, 100, 65,
                      98, 71, 96, 44, 35, 80, 155, 95, 146, 92, 144, 196, 301, 177, 14, 180, 111, 186, 40, 89)
  expected_table <- table(las$treeID)

  expect_true("treeID" %in% names(las))
  expect_equal(sort(unique(las$treeID)), 1:n)
  expect_equivalent(expected_table, expected_table)
})

test_that("Watershed algorithm works", {

  skip_if_not_installed("EBImage")

  las <- segment_trees(las, watershed(chm))

  expected_table <- c(336, 271, 208, 254, 198, 289, 140, 35, 195, 258, 287, 156, 167, 174, 233,
                      150, 251, 225, 121, 44, 103, 180, 109, 87, 198, 92, 43, 78, 15)
  expected_table <- table(las$treeID)

  expect_true("treeID" %in% names(las))
  expect_equal(sort(unique(las$treeID)), 1:29)
  expect_equivalent(expected_table, expected_table)
})

test_that("Dalponte algorithm works with sfc", {

  ttops <- sf::st_geometry(ttops)
  las <- segment_trees(las, dalponte2016(chm, ttops))

  expected_table <- c(46, 80, 143, 98, 175, 15, 175, 196, 222, 132, 91, 124, 77, 147, 100, 52, 35, 81, 102, 81,
                      49, 95, 170, 89, 72, 60, 123, 91, 98, 125, 101, 69, 159, 118, 134, 128, 136, 99, 47, 58)
  expected_table <- table(las$treeID)

  expect_true("treeID" %in% names(las))
  expect_equal(sort(unique(las$treeID)), 1:n)
  expect_equivalent(expected_table, expected_table)
})

test_that("Silva algorithm works with sfc", {

  ttops <- sf::st_geometry(ttops)
  las <- segment_trees(las, silva2016(chm, ttops))

  expected_table <- c(96, 64, 71, 131, 134, 126, 141, 110, 163, 93, 134, 104, 145, 92, 68, 81, 183, 93, 100, 65,
                      98, 71, 96, 44, 35, 80, 155, 95, 146, 92, 144, 196, 301, 177, 14, 180, 111, 186, 40, 89)
  expected_table <- table(las$treeID)

  expect_true("treeID" %in% names(las))
  expect_equal(sort(unique(las$treeID)), 1:n)
  expect_equivalent(expected_table, expected_table)
})

test_that("Dalponte algorithm catches invalid IDs in input", {

  ttops$treeID[2:3] <- 1L
  expect_error(segment_trees(las, dalponte2016(chm, ttops3), "Duplicated tree IDs found"))
})

test_that("Silva algorithm catches invalid IDs in input", {

  ttops$treeID[2:3] <- 1L
  expect_error(segment_trees(las, silva2016(chm, ttops3), "Duplicated tree IDs found"))
})

test_that("Dalponte algorithm returns NAs if no seed in chm", {

  expect_warning(las <- segment_trees(las, dalponte2016(chm, ttops_shifted500)), "No tree can be used as seed")
  expect_true("treeID" %in% names(las))
  expect_true(all(is.na(las$treeID)))
  expect_equal(storage.mode(las$treeID), storage.mode(ttops_shifted500$treeID))
})

test_that("Silva algorithm returns NAs if no seed in chm", {

  expect_warning(las <- segment_trees(las, silva2016(chm, ttops_shifted500)), "No tree can be used as seed")
  expect_true("treeID" %in% names(las))
  expect_true(all(is.na(las$treeID)))
  expect_equal(storage.mode(las$treeID), storage.mode(ttops_shifted500$treeID))
})

test_that("Dalponte algorithm does not fail with no seed at all", {

  ttops <- ttops[0,]
  expect_message(segment_trees(las, dalponte2016(chm, ttops)), "No tree found")
})

test_that("Silva algorithm does not fail with no seed at all", {

  ttops <- ttops[0,]
  expect_message(segment_trees(las, silva2016(chm, ttops)), "No tree found")
})

test_that("Dalponte algorithm works standalone", {

  expect_error(trees <- dalponte2016(chm, ttops)(), NA)
  expect_true(is(trees, "stars"))
  expect_equal(sort(unique(as.numeric(trees[[1]]))), ttops$treeID)
  expect_warning(trees <- dalponte2016(chm, ttops_shifted500)(), "No tree can be used as seed")
  expect_true(is(trees, "stars"))
  expect_equal(sort(unique(as.numeric(trees[[1]]))), numeric(0))
})

test_that("Silva algorithm works standalone", {

  expect_error(trees <- silva2016(chm, ttops)(), NA)
  expect_true(is(trees, "stars"))
  expect_equal(sort(unique(as.numeric(trees[[1]]))), ttops$treeID)
  expect_warning(trees <- silva2016(chm, ttops_shifted500)(), "No tree can be used as seed")
  expect_true(is(trees, "stars"))
  expect_equal(sort(unique(as.numeric(trees[[1]]))), numeric(0))
})

test_that("Watershed algorithm works standalone", {

  skip_if_not_installed("EBImage")

  expect_error(trees <- watershed(chm)(), NA)
  expect_true(is(trees, "stars"))
  expect_equal(sort(unique(as.numeric(trees[[1]]))), 1:29)
})

test_that("Dalponte algorithm works standalone and is backward compatible", {
  skip_if_not_installed("raster")
  ttops <- locate_trees(chm, lmf(3, 2))
  chm   <- as(chm, "Raster")
  trees <- dalponte2016(chm, ttops)()

  expect_true(is(trees, "RasterLayer"))
})

test_that("Silva algorithm works standalone and is backward compatible", {
  skip_if_not_installed("raster")
  ttops <- locate_trees(chm, lmf(3, 2))
  chm   <- as(chm, "Raster")
  trees <- silva2016(chm, ttops)()

  expect_true(is(trees, "RasterLayer"))
})

test_that("Watershed algorithm works standalone and is backward compatible", {

  skip_if_not_installed("EBImage")
  skip_if_not_installed("raster")

  chm   <- as(chm, "Raster")
  trees <- watershed(chm)()

  expect_true(is(trees, "RasterLayer"))
})

test_that("Li algorithm method works", {

  las <- segment_trees(las, li2012(speed_up = 5))

  expect_equal(sort(unique(las$treeID)), 1:48L)
  expect_true(is.integer(las$treeID))

  expect_warning(segment_trees(las, li2012(hmin =  100)), "No tree segmented")
})

# nocov start
catalog_fakerun = function(ctg, sleep = 0.05, pwarning = 0.05, perror = 0.01)
{
  opt_wall_to_wall(ctg) <- FALSE
  options <- list(need_buffer = FALSE, drop_null = FALSE)

  fake = function(x)
  {
    Sys.sleep(sleep)
    if (stats::runif(1) < pwarning) warning("Fake warning", call. = FALSE)
    if (stats::runif(1) < perror) stop("Fake error", call. = FALSE)
    return(0)
  }

  return(catalog_apply(ctg, fake, .options = options))
}
# nocov end

round_any <- function(x, accuracy)
{
  roundc(x / accuracy) * accuracy
}

group_grid_3d = function(x, y, z, res, start = c(0,0,0))
{
  xgrid = f_grid(x, res[1], start[1])
  ygrid = f_grid(y, res[1], start[2])
  zgrid = f_grid(z, res[2], start[3])

  return(list(Xgrid = xgrid, Ygrid = ygrid, Zgrid = zgrid))
}

f_grid = function(x, res, start)
{
  round_any(x - 0.5 * res - start, res) + 0.5 * res + start
}

verbose = function(...)
{
  if (getOption("lidR.verbose") || getOption("lidR.debug"))
    cat(..., "\n")
}

dummy_las <- generate_las

uuid <- function()
{
  hex_digits <- c(as.character(0:9), letters[1:6])
  y_digits <- hex_digits[9:12]
  paste(
    paste0(sample(hex_digits, 8), collapse=''),
    paste0(sample(hex_digits, 4), collapse=''),
    paste0('4', sample(hex_digits, 3), collapse=''),
    paste0(sample(y_digits,1),
           sample(hex_digits, 3),
           collapse=''),
    paste0(sample(hex_digits, 12), collapse=''),
    sep='-')
}

fast_count_equal = function(x, v)
{
  if (is_compact(x))
  {
    if (x[1] == v)
      return(length(x))
    else
      return(0)
  }
  else
  {
    return(fast_countequal(x, v))
  }
}

#' Intensity normalization algorithm
#'
#' This function is made to be used in \link{normalize_intensity}. It corrects intensity with a
#' range correction according to the formula (see references):
#' \deqn{I_{norm} = I_{obs} \left(\frac{R}{Rs}\right)^f}{Inorm = Iobs * (R/Rs)^f}
#' To achieve the range correction the position of the sensor must be known at different discrete times.
#' Using the 'gpstime' of each point, the position of the sensor is interpolated from the reference
#' and a range correction is applied.
#'
#' @param sensor `SpatialPointsDataDrame` or `sf` object containing the coordinates of
#' the sensor at different timepoints t. The time and elevation are stored as attributes
#' (default names are 'gpstime' and 'Z'). Z can also come from the geometry if the input records XYZ
#' coordinates. It can be computed with \link{track_sensor}.
#' @param Rs numeric. Range of reference.
#' @param f numeric. Exponent. Usually between 2 and 3 in vegetation contexts.
#' @param gpstime,elevation character. The name of the attributes that store the gpstime of the
#' position and the elevation of the sensor respectively. If the input contains 3 coordinates points,
#' `elevation` is not considered.
#' @param las an object of class LAS. \code{get_range()} is a regular function documented here for
#' convenience.
#' @export
#' @references
#' Gatziolis, D. (2011). Dynamic Range-based Intensity Normalization for Airborne, Discrete Return
#' Lidar Data of Forest Canopies. Photogrammetric Engineering & Remote Sensing, 77(3), 251–259.
#' https://doi.org/10.14358/pers.77.3.251
#' @examples
#' # A valid file properly populated
#' LASfile <- system.file("extdata", "Topography.laz", package="lidR")
#' las <- readLAS(LASfile)
#'
#' # pmin = 15 because it is an extremely tiny file
#' # strongly decimated to reduce its size. There are
#' # actually few multiple returns
#' sensor <- track_sensor(las, Roussel2020(pmin = 15))
#'
#' # Here the effect is virtually null because the size of
#' # the sample is too small to notice any effect of range
#' las <- normalize_intensity(las, range_correction(sensor, Rs = 2000))
#'
#' # This might be useful for some applications
#' R = get_range(las, sensor)
range_correction = function(sensor, Rs, f = 2.3, gpstime = "gpstime", elevation = "Z")
{
  assert_is_a_number(Rs)
  assert_is_a_number(f)
  assert_all_are_positive(Rs)
  assert_all_are_positive(f)

  Rs <- lazyeval::uq(Rs)
  f <- lazyeval::uq(f)
  gpstime <- lazyeval::uq(gpstime)
  elevation <- lazyeval::uq(elevation)

  fun = function(las)
  {
    assert_is_valid_context(LIDRCONTEXTNIT, "range_correction")
    fl <- prepare_sensor(sensor, gpstime, elevation, las)
    return(C_lasrangecorrection(las, fl, Rs, f))
  }

  class(fun) <- LIDRALGORITHMNIT
  return(fun)
}

#' @rdname range_correction
#' @export
get_range = function(las, sensor, gpstime = "gpstime", elevation = "Z")
{
  stopifnotlas(las)
  fl <- prepare_sensor(sensor, gpstime, elevation, las)
  R <- C_lasrange(las, fl)
  return(round(R, 3))
}

prepare_sensor = function(sensor, gpstime, elevation, las)
{
  if (is(sensor, "SpatialPointsDataFrame"))
    sensor <- sf::st_as_sf(sensor)

  if (!is(sensor, "sf"))
    stop("'ttops' is not a SpatialPointsDataFrame or sf")

  geom <- sf::st_geometry(sensor)
  if (!is(geom, "sfc_POINT"))
    stop("Only point geometry types are supported", call. = FALSE)

  if (!"gpstime" %in% names(las))
    stop("No 'gpstime' attribute found in las",   call. = FALSE)

  if (!"Intensity" %in% names(las))
    stop("No 'Intensity' attribute found", call. = FALSE)

  if (!gpstime %in% names(sensor))
    stop(glue::glue("No '{gpstime}' attribute found in sensor."), call. = FALSE)

  if (is(sf::st_geometry(sensor)[[1]], "XYZ"))
  {
    if (elevation %in% names(sensor))
      message("3 coordinates detected in the sensor positions, parameter 'elevation' is not considered.")

    elevation = NULL
  }

  fl <- sf::st_coordinates(sensor)
  fl <- data.table::data.table(fl)

  trange.las <- range(las[["gpstime"]])
  trange.sensor <- range(sensor[[gpstime]])

  if ((trange.las[1] < trange.sensor[1] - 4) || (trange.las[2] > trange.sensor[2] + 4))
    stop("'gpstime range from the sensor does not contain gpstime range from the point-cloud", call. = FALSE)

  if (!is.null(elevation))
  {
    if (!elevation %in% names(sensor))
      stop(glue::glue("No '{elevation}' attribute found in sensor."), call. = FALSE)

    fl[["Z"]] <- sensor[[elevation]]
  }
  else
  {
    if (ncol(fl) != 3)
      stop("There are only two dimensions in the coordinates of the sensor.", call. = FALSE)
  }

  fl[["gpstime"]] <- sensor[[gpstime]]

  fl <- data.table::data.table(fl)
  data.table::setorder(fl, gpstime)
  data.table::setnames(fl, c("X", "Y", "Z", "gpstime"))
  dup <- duplicated(fl, by = "gpstime")

  if (any(dup))
  {
    warning("Duplicated gpstime found. Duplicated sensor positions were removed.", call. = FALSE)
    fl <- fl[!dup]
  }

 x = diff(fl$gpstime) > 30
 conseq_true <- with(rle(x), sum(lengths[values] >= 2))
 if (x[1] == TRUE | conseq_true > 0)
   warning("Detection of isolated gpstime likely corresponding to flightlines with a single sensor position.", call. = FALSE)

  return(fl)
}

#' Create an object of class LAScatalog
#'
#' Create an object of class \link[=LAScatalog-class]{LAScatalog} from a folder
#' or a collection of filenames. A LAScatalog is a representation of a collection
#' of las/laz files. A computer cannot load all the data at once. A \code{LAScatalog}
#' is a simple way to manage all the files sequentially. Most functions from
#' `lidR` can be used seamlessly with a LAScatalog using the internal
#' `LAScatalog` processing engine. To take advantage of the `LAScatalog`
#' processing engine the user must first adjust some processing options using the
#' \link[=engine_options]{appropriate functions}. Careful reading of the
#' \link[=LAScatalog-class]{LAScatalog class documentation} is required to use the
#' `LAScatalog` class correctly.\cr\cr
#' `readLAScatalog` is the original function and always works. Using one of the `read*LAScatalog` functions
#' adds information to the returned object to register a point-cloud type. Registering the correct point
#' type **may** improve the performance of some functions by enabling users to select an appropriate spatial index.
#' See \link[=lidR-spatial-index]{spatial indexing}. Notice that by legacy and for backwards-compatibility
#' reasons `readLAScatalog()` and `readALSLAScatalog()` are equivalent because lidR was originally designed
#' for ALS and thus the original function `readLAScatalog()` was (supposedly) used for ALS.
#'
#' @param folder string. The path of a folder containing a set of las/laz files.
#' Can also be a vector of file paths.
#' @param progress,select,filter,chunk_size,chunk_buffer Easily accessible processing
#' options tuning. See \link{LAScatalog-class} and \link{engine_options}.
#' @param \dots Extra parameters to \link[base:list.files]{list.files}. Typically
#' `recursive = TRUE`. Propagates also to `readLAScatalog`
#'
#' @return A \code{LAScatalog} object
#'
#' @export
#'
#' @examples
#' # A single file LAScatalog using data provided with the package
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' ctg = readLAScatalog(LASfile)
#' plot(ctg)
#'
#' \dontrun{
#' ctg <- readLAScatalog("</path/to/folder/of/las/>")
#'
#' # Internal engine will sequentially process chunks of size 500 x 500 m
#' opt_chunk_size(ctg) <- 500
#'
#' # Internal engine will align the 500 x 500 m chunks on x = 250 and y = 300
#' opt_alignment(ctg) <- c(250, 300)
#'
#' # Internal engine will not display a progress estimation
#' opt_progress(ctg) <- FALSE
#'
#' # Internal engine will not return results into R.
#' # Instead it will write results in files.
#' # Files will be named e.g.
#' # filename_256000_1.ext
#' # filename_257000_2.ext
#' # filename_258000_3.ext
#' # ...
#' opt_output_files(ctg) <- "/path/filename_{XBOTTOM}_{ID}"
#'
#' # More details in the documentation
#' help("LAScatalog-class", "lidR")
#' help("engine_options", "lidR")
#' }
#' @md
readLAScatalog <- function(folder, progress = TRUE, select = "*", filter = "", chunk_size = 0, chunk_buffer = 30, ...)
{
  assert_is_character(folder)

  is_remote_flags = is_remote(folder)

  finfo <- file.info(folder[!is_remote_flags])  # only stat local paths

  local_paths  <- folder[!is_remote_flags]
  remote_paths <- folder[is_remote_flags]

  # Expand local directories into file lists
  local_files <- character(0)
  if (length(local_paths) > 0)
  {
    local_finfo <- file.info(local_paths)

    # Check that all local paths exist
    if (any(is.na(local_finfo$isdir)))
    {
      missing <- local_paths[is.na(local_finfo$isdir)]
      stop(glue::glue("'{paste(missing, collapse = ', ')}' does not exist."), call. = FALSE)
    }

    dirs  <- local_paths[local_finfo$isdir]
    files <- local_paths[!local_finfo$isdir]

    # Expand directories
    if (length(dirs) > 0)
    {
      p <- list(...)
      p$full.names <- TRUE
      if (is.null(p$pattern)) p$pattern <- "(?i)\\.la(s|z)$"
      expanded <- unlist(lapply(dirs, function(d) { p$path <- d; do.call(list.files, p) }))
      files <- c(files, expanded)
    }

    local_files <- normalizePath(files, mustWork = FALSE)
  }

  files <- c(local_files, remote_paths)

  if (length(files) == 0)
    stop("No LAS/LAZ files found.", call. = FALSE)

  verbose("Reading files...")

  if (length(files) == 1L && tools::file_ext(files) == "vpc")
  {
    files = normalizePath(files)
    headers = read_vpc(files)
    files = sapply(headers, function(x) x$filename)
    crs = sf::st_crs(headers[[1]]$CRS)
  }
  else
  {
    header <- LASheader(rlas::read.lasheader(files[1]))
    crs    <- st_crs(header)
    phblab <- make.names(names(phb(header)))
    phblab[4] <- "GUID"

    # Delayed progress bar
    t0 <- Sys.time()
    pb <- NULL
    i  <- 0

    headers <- lapply(files, function(x)
    {
      header <- rlas::read.lasheader(x)
      if (length(header) == 0)
      {
        warning(paste0("Corrupted file ", x, " is not readable and was skipped in the LAScatalog"))
        return(NULL)
      }

      header        <- LASheader(header)
      PHB           <- header@PHB
      names(PHB)    <- phblab

      if (use_wktcs(header))
        PHB[["CRS"]] <- wkt(header)
      else
        PHB[["CRS"]] <- epsg(header)

      # Compatibility with rlas 1.3.0
      if (!is.null( PHB[["Number.of.points.by.return"]]))
      {
        PHB[["Number.of.1st.return"]] <- PHB[["Number.of.points.by.return"]][1]
        PHB[["Number.of.2nd.return"]] <- PHB[["Number.of.points.by.return"]][2]
        PHB[["Number.of.3rd.return"]] <- PHB[["Number.of.points.by.return"]][3]
        PHB[["Number.of.4th.return"]] <- PHB[["Number.of.points.by.return"]][4]
        PHB[["Number.of.5th.return"]] <- PHB[["Number.of.points.by.return"]][5]
        PHB[["Number.of.points.by.return"]] <- NULL
        PHB[["Global.Encoding"]] <- NULL
      }

      if (progress && Sys.time() - t0 > getOption("lidR.progress.delay")) {
        if (is.null(pb))
          pb <<- utils::txtProgressBar(min = 0, max = length(files), initial = i, style = 3)

        utils::setTxtProgressBar(pb, i)
      }

      i <<- i + 1

      return(PHB)
    })
  }

  rm = sapply(headers, is.null)
  files = files[!rm]
  headers <- data.table::rbindlist(headers)
  headers$filename <- files
  data.table::setDF(headers)

  xmin <- headers$Min.X
  xmax <- headers$Max.X
  ymin <- headers$Min.Y
  ymax <- headers$Max.Y
  zmin <- headers$Min.Z
  zmax <- headers$Max.Z
  ids  <- as.character(seq_along(files))

  geom <- lapply(seq_along(ids), function(xi) {
    mtx <- matrix(c(xmin[xi], xmax[xi], ymin[xi], ymax[xi])[c(1, 1, 2, 2, 1, 3, 4, 4, 3, 3)], ncol = 2)
    sf::st_polygon(list(mtx))
  })

  geom <-sf::st_sfc(geom)
  sf::st_crs(geom) <- crs
  headers <- sf::st_set_geometry(headers, geom)

  res <- new("LAScatalog")
  res@data <- headers

  opt_filter(res) <- filter
  opt_select(res) <- select
  opt_chunk_size(res) <- chunk_size
  opt_chunk_buffer(res) <- chunk_buffer
  opt_progress(res) <- progress

  xrange = xmax - xmin
  yrange = ymax - ymin
  zrange = zmax - zmin
  area = sum(xrange*yrange)
  if (area > 0)
  {
    n = sum(res$Number.of.point.records)
    density = n/area
    zratio = min(zrange/xrange, zrange/yrange)
  }
  else
  {
    zratio = 0
    density = 0
  }


  if (is.na(zratio)) # VPC file
    res@index <- LIDRALSINDEX
  else if (zratio < 10/100)
    res@index <- LIDRALSINDEX
  else if ((zratio >= 10/100 & density > 100) || density > 1000)
    res@index <- LIDRTLSINDEX
  else
    res@index <- LIDRALSINDEX



  if (is.overlapping(res))
    message("Be careful, some tiles seem to overlap each other. lidR may return incorrect outputs with edge artifacts when processing this catalog.")

  # if (!is.indexed(res))
  # message("las or laz files are not associated with lax files. This is not mandatory but may greatly speed up some computations. See help('writelax', 'rlas').")

  return(res)
}

#' @export
#' @rdname readLAScatalog
readALScatalog = function(folder, ...)
{
  ctg <- readLAScatalog(folder, ...)
  ctg@index <- LIDRALSINDEX
  return(ctg)
}


#' @export
#' @rdname readLAScatalog
readTLScatalog = function(folder, ...)
{
  ctg <- readLAScatalog(folder, ...)
  ctg@index <- LIDRTLSINDEX
  return(ctg)
}

#' @export
#' @name readLAScatalog
catalog <- function(folder, ...) { readLAScatalog(folder, FALSE, ...) }

read_vpc <- function(f)
{
  vpc = rjson::fromJSON(file = f)
  features = vpc$features
  headers = vector("list", length(features))
  for (i in seq_along(features))
  {
    feature = features[[i]]
    if (length(feature$properties[["proj:bbox"]]) == 6)
    {
      headers[[i]]$Min.X = feature$properties[["proj:bbox"]][1]
      headers[[i]]$Min.Y = feature$properties[["proj:bbox"]][2]
      headers[[i]]$Min.Z = feature$properties[["proj:bbox"]][3]
      headers[[i]]$Max.X = feature$properties[["proj:bbox"]][4]
      headers[[i]]$Max.Y = feature$properties[["proj:bbox"]][5]
      headers[[i]]$Max.Z = feature$properties[["proj:bbox"]][6]
    }
    else if (length(feature$properties[["proj:bbox"]]) == 4)
    {
      headers[[i]]$Min.X = feature$properties[["proj:bbox"]][1]
      headers[[i]]$Min.Y = feature$properties[["proj:bbox"]][2]
      headers[[i]]$Max.X = feature$properties[["proj:bbox"]][3]
      headers[[i]]$Max.Y = feature$properties[["proj:bbox"]][4]
    }
    else
    {
      stop("Invalid VPC file with no valid bbox")
    }

    headers[[i]]$Number.of.point.records = feature$properties[["pc:count"]]

    path = feature$assets$data$href[1]

    if (!is_absolute(path)) #771
    {
      relative_path <- path
      parent = dirname(f)
      absolute_path <- file.path(parent, relative_path)
      absolute_path <- normalizePath(absolute_path)
    }
    else
    {
      absolute_path = path
    }

    headers[[i]]$Min.Z = NA_real_
    headers[[i]]$Max.Z = NA_real_
    headers[[i]]$filename = absolute_path

    wkt = feature$properties[["proj:wkt2"]]
    epsg = feature$properties[["proj:epsg"]]

    if (!is.null(wkt))
      headers[[i]]$CRS = wkt
    else if (!is.null(epsg))
      headers[[i]]$CRS = epsg
    else
      headers[[i]]$CRS = 0
  }

  return (headers)
}

is_absolute <- function(path)
{
  if (.Platform$OS.type == "windows") {
    grepl("^(?:[A-Za-z]:[\\\\/]|\\\\\\\\)", path)  # Absolute paths start with "C:\" or "\\" on Windows
  } else {
    substr(path, 1, 1) == "/"  # Absolute paths start with "/" on Unix-like systems
  }
}

#' @export
#' @rdname rasterize
rasterize_terrain = function(las, res = 1, algorithm = tin(), use_class = c(2L,9L), shape = "convex", ...)
{
  UseMethod("rasterize_terrain", las)
}

#' @export
rasterize_terrain.LAS = function(las, res = 1, algorithm = tin(), use_class = c(2L,9L), shape = "convex", ...)
{
  # NOTE: the complexity of the code comes from the fact that res can be a number, a RasterLayer
  # a stars or a SpatRaster, and the function must return a corresponding raster. This implies
  # to use some raster agnostic wrappers

  # Defensive programming
  if (is_a_number(res))
    assert_all_are_non_negative(res)
  else if (!raster_is_supported(res))
    stop("'res' must be a number or a raster.", call. = FALSE)


  assert_is_algorithm(algorithm)
  assert_is_algorithm_spi(algorithm)

  if (is.character(shape))
    shape <- match.arg(shape, c("convex", "concave", "bbox"))
  else if (!is(shape, "sfc_POLYGON") & !is(shape, "sfc_MULTIPOLYGON"))
    stop("Argument 'shape' must be a string or a sfc", call. = FALSE)

  if (!"Classification" %in% names(las)) stop("LAS object does not contain 'Classification' attribute", call. = FALSE)
  if (any(as.integer(use_class) != use_class)) stop("'use_class' is not a vector of integers'", call. = FALSE)
  use_class <- as.integer(use_class)

  # Ellipsis parsing
  dots <- list(...)
  Wdegenerated <- isTRUE(dots$Wdegenerated)
  keep_lowest <- isTRUE(dots$keep_lowest)
  force_crs <- isTRUE(dots$force_crs)
  pkg <- if (is.null(dots$pkg)) getOption("lidR.raster.default") else dots$pkg

  # Non standard evaluation (R CMD check)
  . <- Z <- Zref <- X <- Y <- Classification <- NULL

  # Select the ground points
  ground <- las@data[Classification %in% c(use_class), .(X,Y,Z)]
  if (nrow(ground) == 0) stop("No ground points found. Impossible to compute a DTM.", call. = FALSE)
  ground <- check_degenerated_points(ground, Wdegenerated)

  # Find where to interpolate the DTM. Generate the XY where we want a DTM
  # It creates a stars object no matter the input which might be a number
  # a Raster, a stars or a SpatRaster
  layout <- if (!is_a_number(res)) raster_template(res) else raster_layout(las, res, format = "template")
  layout <- raster_materialize(layout, values = 0, pkg = "terra")

  if (inherits(layout, "Raster")) sf::st_crs(layout) <- st_crs(las) # Fix #517 because if res is a RasterLayer the crs is proj4 and it mess up everything

  if (is.character(shape) && shape %in% c("convex", "concave"))
  {
    if (shape == "concave")
      hull <- st_concave_hull(las, concavity = 10, length_threshold = 100)
    else
      hull <- st_convex_hull(las)

    shape <- sf::st_buffer(hull, dist = raster_res(layout)[1]/2)
  }

  if (is(shape, "sfc"))
  {
    shape <- terra::vect(shape)
    layout <- terra::mask(layout, shape)
  }

  grid <- raster_as_dataframe(layout, xy = TRUE, na.rm = TRUE)

  # Interpolate the terrain providing what to interpolate (ground) and where
  # to interpolate (grid)
  lidR.context <- "rasterize_terrain"
  ground <- LAS(ground, las@header, crs = st_crs(las), check = FALSE, index = las@index)
  Zg <- algorithm(ground, grid)
  Zg[is.nan(Zg)] <- NA_real_

  # If it remains NAs it means that we have points very far from ground points
  # and they cannot be interpolated. But we will interpolate them anyway. Previously
  # the function stopped. It now forces interpolation with NN.
  isna <- is.na(Zg)
  nnas <- sum(isna)
  if (nnas > 0)
  {
    nn <- knnidw(1, rmax = .Machine$double.xmax)
    sub_grid <- data.frame(X = grid$X[isna],  Y = grid$Y[isna])
    znn <- nn(ground, sub_grid)
    Zg[isna] <- znn
    warning(glue::glue("Interpolation of {nnas} points failed because they are too far from ground points. Nearest neighbour was used but interpolation is weak for those points"), call. = FALSE)
  }

  # Quantize the elevation for not returning absurdly accurate elevation
  fast_quantization(Zg, las[["Z scale factor"]], las[["Z offset"]])

  # Assignment of the values to a raster. It might be a raster from raster/stars/terra
  # This part of the code is raster agnostic. It returns a raster from the default format
  # registered in lidR or from pkg argument or the format given in 'res'
  cells  <- get_group(layout, grid)
  layout <- raster_layout(las, res)
  layout <- raster_materialize(layout, pkg = pkg)
  layout <- raster_set_values(layout, Zg, cells = cells)

  # Replace the interpolated value by the lowest point (legacy code) for option
  # keep_lowest that has been removed
  if (keep_lowest)
  {
    res    <- raster_res(layout)[1]
    lasg   <- filter_poi(las, Classification %in% c(use_class))
    rmin   <- template_metrics(lasg, ~list(Z = min(Z)), layout)
    Z1     <- raster_values(layout)
    Z2     <- raster_values(rmin)
    Zg     <- pmin(Z1, Z2, na.rm = TRUE)
    layout <- raster_set_values(layout, Zg)
  }

  raster_names(layout) <- "Z"
  return(layout)
}

#' @export
rasterize_terrain.LAScluster = function(las, res = 1, algorithm = tin(), use_class = c(2L,9L), shape = "convex", ...)
{
  x <- readLAS(las)
  if (is.empty(x)) return(NULL)

  # crop the raster to the extent of the chunk because the raster can be a very large proxy
  # that will be materialized by rasterize_terrain LAS
  if (is_raster(res)) res <- raster_crop(res, st_adjust_bbox(x, raster_res(res)))

  dtm <- rasterize_terrain(x, res, algorithm, use_class = use_class, shape = shape, ...)
  dtm <- raster_crop(dtm, st_bbox(las))
  return(dtm)
}

#' @export
rasterize_terrain.LAScatalog = function(las, res = 1, algorithm = tin(), use_class = c(2L,9L), shape = "convex", ...)
{
  assert_is_algorithm(algorithm)
  assert_is_algorithm_spi(algorithm)

  if (is_a_number(res))
    assert_all_are_non_negative(res)
  else if (!raster_is_supported(res))
    stop("'res' must be a number or a raster.", call. = FALSE)

  # subset the collection to the size of the layout (if any)
  if (!is_a_number(res)) las <- catalog_intersect(las, res)

  # Workaround for #690 (copied from #580). If rasterize_terrain is ran in parallel it will fail with
  # SpatRaster because they are not serializable. SpatRaster are converted to RasterLayer
  # for multicore strategies
  if (is_raster(res) && raster_pkg(res) == "terra")
  {
    ncores <- try_to_get_num_future_cores()
    if (!is.null(ncores) && ncores >= 2L)
      res <- raster::raster(res)
  }

  # Enforce some options
  opt_select(las) <- "xyzc"
  opt_filter(las) <-  paste("-keep_class", paste(use_class, collapse = " "), opt_filter(las))

  # Compute the alignment options including the case when res is a raster/stars/terra
  alignment <- raster_alignment(res)

  if (opt_chunk_size(las) > 0 && opt_chunk_size(las) < 2*alignment$res)
    stop("The chunk size is too small. Process aborted.", call. = FALSE)

  if (is(shape, "sfc"))
    las = catalog_intersect(las, shape, subset = "flag_processed")

  # Processing
  options <- list(need_buffer = TRUE, drop_null = TRUE, raster_alignment = alignment, automerge = TRUE)
  output  <- catalog_apply(las, rasterize_terrain, res = res, algorithm = algorithm, shape = shape, use_class = use_class, ..., .options = options)
  return(output)
}

#' Connected-Component Labeling
#'
#' Assigns an ID to each cluster of connected components in a point cloud. The point cloud is subdivided
#' into a 3D grid, and a classical Connected-Component Labeling algorithm is applied in 3D.
#'
#' @param las A LAS object representing the point cloud data.
#' @param res Grid resolution. If two non-empty voxels are contiguous, they are considered part of
#' the same component.
#' @param min_pts Minimum number of points in a cluster. If a cluster contains fewer than `min_pts`
#' points, it is assigned an ID of 0, indicating that it does not belong to any group.
#' @param name A string specifying the name of the new attribute used to store the ID of each cluster.
#' @param connectivity integer. 6, 18 or 26 neighbors search.
#'
#' @return A LAS object with an additional attribute named as specified by `name`.
#' @md
#' @export
connected_components = function(las, res, min_pts, name = "clusterID", connectivity = 6)
{
  .N <- N <- clusterID <- gpstime <- NULL

  allowed <- c(6L, 18L, 26L)
  if (!connectivity %in% allowed) stop(sprintf("Invalid connectivity: %d. Allowed values are %s", connectivity, paste(allowed, collapse = ", ")))

  u = C_connected_component(las, res, connectivity)
  las = add_lasattribute(las, u, name, "connected component ID")
  grp = las@data[, .N, by = clusterID]
  grp = grp[N < min_pts]
  invalid = las@data[[name]] %in% grp$clusterID
  las@data[[name]][invalid] = 0L
  return(las)
}



#' Computes the Distance to k-Nearest Neighbors
#'
#' Computes the average distance between each point and its k-nearest neighbors in a point cloud.
#' The results are stored in a new attribute.
#'
#' @param las A LAS object representing the point cloud data.
#' @param k The number of nearest neighbors.
#' @param name A string specifying the name of the new attribute used to store the computed distances.
#'
#' @return A LAS object with an additional attribute named as specified by `name`.
#' @export
#' @examples
#' LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
#' las <- readLAS(LASfile, select = "xyz", filter = "-inside 481250 3812980 481300 3813030")
#' las = knn_distance(las)
#' #plot(las, color = "distance", breaks = "quantile", legend = TRUE)
knn_distance = function(las, k = 10, name = "distance")
{
  d = C_knn_distance(las, k, getThreads())
  las = add_lasattribute(las, d, name, "knn average distance")
  return(las)
}

#' Compression of the point cloud
#'
#' Package rlas 1.6.0 supports compact representation of non populated attributes. For example `UserData`
#' is usually populated with zeros (not populated). Yet it takes 32 bits per point to store each 0.
#' With rlas 1.6.0 it can now use 644 bits no matter the number of points loaded if it is not
#' populated or populated with a unique value.
#'
#' `las_is_compressed` test each attributes and returns a named vector with TRUE if the attribute is
#' compressed FALSE otherwise.\cr\cr
#' `las_size` returns the true size of a LAS object by considering the compression. `object.size` from
#' base R does not account for ALTREP and consequently cannot measure properly the size of a LAS object
#'
#' @param las A LAS object.
#'
#' @examples
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' las = readLAS(LASfile)
#' las_is_compressed(las)
#'
#' format(object.size(las), units = "MB")
#' format(las_size(las), units = "MB")
#'
#' @name las_compression
#' @rdname las_compression
#' @md
NULL

#' @rdname las_compression
#' @export
las_is_compressed <- function(las)
{
  stopifnotlas(las)
  return(sapply(payload(las), is_compact))
}

#' @rdname las_compression
#' @export
las_size <- function(las)
{
  stopifnotlas(las)
  size <- 0
  gz <- las_is_compressed(las)

  for (i in seq_along(gz))
  {
    if (isTRUE(gz[i]))
      size <- size + 644L
    else
      size <- size + utils::object.size(payload(las)[[i]])
  }

  size + utils::object.size(header(las)) + utils::object.size(st_crs(las))  + utils::object.size(las@index)

  class(size) <- 'object_size'
  return(size)
}

is_compact <- function(x)
{
  altrep <- altrep_full_class(x)
  if (is.null(altrep)) return(FALSE)
  if (altrep[[2]] == "rlas") {
    if (is_materialized(x)) return(FALSE)
    return(TRUE)
  }
  return(FALSE)
}

#' Normalize point cloud
#'
#' Normalize elevation or intensity values using multiple methods.
#'
#' \describe{
#' \item{normalize_height}{Subtract digital terrain model (DTM) from a LiDAR point cloud to create a
#' dataset normalized with the ground at 0. The DTM can be a raster, but it can also be computed
#' on-the-fly. In this case the algorithm does not use rasterized data and each point is interpolated.
#' There is no inaccuracy due to the discretization of the terrain and the resolution of the terrain
#' is virtually infinite. A new attribute 'Zref' records the former elevation values, which enables
#' the use of \link{unnormalize_height} to restore original point elevations.}
#' \item{normalize_intensity}{Normalize intensity values using multiple methods. The attribute 'Intensity'
#' records the normalized intensity. An extra attribute named 'RawIntensity' records the original
#' intensities.}
#' \item{height_above_ground}{instead of normalizing the point cloud, records the height above ground (HAG)
#' in a new attribute named 'hag'}
#' }
#'
#' @section Non-supported LAScatalog options:
#' The option `select` is not supported and not respected because it always preserves the file format
#' and all the attributes. `select = "*"` is imposed internally.
#'
#' @template param-las
#' @param algorithm (1) An algorithm for spatial interpolation. \code{lidR} has \link{tin},
#' \link{kriging}, \link{knnidw} or a raster representing a digital terrain
#' model. (2) An algorithm for intensity normalization. \code{lidR} currently has \link{range_correction}.
#' @param use_class integer vector. By default the terrain is computed by using ground points
#' (class 2) and water points (class 9). Relevant only for a normalization without a raster DTM.
#' @param ... passed to `normalized_height()`
#' @param dtm raster. If `dtm` is provided, then the DTM is used in place of ground points. This is
#' different than providing a DTM in `algorithm`. If `algorithm = dtm` the dtm is subtracted naively.
#' If `algorithm = tin()` and `dtm = raster` the ground points are not used and the DTM is
#' interpolated as if it were made of regularly-spaced ground points.
#'
#' @name normalize
#' @rdname normalize
#' @md
#' @examples
#' LASfile <- system.file("extdata", "Topography.laz", package="lidR")
#' las <- readLAS(LASfile)
#'
#' # ====================
#' # Normalize elevation
#' # ====================
#'
#' # First option: use a raster as DTM
#' # --------------------------------------
#'
#' dtm <- rasterize_terrain(las, 1, knnidw(k = 6L, p = 2))
#' nlas <- normalize_height(las, dtm)
#'
#' # restore original elevations
#' las <- unnormalize_height(nlas)
#'
#' # operator - can be used. This is equivalent to the previous
#' nlas <- las - dtm
#'
#' # restore original elevations
#' las <- unnormalize_height(las)
#'
#' # Second option: interpolate each point (no discretization)
#' # ---------------------------------------------------------
#'
#' nlas <- normalize_height(las, tin())
#'
#' # operator - can be used. This is equivalent to the previous
#' las <- unnormalize_height(nlas)
#' nlas <- las - tin()
#'
#' \dontrun{
#' # All the following syntaxes are correct
#' las <- normalize_height(las, knnidw())
#' las <- normalize_height(las, knnidw(k = 8, p = 2))
#' las <- las - knnidw()
#' las <- las - knnidw(k = 8)
#' las <- normalize_height(las, kriging())
#' las <- las - kriging(k = 8)
#' }
#'
#' # ====================
#' # Normalize intensity
#' # ====================
#'
#' # pmin = 15 because it is an extremely small file
#' # strongly decimated to reduce its size. There are
#' # actually few multiple returns
#' sensor <- track_sensor(las, Roussel2020(pmin = 15))
#'
#' # Here the effect is virtually null because the size of
#' # the sample is too small to notice any effect of range
#' las <- normalize_intensity(las, range_correction(sensor, Rs = 2000))
NULL

#' Coordinates of a LAS* object in a matrix form
#'
#' Retrieve coordinates of a `LAS*` object in matrix form. It creates a copy of the coordinates
#' because of the coercion from `data.frame` to `matrix`. This function inherits `sf::st_coordinates`
#'
#' @param x A LAS* object
#' @param z bool. Return XY or XYZ matrix
#' @param ... unused.
#'
#' @return matrix
#'
#' @name st_coordinates
#' @rdname st_coordinates
#' @importFrom sf st_coordinates
#' @md
#' @examples
#' LASfile <- system.file("extdata", "example.laz", package="rlas")
#' las <- readLAS(LASfile)
#' sf::st_coordinates(las)
NULL

#' @rdname st_coordinates
#' @export
st_coordinates.LAS <- function(x, z = TRUE, ...)
{
  if (isTRUE(z))
    return(as.matrix(coordinates3D(x)))
  else
    return(as.matrix(coordinates(x)))
}

#' @rdname st_coordinates
#' @export
st_coordinates.LAScatalog <- function(x, ...)
{
  return(sf::st_coordinates(x@data))
}

coordinates = function(las)
{
  X  <- las[["X"]]
  Y  <- las[["Y"]]
  DF <- data.frame(X,Y)
  data.table::setDT(DF)
  return(DF)
}

coordinates3D = function(las)
{
  X  <- las[["X"]]
  Y  <- las[["Y"]]
  Z  <- las[["Z"]]
  DF <- data.frame(X,Y,Z)
  data.table::setDT(DF)
  return(DF)
}

#' Older R Spatial Packages
#'
#' lidR 4.0.0 no longer uses the `sp` and `raster` packages. New functions are based on `sf`, `terra` and `stars`.
#' However, to maintain backward compatibility the old functions from v<4.0.0 were preserved.\cr\cr
#' `rgdal` and `rgeos` will be retired on Jan 1st 2024. The `raster` and `sp` packages are based on
#' `rgdal` and `rgeos`. `lidR` was based on `raster` and `sp` because it was created before the `sf`, `terra`
#' and `stars` packages. This means that sooner or later users and packages that are still based on
#' old R spatial packages will run into trouble. According to Edzer Pebesma, Roger Bivand: \cr\cr
#' *R users who have been around a bit longer, in particular before packages like `sf` and `stars` were
#' developed, may be more familiar with older packages like `maptools`, `sp`, `rgeos`, and `rgdal`. A fair
#' question is whether they should migrate existing code and/or existing R packages depending on these
#' packages. The answer is: yes (see reference).*\cr\cr
#' The following functions are not formally deprecated but users should definitely move their workflow to modern
#' spatial packages. lidR will maintain the old functions as long as it does not generate issues
#' on CRAN. So, it might be until Jan 1st 2024 or later, who knows...
#'
#' @param x,las an object of class LAS*
#' @param res,start see \link{pixel_metrics}
#' @param func see \link{template_metrics}
#' @param attribute,type see \link{crown_metrics}
#' @param ... ignored
#' @param algorithm see \link{rasterize_canopy}, \link{rasterize_terrain}
#' @param full_raster,use_class,Wdegenerated,is_concave,keep_lowest see \link{rasterize_density}
#' @param filter,by_echo see \link{template_metrics}
#' @param uniqueness see \link{crown_metrics}
#' @param concavity,length_threshold see \link{concaveman}
#' @references Edzer Pebesma, Roger Bivand Spatial Data Science with applications in R
#' https://keen-swartz-3146c4.netlify.app/older.html
#' @md
#' @name old_spatial_packages
#' @rdname old_spatial_packages
NULL

#' @export
#' @rdname old_spatial_packages
as.spatial = function(x) UseMethod("as.spatial", x)

#' @export
#' @rdname old_spatial_packages
as.spatial.LAS = function(x) { sp::SpatialPointsDataFrame(x@data[, 1:2], x@data[, 3:ncol(x@data)], proj4string = as(st_crs(x), "CRS")) }

#' @export
#' @rdname old_spatial_packages
as.spatial.LAScatalog = function(x)
{
  # Workaround to repair LAScatalog v3 and minimize backward incompatibilities with v4
  # nocov start
  if (is_lascatalog_v3(x))
  {
    res <- new("SpatialPolygonsDataFrame")
    res@bbox <- x@bbox
    res@proj4string <- x@proj4string
    res@plotOrder <- x@plotOrder
    res@data <- x@data
    res@polygons <- x@polygons
    return(res)
  }
  # nocov end

  return(sf::as_Spatial(x@data))
}

#' @export
#' @rdname old_spatial_packages
tree_metrics <- function(las, func = ~list(Z = max(Z)), attribute = "treeID", ...)
{
  res <- crown_metrics(las, func, geom = "point", attribute, ...)
  if (is(res, "sf")) res <- sf::as_Spatial(res)
  return(res)
}

#' @export
#' @rdname old_spatial_packages
grid_canopy = function(las, res, algorithm)
{
  rasterize_canopy(las, res, algorithm, pkg = "raster")
}

#' @export
#' @rdname old_spatial_packages
grid_density = function(las, res = 4)
{
  rasterize_density(las, res, pkg = "raster")
}

#' @export
#' @rdname old_spatial_packages
grid_terrain = function(las, res = 1, algorithm, ..., keep_lowest = FALSE, full_raster = FALSE, use_class = c(2L,9L), Wdegenerated = TRUE, is_concave = FALSE)
{
  shape <- "convex"
  if (full_raster) shape <- "bbox"
  if (is_concave) shape <- "concave"

  rasterize_terrain(las, res, algorithm, shape = shape, use_class = use_class, keep_lowest = keep_lowest, Wdegenerated = Wdegenerated, pkg = "raster")
}

#' @export
#' @rdname old_spatial_packages
grid_metrics = function(las, func, res = 20, start = c(0,0), filter = NULL, by_echo = "all")
{
  res <- pixel_metrics(las, func, res, start, filter = filter, by_echo = by_echo, pkg = "raster")
  return(res)
}

#' @export
#' @rdname old_spatial_packages
find_trees = function(las, algorithm, uniqueness = 'incremental')
{
  res <- locate_trees(las, algorithm, uniqueness)
  if (is(res, "sf"))
  {
    if (nrow(res) == 0L)
    {
      coords <- matrix(0, ncol = 2)
      data   <- data.frame(treeID = integer(1), Z = numeric(1))
      res    <- sp::SpatialPointsDataFrame(coords, data, proj4string = as(st_crs(las), "CRS"))
      res    <- res[0,]
    }
    else
    {
      res <- sf::st_zm(res)
      res <- sf::as_Spatial(res)
    }
  }

  return(res)
}

#' @export
#' @rdname old_spatial_packages
delineate_crowns = function(las, type = c("convex", "concave", "bbox"), concavity = 3, length_threshold = 0, func = NULL, attribute = "treeID")
{
  type <- match.arg(type)
  res <- crown_metrics(las, func = func, geom = type, concaveman = c(concavity, length_threshold), attribute = attribute, xyz = TRUE)

  invalid <- !sf::st_is_valid(res)
  ninvalid <- sum(invalid)
  if (ninvalid > 0)
  {
    res <- res[!invalid,]
    warning(glue::glue("{ninvalid} geometries were discarded before to convert from sf to sp. To keep them use crown_metrics()"), call. = FALSE)
  }

  if (is(res, "sf")) res <- sf::as_Spatial(res)
  res
}


is_las_v3 <- function(x) { !methods::.hasSlot(x, "crs") } # nocov
las_v3_repair <- function(x) { if (is_las_v3(x)) return(LAS(x)) else return(x) } # nocov

is_lascatalog_v3 <- function(x) { !is(x@data, "sf") }
lascatalog_v3_repair <- function(x)
{
  if (is_lascatalog_v3(x))
  {
    y <- new("LAScatalog")
    y@data <- sf::st_as_sf(as.spatial(x))
    opt_copy(y) <- x
    return(y)
  }

  return(x)
}

# nocov start
rOverlay = function(las, res, start = c(0,0), buffer = 0)
{
  if (is(res, "RasterLayer"))
  {
    resolution <- raster::res(res)
    if (round(resolution[1], 4) != round(resolution[2], 4))
      stop("Rasters with different x y resolutions are not supported", call. = FALSE)

    return(raster::raster(res))
  }

  bbox      <- raster::extent(las) + 2 * buffer
  bbox@xmin <- round_any(bbox@xmin - 0.5 * res - start[1], res) + start[1]
  bbox@xmax <- round_any(bbox@xmax - 0.5 * res - start[1], res) + res + start[1]
  bbox@ymin <- round_any(bbox@ymin - 0.5 * res - start[2], res) + start[2]
  bbox@ymax <- round_any(bbox@ymax - 0.5 * res - start[2], res) + res + start[2]
  layout    <- suppressWarnings(raster::raster(bbox, res = res, crs = crs(las)))
  layout@data@values <- rep(NA, raster::ncell(layout))
  raster::crs(layout) <- raster::crs(las)
  return(layout)
}
# nocov end

#' Add attributes into a LAS object
#'
#' A \link[=LAS-class]{LAS} object represents a las file in R. According to the
#' \href{https://community.asprs.org/leadership-restricted/leadership-content/public-documents/standards}{LAS specifications}
#' a las file contains a core of defined attributes, such as XYZ coordinates, intensity, return number,
#' and so on, for each point. It is possible to add supplementary attributes.
#'
#' Users cannot assign names that are the same as the names of the core attributes. These functions are
#' dedicated to adding data that are not part of the LAS specification. For example, \code{add_lasattribute(las, x, "R")}
#' will fail because \code{R} is a name reserved for the red channel of a .las file that contains RGB
#' attributes. Use \code{add_lasrgb} instead.
#' \describe{
#' \item{\code{add_attribute}}{Simply adds a new column in the data but does not update the header. Thus the LAS
#' object is not strictly valid. These data will be temporarily usable at the R level but will not
#' be written in a las file with \link{writeLAS}.}
#'
#'\item{ \code{add_lasattribute}}{Does the same as \code{add_attribute} but automatically updates the header of
#' the LAS object. Thus, the LAS object is valid and the new data is considered as "extra bytes". This new
#' data will be written in a las file with \link{writeLAS}.}
#'
#' \item{\code{add_lasattribute_manual}}{Allows the user to manually write all the extra bytes metadata.
#' This function is reserved for experienced users with a good knowledge of the LAS specifications.
#' The function does not perform tests to check the validity of the information.
#' When using \code{add_lasattribute} and \code{add_lasattribute_manual}, \code{x} can only be of type numeric,
#' (\code{integer} or \code{double}). It cannot be of type \code{character} or \code{logical} as these are
#' not supported by the LAS specifications. The types that are supported in lidR are types 0 to 10
#' (Table 24 on page 25 of the specification). Types greater than 10 are not supported.}
#'
#' \item{\code{add_lasrgb}}{Adds 3 columns named RGB and updates the point format of the LAS object
#' for a format that supports RGB attributes. If the RGB values are ranging from 0 to 255 they are
#' automatically scaled on 16 bits.}
#' }
#'
#' @param las An object of class \link[=LAS-class]{LAS}
#' @param x a vector that needs to be added in the LAS object. For \code{add_lasattribute*} it can
#' be missing (see details).
#' @param name character. The name of the extra bytes attribute to add in the file.
#' @param desc character. A short description of the extra bytes attribute to add in the file (32 characters).
#' @param type character. The data type of the extra bytes attribute. Can be "uchar", "char", "ushort",
#' "short", "uint", "int", "uint64", "int64", "float", "double".
#' @param scale,offset numeric. The scale and offset of the data. NULL if not relevant.
#' @param NA_value numeric or integer. NA is not a valid value in a las file. At time of writing it will
#' be replaced by this value that will be considered as NA. NULL if not relevant.
#' @param R,G,B,NIR integer. RGB and NIR values. Values are automatically scaled to 16 bits if they are
#' coded on 8 bits (0 to 255).
#'
#' @return An object of class \link[=LAS-class]{LAS}
#'
#' @name add_attribute
#'
#' @examples
#' LASfile <- system.file("extdata", "example.laz", package="rlas")
#' las <- readLAS(LASfile, select = "xyz")
#'
#' print(las)
#' print(header(las))
#'
#' x <- 1:30
#'
#' las <- add_attribute(las, x, "mydata")
#' print(las)        # The las object has a new attribute called "mydata"
#' print(header(las)) # But the header has not been updated. This new data will not be written
#'
#' las <- add_lasattribute(las, x, "mydata2", "A new data")
#' print(las)        # The las object has a new attribute called "mydata2"
#' print(header(las)) # The header has been updated. This new data will be written
#'
#' # Optionally if the data is already in the LAS object you can update the header skipping the
#' # parameter x
#' las <- add_attribute(las, x, "newattr")
#' las <- add_lasattribute(las, name = "newattr", desc = "Amplitude")
#' print(header(las))
#'
#' # Remove an extra bytes attribute
#' las <- remove_lasattribute(las, "mydata2")
#' print(las)
#' print(header(las))
#'
#' las <- remove_lasattribute(las, "mydata")
#' print(las)
#' print(header(las))
NULL

#' @export
#' @rdname add_attribute
add_attribute = function(las, x, name)
{
  stopifnotlas(las)
  assert_is_a_string(name)
  stopif_forbidden_name(name)

  las@data[[name]] <- x
  return(las)
}

#' @export
#' @rdname add_attribute
add_lasattribute = function(las, x, name, desc)
{
  stopifnotlas(las)
  assert_is_a_string(name)
  assert_is_a_string(desc)
  stopif_forbidden_name(name)

  if (missing(x))
  {
    if (!name %in% names(las))
      stop(glue::glue("{name} is not an attribute of the LAS object."), call. = FALSE)

    x <- las@data[[name]]

    if (!is.numeric(x))
      stop(glue::glue("'{name}' must be numeric. LAS format specifications do not allow storing of '{class(las@data[[name]])}' extra bytes."), call. = FALSE)
  }
  else
  {
    assert_is_vector(x)

    if (!is.numeric(x))
      stop(glue::glue("'x' must be numeric. LAS format specifications do not allow storing of '{class(x)}' extra bytes."), call. = FALSE)

    las <- add_attribute(las, x, name)
  }

  header     <- as.list(las@header)
  new_header <- rlas::header_add_extrabytes(header, x, name, desc)
  new_header <- LASheader(new_header)
  las@header <- new_header
  return(las)
}

#' @export
#' @rdname add_attribute
add_lasattribute_manual = function(las, x, name, desc, type, offset = NULL, scale = NULL, NA_value = NULL)
{
  stopifnotlas(las)
  assert_is_a_string(name)
  assert_is_a_string(desc)
  assert_is_a_string(type)
  stopif_forbidden_name(name)

  types = c("uchar", "char", "ushort", "short", "uint", "int", "uint64", "int64", "float", "double")
  type = match.arg(type, types)
  type = which(type == types)

  if (missing(x))
  {
    if (!name %in% names(las))
      stop(glue::glue("{name} is not an attribute of the LAS object."), call. = FALSE)

    x <- las@data[[name]]

    if (!is.numeric(x))
      stop(glue::glue("'{name}' must be numeric. LAS format specifications do not allow storing of '{class(las@data[[name]])}' extra bytes."), call. = FALSE)
  }
  else
  {
    assert_is_vector(x)

    if (!is.numeric(x))
      stop(glue::glue("'x' must be numeric. LAS format specifications do not allow storing of '{class(x)}' extra bytes."), call. = FALSE)

    las <- add_attribute(las, x, name)
  }

  header     <- as.list(las@header)

  new_header <- rlas::header_add_extrabytes_manual(header, name, desc, type, offset, scale, NULL, NULL, NA_value)
  new_header <- LASheader(new_header)
  las@header <- new_header
  return(las)
}

#' @export
#' @rdname add_attribute
add_lasrgb <- function(las, R, G, B)
{
  stopifnotlas(las)
  stopifnot(is.integer(R), is.integer(G), is.integer(B))
  assert_are_same_length(R, G)
  assert_are_same_length(R, B)
  assert_is_of_length(R, npoints(las))

  maxr <- max(R, na.rm = TRUE)
  maxg <- max(G, na.rm = TRUE)
  maxb <- max(B, na.rm = TRUE)

  scale <- 1L
  if (maxr <= 255 & maxg <= 255 & maxb <= 255)
    scale <- 257L

  las@data$R <- R*scale
  las@data$G <- G*scale
  las@data$B <- B*scale

  format <- las@header@PHB[["Point Data Format ID"]]

  if (format %in% c(2,3,8))
  {
    # nothing to do
  }
  else if ("NIR" %in% names(las))
  {
    format <- 8L
  }
  else if ("gpstime" %in% names(las))
  {
    format <- 3L
  }
  else
  {
    format <- 2L
  }

  las@header@PHB[["Point Data Format ID"]] <- format
  return(las)
}

#' @export
#' @rdname add_attribute
add_lasnir <- function(las, NIR)
{
  stopifnotlas(las)
  stopifnot(is.integer(NIR))
  assert_is_of_length(NIR, npoints(las))

  maxnir <- max(NIR, na.rm = TRUE)

  scale <- 1L
  if (maxnir <= 255)
    scale <- 257L

  las@data$NIR <- NIR*scale
  las@header@PHB[["Point Data Format ID"]] <- 8L
  return(las)
}

#' @export
#' @rdname add_attribute
remove_lasattribute = function(las, name)
{
  stopifnotlas(las)
  assert_is_a_string(name)
  stopif_forbidden_name(name)

  if (!name %in% names(las))
  {
     message(glue::glue("{name} is not an attribute of the LAS object."))
     return(las)
  }

  eb <- las@header@VLR$Extra_Bytes$`Extra Bytes Description`[[name]]

  if (is.null(eb))
  {
    #message(glue::glue("{name} is not an extrabytes attribute of the LAS object."))
    las@data[[name]] <- NULL
    return(las)
  }
  else
  {
    las@header@VLR$Extra_Bytes$`Extra Bytes Description`[[name]] <- NULL
    las@data[[name]] <- NULL

    if (length(las@header@VLR$Extra_Bytes$`Extra Bytes Description`) == 0)
      las@header@VLR$Extra_Bytes <- NULL

    return(las)
  }
}

# type = 0 : undocumented
# type = 1 : unsigned char
# type = 2 : char
# type = 3 : unsigned short
# type = 4 : short
# type = 5 : unsigned int
# type = 6 : int
# type = 7 : unsigned int64
# type = 8 : int64
# type = 9 : float  (try not to use)
# type = 10 : double (try not to use)

context("add_attribute")

las     <- example
las@data <- lidR:::coordinates3D(las)
x = runif(30)
k = sample(1:100, 30)
n = npoints(las)
R = sample(1:255, n, TRUE)
G = sample(1:255, n, TRUE)
B = sample(1:255, n, TRUE)

test_that("add_attribute works", {
   las2 = add_attribute(las, x, "plop")
   n = names(las2@data)

   expect_true("plop" %in% n)
   expect_equal(las@header, las2@header)
})

test_that("add_lasattribute works", {
  las2 = add_lasattribute(las, x, "plop", "extra attr")
  n = names(las2@data)

  expect_true("plop" %in% n)
  expect_true(!is.null(las2@header@VLR$Extra_Bytes))
  expect_true(!is.null(las2@header@VLR$Extra_Bytes$`Extra Bytes Description`$plop))
})

test_that("add_lasattribute manual works", {
  las2 = add_lasattribute_manual(las, x, "plop", "extra attr", "float")
  n = names(las2@data)

  expect_true("plop" %in% n)
  expect_true(!is.null(las2@header@VLR$Extra_Bytes))
  expect_true(!is.null(las2@header@VLR$Extra_Bytes$`Extra Bytes Description`$plop))
})

test_that("lasremoveextrabyes works", {
  las2 = add_lasattribute(las, x, "plop", "extra attr")
  las3 = remove_lasattribute(las2, "plop")

  expect_equal(las3@header, las@header)
})

test_that("add_lasrgb works", {
  las2 = add_lasrgb(las, R, G, B)

  expect_equal(las2@header@PHB$`Point Data Format ID`, 2L)
  expect_gt(mean(las2$R), 255)
})

test_that("header extrabyte is updated", {
  x = runif(n)

  las <- add_attribute(las, x, "x")
  expect_true("x" %in% names(las))
  expect_true(is.null(las@header@VLR$Extra_Bytes))


  las <- add_lasattribute(las, name = "x", desc = "test")
  expect_true("x" %in% names(las))
  expect_true(!is.null(las@header@VLR$Extra_Bytes))
  expect_equal(las@header@VLR$Extra_Bytes$`Extra Bytes Description`$x$data_type, 10L)

  las <- add_lasattribute_manual(las, name = "x", desc = "test", type = "int")
  expect_true("x" %in% names(las))
  expect_true(!is.null(las@header@VLR$Extra_Bytes))
  expect_equal(las@header@VLR$Extra_Bytes$`Extra Bytes Description`$x$data_type, 6L)
})

test_that("header extrabyte fails", {

  expect_error(add_lasattribute(las, name = "y", desc = "test"), "y is not an attribute of the LAS object.")
  expect_error(add_lasattribute_manual(las, name = "y", desc = "test", type = "uint"), "y is not an attribute of the LAS object.")

  x = logical(n)

  expect_error(add_lasattribute(las, x, name = "y", desc = "test"), "LAS format specifications do not allow storing of 'logical' extra bytes")
  expect_error(add_lasattribute_manual(las, x, name = "y", desc = "test", type = "uint"), "LAS format specifications do not allow storing of 'logical' extra bytes")
})

#' @param bboxes list. list of bbox
#' @param shape numeric LIDRRECTANGLE|LIDRCIRCLE
#' @param buffer numeric
#' @param process. logical
#' @param outside_catalog_is_null bool
#' @param by_file bool
#' @noRd
engine_index =	function(ctg, bboxes, shape = LIDRRECTANGLE, buffer = 0, process = TRUE, outside_catalog_is_null = TRUE, by_file = FALSE)
{
  stopifnot(is.list(bboxes))

  MinX <- ctg[["Min.X"]]
  MaxX <- ctg[["Max.X"]]
  MinY <- ctg[["Min.Y"]]
  MaxY <- ctg[["Max.Y"]]

  if (length(process) == 1L)
    process <- rep(process, nrow(ctg@data))

  queries <- vector("list", length(bboxes))

  for (i in 1:length(bboxes))
  {
    bbox  <- bboxes[[i]]
    bbbox <- st_expand_bbox(bbox, buffer)

    tile_is_in_bbox          <- !(MinX >= bbox$xmax  | MaxX <= bbox$xmin  | MinY >= bbox$ymax  | MaxY <= bbox$ymin)
    tile_is_in_buffered_bbox <- !(MinX >= bbbox$xmax | MaxX <= bbbox$xmin | MinY >= bbbox$ymax | MaxY <= bbbox$ymin)

    if (all(!process[tile_is_in_bbox]))
      select <- FALSE
    else if (sum(tile_is_in_bbox) > 0)
      select <- tile_is_in_buffered_bbox
    else
      select <- FALSE

    files  <- ctg$filename[select]

    if (length(files) == 0 & outside_catalog_is_null)
      next
    else if (length(files) == 0 & !outside_catalog_is_null)
      files <- ""

    # If one file is considered the main one put it in first place so it get the header precedence
    # when merged-reading the las files.
    if (length(files) > 1 && by_file)
    {
      main <- ctg$filename[process][i]
      j <- which(files == main)
      if (length(j) == 0) stop("Internal error: the indexation algorithm generated an incorrect list of files. Please report this error.") # nocov
      sfiles <- c(main, files[-j[1]])
      files <- sfiles
    }

    # If one file that encompasses the bbox is set to 'non processing' resize the chunk
    if (!by_file && any(!process[tile_is_in_bbox]))
    {
      k <- process & tile_is_in_bbox
      bbox$xmin <- max(min(MinX[k]), bbox$xmin)
      bbox$ymin <- max(min(MinY[k]), bbox$ymin)
      bbox$xmax <- min(max(MaxX[k]), bbox$xmax)
      bbox$ymax <- min(max(MaxY[k]), bbox$ymax)
      bbox      <- sf::st_bbox(unlist(bbox))
      bbbox     <- st_expand_bbox(bbox, buffer)
    }

    center  <- list(x = (bbox$xmax + bbox$xmin)/2, y = (bbox$ymax + bbox$ymin)/2)
    width   <- (bbox$xmax - bbox$xmin)
    height  <- (bbox$ymax - bbox$ymin)
    cluster <- LAScluster(center, width, height, buffer, shape, files, "noname", crs = st_crs(ctg), index = ctg@index)

    cluster@select <- opt_select(ctg)
    cluster@filter <- paste(cluster@filter, opt_filter(ctg))

    queries[[i]] <- cluster
  }

  return(queries)
}

#' Retile a LAScatalog
#'
#' Splits or merges files to reshape the original files collection (.las or .laz) into smaller or larger
#' files. It also enables the addition or removal of a buffer around the tiles. Internally, the
#' function reads and writes the chunks defined by the internal processing options of a
#' \link[=LAScatalog-class]{LAScatalog}. Thus, the function is flexible and enables the user to retile
#' the dataset, retile while adding or removing a buffer (negative buffers are allowed), or optionally
#' to compress the data by retiling without changing the pattern but by changing
#' the format (las/laz). **This function is does not load the point cloud into R memory** It streams
#' from input file(s) to output file(s) and can be applied to large point-cloud with low memory
#' computer.\cr\cr
#' Note that this function is not actually very useful because `lidR` manages everything
#' (clipping, processing, buffering, ...) internally using the proper options. Thus, retiling may be
#' useful for working in other software, for example, but not in `lidR`
#'
#' @section Non-supported LAScatalog options:
#' The option `select` is not supported and not respected because it always preserves the file
#' format and all the attributes. `select = "*"` is imposed internally.
#'
#' @param ctg A \link[=LAScatalog-class]{LAScatalog} object
#'
#' @return A new \code{LAScatalog} object
#'
#' @export
#' @md
#'
#' @examples
#' \donttest{
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' ctg = readLAScatalog(LASfile)
#' plot(ctg)
#'
#' # Create a new set of 200 x 200 m.las files with first returns only
#' opt_chunk_buffer(ctg) <- 0
#' opt_chunk_size(ctg) <- 200
#' opt_filter(ctg) <- "-keep_first"
#' opt_chunk_alignment(ctg) <- c(275, 90)
#' opt_output_files(ctg) <- paste0(tempdir(), "/retile_{XLEFT}_{YBOTTOM}")
#'
#' # preview the chunk pattern
#' plot(ctg, chunk = TRUE)
#'
#' newctg = catalog_retile(ctg)
#'
#' plot(newctg)
#'
#' # Create a new set of 200 x 200 m.las files
#' # but extended with a 50 m buffer in the folder
#'
#' opt_chunk_buffer(ctg) <- 25
#' opt_chunk_size(ctg) <- 200
#' opt_filter(ctg) <- ""
#' opt_chunk_alignment(ctg) <- c(275, 90)
#' opt_output_files(ctg) <- paste0(tempdir(), "/{XLEFT}_{YBOTTOM}_buffered")
#' newctg = catalog_retile(ctg)
#'
#' plot(newctg)
#' }
#'
#' \dontrun{
#' # Create a new set of compressed .laz file equivalent to the original, keeping only
#' # first returns above 2 m
#'
#' opt_chunk_buffer(ctg) <- 0
#' opt_chunk_size(ctg) <- 0
#' opt_laz_compression(ctg) <- TRUE
#' opt_filter(ctg) <- "-keep_first -drop_z_below 2"
#' opt_output_files(ctg) <- paste0(tempdir(), "/{ORIGINALFILENAME}_first_2m")
#' newctg = catalog_retile(ctg)
#' }
catalog_retile = function(ctg)
{
  laz  <- opt_laz_compression(ctg)
  path <- opt_output_files(ctg)
  path <- if (laz) paste0(path, ".laz") else paste0(path, ".las")
  ctg@output_options$output_files <- path
  path <- dirname(path)

  reshape_func = function(cluster)
  {
    streamLAS(cluster, cluster@save)

    if (file.exists(cluster@save))
    {
      ret <- structure(cluster@save, class = "lidr_internal_skip_write")
      return(ret)
    }
    else
    {
      return(NULL)
    }
  }

  opt_wall_to_wall(ctg) <- FALSE

  options <- list(need_buffer = FALSE, drop_null = TRUE, need_output_file = TRUE, automerge = TRUE)
  return(catalog_apply(ctg, reshape_func, .options = options))
}

generate_las <- function(n, seeds = 1) {
  set.seed(seeds)
  X <- round_any(stats::runif(n, 0, 100), 0.001)
  set.seed(seeds + 1)
  Y <- round_any(stats::runif(n, 0, 100), 0.001)
  set.seed(seeds + 2)
  Z <- round_any(c(stats::runif(0.8 * n, 0, 25), rep(0, 0.2 * n)), 0.001)
  Classification <- as.integer(c(rep(1, 0.8 * n), rep(2, 0.2 * n)))
  set.seed(seeds + 3)
  Intensity <- as.integer(stats::runif(n, 10, 50))
  ReturnNumber <- as.integer(rep(c(1, 1, 1, 2, 3, 1, 2, 1, 2, 1), n / 10))
  NumberOfReturns <- as.integer(rep(c(1, 1, 3, 3, 3, 2, 2, 2, 2, 1), n / 10))

  data <- data.table::data.table(X, Y, Z, Classification, Intensity, ReturnNumber, NumberOfReturns)
  header <- rlas::header_create(data)
  header[["X scale factor"]] <- 0.001
  header[["Y scale factor"]] <- 0.001
  header[["Z scale factor"]] <- 0.001
  header[["X offset"]] <- 0
  header[["Y offset"]] <- 0
  header[["Z offset"]] <- 0
  las <- suppressMessages(suppressWarnings(LAS(data, header, check = FALSE)))
  st_crs(las) <- 26917
  return(las)
}

catalog_generator <- function(ntiles = 4, npoints = 250, size = 100, seed = 1) {
  X <- Y <- NULL

  xntiles = as.integer(floor(ntiles/2))
  yntiles = as.integer(ceiling(ntiles/xntiles))

  # Generate a catalog
  xshift <- seq(0, (xntiles-1)*size, by = size)
  yshift <- seq(0, (yntiles-1)*size, by = size)
  ntiles <- length(xshift)*length(yshift)
  temp   <- sapply(1:ntiles, function(x) tempfile(fileext = ".las"))

  k = 1
  for (i in 1:length(xshift))
  {
    for (j in 1:length(yshift))
    {
      las <- generate_las(npoints, seeds = k + seed)
      las@data[, X := X + xshift[i]]
      las@data[, Y := Y + yshift[j]]
      writeLAS(las, temp[k])
      k = k+1
    }
  }

  ctg <- readLAScatalog(temp)
  opt_chunk_buffer(ctg)    <- 0
  opt_chunk_size(ctg)      <- 0
  opt_progress(ctg)        <- FALSE
  opt_output_files(ctg)    <- ""

  return(ctg)
}

context("progress bar")

test_that("progress bar work at C++ level", {
  sink(tempfile())
  las = lidR:::generate_las(500)
  options(lidR.progress = TRUE)
  options(lidR.progress.delay = 0)
  expect_error(smooth_height(las, 3), NA)
  options(lidR.progress = FALSE)
  options(lidR.progress.delay = 2)
  sink(NULL)
})

las <- lidR:::generate_las(2000)
f1  <- ~list(Zmean = mean(Z))
f2  <- ~list(meanZ = mean(Z), maxZ = max(Z))

test_that("hexagon_metrics returns a named sf", {

  x <- hexagon_metrics(las, f1, area = 800)

  expect_true(is(x, "sf"))
  expect_equivalent(nrow(x), 22)
  expect_equivalent(names(x)[1], "Zmean")

  x <- hexagon_metrics(las, f2, area = 800)

  expect_true(is(x, "sf"))
  expect_equivalent(nrow(x), 22)
  expect_equivalent(names(x)[1:2], c("meanZ", "maxZ"))
})


test_that("hexagon_metrics work with by_echo and filter", {

  x <- hexagon_metrics(las, f2, area = 800, filter = ~Z > 2, by_echo = c("all", "lastofmany", "single"))

  expect_true(is(x, "sf"))
  expect_equivalent(dim(x), c(22, 7))
})

#' @param geometry A spatial vector object. `sp` and `sf`' objects are supported. `plot_metrics()`
#' supports point and polygons but `polygon_metrics()` supports only polygons.
#' @param radius numeric. If the geometry is spatial points a radius must be defined. Support one
#' radius or a vector of radii for variable plot sizes.
#' @rdname aggregate
#' @export
#' @md
plot_metrics <- function(las, func, geometry, ..., radius)
{
  if (missing(radius))
    rois <- suppressWarnings(clip_roi(las, geometry))
  else
    rois <- suppressWarnings(clip_roi(las, geometry, radius = radius))

  if (is(rois, "LAScatalog"))
  {
    opt_independent_files(rois) <- TRUE
    opt_output_files(rois) <- ""
    opt_progress(rois) <- FALSE
    metrics <- catalog_apply(rois, cloud_metrics, func = func, ...)
  }
  else
  {
    # Fix 664: clip_roi returns inconsistently a LAS or a list depending if there is
    # one or muliple queries. If there is only one query rois in not a list and sapply fails
    if (is(rois, "LAS"))
      rois = list(rois)

    n <- length(rois)
    empty <- sapply(rois, is.empty)
    any_empty <- any(empty)
    if (any_empty)
    {
      k <- which(empty)
      warning(glue::glue("Plots [{glue::glue_collapse(k, sep = ', ')}] are empty."), call. = FALSE)
      rois <- rois[-k]
    }

    metrics <- lapply(rois, template_metrics, template = NULL, func = func, ...)

    if (any_empty)
    {
      incomplete_metrics <- metrics
      ref <- metrics[[1]]
      metrics <- vector("list", n)
      ref <- lapply(ref, function(x) return(NA))
      j = 1
      for (i in 1:n)
      {
        if (empty[i])
        {
          metrics[[i]] <- ref
        }
        else
        {
          metrics[[i]] <- incomplete_metrics[[j]]
          j = j+1
        }
      }

    }
  }

  if (length(metrics[[1]]) == 1)
  {
    name <- names(metrics[[1]])
    if (is.null(name)) name = "V1"
    metrics <- data.table::data.table(unlist(metrics))
    names(metrics) <- name
  }
  else
  {
    metrics <- data.table::rbindlist(metrics)
  }

  data.table::setDF(metrics)
  metrics <- cbind(geometry, metrics)
  return(metrics)
}

test_that("plot_metrics works if some plot contain 0 points (#551)", {

  shpfile <- system.file("extdata", "efi_plot.shp", package="lidR")
  inventory <- sf::st_read(shpfile, quiet = TRUE)
  inventory$geometry[4][[1]][2] = 5017891 + 300

  expect_warning(m <- plot_metrics(megaplot, .stdmetrics_z, inventory, radius = 11.28), "are empty")
  expect_equal(nrow(m), nrow(inventory))
  expect_equal(m[4,]$zmax, NA_real_)
  expect_equal(m[4,]$zq5, NA_real_)
})

test_that("plot_metrics works with 1 plot on LAScatalog (#664)", {

  shpfile <- system.file("extdata", "efi_plot.shp", package="lidR")
  inventory <- sf::st_read(shpfile, quiet = TRUE)
  inventory <- inventory[1,]

  expect_error(m <- plot_metrics(megaplot_ctg, .stdmetrics_z, inventory, radius = 11.28), NA)

})

#' Inspect a LAS object
#'
#' Performs a deep inspection of a LAS or LAScatalog object and prints a report.\cr\cr
#' For a LAS object it checks:
#' \itemize{
#' \item if the point cloud is valid according to las specification
#' \item if the header is valid according to las specification
#' \item if the point cloud is in accordance with the header
#' \item if the point cloud has duplicated points and degenerated ground points
#' \item if gpstime and pulses are consistent
#' \item it the coordinate reference sytem is correctly recorded
#' \item if some pre-processing, such as normalization or ground filtering, is already done.
#' \item and much more
#' }
#' For a LAScatalog object it checks:
#' \itemize{
#' \item if the headers are consistent across files
#' \item if the files are overlapping
#' \item if some pre-processing, such as normalization, is already done.
#' }
#' For the pre-processing tests the function only makes an estimation and may not be correct.
#'
#' @template param-las
#' @param print logical. By default, prints a report and returns a \code{list} invisibly. If
#' \code{print = FALSE} the functions returns a \code{list} visibly and do not print the report.
#' @param ... Use \code{deep = TRUE} on a LAScatalog only. Instead of a shallow inspection it reads
#' all the files and performs a deep inspection.
#'
#' @return A list with three elements named \code{message}, \code{warnings} and \code{errors}. This list is returned
#' invisibly if \code{print = TRUE}. If \code{deep = TRUE} a nested list is returned with one element
#' per file.
#'
#' @examples
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' las <- readLAS(LASfile)
#' las_check(las)
#' @export
#' @family las utilities
las_check = function(las, print = TRUE, ...)
{
  UseMethod("las_check", las)
}

#' @export
las_check.LASheader = function(las, print = TRUE, ...)
{
  xscale  <- las[["X scale factor"]]
  xoffset <- las[["X offset"]]
  yscale  <- las[["Y scale factor"]]
  yoffset <- las[["Y offset"]]
  zscale  <- las[["Z scale factor"]]
  zoffset <- las[["Z offset"]]

  head <- as.list(las)
  g    <- glue::glue

  warnings <- character(0)
  errors <- character(0)
  infos <- character(0)

  .h1("Checking the header")

  .h2("Checking header completeness...")

  msg = character(0)
  msg = c(msg, rlas::is_defined_offsets(head, "vector"))
  msg = c(msg, rlas::is_defined_scalefactors(head, "vector"))
  msg = c(msg, rlas::is_defined_version(head, "vector"))
  msg = c(msg, rlas::is_defined_pointformat(head, "vector"))
  msg = c(msg, rlas::is_defined_date(head, "vector"))
  msg = c(msg, rlas::is_defined_globalencoding(head, "vector"))

  .fail(msg)

  .h2("Checking scale factor validity...")

  .fail(rlas::is_valid_scalefactors(head, "vector"))

  .h2("Checking point data format ID validity...")

  .fail(rlas::is_valid_pointformat(head, "vector"))

  .h2("Checking extra bytes attributes validity...")

  .fail(rlas::is_valid_extrabytes(head, "vector"))

  .h2("Checking the bounding box validity...")

  bbx = c(head[["Min X"]], head[["Max X"]])
  bby = c(head[["Min Y"]], head[["Max Y"]])
  bbz = c(head[["Min Z"]], head[["Max Z"]])

  i <- fast_countunquantized(bbx, xscale, xoffset)
  j <- fast_countunquantized(bby, yscale, yoffset)
  k <- fast_countunquantized(bbz, zscale, zoffset)

  if (i + j + k > 0)
  {
    if (i > 0) .fail("Stored resolution of 'Min X' and/or 'Max X' not compatible with 'X offset' and 'X scale factor'")
    if (j > 0) .fail("Stored resolution of 'Min Y' and/or 'Max Y' not compatible with 'Y offset' and 'Y scale factor'")
    if (k > 0) .fail("Stored resolution of 'Min Z' and/or 'Max Z' not compatible with 'Z offset' and 'Z scale factor'")
  }
  else
  {
    .ok()
  }

  .h2("Checking coordinate reference system...")

  code    <- if (use_epsg(las)) epsg(las) else 0
  swkt    <- wkt(las)
  failure <- FALSE

  if (use_epsg(las) && swkt != "")
  { .fail("Global encoding WKT bits set to 0 but a WKT string found in the header") ; failure = TRUE }

  if (use_wktcs(las) && code != 0)
  { .fail("Global encoding WKT bits set to 1 but an epsg code found in the header") ; failure = TRUE }

  if (!failure)
    .ok()

  warnerr = list(
    messages = infos,
    warnings = warnings,
    errors = errors)

  if (print)
    return(invisible(warnerr))
  else
    return(warnerr)
}

#' @export
las_check.LAS = function(las, print = TRUE, ...)
{
  assert_is_a_bool(print)

  warnings <- character(0)
  errors <- character(0)
  infos <- character(0)
  g <- glue::glue

  #nocov start
  if (is_las_v3(las))
  {
    .h1("Checking format")
    .h2("Checking lidR format version is v4...")
    .fail("This LAS is in old format from lidR v3")
    las <- las_v3_repair(las)
  }
  #nocov end

  data <- payload(las)
  head <- as.list(header(las))

  xscale <- las[["X scale factor"]]
  xoffset <- las[["X offset"]]
  yscale <- las[["Y scale factor"]]
  yoffset <- las[["Y offset"]]
  zscale <- las[["Z scale factor"]]
  zoffset <- las[["Z offset"]]

  # ==== data =====

  .h1("Checking the data")

  if (npoints(las) == 0)
  {
    .fail(glue::glue("0 point, cannot check this object. Aborting."))
    warnerr = list(
      messages = infos,
      warnings = warnings,
      errors = errors)

    if (print)
      return(invisible(warnerr))
    else
      return(warnerr)
  }

  .h2("Checking coordinates...")

  .fail(rlas::is_defined_coordinates(data, "vector"))

  .h2("Checking coordinates type...")

  .fail(rlas::is_valid_XYZ(data, "vector"))

  .h2("Checking coordinates range...")

  xvalid_range <- storable_coordinate_range(xscale, xoffset)
  yvalid_range <- storable_coordinate_range(yscale, yoffset)
  zvalid_range <- storable_coordinate_range(zscale, zoffset)
  xrange <- range(data$X)
  yrange <- range(data$Y)
  zrange <- range(data$Z)

  failure = FALSE

  if (xrange[1] < xvalid_range[1] | xrange[2] > xvalid_range[2])
  {
    .fail(glue::glue("X coordinates range in [{xrange[1]}, {xrange[2]}] but storable range is  [{xvalid_range[1]}, {xvalid_range[2]}]"))
    failure = TRUE
  }

  if (yrange[1] < yvalid_range[1] | yrange[2] > yvalid_range[2])
  {
    .fail(glue::glue("Y coordinates range in [{yrange[1]}, {yrange[2]}] but storable range is  [{yvalid_range[1]}, {yvalid_range[2]}]"))
    failure = TRUE
  }

  if (zrange[1] < zvalid_range[1] | zrange[2] > zvalid_range[2])
  {
    .fail(glue::glue("Z coordinates range in [{zrange[1]}, {zrange[2]}] but storable range is  [{zvalid_range[1]}, {zvalid_range[2]}]"))
    failure = TRUE
  }

  if (failure == FALSE)
    .ok()

  .h2("Checking coordinates quantization...")

  i <- fast_countunquantized(las$X, xscale, xoffset)
  j <- fast_countunquantized(las$Y, yscale, yoffset)
  k <- fast_countunquantized(las$Z, zscale, zoffset)

  if (i + j + k > 0)
  {
    if (i > 0) .fail(glue::glue("{i} X coordinates were not stored with a resolution compatible with scale factor {xscale} and offset {xoffset}"))
    if (j > 0) .fail(glue::glue("{j} Y coordinates were not stored with a resolution compatible with scale factor {yscale} and offset {yoffset}"))
    if (k > 0) .fail(glue::glue("{k} Z coordinates were not stored with a resolution compatible with scale factor {zscale} and offset {zoffset}"))
  }
  else
  {
    .ok()
  }

  .h2("Checking attributes type...")

  msg = character(0)
  msg = c(msg, rlas::is_valid_gpstime(data, "vector"))
  msg = c(msg, rlas::is_valid_Intensity(data, "vector"))
  msg = c(msg, rlas::is_valid_ReturnNumber(data, head, "vector"))
  msg = c(msg, rlas::is_valid_EdgeOfFlightline(data, "vector"))
  msg = c(msg, rlas::is_valid_Classification(data, head, "vector"))
  msg = c(msg, rlas::is_valid_UserData(data, "vector"))
  msg = c(msg, rlas::is_valid_ScanAngleRank(data, "vector"))
  msg = c(msg, rlas::is_valid_ScanAngle(data, "vector"))
  msg = c(msg, rlas::is_valid_PointSourceID(data, "vector"))
  msg = c(msg, rlas::is_valid_RGB(data, "vector"))
  msg = c(msg, rlas::is_valid_NIR(data, "vector"))
  msg = c(msg, rlas::is_valid_SyntheticFlag(data, "vector"))
  msg = c(msg, rlas::is_valid_KeypointFlag(data, "vector"))
  msg = c(msg, rlas::is_valid_WithheldFlag(data, "vector"))

  .fail(msg)


  .h2("Checking ReturnNumber validity...")

  .warn(rlas::is_compliant_ReturnNumber(data, "vector"))

  .h2("Checking NumberOfReturns validity...")

  .warn(rlas::is_compliant_NumberOfReturns(data, "vector"))

  .h2("Checking ReturnNumber vs. NumberOfReturns...")

  .warn(rlas::is_compliant_ReturnNumber_vs_NumberOfReturns(data, "vector"))

  .h2("Checking RGB validity...")

  .warn(rlas::is_compliant_RGB(data, "vector"))

  .h2("Checking absence of NAs...")

  nas = data[, lapply(.SD, anyNA)]
  nas = unlist(as.list(nas))
  nas = nas[nas == TRUE]
  whichnas = names(nas)
  whichnas = paste(whichnas, collapse = ", ")

  if (length(nas) > 0)
  {
    string = paste("The following attributes contain NAs:", whichnas)
    .fail(string)
  }
  else
    .ok()

  .h2("Checking duplicated points...")

  s = sum(duplicated(data, by = c("X", "Y", "Z")))

  if (s > 0)
    .warn(g("{s} points are duplicated and share XYZ coordinates with other points"))
  else
    .ok()

  .h2("Checking degenerated ground points...")

  if (!is.null(data$Classification))
  {
    s = fast_count_equal(data$Classification, 2L)

    if (s > 0)
    {
      Classification <- NULL
      gnd = payload(filter_ground(las))

      s1 = duplicated(gnd, by = c("X", "Y", "Z"))
      s2 = duplicated(gnd, by = c("X", "Y"))  & !s1
      s1 = sum(s1)
      s2 = sum(s2)

      if (s1 == 0 & s2 == 0)
        .ok()
      else
      {
        if (s1 > 0)
          .warn(g("There were {s1} degenerated ground points. Some X Y Z coordinates were repeated"))

        if (s2 > 0)
          .warn(g("There were {s2} degenerated ground points. Some X Y coordinates were repeated but with different Z coordinates"))

      }
    }
    else
      .skip()
  }
  else
    .skip()

  .h2("Checking attribute population...")

  msg = character(0)

  if (!is.null(data[["gpstime"]]))
  {
    s = all(data[["gpstime"]] == 0)

    if (s)
      msg = c(msg, g("'gpstime' attribute is not populated"))
  }

  if (!is.null(data[["PointSourceID"]]))
  {
    s = fast_count_equal(data[["PointSourceID"]], 0L)

    if (s == nrow(data))
      msg = c(msg, g("'PointSourceID' attribute is not populated"))
  }

  if (!is.null(data[["ScanDirectionFlag"]]))
  {
    s = fast_count_equal(data[["ScanDirectionFlag"]], 0L)

    if (s == nrow(data))
      msg = c(msg, g("'ScanDirectionFlag' attribute is not populated"))
  }

  if (!is.null(data[["EdgeOfFlightline"]]))
  {
    s = fast_count_equal(data[["EdgeOfFlightline"]], 0L)

    if (s == nrow(data))
      msg = c(msg, g("'EdgeOfFlightline' attribute is not populated"))
  }

  .info(msg)

  .h2("Checking gpstime incoherances")

  if (!is.null(data[["gpstime"]]) && !is.null(data[["ReturnNumber"]]))
  {
    s1 <- C_check_gpstime(data[["gpstime"]], data[["ReturnNumber"]])
    if (s1 > 0)
      .fail(g("{s1} pulses (points with the same gpstime) have points with identical ReturnNumber"))
    else
      .ok()
  }
  else
    .skip()


  .h2("Checking flag attributes...")

  msg = character(0)

  if (!is.null(data[["Withheld_flag"]]))
  {
    s = sum(data[["Withheld_flag"]])

    if (s > 0)
      msg = c(msg, g("{s} points flagged 'withheld'"))
  }

  if (!is.null(data[["Synthetic_flag"]]))
  {
    s = sum(data[["Synthetic_flag"]])

    if (s > 0)
      msg = c(msg, g("{s} points flagged 'synthetic'"))
  }

  if (!is.null(data[["Keypoint_flag"]]))
  {
    s = sum(data[["Keypoint_flag"]])

    if (s > 0)
      msg = c(msg, g("{s} points flagged 'keypoint'"))
  }

  .info(msg)

  .h2("Checking user data attribute...")

  msg = character(0)

  if (!is.null(data[["UserData"]]))
  {
    s <- fast_count_equal(data[["UserData"]], 0L)
    s <- npoints(las) - s

    if (s > 0)
      .info(g("{s} points have a non 0 UserData attribute. This probably has a meaning"))
    else
      .ok()
  }
  else
  {
    .skip()
  }

  # ==== header ====

  head_chk <- las_check(header(las), print = print, ...)
  infos <- c(infos, head_chk[["infos"]])
  errors <- c(errors, head_chk[["errors"]])
  warnings <- c(warnings, head_chk[["warnings"]])

  # ==== data vs header ====

  .h1("Checking header vs data adequacy")

  .h2("Checking attributes vs. point format...")

  msg = character(0)
  msg = c(msg, rlas::is_NIR_in_valid_format(head, data, "vector"))
  msg = c(msg, rlas::is_gpstime_in_valid_format(head, data, "vector"))
  msg = c(msg, rlas::is_RGB_in_valid_format(head, data, "vector"))

  .fail(msg)

  .h2("Checking header bbox vs. actual content...")

  if (any(c("X", "Y", "Z") %in% whichnas)) {
    .skip()
  }
  else {
    msg = character(0)
    msg = c(msg, rlas::is_XY_larger_than_bbox(head, data, "vector"))
    msg = c(msg, rlas::is_XY_smaller_than_bbox(head, data, "vector"))
    msg = c(msg, rlas::is_Z_in_bbox(head, data, "vector"))
    .warn(msg)
  }

  .h2("Checking header number of points vs. actual content...")

  .warn(rlas::is_number_of_points_in_accordance_with_header(head, data, "vector"))

  .h2("Checking header return number vs. actual content...")

  .warn(rlas::is_number_of_points_by_return_in_accordance_with_header(head, data, "vector"))

  # ==== CRS ====

  .h1("Checking coordinate reference system...")

  .h2("Checking if the CRS was understood by R...")

  code    <- if (use_epsg(las)) epsg(las) else 0
  swkt    <- wkt(las)
  lasproj <- st_crs(las)
  failure <- FALSE

  if (use_epsg(las) && code != 0)
  {
    codeproj <- epsg2crs(code, fail = FALSE)

    if (is.na(codeproj))
    { .fail(glue::glue("EPSG code {code} unknown")) ; failure = TRUE }

    if (is.na(codeproj) && !is.na(lasproj))
    { .warn(glue::glue("EPSG code is unknown but a CRS found")) ; failure = TRUE }

    if (!is.na(codeproj) && is.na(lasproj))
    { .warn("ESPG code is valid but no CRS found") ; failure = TRUE }

    if (!is.na(codeproj) && !is.na(lasproj))
    {
      if (codeproj != lasproj)
      { .fail("ESPG code and CRS do not match") ; failure = TRUE }
    }

    if (!failure)
      .ok()
  }

  if (use_wktcs(las) && swkt != "")
  {
    codeproj = wkt2crs(swkt, fail = FALSE)

    if (is.na(codeproj))
    { .fail("WKT OGC CS not parsed") ; failure = TRUE }

    if (is.na(codeproj) & !is.na(lasproj))
    { .warn("WKT OGC CS not parsed but a CRS found") ; failure = TRUE }

    if (!is.na(codeproj) & is.na(lasproj))
    { .warn("WKT OGC CS is valid but no CRS found") ; failure = TRUE }

    if (!is.na(codeproj) & !is.na(lasproj))
    {
      if (codeproj != lasproj)
      { .fail("WKT OGC CS and CRS do not match") ; failure = TRUE }
    }

    if (!failure)
      .ok()
  }

  if (code == 0 & swkt == "")
  {
    if (!is.na(lasproj))
    { .warn("A CRS found but no CRS in the header") ; failure = TRUE }

    if (!failure)
      .ok()
  }

  # ==== Preprocessing ====

  .h1("Checking preprocessing already done ")

  .h2("Checking ground classification...")

  if (!is.null(data$Classification))
  {
    s = fast_count_equal(data$Classification, 2L)

    if (s > 0) {
      .yes()
      infos <- append(infos, "The point cloud is ground classified")
    }
    else {
      .no()
      infos <- append(infos, "The point cloud is not ground classified")
    }
  }
  else
    .skip()

  .h2("Checking normalization...")

  if (any(c("X", "Y", "Z") %in% whichnas)) {
    .skip()
  }
  else
  {
    min <- rasterize_fast(las, res = 20, method = "min")
    val <- raster_values(min)
    mean_min <- mean(abs(val), na.rm = TRUE)

    if (mean_min <= 0.1) {
      .yes()
      infos <- append(infos, "The point cloud is height normalized")
    }
    else if (mean_min > 0.1 & mean_min < 1) {
      .maybe()
      infos <- append(infos, "The point cloud is maybe height normalized")
    }
    else {
      .no()
      infos <- append(infos, "The point cloud is not height normalized")
    }
  }

  .h2("Checking negative outliers...")

  s = fast_countbelow(data$Z, 0)

  if (s > 0)
    .warn(g("{s} points below 0"))
  else
    .ok()

  .h2("Checking flightline classification...")

  if (!is.null(data$PointSourceID))
  {
    s = fast_count_equal(data$PointSourceID, 0L)

    if (s == nrow(data)) {
      .no()
    }
    else if (s > 0 & s < nrow(data)) {
      .maybe()
    }
    else {
      .yes()
    }
  }
  else
    .skip()

  # ==== Compression ====

  .h1("Checking compression")

  gz <- las_is_compressed(las)
  gz <- gz[gz == TRUE]

  .h2("Checking attribute compression...")

  if (length(gz) == 0)
  {
    if (utils::packageVersion("rlas") < "1.6.0")
      .info("Compression supported only from rlas 1.6.0")
    else
      .no()
  }
  else
  {
    if (print)
    {
      cat("\n")
      for (name in names(gz))
        cat("   - ", name, "is compressed\n")
    }
  }

  warnerr = list(
    messages = infos,
    warnings = warnings,
    errors = errors)

  if (print)
    return(invisible(warnerr))
  else
    return(warnerr)
}

#' @export
las_check.LAScluster = function(las, print = TRUE, ...)
{
  f <- basename(las@files)
  if (length(f) > 1) stop("Internal error: several files in the LAScluster. Please report this issue")

  if (print)
  {
    title <- glue::glue("Checking the file: {f}")
    n = nchar(title)
    header <- strrep("_", n)
    cat("\n\n")
    cat(header)
    cat("\n")
    cat(title)
    cat("\n")
    cat(header)
    cat("\n")
  }

  x <- readLAS(las)
  u <- las_check(x, print)

  if (print)
    return(invisible(u))
  else
    return(u)
}

#' @export
las_check.LAScatalog = function(las, print = TRUE, deep = FALSE, ...)
{
  assert_is_a_bool(print)
  assert_is_a_bool(deep)

  # Deep inspection of each file
  if (deep)
  {
    opt_chunk_size(las) <- 0
    opt_chunk_buffer(las) <- 0
    opt_select(las) <- "*"
    opt_filter(las) <- ""
    opt_output_files(las) <- ""
    opt_wall_to_wall(las) <- FALSE
    if (print) opt_progress(las) <-  FALSE
    out <- catalog_apply(las, las_check, print = print)
    names(out) <- basename(las@data[["filename"]])

    if (print)
      return(invisible(out))
    else
      return(out)
  }

  warnings <- character(0)
  errors <- character(0)
  infos <- character(0)
  g <- glue::glue

  # nocov start
  if (is_lascatalog_v3(las))
  {
    .h1("Checking format")
    .h2("Checking lidR format version is v4...")
    .fail("This LAScatalog is in old format from lidR v3")
    las <- lascatalog_v3_repair(las)
  }
  # nocov end

  data <- las@data

  # ==== data =====


  .h1("Checking headers consistency")

  .h2("Checking file version consistency...")

  s = length(unique(paste0(data$Version.Major, ".", data$Version.Minor)))

  if (s > 1L)
    .warn("Inconsistent file versions")
  else
    .ok()

  .h2("Checking scale consistency...")

  s1 = length(unique(data$X.scale.factor))
  s2 = length(unique(data$Y.scale.factor))
  s3 = length(unique(data$Z.scale.factor))

  if (s1 + s2 + s3 > 3L)
    .warn("Inconsistent scale factors")
  else
    .ok()

  .h2("Checking offset consistency...")

  s1 = length(unique(data$X.offset))
  s2 = length(unique(data$Y.offset))
  s3 = length(unique(data$Z.offset))

  if (s1 + s2 + s3 > 3L)
    .warn("Inconsistent offsets")
  else
    .ok()

  .h2("Checking point type consistency...")

  s = length(unique(data$Point.Data.Format.ID))

  if (s > 1L)
    .warn("Inconsistent point formats")
  else
    .ok()

  .h2("Checking VLR consistency...")

  s = length(unique(data$Number.of.variable.length.record))

  if (s > 1L)
    .fail("Inconsistent number of VLR")
  else
    .ok()

  .h2("Checking CRS consistency...")

  s = length(unique(data[["CRS"]]))

  if (s > 1L)
    .fail("Inconsistent CRS accross files")
  else
    .ok()

  .h1("Checking the headers")

  .h2("Checking scale factor validity...")

  failure = FALSE

  s = c(1,10,100,1000,10000)
  valid = c(1/s, 0.5/s, 0.25/s)

  if (any(!data$X.scale.factor %in% valid))
  { .fail("Invalid header: X scale factor should be factor ten of 0.1 or 0.5 or 0.25") ; failure = TRUE }

  if (any(!data$Y.scale.factor %in% valid))
  { .fail("Invalid header: Y scale factor should be factor ten of 0.1 or 0.5 or 0.25") ; failure = TRUE }

  if (any(!data$Z.scale.factor %in% valid))
  { .fail("Invalid header: Z scale factor should be factor ten of 0.1 or 0.5 or 0.25") ; failure = TRUE }

  if (!failure)
    .ok()

  .h2("Checking Point Data Format ID validity...")

  if (any(data$Point.Data.Format.ID %in% c(4,5,9,10)))
    .warn("Invalid headers: point data format not supported yet")
  else if (any(data$Point.Data.Format.ID < 0 | data$Point.Data.Format.ID > 10))
    .fail("Invalid header: point data format invalid")
  else
    .ok()

  .h1("Checking preprocessing already done ")

  .h2("Checking negative outliers...")

  s = sum(data$Min.Z < 0)

  if (s > 0)
    .warn(g("{s} file(s) with points below 0"))
  else
    .ok()

  .h2("Checking normalization...")

  mean_min = mean(abs(data$Min.Z))

  if (mean_min <= 0.1) {
    .yes()
    infos <<- append(infos, "The point cloud is height normalized")
  }
  else if (mean_min > 0.1 & mean_min < 2) {
    .maybe()
    infos <<- append(infos, "The point cloud is maybe height normalized")
  }
  else {
    .no()
    infos <<- append(infos, "The point cloud is not height normalized")
  }

  .h1("Checking the geometry")

  .h2("Checking overlapping tiles...")

  if (is.overlapping(las))
    .warn("Some tiles seem to overlap each other")
  else
    .ok()

  .h2("Checking point indexation...")

  if (is.indexed(las)) {
    .yes()
    #infos <- append(infos, "The LAS files are spatially indexed")
  }
  else {
    .no()
    infos <- append(infos, "The LAS files are not spatially indexed")
  }

  warnerr = list(
    messages = infos,
    warnings = warnings,
    errors = errors)

  if (print)
    return(invisible(warnerr))
  else
    return(warnerr)
}

.h1    <- function(x) {if (get("print", envir = parent.frame())) cat("\n", x)}
.h2    <- function(x) {if (get("print", envir = parent.frame())) cat("\n  -", x)}
.ok    <- function()  {if (get("print", envir = parent.frame())) cat(.colourise(" \u2713", "green"))}
.skip  <- function()  {if (get("print", envir = parent.frame())) cat(.colourise(" skipped", "light gray"))}
.no    <- function()  {if (get("print", envir = parent.frame())) cat(.colourise(" no", "red"))}
.yes   <- function()  {if (get("print", envir = parent.frame())) cat(.colourise(" yes", "green"))}
.maybe <- function()  {if (get("print", envir = parent.frame())) cat(.colourise(" maybe", "yellow"))}

.fail  <- function(msg) {
  print <- get("print", envir = parent.frame())
  if (print) {
    if (length(msg) == 0) {
      .ok()
    } else {
      for (x in msg) cat("\n", .colourise(glue::glue("   \U2717 {x}"), "red"))
    }
  }

  if (length(msg) > 0) {
    errors <- get("errors", envir = parent.frame())
    for (x in msg) errors <- append(errors, x)
    assign("errors", errors, envir = parent.frame())
  }
}

.warn  <- function(msg) {
  print <- get("print", envir = parent.frame())
  if (print) {
    if (length(msg) == 0) {
      .ok()
    } else {
      for (x in msg) cat("\n", .colourise(glue::glue("   \U26A0 {x}"), "yellow"))
    }
  }

  if (length(msg) > 0) {
    warnings <- get("warnings", envir = parent.frame())
    for (x in msg) warnings <- append(warnings, x)
    assign("warnings", warnings, envir = parent.frame())
  }
}

.info  <- function(msg) {
  print <- get("print", envir = parent.frame())
  if (print) {
    if (length(msg) == 0) {
      .ok()
    } else {
      for (x in msg) cat("\n", .colourise(glue::glue("   \U1F6C8 {x}"), "green"))
    }
  }

  if (length(msg) > 0) {
    infos <- get("infos", envir = parent.frame())
    for (x in msg) infos <- append(infos, x)
    assign("infos", infos, envir = parent.frame())
  }
}

# code from testthat
# https://github.com/r-lib/testthat/blob/717b02164def5c1f027d3a20b889dae35428b6d7/R/colour-text.r
# nocov start
.colourise <- function(text, fg = "black", bg = NULL) {
  term <- Sys.getenv()["TERM"]
  colour_terms <- c("xterm-color","xterm-256color", "screen", "screen-256color")

  if (nchar(Sys.getenv('R_TESTS')) != 0 || !any(term %in% colour_terms, na.rm = TRUE)) {
    return(text)
  }

  .fg_colours <- c(
    "black" = "0;30",
    "blue" = "0;34",
    "green" = "0;32",
    "cyan" = "0;36",
    "red" = "0;31",
    "purple" = "0;35",
    "brown" = "0;33",
    "light gray" = "0;37",
    "dark gray" = "1;30",
    "light blue" = "1;34",
    "light green" = "1;32",
    "light cyan" = "1;36",
    "light red" = "1;31",
    "light purple" = "1;35",
    "yellow" = "1;33",
    "white" = "1;37"
  )

  .bg_colours <- c(
    "black" = "40",
    "red" = "41",
    "green" = "42",
    "brown" = "43",
    "blue" = "44",
    "purple" = "45",
    "cyan" = "46",
    "light gray" = "47"
  )

  col_escape <- function(col) {
    paste0("\033[", col, "m")
  }

  col <- .fg_colours[tolower(fg)]
  if (!is.null(bg)) {
    col <- paste0(col, .bg_colours[tolower(bg)], sep = ";")
  }

  init <- col_escape(col)
  reset <- col_escape("0")
  paste0(init, text, reset)
}
# nocov end

#' Voxelize a point cloud
#'
#' Reduce the number of points by voxelizing the point cloud. If the Intensity is part of the attributes
#' it is preserved and aggregated as \code{mean(Intensity)}. Other attributes cannot be aggregated and
#' are lost.
#'
#' @template param-las
#' @param res numeric. The resolution of the voxels. \code{res = 1} for a 1x1x1 cubic voxels. Optionally
#' \code{res = c(1,2)} for non-cubic voxels (1x1x2 cuboid voxel).
#'
#' @template return-lasfilter-las-lascatalog
#'
#' @export
#'
#' @examples
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' las = readLAS(LASfile, select = "xyz")
#'
#' las2 = voxelize_points(las, 5)
#' #plot(las2, voxel = TRUE)
voxelize_points = function(las, res)
{
  assert_all_are_non_negative(res)

  UseMethod("voxelize_points", las)
}

#' @export
voxelize_points.LAS = function(las, res)
{
  Intensity <- NULL

  if (length(res) == 1L)
    res <- c(res,res)
  else if (length(res) > 2L)
    stop("Wrong resolution provided.")

  by <- group_grid_3d(las@data$X, las@data$Y, las@data$Z, res, c(0,0,0.5*res[2]))

  if ("Intensity" %in% names(las))
  {
    voxels <- las@data[, list(Intensity = as.integer(mean(Intensity))), by = by]
    data.table::setnames(voxels, c("X", "Y", "Z", "Intensity"))
  }
  else
  {
    data.table::setDT(by)
    data.table::setnames(by, c("X", "Y", "Z"))
    voxels <- unique(by)
  }

  output <- LAS(voxels, header = las@header, crs = st_crs(las), check = FALSE, index = las@index)
  return(output)
}

#' @export
voxelize_points.LAScluster = function(las, res)
{
  x <- readLAS(las)
  if (is.empty(x)) return(NULL)

  output <- voxelize_points(x, res)
  output <- clip_roi(output, st_bbox(las))
  return(output)
}

#' @export
voxelize_points.LAScatalog = function(las, res)
{
  opt_select(las) <- "xyzi"
  if (opt_wall_to_wall(las))
    opt_chunk_buffer(las) <- res[1]

  options <- list(need_buffer = FALSE, drop_null = TRUE, need_output_file = TRUE)
  output  <- catalog_apply(las, voxelize_points, res = res, .options = options)
  output  <- unlist(output)
  ctg     <- suppressMessages(suppressWarnings(readLAScatalog(output)))

  opt_copy(ctg) <- las
  return(ctg)
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

filterTimeBlockPulses <- function(pulsedt) {
    .Call(`_lidR_filterTimeBlockPulses`, pulsedt)
}

cmpCPA <- function(pulsedt) {
    .Call(`_lidR_cmpCPA`, pulsedt)
}

C_chm_prep <- function(data, snlin, sncol, lap_size, thr_cav, thr_spk, med_size, dil_radius, nodata) {
    .Call(`_lidR_C_chm_prep`, data, snlin, sncol, lap_size, thr_cav, thr_spk, med_size, dil_radius, nodata)
}

C_dalponte2016 <- function(Image, Seeds, th_seed, th_crown, th_tree, DIST) {
    .Call(`_lidR_C_dalponte2016`, Image, Seeds, th_seed, th_crown, th_tree, DIST)
}

C_delaunay <- function(P, scales, offsets, trim = 0) {
    .Call(`_lidR_C_delaunay`, P, scales, offsets, trim)
}

C_interpolate_delaunay <- function(P, L, scales, offsets, trim = 0, min_normal_z = 0, ncpu = 1L) {
    .Call(`_lidR_C_interpolate_delaunay`, P, L, scales, offsets, trim, min_normal_z, ncpu)
}

C_tinfo <- function(D, P) {
    .Call(`_lidR_C_tinfo`, D, P)
}

C_tsearch <- function(D, P, X, ncpu) {
    .Call(`_lidR_C_tsearch`, D, P, X, ncpu)
}

C_ptd <- function(df, res, angle, distance, spacing, verbose, ncores) {
    .Call(`_lidR_C_ptd`, df, res, angle, distance, spacing, verbose, ncores)
}

C_spikefree <- function(df, grid, d_f, h_b, slope = 1.96, intercept = 1.42) {
    .Call(`_lidR_C_spikefree`, df, grid, d_f, h_b, slope, intercept)
}

C_lmf <- function(las, ws, min_height, circular, ncpu) {
    .Call(`_lidR_C_lmf`, las, ws, min_height, circular, ncpu)
}

C_smooth <- function(las, size, method, shape, sigma, ncpu) {
    .Call(`_lidR_C_smooth`, las, size, method, shape, sigma, ncpu)
}

C_highest <- function(las, layout) {
    .Call(`_lidR_C_highest`, las, layout)
}

C_lowest <- function(las, layout) {
    .Call(`_lidR_C_lowest`, las, layout)
}

C_in_polygon <- function(las, polygons, by_poly) {
    .Call(`_lidR_C_in_polygon`, las, polygons, by_poly)
}

C_lasdetectshape <- function(las, method, th, k, filter, ncpu) {
    .Call(`_lidR_C_lasdetectshape`, las, method, th, k, filter, ncpu)
}

C_Wing2015 <- function(las, neigh_radii, low_int_thrsh, uppr_int_thrsh, pt_den_req, BBPRthrsh_mat, ncpu) {
    .Call(`_lidR_C_Wing2015`, las, neigh_radii, low_int_thrsh, uppr_int_thrsh, pt_den_req, BBPRthrsh_mat, ncpu)
}

C_li2012 <- function(las, dt1, dt2, Zu, R, th_tree, radius) {
    .Call(`_lidR_C_li2012`, las, dt1, dt2, Zu, R, th_tree, radius)
}

C_pmf <- function(las, ws, th, filter) {
    .Call(`_lidR_C_pmf`, las, ws, th, filter)
}

C_rasterize <- function(las, layout, subcircle = 0, method = 1L) {
    .Call(`_lidR_C_rasterize`, las, layout, subcircle, method)
}

C_knnidw <- function(las, x, y, k, p, rmax, ncpu) {
    .Call(`_lidR_C_knnidw`, las, x, y, k, p, rmax, ncpu)
}

C_point_metrics <- function(las, k, r, nalloc, call, env, filter) {
    .Call(`_lidR_C_point_metrics`, las, k, r, nalloc, call, env, filter)
}

is_disable_point_metrics <- function() {
    .Call(`_lidR_is_disable_point_metrics`)
}

C_knn_distance <- function(las, k, cpu) {
    .Call(`_lidR_C_knn_distance`, las, k, cpu)
}

C_lasrangecorrection <- function(las, flightlines, Rs, f) {
    .Call(`_lidR_C_lasrangecorrection`, las, flightlines, Rs, f)
}

C_lasrange <- function(las, flightlines) {
    .Call(`_lidR_C_lasrange`, las, flightlines)
}

C_local_maximum <- function(las, ws, filter, ncpu) {
    .Call(`_lidR_C_local_maximum`, las, ws, filter, ncpu)
}

C_isolated_voxel <- function(las, res, isolated) {
    .Call(`_lidR_C_isolated_voxel`, las, res, isolated)
}

C_check_gpstime <- function(t, rn) {
    .Call(`_lidR_C_check_gpstime`, t, rn)
}

C_eigen_metrics <- function(las, k, r, coeffs, filter, ncpu) {
    .Call(`_lidR_C_eigen_metrics`, las, k, r, coeffs, filter, ncpu)
}

C_fast_eigen_decomposition <- function(las, k, r, coeffs, ncpu) {
    .Call(`_lidR_C_fast_eigen_decomposition`, las, k, r, coeffs, ncpu)
}

C_connected_component <- function(las, res, connectivity) {
    .Call(`_lidR_C_connected_component`, las, res, connectivity)
}

C_voxel_id <- function(las, res) {
    .Call(`_lidR_C_voxel_id`, las, res)
}

fast_table <- function(x, size = 5L) {
    .Call(`_lidR_fast_table`, x, size)
}

fast_countequal <- function(x, t) {
    .Call(`_lidR_fast_countequal`, x, t)
}

fast_countbelow <- function(x, t) {
    .Call(`_lidR_fast_countbelow`, x, t)
}

fast_quantization <- function(x, scale, offset) {
    invisible(.Call(`_lidR_fast_quantization`, x, scale, offset))
}

fast_countunquantized <- function(x, scale, offset) {
    .Call(`_lidR_fast_countunquantized`, x, scale, offset)
}

fast_countover <- function(x, t) {
    .Call(`_lidR_fast_countover`, x, t)
}

roundc <- function(x, digit = 0L) {
    .Call(`_lidR_roundc`, x, digit)
}

bitmerge <- function(u, v) {
    .Call(`_lidR_bitmerge`, u, v)
}

fast_eigen_values <- function(A) {
    .Call(`_lidR_fast_eigen_values`, A)
}

C_knn <- function(X, Y, x, y, k, ncpu) {
    .Call(`_lidR_C_knn`, X, Y, x, y, k, ncpu)
}

C_circle_lookup <- function(las, x, y, r) {
    .Call(`_lidR_C_circle_lookup`, las, x, y, r)
}

C_orectangle_lookup <- function(las, x, y, w, h, angle) {
    .Call(`_lidR_C_orectangle_lookup`, las, x, y, w, h, angle)
}

C_knn2d_lookup <- function(las, x, y, k) {
    .Call(`_lidR_C_knn2d_lookup`, las, x, y, k)
}

C_knn3d_lookup <- function(las, x, y, z, k) {
    .Call(`_lidR_C_knn3d_lookup`, las, x, y, z, k)
}

is_altrep <- function(x) {
    .Call(`_lidR_is_altrep`, x)
}

is_materialized <- function(x) {
    .Call(`_lidR_is_materialized`, x)
}

altrep_full_class <- function(x) {
    .Call(`_lidR_altrep_full_class`, x)
}

cpp_concaveman <- function(x, y, concavity, lengthThreshold, chull) {
    .Call(`_lidR_cpp_concaveman`, x, y, concavity, lengthThreshold, chull)
}

cpp_knn <- function(data, k, ncpu) {
    .Call(`_lidR_cpp_knn`, data, k, ncpu)
}

cpp_knnx <- function(data, query, k, ncpu) {
    .Call(`_lidR_cpp_knnx`, data, query, k, ncpu)
}

R_omp_get_max_threads <- function() {
    .Call(`_lidR_R_omp_get_max_threads`)
}

context("utils_geometry")

x = c(0,1,1,0,0.5,0.2,0.3,0.4,0.2,0.8,0.9)
y = c(0,0,1,1,0.1,0.5,0.8,0.3,0.1,0.4,0.7)

vertx = c(0,1,0)
verty = c(0,0,1)

las <- random_500_points
ctg <- random_2files_250points

test_that("convex hull works", {
  expect_equal(lidR:::convex_hull(x,y), data.frame(x = c(1,0,0,1,1), y = c(0,0,1,1,0)))
})

test_that("area_convex_hull works", {
  expect_equal(lidR:::area_convex_hull(x,y), 1)
  expect_equal(lidR:::area_convex_hull(vertx, verty), 0.5)
})

test_that("area works with a LAS*", {
  expect_equal(area(las), 10000, tolerance = 300)
  expect_equal(area(ctg), 40000, tolerance = 400)
  expect_equal(area(las@header), 1000, tolerance = 100)
})

test_that("area works with a 0 points", {
  las <- filter_poi(las, Z > 100)
  expect_equal(area(las), 0)
})

test_that("npoints works with a LAS*", {
  expect_equal(npoints(las), 500)
  expect_equal(npoints(ctg), 500)
  expect_equal(npoints(las@header), 500)
})

test_that("density works with a LAS*", {
  expect_equal(density(las), 0.05, tolerance = 0.002)
  expect_equal(density(ctg), 0.025, tolerance = 0.001)
  expect_equal(density(las@header), 0.05, tolerance = 0.001)
})


test_that("tsearch works", {
  x <- c(-1, -1, 1)
  y <- c(-1, 1, -1)
  P <- cbind(x,y)
  D <- matrix(c(1L, 2L, 3L), 1, 3)

  ## Should be in triangle #1
  ts <- lidR:::C_tsearch(D, P, cbind(-1,-1), 1L)
  expect_equal(ts, 1)

  ## Should be in triangle #1
  ts <- lidR:::C_tsearch(D, P, cbind(1,-1), 1L)
  expect_equal(ts, 1)

  ## Should be in triangle #1
  ts <- lidR:::C_tsearch(D, P, cbind(-1,1), 1L)
  expect_equal(ts, 1)

  ## Centroid
  ts <- lidR:::C_tsearch(D, P, cbind(-1/3,-1/3), 1L)
  expect_equal(ts, 1)

  ## Should be outside triangle #1, so should return NA
  ts <- lidR:::C_tsearch(D, P, cbind(1,1), 1)
  expect_true(is.na(ts))
})

test_that("tsearch passes computer precision tests", {
  x <- c(6.89, 7.15, 7.03)
  y <- c(7.76, 7.75, 8.35)
  P <- cbind(x,y)
  D <- matrix(c(1L, 2L, 3L), 1, 3)

  ts <- lidR:::C_tsearch(D, P, cbind(7.125, 7.875), 1)
  expect_equal(ts, 1)

  x <- c(278287.03, 278286.89, 278287.15)
  y <- c(602248.35, 602247.76, 602247.75)
  P <- cbind(x,y)

  D = matrix(c(1L,2L,3L), 1,3)
  ts <- lidR:::C_tsearch(D, P, cbind(278287.125, 602247.875), 1L)
  expect_equal(ts, 1L)

  tri = matrix(c(3L,2L,1L), 1,3)
  ts <- lidR:::C_tsearch(D, P, cbind(278287.125, 602247.875), 1L)
  expect_equal(ts, 1L)

  tri = matrix(c(2L,3L,1L), 1,3)
  ts <- lidR:::C_tsearch(D, P, cbind(278287.125, 602247.875), 1L)
  expect_equal(ts, 1L)

  tri = matrix(c(2L,1L,3L), 1,3)
  ts <- lidR:::C_tsearch(D, P, cbind(278287.125, 602247.875), 1L)
  expect_equal(ts, 1)

  tri = matrix(c(3L,1L,2L), 1,3)
  ts <- lidR:::C_tsearch(D, P, cbind(278287.125, 602247.875), 1L)
  expect_equal(ts, 1L)

  tri <- matrix(c(1L, 2L, 3L), 1, 3)
  ts <- lidR:::C_tsearch(D, P, cbind(278287.125, 602247.875), 1L)
  expect_equal(ts, 1L)
})

test_that("tsearch passes computer precision tests", {

  P = structure(c(
    488094.617850573, 488094.682850573, 488095.461850573,
    5189348.04894701, 5189348.79694701, 5189348.25294701),
    .Dim = 3:2, .Dimnames = list(NULL, c("X", "Y")))

  X = matrix(c(488094.75, 5189348.75), nrow = 1)

  D = matrix(3:1, nrow = 1)

  expect_equal(lidR:::tSearch(D,P, X, 1), 1L)

  D = matrix(c(2,3,1), nrow = 1)

  expect_equal(lidR:::tSearch(D,P, X, 1), 1L)

  D = matrix(1:3, nrow = 1)

  expect_equal(lidR:::tSearch(D,P, X, 1), 1L)

  P = structure(c(
    488100.025850573, 488099.663850573, 488099.916850573,
    5189337.35594701, 5189337.44894701, 5189336.86194701),
    .Dim = 3:2, .Dimnames = list(NULL, c("X", "Y")))

  X = structure(c(488099.75, 5189337.25), .Dim = 1:2)

  D = matrix(1:3, nrow = 1)

  expect_equal(lidR:::tSearch(D,P, X, 1), 1L)

  D = matrix(3:1, nrow = 1)

  expect_equal(lidR:::tSearch(D,P, X, 1), 1L)

  D = matrix(c(2,3,1), nrow = 1)

  expect_equal(lidR:::tSearch(D,P, X, 1), 1L)
})

test_that("tinfo works", {
  x <- c(0, 1, 0)
  y <- c(0, 0, 1)
  z <- c(1, 0, 0)
  X <- cbind(x, y, z)
  D <- matrix(c(1, 2, 3, 1), 1, 4)

  info = lidR:::C_tinfo(D, X)

  # normal vector is (1,1,1)
  n = c(info[,1], info[,2], info[,3]) %>% as.numeric
  expect_equal(n, c(1,1,1))

  # intercept is -1
  expect_equal(as.numeric(info[,4]), -1)

  # area is sqrt(3)/2
  expect_equal(as.numeric(info[,5]), sqrt(3)/2)

  # projected area is 0.5
  expect_equal(as.numeric(info[,6]), 0.5)

  # Max edge size is sqrt(2)
  expect_equal(as.numeric(info[,7]), sqrt(2))

  D <- matrix(c(1, 2, 3, 1), 1, 4)

  x = c(0,0,1)
  y = c(0,1,1)
  z = c(0,0,1)
  X = cbind(x,y,z)

  I = as.numeric(lidR:::C_tinfo(D, X))

  expect_equal(I[5], sqrt(2)/2)
  expect_equal(I[6], 1/2)

  x = c(0,0,1)
  y = c(0,1,0)
  z = c(0,1,1)
  X = cbind(x,y,z)

  I = as.numeric(lidR:::C_tinfo(D, X))

  expect_equal(I[5], sqrt(3)/2)
  expect_equal(I[6], 1/2)

  x = c(0,0,1)
  y = c(0,1,1)
  z = c(0,0,2)
  X = cbind(x,y,z)

  I = as.numeric(lidR:::C_tinfo(D, X))

  expect_equal(I[5], sqrt(5)/2)
  expect_equal(I[6], 1/2)
})

X = c(0,1,0,1,0,1)
Y = c(0,0,0.5,0.5,1,1)
Z = c(0,1,0.5,1.5,1,2)
I = 1:6
J = rep(TRUE, 6)
D = data.table::data.table(X,Y,Z,I,J)
las = LAS(D)

#skip_if(lidR:::is_disable_point_metrics(), "Point metrics disabled")

test_that("points_metrics works with a single metric (knn)", {

  m = point_metrics(las, ~mean(Z), k = 3L)

  expect_is(m, "data.table")
  expect_equal(dim(m), c(npoints(las), 2))
  expect_equal(m$V1, c(0.5, 1, 0.5, 1.5, 1, 1.5))
  expect_equal(names(m), c("pointID", "V1"))

  m = point_metrics(las, ~list(M = mean(Z)), k = 3L)

  expect_equal(dim(m), c(npoints(las), 2))
  expect_equal(m$M, c(0.5, 1, 0.5, 1.5, 1, 1.5))
  expect_equal(names(m), c("pointID", "M"))

  m = point_metrics(las, ~list(M = mean(Z)), k = 3L, xyz = TRUE)

  expect_equal(names(m), c("X", "Y", "Z", "M"))
})

test_that("points_metrics works with a single metric (shpere)", {

  m = point_metrics(las, ~length(Z), r = 0.8)

  expect_is(m, "data.table")
  expect_equal(dim(m), c(npoints(las), 2))
  expect_equal(m$V1, c(2,2,3,3,2,2))
  expect_equal(names(m), c("pointID", "V1"))

  m = point_metrics(las, ~list(M = mean(Z)), r = 0.8)

  expect_equal(dim(m), c(npoints(las), 2))
  expect_equal(m$M, c(0.25, 1.25, 0.5, 1.5, 0.75, 1.75))
  expect_equal(names(m), c("pointID", "M"))

  m = point_metrics(las, ~list(M = mean(Z)), r = 3L, xyz = TRUE)

  expect_equal(names(m), c("X", "Y", "Z", "M"))
})

test_that("points_metrics works with a single metric (knn + sphere)", {

  m1 = point_metrics(las, ~length(Z), k = 2, r = 0.8)
  m2 = point_metrics(las, ~length(Z), k = 3, r = 0.8)

  expect_equal(m1$V1, c(2,2,2,2,2,2))
  expect_equal(m2$V1, c(2,2,3,3,2,2))
})

test_that("points_metrics restpect the filter argument (knn)", {

  m = point_metrics(las, ~mean(Z), k = 3L, filter = ~I>2)

  expect_is(m, "data.table")
  expect_equal(dim(m), c(npoints(las)-2, 2))
  expect_equal(m$V1, c(1, 1.5, 1, 1.5))
})

test_that("points_metrics restpect the filter argument (shpere)", {

  m = point_metrics(las, ~mean(Z), r = 0.8, filter = ~I>2)

  expect_is(m, "data.table")
  expect_equal(dim(m), c(npoints(las)-2, 2))
  expect_equal(m$V1, c(0.75, 1.75, 0.75, 1.75))
})

test_that("points_metrics fails with 0 points after filter ", {

  expect_error(point_metrics(las, ~mean(Z), r = 0.8, filter = ~I>2000))
})

test_that("points_metrics works with a multiple metrics (knn)", {

  m = point_metrics(las, ~list(mean(Z), max(Z), Z[1]), k = 3L)

  expect_is(m, "data.table")
  expect_equal(dim(m), c(npoints(las), 4))
  expect_equal(m$V3, c(0, 1, 0.5, 1.5, 1, 2))

  m = point_metrics(las, ~list(A = mean(Z), B = max(Z), C = Z[1]), k = 3L, xyz = TRUE)

  expect_equal(names(m), c("X", "Y", "Z", "A", "B", "C"))
})

test_that("points_metrics works with a multiple metrics (sphere)", {

  m = point_metrics(las, ~list(mean(Z), max(Z), Z[1]), r = 0.8)

  expect_is(m, "data.table")
  expect_equal(dim(m), c(npoints(las), 4))
  expect_equal(m$V1, c(0.25, 1.25, 0.5, 1.5, 0.75, 1.75))

  m = point_metrics(las, ~list(A = mean(Z), B = max(Z), C = Z[1]), r = 0.8, xyz = TRUE)

  expect_equal(names(m), c("X", "Y", "Z", "A", "B", "C"))
})


test_that("points_metrics works with lidR metrics", {

  m = point_metrics(las, .stdmetrics_z, k = 3L, xyz = FALSE)

  expect_is(m, "data.table")
  expect_equal(dim(m), c(npoints(las), 37))

  m = point_metrics(las, .stdmetrics_z, r = 0.8, xyz = FALSE)

  expect_is(m, "data.table")
  expect_equal(dim(m), c(npoints(las), 37))
})

test_that("points_metrics works with nested function", {

  f <- function(x, y, z) {  return(max(c(x,y,z))) }
  g <- function(las) { point_metrics(las, ~f(X,Y,Z), k = 3L, xyz = FALSE) }
  m <- g(las)

  expect_is(m, "data.table")
  expect_equal(dim(m), c(npoints(las), 2))
  expect_equal(m$V1, c(1, 1.5, 1, 2, 1.5, 2))

  f <- function(x, y, z) {  return(max(c(x,y,z))) }
  g <- function(las) { point_metrics(las, ~f(X,Y,Z), r = 0.8, xyz = FALSE) }
  m <- g(las)

  expect_is(m, "data.table")
  expect_equal(dim(m), c(npoints(las), 2))
  expect_equal(m$V1, c(0.5, 1.5, 1, 2, 1, 2))
})

test_that("points_metrics realloc memory", {

  m1 <- point_metrics(las, ~list(mean(Z), length(Z)), r = 2, alloc = 1)
  m2 <- point_metrics(las, ~list(mean(Z), length(Z)), r = 2)
  expect_equal(m1, m2)
})

test_that("points_metrics fails nicely if error in func", {

  f <- function(x) {  stop("Dummy error") }

  expect_error(point_metrics(las, ~f(Z), k = 3L), "Dummy")
  expect_error(point_metrics(las, ~f(Z), r = 3L), "Dummy")
})

test_that("points_metrics fails with non atomic output", {

  expect_error(point_metrics(las, ~c(1,2), k = 3L))
  expect_error(point_metrics(las, ~c(1,2), r = 0.8))
})

test_that("points_metrics fails", {

  expect_error(point_metrics(las, ~mean(Z)), "'k' or 'r' is missing")

  las@data$test = letters[1:6]
  expect_error(point_metrics(las, ~mean(Z), k = 3), "Incompatible type encountered")
})


test_that("points_metrics return references on coordinates", {

  m = point_metrics(las, ~list(mean(Z), max(Z), Z[1]), k = 3L, xyz = TRUE)

  expect_reference(m$X, las$X)
  expect_reference(m$Y, las$Y)
  expect_reference(m$Z, las$Z)
})

context("spatial index")

LASfile <- example_las_path

test_that("read*LAS work", {

  las = readLAS(LASfile)

  expect_equal(las@index$sensor, lidR:::ALSLAS)
  expect_equal(las@index$index, 0L)

  las = readALSLAS(LASfile)

  expect_equal(las@index$sensor, lidR:::ALSLAS)
  expect_equal(las@index$index, 0L)

  las = readTLSLAS(LASfile)

  expect_equal(las@index$sensor, lidR:::TLSLAS)
  expect_equal(las@index$index, 0L)
})

test_that("read*LAScatalog work", {

  las = readLAScatalog(LASfile)

  expect_equal(las@index$sensor, lidR:::ALSLAS)
  expect_equal(las@index$index, 0L)

  las = readALSLAScatalog(LASfile)

  expect_equal(las@index$sensor, lidR:::ALSLAS)
  expect_equal(las@index$index, 0L)

  las = readTLSLAScatalog(LASfile)

  expect_equal(las@index$sensor, lidR:::TLSLAS)
  expect_equal(las@index$index, 0L)
})

test_that("index works", {

  las = readLAS(LASfile)

  expect_equal(index(las), 0L)

  index(las) <- 2

  expect_equal(index(las), 2L)

  index(las) <- "quadtree"

  expect_equal(index(las), 3L)
  expect_equal(index(las, h = TRUE), "quadtree")

  expect_error(index(las) <- "plop")
  expect_error(index(las) <- 8)
})

test_that("sensor works", {

  las = readLAS(LASfile)

  expect_equal(sensor(las),lidR:::ALSLAS)

  sensor(las) <- lidR:::TLSLAS

  expect_equal(sensor(las), lidR:::TLSLAS)

  sensor(las) <- "tls"

  expect_equal(sensor(las), lidR:::TLSLAS)
  expect_equal(sensor(las, h = TRUE), "TLS")

  expect_error(sensor(las) <- "plop")
  expect_error(sensor(las) <- 8)
})

test_that("knn works in 2D", {

  X = c(0, 1, 1, 0, 0.5)
  Y = c(0, 0, 1, 1, 0.5)
  Z = c(0, 1, 1, 0, 0.5)

  x = 0.1
  y = 0.1

  nn = lidR:::C_knn(X,Y, x, y, 1, 1)

  expect_equal(nn$nn.idx, matrix(1))
  expect_equal(nn$nn.dist, matrix(sqrt(2)/10))

  nn = lidR:::C_knn(X,Y, x, y, 2, 1)

  expect_equal(nn$nn.idx, matrix(c(1,5), ncol = 2))
  expect_equal(nn$nn.dist, matrix(c(matrix(sqrt(2)/10), 4*sqrt(2)/10), ncol = 2))
})

test_that("Spatial index knn works even with npoints < k", {

  X = c(0, 1, 1, 0, 0.5)
  Y = c(0, 0, 1, 1, 0.5)
  Z = c(0, 1, 1, 0, 0.5)

  nn = lidR:::C_knn(X,Y, c(0,2), c(0.8,2), 8, 1)

  expect <- matrix(c(4,5,1,3,2,0,0,0,3,5,4,2,1,0,0,0), ncol = 8, byrow = T)

  expect_equal(nn$nn.idx, expect)
})

test_that("knnidw works", {

  X = c(0, 1, 1, 0, 0.5)
  Y = c(0, 0, 1, 1, 0.5)
  Z = c(0, 1, 1, 0, 0.5)
  D = data.frame(X,Y,Z)
  las = LAS(D)

  val1 = lidR:::C_knnidw(las, 0.1, 0.1, 1, 1, 50, 1)
  val2 = lidR:::C_knnidw(las, 0.2, 0.2, 1, 1, 50, 1)
  val3 = lidR:::C_knnidw(las, 0.3, 0.3, 1, 1, 50, 1)
  val4 = lidR:::C_knnidw(las, 0.4, 0.5, 1, 1, 50, 1)

  expect_equal(val1, 0)
  expect_equal(val2, 0)
  expect_equal(val3, 0.5)
  expect_equal(val4, 0.5)

  val1 = lidR:::C_knnidw(las, 0.25, 0.25, 2, 1, 50, 1)
  val2 = lidR:::C_knnidw(las, 0.2, 0.2, 2, 1, 50, 1)

  expect_equal(val1, 0.25)
  expect_equal(val2, 0.2)
})

X = c(1, 2, 3, 10, 11, 12, 1, 2, 3)
Y = c(3, 2, 1, 5, 7, 6, 8, 9, 7)
Z = c(20, 1, 1, 5, 7, 6, 1, 1, 4)
D <- data.frame(X,Y,Z)
las = LAS(D)

test_that("circle lookup works for each spatial index", {

  for(index in 0:3)
  {
    index(las) <- index

    id = lidR:::C_circle_lookup(las, 2,2,1)
    expect_equal(id, 2)

    id = lidR:::C_circle_lookup(las, 1.9,3,1)
    expect_equal(id, 1)

    id = lidR:::C_circle_lookup(las, 2,2,2)
    id = sort(id)
    expect_equal(id, 1:3)

    id = lidR:::C_circle_lookup(las, 6,6,2)
    expect_equal(length(id), 0)
  }
})

test_that("oriented rectangle lookup works for each spatial index", {

  for(index in 0:3)
  {
    index(las) <- index

    id = lidR:::C_orectangle_lookup(las, 2,2,3,3, -pi/4)
    id = sort(id)
    expect_equal(id, 1:3)

    id = lidR:::C_orectangle_lookup(las, 10, 6, 5, 2, 0)
    expect_equal(id, 6)

    id = lidR:::C_orectangle_lookup(las, 10, 6, 5, 2, pi/8)
    id = sort(id)
    expect_equal(id, 4:6)
  }
})

test_that("knn in 2d works for each spatial indexes", {

  for(index in 0:3)
  {
    index(las) <- index

    id = lidR:::C_knn2d_lookup(las, 2,2,1)
    expect_equal(id, 2)

    id = lidR:::C_knn2d_lookup(las, 2,2.1,4)
    expect_equal(id, c(2,1,3,9))

    id = lidR:::C_knn2d_lookup(las, 6,6,2)
    expect_equal(id, c(9, 4))
  }
})

test_that("knn in 3d works for each spatial indexes", {

  for(index in 0:3)
  {
    index(las) <- index

    id = lidR:::C_knn3d_lookup(las, 2,2,0,1)
    expect_equal(id, 2)

    id = lidR:::C_knn3d_lookup(las, 2,2,2,4)
    expect_equal(id, c(2,3,9,7))

    id = lidR:::C_knn3d_lookup(las, 6,6,50,2)
    expect_equal(id, c(1, 5))
  }
})

test_that("Spatial index fails with inccorect code", {

    las@index$index <- 12
    expect_error(lidR:::C_circle_lookup(las, 2,2,1), "index code inccorect")
})

test_that("Spatial indexes work with more points (coverage)", {

  las = lidR:::generate_las(2000)

  x = runif(1, 10, 900)
  y = runif(1, 10, 900)
  z = runif(1, 1, 19)

  u = vector("list", 3)
  for(index in 0:3)
  {
    index(las) <- index
    id = lidR:::C_knn3d_lookup(las, x,y,z,10)
    u[[index+1]] = id
  }

  expect_true(all(sapply(u, identical, u[[1]])))

  u = vector("list", 3)
  for(index in 0:3)
  {
    index(las) <- index
    id = lidR:::C_circle_lookup(las, x, y, 5)
    id = sort(id)
    u[[index+1]] = id
  }

  expect_true(all(sapply(u, identical, u[[1]])))
})

test_that("Spatial indexes work with 0 point", {

  las = lidR:::generate_las(0)

  x = runif(1, 10, 900)
  y = runif(1, 10, 900)
  z = runif(1, 1, 19)

  u = vector("list", 3)
  for(index in 0:3)
  {
    index(las) <- index
    id = lidR:::C_knn3d_lookup(las, x, y, z,10)
    u[[index+1]] = id
  }

  expect_true(all(sapply(u, identical, integer(0))))

  u = vector("list", 3)
  for(index in 0:3)
  {
    index(las) <- index
    id = lidR:::C_circle_lookup(las, x, y, 5)
    id = sort(id)
    u[[index+1]] = id
  }

  expect_true(all(sapply(u, identical, integer(0))))
})

test_that("Spatial indexes work with 1 point", {

  las = LAS(data.frame(X = 10, Y = 10, Z = 10))

  x = runif(1, 10, 11)
  y = runif(1, 10, 11)
  z = runif(1, 1, 19)

  u = vector("list", 3)
  for(index in 0:3)
  {
    index(las) <- index
    id = lidR:::C_knn3d_lookup(las, x, y, z, 10)
    u[[index+1]] = id
  }

  expect_true(all(sapply(u, identical, 1L)))

  u = vector("list", 3)
  for(index in 0:3)
  {
    index(las) <- index
    id = lidR:::C_circle_lookup(las, x, y, 5)
    id = sort(id)
    u[[index+1]] = id
  }

  expect_true(all(sapply(u, identical, 1L)))
})

test_that("Spatial indexes work no point to find", {

  las = lidR:::generate_las(10)

  x = -100
  y = -100
  z = 0

  u = vector("list", 3)
  for(index in 0:3)
  {
    index(las) <- index
    id = lidR:::C_circle_lookup(las, x, y, 5)
    id = sort(id)
    u[[index+1]] = id
  }

  expect_true(all(sapply(u, identical, integer(0))))
})

test_that("Forcing autoindex work", {

  las = readLAS(LASfile)

  lidR:::force_autoindex(las) <- lidR:::LIDRGRIDPARTITION

  expect_equal(index(las), 1L)

  lidR:::force_autoindex(las) <- lidR:::LIDRQUADTREE

  expect_equal(index(las), 1L)

  index(las) <- lidR:::LIDRAUTOINDEX

  lidR:::force_autoindex(las) <- lidR:::LIDRQUADTREE

  expect_equal(index(las), 3L)
})

#' Convert full waveform data into a regular point cloud
#'
#' Full waveform can be difficult to manipulate and visualize in R. This function converts
#' a LAS object with full waveform data into a regular point cloud. Each waveform record
#' becomes a point with XYZ coordinates and an amplitude (units: volts) and an ID that records
#' each original pulse. Notice that this has the effect of drastically inflating the size of the
#' object in memory, which is likely already very large
#'
#' @param las An object of class LAS with full waveform data
#' @return An object of class LAS 1.2 format 0 with one point per records
#'
#' @template section-fwf
#' @export
#' @examples
#' \dontrun{
#' LASfile <- system.file("extdata", "fwf.laz", package="rlas")
#' fwf <- readLAS(LASfile)
#' las <- interpret_waveform(fwf)
#' x <- plot(fwf, size = 3, pal = "red")
#' plot(las, color = "Amplitude", bg = "white", add = x, size = 2)
#' }
interpret_waveform <- function(las)
{
  if (utils::packageVersion("rlas") <= package_version("1.4.0"))
    stop("Package rlas > 1.4.0 is required for this function")

  # The temporal spacing is in picosecond. The light travels 300 mm per nanosecond. For a temporal
  # spacing of 1 ns two points emitted at one degree off nadir (worst case considered) are separated
  # by 300*sin(pi/180) = 5.2 mm. We need a resolution of 1 mm to store distinguishable values. With
  # a scale factor of 0.001 we are good. We use a 10th of this to be sure.
  temporal_spacing <- vlr(las)[["Full WaveForm Description"]][["Full WaveForm"]][["Temporal Spacing"]]
  one_picosecond_light <- 0.0003 # 0.3 mm
  distance_between_two_samples <- temporal_spacing * one_picosecond_light
  minimum_distance_at_one_degree_off_nadir <- distance_between_two_samples * pi/180
  s <- 10^(0:7)
  valid <- sort(c(1/s, 0.5/s))
  i <- which(valid < minimum_distance_at_one_degree_off_nadir)
  scale_factor <- valid[max(i)]/10

  header <- as.list(las@header)
  data <- las@data
  fwf <- rlas::fwf_interpreter(header, data)
  fwf <- data.table::rbindlist(fwf)
  fwfheader <- rlas::header_create(fwf)
  fwfheader[["X scale factor"]] <- scale_factor
  fwfheader[["Y scale factor"]] <- scale_factor
  fwfheader[["Z scale factor"]] <- scale_factor
  fwf <- LAS(fwf, fwfheader, crs = st_crs(las), check = FALSE)
  las_quantize(fwf, by_reference = TRUE)
  fwf <- las_update(fwf)
  fwf <- add_lasattribute(fwf, name = "Amplitude", desc = "Full waveform amplitude in volts")
  return(fwf)
}

#' Computes the polygon that encloses the points
#'
#' Computes the polygon that encloses the points. It reads all the files one by one and computes a
#' concave hull using the \link{st_concave_hull} function. When all the hulls are computed it updates the
#' LAScatalog to set the true polygons instead of the bounding boxes.
#'
#' @section Non-supported LAScatalog options:
#' The options `select`, `output files`, `chunk size`, `chunk buffer`, `chunk alignment` are not
#' supported and not respected in  `catalog_boundaries*` because the function must always process by
#' file, without buffer and knows which attributes to load.
#'
#' @param ctg A LAScatalog
#' @param ... propagated to \link{st_concave_hull}
#'
#' @return A LAScatalog with true boundaries
#' @export
#' @examples
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' ctg <- readLAScatalog(LASfile, filter = "-drop_z_below 0.5")
#' ctg2 <- catalog_boundaries(ctg, concavity = 1, length_threshold = 15)
#' plot(ctg)
#' plot(ctg2, add = TRUE)
catalog_boundaries = function(ctg, ...)
{
  if (is_lascatalog_v3(ctg)) ctg <- lascatalog_v3_repair(ctg)

  opt_chunk_buffer(ctg) <- 0
  opt_chunk_size(ctg) <- 0
  opt_select(ctg) <- "xyz"
  opt_output_files(ctg) <- ""
  opt_independent_files(ctg) <- TRUE
  res <- catalog_map(ctg, st_concave_hull, ...)

  sf::st_geometry(ctg@data) <- res
  attr(ctg, "trueshape") <- TRUE
  return(ctg)
}

#' Split a catalog into spatially clustered subsets
#'
#' This function groups elements of a catalog into spatial clusters based on
#' contiguity of the tiles. The original catalog is then split into `n` catalogs
#'
#' @param ctg A LAScatalog
#' @param buffer Numeric. Buffer distance applied to tile geometry before clustering. Defaults to 1.
#'
#' @return A list of LAScatalog
#'
#' @export
catalog_split_clusters = function(ctg, buffer = 1)
{
  sfctg = sf::st_as_sf(ctg)
  sfctg = sf::st_buffer(sfctg, buffer)
  clustered = sf::st_union(sfctg)
  clustered = sf::st_cast(clustered, "POLYGON")
  idx = sf::st_intersects(clustered, sfctg)
  subctg = lapply(idx, function(i) ctg[i,])
  subctg
}

catalog_laxindex = function(ctg)
{
  stopifnot(is(ctg, "LAScatalog"))

  opt_chunk_size(ctg)   <- 0
  opt_chunk_buffer(ctg) <- 0
  opt_wall_to_wall(ctg) <- FALSE
  opt_output_files(ctg) <- ""

  create_lax_file = function(cluster) {
    rlas::writelax(cluster@files)
    return(0)
  }

  options <- list(need_buffer = FALSE, drop_null = FALSE)

  catalog_apply(ctg, create_lax_file,.options = options)
  return(invisible())
}

context("catalog_laxindex")

test_that("catalog_laxindex works", {
  skip_on_cran() # ASAN/USBAN to fix in rlas

  ctg <- random_2files_250points
  expect_error(lidR:::catalog_laxindex(ctg), NA)
  expect_true(is.indexed(ctg))
})

#' @rdname engine
#' @export
#' @param .CHUNKS list. list of LAScluster
#' @param .FUN function. function that respects a template (see \link{catalog_apply})
#' @param .PROCESSOPT list. Processing option
#' @param .OUTPUTOPT list. Output option
#' @param .GLOBALS list. Force export of some object in workers
#' @param .AUTOREAD bool. Enable autoread
#' @param .AUTOCROP bool. Enable autocrop
#' @param ... parameters of .FUN
engine_apply = function(.CHUNKS, .FUN, .PROCESSOPT, .OUTPUTOPT, .GLOBALS = NULL, .AUTOREAD = FALSE, .AUTOCROP = FALSE, ...)
{
  # Parse ellipsis
  params  <- list(...)

  # Retrieve the names of the param in user-defined function
  first_p <- names(formals(.FUN))[1]
  second_p <- names(formals(.FUN))[2]

  # Initialize output
  nclusters  <- length(.CHUNKS)
  futures    <- vector("list", nclusters)
  output     <- vector("list", nclusters)
  writemode  <- .CHUNKS[[1]]@save != ""
  drivers    <- .OUTPUTOPT$drivers

  # Initialize progress bar
  prgrss     <- .PROCESSOPT$progress
  abort      <- .PROCESSOPT$stop_early
  states     <- rep(CHUNK_WAINTING, nclusters)
  messages   <- rep("", nclusters)
  pb         <- engine_progress_bar(nclusters, prgrss)
  percentage <- 0

  raster.default <- getOption("lidR.raster.default")
  on.exit(engin_close_pb(pb))

  # Disable OpenMP? The different between  LIDRTHREADS$n and LIDRTHREADS$input
  # is that n is the corrected number of workers. e.g on a quadcore set_lidr_threads(0)
  # set 4 and set_lidr_threads(12) set 4 too. On the contrary input will be 0 and 12 respectively
  # Consequently so on remote machines with different capabilities than the master worker
  # the information is not overwritten by the master worker.
  threads = if (must_disable_openmp()) 1L else if (isFALSE(getOption("lidR.check.nested.parallelism"))) LIDRTHREADS$input else LIDRTHREADS$n

  # ==== PROCESSING ====
  use_future <- engine_use_future()
  if (!use_future) verbose("Future is disabled")
  verbose(glue::glue("Start processing {nclusters} chunks..."))

  future <- if (!use_future)  no_future else future::future
  value  <- if (!use_future) function(x) { attr(x, "state") <- NULL ; x } else function(x) { suppressWarnings(future::value(x)) }

  for (i in seq_along(.CHUNKS))
  {
    chunk <- .CHUNKS[[i]]
    params[[first_p]] <- chunk
    save <- chunk@save

    states[i] <- CHUNK_PROCESSING
    engine_update_progress(pb, .CHUNKS[[i]], states[i], percentage, i)

    # Asynchronous computation of .FUN on the chunk
    futures[[i]] <- future(
    {
      setThreads(threads)
      options(lidR.progress = FALSE)
      options(lidR.verbose = FALSE)
      options(lidR.raster.default = raster.default)
      y <- NULL

      # Regular behaviour before v3.0.0
      if (.AUTOREAD == FALSE)
      {
        y <- do.call(.FUN, params)
      }
      # New option autoread from v3.0.0 (was not a good idea but it exists)
      else if (.AUTOREAD == TRUE & .AUTOCROP == FALSE)
      {
        bbox <- st_bbox(chunk)
        las  <- readLAS(chunk)
        y    <- NULL
        if (!is.empty(las))
        {
          params[[first_p]] <- las
          params[[second_p]] <- bbox
          y <- do.call(.FUN, params)
        }
      }
      # New option autocrop = TRUE from v4.0.0 intends to make simplify the logic of catalog apply.
      # it is not documented and is only used in apply.LAScatalog
      else if (.AUTOREAD == TRUE & .AUTOCROP == TRUE)
      {
        bbox <- st_bbox(chunk)
        las  <- readLAS(chunk)
        y    <- NULL
        if (!is.empty(las))
        {
          params[[first_p]] <- las
          y <- do.call(.FUN, params)
        }

        if (!is.null(y))
          y <- engine_crop(y, bbox)
      }
      else
        stop("Internal error in engine_apply: AUTO options are invalid. Please report.", call. = FALSE)

      if (is.null(y)) y <- NULL
      if (!is.null(y) && writemode) y <- engine_write(y, save, drivers)

      # Fix #523
      if (is(y, "SpatRaster") | is(y, "SpatVector")) y <- terra::wrap(y)

      y
    }, substitute = TRUE, globals = structure(TRUE, add = .GLOBALS), seed = TRUE)

    # Evaluation of the state of the futures
    for (j in 1:i)
    {
      # Skip chunks that were already evaluated and for which the state is known
      if (states[j] != CHUNK_PROCESSING) next

      # Evaluate the state of the chunk
      state       <- engine_eval_state(futures[[j]])
      states[j]   <- state[["state"]]
      messages[j] <- state[["msg"]]

      # The state is unchanged: the chunk is still processing
      if (states[j] == CHUNK_PROCESSING) next # nocov

      # The state changed: the chunk was processed. Update the progress
      percentage <-  engine_compute_progress(states)
      engine_update_progress(pb, .CHUNKS[[j]], states[j], percentage, j)

      # The state is ERROR: abort the process nicely
      if (states[j] == CHUNK_ERROR && abort)
      {
        # If it fails in first chunk it is likely to be an error in code.
        # Stop and display the error message
        if (j == 1)
        {
          stop(messages[j], call. = FALSE)
        }
        # If it fails somewhere else it is likely to be an error in a specific point cloud.
        # Return a partial output and display the logs
        else
        {
          engine_save_logs(.CHUNKS[[j]], j)
          message(messages[j])
          return(output)
        }
      }

      if (states[j] == CHUNK_ERROR && !abort) {
        output[[j]] <- NULL
        next
      }

      # The state is NULL: do nothing
      if (states[j] == CHUNK_NULL) next

      # The state is OK or WARNING: get the value
      res <- value(futures[[j]])

      # Fix #523
      if (is(res, "PackedSpatRaster"))
        res <- terra::rast(res)
      if (is(res, "PackedSpatVector"))
        res <- terra::vect(res)

      output[[j]] <- res
    }
  }

  # ==== PROGRESS ENDING ====

  # Because of asynchronous computation, the loop may be ended
  # but not the computations. Wait until the end & check states.
  # no cov because tested with a single core on CRAN

  # nocov start
  while (any(states == CHUNK_PROCESSING))
  {
    i <- which(states == CHUNK_PROCESSING)

    for (j in i)
    {
      if (states[j] != CHUNK_PROCESSING) next

      state       <- engine_eval_state(futures[[j]])
      states[j]   <- state[["state"]]
      messages[j] <- state[["msg"]]

      if (states[j] == CHUNK_PROCESSING) next

      percentage <-  engine_compute_progress(states)
      engine_update_progress(pb, .CHUNKS[[j]], states[j], percentage, j)

      if (states[j] == CHUNK_ERROR && abort)
      {
        if (j == 1)
        {
          stop(messages[j], call. = FALSE)
        }
        else
        {
          engine_save_logs(.CHUNKS[[j]], j)
          message(messages[j])
          return(output)
        }
      }

      if (states[j] == CHUNK_ERROR && !abort) {
        output[[j]] <- NULL
        next
      }

      if (states[j] == CHUNK_NULL) next

      res <- value(futures[[j]])

      # Fix #523
      if (is(res, "PackedSpatRaster"))
        res <- terra::rast(res)
      if (is(res, "PackedSpatVector"))
        res <- terra::vect(res)

      output[[j]] <- res
      attr(output[[j]], "state") <- NULL ;
    }

    Sys.sleep(0.5)
  }

  return(output)
}
# nocov end

engine_eval_state <- function(future)
{

  # If future is not a future it means that we are in "no future" mode
  # The result has already been evaluated and the state is stored in an attribute
  if (!is(future, "Future")) return(attr(future, "state"))

  # future is a future, we must check if it is resolved. If it is resolved we evaluate
  # the state of the result

  # Fix #414
  sink(paste0(tempdir(), "/dev_null"))
  on.exit(sink(NULL))

  cluster_state <- list(state = CHUNK_PROCESSING, msg = "")

  if (future::resolved(future))
  {
    cluster_state <- list(state = CHUNK_OK, msg = "")

    tryCatch(
    {
      withCallingHandlers(
      {
        y <- future::value(future)
        if (is.null(y)) cluster_state <- list(state = CHUNK_NULL, msg = "")
      },
      warning = function(w)
      {
        cluster_state <<- list(state = CHUNK_WARNING, msg = w["message"])
      })
    },
    error = function(e)
    {
      cluster_state <<- list(state = CHUNK_ERROR, msg = e["message"])
    })
  }

  return(cluster_state)
}

engine_progress_bar <- function(n, prgss = FALSE)
{
  pb <- NULL

  if (!prgss)
    return(pb)

  # nocov start
  if (!interactive())
    return(n)

  if (requireNamespace("progress", quietly = TRUE))
    pb <- progress::progress_bar$new(format = glue::glue("Processing [:bar] :percent (:current/:total) eta: :eta"), total = n, clear = FALSE)
  else
    pb <- utils::txtProgressBar(min = 0, max = 1, style = 3)

  graphics::legend("topright", title = "Colors", legend = c("Processing", "Empty","Ok","Warning", "Error"), fill = c("cornflowerblue", "gray","green3", "orange", "red"), cex = 0.8)

  return(pb) # nocov end
}

engine_update_progress <- function(pb, cluster, state, p, j)
{
  if (is.null(pb))
    return(invisible(NULL))

  # nocov start
  if (state == CHUNK_OK) { col <- "green3" ; sym <- "\u2713" }
  else if (state == CHUNK_NULL) { col <- "gray" ; sym <- "\u2205" }
  else if (state == CHUNK_WARNING) { col <- "orange" ; sym <- "\u26A0" }
  else if (state == CHUNK_ERROR) { col <- "red" ; sym <- "\u2717" }
  else if (state == CHUNK_PROCESSING) { col <- "cornflowerblue" ; sym <- "\u21BB" }

  if (!interactive())
  {
    if (state == CHUNK_PROCESSING)
      return(invisible())

    cat(glue::glue("Chunk {j} of {pb} ({round(p*100,1)}%): state {sym}"))
    cat("\n")
    return(invisible())
  }

  if (cluster@shape == LIDRRECTANGLE) {
    if (cluster@wkt == "") {
      bbox <- cluster@bbox
      graphics::rect(bbox[1], bbox[2], bbox[3], bbox[4], border = "black", col = col)
    } else {
      poly <- sf::st_as_sfc(cluster@wkt)
      plot(poly, add = TRUE, col = col)
    }
  } else if (cluster@shape == LIDRCIRCLE) {
    center <- cluster@center
    width  <- cluster@width
    theta <- seq(0, 2 * pi, length = 32)
    graphics::polygon(x = center$x + width/2 * cos(theta), y = center$y + width/2 * sin(theta), col = col)
  }

  if (is(pb, "txtProgressBar"))
    utils::setTxtProgressBar(pb, p)
  else
    pb$update(p)

  return(invisible(NULL)) # nocov end
}

# nocov start
engin_close_pb <- function(pb)
{
  if (is.null(pb))
    return(invisible(NULL))

  if (!interactive())
    return(invisible(NULL))

  if (is(pb, "txtProgressBar"))
    close(pb)
  else if (!pb$finished)
    pb$terminate()

  return(invisible(NULL))
}
# nocov end

engine_save_logs <- function(cluster, index)
{
  log <- glue::glue("{tempdir()}/chunk{index}.rds")
  saveRDS(cluster, log)
  message(glue::glue("\nAn error occurred when processing the chunk {index}. Try to load this chunk with:\n chunk <- readRDS(\"{log}\")\n las <- readLAS(chunk)"))
}

engine_compute_progress = function(states)
{
  sum(states != CHUNK_WAINTING & states != CHUNK_PROCESSING)/length(states)
}

engine_use_future <- function()
{
  if (isTRUE(getOption("lidR.debug")))
    return(FALSE)

  if (isTRUE(getOption("lidR.no.future")))
    return(FALSE)

  b <- requireNamespace("future", quietly = TRUE)
  return(b)
}

no_future <- function(expr, ...)
{
  cluster_state <- list(state = CHUNK_OK, msg = "")

  y <- tryCatch(
  {
    withCallingHandlers(
    {
      y <- eval(expr)
      if (is.null(y)) cluster_state <- list(state = CHUNK_NULL, msg = "")
      y
    },
    warning = function(w)
    {
      cluster_state <<- list(state = CHUNK_WARNING, msg = w["message"])
    })
  },
  error = function(e)
  {
    cluster_state <<- list(state = CHUNK_ERROR, msg = e["message"])
  })

  if (is.null(y)) y <- list(NULL)
  attr(y, "state") <- cluster_state
  return(y)
}

#' Point-based metrics
#'
#' Computes a series of user-defined descriptive statistics for a LiDAR dataset for each point based
#' on its k-nearest neighbours or its sphere neighbourhood.
#'
#' When the neighbourhood is knn the user-defined function is fed with the current
#' processed point and its k-1 neighbours. The current point being considered as
#' the 1-neighbour with a distance 0 to the reference point. The points are ordered
#' by distance to the central point. When the neighbourhood is a sphere the processed
#' point is also included in the query but points are coming in a random order. `point_eigenmetrics`
#' computes the eigenvalues of the covariance matrix and computes associated metrics following
#' Lucas et al, 2019 (see references). It is equivalent to `point_metrics(las, .stdshapemetrics)`
#' but much faster because it is optimized and parallelized internally.
#'
#' @section Performances:
#' It is important to bear in mind that this function is very fast for the feature it provides i.e.
#' mapping a user-defined function at the point level using optimized memory management. However, it
#' is still computationally demanding.\cr\cr
#' To help users to get an idea of how computationally demanding this function is, let's compare it to
#' \link{pixel_metrics}. Assuming we want to apply `mean(Z)` on a 1 km² tile with 1 point/m²
#' with a resolution of 20 m (400 m² cells), then the function `mean` is called roughly 2500
#' times (once  per cell). On the contrary, with `point_metrics`, `mean` is called 1000000
#' times (once per point). So the function is expected to be more than 400 times slower in this specific
#' case (but it does not provide the same feature).\cr\cr
#' This is why the user-defined function is expected to be well-optimized, otherwise it might drastically
#' slow down this already heavy computation. See examples.
#'
#' @param las An object of class LAS
#' @param func formula. An expression to be applied to each point neighbourhood (see also
#' \link{template_metrics}).
#' @param k,r integer and numeric respectively for k-nearest neighbours and radius of the neighborhood
#' sphere. If k is given and r is missing, computes with the knn, if r is given and k is missing
#' computes with a sphere neighborhood, if k and r are given computes with the knn and a limit on the
#' search distance.
#' @param xyz logical. Coordinates of each point are returned in addition to each metric. Otherwise an
#' ID referring to each point.
#' @param coeffs logical. Principal component coefficients are returned
#' @param filter formula of logical predicates. Enables the function to run only on points of interest
#' in an optimized way. See examples.
#' @param ... unused.
#' @param metrics logical. Compute metrics or not
#'
#' @examples
#' \dontrun{
#' LASfile <- system.file("extdata", "Topography.laz", package="lidR")
#'
#' # Read only 0.5 points/m^2 for the purposes of this example
#' las = readLAS(LASfile, filter = "-thin_with_grid 2")
#'
#' # Computes the eigenvalues of the covariance matrix of the neighbouring
#' # points and applies a test on these values. This function simulates the
#' # 'shp_plane()' algorithm from 'segment_shape()'
#' plane_metrics1 = function(x,y,z, th1 = 25, th2 = 6) {
#'   xyz <- cbind(x,y,z)
#'   cov_m <- cov(xyz)
#'   eigen_m <- eigen(cov_m)$value
#'   is_planar <- eigen_m[2] > (th1*eigen_m[3]) && (th2*eigen_m[2]) > eigen_m[1]
#'   return(list(planar = is_planar))
#' }
#'
#' # Apply a user-defined function
#' M <- point_metrics(las, ~plane_metrics1(X,Y,Z), k = 25)
#' #> Computed in 6.3 seconds
#'
#' # We can verify that it returns the same as 'shp_plane'
#' las <- segment_shapes(las, shp_plane(k = 25), "planar")
#' #> Computed in 0.1 seconds
#'
#' all.equal(M$planar, las$planar)
#'
#' # At this stage we can be clever and find that the bottleneck is
#' # the eigenvalue computation. Let's write a C++ version of it with
#' # Rcpp and RcppArmadillo
#' Rcpp::sourceCpp(code = "
#' #include <RcppArmadillo.h>
#' // [[Rcpp::depends(RcppArmadillo)]]
#'
#' // [[Rcpp::export]]
#' SEXP eigen_values(arma::mat A) {
#' arma::mat coeff;
#' arma::mat score;
#' arma::vec latent;
#' arma::princomp(coeff, score, latent, A);
#' return(Rcpp::wrap(latent));
#' }")
#'
#' plane_metrics2 = function(x,y,z, th1 = 25, th2 = 6) {
#'   xyz <- cbind(x,y,z)
#'   eigen_m <- eigen_values(xyz)
#'   is_planar <- eigen_m[2] > (th1*eigen_m[3]) && (th2*eigen_m[2]) > eigen_m[1]
#'   return(list(planar = is_planar))
#' }
#'
#' M <- point_metrics(las, ~plane_metrics2(X,Y,Z), k = 25)
#' #> Computed in 0.5 seconds
#'
#' all.equal(M$planar, las$planar)
#' # Here we can see that the optimized version is way better but is still 5-times slower
#' # because of the overhead of calling R functions and switching back and forth from R to C++.
#'
#' M <- point_eigenvalues(las, k = 25)
#' is_planar = M$eigen_medium > (25*M$eigen_smallest) & (6*M$eigen_medium) > M$eigen_largest
#'
#' # Use the filter argument to process only first returns
#' M1 <- point_metrics(las, ~plane_metrics2(X,Y,Z), k = 25, filter = ~ReturnNumber == 1)
#' dim(M1) # 13894 instead of 17182 previously.
#' }
#' @export
#' @family metrics
#' @md
point_metrics <- function(las, func, k, r,  xyz = FALSE, filter = NULL, ...) {

  #if (is_disable_point_metrics())
  #  stop("Because of a new CRAN policy that affects numerous packages, we temporarily removed this function to give us time to find a workaround and prevent the package from being removed from CRAN.
#To get a version of the package with this function enabled you can install from r-universe rather than from CRAN: https://r-lidar.r-universe.dev/lidR", call. = FALSE)

  UseMethod("point_metrics", las)
}

#' @export
point_metrics.LAS <- function(las, func, k, r, xyz = FALSE, filter = NULL, ...) {

  #if (is_disable_point_metrics())
  #{
  #  C_point_metrics = function(a, b, c, d, e, f) {}
  #}

  if (missing(k) && missing(r))  stop("'k' or 'r' is missing", call. = FALSE)

  # knn + radius
  if (!missing(r) && !missing(k)) {
    assert_is_a_number(k)
    assert_is_a_number(r)
    assert_all_are_positive(k)
    assert_all_are_positive(r)
    k <- as.integer(k)
    mode <- 3L
  }

  # knn
  if (!missing(k) && missing(r)) {
    assert_is_a_number(k)
    assert_all_are_positive(k)
    k <- as.integer(k)
    r <- 0
    mode <- 0L
  }

  # radius
  if (!missing(r) && missing(k)) {
    assert_is_a_number(r)
    assert_all_are_positive(r)
    k <- 0
    mode <- 1L
  }

  assert_is_a_bool(xyz)

  formula <- tryCatch(lazyeval::is_formula(func), error = function(e) FALSE)
  if (!formula) func <- lazyeval::f_capture(func)

  # Preparation of the objects
  func <- lazyeval::f_interp(func)
  call <- lazyeval::as_call(func)

  # Memory allocation for the query. This memory will be recycled in each iteration
  p <- list(...)
  if (is.null(p$alloc))
    n <- if (mode == 0L) k else as.integer(density(las) * pi * r^3)
  else
    n <- as.integer(p$alloc)

  if (n < 1) n <- 1

  # Creation of a call environment
  env <- new.env(parent = parent.frame())

  filter <- parse_filter(las, filter)

  output <- C_point_metrics(las, k, r, n, call, env, filter)

  if (length(output[[1]]) == 1)
  {
    name <- names(output[[1]])
    output <- data.table::data.table(unlist(output))
    if (!is.null(name)) data.table::setnames(output, name)
  }
  else
  {
    output <- data.table::rbindlist(output)
  }

  if (xyz)
  {
    xyz <- coordinates3D(las)
    data.table::setDT(xyz)
    if (length(filter) > 1 && !all(filter)) xyz <- xyz[filter]
    coln <- names(output)
    coln <- c("X", "Y", "Z", coln)
    output[["X"]] <- xyz$X
    output[["Y"]] <- xyz$Y
    output[["Z"]] <- xyz$Z
    data.table::setcolorder(output, coln)
    output[]
  }
  else
  {
    pointID <- NULL
    colnames <- data.table::copy(names(output))

    if (length(filter) > 1)
      output[, pointID := which(filter)]
    else
      output[, pointID := 1:npoints(las)]

    data.table::setcolorder(output, c("pointID", colnames))
    output[]
  }

  return(output)
}

#' @export
#' @rdname point_metrics
point_eigenvalues = function(las, k, r, xyz = FALSE, metrics = FALSE, coeffs = FALSE, filter = NULL)
{
  pointID <- NULL

  if (missing(k) && missing(r))  stop("'k' or 'r' is missing", call. = FALSE)

  # knn + radius
  if (!missing(r) && !missing(k)) {
    assert_is_a_number(k)
    assert_is_a_number(r)
    assert_all_are_positive(k)
    assert_all_are_positive(r)
    k <- as.integer(k)
    mode <- 2L
  }

  # knn
  if (!missing(k) && missing(r)) {
    assert_is_a_number(k)
    assert_all_are_positive(k)
    k <- as.integer(k)
    r <- 0
    mode <- 0L
  }

  # radius
  if (!missing(r) && missing(k)) {
    assert_is_a_number(r)
    assert_all_are_positive(r)
    k <- 0
    mode <- 1L
  }


  assert_is_a_bool(xyz)

  if (!is.null(filter))
  {
    filter <- parse_filter(las, filter)
    M <- C_eigen_metrics(las, k, r, coeffs, filter, getThreads())
  }
  else
  {
    M <- C_fast_eigen_decomposition(las, k, r, coeffs, getThreads())
    M <- cbind(pointID = 1:npoints(las), M)
  }

  data.table::setDT(M)
  data.table::setorder(M, pointID)

  if (xyz)
  {
    xyz <- coordinates3D(las)
    data.table::setDT(xyz)
    xyz <- xyz[M[["pointID"]]]
    coln <- names(M)
    coln <- c("X", "Y", "Z", coln)
    M[["X"]] <- xyz$X
    M[["Y"]] <- xyz$Y
    M[["Z"]] <- xyz$Z
    data.table::setcolorder(M, coln)
    M[]
  }

  if (metrics)
  {
    M[["curvature"]]  <- M[["eigen_smallest"]] / (M[["eigen_largest"]] + M[["eigen_medium"]] + M[["eigen_smallest"]])   # Luca et al 2019
    M[["linearity"]]  <- (M[["eigen_largest"]] - M[["eigen_medium"]]) / M[["eigen_largest"]]                            # Luca et al 2019
    M[["planarity"]]  <- (M[["eigen_medium"]] - M[["eigen_smallest"]]) / M[["eigen_largest"]]                           # Luca et al 2019
    M[["sphericity"]] <- M[["eigen_smallest"]] / M[["eigen_largest"]];
    M[["anisotropy"]] <- (M[["eigen_largest"]] - M[["eigen_smallest"]]) / M[["eigen_largest"]]
  }

  return(M)
}

#' Sensor tracking algorithm
#'
#' This function is made to be used in \link{track_sensor}. It implements an algorithm from Roussel et
#' al. 2020 (see reference) for sensor tracking using multiple returns to estimate the positioning of
#' the sensor by computing the intersection in space of the lines passing through the first and last
#' returns.
#'
#' When multiple returns from a single pulse are detected, the sensor computes their positions as being
#' in the center of the footprint and thus all aligned. Because of that behavior, a line
#' drawn between and beyond those returns must cross the sensor. Thus, several consecutive pulses
#' emitted in a tight interval (e.g. 0.5 seconds) can be used to approximate an intersection
#' point in the sky that corresponds to the sensor position given that the sensor carrier hasn't
#' moved much during this interval. A weighted least squares method gives an approximation of the
#' intersection by minimizing the squared sum of the distances between the intersection point and all
#' the lines.
#'
#' @param interval numeric. Interval used to bin the gps times and group the pulses to compute
#' a position at a given timepoint t.
#' @param pmin integer. Minimum number of pulses needed to estimate a sensor position.
#' For a given interval, the sensor position is not computed if the number of pulses is lower than
#' \code{pmin}.
#' @examples
#' # A valid file properly populated
#' LASfile <- system.file("extdata", "Topography.laz", package="lidR")
#' las = readLAS(LASfile)
#'
#' # pmin = 15 because it is an extremely tiny file
#' # strongly decimated to reduce its size. There are
#' # actually few multiple returns
#' flightlines <- track_sensor(las, Roussel2020(pmin = 15))
#'
#' plot(las@header)
#' plot(flightlines, add = TRUE)
#' @references Roussel Jean-Romain, Bourdon Jean-Francois, Achim Alexis, (2020) Range-based intensity
#' normalization of ALS data over forested areas using a sensor tracking method from multiple returns
#' (preprint) Retrieved from eartharxiv.org/k32qw https://doi.org/10.31223/osf.io/k32qw
#' @export
#' @name track_sensor_roussel2020
Roussel2020 = function(interval = 0.5, pmin = 50)
{
  assert_is_a_number(interval)
  assert_is_a_number(pmin)
  interval <- lazyeval::uq(interval)
  pmin <- lazyeval::uq(pmin)

  f = function(data)
  {
    assert_is_valid_context(LIDRCONTEXTTRK, "Roussel2020")

    . <- X <- Y <- Z <- ReturnNumber <- PointSourceID <- pulseID <- gpstime <- UserData <- SCORE <- npulses <- NULL
    # Generate the bins
    bins <- round_any(data$gpstime, interval)
    # Find the position P of the sensor in each interval
    P  <- data[, if (.N > 2*pmin) sensor_positions(X,Y,Z, ReturnNumber), by = .(gpstime = bins, PointSourceID = PointSourceID)]
    return(P)
  }

  f <- plugin_track(f)
  return(f)
}

#' Sensor tracking algorithm
#'
#' This function is made to be used in \link{track_sensor}. It implements an algorithm from Gatziolis
#' and McGaughey 2019 (see reference) for sensor tracking using multiple returns to estimate the positioning
#' of the sensor by computing the intersection in space of the lines passing through the first and
#' last returns.
#'
#' In the original paper, two steps are described: (1) closest point approach (CPA) and (2) cubic
#' spline fitting. Technically, the cubic spline fitting step is a post-processing step and is not
#' included in this algorithm.\cr\cr
#' The source code of the algorithm is a slight modification of the original source code provided
#' with the paper to fit with the lidR package.
#'
#' @param SEGLENFactor scalar. Weighting factor for the distance b/w 1st and last pulse returns
#' @param AngleFactor scalar. Weighting factor for view angle of mother pulse of a return
#' @param deltaT scalar. TimeBlock duration (in seconds)
#' @references Gatziolis, D., & McGaughey, R. J. (2019). Reconstructing Aircraft Trajectories from
#' Multi-Return Airborne Laser-Scanning Data. Remote Sensing, 11(19), 2258.
#' @examples
#' # A valid file properly populated
#' LASfile <- system.file("extdata", "Topography.laz", package="lidR")
#' las = readLAS(LASfile)
#' flightlines <- track_sensor(las, Gatziolis2019())
#'
#' plot(las@header)
#' plot(flightlines, add = TRUE)
#' @author Demetrios Gaziolis and Jean-Romain Roussel
#' @export
#' @name track_sensor_gatziolis2019
Gatziolis2019 <- function(SEGLENFactor = 1.0059, AngleFactor = 0.8824, deltaT = 0.5)
{
  assert_is_a_number(SEGLENFactor)
  assert_is_a_number(AngleFactor)
  assert_is_a_number(deltaT)

  f = function(data)
  {
    assert_is_valid_context(LIDRCONTEXTTRK, "Gatziolis2019")

    XFIRST <- XLAST <- YFIRST <- YLAST <- ZFIRST <- ZLAST <- SEGLEN <- SCANANGLE <- SCORE <- TBLOCK <- WT <- PULSEFLAG <- DIST <- NULL

    if ("ScanAngleRank" %in% names(data))
      angleattribute = "ScanAngleRank"
    else if ("ScanAngle" %in% names(data))
      angleattribute = "ScanAngle"
    else
      stop("Error: Scan angles are missing", call. = FALSE)

    select = c("X", "Y", "Z", "gpstime", angleattribute, "PointSourceID")

    # Reshape the data to get one pulse (pair of points) per rows
    v <- seq( 2, nrow(data), by = 2L)
    pulse.dt <- data.table::as.data.table(data[v, c("X", "Y", "Z")])
    names(pulse.dt) <- c("XFIRST", "YFIRST", "ZFIRST")
    pulse.dt[, c("XLAST", "YLAST", "ZLAST", "T", "SCANANGLE", "PointSourceID") := data[v-1, select, with = FALSE]]
    pulse.dt[, SEGLEN := sqrt( (XFIRST - XLAST)^2 + (YFIRST - YLAST)^2 + (ZFIRST - ZLAST)^2 ) ]

    ## Define time blocks of pulses
    minT <- min(pulse.dt[, "T"])
    minT <- floor(minT / deltaT) * deltaT                 ## aligns minT to increments of deltaT
    pulse.dt[, TBLOCK := floor( (T - minT) / deltaT )]    ## generates time blocks of returns labeled 0 to k.
    maxT <- ceiling(max(pulse.dt[, "T"]) / deltaT) * deltaT

    # >>>>>>>>>>>>>>>>
    # (Removed from original code because it is already performed upstream)
    #
    ## check if at least two eligible pulses are present in each timeblock. If only one, remove it
    #tblBlock.df <- count(pulse.dt$TBLOCK)
    #w <- tblBlock.df$freq > 1 ## you need at least two pulses to compute a platform location estimate
    #pulse.dt <- subset(pulse.dt, TBLOCK %in% tblBlock.df$x[w])
    # <<<<<<<<<<<<<<<<

    ## compute pulse weights (and convert pulses with negative SCANANGLE to negative weight)
    pulse.dt[, WT := exp(SEGLENFactor)*SEGLEN + exp(AngleFactor)*abs(SCANANGLE)]
    pulse.dt[SCANANGLE < 0, WT := -1.0 * WT]

    ## Filter to two 'best' pulses per time Block
    pulseflag = filterTimeBlockPulses(pulse.dt)
    pulse.dt[, PULSEFLAG := pulseflag ]
    pulse.dt <- pulse.dt[PULSEFLAG == 1][, PULSEFLAG := NULL][]

    ## calculate cpa
    trj.df <- cmpCPA(pulse.dt)
    data.table::setDT(trj.df)
    trj.df[, DIST := NULL]
    data.table::setcolorder(trj.df, c("T", "PointSourceID", "X", "Y", "Z", "WT"))
    data.table::setnames(trj.df, c("gpstime", "PointSourceID", "X", "Y", "Z", "SCORE"))
    return(trj.df)
  }

  f <- plugin_track(f)
  return(f)
}

sensor_positions <- function(x, y, z, rn)
{
  first <- rn == 1L
  last  <- rn > 1L

  # Adapted from MATLAB (Anders Eikenes, 2012)
  # http://www.mathworks.com/matlabcentral/fileexchange/37192-intersection-point-of-lines-in-3d-space

  # Matrix n x 3 containing XYZ coordinates for each first/last return
  A <- matrix(c(x[first], y[first], z[first]), ncol = 3L)
  B <- matrix(c(x[last], y[last], z[last]), ncol = 3L)

  # This should never happen but it might happen in wrongly populated datasets
  if (nrow(A) != nrow(B))
    return(list(X = NA_real_, Y = NA_real_, Z = NA_real_, npulses = NA_integer_))

  # Lines described as vectors V
  V <- B - A

  # Weights lines by distances
  D <- matrix(sqrt(.rowSums(V^2, nrow(V), ncol(V))))
  W <- D / sum(D)

  # Normalized vectors (|V| = 1)
  ones <- matrix(1, ncol = 3)
  N  <- V / (D %*% ones)
  Nx <- N[, 1L, drop = FALSE]
  Ny <- N[, 2L, drop = FALSE]
  Nz <- N[, 3L, drop = FALSE]

  Wxx <- W * (Nx^2 - 1)
  Wyy <- W * (Ny^2 - 1)
  Wzz <- W * (Nz^2 - 1)
  Wxy <- W * Nx * Ny
  Wxz <- W * Nx * Nz
  Wyz <- W * Ny * Nz

  Sxx <- sum(Wxx)
  Syy <- sum(Wyy)
  Szz <- sum(Wzz)
  Sxy <- sum(Wxy)
  Sxz <- sum(Wxz)
  Syz <- sum(Wyz)
  S   <- matrix(c(Sxx, Sxy, Sxz, Sxy, Syy, Syz, Sxz, Syz, Szz), nrow = 3L)

  Cx <- sum(A[,1, drop = FALSE] * Wxx + A[,2, drop = FALSE] * Wxy + A[,3, drop = FALSE] * Wxz)
  Cy <- sum(A[,1, drop = FALSE] * Wxy + A[,2, drop = FALSE] * Wyy + A[,3, drop = FALSE] * Wyz)
  Cz <- sum(A[,1, drop = FALSE] * Wxz + A[,2, drop = FALSE] * Wyz + A[,3, drop = FALSE] * Wzz)
  C  <- matrix(c(Cx,Cy,Cz))

  M  <- t(solve(S,C))
  M  <- round(M, 3)

  return(list(X = M[1], Y = M[2], Z = M[3], SCORE = nrow(A)))
}

#' @export
#' @rdname segment
segment_snags = function(las, algorithm, attribute = "snagCls")
{
  UseMethod("segment_snags", las)
}

#' @export
segment_snags.LAS = function(las, algorithm, attribute = "snagCls")
{
  assert_is_algorithm(algorithm)
  assert_is_algorithm_sng(algorithm)
  stopif_forbidden_name(attribute)

  lidR.context <- "segment_snags"
  snags <- algorithm(las)

  las <- add_lasattribute(las, snags, attribute, "Number identifying a snag class")
  return(las)
}

#' @export
segment_snags.LAScatalog = function(las, algorithm, attribute = "snagCls")
{
  assert_is_algorithm(algorithm)
  assert_is_algorithm_sng(algorithm)
  stopif_forbidden_name(attribute)

  opt_select(las) <- "*"

  options <- list(need_buffer = TRUE, drop_null = TRUE, need_output_file = TRUE)
  output  <- catalog_map(las, segment_snags,  algorithm = algorithm, .options = options)
  return(output)
}

#' Search Nearest Neighbors
#'
#' Fast parallelized k-neareast neighbor searching algorithms for point cloud in LAS format
#'
#' @param data LAS object for input point cloud
#' @param query LAS object for query locations
#' @param k number of nearest neighbors to search.
#'
#' @return a list contains:
#' \enumerate{
#' \item **nn.index** an n x k matrix for the nearest neighbor indice.
#' \item **nn.dist** an n x k matrix for the nearest neighbor Euclidean distances.
#' }
#' @export
knn = function(data, k = 10)
{
  stopifnotlas(data)
  assert_all_are_non_negative(k)
  return(cpp_knn(data, k, getThreads()))
}

#' @rdname knn
#' @export
knnx = function(data, query, k = 10)
{
  stopifnotlas(data)
  stopifnotlas(query)
  assert_all_are_non_negative(k)
  return(cpp_knnx(data, query, k, getThreads()))
}

#' Plot a LAS* object
#'
#' Plot displays a 3D interactive windows based on rgl for \link{LAS} objects\cr\cr
#' Plot displays an interactive view for \link[=LAScatalog-class]{LAScatalog} objects with pan and
#' zoom capabilities based on `mapview()` from package `mapview`. If the coordinate reference
#' system (CRS) of the \code{LAScatalog} is non empty, the plot can be displayed on top of base maps
#' (satellite data, elevation, street, and so on).\cr\cr
#' Plot displays a \link[=LASheader-class]{LASheader} object exactly like it displays a LAScatalog
#' object.
#'
#'
#' @param x A \code{LAS*} object
#' @param y Unused (inherited from R base)
#' @param color characters. The attribute used to color the point cloud. Default is Z coordinates. RGB
#' is an allowed string even if it refers to three attributes simultaneously.
#' @param pal palette function, similar to heat.colors, or palette values. Default is \code{"auto"}
#' providing an automatic coloring depending on the attribute \code{color}
#' @param bg The color for the background. Default is black.
#' @param clear_artifacts logical. It is a known and documented issue that the 3D visualisation with
#' \code{rgl} displays artifacts. The points look aligned and/or regularly spaced in some view angles.
#' This is because \code{rgl} computes with single precision \code{float}. To fix that the point
#' cloud is shifted to (0,0) to reduce the number of digits needed to represent its coordinates.
#' The drawback is that the point cloud is not plotted at its actual coordinates.
#' @param breaks either a numeric vector with the actual breaks, or a name of a method accepted
#' by the style argument of \link[classInt:classIntervals]{classIntervals}
#' @param nbreaks Number of colors breaks.
#' @param axis logical. Display axis on XYZ coordinates.
#' @param legend logical. Display a gradient colour legend.
#' @param backend character. Can be \code{"rgl"} or \code{"lidRviewer"}. If \code{"rgl"} is chosen
#' the display relies on the \code{rgl} package. If \code{"lidRviewer"} is chosen it relies on the
#' \code{lidRviewer} package, which is much more efficient and can handle million of points
#' using less memory. \code{lidRviewer} is not available on CRAN yet and should
#' be installed from github (see. \url{https://github.com/r-lidar/lidRviewer}).
#' @param add If \code{FALSE} normal behaviour otherwise must be the output of a prior plot function
#' to enable the alignment of a second point cloud.
#' @param voxel boolean or numeric. Displays voxels instead of points. Useful to render the output
#' of \link{voxelize_points}, for example. However it is computationally demanding to render and can
#' easily take 15 seconds for 10000 voxels. It should be reserved for small scenes. If boolean the voxel
#' resolution is guessed automatically. Otherwise users can provide the size of the voxels. To reduce the rendering time,
#' an internal optimization removes voxels that are not visible when surrounded by other voxels.
#' @param NAcol a color for NA values.
#'
#' @param mapview logical. If \code{FALSE} the catalog is displayed in a regular plot from R base.
#' Since v4.0.4 `mapview = TRUE` is also possible with LAS objects.
#' @param chunk_pattern logical. Display the current chunk pattern used to process the catalog.
#' @param overlaps logical. Highlight the overlaps between files.
#'
#' @param ... Will be passed to \link[rgl:3dobjects]{points3d} (LAS) or \link[graphics:plot.default]{plot}
#' if \code{mapview = FALSE} or to `mapview()` if \code{mapview = TRUE} (LAScatalog).
#'
#' @examples
#' \dontrun{
#' LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
#' las <- readLAS(LASfile)
#'
#' plot(las)
#' plot(las, color = "Intensity")
#' plot(las, color = "ScanAngleRank", pal = rainbow)
#'
#' # If outliers break the color range, use the breaks parameter
#' las$Intensity[150] <- 1000L
#' plot(las, color = "Intensity")
#' plot(las, color = "Intensity", breaks = "quantile", nbreaks = 50)
#'
#' plot(las, color = "Classification")
#'
#' # This dataset is already tree segmented
#' plot(las, color = "treeID")
#' plot(las, color = "treeID", pal = random.colors)
#'
#'
#' # single file LAScatalog using data provided in lidR
#' ctg = readLAScatalog(LASfile)
#' plot(ctg)
#' plot(ctg, map = T, map.types = "Esri.WorldImagery")
#' }
#'
#' @export
#' @method plot LAS
setGeneric("plot", function(x, y, ...)
  standardGeneric("plot"))

#' @rdname plot
setMethod("plot", signature(x = "LAS", y = "missing"),
          function(x, y, ...,
                   color = "Z",
                   pal = "auto",
                   bg = "black",
                   breaks = "pretty",
                   nbreaks = "auto",
                   backend = "rgl",
                   clear_artifacts = TRUE,
                   axis = FALSE,
                   legend = FALSE,
                   add = FALSE,
                   voxel = FALSE,
                   NAcol = "lightgray",
                   mapview = FALSE)
{
  plot.LAS(x, y, ..., color = color, pal = pal, bg = bg, breaks = breaks, nbreaks = nbreaks, backend = backend, clear_artifacts = clear_artifacts, axis = axis, legend = legend, add = add, voxel = voxel, NAcol = NAcol, mapview = mapview)
})

#' @export
#' @rdname plot
setMethod("plot", signature(x = "LAScatalog", y = "missing"), function(x, y, mapview = FALSE, chunk_pattern = FALSE, overlaps = FALSE, ...)
{
  plot.LAScatalog(x, y, mapview, chunk_pattern, overlaps, ...)
})

#' @export
#' @rdname plot
setMethod("plot", signature(x = "LASheader", y = "missing"), function(x, y, mapview = FALSE, ...)
{
  PHB  <- x@PHB
  crs  <- st_crs(x)
  xmin <- PHB[["Min X"]]
  xmax <- PHB[["Max X"]]
  ymin <- PHB[["Min Y"]]
  ymax <- PHB[["Max Y"]]
  mtx  <- matrix(c(xmin, xmax, ymin, ymax)[c(1, 1, 2, 2, 1, 3, 4, 4, 3, 3)], ncol = 2)
  geom <- sf::st_polygon(list(mtx))
  geom <- sf::st_sfc(geom)
  sf::st_crs(geom) <- crs

  names(PHB) <- make.names(names(PHB))

  if (!is.null(PHB[["Number.of.points.by.return"]]))
  {
    PHB[["Number.of.1st.return"]] <- PHB[["Number.of.points.by.return"]][1]
    PHB[["Number.of.2nd.return"]] <- PHB[["Number.of.points.by.return"]][2]
    PHB[["Number.of.3rd.return"]] <- PHB[["Number.of.points.by.return"]][3]
    PHB[["Number.of.4th.return"]] <- PHB[["Number.of.points.by.return"]][4]
    PHB[["Number.of.5th.return"]] <- PHB[["Number.of.points.by.return"]][5]
    PHB[["Number.of.points.by.return"]] <- NULL
    PHB[["Global.Encoding"]] <- NULL
  }

  data <- data.table::as.data.table(PHB)
  data <- sf::st_set_geometry(data, geom)

  res <- new("LAScatalog")
  res@data <- data

  plot.LAScatalog(res, mapview = mapview, ...)
})

plot.LAScatalog = function(x, y, mapview = FALSE, chunk_pattern = FALSE, overlaps = FALSE, ...)
{
  assert_is_a_bool(mapview)
  assert_is_a_bool(chunk_pattern)
  assert_is_a_bool(overlaps)

  if (mapview)
  {
    if (!requireNamespace("mapview", quietly = TRUE))
    {
      message("'mapview' is required to display the LAScatalog interactively.") # nocov
      mapview <- FALSE # nocov
    }

    if (overlaps)
      message("overlaps = TRUE is not supported yet with mapview") # nocov
  }

  if (mapview)
  {
    mapview::mapview(x@data, ...)
  }
  else if (chunk_pattern)
  {
    opt_progress(x) <- TRUE
    opt_output_files(x) <- ""
    engine_chunks(x)
    return(invisible())
  }
  else
  {
    # New feature from v2.2.0 to do not process some tiles
    process <- x@data$process
    if (is.null(process)) process <- rep(TRUE, nrow(x@data))
    if (!is.logical(process)) {
      warning("The attribute 'process' of the catalog is not logical.", call. = FALSE)
      process <- rep(TRUE, nrow(x@data))
    }

    alpha   <- ifelse(process, 0.15, 0.03)
    param   <- list(...)
    xmin    <- min(x@data$Min.X)
    xmax    <- max(x@data$Max.X)
    ymin    <- min(x@data$Min.Y)
    ymax    <- max(x@data$Max.Y)
    xcenter <- (xmin + xmax)/2
    ycenter <- (ymin + ymax)/2
    col    <- grDevices::rgb(0, 0, 1, alpha = alpha)

    if (is.null(param$xlim)) param$xlim <- c(xmin, xmax)
    if (is.null(param$ylim)) param$ylim <- c(ymin, ymax)
    if (is.null(param$xlab)) param$xlab <- ""
    if (is.null(param$ylab)) param$ylab <- ""
    if (is.null(param$asp))  param$asp  <- 1
    if (!is.null(param$col)) col <- param$col
    param$map.types = NULL

    param$col <- "white"
    param$x   <- xcenter
    param$y   <- ycenter

    op <- graphics::par(mar = c(2.5,2.5,1,1) + 0.1)

    if (is.null(param$add)) do.call(graphics::plot, param)

    if (!isTRUE(attr(x, "trueshape")))
    {
      graphics::rect(x@data$Min.X, x@data$Min.Y, x@data$Max.X, x@data$Max.Y, col = col)
      graphics::par(op)
    }
    else
    {
      plot(sf::st_geometry(sf::st_as_sf(x)), axes = TRUE)
      graphics::par(op)
    }

    if (overlaps) {
      plot(catalog_overlaps(x), add = T, col = "red", border = "red")
    }

    return(invisible())
  }
}

plot.LAS = function(x, y, ...,
                    color = "Z",
                    pal = "auto",
                    bg = "black",
                    breaks = "pretty",
                    nbreaks = "auto",
                    backend = "rgl",
                    clear_artifacts = TRUE,
                    axis = FALSE,
                    legend = FALSE,
                    add = FALSE,
                    voxel = FALSE,
                    NAcol = "lightgray",
                    mapview = FALSE)
{
  if (mapview)
  {
    return(plot(header(x), mapview = TRUE))
  }

  args <- list(...)
  if (is.null(args$size))
    args$size <- 1.5

  # Backward compatibility
  #nocov start
  if (!is.null(args$colorPalette))
  {
    if (length(args$colorPalette) == 1L && args$colorPalette == "auto")
      pal <- "auto"
    else
      pal <- grDevices::colorRampPalette(args$colorPalette)

    args$colorPalette <- NULL
    message("The argument 'coloPalette' is deprecated. Use 'pal' instead")
  }

  # Backward compatibility
  if (!is.null(args$trim))
  {
    args$trim <- NULL
    message("The argument 'trim' is not longer supported. Use breaks = 'quantile' instead")
  }

  # Backward compatibility
  if (!is.null(args$nbits))
  {
    args$nbits <- NULL
    message("The argument 'nbits' is not longer supported. It is now infered automatically.")
  }
  #nocov end

  backend <- match.arg(backend, c("rgl", "lidRviewer"))
  use_pcv <- backend == "lidRviewer"
  use_rgl <- !use_pcv
  use_vox <- !isFALSE(voxel)
  has_pcv <- "lidRviewer" %in% rownames(utils::installed.packages())
  has_col <- color %in% names(x)
  use_rgb <- color == "RGB"
  has_rgb <- all(c("R", "G", "B") %in% names(x))
  autocol <- is.character(pal) && length(pal) == 1 && pal == "auto"

  # Error handling
  assert_is_a_bool(clear_artifacts)
  assert_is_a_bool(axis)
  assert_is_a_bool(legend)
  if (!isFALSE(add))
  {
    assert_is_numeric(add)
    assert_is_of_length(add, 2)
  }
  if (is.empty(x))         stop("Cannot display an empty point cloud", call. = FALSE)
  if (use_pcv & !has_pcv)  stop("'lidRviewer' package is needed. Please read documentation.", call. = FALSE) # nocov
  if (length(color) > 1)   stop("'color' should contain a single value.", call. = FALSE)
  if (!use_rgb & !has_col) stop("'color' should refer to an attribute of the LAS data.", call. = FALSE)
  if (use_rgb & !has_rgb)  stop("No 'RGB' attributes found.", call. = FALSE)

  # Retrieve voxels size
  if (use_vox)
  {
    verbose("Plotting in voxel mode")

    if (!is.null(attr(x, 'res')) && !is.numeric(voxel))
    {
      voxel <- attr(x, 'res')
      verbose(glue::glue("Resolution of the voxels found in the attribute 'res': {voxel}"))
    }
    else if (isTRUE(voxel))
    {
      xres <- min(diff(sort(unique(x$X))))
      yres <- min(diff(sort(unique(x$Y))))
      zres <- min(diff(sort(unique(x$Z))))
      voxel  <- min(xres, yres, zres)
      verbose(glue::glue("Resolution automatically detected: {voxel}"))
    }
    else if (is.numeric(voxel))
    {
      verbose(glue::glue("Resolution defined by user: {voxel}"))
    }
    else
    {
      stop("Parameter 'voxel' must be a boolean or a number")
    }
  }

  # Create a colour scheme for each point as a function of the attribute plotted
  nbreaks <- if (nbreaks == "auto") 50 else nbreaks
  idcolor <- NULL

  if (autocol)
  {
    if (color == "Z")
      pal <- height.colors
    else if (color == "Intensity")
      pal <- grDevices::heat.colors
    else if (color  == "Classification")
      pal <- lasclass.colors
    else if (color == "ScanAngleRank" | color == "ScanAngle")
      pal <- height.colors
    else if (color == "ReturnNumber")
      pal <- grDevices::colorRampPalette(rev(c("#440154FF", "#3B528BFF", "#21908CFF", "#5DC863FF", "#FDE725FF")))
    else if (color == "treeID")
      pal <- pastel.colors
    else if (color == "Amplitude")
      pal <- amplitude.color
    else
      pal <- height.colors

    if (color  == "Classification")
      idcolor <- x[["Classification"]] + 1L

    if (color == "treeID")
      nbreaks <- data.table::uniqueN(x[[color]])
  }
  else if (is.character(pal))
  {
    pal <- grDevices::colorRampPalette(pal)
  }

  if (use_rgb) # Case RGB
  {
    idcolor <- "RGB"
  }
  else if (!is.null(idcolor)) # Case classification
  {
    pal <- pal()
    idcolor[idcolor > length(pal)] <- length(pal) # No colordefined after class 19. Everything is yellow
  }
  else
  {
    colorattr <- x[[color]]
    nunique   <- data.table::uniqueN(colorattr)

    if (nunique == 1L)
    {
      idcolor <- 1L
      pal   <- pal(1)
    }
    else
    {
      if (is.logical(colorattr)) colorattr <- as.integer(colorattr)
      if (is.character(breaks)) breaks <- suppressWarnings(classInt::classIntervals(colorattr, min(nbreaks, nunique), breaks)$brk)
      nbreaks   <- length(breaks)
      pal       <- pal(nbreaks-1)
      idcolor   <- cut(colorattr, breaks, include.lowest = TRUE, label = FALSE)
    }

    if (anyNA(idcolor))
    {
      idcolor[is.na(idcolor)] <- nbreaks
      pal <- c(pal, NAcol)
    }
  }

  lasplot <- if (use_rgl) .plot_with_rgl else .plot_with_pcv # nocov

  return(lasplot(x, bg, idcolor, pal, clear_artifacts, axis, legend, args, add, voxel, breaks))
}

.plot_with_rgl = function(x, bg, idcolor, pal, clear_artifacts, axis, legend, args, add, use_voxels, breaks)
{
  # The color of the foreground  (axis actually) is the opposite of the background
  # so usually white, maybe black if bg = "white" or something fancy in other cases
  fg <- grDevices::col2rgb(bg)
  fg <- grDevices::rgb(t(255 - fg)/255)

  # If it is a new layer, compute the offset used to get rid of floating point accuracy with rgl
  # otherwise we use the input offset
  if (isFALSE(add))
  {
    minx <- min(x@data$X)
    miny <- min(x@data$Y)
  }
  else
  {
    minx <- add[1]
    miny <- add[2]
  }

  if (is.character(idcolor) && idcolor == "RGB")
  {
    colmax <- max(max(x@data[["R"]]), max(x@data[["G"]]), max(x@data[["B"]]))
    nbits  <- if (colmax > 255) 16 else 8
    maxcol <- 2^nbits - 1
    col <- grDevices::rgb(x@data[["R"]]/maxcol, x@data[["G"]]/maxcol, x@data[["B"]]/maxcol)
  }
  else if (is.integer(idcolor))
  {
    col <- pal[idcolor]
  }
  else
    stop("Internal error in color attribution. Please report.") # nocov

  # Optimize the rendering of the voxels by removing voxels than can't be seen
  if(!isFALSE(use_voxels) && !is_disable_point_metrics())
  {
    nvoxels1 <- npoints(x)
    res <- as.numeric(use_voxels)
    u <- point_metrics(x, ~length(Z), r = res)
    keep <- u$V1 <= 6
    x <- filter_poi(x, keep)
    col <- col[keep]
    nvoxels2 <- npoints(x)
    verbose(glue::glue("Rendering {nvoxels2} voxels. {nvoxels1-nvoxels2} being hidden."))
  }

  with <- c(list(x = x@data[["X"]], y = x@data[["Y"]], z = x@data[["Z"]], col = col), args)

  if (clear_artifacts)
  {
    with$x <- with$x - minx
    with$y <- with$y - miny
  }
  # If it is a new layer, open an rgl windows
  if (isFALSE(add))
  {
    rgl::open3d()
    rgl::bg3d(color = bg)
    rgl::material3d(specular = "black")
  }

  # Two modes, point-cloud rendering or voxel rendering
  if (isFALSE(use_voxels))
  {
    do.call(rgl::points3d, with)
  }
  else
  {
    res <- as.numeric(use_voxels)
    res <- res / 2
    n <- npoints(x)
    u <- vector("list", n)
    for(i in 1:n)
    {
      voxel <- rgl::cube3d(col = col[i])
      voxel <- rgl::scale3d(voxel, res, res, res)
      voxel <- rgl::translate3d(voxel , with$x[i], with$y[i], with$z[i])
      u[[i]] <- voxel
    }

    rgl::shapelist3d(u)
  }

  if (axis)
  {
    rgl::axis3d("x", col = fg)
    rgl::axis3d("y", col = fg)
    rgl::axis3d("z", col = fg)
  }

  if (legend)
  {
    # nocov because this fails on some flavors on CRAN
    f <- .plot_scale_gradient(breaks, fg, pal, bg) # nocov
    rgl::bg3d(texture = f, col = "white") # nocov
  }

  .pan3d(2)

  if (clear_artifacts)
    return(invisible(c(minx, miny)))
  else
    return(invisible(c(0,0)))
}

# nocov start
.plot_with_pcv = function(x, bg, idcolor, pal, clear_artifacts, axis, legend, args, add, use_voxel, breaks)
{
  if (!isFALSE(add)) stop("Argument 'add' is not supported with lidRviewer")
  if (!isFALSE(use_voxel)) stop("Argument 'use_voxel' is not supported with lidRviewer")

  if (is.character(idcolor))
  {
    if (idcolor == "RGB")
      eval(parse(text = "lidRviewer::plot_xyzrgb(x@data$X, x@data$Y, x@data$Z, x@data$R, x@data$G, x@data$B, args$size)"))
    else
      stop("Unexpected error.", call. = FALSE)
  }
  else
  {
    eval(parse(text = "lidRviewer::plot_xyzcol(x@data$X, x@data$Y, x@data$Z, pal, idcolor, args$size)"))
  }

  return(invisible(c(0,0)))
}
# nocov end

# nocov start
.plot_scale_gradient = function(breaks, text.col, scale.col, bg)
{
  f <- tempfile(fileext = ".png")
  labels <- round(breaks, 2)
  ncol   <- length(scale.col)
  nlab   <- length(labels)
  xl <- 1 ; yb <- 1 ; xr <- 1.1 ; yt <- 2
  grDevices::png(f, 1920, 1080, bg = bg)
  graphics::layout(matrix(1:2, nrow = 1), widths = c(0.9,0.1))
  graphics::par(mar = c(5.1, 4.1, 4.1, 2.1))
  graphics::plot(0, ann = FALSE, type = "n", axes = FALSE)
  graphics::par(mar = c(5.1, 0.5, 4.1, 0))
  graphics::plot(NA, type = "n", ann = FALSE, xlim = c(1,2), ylim = c(1,2), xaxt = "n", yaxt = "n", bty = "n")
  graphics::rect(xl, utils::head(seq(yb, yt, (yt - yb)/ncol), -1), xr, utils::tail(seq(yb, yt, (yt - yb)/ncol), -1), col = scale.col, border = NA)
  graphics::mtext(labels, side = 2, at = seq(yb, yt, length.out = nlab), las = 2, cex = 1.2, col = text.col)
  grDevices::dev.off()
  return(f)
}
# nocov end

# From rgl.setMouseCallbacks man page
# nocov start
.pan3d <- function(button, dev = rgl::cur3d(), subscene = rgl::currentSubscene3d(dev))
{
  start <- list()

  begin <- function(x, y)
  {
    activeSubscene <- rgl::par3d("activeSubscene", dev = dev)
    start$listeners <<- rgl::par3d("listeners", dev = dev, subscene = activeSubscene)

    for (sub in start$listeners)
    {
      init <- rgl::par3d(c("userProjection","viewport"), dev = dev, subscene = sub)
      init$pos <- c(x/init$viewport[3], 1 - y/init$viewport[4], 0.5)
      start[[as.character(sub)]] <<- init
    }
  }

  update <- function(x, y)
  {
    for (sub in start$listeners)
    {
      init <- start[[as.character(sub)]]
      xlat <- 2*(c(x/init$viewport[3], 1 - y/init$viewport[4], 0.5) - init$pos)
      mouseMatrix <- rgl::translationMatrix(xlat[1], xlat[2], xlat[3])
      rgl::par3d(userProjection = mouseMatrix %*% init$userProjection, dev = dev, subscene = sub )
    }
  }
  rgl::rgl.setMouseCallbacks(button, begin, update, dev = dev, subscene = subscene)
}
# nocov end

#' Tools inherited from base R for LAS* objects
#'
#' Tools inherited from base R for LAS* objects
#'
#' @param x a LAS* object
#' @param ... LAS* objects if it is the sole argurment (e.g. in rbind())
#' @name tools
#' @rdname tools
NULL

#' @export
#' @rdname tools
dim.LAS <- function(x) return(dim(x@data))

#' @export
#' @rdname tools
dim.LAScatalog <- function(x) return(dim(x@data))

#' @export
#' @rdname tools
ncol.LAS <- function(x) return(ncol(x@data))

#' @export
#' @rdname tools
nrow.LAScatalog <- function(x) return(nrow(x@data))

#' @export
#' @rdname tools
names.LAS <- function(x) return(names(x@data))

#' @export
#' @rdname tools
names.LASheader <- function(x) return(c(names(x@PHB), names(x@VLR), names(x@EVLR)))

#' @export
names.LAScatalog <- function(x) return(names(x@data))

#' @name tools
#' @export
rbind.LAS <- function(...)
{
  dots <- list(...)
  names(dots) <- NULL
  assert_all_are_same_crs(dots)

  xscales <- sapply(dots, function(x) x[["X scale factor"]])
  yscales <- sapply(dots, function(x) x[["Y scale factor"]])
  zscales <- sapply(dots, function(x) x[["Z scale factor"]])
  xoffsets <- sapply(dots, function(x) x[["X offset"]])
  yoffsets <- sapply(dots, function(x) x[["Y offset"]])
  zoffsets <- sapply(dots, function(x) x[["Z offset"]])

  need_quantization <- FALSE
  if (length(unique(xscales)) != 1L || length(unique(yscales)) != 1L || length(unique(zscales)) != 1L ||
      length(unique(xoffsets)) != 1L || length(unique(yoffsets)) != 1L || length(unique(zoffsets)) != 1L)
  {
    warning("Different LAS objects have different scales and/or offsets. The first object was used as reference to quantize the others.", call. = FALSE)
    need_quantization <- TRUE
  }

  data <- data.table::rbindlist(lapply(dots, function(x) { x@data } ))

  if (need_quantization)
  {
    quantize(data[["X"]], xscales[1], xoffsets[1])
    quantize(data[["Y"]], yscales[1], yoffsets[1])
    quantize(data[["Z"]], zscales[1], zoffsets[1])
  }

  return(LAS(data, dots[[1]]@header, st_crs(dots[[1]]), index = dots[[1]]@index))
}

#' @name rbind
#' @export
rbind.LAScatalog <- function(...)
{
  dots <- list(...)
  names(dots) <- NULL
  assert_all_are_same_crs(dots)

  data <- do.call(rbind, lapply(dots, function(x) { x@data } ))
  ctg <- dots[[1]]
  ctg@data <- data

  chk <- las_check(ctg, print = FALSE)
  for (msg in chk$warnings) warning(msg, call. = FALSE)
  for (msg in chk$errors) warning(msg, call. = FALSE)

  return(ctg)
}

#' @export
#' @rdname tools
npoints <- function(x, ...) { UseMethod("npoints", x) }

#' @export
npoints.LAS <- function(x, ...) { return(nrow(x@data)) }

#' @export
npoints.LASheader <- function(x, ...) { return(x@PHB[["Number of point records"]]) }

#' @export
npoints.LAScatalog <- function(x, ...) { return(sum(x[["Number.of.point.records"]])) }

#' @rdname tools
#' @export
setGeneric("density", function(x, ...) standardGeneric("density"))

#' @export
#' @rdname tools
setMethod("density", "LAS", function(x, ...) { return(npoints(x)/area(x)) })

#' @export
#' @rdname tools
setMethod("density", "LASheader", function(x, ...) { return(npoints(x)/area(x)) })

#' @rdname tools
#' @export
setMethod("density", "LAScatalog",  function(x, ...) { return(npoints(x)/area(x)) })

context("filter_duplicates")

las     <- example

test_that("filter_duplicates works with LAS", {
   las@data <- rbind(las@data, las@data)
   las = lidR:::lasupdateheader(las)
   las = filter_duplicates(las)
   expect_equal(npoints(las), 30L)
})

#test_that("filter_duplicates works with LAScatalog", {
#
#})

raster_template <- function(raster)
{
  if (is(raster, "raster_template"))
    return(raster)

  if (inherits(raster, "Raster") | is(raster, "SpatRaster"))
  {
    bbox <- raster_bbox(raster)
    xmin <- bbox$xmin
    xmax <- bbox$xmax
    ymin <- bbox$ymin
    ymax <- bbox$ymax
    nrow <- nrow(raster)
    ncol <- ncol(raster)
    xres <- round((xmax - xmin)/ncol, 5)
    yres <- round((ymax - ymin)/nrow, 5)
    crs  <- sf::st_crs(bbox)
  }
  else if (is(raster, "stars"))
  {
    sdim   <- stars::st_dimensions(raster)
    stattr <- attr(stars::st_dimensions(raster), "raster")

    # https://r-spatial.github.io/stars/articles/stars4.html
    if (sdim$y$delta > 0) stop("stars objects with delta > 0 are not yet supported", call. = FALSE)
    if (!all(stattr$affine == 0)) stop("Rotated and sheared stars objects are not yet supported", call. = FALSE)
    if (stattr$curvilinear == TRUE) stop("Curvilinear stars objects are not yet supported", call. = FALSE)

    bbox <- raster_bbox(raster)
    xmin <- bbox$xmin
    xmax <- bbox$xmax
    ymin <- bbox$ymin
    ymax <- bbox$ymax
    xres <- sdim$x$delta
    yres <- abs(sdim$y$delta)
    ncol <- as.integer(round((xmax-xmin)/xres))
    nrow <- as.integer(round((ymax-ymin)/yres))
    crs  <- raster_crs(raster)
  }
  else
  {
    raster_error() # nocov
  }

  layout <- list(xmin = xmin, xmax = xmax,
                 ymin = ymin, ymax = ymax,
                 ncol = ncol, nrow = nrow,
                 xres = xres, yres = yres,
                 crs  = crs)

  class(layout) <- "raster_template"
  return(layout)
}

raster_cell_from_xy = function(raster, x, y)
{
  e    <- raster_template(raster)
  xmin <- e$xmin
  xmax <- e$xmax
  ymin <- e$ymin
  ymax <- e$ymax
  xres <- e$xres
  yres <- e$yres
  ncol <- e$ncol
  nrow <- e$nrow

  col  <- floor((x - xmin) / xres);
  col[x == xmax] <- ncol-1;
  col[col >= ncol | col < 0L] <- NA_integer_

  row  <- floor((ymax - y) / yres);
  row[y == ymin] <- nrow-1;
  row[row >= nrow | row < 0L] <- NA_integer_

  cell <- as.integer(row * ncol + col + 1)
  return(cell)
}

raster_value_from_xy = function(raster, x, y, layer = 1)
{
  if (is(raster, "stars_proxy"))
  {
    xrange <- range(x)
    yrange <- range(y)
    bbox   <- sf::st_bbox(c(xmin = xrange[1], xmax = xrange[2], ymin = yrange[1], ymax = yrange[2]), crs = raster_crs(raster))
    raster <- raster[bbox]
    raster <- stars::st_as_stars(raster)
  }

  cells <- raster_cell_from_xy(raster, x, y)
  return(raster_value_from_cells(raster, cells, layer))
}

raster_value_from_cells = function(raster, cells, layer = 1)
{
  if (inherits(raster, "Raster"))
  {
    if (raster_nlayer(raster) > 1)
      return(raster::extract(raster, cells, layer = layer, nl = 1)[,1])
    else
      return(raster::extract(raster, cells, layer = layer, nl = 1))
  }

  if (is(raster, "stars_proxy"))
    stop("stars_proxy not yet supported in 'raster_value_from_cells()'", call. = FALSE) # nocov

  if (is(raster, "stars"))
  {
    dims <- dim(raster)

    if (is.na(dims[3]))
      return(raster[[1]][cells])

    return(raster[,,,layer][[1]][cells])
  }

  if (is(raster, "SpatRaster"))
    return(terra::extract(raster, cells)[[layer]])

  raster_error() # nocov
}

raster_crop <- function(raster, bbox)
{
  if (inherits(raster, "Raster"))
  {
    bbox <- raster::extent(bbox)
    return(raster::crop(raster, bbox))
  }

  if (is(raster, "stars"))
  {
    # Workaround for https://github.com/r-spatial/stars/issues/463
    rbbox <- sf::st_bbox(raster)
    xmin  <- max(rbbox$xmin, bbox$xmin)
    ymin  <- max(rbbox$ymin, bbox$ymin)
    xmax  <- min(rbbox$xmax, bbox$xmax)
    ymax  <- min(rbbox$ymax, bbox$ymax)
    bbox  <- sf::st_bbox(c(xmin = xmin , ymin = ymin, xmax = xmax, ymax = ymax), crs = sf::st_crs(bbox))
    return(sf::st_crop(raster, bbox))
  }

  if (is(raster, "SpatRaster"))
  {
    bbox <- as.numeric(bbox)
    bbox <- bbox[c(1,3,2,4)]
    bbox <- terra::ext(bbox)
    return(terra::crop(raster, bbox))
  }

  raster_error() # nocov
}

raster_as_matrix <- function(raster, downsample = FALSE)
{
  size <- 1000*1000*10

  if (raster_is_proxy(raster))
  {
    if (raster_ncell(raster) > size)
    {
      if (downsample == FALSE)
        stop("Cannot convert this on-disk raster into a matrix without downsampling")

      raster <- raster_downsample(raster, size)
    }
    else
    {
      if (is(raster, "stars"))
        raster <- stars::st_as_stars(raster)
    }
  }

  if (is(raster, "RasterLayer"))
  {
    mx <-  t(apply(raster::as.matrix(raster), 2, rev))
    x  <- sort(raster::xFromCol(raster, 1:raster::ncol(raster)))
    y  <- sort(raster::yFromRow(raster, 1:raster::nrow(raster)))
    return(list(x = x, y = y, z = mx))
  }

  if (is(raster, "stars"))
  {
    sdim   <- stars::st_dimensions(raster)
    if (sdim$y$delta > 0) stop("stars objects with delta > 0 are not yet supported", call. = FALSE)

    dims <- dim(raster)

    if (is.na(dims[3]))
      mx <-  raster[[1]]
    else
      mx <- raster[,,,1][[1]]

    x <- stars::st_get_dimension_values(raster, 'x')
    y <- stars::st_get_dimension_values(raster, 'y')
    mx <- t(apply(mx, 1, rev))
    return(list(x = x, y = rev(y), z = mx))
  }

  if (is(raster, "SpatRaster"))
  {
    mx <- t(apply(terra::as.matrix(raster, wide = TRUE), 2, rev))
    x  <- sort(terra::xFromCol(raster, 1:terra::ncol(raster)))
    y  <- sort(terra::yFromRow(raster, 1:terra::nrow(raster)))
    return(list(x = x, y = y, z = mx))
  }

  raster_error() # nocov
}

#' @importFrom stats na.omit
raster_as_dataframe <- function(raster,  xy = TRUE, na.rm = TRUE)
{
  ondisk <- raster_is_proxy(raster)

  # Small rasters can be loaded on the fly
  if (ondisk & raster_fits_in_memory(raster, n = 10))
  {
    raster <- raster_in_memory(raster)
    ondisk <- FALSE
  }

  if (ondisk)
    stop("On-disk rasters not supported in 'raster_as_dataframe()'", call. = FALSE) # nocov

  m <- raster_as_matrix(raster)
  z <- as.numeric(t(apply(m$z, 1, rev)))
  x <- rep(m$x, length(m$y))
  y <- rep(rev(m$y), each = length(m$x))
  d <- data.frame(X = x, Y = y)

  if (!xy) d$Z = z

  data.table::setDT(d)

  if (na.rm)
  {
    nas <- is.na(z)
    d <- d[!nas]
  }

  return(d)
}

raster_as_las <- function(raster, bbox = NULL)
{
  ondisk <- raster_is_proxy(raster)

  # Small rasters can be loaded on the fly
  if (ondisk & raster_fits_in_memory(raster, n = 10))
  {
    raster <- raster_in_memory(raster)
    ondisk <- FALSE
  }

  if (ondisk & is.null(bbox))
    stop("On-disk rasters not supported without a bbox in 'raster_as_las()'", call. = FALSE) # nocov

  if (ondisk & !is.null(bbox))
    raster <- raster_crop(raster, bbox)

  data <- raster_as_dataframe(raster, xy = FALSE, na.rm = TRUE)
  header <- rlas::header_create(data)
  las <- LAS(data, header, crs = raster_crs(raster), check = FALSE)
  return(las)
}

raster_res <- function(raster)
{
  res <- raster_template(raster)
  return(unname(c(res$xres, res$yres)))
}

raster_ncell <- function(raster)
{
  res <- raster_template(raster)
  return(unname(res$nrow * res$ncol))
}

raster_names <- function(raster)
{
  if (inherits(raster, "Raster") | is(raster, "SpatRaster"))
    return(names(raster))

  if (is(raster, "stars"))
  {
    dims <- dim(raster)

    if (is.na(dims[3]))
      return(names(raster))
    else
      return(stars::st_get_dimension_values(raster, 3))
  }

  raster_error() # nocov
}

`raster_names<-` <- function(raster, value)
{
  if (inherits(raster, "Raster") | is(raster, "SpatRaster"))
  {
    names(raster) <- value
    return(raster)
  }

  if (is(raster, "stars"))
  {
    dims <- dim(raster)

    if (is.na(dims[3]))
      names(raster) <- value
    else
      raster <- stars::st_set_dimensions(raster, 3, values = value)

    return(raster)
  }

  raster_error() # nocov
}

raster_is_supported <- function(raster)
{
  return(inherits(raster, "Raster") | is(raster, "stars") | is(raster, "SpatRaster") | is(raster, "raster_template"))
}

raster_build_vrt = function(file_list, vrt)
{
  file_list <- unlist(file_list)
  layers    <- names(terra::rast(file_list[1]))
  folder    <- dirname(file_list[1])
  file      <- paste0("/", vrt, ".vrt")
  vrt       <- paste0(folder, file)
  sf::gdal_utils("buildvrt", source = file_list, destination = vrt, quiet = TRUE)
  return(vrt)
}

raster_layout <- function(las, res, start = c(0,0), buffer = 0, format = "template")
{
  if (is_raster(res))
  {
    if (raster_nlayer(res) > 1)
      stop("Multilayer rasters are not supported as a template", call. = FALSE)

    resolution <- raster_res(res)
    if (resolution[1] != resolution[2])
      stop("Rasters with different x y resolutions are not supported as a template", call. = FALSE)

    pkg <- raster_pkg(res)
    res <- raster_template(res)
    res <- raster_materialize(res, pkg, NA_real_)
    return(res)
  }

  assert_all_are_non_negative(res)
  bbox <- st_adjust_bbox(las, res, start, buffer)

  if (format == "stars")
  {
    layout  <- stars::st_as_stars(bbox, dx = res, value = NA_real_)
    return(layout)
  }

  if (format == "raster")
  {
    bbox   <- raster::extent(bbox)
    layout <- suppressWarnings(raster::raster(bbox, res = res, crs = as(st_crs(las), "CRS")))
    suppressWarnings(layout[] <- NA_real_)
    raster::crs(layout) <- raster::crs(las)
    return(layout)
  }

  if (format == "terra")
  {
    crs    <- st_crs(las)$wkt
    bbox   <- as.numeric(bbox)
    bbox   <- bbox[c(1,3,2,4)]
    bbox   <- terra::ext(bbox)
    layout <- terra::rast(bbox)
    terra::res(layout) <- res
    suppressWarnings(layout[] <- NA_real_)
    terra::crs(layout) <- crs
    return(layout)
  }

  xmin <- bbox$xmin
  xmax <- bbox$xmax
  ymin <- bbox$ymin
  ymax <- bbox$ymax
  xres <- res
  yres <- res
  ncol <- round((xmax - xmin)/xres)
  nrow <- round((ymax - ymin)/yres)

  layout <- list(xmin = xmin, xmax = xmax,
                 ymin = ymin, ymax = ymax,
                 ncol = ncol, nrow = nrow,
                 xres = xres, yres = yres,
                 crs  = st_crs(las))

  class(layout) <- "raster_template"
  return(layout)
}

raster_materialize <- function(raster, pkg = getOption("lidR.raster.default"), values = NA_real_)
{
  if (!is(raster, "raster_template"))
    return(raster)

  bbox <- raster_bbox(raster)

  if (pkg == "raster")
  {
    crs  <- as(st_crs(bbox), "CRS")
    bbox <- as.numeric(bbox)
    bbox <- bbox[c(1,3,2,4)]
    bbox <- raster::extent(bbox)
    res  <- as.numeric(raster$xres)
    out  <- suppressWarnings(raster::raster(bbox, res = res, crs = crs))
    suppressWarnings(out[] <- values)
    return(out)
  }

  if (pkg == "stars")
  {
    res  <- stars::st_as_stars(bbox, nx = raster$ncol, ny = raster$nrow, values = values)
    return(res)
  }

  if (pkg == "terra")
  {
    crs  <- sf::st_crs(bbox)$wkt
    bbox <- as.numeric(bbox)
    bbox <- bbox[c(1,3,2,4)]
    bbox <- terra::ext(bbox)
    res  <- terra::rast(bbox)
    terra::res(res) <- raster$xres
    suppressWarnings(res[] <- values)
    terra::crs(res) <- crs
    return(res)
  }

  raster_error() # nocov
}

raster_alignment <- function(res, start = c(0, 0))
{
  if (is_raster(res))
  {
    ext       <- raster_bbox(res)
    r         <- round(raster_res(res)[1], 5)
    start     <- round(c(ext$xmin, ext$ymin), 5)
    alignment <- list(res = r, start = start)
    return(alignment)
  }

  alignment   <- list(res = res, start = start)
  return(alignment)
}

raster_pkg <- function(raster)
{
  if (inherits(raster, "Raster"))
    return("raster")

  if (is(raster, "stars"))
    return("stars")

  if (is(raster, "SpatRaster"))
    return("terra")

  raster_error() # nocov
}

raster_nlayer <- function(raster)
{
  if (inherits(raster, "Raster") | is(raster, "SpatRaster"))
  {
    return(dim(raster)[3])
  }

  if (is(raster, "stars"))
  {
    dims <- dim(raster)
    if (is.na(dims[3])) return(1L)
    return(dims[3])
  }

  raster_error() # nocov
}

raster_set_values <- function(raster, values, cells = NULL)
{
  if (is(raster, "raster_template"))
    stop("Internal error in raster_value: the raster must be materialized. Please report.", call. = FALSE) # nocov

  if (raster_nlayer(raster) > 1)
    stop("Internal error in raster_values: the raster must be a single layer. Please report.", call. = FALSE) # nocov

  if (is.null(cells))
    cells <- 1:raster_ncell(raster)

  if (inherits(raster, "Raster") | is(raster, "SpatRaster"))
  {
    suppressWarnings(raster[cells] <- values)
    return(raster)
  }

  if (is(raster, "stars"))
  {
    raster[[1]][cells] <- values
    storage.mode(raster[[1]]) <- storage.mode(values)
    return(raster)
  }

  raster_error() # nocov
}

raster_values <- function(raster)
{
  if (is(raster, "raster_template"))
    stop("Internal error in raster_value: the raster must be materialized. Please report.", call. = FALSE) # nocov

  if (raster_nlayer(raster) > 1)
    stop("Internal error in raster_values: the raster must a single layer. Please report.", call. = FALSE) # nocov

  if (inherits(raster, "Raster") | is(raster, "SpatRaster"))
    return(raster[])

  if (is(raster, "stars"))
    return(as.numeric(raster[[1]][]))

  raster_error() # nocov
}

raster_replace_na <- function(raster, value = 0)
{
  if (is(raster, "raster_template"))
    stop("Internal error in replace_na: the raster must be materialized. Please report.", call. = FALSE) # nocov

  if (inherits(raster, "Raster") | is(raster, "SpatRaster"))
  {
    raster[is.na(raster)] <- value
    return(raster)
  }

  if (is(raster, "stars"))
  {
    raster[[1]][is.na(raster[[1]])] <- value
    return(raster)
  }

  raster_error() # nocov
}

raster_error <- function()
{
  parent = deparse(sys.calls()[[sys.nframe()-1]])
  stop(glue::glue("Internal error in {parent}: raster must be a raster from raster, stars,7 or terra. Please report"), call. = FALSE) # nocov
}

raster_is_proxy <- function(raster)
{
  if (is(raster, "stars_proxy"))
    return(TRUE)

  if (inherits(raster, "Raster"))
    return(!raster::inMemory(raster))

  if (is(raster,"SpatRaster"))
    return(!terra::inMemory(raster))

  return(FALSE)
}

raster_downsample <- function(raster, size)
{
  cat("downsample to", size, "cells\n")

  if (raster_ncell(raster) < size)
    return(raster)

  if (is(raster, "stars"))
  {
    downsample <- floor(1/sqrt(size/raster_ncell(raster)))
    raster <- stars::st_as_stars(raster, downsample = downsample)
    return(raster)
  }

  if (inherits(raster, "Raster"))
  {
    raster = raster::sampleRegular(raster, size, asRaster = TRUE, useGDAL = TRUE)
    return(raster)
  }

  if (is(raster, "SpatRaster"))
  {
    raster = terra::spatSample(raster, size, method = "regular", as.raster = TRUE)
    return(raster)
  }

  raster_error() # nocov
}

raster_bbox <- function(raster)
{
  if (is(raster, "raster_template"))
    return(sf::st_bbox(raster))

  if (is(raster, "stars"))
    return(sf::st_bbox(raster))

  if (inherits(raster, "Raster"))
    return(sf::st_bbox(raster))

  if (is(raster, "SpatRaster"))
  {
    # sf::st_bbox now works properly on SpatRaster.
    return(sf::st_bbox(raster))
  }

  raster_error() # nocov
}

raster_crs <- function(raster)
{
  if (is(raster, "SpatRaster"))
  {
    if (terra::crs(raster) == "")
      return(sf::NA_crs_)
  }

  return(sf::st_crs(raster))
}

raster_pkg <- function(raster)
{
  if (is(raster, "stars"))
    return("stars")

  if (inherits(raster, "Raster"))
    return("raster")

  if (is(raster,"SpatRaster"))
    return("terra")

  raster_error()
}

raster_size <- function(raster)
{
  if (is(raster, "stars"))
  {
    if (raster_nlayer(raster) == 1)
      return(c(dim(raster), 1))
    else
      return(dim(raster))
  }

  if (inherits(raster, "Raster") | is(raster,"SpatRaster"))
    return(dim(raster))

  raster_error()
}

raster_multilayer_class <- function(pkg = getOption("lidR.raster.default"))
{
  if (pkg == "raster") return("RasterBrick")
  if (pkg == "terra") return("SpatRaster")
  if (pkg == "stars") return("stars")
}

raster_class <- function(pkg = getOption("lidR.raster.default"))
{
  if (pkg == "raster") return("RasterLayer")
  if (pkg == "terra") return("SpatRaster")
  if (pkg == "stars") return("stars")
}

raster_fits_in_memory <- function(raster, n = 1)
{
  if (!raster_is_proxy(raster)) return(TRUE)

  nc <- raster_ncell(raster)
  n <- n * raster_nlayer(raster)
  memneed <- nc * n * 8L
  memavail <- terra::free_RAM()*1000
  memavail <- 0.6 * memavail
  return(memneed < memavail)
}

raster_in_memory <- function(raster)
{
  if (is(raster, "stars"))
    return(stars::st_as_stars(raster))

  if (inherits(raster, "Raster"))
    return(raster::readAll(raster))

  if (is(raster,"SpatRaster"))
    return(raster*1)

  raster_error()
}


# Workaround for #580 #622. If normalize height is ran in parallel it will fail with
# SpatRaster because they are not serializable. SpatRaster are converted to RasterLayer
# for multicore strategies
convert_ondisk_spatraster_into_serializable_raster_if_necessary <- function(x)
{
  if (is_raster(x) && raster_pkg(x) == "terra")
  {
    ncores <- try_to_get_num_future_cores()
    if (!is.null(ncores) && ncores >= 2L)
      x <- raster::raster(x)
  }

  return(x)
}

context("rasterize_canopy")

slr = lidR:::raster_class()

test_that("rasterize_canopy with p2r() returns a georeferenced RasterLayer", {

  las <- lidR:::generate_las(2000)
  x   <- rasterize_canopy(las, 4, p2r())

  expected_bbox = sf::st_bbox(c(xmin = 0,xmax = 100, ymin = 0, ymax = 100), crs = st_crs(las))

  expect_true(is(x, slr))
  expect_equal(lidR:::raster_res(x), c(4,4))
  expect_equivalent(lidR:::raster_size(x), c(25,25,1))
  expect_equivalent(sf::st_bbox(x), expected_bbox)
  expect_equal(lidR:::raster_names(x), "Z")
})

test_that("grid_canopy with p2r() is backward compatible", {
  skip_if_not_installed("raster")
  las <- lidR:::generate_las(2000)
  x   <- grid_canopy(las, 4, p2r())

  expected_bbox = sf::st_bbox(c(xmin = 0,xmax = 100, ymin = 0, ymax = 100), crs = st_crs(las))

  expect_true(is(x, "RasterLayer"))
  expect_equal(lidR:::raster_res(x), c(4,4))
  expect_equal(lidR:::raster_size(x), c(25,25,1))
  expect_equivalent(sf::st_bbox(x), expected_bbox)
  expect_equal(lidR:::raster_names(x), "Z")
})


test_that("rasterize_canopy with p2r() returns the good bounding box", {

  dt  <- data.table::data.table(X = seq(0,10, 0.5), Y = seq(0,10, 0.5), Z = runif(21))
  las <- LAS(dt, crs = sf::st_crs(26917))

  x <- rasterize_canopy(las, 2, p2r())
  expected_bbox = sf::st_bbox(c(xmin = -2, xmax = 12, ymin = -2,ymax = 12), crs = st_crs(las))

  expect_true(is(x, slr))
  expect_equivalent(sf::st_bbox(x), expected_bbox )

  x <- rasterize_canopy(las, 1, p2r())
  expected_bbox <- sf::st_bbox(c(xmin = -1, xmax = 11, ymin = -1,ymax = 11), crs = st_crs(las))

  expect_true(is(x, slr))
  expect_equivalent(sf::st_bbox(x), expected_bbox )
})

test_that("rasterize_canopy with p2r() works with subcircle option", {

  dt  <- data.table::data.table(X = c(0,10), Y = c(0,20), Z = c(0, 5))
  las <- suppressWarnings(LAS(dt, crs = sf::st_crs(26917)))
  x   <- rasterize_canopy(las, 0.5, p2r(10))
  expected_bbox <- sf::st_bbox(c(xmin = -0.5, xmax = 10.5, ymin = -0.5,ymax = 20.5), crs = st_crs(las))

  expect_true(is(x, slr))
  expect_equivalent(sf::st_bbox(x), expected_bbox)
})


test_that("rasterize_canopy pit-free works subcircle", {

  skip_on_cran()

  las <- lidR:::generate_las(2000)

  f <- pitfree(thresholds = c(0,2,5,10,15), max_edge = c(0, 1.5), subcircle = 5)
  x <- rasterize_canopy(las, 1, f)
  expected_bbox = sf::st_bbox(c(xmin = 0,xmax = 100, ymin = 0, ymax = 100), crs = st_crs(las))

  expect_true(is(x, slr))
  expect_equal(lidR:::raster_res(x), c(1,1))
  expect_equivalent(sf::st_bbox(x), expected_bbox)
})

test_that("rasterize_canopy with p2r() works with na.fill", {

  las <- lidR:::generate_las(1000)
  x   <- rasterize_canopy(las, 4, p2r())

  expect_equal(sum(is.na(lidR:::raster_values(x))), 118L)

  x   <- rasterize_canopy(las, 4, p2r(na.fill = knnidw(3)))

  expect_equal(sum(is.na(lidR:::raster_values(x))), 0L)
})

las <- readLAS(mixedconifer_las_path, select = "xyzr", filter = "-thin_with_grid 1")
ctg <- mixedconifer_ctg

opt_filter(ctg)          <- "-thin_with_grid 1"
opt_chunk_size(ctg)      <- 100
opt_chunk_alignment(ctg) <- c(5,15)
opt_progress(ctg)        <- FALSE

test_that("rasterize_canopy tin works both with LAS and LAScatalog", {

  f <- dsmtin()
  x <- rasterize_canopy(las, 1, f)
  y <- rasterize_canopy(ctg, 1, f)
  b <- sf::st_bbox(sf::st_buffer(sf::st_as_sfc(st_bbox(x)), -1))
  x <- lidR:::raster_crop(x, b)
  y <- lidR:::raster_crop(y, b)
  expected_bbox =  sf::st_bbox(c(xmin = 481261, xmax = 481349, ymin = 3812922,ymax = 3813010), crs = st_crs(las))

  expect_true(is(x, slr))
  expect_equal(lidR:::raster_res(x), c(1,1))
  expect_equivalent(lidR:::raster_size(x), c(88,88,1))
  expect_equivalent(sf::st_bbox(x), expected_bbox)
  expect_equal(lidR:::raster_names(x), "Z")
  expect_equal(lidR:::raster_values(x), lidR:::raster_values(y), tolerance = 0.00065)
})

test_that("rasterize_canopy pit-free works both with LAS and LAScatalog", {

  f <- pitfree(thresholds = c(0,2,5,10,15), max_edge = c(0, 1.5))
  x <- rasterize_canopy(las, 1, f)
  y <- rasterize_canopy(ctg, 1, f)
  b <- sf::st_bbox(sf::st_buffer(sf::st_as_sfc(st_bbox(x)), -1))
  x <- lidR:::raster_crop(x, b)
  y <- lidR:::raster_crop(y, b)
  expected_bbox =  sf::st_bbox(c(xmin = 481261, xmax = 481349, ymin = 3812922,ymax = 3813010), crs = st_crs(las))

  expect_true(is(x, slr))
  expect_equal(lidR:::raster_res(x), c(1,1))
  expect_equivalent(lidR:::raster_size(x), c(88,88,1))
  expect_equivalent(sf::st_bbox(x), expected_bbox)
  expect_equal(lidR:::raster_names(x), "Z")
  expect_equal(lidR:::raster_values(x), lidR:::raster_values(y), tolerance = 0.00079)
})

test_that("triangulation does not return negative values", {
  # See https://gis.stackexchange.com/questions/376261/negative-values-after-chm-rasterization-lidr
  las <- clip_rectangle(megaplot, 684750, 5017800, 684850, 5017900)
  chm <- rasterize_canopy(las, 1, dsmtin())
  expect_equal(min(lidR:::raster_values(chm), na.rm = T), 0)
})

test_that("rasterize_canopy spike-free works", {

  f <- spikefree(1.5)
  x <- rasterize_canopy(mixedconifer, 0.5, f)

  expect_true(is(x, slr))
  expect_equal(lidR:::raster_res(x), c(0.5,0.5))
  expect_equivalent(lidR:::raster_size(x), c(180,180,1))
  expect_equal(mean(x[], na.rm = T), 12.455, tolerance = 0.001)
  expect_equal(sum(is.na(x[])), 25, tolerance = 0.001)

  f <- spikefree(1)
  x <- rasterize_canopy(mixedconifer, 0.25, f)

  expect_equal(mean(x[], na.rm = T), 11.774, tolerance = 0.001)
  expect_equal(sum(is.na(x[])), 108, tolerance = 0.001)
})

las <- lidR:::generate_las(5000)
st_crs(las) <- 2949
las@data[, Z := round(Z + 0.1*X + 0.1*Y + sin(0.01*X) - sin(0.1*Y) + sin(0.003*X*Y),3)]

tdtm <- lidR:::raster_layout(las, 1, format = getOption("lidR.raster.default"))
xy   <- lidR:::raster_as_dataframe(tdtm, na.rm = FALSE, xy = TRUE)
X    <- xy$X
Y    <- xy$Y
cell <- lidR:::raster_cell_from_xy(tdtm, X, Y)
gnd  <- round(0.1*X + 0.1*Y + sin(0.01*X) - sin(0.1*Y) + sin(0.003*X*Y),3)
tdtm <- lidR:::raster_set_values(tdtm, gnd, cell)

slr = lidR:::raster_class()

test_that("rasterize_terrain works with knnidw", {

  dtm <- rasterize_terrain(las, 1, knnidw(k = 10L))

  expect_true(is(dtm, slr))
  expect_equal(lidR:::raster_res(dtm), c(1,1))
  expect_equal(dim(dtm), dim(tdtm))
  expect_equivalent(sf::st_bbox(dtm), sf::st_bbox(tdtm))
  expect_equal(lidR:::raster_names(dtm), "Z")
  expect_equal(sum(is.na(lidR:::raster_values(dtm))), 1L)

  error <- abs(dtm - tdtm)
  expect_equal(mean(lidR:::raster_values(error), na.rm = TRUE), 0.1558, tolerance = 0.0001)

  z <- lidR:::raster_value_from_xy(dtm, las$X, las$Y)
  expect_true(!anyNA(z))
})

test_that("grid_terrain is backward compatible", {
  skip_if_not_installed("raster")
  dtm <- grid_terrain(las, 1, knnidw(k = 10L))

  expect_true(is(dtm, "RasterLayer"))
  expect_equal(lidR:::raster_res(dtm), c(1,1))
  expect_equivalent(lidR:::raster_size(dtm), c(100, 100, 1))
  expect_equivalent(sf::st_bbox(dtm), sf::st_bbox(tdtm))
  expect_equal(lidR:::raster_names(dtm), "Z")
  expect_equal(sum(is.na(lidR:::raster_values(dtm))), 1L)
})

test_that("rasterize_terrain works with delaunay", {

  dtm <- suppressWarnings(rasterize_terrain(las, 1, tin()))

  expect_true(is(dtm, slr))
  expect_equal(lidR:::raster_res(dtm), c(1,1))
  expect_equal(dim(dtm), dim(tdtm))
  expect_equivalent(sf::st_bbox(dtm), sf::st_bbox(tdtm))
  expect_equal(lidR:::raster_names(dtm), "Z")
  expect_equal(sum(is.na(lidR:::raster_values(dtm))), 1L)

  error <- abs(dtm - tdtm)
  expect_equal(mean(lidR:::raster_values(error), na.rm = TRUE), 0.074, tolerance = 0.0001)

  z <- lidR:::raster_value_from_xy(dtm, las$X, las$Y)
  expect_true(!anyNA(z))
})

test_that("rasterize_terrain works with kriging", {

  skip_if_not_installed("gstat")

  dtm <- rasterize_terrain(las, 1, kriging(k = 10L))

  expect_true(is(dtm, slr))
  expect_equal(lidR:::raster_res(dtm), c(1,1))
  expect_equal(dim(dtm), dim(tdtm))
  expect_equivalent(sf::st_bbox(dtm), sf::st_bbox(tdtm))
  expect_equal(lidR:::raster_names(dtm), "Z")
  expect_equal(sum(is.na(lidR:::raster_values(dtm))), 1L)

  error <- abs(dtm - tdtm)
  expect_equal(mean(lidR:::raster_values(error), na.rm = TRUE), 0.0604, tolerance = 0.0001)

  z <- lidR:::raster_value_from_xy(dtm, las$X, las$Y)
  expect_true(!anyNA(z))
})

test_that("rasterize_terrain option keep_lowest works", {

  dtm <- rasterize_terrain(las, 1, tin(), keep_lowest = TRUE)

  expect_true(is(dtm, slr))
  expect_equal(lidR:::raster_res(dtm), c(1,1))
  expect_equal(dim(dtm), dim(tdtm))
  expect_equivalent(sf::st_bbox(dtm), sf::st_bbox(tdtm))
  expect_equal(lidR:::raster_names(dtm), "Z")
  expect_equal(sum(is.na(lidR:::raster_values(dtm))), 1L)

  error <- abs(dtm - tdtm)
  expect_equal(mean(lidR:::raster_values(error), na.rm = TRUE), 0.0754, tolerance = 0.0001)

  z <- lidR:::raster_value_from_xy(dtm, las$X, las$Y)
  expect_true(!anyNA(z))
})

las2 <- clip_circle(las, 50, 50, 40)

test_that("rasterize_terrain returns a circular dtm ", {

  dtm <- rasterize_terrain(las2, 1, tin())

  expected_bbox = sf::st_bbox(c(xmin = 10, xmax = 90, ymin = 10, ymax = 90), crs = st_crs(las2))

  expect_true(is(dtm,slr))
  expect_equal(lidR:::raster_res(dtm), c(1,1))
  expect_equal(lidR:::raster_names(dtm), "Z")
  expect_equivalent(sf::st_bbox(dtm), expected_bbox)

  z <- lidR:::raster_value_from_xy(dtm, las2$X, las2$Y)
  expect_true(!anyNA(z))
})

ctg  <- topography_ctg
las  <- topography

opt_chunk_size(ctg)      <- 300
opt_chunk_buffer(ctg)    <- 30
opt_chunk_alignment(ctg) <- c(50, 200)
opt_progress(ctg)        <- FALSE

test_that("rasterize_terrain returns the same both with LAScatalog and LAS", {

  dtm1 <- rasterize_terrain(las, 1, tin())
  dtm2 <- rasterize_terrain(ctg, 1, tin())
  bbox <- sf::st_bbox(sf::st_buffer(sf::st_as_sfc(st_bbox(las), crs = st_crs(las)), -4))

  cdtm1 <- lidR:::raster_crop(dtm1, bbox)
  cdtm2 <- lidR:::raster_crop(dtm2, bbox)

  error <- abs(cdtm1 - cdtm2)

  expect_equal(mean(lidR:::raster_values(error)), 0, tolerance = 2e-5)

  z <- lidR:::raster_value_from_xy(dtm1, las$X, las$Y)
  expect_true(!anyNA(z))
})

test_that("rasterize_terrain with LAScatalog and file", {

  skip_on_cran()

  opt_output_files(ctg) <- "{tempdir()}/{ID}"
  dtm <- rasterize_terrain(ctg, 1, tin())

  expect_is(dtm, "SpatRaster")
})

test_that("rasterize_terrain fails in some specific case", {

  skip_if_not_installed("geometry")

  las@header@PHB$`X scale factor` <- 0.002
  las@header@PHB$`Y scale factor` <- 0.002

  expect_message(rasterize_terrain(las, 1, tin()), "reverted to the old slow method")

  las = data.frame(
    X = runif(10, 0,10),
    Y = runif(10, 0,10),
    Z = 1,
    Classification = 2L)

  las = LAS(las)

  expect_error(rasterize_terrain(las, 1, tin()), NA)

  las@data$Classification <- 1L
  expect_error(rasterize_terrain(las, 1, tin()), "No ground points found")

  expect_error(rasterize_terrain(las, 1, tin(), use_class = 1), NA)

  las@data$Classification <- NULL
  expect_error(rasterize_terrain(las, 1, tin()), "LAS object does not contain 'Classification' attribute")
})

#' Spatial Interpolation Algorithm
#'
#' This function is made to be used in \link{rasterize_terrain} or \link{normalize_height}. It
#' implements an algorithm for spatial interpolation. Spatial interpolation is based on a Delaunay
#' triangulation, which performs a linear interpolation within each triangle. There are usually a
#' few points outside the convex hull, determined by the ground points at the very edge of the dataset,
#' that cannot be interpolated with a triangulation. Extrapolation can be performed with another algorithm.
#'
#' @param ... unused
#' @param extrapolate There are usually a few points outside the convex hull, determined by the ground
#' points at the very edge of the dataset, that cannot be interpolated with a triangulation.
#' Extrapolation is done using \link{knnidw} by default.
#'
#' @export
#'
#' @family dtm algorithms
#'
#' @examples
#' LASfile <- system.file("extdata", "Topography.laz", package="lidR")
#' las = readLAS(LASfile, filter = "-inside 273450 5274350 273550 5274450")
#'
#' #plot(las)
#'
#' dtm = rasterize_terrain(las, algorithm = tin())
#'
#' #plot(dtm)
#' #plot_dtm3d(dtm)
#' @name dtm_tin
tin = function(..., extrapolate = knnidw(3,1,50))
{
  assert_is_algorithm_spi(extrapolate)
  extrapolate <- lazyeval::uq(extrapolate)

  f = function(las, where)
  {
    assert_is_valid_context(LIDRCONTEXTSPI, "tin")
    z <- interpolate_delaunay(las, where, trim = 0, min_normal_z = 3e-2)

    # Extrapolate beyond the convex hull
    isna <- is.na(z)
    nnas <- sum(isna)
    if (nnas > 0)
    {
      lidR.context <- "spatial_interpolation"
      where2 <- data.frame(X = where$X[isna],  Y = where$Y[isna])
      zknn <- extrapolate(las, where2)
      z[isna] <- zknn
    }

    return(z)
  }

  f <- plugin_dtm(f, omp = TRUE)
  return(f)
}

#' Spatial Interpolation Algorithm
#'
#' This function is made to be used in \link{rasterize_terrain} or \link{normalize_height}. It implements an algorithm
#' for spatial interpolation. Interpolation is done using a k-nearest neighbour (KNN) approach with
#' an inverse-distance weighting (IDW).
#'
#' @param k integer. Number of k-nearest neighbours. Default 10.
#' @param p numeric. Power for inverse-distance weighting. Default 2.
#' @param rmax numeric. Maximum radius where to search for knn. Default 50.
#'
#' @export
#'
#' @family dtm algorithms
#'
#' @examples
#' LASfile <- system.file("extdata", "Topography.laz", package="lidR")
#' las = readLAS(LASfile)
#'
#' #plot(las)
#'
#' dtm = rasterize_terrain(las, algorithm = knnidw(k = 6L, p = 2))
#'
#' #plot(dtm)
#' #plot_dtm3d(dtm)
#' @name dtm_idw
knnidw = function(k = 10, p = 2, rmax = 50)
{
  k <- lazyeval::uq(k)
  p <- lazyeval::uq(p)
  rmax <- lazyeval::uq(rmax)

  f = function(las, where)
  {
    assert_is_valid_context(LIDRCONTEXTSPI, "knnidw")
    return(interpolate_knnidw(las, where, k, p, rmax))
  }

  f <- plugin_dtm(f, omp = TRUE)
  return(f)
}

#' Spatial Interpolation Algorithm
#'
#' This function is made to be used in \link{rasterize_terrain} or \link{normalize_height}. It
#' implements an algorithm for spatial interpolation. Spatial interpolation is based on universal
#' kriging using the `krige()` function from \code{gstat}. This method combines the
#' KNN approach with the kriging approach. For each point of interest it kriges the terrain using
#' the k-nearest neighbour ground points. This method is more difficult to manipulate but it is also
#' the most advanced method for interpolating spatial data.
#'
#' @param k numeric. Number of k-nearest neighbours. Default 10.
#' @param model A variogram model computed with `vgm()` from package `gstat`. If NULL it performs an ordinary
#' or weighted least squares prediction.
#'
#' @export
#'
#' @family dtm algorithms
#'
#' @examples
#' \dontrun{
#' LASfile <- system.file("extdata", "Topography.laz", package="lidR")
#' las = readLAS(LASfile)
#'
#' plot(las)
#'
#' dtm = rasterize_terrain(las, algorithm = kriging())
#'
#' plot(dtm)
#' plot_dtm3d(dtm)
#' }
#' @name dtm_kriging
#' @md
kriging = function(model = gstat::vgm(.59, "Sph", 874), k = 10L)
{
  assert_package_is_installed("gstat")

  f = function(las, where)
  {
    assert_is_valid_context(LIDRCONTEXTSPI, "kriging")
    return(interpolate_kriging(las, where, model, k))
  }

  f <- plugin_dtm(f)
  return(f)
}

interpolate_knnidw = function(points, coord, k, p, rmax = 50)
{
  if (!inherits(points, "LAS")) {
    h <- rlas::header_create(points)
    points <- LAS(points, h, check = F)
  }

  force_autoindex(points) <- LIDRGRIDPARTITION
  return(C_knnidw(points, coord$X, coord$Y, k, p, rmax, getThread()))
}

interpolate_kriging = function(points, coord, model, k)
{
  X <- Y <- Z <- NULL
  if (!getOption("lidR.verbose")) sink(tempfile())
  if (inherits(points, "LAS")) points <- points@data
  x  <- gstat::krige(Z~X+Y, location = ~X+Y, data = points, newdata = coord, model, nmax = k)
  sink()
  return(x$var1.pred)
}

interpolate_delaunay <- function(points, coord, trim = 0, scales = c(1,1), offsets = c(0,0), options = "QbB", min_normal_z = 0)
{
  # /!\ TODO: triangulation does not respect spatial index and always use grid partition

  stopifnot(is.numeric(trim), length(trim) == 1L)
  stopifnot(is.numeric(scales), length(scales) == 2L)
  stopifnot(is.numeric(offsets), length(offsets) == 2L)
  stopifnot(is.data.frame(coord))

  boosted_triangulation <- TRUE

  if (inherits(points, "LAS")) {
    xscale  <- points[["X scale factor"]]
    yscale  <- points[["Y scale factor"]]
    xoffset <- points[["X offset"]]
    yoffset <- points[["Y offset"]]
    scales  <- c(xscale, yscale)
    offsets <- c(xoffset, yoffset)
    points  <- points@data
  }

  stopifnot(is.data.frame(points))

  if (scales[1] != scales[2]) {
    message("The Delaunay triangulation reverted to the old slow method because xy scale factors are different so the fast method cannot be applied.")
    boosted_triangulation <- FALSE
  }

  # Check if coordinates actually match the resolution
  # Check only 100 of them
  n <- min(100L, length(points$X))
  s <- as.integer(seq(1L, length(points$X), length.out = n))
  X <- points$X[s]
  Y <- points$Y[s]
  x <- fast_countunquantized(X, scales[1], offsets[1])
  y <- fast_countunquantized(Y, scales[2], offsets[2])

  if (x > 0 | y > 0)
  {
    message("The Delaunay triangulation reverted to the old slow method because xy coordinates were not convertible to integer values. xy scale factors and offsets are likely to be invalid")
    boosted_triangulation <- FALSE
  }

  if (boosted_triangulation) {
    return(C_interpolate_delaunay(points, coord, scales, offsets, trim, min_normal_z, getThreads()))
  }
  else {
    P <- as.matrix(points)
    X <- as.matrix(coord)
    D <- tDelaunay(P, trim = trim)
    return(tInterpolate(D, P, X, getThreads()))
  }
}

#' Metric derivation at different levels of regularization
#'
#' `template_metrics()` computes a series of user-defined descriptive statistics for a LiDAR dataset
#' within each element of a template. Depending on the template it can be for each pixel of a raster
#' (area-based approach), or each polygon, or each segmented tree, or on the whole point cloud. Other
#' functions are convenient and simplified wrappers around `template_metrics()` and are expected to be
#' the actual functions used. See Details and Examples.
#'
#' \describe{
#' \item{`pixel_metrics`}{Area-based approach. Computes metrics in a square tessellation. The output is a
#' raster.}
#' \item{`hexagon_metrics`}{Computes metrics in an hexagon tessellation. The output is a `sf/sfc_POLYGON`}
#' \item{`plot_metrics`}{Computes metrics for each plot of a ground inventory by 1. clipping the plot
#' inventories with \link{clip_roi}, 2. computing the user's metrics for each plot with  \link{cloud_metrics}, and
#' 3. combining spatial data and metrics into one data.frame ready for statistical modelling with
#' `cbind`. The output is of the class of the input.}
#' \item{`cloud_metrics`}{Computes a series of user-defined descriptive statistics for an entire point cloud.
#' The output is a `list`}
#' \item{`crown_metrics`}{Once the trees are segmented, i.e. attributes exist in the
#' point cloud that reference each tree, computes a set of user-defined descriptive statistics for
#' each individual tree. The output can be spatial points or spatial polygons (`sf/sfc_POINT` or `sf/sfc_POLYGON`)}
#' \item{`voxel_metrics`}{Is a 3D version of `pixel_metrics`. It creates a 3D matrix of voxels with a given
#' resolution. It creates a voxel from the cloud of points if there is at least one point. The output is
#' a `data.frame`}
#' \item{`point_metrics`}{Is a bit more complex and is documented in \link{point_metrics}}
#' }
#'
#' @template param-las
#' @param func formula or expression. An expression to be applied to each element of the template (see
#' section "Parameter func").
#' @param template can be of many types and corresponds to the different levels of regularization.
#' `RasterLayer/stars/SpatRaster`, `sf/sfc` (polygons), `numeric`, `bbox`, `NULL`. The metrics are
#' computed for each element of the template. See examples.
#' @param filter formula of logical predicates. Enables the function to run only on points of interest
#' in an optimized way. See examples.
#' @param by_echo characters. The metrics are computed multiple times for different echo types. Can
#' be one or more of "all", "first", "intermediate", "lastofmany", "single", and "multiple". See examples.
#' Default is "all" meaning that it computes metrics with all points provided.
#' @param ... propagated to `template_metrics` i.e. `filter` and `by_echo`. `pixel_metrics()` also
#' supports `pkg = "terra|raster|stars"` to get an output in `SpatRaster`, `Raster*`
#' or `stars` format. Default is `getOption("lidR.raster.default")`.
#'
#' @section Parameter \code{func}:
#' The function to be applied to each cell is a classical function (see examples) that
#' returns a labelled list of metrics. For example, the following function \code{f} is correctly formed.
#' \preformatted{
#' f = function(x) {list(mean = mean(x), max = max(x))}
#' }
#' And could be applied either on the \code{Z} coordinates or on the intensities. These two
#' statements are valid:
#' \preformatted{
#' pixel_metrics(las, f(Z), res = 20)
#' voxel_metrics(las, f(Intensity), res = 2)
#' }
#' The following existing functions allow the user to
#' compute some predefined metrics: \link[=stdmetrics]{stdmetrics}
#' \link[=entropy]{entropy}, \link[=VCI]{VCI}, \link[=LAD]{LAD}. But usually users must write their own
#' functions to create metrics. \code{template_metrics} will dispatch the point cloud in the user's
#' function.
#'
#' @return Depends on the function, the template and the number of metrics. Can be a `RasterLayer`,
#' a `RasterBrick`, a `stars`, a `SpatRaster` a `sf/sfc`, a `list`, a `SpatialPolygonDataFrame`, or
#' a `data.table`. Functions are supposed to return an object that is best suited for storing the level
#' of regularization needed.
#'
#' @examples
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' las <- readLAS(LASfile, filter = "-keep_random_fraction 0.5")
#' col <- sf::sf.colors(15)
#' fun1 <- ~list(maxz = max(Z))
#' fun2 <- ~list(q85 = quantile(Z, probs = 0.85))
#'
#' set_lidr_threads(1) ; data.table::setDTthreads(1) # for cran only
#'
#' # ================
#' # CLOUD METRICS
#' # ================
#'
#' cloud_metrics(las, .stdmetrics_z)
#'
#' # ================
#' # PIXEL METRICS
#' # ================
#'
#' m <- pixel_metrics(las, fun1, 20)
#' #plot(m, col = col)
#'
#' # ================
#' # PLOT METRICS
#' # ================
#'
#' shpfile <- system.file("extdata", "efi_plot.shp", package="lidR")
#' inventory <- sf::st_read(shpfile, quiet = TRUE)
#' inventory # contains an ID and a Value Of Interest (VOI) per plot
#'
#' m <- plot_metrics(las, fun2, inventory, radius = 11.28)
#' #plot(header(las))
#' #plot(m["q85"], pch = 19, cex = 3, add = TRUE)
#'
#' \donttest{
#' # Works with polygons as well
#' inventory <- sf::st_buffer(inventory, 11.28)
#' #plot(header(las))
#' #plot(sf::st_geometry(inventory), add = TRUE)
#' m <- plot_metrics(las, .stdmetrics_z, inventory)
#' #plot(m["zq85"], pch = 19, cex = 3, add = TRUE)
#' }
#'
#' # ================
#' # VOXEL METRICS
#' # ================
#'
#' m <- voxel_metrics(las, length(Z), 8)
#' m <- voxel_metrics(las, mean(Intensity), 8)
#' #plot(m, color = "V1", colorPalette = heat.colors(50), trim = 60)
#' #plot(m, color = "V1", colorPalette = heat.colors(50), trim = 60, voxel = TRUE)
#'
#' # ================
#' # CROWN METRICS
#' # ================
#'
#' # Already tree-segmented point cloud
#' LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
#' trees <- readLAS(LASfile, filter = "-drop_z_below 0")
#'
#' metrics <- crown_metrics(trees, .stdtreemetrics)
#' #plot(metrics["Z"], pch = 19)
#'
#' metrics <- crown_metrics(trees, .stdtreemetrics, geom = "convex")
#' #plot(metrics["Z"])
#'
#' metrics <- crown_metrics(trees, .stdtreemetrics, geom = "bbox")
#' #plot(metrics["Z"])
#'
#' \donttest{
#' metrics <- crown_metrics(trees, .stdtreemetrics, geom = "concave")
#' #plot(metrics["Z"])
#' }
#' # ================
#' # ARGUMENT FILTER
#' # ================
#'
#' # Compute using only some points: basic
#' first = filter_poi(las, ReturnNumber == 1)
#' metrics = pixel_metrics(first, mean(Z), 20)
#'
#' # Compute using only some points: optimized
#' # faster and uses less memory. No intermediate object
#' metrics = pixel_metrics(las, mean(Z), 20, filter = ~ReturnNumber == 1)
#'
#' # Compute using only some points: best
#' # ~50% faster and uses ~10x less memory
#' las = readLAS(LASfile, filter = "-keep_first")
#' metrics = pixel_metrics(las, mean(Z), 20)
#'
#' # ================
#' # ARGUMENT BY_ECHO
#' # ================
#'
#' func = ~list(avgI = mean(Intensity))
#' echo = c("all", "first","multiple")
#'
#' # func defines one metric but 3 are computed respectively for: (1) all echo types,
#' # (2) for first returns only and (3) for multiple returns only
#' metrics <- pixel_metrics(las, func, 20, by_echo = echo)
#' #plot(metrics, col = heat.colors(25))
#'
#' cloud_metrics(las, func, by_echo = echo)
#'
#' \dontrun{
#' # ================
#' # TEMPLATE METRICS
#' # ================
#'
#' # a raster as template
#' template <- raster::raster(extent(las), nrow = 15, ncol = 15)
#' raster::crs(template) <- crs(las)
#' m <- template_metrics(las, fun1, template)
#' #plot(m, col = col)
#'
#' # a sfc_POLYGON as template
#' sfc <- sf::st_as_sfc(st_bbox(las))
#' template <- sf::st_make_grid(sfc, cellsize = 20, square = FALSE)
#' m <- template_metrics(las, fun1, template)
#' #plot(m)
#'
#' # a bbox as template
#' template <- st_bbox(las) + c(50,30,-50,-70)
#' plot(sf::st_as_sfc(st_bbox(las)), col = "gray")
#' plot(sf::st_as_sfc(template), col = "darkgreen", add = TRUE)
#' m <- template_metrics(las, fun2, template)
#' print(m)
#'
#' # ================
#' # CUSTOM METRICS
#' # ================
#'
#' # Define a function that computes custom metrics
#' # in an R&D perspective.
#' myMetrics = function(z, i) {
#'   metrics = list(
#'      zwimean = sum(z*i)/sum(i), # Mean elevation weighted by intensities
#'      zimean  = mean(z*i),       # Mean products of z by intensity
#'      zsqmean = sqrt(mean(z^2))) # Quadratic mean
#'
#'    return(metrics)
#' }
#'
#' # example with a stars template
#' template <- stars::st_as_stars(st_bbox(las), dx = 10, dy = 10)
#' m <- template_metrics(las, myMetrics(Z, Intensity), template)
#' #plot(m, col = col)
#' }
#' @name aggregate
#' @rdname aggregate
#' @md
NULL

#' @export
#' @rdname aggregate
template_metrics <- function(las, func, template, filter = NULL, by_echo = "all", ...)
{
  UseMethod("template_metrics", las)
}

#' @export
template_metrics.LAS <- function(las, func, template, filter = NULL, by_echo = "all", ...)
{
  . <- echo <- NULL

  # Defensive programming
  if (!inherits(las, "LAS")) stop("template metrics only supports LAS objects", call. = FALSE)
  formula <- tryCatch(lazyeval::is_formula(func), error = function(e) FALSE)
  if (!formula) func <- lazyeval::f_capture(func)
  echo_types <- c("all", "first", "intermediate", "lastofmany", "single", "multiple")
  stopifnot(all(by_echo %in% echo_types))

  dots <- list(...)
  pkg <- if (is.null(dots$pkg)) getOption("lidR.raster.default") else dots$pkg

  # Compute on different subsets of echos
  all_echos   <- any(by_echo == "all")
  split_echos <- any(by_echo != "all")

  # User-defined call parsing
  func   <- lazyeval::f_interp(func)
  call   <- lazyeval::as_call(func)

  # The way the user's expression is evaluated makes the scoping rule counterintuitive and
  # somewhat buggy. It gives precedence to lidR in all cases. For example, it will call an internal
  # function of lidR instead of its globalenv version if they have the same name. The following
  # tries to parse the expression and evaluates whether ambiguous definitions exist.
  call_names <- all.names(call)
  las_names  <- names(las)
  call_names <- call_names[!call_names %in% las_names]
  lapply(call_names, stop_if_ambiguous_definition)

  call <- deparse(call)
  call <- trimws(call)
  call <- paste0(call, collapse = " ")

  if (is.null(template)) template <- st_bbox(las)

  # Pre-computation of the groups
  grp  <- get_group(template, las)

  # If the template is not numeric it is a spatial object with a bbox
  # We check if the template matches the point cloud
  if (!is.numeric(template))
  {
    bbox_template <- sf::st_as_sfc(sf::st_bbox(template))
    bbox_template <- sf::st_set_crs(bbox_template, sf::NA_crs_)     # Workaround for RasterLayer
    bbox_template <- sf::st_set_crs(bbox_template, sf::st_crs(las)) # Workaround for RasterLayer
    bbox_las      <- sf::st_as_sfc(st_bbox(las))
    i <- sf::st_intersects(bbox_template, bbox_las)

    if (is(template, "raster_template"))
      template <- raster_materialize(template, pkg = pkg)

    if (length(i[[1]]) == 0L)
    {
      if (is_raster(template))
      {
        # For backward compatibility we return the template with NAs
        # It means that pixel_metrics and grid_metrics never fails
        warning("No points fall in the raster. Bounding boxes are not intersecting.", call. = FALSE)
        return(template)
      }

      stop("Bounding boxes of the template and the point-cloud are not intersecting.", call. = FALSE)
    }
  }

  # Aggregation of the point cloud
  if (all_echos)
    M1 <- metrics_classic(las, call, grp, filter)

  if (split_echos)
    M2 <- metrics_by_echo_type(las, call, grp, filter, by_echo)

  # Merge subsets
  if (all_echos && split_echos)
    M <- rbind(M1, M2)
  else if (all_echos)
    M <- M1
  else if (split_echos)
    M <- M2
  else
    stop("Internal error") #nocov

  data.table::setDT(M)
  data.table::setkey(M, echo)

  # Cast the table into appropriate objects
  computed_echos <- sort(match(by_echo, echo_types) - 1L)
  output <- vector("list", length(computed_echos))
  i <- 1
  for (k in computed_echos)
  {
    echo_type <- echo_types[k+1]
    m <- M[.(k)]
    cells <- m[[1]]
    m[, cells := NULL]
    m[, echo := NULL]
    if (k > 0) names(m) <- paste0(names(m), "." , echo_type)
    output[[i]] <- assign_to_template(template, cells, m, pkg = pkg)
    i <- i + 1
  }
  output <- merge_list(template, output)

  return(output)
}

# ==== CORE ====

metrics_classic = function(las, call, cells, filter)
{
  .BY <- NULL
  las@data[["cells"]] <- cells
  las@data[["echo"]] <- LASALLECHOS

  verbose("Computing metrics for all echos...")

  if (is.null(filter))
  {
    verbose("Argument filter is NULL...")
    cmd <- paste0('las@data[, if (!anyNA(.BY)) ', call, ', by = .(cells, echo)]')
  }
  else
  {
    filter  <- lasfilter_(las, list(filter))
    verbose(glue::glue("Argument filter retained {sum(filter)} points..."))
    cmd <- paste0('las@data[filter, if (!anyNA(.BY)) ', call, ', by = .(cells, echo)]')
  }

  metrics <- eval(parse(text = cmd))

  # This may append because new versions of data.table are more flexible than before
  if (any(duplicated(metrics[[1]])))
    stop("Duplicated elements found. At least one of the metrics was not a number. Each metric should be a single number.", call. = FALSE)

  return(metrics)
}

metrics_by_echo_type = function(las, call, cells, filter, by_echo)
{
  .BY <- echo <- . <- NULL

  header <- as.list(header(las))

  # Somewhat aggressive defensive programming
  #nocov start
  if (is.null(las$ReturnNumber) || is.null(las$NumberOfReturns))
    stop("ReturnNumber and NumberOfReturns must be loaded to perform computation by echo type", call. = FALSE)

  if (!rlas::is_valid_NumberOfReturns(las@data, header))
    stop("NumberOfReturns is not valid. See las_check(). Operation aborted.", call. = FALSE)

  if (!rlas::is_compliant_NumberOfReturns(las@data))
    stop("NumberOfReturns is not compliant with ASPRS specification. See las_check(). Operation aborted.", call. = FALSE)

  if (!rlas::is_valid_ReturnNumber(las@data, header))
    stop("ReturnNumber is not valid. See las_check(). Operation aborted.", call. = FALSE)

  if (!rlas::is_compliant_ReturnNumber(las@data))
    stop("ReturnNumber is not compliant with ASPRS specification. See las_check(). Operation aborted.", call. = FALSE)

  if(!rlas::is_compliant_ReturnNumber_vs_NumberOfReturns(las@data))
    stop("ReturnNumber and NumberOfReturns are not compatible. See las_check(). Operation aborted.", call. = FALSE)
  #nocov end

  echo_types = character(5)
  echo_types[LASFIRST] <- "first"
  echo_types[LASSINGLE] <- "single"
  echo_types[LASMULTIPLE] <- "multiple"
  echo_types[LASLASTOFMANY] <- "lastofmany"
  echo_types[LASINTERMEDIATE] <- "intermediate"
  echo_types1 <- echo_types[c(LASFIRST, LASINTERMEDIATE, LASLASTOFMANY)]
  echo_types2 <- echo_types[c(LASSINGLE, LASMULTIPLE)]
  echo_class  <- get_echo_type(las$ReturnNumber, las$NumberOfReturns)
  echo_class1 <- echo_class[[1]]
  echo_class2 <- echo_class[[2]]
  compute_type1 <- any(by_echo %in% echo_types1)
  compute_type2 <- any(by_echo %in% echo_types2)

  las@data[["cells"]] <- cells

  verbose("Computing metrics for selected echos...")

  if (is.null(filter))
  {
    verbose("Argument filter is NULL...")

    metrics1 <- NULL
    metrics2 <- NULL

    if (compute_type1)
    {
      verbose("Computing metrics for first intermediate and last...")
      las@data[["echo"]] <- echo_class1
      cmd <- paste0('las@data[, if (!anyNA(.BY)) ', call, ', by = .(cells, echo)]')
      metrics1 <- eval(parse(text = cmd))
    }

    if (compute_type2)
    {
      verbose("Computing metrics for single and multiple...")
      las@data[["echo"]] <- echo_class2
      cmd <- paste0('las@data[, if (!anyNA(.BY)) ', call, ', by = .(cells, echo)]')
      metrics2 <-eval(parse(text = cmd))
    }

    metrics12 <- rbind(metrics1, metrics2)
    data.table::setkey(metrics12, echo)
  }
  else
  {
    filter <- lasfilter_(las, list(filter))
    cells <- cells[filter]
    echo_class1 <- echo_class1[filter]
    echo_class2 <- echo_class2[filter]

    verbose(glue::glue("Argument filter retained {sum(filter)} points..."))

    metrics1 <- NULL
    metrics2 <- NULL

    if (compute_type1)
    {
      cmd <- paste0('las@data[filter, if (!anyNA(.BY)) ', call, ', by = list(cells = cells, echo = echo_class1)]')
      metrics1 <- eval(parse(text = cmd))
    }

    if (compute_type2)
    {
      cmd <- paste0('las@data[filter, if (!anyNA(.BY)) ', call, ', by = list(cells = cells, echo = echo_class2)]')
      metrics1 <- eval(parse(text = cmd))
    }

    metrics12 <- rbind(metrics1, metrics2)
    data.table::setkey(metrics12, echo)
  }

  # This may append because new versions of data.table are more flexible than before
  if (any(duplicated(metrics12[LASFIRST][[1]])))
    stop("Duplicated elements found. At least one of the metrics was not a number. Each metric should be a single number.", call. = FALSE)

  return(metrics12)
}

get_echo_type <- function(ReturnNumber, NumberOfReturns)
{
  n <- length(ReturnNumber)

  class1 <- rep(LASINTERMEDIATE, n)
  class1[ReturnNumber == 1L] <- LASFIRST
  class1[ReturnNumber == NumberOfReturns & ReturnNumber > 1L] <- LASLASTOFMANY

  class2 <- rep(LASMULTIPLE, n)
  class2[NumberOfReturns == 1L] <- LASSINGLE

  return(list(type1 = class1, type2 = class2))
}

# ==== GROUPING ====

get_group <- function(template, las) { UseMethod("get_group", template) }

get_group.stars <- function(template, las)
{
  return(raster_cell_from_xy(template, las$X, las$Y))
}

get_group.RasterLayer <- function(template, las)
{
  return(raster_cell_from_xy(template, las$X, las$Y))
}

get_group.SpatRaster <- function(template, las)
{
  return(raster_cell_from_xy(template, las$X, las$Y))
}

get_group.raster_template <- function(template, las)
{
  return(raster_cell_from_xy(template,las$X, las$Y))
}

get_group.sfc <- function(template, las)
{
  return(point_in_polygons(las, template))
}

get_group.bbox <- function(template, las)
{
  group <- integer(length(las$x))
  group[] <- NA_integer_
  group[las$X >= template$xmin & las$X <= template$xmax &
        las$Y >= template$ymin & las$Y <= template$ymax] <- 1L
  return(group)
}

get_group.numeric <- function(template, las)
{
  if (length(template) != length(las$X))
    stop("Length of template is not the size of the point cloud", call. = FALSE)

  return(template)
}

# ==== ASSIGN ====

assign_to_template <- function(template, cells, metrics, ...) { UseMethod("assign_to_template", template) }

assign_to_template.RasterLayer <- function(template, cells, metrics, ...)
{
  suppressWarnings(template[] <- NA_real_)
  nmetrics <- ncol(metrics)

  if (nmetrics == 1L)
  {
    template[cells] <- metrics[[1]]
    raster_names(template) <- names(metrics)
    return(template)
  }

  suppressWarnings(output <- raster::brick(template, nl = nmetrics))
  ncells <- raster_ncell(template)
  for (i in 1:nmetrics)
  {
    values <- vector(mode = class(metrics[[i]]), length = ncells)
    values[] <- NA_real_
    values[cells] <- metrics[[i]]
    output <- raster::setValues(output, values, layer = i)
  }

  raster_names(output) <- names(metrics)
  return(output)
}

assign_to_template.stars <- function(template, cells, metrics, ...)
{
  n <- ncol(metrics)
  s <- dim(template)

  if (n == 1L)
  {
    template[[1]][cells] <- metrics[[1]]
    raster_names(template) <- names(metrics)
    return(template)
  }

  bbox   <- sf::st_bbox(template)
  output <- stars::st_as_stars(bbox, nx = s[1], ny = s[2], value = NA_real_, nz = n)
  raster_names(output) <- names(metrics)
  for (i in 1:n) output[[1]][,,i][cells] <- metrics[[i]]
  return(output)
}

assign_to_template.SpatRaster <- function(template, cells, metrics, ...)
{
  nmetrics <- ncol(metrics)
  output <- terra::init(template, NA_real_)
  output[cells] <- metrics[[1]]

  if (nmetrics == 1L)
  {
    raster_names(output) <- names(metrics)
    return(output)
  }

  rlist <- vector("list", nmetrics)
  rlist[[1]] <- output

  for (i in 2:nmetrics)
  {
    terra::set.values(template, cells, metrics[[i]])
    rlist[[i]] <- terra::deepcopy(template)
  }

  output <- terra::rast(rlist)

  # Best option but add is not part of terra if library(terra) is not called
  #for (i in 2:nmetrics)
  #{
  #  terra::set.values(template, cells, metrics[[i]])
  #  add(output) <- template
  #}

  raster_names(output) <- names(metrics)
  return(output)
}

assign_to_template.raster_template <- function(template, cells, metrics, ...)
{
   dots <- list(...)
   pkg <- if (is.null(dots$pkg)) getOption("lidR.raster.default") else dots$pkg
   template <- raster_materialize(template, pkg = pkg)
   return(assign_to_template(template, cells, metrics, ...))
}

assign_to_template.sfc <- function(template, cells, metrics, ...)
{
  data.table::setDF(metrics)

  template <- sf::st_as_sf(template)
  for(j in names(metrics))
  {
    template[[j]] <- NA
    template[[j]][cells] <- metrics[[j]]
  }

  geom <- sf::st_geometry(template)
  template <- sf::st_drop_geometry(template)
  template <- sf::st_set_geometry(template, geom)
  return(template)
}

assign_to_template.bbox <- function(template, cells, metrics, ...)
{
  return(as.list(metrics))
}

assign_to_template.numeric <- function(template, cells, metrics, ...)
{
  GRPID <- NULL
  metrics[, GRPID := cells]
  data.table::setcolorder(metrics, c("GRPID"))
  data.table::setkey(metrics, GRPID)
  return(metrics)
}

# ==== MERGE ====

merge_list <- function(template, list)  { UseMethod("merge_list", template) }

merge_list.raster_template <- function(template, list)
{
  template <- raster_materialize(template)
  return(merge_list(template, list))
}

merge_list.RasterLayer <- function(template, list)
{
  list <- Filter(Negate(is.null), list)
  if (length(list) > 1)
    return(raster::brick(list))
  else
    return(list[[1]])
}

merge_list.SpatRaster <- function(template, list)
{
  list <- Filter(Negate(is.null), list)
  if (length(list) > 1)
    return(terra::rast(list))
  else
    return(list[[1]])
}

merge_list.stars <- function(template, list)
{
  list <- Filter(Negate(is.null), list)
  n <- length(list)

  if (n == 1) return(list[[1]])

  dims  <- sapply(list, dim)
  multilayer <- nrow(dims) == 3

  if (multilayer)
  {
    names <- unlist(lapply(list, function(x) stars::st_dimensions(x)$z$value))
    nlayers <- sum(dims[3,])
    output <- list[[1]]
    for (i in 2:n) output <- c(output, list[[i]], along = 3L)
  }
  else
  {
    names <- sapply(list, names)
    nlayers <- n
    bs <- raster_template(template)
    output <- stars::st_as_stars(sf::st_bbox(template), dx = bs$xres, dy = bs$yres, nz = nlayers, values = NA_real_)
    for (i in 1:n) output[[1]][,,i] <- list[[i]][[1]]
  }

  raster_names(output) <- names

  return(output)
}

merge_list.sfc <- function(template, list)
{
  list <- Filter(Negate(is.null), list)
  n <- length(list)

  if (n == 1) return(list[[1]])

  geom <- sf::st_geometry(list[[1]])
  output <- sf::st_drop_geometry(list[[1]])
  for (i in 2:length(list)) output <- cbind(output, sf::st_drop_geometry(list[[i]]))
  output <- sf::st_set_geometry(output, geom)
  return(output)
}


merge_list.bbox<- function(template, list)
{
  list <- Filter(Negate(is.null), list)
  if (length(list) > 1)
    return(do.call(c, list))
  else
    return(list[[1]])
}

merge_list.numeric <- function(template, list)
{
  list <- Filter(Negate(is.null), list)
  n <- length(list)

  if (n == 1) return(list[[1]])

  output <- list[[1]]
  for (i in 2:length(list)) output <- merge(output, list[[i]], all = TRUE)
  return(output[])

}


# ==== ERROR ====

stop_if_ambiguous_definition <- function(name)
{
  # Check if this name exists in lidR (internal or not)
  u <- tryCatch(utils::getFromNamespace(name, 'lidR'), error = function(e) FALSE)
  u <- is.function(u)

  # If FALSE, no ambiguity
  if (!u) return(invisible(NULL))

  # Check in the parent frame if the name has been defined
  e <- parent.frame(1)
  u <- tryCatch(mget(name, envir = e), error = function(e) FALSE)
  u <- !isFALSE(u)
  n <- environmentName(e)
  if (n == "") n <- "parent.frame(1)"

  # If TRUE, there is an ambiguity, two objects have the same name, lidR object will take
  # the precedence. We don't want that
  if (u) stop(glue::glue("The function '{name}' exists in the package lidR but is also defined in {n}. The scoping rules give precedence to lidR and the result may not be the one expected. Please rename your function."), call. = FALSE)

  # We do that while we did not reach R_GlobalEnv
  i <- 2
  while(!identical(parent.frame(i-1), globalenv()) & isFALSE(u))
  {
    e <- parent.frame(i)
    u <- tryCatch(mget(name, envir = e), error = function(e) FALSE)
    u <- !isFALSE(u)
    n <- environmentName(e)
    i <- i+1
  }

  # u is FALSE, we did not find ambiguous definitions
  if (!u) return(invisible(NULL))

  if (n == "") n <- glue::glue("parent.frame({i-1})")

  # u is TRUE, we found ambiguous definitions
  stop(glue::glue("The function '{name}' exists in the package lidR but is also defined in {n}. The scoping rules give precedence to lidR and the result may not be the one expected. Please rename your function."), call. = FALSE)
}

#' @param ctg an object of class \code{LAScatalog}
#' @param realignment \code{FALSE} or \code{list(res = x, start = c(y,z))}. Sometimes the chunk must
#' be aligned with a raster, for example to ensure the continuity of the output. If the chunk size is
#' 800 and the expected product is a raster with a resolution of 35, 800 and 35 are not compatible
#' and will create 2 different partial pixels on the edges. The realignment option forces the
#' chunk to fit the grid alignment.
#' @param plot logical. Displays the chunk pattern.
#' @export
#' @rdname engine
engine_chunks = function(ctg, realignment = FALSE, plot = opt_progress(ctg))
{
  assert_is_all_of(ctg, "LAScatalog")

  # Bounding box of the catalog
  xmin <- ymin <- xmax <- ymax <- 0

  # Put some options in variables with short names
  buffer <- opt_chunk_buffer(ctg)
  by_file <- opt_chunk_is_file(ctg)
  chunk_alignment <- opt_chunk_alignment(ctg)
  width <- opt_chunk_size(ctg)
  realign <- !isFALSE(realignment) && opt_wall_to_wall(ctg)
  realigned <- FALSE

  # New feature from v2.2.0 to not process some tiles
  processed <- ctg$processed
  if (is.null(processed)) processed <- rep(TRUE, nrow(ctg@data))
  if (!is.logical(processed)) stop("The attribute 'processed' of the catalog is not logical.", call. = FALSE)
  files <- ctg$filename[processed]

  if (by_file)
  {
    # If processing by file the chunks are easy to make. It corresponds
    # to the bounding boxes of the files
    xmin <- ctg[["Min.X"]]
    xmax <- ctg[["Max.X"]]
    ymin <- ctg[["Min.Y"]]
    ymax <- ctg[["Max.Y"]]

    # Remove the files that are flagged for buffer only but no actual processing
    xmin <- xmin[processed]
    xmax <- xmax[processed]
    ymin <- ymin[processed]
    ymax <- ymax[processed]

    # Realignment happens when the chunks need a specific alignment (e.g. with a raster)
    # If a raster has a resolution of 16 m and the files are 1000 x 1000 meters
    # the file pattern implies the edge pixels will be split into two equal parts. A size of 1000 is
    # thus not valid and must be resized to 1008 to fit with the grid.
    if (realign)
    {
      res <- realignment$res
      xscale <- ctg[["X.scale.factor"]][processed]
      yscale <- ctg[["Y.scale.factor"]][processed]

      new_xmin <- round_any(xmin, res)
      new_ymin <- round_any(ymin, res)
      new_xmax <- round_any(xmax, res)
      new_ymax <- round_any(ymax, res)

      resize_xmin = new_xmin < xmin - xscale | new_xmin > xmin + xscale
      resize_ymin = new_ymin < ymin - yscale | new_ymin > ymin + yscale
      resize_xmax = new_xmax < xmax - xscale | new_xmax > xmax + xscale
      resize_ymax = new_ymax < ymax - yscale | new_ymax > ymax + yscale

      if (any(resize_xmin) || any(resize_ymin) || any(resize_xmax) || any(resize_ymax))
      {
        xmin[resize_xmin] <- new_xmin[resize_xmin] - ifelse(new_xmin[resize_xmin] >= xmin[resize_xmin], res, 0)
        ymin[resize_ymin] <- new_ymin[resize_ymin] - ifelse(new_ymin[resize_ymin] >= ymin[resize_ymin], res, 0)
        xmax[resize_xmax] <- new_xmax[resize_xmax] + ifelse(new_xmax[resize_xmax] <= xmax[resize_xmax], res, 0)
        ymax[resize_ymax] <- new_ymax[resize_ymax] + ifelse(new_ymax[resize_ymax] <= ymax[resize_ymax], res, 0)

        ns <- sum(resize_xmin | resize_ymin | resize_xmax | resize_ymax)
        nt <- length(resize_xmin)
        # message(glue::glue("The tiling pattern does not match with the resolution {res}. {ns}/{nt} chunks were extended to avoid partial pixels."))
        realigned <- TRUE
      }
    }
  }
  else
  {
    # Realignment happens when the chunks need a specific alignment (e.g. with a raster)
    # If a raster has a resolution of 16 m and the chunk are 500 x 500 meters
    # the chunk pattern implies the edge pixels will be split into two unequal parts.
    # A size of 500 is thus not valid and must be resized to 512 to fit with the grid.
    if (realign)
    {
      res <- realignment$res
      start <- realignment$start

      # If the chunk_size option does not match with the resolution
      new_width <- round_any(width, res)

      if (new_width != width)
      {
        message(glue::glue("Chunk size does not match with the resolution of the raster. Chunk size changed to {new_width} instead of {width} to ensure continuity of the output."))
        width <- new_width
        realigned <- TRUE
      }

      # If the alignment of the chunks does not match the start point of the raster
      new_chunk_alignment <- abs((chunk_alignment - start)) %% res + chunk_alignment
      if (any(new_chunk_alignment != chunk_alignment))
      {
        message(glue::glue("Alignment of the chunks does not match with the starting points of the raster. Alignment changed to ({new_chunk_alignment[1]}, {new_chunk_alignment[2]}) to ensure continuity of the output."))
        chunk_alignment <- new_chunk_alignment
        realigned <- TRUE
      }
    }

    # Bounding box of the catalog
    bbox <- with(ctg@data, c(min(Min.X), min(Min.Y), max(Max.X), max(Max.Y)))

    # Shift to align the grid
    shift <- numeric(2)
    shift[1] <- (bbox[1] - chunk_alignment[1]) %% width
    shift[2] <- (bbox[2] - chunk_alignment[2]) %% width

    # Generate coordinates of bottom-left corner
    xmin <- seq(bbox[1] - shift[1], bbox[3], width)
    ymin <- seq(bbox[2] - shift[2], bbox[4], width)
    grid <- expand.grid(xmin = xmin, ymin = ymin)

    xmin <- grid$xmin
    ymin <- grid$ymin
    xmax <- xmin + width
    ymax <- ymin + width
  }
  verbose("Creating a set of clusters for the catalog...")

  # Generate center and width of each chunk
  xcenter <- (xmin + xmax)/2
  ycenter <- (ymin + ymax)/2
  width   <- xmax - xmin
  height  <- ymax - ymin

  # Creation of a set of LASclusters from the rectangles
  if (by_file && buffer <= 0 && realigned == FALSE)
  {
    filenames <- ctg$filename[processed]

    clusters = lapply(1:length(xcenter), function(i)
    {
      center  <- list(x = xcenter[i], y = ycenter[i])
      cluster <- LAScluster(center, width[i], height[i], buffer, LIDRRECTANGLE, filenames[i], "noname", crs = st_crs(ctg), ctg@index)

      cluster@select <- ctg@input_options$select
      cluster@filter <- paste(cluster@filter, ctg@input_options$filter)

      return(cluster)
    })
  }
  else
  {
    bboxes   <- st_make_bboxes(xcenter - width/2, xcenter + width/2, ycenter - height/2, ycenter + height/2)
    clusters <- suppressWarnings(engine_index(ctg, bboxes, LIDRRECTANGLE, buffer, processed, TRUE, by_file))
    clusters <- clusters[!sapply(clusters, is.null)]
  }

  # Post-process the clusters

  # Specific optimisation to remove the "-inside" filter
  if (by_file && buffer == 0 && realigned == FALSE)
  {
    clusters <- lapply(clusters, function(x)
    {
      x@filter <- ctg@input_options$filter
      return(x)
    })
  }

  # Record the path indicating where to write the raster if requested
  if (opt_output_files(ctg) != "")
  {
    clusters <- lapply(seq_along(clusters), function(i)
    {
      X         <- list()
      X$ID      <- i
      X$XCENTER <- format(clusters[[i]]@center$x, scientific = F)
      X$YCENTER <- format(clusters[[i]]@center$y, scientific = F)
      X$XLEFT   <- format(clusters[[i]]@bbox[1], scientific = F)
      X$XRIGHT  <- format(clusters[[i]]@bbox[3], scientific = F)
      X$YBOTTOM <- format(clusters[[i]]@bbox[2], scientific = F)
      X$YTOP    <- format(clusters[[i]]@bbox[4], scientific = F)

      usefilename <- grepl("\\{ORIGINALFILENAME\\}",  opt_output_files(ctg))

      if (usefilename && !by_file)
        stop("The template {ORIGINALFILENAME} makes sense only when processing by file (chunk size = 0). It is undefined otherwise.", call. = FALSE)

      if (by_file)
        X$ORIGINALFILENAME <- tools::file_path_sans_ext(basename(files[i]))

      filepath  <- paste0(glue::glue_data(X, opt_output_files(ctg)))
      n         <- length(filepath)

      if (n > 1)
        stop(glue::glue("Ill-formed template string in the catalog: {n} filenames were generated for each chunk"), call. = FALSE)

      clusters[[i]]@save <- filepath
      return(clusters[[i]])
    })
  }

  # Record the alternative directories for remote computation
  if (ctg@input_options[["alt_dir"]][1] != "")
  {
    clusters <- lapply(seq_along(clusters), function(i)
    {
      clusters[[i]]@alt_dir <- ctg@input_options[["alt_dir"]]
      return(clusters[[i]])
    })
  }

  # New feature 3.2.0: skip some chunks (allow to restart a code that crashed)
  if (!is.null(ctg@chunk_options[["drop"]]) && length(ctg@chunk_options[["drop"]]) > 0)
    clusters <- clusters[-ctg@chunk_options[["drop"]]]

  # Plot the catalog and the clusters if progress
  if (plot)
  {
    xrange = c(min(xmin, ctg[["Min.X"]]), max(xmax, ctg[["Max.X"]]))
    yrange = c(min(ymin, ctg[["Min.Y"]]), max(ymax, ctg[["Max.Y"]]))
    title  = "Pattern of chunks"
    plot.LAScatalog(ctg, mapview = FALSE, chunk_pattern = FALSE, main = title, xlim = xrange, ylim = yrange)

    lapply(clusters, function(x)
    {
      graphics::rect(x@bbox[1], x@bbox[2], x@bbox[3], x@bbox[4], border = "red")

      if (x@buffer != 0)
        graphics::rect(x@bbbox[1], x@bbbox[2], x@bbbox[3], x@bbbox[4], border = "darkgreen", lty = "dotted")
    })
  }

  return(clusters)
}

#' Smooth a point cloud
#'
#' Point cloud-based smoothing algorithm. Two methods are available: average within a window and
#' Gaussian smooth within a window. The attribute \code{Z} of the returned LAS object is the smoothed Z.
#' A new attribute \code{Zraw} is added to store the original values and can be used to restore the
#' point cloud with \code{unsmooth_height}.
#'
#' This method does not use raster-based methods to smooth the point cloud. This is a true point cloud
#' smoothing. It is not really useful by itself but may be interesting in combination with filters,
#' for example to develop new algorithms.
#'
#' @param las An object of class \code{LAS}
#' @param size numeric. The size of the windows used to smooth.
#' @param method character. Smoothing method. Can be 'average' or 'gaussian'.
#' @param shape character. The shape of the windows. Can be circle or square.
#' @param sigma numeric. The standard deviation of the gaussian if the method is gaussian.
#'
#' @return An object of the class \code{LAS}.
#'
#' @export
#'
#' @examples
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' las <- readLAS(LASfile, select = "xyz")
#'
#' las <- decimate_points(las, highest(1))
#' #plot(las)
#'
#' las <- smooth_height(las, 5, "gaussian", "circle", sigma = 2)
#' #plot(las)
#'
#' las <- unsmooth_height(las)
#' #plot(las)
smooth_height = function(las, size, method = c("average", "gaussian"), shape = c("circle", "square"), sigma = size/6)
{
  stopifnotlas(las)
  assert_is_a_number(size)
  assert_all_are_positive(size)
  assert_is_a_number(sigma)
  assert_all_are_positive(sigma)
  method <- match.arg(method)
  shape  <- match.arg(shape)

  if (method == "average") method <- 1  else method <- 2
  if (method == "circle") shape   <- 1  else shape  <- 2

  force_autoindex(las) <- LIDRGRIDPARTITION
  Zs <- C_smooth(las, size, method, shape, sigma, getThread())
  fast_quantization(Zs, las[["Z scale factor"]], las[["Z offset"]])

  if (!"Zraw" %in% names(las))
    las@data[["Zraw"]] <- las@data[["Z"]]

  las@data[["Z"]] <- Zs
  return(las)
}

#' @export
#' @rdname smooth_height
unsmooth_height = function(las)
{
  stopifnotlas(las)
  Z <- Zraw <- NULL

  if ("Zraw" %in% names(las))
  {
    las@data[["Z"]] <- las@data[["Zraw"]]
    las@data[["Zraw"]] <- NULL
  }
  else
    stop("No attribute named 'Zraw' found. Unsmoothing is not possible.")

  return(las)
}

context("smooth_height")

data = data.frame(X = c(0,10,20), Y = c(0,0,0), Z = c(0,1,2))
las = LAS(data)

test_that("smooth_height works with average", {

  las <- smooth_height(las, 1, "average")
  expect_equal(las@data$Z, las@data$Zraw)

  las <- unsmooth_height(las)

  las <- smooth_height(las, 20, "average")
  expect_equal(las@data$Z, c(0.5, 1, 1.5))

  las <- unsmooth_height(las)

  las <- smooth_height(las, 50, "average")
  expect_equal(las@data$Z, c(1, 1, 1))

  unsmooth_height(las)
})

test_that("smooth_height works with gaussian", {

  las <- smooth_height(las, 1, "gaussian", sigma = 2)
  expect_equal(las@data$Z, las@data$Zraw)

  las <- unsmooth_height(las)

  las <- smooth_height(las, 20, "gaussian", sigma = 2)
  expect_equal(las@data$Z, las@data$Zraw, tolerance = 0.001)

  las <- unsmooth_height(las)

  las <- smooth_height(las, 20, "gaussian", sigma = 10)

  w1 = 1/(20*pi)*exp(-100/200)
  w2 = 1/(20*pi)
  z1 = (w2*data$Z[1] + w1*data$Z[2])/(w1+w2)
  z3 = (w1*data$Z[2] + w2*data$Z[3])/(w1+w2)
  expected = c(z1, 1, z3)
  lidR:::fast_quantization(expected, las@header@PHB$`Z scale factor`, las@header@PHB$`Z scale factor`)
  expect_equal(las@data$Z, expected)

  las <- unsmooth_height(las)
})

#' Concave and convex hulls for LAS objects
#'
#' Concave and convex hulls for LAS objects. `st_convex_hull` extends `sf::st_convex_hull` for LAS
#' objects. Both functions return a `sfc_POLYGON`. `concaveman` is very a fast 2D concave hull algorithm
#' for a set of points.
#'
#' The concaveman algorithm is based on ideas from Park and Oh (2012). A first implementation in
#' JavaScript was proposed by Vladimir Agafonkin in \href{https://github.com/mapbox/concaveman}{mapbox}.
#' This implementation dramatically improved performance over the one stated in the paper
#' using a spatial index. The algorithm was then ported to R by Joël Gombin in the R package
#' \href{https://github.com/joelgombin/concaveman}{concaveman} that runs the JavaScript
#' implementation proposed by Vladimir Agafonkin. Later, a C++ version of Vladimir Agafonkin's
#' JavaScript implementation was proposed by Stanislaw Adaszewski in
#' \href{https://github.com/sadaszewski/concaveman-cpp}{concaveman-cpp}. This concaveman
#' function uses Stanislaw Adaszewski's C++ code making the concaveman algorithm an
#' order of magnitude (up to 50 times) faster than the Javascript version.
#'
#' @param x,y An object of class LAS or XY coordinates of points in case of `concaveman`. This can be
#' specified as two vectors x and y, a 2-column matrix x, a list with two components, etc.
#' @param method string. currently supports "concaveman".
#' @param ... Propagate to the method.
#' @param concavity numeric a relative measure of concavity. 1 results in a relatively detailed shape,
#' Infinity results in a convex hull. You can use values lower than 1, but they can produce pretty crazy
#' shapes.
#' @param length_threshold numeric. When a segment length is below this threshold, it stops being
#' considered for further detailed processing. Higher values result in simpler shapes.
#' @return A `sfc_POLYGON` from `sf` or a `data.frame` in the case of `concaveman`
#' @importFrom sf st_convex_hull
#' @name st_hull
#' @rdname st_hull
#' @md
#' @references Park, J.-S & Oh, S.-J. (2013). A New Concave Hull Algorithm and Concaveness Measure
#' for n-dimensional Datasets. Journal of Information Science and Engineering. 29. 379-392.
#' @examples
#' x <- runif(35)
#' y <- runif(35)
#' hull <- concaveman(x,y)
#' plot(x,y, asp = 1)
#' lines(hull, lwd = 3, col = "red")
#'
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' las = readLAS(LASfile, filter = "-drop_z_below 1")
#' hull = st_concave_hull(las, length_threshold = 10)
#' plot(hull)
NULL

#' @export
#' @rdname st_hull
st_concave_hull <- function(x, method = "concaveman",  ...) UseMethod("st_concave_hull")


#' @export
#' @rdname st_hull
st_convex_hull.LAS = function(x)
{
  chull <- convex_hull(x$X, x$Y)
  chull <- sf::st_polygon(list(as.matrix(chull)))
  chull <- sf::st_geometry(chull)
  sf::st_crs(chull) <- st_crs(x)
  return(chull)
}

#' @export
st_concave_hull.LAS = function(x, method = "concaveman",  ...)
{
  if (method == "concaveman")
  {
    chull <- concaveman(x$X, x$Y, ...)
    chull <- sf::st_polygon(list(as.matrix(chull)))
    chull <- sf::st_sfc(chull, crs = st_crs(x))
    if (!sf::st_is_valid(chull)) chull <- sf::st_make_valid(chull)
    if (sf::st_geometry_type(chull) == "MULTIPOLYGON")
    {
      chull = sf::st_cast(chull, "POLYGON")
      warning("The output of 'st_concave_hull' was a MULTIPOLYGON converted into multiple POLYGON(s)")
    }
    return(chull)
  }

  if (method == "concavetin")
  {
    chull <- concavetin(x, ...)
    chull <- sf::st_polygon(list(as.matrix(chull)))
    chull <- sf::st_sfc(chull, crs = st_crs(x))
    if (!sf::st_is_valid(chull)) chull <- sf::st_make_valid(chull)
    return(chull)
  }

  stop("Method not supported. It should be one of 'concaveman' or 'concavetin'")
}

#' @rdname st_hull
#' @export
concaveman <- function(x, y = NULL, concavity = 2, length_threshold = 0)
{
  stopifnot(is.numeric(concavity), is.numeric(length_threshold), length(concavity) == 1L, length(length_threshold) == 1L)

  if (is.null(y))
  {
    if (!is.matrix(x) && !is.data.frame(x) && !is.list(x)) stop("A matrix, a list, or a data.frame is expected.")
    if (dim(x)[2] != 2) stop("Two columns are expected.")

    if (is.matrix(x))
    {
      y <- x[,2]
      x <- x[,1]
    } else
    {
      y <- x[[2]]
      x <- x[[1]]
    }
  }

  stopifnot(is.numeric(x), is.numeric(y), length(x) == length(y), length(x) >= 3L)

  h <- grDevices::chull(x, y) - 1
  return(cpp_concaveman(x, y, concavity, length_threshold, h))
}

concavetin <- function(las, max_length = 8)
{
  assert_is_a_number(max_length)
  assert_all_are_non_negative(max_length)

  if (npoints(las) < 3) stop("Impossible to triangulate fewer than 3 points.", call. = FALSE)

  # Triangulate the points and trim small triangles
  D <- tDelaunay(las, trim = max_length)
  X <- st_coordinates(las, z = FALSE)
  #geometry::trimesh(D,X, asp = 1)

  # Compute the contours of the triangulation
  hull <- tin_exterior_ring(D, X)
  return(hull)
}

tin_exterior_ring = function(D, X, s = 5)
{
  used <- N <- . <- P1 <- P2 <- NULL

  # Get the contours of the partial triangulation
  # *********************************************
  D <- data.table::as.data.table(D)

  # Order the Delaunay triangulation clockwise
  D = t(apply(D, 1, function(x)
  {
    p1 <- c(X[x[1], 1], X[x[1], 2], 1)
    p2 <- c(X[x[2], 1], X[x[2], 2], 1)
    p3 <- c(X[x[3], 1], X[x[3], 2], 1)
    M  <- rbind(p1, p2, p3)
    clockwise <-  det(M) < 0
    if (clockwise) return(x)
    return(c(x[3], x[2], x[1]))
  }))

  # Compute the vertices (edge of each triangle)
  n  <- nrow(D)
  p1 <- integer(n*3)
  p2 <- integer(n*3)
  for (i in 1:n)
  {
    j <- 3*i-2
    tri <- as.numeric(D[i,])
    p1[j]     <- tri[1]
    p2[j]     <- tri[2]
    p1[1 + j] <- tri[1]
    p2[1 + j] <- tri[3]
    p1[2 + j] <- tri[2]
    p2[2 + j] <- tri[3]
  }

  vertice  <- matrix(c(p1, p2), ncol = 2)
  svertice <- t(apply(vertice, 1, sort))
  svertice <- data.table::as.data.table(svertice)
  vertice  <- data.table::as.data.table(vertice)
  names(vertice) <- c("p1", "p2")

  # Filter the contour vertices. Contour vertices are those
  # that appear only once
  notcontour1  <- duplicated(svertice)
  notcontour2  <- duplicated(svertice, fromLast = TRUE)
  contour      <- vertice[!(notcontour1|notcontour2)]
  contour$used <- FALSE

  # Extract the polygons from the contour
  # *************************************
  # Initialization
  n  <- nrow(contour)
  p1 <- numeric(n)
  p2 <- numeric(n)

  # The first vertice initiates a polygon
  p1[1] <- contour$p1[1]
  p2[1] <- contour$p2[1]
  data.table::set(contour, 1L, 3L, TRUE)

  # Find the last vertice that closes the polygon
  j <- which(contour$p1 == p1[1] & contour$used == FALSE)
  s <- length(j) == 0L

  if (s) j <- which(contour$p2 == p1[1] & contour$used == FALSE)
  if (length(j) == 0L) stop("Internal error in tHull: please report this error (length(j) == 0L)")
  if (length(j) > 1) j <- j[1]

  if (s) {
    p1[n] <- contour$p1[j]
    p2[n] <- contour$p2[j]
  } else {
    p1[n] <- contour$p2[j]
    p2[n] <- contour$p1[j]
  }

  data.table::set(contour, j, 3L, TRUE)

  # Continue until we reach the last vertice
  for (i in 2:n)
  {
    j <- which(contour$p1 == p2[i - 1] & contour$used == FALSE)
    s <- length(j) == 0L

    if (s) j <- which(contour$p2 == p2[i - 1] & contour$used == FALSE)
    if (length(j) == 0L) stop("Internal Error: please report this error (tHull L147)")
    if (length(j) > 1) j <- j[1]

    if (!s)
    {
      p1[i] <- contour$p1[j]
      p2[i] <- contour$p2[j]
    }
    else
    {
      p1[i] <- contour$p2[j]
      p2[i] <- contour$p1[j]
    }

    data.table::set(contour, j, 3L, TRUE)

    if (p2[i] == p1[n]) {
      contour <- contour[used == FALSE]
      break
    }
  }

  orderedcontour <- data.table::data.table(p1, p2)
  orderedcontour <- orderedcontour[orderedcontour$p1 != 0,]

  XX <- data.frame(P1 = X[orderedcontour$p1,1], P2 = X[orderedcontour$p1,2])
  data.table::setDT(XX)

  # Fix self intersection
  XX[, N := .N , by = .(P1,P2)]
  XX[c(1, .N), N := 1]
  XX = XX[N == 1][, N := NULL][]

  # Close polygon
  XX <-rbind(XX, XX[1,])
  return(XX)
}

convex_hull = function(x, y)
{
  i <- grDevices::chull(x,y)
  i <- c(i, i[1])
  coords <- list(x = x[i], y = y[i])
  data.table::setDF(coords)
  return(coords)
}

area_convex_hull = function(x, y)
{
  stopifnot(length(x) == length(y))
  hull <- convex_hull(x, y)
  area <- polygon_area(hull$x, hull$y)
  return(area)
}

polygon_area = function(x, y)
{
  if (length(x) == 0 && length(y) == 0) return(0)
  if (!is.numeric(x) || !is.numeric(y) ) stop("Arguments 'x' and 'y' must be real")  # nocov
  if (length(x) != length(y)) stop("Argument 'x' and 'y' must be of same size") # nocov

  area <- 0
  j <- length(x)

  for (i in 1:j)
  {
    area <- area + (x[j] + x[i])*(y[j] - y[i]);
    j <- i;
  }

  area <- abs(area*0.5)
  return(area)
}

points = filter_poi(megaplot, Z > 2)
set.seed(42)
i = sample(1:npoints(points), 5000)
points = points[i]

n = 342

test_that("st_concave_hull works", {

  # Has been successfully tested on Solaris with rhub...
  # I'm tired to fix stuff on Solaris...
  skip_on_os("solaris")

  # skip on CRAN because it suddenly fails with M1mac. What can I do? How am I supposed to reproduce?
  skip_on_cran()

  pts = points
  hull <- st_concave_hull(pts)
  expect_is(hull, "sfc")
  hull <- sf::st_coordinates(hull)
  expect_equal(dim(hull), c(n,4))
  expect_equivalent(hull[1,], hull[n,])

  hull <- st_concave_hull(pts, length_threshold = 10)
  hull <- sf::st_coordinates(hull)
  expect_equal(dim(hull), c(175,4))
  expect_equivalent(hull[1,], hull[175,])

  # It works with 3 points
  pts = points[1:3]
  hull <- st_concave_hull(pts)
  hull <- sf::st_coordinates(hull)
  expect_equal(dim(hull), c(4,4))
  expect_equivalent(hull[1,], hull[4,])

  # It does not work with < 3 points
  pts = points[1:2]
  expect_error(st_concave_hull(pts))

  # Large concavity result in convex hull
  hull <- st_concave_hull(points, concavity = 1e8)
  hull <- sf::st_coordinates(hull)
  expect_equal(dim(hull), c(29,4))
  expect_equivalent(hull[1,], hull[29,])

  skip_on_cran()

  set.seed(42)
  i = sample(1:npoints(points), 1000)
  points = points[i]
  hull <- st_concave_hull(points, method = "concavetin", max_length = 20)
  hull <- sf::st_coordinates(hull)
  expect_equal(dim(hull), c(75,4))
  expect_equivalent(hull[1,], hull[75,])
})

test_that("concaveman works", {

  # Has been successfully tested on Solaris with rhub...
  # I'm tired to fix stuff on Solaris...
  skip_on_os("solaris")

  # skip on CRAN because it suddenly fails with M1mac. What can I do? How am I supposed to reproduce?
  skip_on_cran()

  # It works with matrix
  pts = as.matrix(sf::st_coordinates(points, z = FALSE))
  hull <- concaveman(pts)
  expect_equal(dim(hull), c(n,2))
  expect_equivalent(hull[1,], hull[n,])

  # It works with vectors
  x = points$X
  y = points$Y
  hull <- concaveman(x, y)
  expect_equal(dim(hull), c(n,2))
  expect_equivalent(hull[1,], hull[n,])
})

#' Rasterize a point cloud
#'
#' Rasterize a point cloud in different ways to compute a DTM, a CHM or a density map. Most
#' raster products can be computed with \link{pixel_metrics} but some are more complex and require
#' dedicated and optimized functions. See Details and Examples.
#'
#' \describe{
#' \item{`rasterize_terrain`}{Interpolates the ground points and creates a rasterized
#' digital terrain model. The algorithm uses the points classified as "ground" and "water"
#' (Classification = 2 and 9, respectively, according to
#' \href{https://community.asprs.org/leadership-restricted/leadership-content/public-documents/standards}{LAS file format specifications})
#' to compute the interpolation. How well the edges of the dataset are interpolated depends on the
#' interpolation method used. A buffer around the region of interest is always recommended to avoid
#' edge effects.}
#' \item{`rasterize_canopy`}{Creates a digital surface model (DSM) using several
#' possible algorithms. If the user provides a normalized point cloud, the output is indeed a canopy
#' height model (CHM).}
#' \item{`rasterize_density`}{Creates a map of the point density. If a "pulseID"
#' attribute is found, also returns a map of the pulse density.}
#' }
#'
#' @section Non-supported LAScatalog options:
#' The option `select` is not supported and not respected in `rasterize_*` because it is internally
#' known what is best to select.\cr
#' The option `chunk_buffer` is not supported and not respected in `rasterize_canopy` and
#' `rasterize_density` because it is not necessary.
#'
#' @template param-las
#' @param algorithm function. A function that implements an algorithm to compute a digital surface model
#' or a digital terrain model. \code{lidR} implements \link{p2r}, \link{dsmtin}, \link{pitfree}
#' for digital surface models, and \link{knnidw}, \link{tin}, and \link{kriging} for digital terrain
#' models (see respective documentation and examples).
#' @param ... Use `pkg = "terra|raster|stars"` to get an output in `SpatRaster`, `RasterLayer`
#' or `stars` format. Default is `getOption("lidR.raster.default")`.
#' @param res numeric. The size of a grid cell in point cloud coordinates units. Can also be
#'  `RasterLayer` or a `stars` or a `SpatRaster` used as layout.
#' @param shape By default the interpolation is made only within the `"convex"` hull of
#' the point cloud to get a DTM with the shape of the point cloud. This prevents meaningless
#' interpolations where there is no data. It can also be `"concave"` or `"bbox"`. It can also be an `sfc`
#' to define a polygon in which to perform the interpolation.
#' @param use_class integer vector. By default the terrain is computed by using ground points
#' (class 2) and water points (class 9).
#'
#' @return `RasterLayer` or a `stars` or a `SpatRaster` depending on the settings.
#' @name rasterize
#' @rdname rasterize
#' @md
#' @examples
#'
#' # =====================
#' # Digital Terrain Model
#' # =====================
#'
#' LASfile <- system.file("extdata", "Topography.laz", package="lidR")
#' las = readLAS(LASfile, filter = "-inside 273450 5274350 273550 5274450")
#' #plot(las)
#'
#' dtm1 = rasterize_terrain(las, algorithm = knnidw(k = 6L, p = 2))
#' dtm2 = rasterize_terrain(las, algorithm = tin())
#'
#' \dontrun{
#' dtm3 = rasterize_terrain(las, algorithm = kriging(k = 10L))
#'
#' plot(dtm1, col = gray(0:25/25))
#' plot(dtm2, col = gray(0:25/25))
#' plot(dtm3, col = gray(0:25/25))
#' plot_dtm3d(dtm1)
#' plot_dtm3d(dtm2)
#' plot_dtm3d(dtm3)
#' }
#'
#' # =====================
#' # Digital Surface Model
#' # =====================
#'
#' LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
#' las <- readLAS(LASfile, filter = "-inside 481280 3812940 481330 3812990")
#' col <- height.colors(15)
#'
#' # Points-to-raster algorithm with a resolution of 1 meter
#' chm <- rasterize_canopy(las, res = 1, p2r())
#' plot(chm, col = col)
#'
#' # Points-to-raster algorithm with a resolution of 0.5 meters replacing each
#' # point by a 20-cm radius circle of 8 points
#' chm <- rasterize_canopy(las, res = 0.5, p2r(0.2))
#' plot(chm, col = col)
#'
#' # Basic triangulation and rasterization of first returns
#' chm <- rasterize_canopy(las, res = 0.5, dsmtin())
#' plot(chm, col = col)
#'
#' # Khosravipour et al. pitfree algorithm
#' chm <- rasterize_canopy(las, res = 0.5, pitfree(c(0,2,5,10,15), c(0, 1.5)))
#' plot(chm, col = col)
#'
#' # ====================
#' # Digital Density Map
#' # ====================
#'
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' las <- readLAS(LASfile,  filter = "-inside 684800 5017800 684900 5017900")
#'
#' d <- rasterize_density(las, 5)
#' plot(d)
#'
#' las <- retrieve_pulses(las)
#' d <- rasterize_density(las)
#' plot(d)
NULL

#' Merge a point cloud with a source of spatial data
#'
#' Merge a point cloud with a source of spatial data. It adds an attribute along each point based on
#' a value found in the spatial data. Sources of spatial data can be a `SpatialPolygons*`, an `sf`/`sfc`,
#' a `Raster*`, a `stars`, or a `SpatRaster`.\cr
#' \itemize{
#' \item{`SpatialPolygons*`, `sf` and `sfc`: it checks if the points belongs within each polygon. If
#' the parameter `attribute` is the name of an attribute in the table of attributes it assigns
#' to the points the values of that attribute. Otherwise it classifies the points as boolean.
#' TRUE if the points are in a polygon, FALSE otherwise.}
#' \item{`RasterLayer`, single band `stars` or single layer `SpatRaster`: it attributes to each point
#' the value found in each pixel of the raster}.
#' \item{`RasterStack`, `RasterBrick`, multibands `stars` or multilayer `SpatRaster` must have 3
#' layers for RGB colors. It colorizes the point cloud with RGB values.}
#' }
#'
#' @param las An object of class `LAS`
#' @param source An object of class `SpatialPolygons*` or `sf` or `sfc` or `RasterLayer` or
#' `RasterStack` or `RasterBrick` or `stars`.
#' @param attribute character. The name of an attribute in the table of attributes or
#' the name of a new column in the LAS object. Not relevant for RGB colorization.
#'
#' @return a `LAS` object
#'
#' @export
#' @md
#'
#' @examples
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' shp     <- system.file("extdata", "lake_polygons_UTM17.shp", package = "lidR")
#'
#' las   <- readLAS(LASfile, filter = "-keep_random_fraction 0.1")
#' lakes <- sf::st_read(shp, quiet = TRUE)
#'
#' # The attribute "inlake" does not exist in the shapefile.
#' # Points are classified as TRUE if in a polygon
#' las    <- merge_spatial(las, lakes, "inlakes")     # New attribute 'inlakes' is added.
#' names(las)
#'
#' forest <- filter_poi(las, inlakes == FALSE)
#' #plot(forest)
#'
#' # The attribute "LAKENAME_1" exists in the shapefile.
#' # Points are classified with the values of the polygons
#' las <- merge_spatial(las, lakes, "LAKENAME_1")     # New column 'LAKENAME_1' is added.
#' names(las)
merge_spatial = function(las, source, attribute = NULL)
{
  UseMethod("merge_spatial", las)
}

#' @export
merge_spatial.LAS = function(las, source, attribute = NULL)
{
  if (is_raster(source) && !raster_nlayer(source) %in% c(1,3))
    stop("rasters must have 1 or 3 bands", call. = FALSE)

  if (inherits(source, "Spatial"))
    source <- sf::st_as_sf(source)

  rgb <- is_raster(source) && raster_nlayer(source) == 3L

  if (is(source, "sf") | is(source, "sfc"))
  {
    geom <- sf::st_geometry(source)
    if (!is(geom, "sfc_POLYGON") & !is(geom, "sfc_MULTIPOLYGON"))
      stop("Only polygon geometry types are supported", call. = FALSE)

    # crop to bbox of las to reduce computation time
    source <- st_crop_if_not_similar_bbox(source, las)
  }

  if (is(source, "sf"))
    values <- merge_sf(las, source, attribute)
  else if (is(source, "sfc"))
    values <- merge_sfc(las, source, attribute)
  else if (is_raster(source) && !rgb)
    values <- raster_value_from_xy(source, las$X, las$Y)
  else if (is_raster(source) && rgb)
    return(colorize_points(las, source))
  else
    stop("No method for this source format.")

  if (is.null(attribute))
    attribute <- "id"

  las <- add_attribute(las, values, attribute)
  return(las)
}

colorize_points = function(las, source)
{
  cells <- raster_cell_from_xy(source, las$X, las$Y)

  R <- raster_value_from_cells(source, cells, 1L)
  G <- raster_value_from_cells(source, cells, 2L)
  B <- raster_value_from_cells(source, cells, 3L)
  R <- as.integer(R)
  G <- as.integer(G)
  B <- as.integer(B)

  if (anyNA(R) | anyNA(G) | anyNA(B))
    stop("Some points were associated with an NA RGB color. RGB cannot be NA in a LAS object. Colorization aborted.", call. = FALSE)

  return(add_lasrgb(las, R, G, B))
}

merge_sf = function(las, source, attribute = NULL)
{
  npoints <- npoints(las)

  # No attribute is provided: assign the number of the polygon
  if (is.null(attribute))
  {
    method <- 0
    values <- rep(NA_integer_, npoints)
  }
  # The attribute is the name of an attribute in the attribute table: assign the value of the attribute
  else if (attribute %in% names(source))
  {
    method <- 1
    data   <- source[[attribute]]

    if (is.factor(data))
      values = factor(rep(NA_integer_, npoints), levels = levels(data))
    else if (is.integer(data))
      values = rep(NA_integer_, npoints)
    else if (is.logical(data))
      values = rep(NA, npoints)
    else if (is.numeric(data))
      values = rep(NA_real_, npoints)
    else if (is.character(data))
      values = rep(NA_character_, npoints)
    else
      stop(glue::glue("The attribute {attribute} in the table of attributes is not of a supported type."))
  }
  # The attribute is not the name of an attribute in the attribute table: assign a boolean value if
  # the point is in a polygon or not.
  else
  {
    method <- 2
    values <- logical(npoints)
  }

  if (nrow(source) == 0)
    return(values)

  verbose("Testing whether points fall in a given polygon...")

  idx <- point_in_polygons(las, sf::st_geometry(source))

  if (method == 1)
    values[!is.na(idx)] <- source[[attribute]][idx[!is.na(idx)]]
  else if (method == 2)
    values <- !is.na(idx)
  else
    values <- idx

  return(values)
}

merge_sfc = function(las, source, attribute = NULL)
{
  npoints <- npoints(las)

  # No attribute is provided: assign the number of the polygon
  if (is.null(attribute)) {
    method <- 0
    values <- rep(NA_integer_, npoints)
  } else {
    method <- 2
    values <- logical(npoints)
  }

  if (length(source) == 0)
    return(values)


  sfc <- sf::st_geometry(source)
  ids <- point_in_polygons(las, sfc)

  if (method == 2)
    values <- !is.na(ids)
  else
    values <- ids

  return(values)
}

#' @param area numeric. Area of the hexagons
#' @rdname aggregate
#' @export
hexagon_metrics <- function(las, func, area = 400, ...) { UseMethod("hexagon_metrics", las) }

#' @export
hexagon_metrics.LAS <- function(las, func, area = 400, ...)
{
  assert_is_a_number(area)
  assert_all_are_non_negative(area)

  # hexagon area : A = sqrt(3)/2*width^2
  # width = sqrt(2*A/sqrt(3))
  width <- sqrt(2*area/sqrt(3))
  sfc = sf::st_as_sfc(st_bbox(las))
  template = sf::st_make_grid(sfc, cellsize = width, square = FALSE)
  return(template_metrics(las, func, template, ...))
}

#' @export
hexagon_metrics.LAScluster <- function(las, func, area = 400, ...)
{
  x <- readLAS(las)
  if (is.empty(x)) return(NULL)
  metrics <- hexagon_metrics(x, func, area, ...)

  bbox <- st_bbox(las)
  sf::st_agr(metrics) <- "constant"
  centroid <- sf::st_centroid(metrics)
  centroid$ID <- 1:nrow(centroid)
  sf::st_agr(centroid) <- "constant"
  centroid <- sf::st_crop(centroid, bbox)
  return(metrics[centroid$ID,])
}

#' @export
hexagon_metrics.LAScluster <- function(las, func, area = 400, ...)
{
  is_formula <- tryCatch(lazyeval::is_formula(func), error = function(e) FALSE)
  if (!is_formula) func <- lazyeval::f_capture(func)
  globals <- future::getGlobalsAndPackages(func)

  options <- list(need_buffer = FALSE, drop_null = TRUE, globals = names(globals$globals), automerge = TRUE)
  output  <- catalog_apply(las, hexagon_metrics, func = func, area = area, ..., .options = options)
  return(output)
}

# ===== LMF ======

#' Individual Tree Detection Algorithm
#'
#' This function is made to be used in \link{locate_trees}. It implements an algorithm for tree
#' detection based on a local maximum filter. The windows size can be fixed or variable and its
#' shape can be square or circular. The internal algorithm works either with a raster or a point cloud.
#' It is deeply inspired by Popescu & Wynne (2004) (see references).
#'
#' @param ws numeric or function. Length or diameter of the moving window used to detect the local
#' maxima in the units of the input data (usually meters). If it is numeric a fixed window size is used.
#' If it is a function, the function determines the size of the window at any given location on the canopy.
#' By default function takes the height of a given pixel or point as its only argument and return the
#' desired size of the search window when centered on that pixel/point. This can be controled with
#' the `ws_args` parameter
#' @param hmin numeric. Minimum height of a tree. Threshold below which a pixel or a point
#' cannot be a local maxima. Default is 2.
#' @param shape character. Shape of the moving window used to find the local maxima. Can be "square"
#' or "circular".
#' @param ws_args list. Named list of argument for the function `ws` if `ws` is a function.
#'
#' @references
#' Popescu, Sorin & Wynne, Randolph. (2004). Seeing the Trees in the Forest: Using Lidar and
#' Multispectral Data Fusion with Local Filtering and Variable Window Size for Estimating Tree Height.
#' Photogrammetric Engineering and Remote Sensing. 70. 589-604. 10.14358/PERS.70.5.589.
#'
#' @export
#'
#' @family individual tree detection algorithms
#'
#' @examples
#' LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
#' las <- readLAS(LASfile, select = "xyzi", filter = "-inside 481250 3812980 481300 3813050")
#'
#' # =================
#' # point-cloud-based
#' # =================
#'
#' # 5x5 m fixed window size
#' ttops <- locate_trees(las, lmf(5))
#'
#' #plot(las) |> add_treetops3d(ttops)
#'
#' # variable windows size
#' f <- function(x) { x * 0.07 + 3}
#' ttops <- locate_trees(las, lmf(f))
#'
#' #plot(las) |> add_treetops3d(ttops)
#'
#' # Very custom variable windows size
#' f <- function(x, y, z) { x * 0.07 + y * 0.01 + z}
#' ws_args <- list(x = "Z", y = "Intensity", z = 3)
#' ttops <- locate_trees(las, lmf(f, ws_args = ws_args))
#'
#' # ============
#' # raster-based
#' # ============
#'
#' chm <- rasterize_canopy(las, res = 1, p2r(0.15), pkg = "terra")
#' ttops <- locate_trees(chm, lmf(5))
#'
#' plot(chm, col = height.colors(30))
#' plot(sf::st_geometry(ttops), add = TRUE, col = "black", cex = 0.5, pch = 3)
#'
#' # variable window size
#' f <- function(x) { x * 0.07 + 3 }
#' ttops <- locate_trees(chm, lmf(f))
#'
#' plot(chm, col = height.colors(30))
#' plot(sf::st_geometry(ttops), add = TRUE, col = "black", cex = 0.5, pch = 3)
#' @name itd_lmf
lmf = function(ws, hmin = 2, shape = c("circular", "square"), ws_args = "Z")
{
  shape <- match.arg(shape)
  circ  <- shape == "circular"
  ws    <- lazyeval::uq(ws)
  hmin  <- lazyeval::uq(hmin)
  ws_args  <- lazyeval::uq(ws_args)

  if (!is.numeric(ws) & !is.function(ws))
    stop("'ws' must be a number or a function", call. = FALSE)

  f = function(las)
  {
    assert_is_valid_context(LIDRCONTEXTITD, "lmf")

    if (is.function(ws))
    {
      args <- lapply(ws_args, function(x) if (x %in% names(las)) las@data[[x]] else x)
      ws <- do.call(ws, args)
      b <- las$Z < hmin
      ws[b] <- min(ws)

      n <- npoints(las)
      if (!is.numeric(ws)) stop("The function 'ws' did not return a correct output. ", call. = FALSE)
      if (any(ws <= 0))    stop("The function 'ws' returned negative or null values.", call. = FALSE)
      if (anyNA(ws))       stop("The function 'ws' returned NA values.",               call. = FALSE)
      if (length(ws) != n) stop("The function 'ws' did not return a correct output.",  call. = FALSE)
    }

    force_autoindex(las) <- LIDRGRIDPARTITION
    return(C_lmf(las, ws, hmin, circ, getThread()))
  }

  f <- plugin_itd(f, omp = TRUE, raster_based = FALSE)
  return(f)
}

# ===== MANUAL ======

#' Individual Tree Detection Algorithm
#'
#' This function is made to be used in \link{locate_trees}. It implements an algorithm for manual
#' tree detection. Users can pinpoint the tree top positions manually and interactively using the mouse.
#' This is only suitable for small-sized plots. First the point cloud is displayed, then the user is
#' invited to select a rectangular region of interest in the scene using the mouse button.
#' Within the selected region the highest point will be flagged as 'tree top' in the scene. Once all the trees
#' are labelled the user can exit the tool by selecting an empty region. Points can also be unflagged.
#' The goal of this tool is mainly for minor correction of automatically-detected tree outputs. \cr
#' **This algorithm does not preserve tree IDs from `detected` and renumber all trees. It also looses
#' all attributes**
#'
#' @param detected `SpatialPoints* or `sf/sfc_POINT` with  2 or 3D points of already found tree tops
#' that need manual correction. Can be NULL
#' @param radius numeric. Radius of the spheres displayed on the point cloud (aesthetic purposes only).
#' @param color character. Colour of the spheres displayed on the point cloud (aesthetic purposes only).
#' @param button Which button to use for selection. One of "left", "middle", "right". lidR using left
#' for rotation and right for dragging using one of left or right will disable either rotation or dragging
#' @param ... supplementary parameters to be passed to \link{plot}.
#'
#' @family individual tree detection algorithms
#'
#' @export
#' @md
#' @examples
#' \dontrun{
#' LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
#' las = readLAS(LASfile)
#'
#' # Full manual tree detection
#' ttops = locate_trees(las, manual())
#'
#' # Automatic detection with manual correction
#' ttops = locate_trees(las, lmf(5))
#' ttops = locate_trees(las, manual(ttops))
#' }
#' @name itd_manual
manual = function(detected = NULL, radius = 0.5, color = "red", button = "middle", ...) # nocov start
{
  f = function(las)
  {
    assert_is_valid_context(LIDRCONTEXTITD, "manual")

    . <- z <- X <- Y <- Z <- treeID <- NULL

    stopifnotlas(las)
    crs <- sf::st_crs(las)

    if (!interactive())
      stop("R is not being used interactively", call. = FALSE)

    if (is.null(detected))
    {
      apice <- data.table::data.table(X = numeric(0), Y = numeric(0), Z = numeric(0))
    }
    else
    {
      if (inherits(detected, "SpatialPoints"))
        detected <- sf::st_as_sf(detected)

      if (!is(detected, "sf") & !is(detected, "sfc"))
        stop("'apice' is not a SpatialPointsDataFrame or sf")

      coords   <- sf::st_coordinates(detected)
      u   <- coords[,1]
      v   <- coords[,2]

      if (ncol(coords) == 3)
        w <- coords[,3]
      else
        w <- detected[["Z"]]

      apice <- data.table::data.table(X = u, Y = v, Z = w)
    }

    minx <- min(las$X)
    miny <- min(las$Y)

    las@data <- las@data[, .(X, Y, Z)]
    las@data[, X := X - minx]
    las@data[, Y := Y - miny]
    apice$X = apice$X - minx
    apice$Y = apice$Y - miny

    plot.LAS(las, ..., clear_artifacts = FALSE)

    id = numeric(nrow(apice))

    # It's very slow to redraw after every sphere, so
    # turn off updates for a bit
    saveSkip = rgl::par3d(skipRedraw = TRUE)
    on.exit(rgl::par3d(saveSkip))  # Just in case of error

    for (i in 1:nrow(apice))
      id[i] = rgl::spheres3d(apice$X[i], apice$Y[i], apice$Z[i], radius = radius, color = color)

    # Now restore drawing mode
    rgl::par3d(saveSkip)
    on.exit() # exit restore no longer needed

    apice$id <- id

    repeat
    {
      # Select a region
      f <- rgl::select3d(button = button)

      # Get the apices in the selected region
      i <- if (nrow(apice) > 0) f(apice) else FALSE

      # There are some apices in the selected region: remove them
      if (sum(i) > 0)
      {
        ii <- which(i == TRUE)
        rgl::pop3d(id = apice[ii]$id)
        apice <- apice[-ii]
      }
      # There is no apex in the selected region: find an apex
      else
      {
        # Get the points in the selected region
        i <- f(las@data)

        # There is 0 points is the region: exit the function
        if (sum(i) == 0)
          break;

        # There are some points: find the highest one and add it to the list of apices
        pts     <- las@data[i, .(X,Y,Z)]
        apex    <- unique(pts[pts$Z == max(pts$Z)])[1]
        apex$id <- as.numeric(rgl::spheres3d(apex$X, apex$Y, apex$Z, radius = radius, color = color))
        apice   <- rbind(apice, apex)
      }
    }

    rgl::close3d()

    apice[, id := NULL]
    apice[, treeID := 1:.N]
    apice[, X := X + minx]
    apice[, Y := Y + miny]
    apice[, z := Z]
    output <- sf::st_as_sf(apice, coords = c("X", "Y", "z"), crs = crs)
    output = output[, c(2,1,3)]
    return(output)
  }

  f <- plugin_itd(f, omp = FALSE, raster_based = FALSE)
  return(f)
} # nocov end

#' @export
#' @rdname rasterize
rasterize_canopy = function(las, res = 1, algorithm = p2r(), ...)
{
  UseMethod("rasterize_canopy", las)
}

#' @export
rasterize_canopy.LAS = function(las, res = 1, algorithm = p2r(), ...)
{
  # Defensive programming
  assert_is_algorithm(algorithm)
  assert_is_algorithm_dsm(algorithm)
  assert_las_is_not_empty(las)

  dots <- list(...)
  pkg <- if (is.null(dots$pkg)) getOption("lidR.raster.default") else dots$pkg

  # Compute the raster that encompass the point cloud
  layout <- if (!is_a_number(res)) raster_template(res) else raster_layout(las, res, format = "template")
  layout <- raster_materialize(layout, pkg = "stars")

  # Compute the elevation for each cells
  lidR.context <- "rasterize_canopy"
  z <- algorithm(las, layout)

  # Quantize
  z <- round(z, 3)

  layout <- raster_layout(las, res)
  layout <- raster_materialize(layout, pkg = pkg)
  layout <- raster_set_values(layout, z)
  raster_names(layout) <- "Z"
  return(layout)
}

#' @export
rasterize_canopy.LAScluster = function(las, res = 1, algorithm = p2r(), ...)
{
  x <- readLAS(las)
  if (is.empty(x)) return(NULL)

  # crop the raster to the extent of the chunk because the raster can be a very large proxy
  # that will be materialized by rasterize_terrain LAS
  if (is_raster(res)) res <- raster_crop(res, st_adjust_bbox(x, raster_res(res)))

  chm <- rasterize_canopy(x, res, algorithm, ...)
  chm <- raster_crop(chm, st_bbox(las))
  return(chm)
}

#' @export
rasterize_canopy.LAScatalog = function(las, res = 1, algorithm = p2r(), ...)
{
  # Defensive programming
  assert_is_algorithm(algorithm)
  assert_is_algorithm_dsm(algorithm)

  if (is_a_number(res))
    assert_all_are_non_negative(res)
  else if (!raster_is_supported(res))
    stop("'res' must be a number of a raster.", call. = FALSE)

  # Compute the alignment options including the case when res is a RasterLayer/stars/terra
  alignment <- raster_alignment(res)

  # subset the collection to the size of the layout (if any)
  if (!is_a_number(res)) las <- catalog_intersect(las, res)

  # Enforce some options
  opt_select(las) <- "xyzr"
  if (opt_wall_to_wall(las)) opt_chunk_buffer(las) <- 2


  if (opt_chunk_size(las) > 0 && opt_chunk_size(las) < 2*alignment$res)
    stop("The chunk size is too small. Process aborted.", call. = FALSE)

  # Processing
  options <- list(need_buffer = TRUE, drop_null = TRUE, raster_alignment = alignment, automerge = TRUE)
  output  <- catalog_apply(las, rasterize_canopy, res = res, algorithm = algorithm, ..., .options = options)
  return(output)
}

#' Reconstruct the trajectory of the LiDAR sensor using multiple returns
#'
#' Use multiple returns to estimate the positioning of the sensor by computing the intersection in
#' space of the line passing through the first and last returns. To work, this function requires a
#' dataset where the 'gpstime', 'ReturnNumber', 'NumberOfReturns' and 'PointSourceID' attributes are
#' properly populated, otherwise the output may be incorrect or weird. For LAScatalog processing
#' it is recommended to use large chunks and large buffers (e.g. a swath width). The point cloud must
#' not be normalized.
#'
#' @section Non-supported LAScatalog options:
#' The option `select` is not supported and not respected  because it is internally known what is the
#' best to select\cr
#' The option `output_files` is not supported and not respected because the output must be post-processed
#' as a whole
#'
#' @section Test of data integrity:
#' In theory, sensor tracking is a simple problem to solve as long as each pulse is properly
#' identified from a well-populated dataset. In practice, many problems may arise from datasets that are populated
#' incorrectly. Here is a list of problems that may happen. Those with a * denote problems already encountered and
#' internally checked to remove weird points:
#' \itemize{
#' \item 'gpstime' does not record the time at which pulses were emitted and thus pulses are not identifiable
#' \item *A pulse (two or more points that share the same gpstime) is made of points from different
#' flightlines (different PointSourceID). This is impossible and denotes an improperly populated PointSourceID
#' attribute.
#' \item 'ReturnNumber' and 'NumberOfReturns' are wrongly populated with either some ReturnNumber > NumberOfReturn
#'  or several first returns by pulses
#' }
#' For a given time interval, when weird points are not filtered, the position is not computed for this
#' interval.
#'
#' @template param-las
#' @param algorithm function. An algorithm to compute sensor tracking. \code{lidR} implements
#' \link{Roussel2020} and  \link{Gatziolis2019} (see respective documentation and examples).
#' @param extra_check boolean. Datasets are rarely perfectly populated, leading to unexpected errors.
#' Time-consuming checks of data integrity are performed. These checks can be skipped as they account
#' for an significant proportion of the computation time. See also section 'Tests of data integrity'.
#' @param thin_pulse_with_time numeric. In practice, it is not useful to compute the position using all
#' multiple returns. It is more computationally demanding but not necessarily more accurate. This keeps
#' only one pulse every x seconds. Set to 0 to use all multiple returns. Use 0 if the file has already
#' been read with \code{filter = "-thin_pulses_with_time 0.001"}.
#' @param multi_pulse logical. TRUE only for systems with multiple pulses. Pulse ID must be recorded
#' in the UserData attribute.
#' @return An sf object with POINT Z geometries. Information about the time interval and the score of
#' the positioning (according to the method used) are also in the table of attributes.
#'
#' @author Jean-Francois Bourdon & Jean-Romain Roussel
#' @export
#' @examples
#' # A valid file properly populated
#' LASfile <- system.file("extdata", "Topography.laz", package="lidR")
#' las = readLAS(LASfile)
#' #plot(las)
#'
#' # pmin = 15 because it is an extremely small file
#' # strongly decimated to reduce its size. There are
#' # actually few multiple returns
#' flightlines <- track_sensor(las, Roussel2020(pmin = 15))
#'
#' plot(las@header)
#' plot(sf::st_geometry(flightlines), add = TRUE)
#'
#' #plot(las) |> add_flightlines3d(flightlines, radius = 10)
#'
#' \dontrun{
#' # With a LAScatalog "-drop_single" and "-thin_pulses_with_time"
#' # are used by default
#' ctg = readLAScatalog("folder/")
#' flightlines <- track_sensor(ctg,  Roussel2020(pmin = 15))
#' plot(flightlines)
#' }
#' @family range
track_sensor <- function(las, algorithm, extra_check = TRUE, thin_pulse_with_time = 0.001, multi_pulse = FALSE)
{
  UseMethod("track_sensor", las)
}

#' @export
track_sensor.LAS <- function(las, algorithm, extra_check = TRUE, thin_pulse_with_time = 0.001, multi_pulse = FALSE)
{
  if (!"PointSourceID" %in% names(las))     stop("No 'PointSourceID' attribute found", call. = FALSE)
  if (!"gpstime" %in% names(las))           stop("No 'gpstime' attribute found", call. = FALSE)
  if (!"ReturnNumber" %in% names(las))      stop("No 'ReturnNumber' attribute found", call. = FALSE)
  if (!"NumberOfReturns" %in% names(las))   stop("No 'NumberOfReturns' attribute found", call. = FALSE)
  if (!any(las@data[["gpstime"]] != 0))          stop("'gpstime' attribute is not populated.", call. = FALSE)
  if (!any(las@data[["ReturnNumber"]] != 0))     stop("'ReturnNumber' attribute is not populated.", call. = FALSE)
  if (!any(las@data[["NumberOfReturns"]] != 0))  stop("'NumberOfReturns' attribute is not populated.", call. = FALSE)
  if (!any(las@data[["PointSourceID"]] != 0))    stop("'PointSourceID' attribute is not populated.", call. = FALSE)

  if (multi_pulse)
  {
    if (!"UserData" %in% names(las))
      stop("No 'UserData' attribute found", call. = FALSE)
  }

  assert_is_algorithm(algorithm)
  assert_is_algorithm_trk(algorithm)
  assert_is_a_bool(extra_check)
  assert_is_a_number(thin_pulse_with_time)
  assert_all_are_non_negative(thin_pulse_with_time)
  assert_is_a_bool(multi_pulse)

  lidR.context = "track_sensor"

  . <- X <- Y <- Z <- ReturnNumber <- NumberOfReturns <- PointSourceID <- pulseID <- gpstime <- UserData <-  missing_psi <- NULL

  data <- las@data

  # Get only the first and last returns of multiple returns
  data <- data[(ReturnNumber == NumberOfReturns | ReturnNumber == 1) & NumberOfReturns > 1]

  # Filter potentially duplicated points (#396)
  if (multi_pulse == FALSE)
    dup <- duplicated(data, by = c("X", "Y", "Z", "ReturnNumber", "NumberOfReturns"))
  else
    dup <- duplicated(data, by = c("X", "Y", "Z", "ReturnNumber", "NumberOfReturns", "UserData"))

  if (any(dup))
  {
    warning(glue::glue("There were {sum(dup)} duplicated points. They were removed for processing."), call. = FALSE)
    data <- data[!dup]
  }

  # Decimate the dataset by gpstime
  if (thin_pulse_with_time > 0)
  {
    ftime <- round_any(data$gpstime - min(data$gpstime), thin_pulse_with_time)
    times <- data[, first(gpstime), by = ftime]$V1
    data <- data[gpstime %in% times]
  }

  # Compute an ID for each pulse
  if (multi_pulse == FALSE)
    data.table::setorder(data, gpstime, ReturnNumber)
  else
    data.table::setorder(data, UserData, gpstime, ReturnNumber) ## set record order

  data$pulseID <- .lagisdiff(data[["gpstime"]])

  # Fix #392
  if (multi_pulse)
    data[["pulseID"]] = data[["pulseID"]] + data[["UserData"]]

  # Count the points per pulse. We kept only first and last so it should be always 2
  count <- fast_table(data$pulseID,  max(data$pulseID))

  # If some pulses still have more than 2 points (#327)
  more_than_2 = fast_countover(count, 2L)
  if (more_than_2 > 0)
    stop(glue::glue("After keeping only first and last returns of multiple returns pulses, {more_than_2} pulses still have more than 2 points. This dataset is corrupted and gpstime is likely to be invalid."), call. = FALSE)

  # Some edge points might not be paired. Removed them.
  ii    <- which(count == 1L)
  data  <- data[!pulseID %in% ii]

  if (extra_check)
  {
    tests <- data[, list(test1 = .N != 2L, test2 = PointSourceID[1] != PointSourceID[2]), by = .(pulseID)]

    unpaired_pulse  <- tests$test1
    multiple_source <- tests$test2

    # Does this really happen? #327 should have made this line obsolete
    if (any(unpaired_pulse))
      warning(glue::glue("{sum(unpaired_pulse)} pulses with multiple returns were not actually paired. The point cloud is likely to be incorrectly populated. These pulses were removed"), call. = FALSE) # nocov

    # This happens if two points share the same 'gpstime' but different 'PointSourceID'
    if (any(multiple_source))
      warning(glue::glue("{sum(multiple_source)} pulses (points with same gpstime) come from different flightlines. The point cloud is likely to be incorrectly populated. These pulses were removed"), call. = FALSE)

    ii    <- tests$pulseID[unpaired_pulse | multiple_source]
    data  <- data[!pulseID %in% ii]
  }

  P <- algorithm(data)

  # Check out if we have at least 2 sensor position by flightline #608
  tmp <- P[, .N, by = PointSourceID]
  ind <- which(tmp$N <= 1L)
  if (length(ind) > 0L)
  {
    str <- glue::glue_collapse(tmp$PointSourceID[ind], ", ", last = " and ")
    warning(glue::glue("Only one sensor position was found for PointSourceID {str}"), call. = FALSE)
  }

  # If no position found return an empty sf
  if (nrow(P) == 0)
  {
    i <- integer(1)
    n <- numeric(1)
    coord <- matrix(i, ncol = 2)
    data  <- data.frame(X = coord[,1], Y = coord[, 2], Z = i, gpstime = i, PointSourceID = n, SCORE = n)
    zero  <- sf::st_as_sf(data, coords = c("X", "Y", "Z"), crs = st_crs(las))
    zero  <- zero[0,]
    return(zero)
  }
  else
  {
    data.table::setcolorder(P, c("X", "Y", "Z", "gpstime", "PointSourceID", "SCORE"))

    # Adress issue 391
    psi <- unique(data$PointSourceID)
    missing  <- !psi %in% unique(P$PointSourceID)
    if (any(missing))
    {
      missing_psi = psi[missing]
      txtids <- paste(missing_psi, collapse = ";")
      missing_psi <- as.character(missing_psi)
      warning(glue::glue("Some swaths present in the point cloud (PointSourceID = [{txtids}]) did not produce any sensor location."), call. = FALSE)
    }
  }

  na <- is.na(P[["X"]])
  P  <- P[!na]
  P  <- sf::st_as_sf(P, coords = c("X", "Y", "Z"), crs = st_crs(las))
  comment(P) <- missing_psi

  if (sum(na) > 0)
    warning(glue::glue("Something went wrong in {sum(na)} bins. The point cloud is likely to be incorrectly populated in a way not handled internally. Positions had not been computed everywere."), call. = FALSE)

  return(P)
}

#' @export
track_sensor.LAScluster <- function(las, algorithm, extra_check = TRUE, thin_pulse_with_time = 0.001, multi_pulse = FALSE)
{
  x <- readLAS(las)
  if (is.empty(x)) return(NULL)
  pos  <- track_sensor(x, algorithm, extra_check, thin_pulse_with_time, multi_pulse)
  return(pos)
}

#' @export
track_sensor.LAScatalog <- function(las, algorithm, extra_check = TRUE, thin_pulse_with_time = 0.001, multi_pulse = FALSE)
{
  assert_is_a_number(thin_pulse_with_time)
  assert_all_are_non_negative(thin_pulse_with_time)

  if (multi_pulse == FALSE)
    opt_select(las) <- "xyzrntpa"
  else
    opt_select(las) <- "xyzrntpau"

  if (thin_pulse_with_time > 0)
    opt_filter(las) <- paste("-drop_single -thin_pulses_with_time", format(thin_pulse_with_time, scientific =  FALSE), opt_filter(las))
  else
    opt_filter(las) <- paste("-drop_single", opt_filter(las))

  if (opt_output_files(las) != "")
  {
    opt_output_files(las) <- ""
    warning("Saving intermediate results is disabled in 'track_sensor' because the output must be post-processed as a whole.", call. = F)
  }

  options <- list(need_buffer = TRUE, drop_null = TRUE, need_output_file = FALSE)
  output  <- catalog_sapply(las, track_sensor, algorithm = algorithm, extra_check = extra_check, thin_pulse_with_time = 0, multi_pulse = multi_pulse, .options = options)

  # Post-processing to remove duplicated positions selecting the one computed with the more pulses
  SCORE <- gpstime <- NULL
  data <- sf::st_drop_geometry(output)
  data.table::setDT(data)
  i <- data[, .I[which.max(SCORE)], by = gpstime]$V1
  data.table::setDF(data)

  ans <- output[i,]
  if (nrow(ans) == 0)  warning("0 sensor locations found. This may be caused by an unsufficient number of multiple returns.", call. = FALSE)
  return(ans)
}

#' Snags Segmentation Algorithm
#'
#' This function is made to be used in \link{segment_snags}. It implements an algorithms for snags segmentation
#' based on Wing et al (2015) (see references). This is an automated filtering algorithm that utilizes
#' three dimensional neighborhood lidar point-based intensity and density statistics to remove lidar
#' points associated with live trees and retain lidar points associated with snags.
#'
#' Note that this algorithm strictly performs a classification based on user input while
#' the original publication's methods also included a segmentation step and some pre-
#' (filtering for first and single returns only) and post-process (filtering for only the
#' snag classified points prior to segmentation) tasks which are now expected to be performed
#' by the user. Also, this implementation may have some differences compared with the original
#' method due to potential mis-interpretation of the Wing et al. manuscript, specifically
#' Table 2 where they present four groups of conditional assessments with their required
#' neighborhood point density and average BBPR values (BBPR = branch and bole point ratio;
#' PDR = point density requirement).\cr\cr
#' This algorithm attributes each point in the point cloud (\code{snagCls} column) into the
#' following five snag classes:
#' \itemize{
#' \item 0: live tree - not a snag\cr
#' \item 1: general snag - the broadest range of snag point situations\cr
#' \item 2: small snag - isolated snags with lower point densities\cr
#' \item 3: live crown edge snag - snags located directly adjacent or intermixing with live trees crowns \cr
#' \item 4: high canopy cover snag - snags protruding above the live canopy in dense conditions (e.g.,
#' canopy cover >= 55\%).
#' }
#'
#' @param neigh_radii numeric. A vector of three radii used in quantifying local-area centered
#' neighborhoods. See Wing et al. (2015) reference page 171 and Figure 4. Defaults are 1.5,
#' 1, and 2 for the sphere, small cylinder and large cylinder neighborhoods, respectively.
#'
#' @param low_int_thrsh numeric. The lower intensity threshold filtering value. See Wing
#' et al. (2015) page 171. Default is 50.
#'
#' @param uppr_int_thrsh numeric. The upper intensity threshold filtering value. See Wing
#' et al. (2015) page 171. Default is 170.
#'
#' @param pt_den_req numeric. Point density requirement based on plot-level point density
#' defined classes. See Wing et al. (2015) page 172. Default is 3.
#'
#' @param BBPRthrsh_mat matrix. A 3x4 matrix providing the four average BBPR (branch and bole
#' point ratio) values for each of the three neighborhoods (sphere, small cylinder and large
#' cylinder) to be used for conditional assessments and classification into the following four snag
#' classes: 1) general snag 2) small snag 3) live crown edge snag 4) high canopy
#' cover snag. See Wing et al. (2015) page 172 and Table 2. This matrix must be provided by
#' the user.
#'
#' @examples
#' LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
#' # Wing also included -keep_single
#' poi ="-keep_first -inside 481260 3812920 481310 3812960"
#' las <- readLAS(LASfile, select = "xyzi", filter = poi)
#'
#' # For the Wing2015 method, supply a matrix of snag BranchBolePtRatio conditional
#' # assessment thresholds (see Wing et al. 2015, Table 2, pg. 172)
#' bbpr_thresholds <- matrix(c(0.80, 0.80, 0.70,
#'                           0.85, 0.85, 0.60,
#'                           0.80, 0.80, 0.60,
#'                           0.90, 0.90, 0.55),
#'                           nrow =3, ncol = 4)
#'
#' # Run snag classification and assign classes to each point
#' las <- segment_snags(las, wing2015(neigh_radii = c(1.5, 1, 2), BBPRthrsh_mat = bbpr_thresholds))
#'
#' # Plot it all, tree and snag points...
#' #plot(las, color="snagCls", colorPalette = rainbow(5))
#'
#' # Filter and plot snag points only
#' snags <- filter_poi(las, snagCls > 0)
#' #plot(snags, color="snagCls", colorPalette = rainbow(5)[-1])
#'
#' # Wing et al's (2015) methods ended with performing tree segmentation on the
#' # classified and filtered point cloud using the watershed method
#'
#' @author
#' Implementation by Andrew Sánchez Meador & Jean-Romain Roussel
#'
#' @references
#' Wing, Brian M.; Ritchie, Martin W.; Boston, Kevin; Cohen, Warren B.; Olsen, Michael J. 2015.
#' Individual snag detection using neighborhood attribute filtered airborne lidar data. Remote
#' Sensing of Environment. 163: 165-179 https://doi.org/10.1016/j.rse.2015.03.013
#'
#' @export
#'
#' @family snags segmentation algorithms
#' @name snag_wing2015
wing2015 = function(neigh_radii = c(1.5,1,2), low_int_thrsh = 50, uppr_int_thrsh = 170, pt_den_req = 3, BBPRthrsh_mat = NULL)
{
  assert_is_numeric(neigh_radii)
  assert_all_are_in_closed_range(neigh_radii, 0, 10)
  assert_is_a_number(low_int_thrsh)
  assert_all_are_in_closed_range(low_int_thrsh, 0, 255)
  assert_is_a_number(uppr_int_thrsh)
  assert_all_are_in_closed_range(uppr_int_thrsh, 0, 255)
  assert_all_are_true(low_int_thrsh < uppr_int_thrsh)
  assert_is_a_number(pt_den_req)
  assert_all_are_in_open_range(pt_den_req, 0, 100)

  if (is.null(BBPRthrsh_mat))
    stop("Branch and bole point ratio thresholds matrix not supplied.")

  f = function(las)
  {
    assert_is_valid_context(LIDRCONTEXTSNG, "wing2015")
    return(C_Wing2015(las, neigh_radii, low_int_thrsh, uppr_int_thrsh, pt_den_req, BBPRthrsh_mat, getThread()))
  }

  f <- plugin_snag(f, omp = TRUE)
  return(f)
}

#' Surface covered by a LAS* object
#'
#' Surface covered by a `LAS*` object. The surface covered by a point cloud is mathematically 0.
#' To compute non zero values the function uses different strategies. The area is computed based on
#' the number of occupied cells, or on the area of the convex hull of the points depending on the density
#' and the size of the point cloud. The result is necessarily an approximation that depends on the method
#' used.\cr For a `LAScatalog` it is computed as the sum of the bounding boxes of the files. For
#' overlapping tiles the value may be larger than the total area covered because some regions are
#' sampled twice. For a `LASheader` it is computed with the bounding box. As a consequence, for the same file
#' `st_area` applied on a LASheader or on a LAS can return slightly different values. `st_area()`
#' extends `sf:st_area()`, `area()` extends `raster:area()`. `area()` is provided for backward
#' compatibility.
#'
#' @param x An object of class \code{LAS*}.
#' @param ... unused.
#'
#' @return numeric. A number in the same units as the coordinate reference system.
#'
#' @export
#' @name st_area
#' @importFrom sf st_area
#' @md
NULL

#' @export
#' @rdname st_area
st_area.LAS = function(x, ...)
{
  if (npoints(x) == 0L) { return(0) }

  r <- area_strategy(x)

  if (r == 0)
    return(round(sf::st_area(st_convex_hull(x)), 1))

  lay  <- raster_layout(x, r)
  temp <- fasterize(x, lay, method = "count")
  area <- sum(!is.na(temp))*r^2

  # Workaround to get area with units, but without depending directly on units
  # and without parsing the CRS.
  bbox <- sf::st_as_sfc(sf::st_bbox(c(xmin = 0, ymin = 0, xmax = sqrt(area), ymax = sqrt(area)), crs = st_crs(x)))
  return(sf::st_area(bbox))
}

#' @export
#' @rdname st_area
st_area.LASheader = function(x, ...)
{
  bbox <- st_bbox(x)
  bbox <- sf::st_as_sfc(bbox)
  return(sf::st_area(bbox))
}

#' @export
#' @rdname st_area
st_area.LAScatalog = function(x, ...)
{
  x <- lascatalog_v3_repair(x)

  if (nrow(x@data) == 0L) { return(0) }
  areas <- sf::st_area(x@data)
  return(sum(areas))
}

#' @importFrom terra area
#' @export
#' @rdname st_area
setMethod("area", "LAS", function(x, ...) { as.numeric(st_area(x)) })

#' @export
#' @rdname st_area
setMethod("area", "LASheader", function(x, ...) { as.numeric(st_area(x)) })

#' @export
#' @rdname st_area
setMethod("area", "LAScatalog",  function(x, ...) { as.numeric(st_area(x)) })


# The area covered by a point cloud is 0 m2. To get another value we must apply a measurement
# strategy. Different strategies return different values. The following function tries to
# infer the more relevant strategy.
area_strategy <- function(x)
{
  bbarea  <- area(header(x))
  npoints <- npoints(x)
  density <- npoints/bbarea
  unitsm  <- st_proj_is_meters(x)
  r <- 0

  # For longlat CRS sf::st_area on the convex hull will return something
  # in the good units. We do not use occupancy grid. Returning 0.
  # For low density, Occupancy grid does not makes sense
  # compute with convex hull (return a resolution of 0)
  # For very tiny point cloud (typically plots) we do not use occupancy grid
  if (isTRUE(sf::st_is_longlat(x)) | density < 0.5 | bbarea < 1000)  return(r)

  # In other cases we use the occupancy grid to estimate the area. The following estimate the best resolution.
  # LAstools uses 2 m or 6 ft (1.8 m) depending of the units. For unknow units we are using 4 (between 2 and 6).
  if (st_crs(x) == sf::NA_crs_)
    r <- 4
  else if (unitsm)
    r <- 2
  else
    r <- 6

  # Previous resolution are good for large dataset. The overestimation is minor. For small coverage
  # the overestimation is significant. See also
  # https://stackoverflow.com/questions/71362024/how-to-get-a-circular-subset-of-a-las-dataset-with-specific-area-using-lidrcli/71369824#71369824
  # If the header tells us the area is less than 1 ha, use a finer resolution for more accuracy. But
  # only if the density if sufficient to compute occupancy grid at this resolution (i.e. there is
  # statically more than 1 point per cell). The threshold is 1.5 point per sq meters.
  one_ha <- 10000
  if (st_crs(x) == sf::NA_crs_ && bbarea < one_ha && density > 3/8)
    r <- r/2
  else if (unitsm && bbarea < one_ha && density > 3/2)
    r <- r/2
  else if (!unitsm && bbarea < one_ha*10.7 && density > 3/18)
    r <- r/2

  return(r)
}

#' lidR: airborne LiDAR for forestry applications
#'
#' lidR provides a set of tools to manipulate airborne LiDAR data in forestry contexts. The package
#' works with .las or .laz files. The toolbox includes algorithms for DSM, CHM, DTM, ABA,
#' normalisation, tree detection, tree segmentation, tree delineation, colourization, validation and
#' other tools, as well as a processing engine to process broad LiDAR coverage split into many files.
#'
#' To learn more about lidR, start with the vignettes: browseVignettes(package = "lidR"). Users can also
#' find unofficial supplementary documentation in the \href{https://r-lidar.github.io/lidRbook/}{lidR book}.
#' To ask "how to" questions please ask on \href{https://gis.stackexchange.com/}{gis.stackexchange.com}
#' with the tag \code{lidr}.
#'
#' @section Package options:
#' \describe{
#' \item{\code{lidR.progress}}{Several functions have a progress bar for long operations (but not all).
#' Should lengthy operations show a progress bar? Default: TRUE}
#' \item{\code{lidR.progress.delay}}{The progress bar appears only for long operations. After how many seconds
#' of computation does the progress bar appear? Default: 2}
#' \item{\code{lidR.raster.default}}{The functions that return a raster are raster agnostic meaning
#' that they work either with rasters from packages `raster`, `stars` or `terra`. By default they return
#' rasters from `stars`. Can be one of "raster", "stars" or "terra". Default: "terra"}
#' \item{\code{lidR.check.nested.parallelism}}{The catalog processing engine (\link{catalog_apply})
#' checks the parallel strategy chosen by the user and verify if C++ parallelization with
#' OpenMP should be disabled to avoid nested parallel loops. Default: TRUE. If FALSE
#' the catalog processing engine will not check for nested parallelism
#' and will respect the settings of \link{set_lidr_threads}.}
#' }
#'
#' @useDynLib lidR, .registration = TRUE
#' @import data.table
#' @import methods
"_PACKAGE"

#' @importFrom sf st_as_sf
NULL

#' @export
st_as_sf.LAS = function(x, ...) { return(sf::st_as_sf(x@data, coords = c("X", "Y", "Z"), crs = st_crs(x))) }

#' @export
st_as_sf.LAScatalog = function(x, ...) { return(x@data) }

#' Filter points of interest
#'
#' Filter points of interest (POI) from a LAS object where conditions are true.
#'
#' \itemize{
#' \item{`filter_poi` Select points of interest based on custom logical criteria.}
#' \item{`filter_first` Select only the first returns.}
#' \item{`filter_firstlast` Select only the first and last returns.}
#' \item{`filter_ground` Select only the returns classified as ground according to LAS specification.}
#' \item{`filter_last` Select only the last returns i.e. the last returns and the single returns.}
#' \item{`filter_nth` Select the returns from their position in the return sequence.}
#' \item{`filter_firstofmany` Select only the first returns from pulses which returned multiple points.}
#' \item{`filter_single` Select only the returns that return only one point.}
#' \item{`filter_duplicates` **Removes** the duplicated points (duplicated by XYZ)}
#' \item{`remove_noise` Removes the returns classified as noise according to LAS specification.}
#' \item{`remove_ground` Removes the returns classified as ground according to LAS specification.}
#' \item{`remove_water` Removes the returns classified as water according to LAS specification.}
#' }
#'
#' @section Non-supported LAScatalog options:
#' The option `select` is not supported and not respected because it always preserves the file format
#' and all the attributes. `select = "*"` is imposed internally.
#'
#' @param las An object of class \link[=LAS-class]{LAS}
#' @param n integer  ReturnNumber == n
#' @param \dots Logical predicates. Multiple conditions are combined with '&' or ','
#'
#' @return An object of class \link[=LAS-class]{LAS}
#'
#' @examples
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' lidar = readLAS(LASfile)
#'
#' # Select the first returns classified as ground
#' firstground = filter_poi(lidar, Classification == 2L & ReturnNumber == 1L)
#'
#' # Multiple arguments are equivalent to &
#' firstground = filter_poi(lidar, Classification == 2L, ReturnNumber == 1L)
#'
#' # Multiple criteria
#' first_or_ground = filter_poi(lidar, Classification == 2L | ReturnNumber == 1L)
#' @name filters
#' @rdname filters
#' @md
NULL

#' @export
#' @rdname filters
filter_poi = function(las, ...)
{
  stopifnotlas(las)
  keep <- lasfilter_(las, lazyeval::dots_capture(...))

  # Memory optimization
  if (sum(keep) == nrow(las@data))
    return(las)

  return(las[keep])
}

lasfilter_ <- function(las, conditions)
{
  n <- nrow(las@data)
  combined_bools <- !logical(n)

  for (condition in conditions)
  {
    bools <- lazyeval::f_eval(condition, las@data)

    if (!is.logical(bools))
      stop("`conditions` must be logical.")

    bools[is.na(bools)] <- FALSE
    combined_bools <- combined_bools & bools
  }

  return(combined_bools)
}

parse_filter = function(las, filter, k)
{
  if (!is.null(filter))
    return(lasfilter_(las, list(filter)))

  return(TRUE)
}

#' @export
#' @rdname filters
filter_first = function(las)
{
  return(filter_nth(las, 1))
}

#' @export
#' @rdname filters
filter_firstlast = function(las)
{
  ReturnNumber <- NumberOfReturns <- NULL
  return(filter_poi(las, ReturnNumber == NumberOfReturns | ReturnNumber == 1))
}

#' @export
#' @rdname filters
filter_firstofmany = function(las)
{
  NumberOfReturns <- ReturnNumber <- NULL
  return(filter_poi(las, NumberOfReturns > 1, ReturnNumber == 1))
}

#' @export
#' @rdname filters
filter_ground = function(las)
{
  Classification <- NULL
  return(filter_poi(las, Classification == 2))
}

#' @export
#' @rdname filters
filter_last = function(las)
{
  NumberOfReturns <- ReturnNumber <- NULL
  return(filter_poi(las, ReturnNumber == NumberOfReturns))
}

#' @export
#' @rdname filters
filter_nth = function(las, n)
{
  ReturnNumber <- NULL
  return(filter_poi(las, ReturnNumber == n))
}

#' @export
#' @rdname filters
filter_single = function(las)
{
  NumberOfReturns <- NULL
  return(filter_poi(las, NumberOfReturns == 1))
}

#' @export
#' @rdname filters
filter_duplicates = function(las)
{
  UseMethod("filter_duplicates", las)
}

#' @export
#' @rdname filters
filter_duplicates.LAS = function(las)
{
  dup_xyz <- duplicated(las@data, by = c("X", "Y", "Z"))
  return(las[!dup_xyz])
}

#' @export
#' @rdname filters
filter_duplicates.LAScatalog = function(las)
{
  opt_select(las) <- "*"
  options <- list(need_buffer = FALSE, drop_null = TRUE, need_output_file = TRUE)
  output  <- catalog_map(las, filter_duplicates, .options = options)
  return(output)
}

#' @export
#' @rdname filters
remove_noise = function(las)
{
  Classification <- NULL
  return(filter_poi(las, Classification != LASNOISE))
}

#' @export
#' @rdname filters
remove_ground = function(las)
{
  Classification <- NULL
  return(filter_poi(las, Classification != LASGROUND))
}

#' @export
#' @rdname filters
remove_water = function(las)
{
  Classification <- NULL
  return(filter_poi(las, Classification != LASWATER))
}

context("filter_poi")

las = random_500_points

test_that("filter works", {
  las2 = filter_poi(las, X > 50)
  expect_equal(npoints(las2), 230)

  las2 = filter_poi(las, Intensity > 25)
  expect_equal(npoints(las2), 294)

  expect_error(filter_poi(las, plop > 25), "'plop'")
})

test_that("filter first return works", {
  expect_equal(dim(filter_first(las)@data)[1], 0.6*500)
})

test_that("filter firstlast return works", {
  expect_equal(dim(filter_firstlast(las)@data)[1], 0.9*500)
})

test_that("filter ground works", {
  expect_equal(dim(filter_ground(las)@data)[1], 0.2*500)
})

test_that("filter firstlast works", {
  expect_equal(dim(filter_last(las)@data)[1], 0.6*500)
})

test_that("filter on non conditionnal statement return an error", {
  expect_error(filter_poi(las, y = 2))
})

test_that("filter memory optimization works", {
  las3 = filter_poi(las, X > 0)
  expect_reference(las@data, las3@data)
})

test_that("filter preserves spatial index metadata", {
  sensor(las) <- "tls"
  index(las) <- "octree"
  las@index$xprt <- "dummy"
  las2 = filter_poi(las, Z > 10)
  expect_equal(index(las), index(las2))
  expect_equal(sensor(las), sensor(las2))
  expect_true(is.null(las2@index$xprt))
})

#' Segment a point cloud
#'
#' Segment a point cloud using different methods. `segment_*` functions add a new attribute to the
#' point cloud to label each point. They segment either individual trees, snags, or
#' geometrical features.
#'
#' \describe{
#' \item{`segment_trees`}{Individual tree segmentation with several possible algorithms. The returned
#' point cloud has a new extra byte attribute named after the parameter `attribute` independently
#' of the algorithm used.}
#' \item{`segment_shapes`}{Computes, for each point, the eigenvalues of the covariance matrix of the
#' neighbouring points. The eigenvalues are later used either to segment linear/planar points or to
#' compute derived metrics. The points that meet a given criterion based on the eigenvalue are labelled
#' as approximately coplanar/colinear or any other shape supported.}
#' \item{`segment_snags`}{Snag segmentation using several possible algorithms. The function attributes
#' a number identifying a snag class (`snagCls` attribute) to each point of the point cloud.
#' The classification/segmentation is done at the point cloud level and currently only one algorithm is
#' implemented, which uses LiDAR intensity thresholds and specified neighbourhoods to differentiate
#' bole and branch from foliage points.}
#' }
#'
#' @section Non-supported LAScatalog options:
#' The option `select` is not supported and not respected because it always preserves the file format
#' and all the attributes. `select = "*"` is imposed internally.
#'
#' @template param-las
#' @param algorithm function. An algorithm for segmentation. For individual tree segmentation, lidR
#' has \link{dalponte2016}, \link{watershed}, \link{li2012}, and \link{silva2016}. More experimental
#' algorithms may be found in the package [lidRplugins](https://github.com/Jean-Romain/lidRplugins).
#' For snag segmentation, \code{lidR} has \link{wing2015}. For geometry segmentation, lidR has
#' \link{shp_plane}, \link{shp_hplane}, and \link{shp_line}.
#' @param attribute character. The returned LAS object as a new attribute (in a new column).
#' This parameter controls the name of the new attribute.
#' @param uniqueness character. A method to compute a unique ID. Can be 'incremental', 'gpstime' or
#' 'bitmerge'. See section 'Uniqueness'. This feature must be considered as 'experimental'.
#' @param filter formula of logical predicates. Enables the function to run only on points of interest
#' in an optimized way. See the examples.
#'
#' @template section-uniqueness
#'
#' @name segment
#' @rdname segment
#' @md
#' @examples
#' # ==============
#' # Segment trees
#' # ==============
#'
#' LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
#' las <- readLAS(LASfile, select = "xyz", filter = "-drop_z_below 0")
#'
#' # Using Li et al. (2012)
#' las <- segment_trees(las, li2012(R = 3, speed_up = 5))
#' #plot(las, color = "treeID")
#'
#' # ==============
#' # Segment shapes
#' # ==============
#'
#' LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
#' las <- readLAS(LASfile, filter = "-keep_random_fraction 0.5")
#'
#' # Use the eigenvalues to estimate if points are part of a local plan
#' las <- segment_shapes(las, shp_plane(k = 15), "Coplanar")
#' #plot(las, color = "Coplanar")
#'
#' \dontrun{
#' # Drop ground point at runtime
#' las <- segment_shapes(las, shp_plane(k = 15), "Coplanar", filter = ~Classification != 2L)
#' #plot(las, color = "Coplanar")
#'
#' # ==============
#' # Segment snags
#' # ==============
#'
#' LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
#' las <- readLAS(LASfile, select = "xyzi", filter="-keep_first") # Wing also included -keep_single
#'
#' # For the Wing2015 method, supply a matrix of snag BranchBolePtRatio conditional
#' # assessment thresholds (see Wing et al. 2015, Table 2, pg. 172)
#' bbpr_thresholds <- matrix(c(0.80, 0.80, 0.70,
#'                             0.85, 0.85, 0.60,
#'                             0.80, 0.80, 0.60,
#'                             0.90, 0.90, 0.55),
#'                             nrow =3, ncol = 4)
#'
#' # Run snag classification and assign classes to each point
#' las <- segment_snags(las, wing2015(neigh_radii = c(1.5, 1, 2), BBPRthrsh_mat = bbpr_thresholds))
#'
#' # Plot it all, tree and snag points...
#' plot(las, color="snagCls", colorPalette = rainbow(5))
#'
#' # Filter and plot snag points only
#' snags <- filter_poi(las, snagCls > 0)
#' plot(snags, color="snagCls", colorPalette = rainbow(5)[-1])
#'
#' # Wing et al's (2015) methods ended with performing tree segmentation on the
#' # classified and filtered point cloud using the watershed method
#' }
NULL

#' Parallel computation in lidR
#'
#' This document explains how to process point clouds taking advantage of parallel processing in the
#' lidR package. The lidR package has two levels of parallelism, which is why it is difficult to
#' understand how it works. This page aims to provide users with a clear overview of how to take advantage
#' of multicore processing even if they are not comfortable with the parallelism concept.
#'
#' @section Algorithm-based parallelism:
#' When processing a point cloud we are applying an algorithm on data. This algorithm may or may not be
#' natively parallel. In lidR some algorithms are fully computed in parallel, but some are not because they are
#' not parallelizable, while some are only partially parallelized. It means that some portions of the code
#' are computed in parallel and some are not. When an algorithm is natively parallel in lidR it is always
#' a C++ based parallelization with OpenMP. The advantage is that the computation is faster without any
#' consequence for memory usage because the memory is shared between the processors. In short,
#' algorithm-based parallelism provides a significant gain without any cost for your R session and
#' your system (but obviously there is a greater workload for the processors). By default lidR uses
#' half of your cores but you can control this with \link{set_lidr_threads}. For example, the \link{lmf}
#' algorithm is natively parallel. The following code is computed in parallel:
#' \preformatted{
#' las  <- readLAS("file.las")
#' tops <- locate_trees(las, lmf(2))
#' }
#' However, as stated above, not all algorithms are parallelized or even parallelizable. For example,
#' \link{li2012} is not parallelized. The following code is computed in serial:
#' \preformatted{
#' las <- readLAS("file.las")
#' dtm <- segment_trees(las, li2012())
#' }
#' To know which algorithms are parallelized users can refer to the documentation or use the
#' function \link{is.parallelised}.
#' \preformatted{
#' is.parallelised(lmf(2))   #> TRUE
#' is.parallelised(li2012()) #> FALSE
#' }
#' @section Chunk-based parallelism:
#' When processing a `LAScatalog`, the internal engine splits the dataset into chunks and each chunk is
#' read and processed sequentially in a loop. This loop can be parallelized with the
#' `future` package. By default the chunks are processed sequentially, but they can be processed
#' in parallel by registering an evaluation strategy. For example, the following code is evaluated
#' sequentially:
#' \preformatted{
#' ctg <- readLAScatalog("folder/")
#' out <- pixel_metrics(ctg, ~mean(Z))
#' }
#' But this one is evaluated in parallel with two cores:
#' \preformatted{
#' library(future)
#' plan(multisession, workers = 2L)
#' ctg <- readLAScatalog("folder/")
#' out <- pixel_metrics(ctg, ~mean(Z))
#' }
#' With chunk-based parallelism any algorithm can be parallelized by processing several subsets of
#' a dataset.  However, there is a strong cost associated with this type of parallelism. When processing several
#' chunks at a time, the computer needs to load the corresponding point clouds. Assuming the user processes
#' one square kilometer chunks in parallel with 4 cores, then 4 chunks are loaded in the computer
#' memory. This may be too much and the speed-up is not guaranteed since there is some overhead involved in
#' reading several files at a time. Once this point is understood, chunk-based parallelism is very
#' powerful since all the algorithms can be parallelized whether or not they are natively parallel.
#' It also allows to parallelize the computation on several machines on the network or to work on a HPC.
#'
#' @section Nested parallelism - part 1:
#' Previous sections stated that some algorithms are natively parallel, such as \link{lmf}, and some are
#' not, such as \link{li2012}. Anyway, users can split the dataset into chunks to process them simultaneously
#' with the LAScatalog processing engine. Let's assume that the user's computer has four cores,
#' what happens in this case:
#' \preformatted{
#' library(future)
#' plan(multisession, workers = 4L)
#' set_lidr_threads(4L)
#' ctg <- readLAScatalog("folder/")
#' out <- locate_trees(ctg, lmf(2))
#' }
#' Here the catalog will be split into chunks that will be processed in parallel. And each computation
#' itself implies a parallelized task. This is a nested parallelism task and it is dangerous! Hopefully
#'  the lidR package handles such cases and chooses by default to give precedence to chunk-based
#'  parallelism. In this case chunks will be processed in parallel and the points will be processed
#'  serially by disabling OpenMP.
#'
#' @section Nested parallelism - part 2:
#' We explained rules of precedence. But actually the user can tune the engine more accurately. Let's
#' define the following function:
#' \preformatted{
#' myfun = function(las, ws, ...)
#' {
#'   las  <- normalize_height(las, tin())
#'   tops <- locate_tree(las, lmf(ws))
#'   return(tops)
#' }
#'
#' out <- catalog_map(ctg, myfun, ws = 5)
#' }
#' This function used two algorithms, one is partially parallelized (`tin`) and one is fully
#' parallelized `lmf`. The user can manually use both OpenMP and future. By default the engine
#' will give precedence to chunk-based parallelism because it works in all cases but the user can
#' impose something else. In the following 2 workers are attributed to future and 2 workers are
#' attributed to OpenMP.
#' \preformatted{
#' plan(multisession, workers = 2L)
#' set_lidr_threads(2L)
#' catalog_map(ctg, myfun, ws = 5)
#' }
#' The rule is simple. If the number of workers needed is greater than the number of
#' available workers then OpenMP is disabled. Let suppose we have a 4 cores:
#' \preformatted{
#' # 2 chunks 2 threads: OK
#' plan(multisession, workers = 2L)
#' set_lidr_threads(2L)
#'
#' # 4 chunks 1 threads: OK
#' plan(multisession, workers = 4L)
#' set_lidr_threads(1L)
#'
#' # 1 chunks 4 threads: OK
#' plan(sequential)
#' set_lidr_threads(4L)
#'
#' # 3 chunks 2 threads: NOT OK
#' # Needs 6 workers, OpenMP threads are set to 1 i.e. sequential processing
#' plan(multisession, workers = 3L)
#' set_lidr_threads(2L)
#' }
#'
#' @section Complex computing architectures:
#' For more complex processing architectures such as multiple computers controlled remotely
#' or HPC a finer tuning might be necessary. Using
#' \preformatted{
#' options(lidR.check.nested.parallelism = FALSE)
#' }
#' lidR will no longer check for nested parallelism and will never automatically disable OpenMP.
#' @md
#' @name lidR-parallelism
NULL

#' A set of boolean tests on objects
#'
#' \code{is.empty} tests if a \code{LAS} object is a point cloud with 0 points.\cr
#' \code{is.overlapping} tests if a \code{LAScatalog} has overlapping tiles.\cr
#' \code{is.indexed} tests if the points of a \code{LAScatalog} are indexed with \code{.lax} files.\cr
#' \code{is.algorithm} tests if an object is an algorithm of the lidR package.\cr
#' \code{is.parallelised} tests if an algorithm of the lidR package is natively parallelised with OpenMP.
#' Returns TRUE if the algorithm is at least partially parallelised i.e. if some portion of the code is
#' computed in parallel.
#'
#' @param x  \code{LAS} object or any R object.
#' @param catalog A \code{LAScatalog} object.
#' @param algorithm An \code{algorithm} object.
#'
#' @return TRUE or FALSE
#'
#' @examples
#' LASfile <- system.file("extdata", "example.laz", package="rlas")
#' las = readLAS(LASfile)
#' is.empty(las)
#'
#' las = new("LAS")
#' is.empty(las)
#'
#' f <- lmf(2)
#' is.parallelised(f)
#'
#' g <- pitfree()
#' is.parallelised(g)
#'
#' ctg <- readLAScatalog(LASfile)
#' is.indexed(ctg)
#' @export
#' @rdname is
#' @name is
NULL

#' @importFrom terra is.empty
#' @rdname is
#' @export
setMethod("is.empty", "LAS", function(x)
{
  stopifnotlas(x)
  return(nrow(x@data) == 0L)
})

#' @rdname is
#' @export
is.overlapping = function(catalog)
{
  sfdf          <- catalog@data
  contour       <- sf::st_union(sfdf)
  actual_area   <- sf::st_area(contour)
  average_area  <- actual_area / length(sfdf)
  measured_area <- sum(sf::st_area(sfdf))
  actual_area   <- actual_area + 0.0000001 * average_area # fix #310
  return(actual_area < measured_area)
}

#' @rdname is
#' @export
is.indexed = function(catalog)
{
  laxfiles <- paste0(tools::file_path_sans_ext(catalog@data$filename), ".lax")
  return(!any(!file.exists(laxfiles)))
}

#' @rdname is
#' @export
is.algorithm = function(x)
{
  return(is(x, LIDRALGORITHM))
}

#' @rdname is
#' @export
is.parallelised = function(algorithm)
{
  if (!is.algorithm(algorithm))
    stop("This function only applies to algorithms from the lidR package")

  return(is(algorithm, LIDRALGORITHMOPENMP))
}

is_raster <- function(raster)
{
  raster_is_supported(raster)
}

#' Transform or convert coordinates of LAS objects
#'
#' Transform or convert coordinates of LAS objects  `st_transform()` extends `sf::st_transform()`

#'
#' @param x, An object of class LAS
#' @param crs crs object from sf or CRS object from sp
#' @param ... additional arguments `scale` and `xoffset`, `yoffset`, `zoffset`
#' for the output LAS objects. It is not  mandatory but recommended to consider providing such
#' information. Otherwise it will be guessed automatically which might not be the best choice.
#'
#' @return A LAS object
#'
#' @name st_transform
#' @importFrom sf st_transform
#' @examples
#' LASfile <- system.file("extdata", "example.laz", package="rlas")
#' las = readLAS(LASfile)
#' st_crs(las)$Name
#' st_bbox(las)
#' tlas <- sf::st_transform(las, sf::st_crs(26918))
#' st_crs(tlas)$Name
#' st_bbox(tlas)
#' @md
#' @export
#' @rdname st_transform
st_transform.LAS <- function(x, crs, ...)
{
  if (is.na(st_crs(x)))
    stop("No transformation possible from NA reference system")

  p <- list(...)

  # Transform the point coordinates
  sfpts <- sf::st_multipoint(as.matrix(coordinates3D(x)))
  sfpts <- sf::st_geometry(sfpts)
  sfpts <- sf::st_set_crs(sfpts, st_crs(x))
  sfpts <- sf::st_transform(sfpts, crs)
  geom  <- sf::st_geometry(sfpts)
  X <- geom[[1]][,1]
  Y <- geom[[1]][,2]
  Z <- geom[[1]][,3]

  # Update the offsets in the header
  offsetx <- floor(min(X))
  offsety <- floor(min(Y))
  offsetz <- floor(min(Z))
  if (!is.null(p$xoffset)) offsetx <- p$xoffset
  if (!is.null(p$yoffset)) offsety <- p$yoffset
  if (!is.null(p$zoffset)) offsetz <- p$zoffset
  x@header@PHB[["X offset"]] <- offsetx
  x@header@PHB[["Y offset"]] <- offsety
  x@header@PHB[["Z offset"]] <- offsetz

  # Update the scale factors
  scalex <- x[["X scale factor"]]
  scaley <- x[["Y scale factor"]]
  scalez <- x[["Z scale factor"]]
  if (!is.null(p$scale)) scalex <- scaley <- scalez <- p$scale
  x@header@PHB[["X scale factor"]] <- scalex
  x@header@PHB[["Y scale factor"]] <- scaley
  x@header@PHB[["Z scale factor"]] <- scalez

  # Quantize the coordinates
  fast_quantization(X, scalex, offsetx)
  fast_quantization(Y, scaley, offsety)
  fast_quantization(Z, scalez, offsetz)

  # Update the LAS object
  x@data[["X"]] <- X
  x@data[["Y"]] <- Y
  x@data[["Z"]] <- Z

  x <- las_update(x)
  st_crs(x) <- crs

  return(x)
}

st_make_bboxes = function(xmin, xmax, ymin, ymax, crs = sf::NA_crs_)
{
  bboxes <- mapply(c,  xmin, xmax, ymin, ymax, SIMPLIFY = FALSE)
  bboxes <- lapply(bboxes, function(x) { names(x) = c("xmin", "xmax", "ymin", "ymax") ; x})
  bboxes <- lapply(bboxes, sf::st_bbox)
  return(bboxes)
}

st_expand_bbox <- function(bbox, size)
{
   return(bbox + c(-size, -size, size, size))
}

st_adjust_bbox <- function(x, res, start = c(0,0), buffer = 0)
{
  bbox    <- sf::st_bbox(x)
  if (length(res) == 1) res <- c(res, res)
  bbox[1] <- round_any(bbox$xmin - buffer - 0.5 * res[1] - start[1], res[1]) + start[1]
  bbox[3] <- round_any(bbox$xmax + buffer - 0.5 * res[1] - start[1], res[1]) + res[1] + start[1]
  bbox[2] <- round_any(bbox$ymin - buffer - 0.5 * res[2] - start[2], res[2]) + start[2]
  bbox[4] <- round_any(bbox$ymax + buffer - 0.5 * res[2] - start[2], res[2]) + res[2] + start[2]
  return(bbox)
}

st_crop_if_not_similar_bbox = function(source, las)
{

  bbox1  <- st_bbox(las)
  bbox2  <- sf::st_bbox(source)
  width  <- (bbox1[3] - bbox1[1])
  height <- (bbox1[4] - bbox1[2])
  bbox   <- bbox1 + c(-width, -height, width, height)

  # If source bbox is within bbox las x 2 do not crop
  if (bbox2$xmin > bbox$xmin && bbox2$ymin > bbox$ymin && bbox2$xmax < bbox$xmax && bbox2$xmin < bbox$ymax)
  {
    return(source)
  }

  # crop a tiny bit  (1%) larger than the point cloud including lon/lat case
  bbox   <- st_bbox(las)
  width  <- (bbox[3] - bbox[1])*0.01 + las[["X scale factor"]]
  height <- (bbox[4] - bbox[2])*0.01 + las[["Y scale factor"]]
  bbox   <- bbox + c(-width, -height, width, height)
  sf::st_crs(bbox) <- sf::st_crs(source)
  if (is(source, "sf")) sf::st_agr(source) <- "constant"
  source <- sf::st_crop(source, bbox)
  return(source)
}

st_proj_is_meters <- function(obj)
{
  !is.na(sf::st_crs(obj)) & !sf::st_is_longlat(obj) & is.null(sf::st_crs(obj)$to_meter)
}

#' @param data a \link[data.table:data.table]{data.table} containing the data of a las or laz file.
#' @param header a \code{list} or a \link[=LASheader-class]{LASheader} containing the header of
#' a las or laz file.
#' @param crs crs object of class \link[sf:st_crs]{crs} from sf
#' @param check logical. Conformity tests while building the object.
#' @param index list with two elements \code{list(sensor = 0L, index = 0L)}.
#' See \link[=lidR-spatial-index]{spatial indexing}
#' @param ... internal use
#' @return An object of class \code{LAS}
#' @export
#' @describeIn LAS-class creates objects of class LAS. The original data is updated by reference to
#' quantize the coordinates according to the scale factor of the header if no header is provided.
#' In this case the scale factor is set to 0.001
LAS <- function(data, header = list(), crs = sf::NA_crs_, check = TRUE, index = NULL, ...)
{
  .N <- X <- Y <- Z <- NULL

  # This is a repair feature for lidR v3 to lidR v4
  #nocov start
  if (is(data, "LAS"))
  {
    if (methods::.hasSlot(data, "proj4string"))
    {
      pts  <- payload(data)
      phb  <- phb(data)
      vlr  <- vlr(data)
      crs  <- st_crs(data)
      evlr <- evlr(data)
      head <- c(phb, vlr, evlr)
      return(LAS(pts, head, crs = crs, check = FALSE))
    }


    return(data)
  }

  # For backward compatibility
  dots <- list(...)
  if (!is.null(dots$proj4string))
  {
    warning("Argument proj4string is deprecated. Use argument crs instead.", call. = TRUE)
    crs <- sf::st_crs(dots$proj4string)
  }
  #nocov end

  if (is(data, "lasmetrics3d"))
    class(data) <- class(data)[-1]

  if (is.data.frame(data) & !data.table::is.data.table(data))
    data.table::setDT(data)

  if (!data.table::is.data.table(data))
    stop("Invalid parameter data in constructor.")

  if (is.null(dots$no_attr_name_check))
    names(data) <- fix_name_convention(names(data))

  rlas::is_defined_coordinates(data, "stop")
  rlas::is_valid_XYZ(data, "stop")

  if (is(header, "LASheader"))
    header <- as.list(header)

  if (!is.list(header))
    stop("Wrong header object provided.")

  if (length(header) == 0) {
    xyzfactor <- 0.001
    header <- rlas::header_create(data)
    header[["X scale factor"]] <- xyzfactor
    header[["Y scale factor"]] <- xyzfactor
    header[["Z scale factor"]] <- xyzfactor
    xoffset <- floor(header[["X offset"]])
    yoffset <- floor(header[["Y offset"]])
    zoffset <- 0
    header[["X offset"]] <- xoffset
    header[["Y offset"]] <- yoffset
    header[["Z offset"]] <- zoffset

    if (nrow(data) > 0)
    {
      fast_quantization(data[["X"]], xyzfactor, xoffset)
      fast_quantization(data[["Y"]], xyzfactor, yoffset)
      fast_quantization(data[["Z"]], xyzfactor, zoffset)
      message(glue::glue("Creation of a LAS object from data but without a header:
      Scale factors were set to {xyzfactor} and XYZ coordinates were quantized to fit the scale factors."))
    }
  }

  header <- rlas::header_update(header, data)

  if (check & nrow(data) > 0)
  {
    rlas::is_defined_offsets(header, "stop")
    rlas::is_defined_scalefactors(header, "stop")
    rlas::is_defined_filesourceid(header, "stop")
    rlas::is_defined_version(header, "stop")
    rlas::is_defined_globalencoding(header, "stop")
    rlas::is_defined_date(header, "stop")

    rlas::is_defined_coordinates(data, "stop")

    rlas::is_valid_scalefactors(header, "warning")
    rlas::is_valid_globalencoding(header, "stop")
    rlas::is_valid_date(header, "stop")
    rlas::is_valid_pointformat(header, "stop")
    rlas::is_valid_extrabytes(header, "stop")

    rlas::is_valid_Intensity(data, "stop")
    rlas::is_valid_ReturnNumber(header, data, "stop")
    rlas::is_valid_NumberOfReturns(header, data, "stop")
    rlas::is_valid_ScanDirectionFlag(data, "stop")
    rlas::is_valid_EdgeOfFlightline(data, "stop")
    rlas::is_valid_Classification(data, header, "stop")
    rlas::is_valid_ScannerChannel(data, "stop")
    rlas::is_valid_SyntheticFlag(data, "stop")
    rlas::is_valid_KeypointFlag(data, "stop")
    rlas::is_valid_WithheldFlag(data, "stop")
    rlas::is_valid_OverlapFlag(data, "stop")
    rlas::is_valid_ScanAngleRank(data, "stop")
    rlas::is_valid_ScanAngle(data, "stop")
    rlas::is_valid_UserData(data, "stop")
    rlas::is_valid_gpstime(data, "stop")
    rlas::is_valid_PointSourceID(data, "stop")
    rlas::is_valid_RGB(data, "stop")
    rlas::is_valid_NIR(data, "stop")

    rlas::is_NIR_in_valid_format(header, data, "warning")
    rlas::is_gpstime_in_valid_format(header, data, "warning")
    rlas::is_RGB_in_valid_format(header, data, "warning")
    rlas::is_ScanAngle_in_valid_format(header, data, "warning")
    rlas::is_ScannerChannel_in_valid_format(header, data, "warning")
    rlas::is_extrabytes_in_accordance_with_data(header, data, "warning")

    rlas::is_compliant_ReturnNumber(data, "warning")
    rlas::is_compliant_NumberOfReturns(data, "warning")
    rlas::is_compliant_RGB(data, "warning")
    rlas::is_compliant_ScanAngleRank(data, "warning")
    rlas::is_compliant_ScanAngle(data, "warning")
    rlas::is_compliant_ReturnNumber_vs_NumberOfReturns(data, "warning")
    rlas::is_XY_larger_than_bbox(header, data, "warning")
    rlas::is_number_of_points_in_accordance_with_header(header, data, "warning")
    rlas::is_number_of_points_by_return_in_accordance_with_header(header, data, "warning")
    rlas::is_XY_smaller_than_bbox(header, data, "warning")
    rlas::is_Z_in_bbox(header, data, "warning")
    rlas::is_RGB_in_valid_format(header, data, "warning")
    rlas::is_NIR_in_valid_format(header, data, "warning")
    rlas::is_gpstime_in_valid_format(header, data, "warning")
    rlas::is_ScanAngle_in_valid_format(header, data, "warning")

    xscale <- header[["X scale factor"]]
    yscale <- header[["Y scale factor"]]
    zscale <- header[["Z scale factor"]]
    xoffset <- header[["X offset"]]
    yoffset <- header[["Y offset"]]
    zoffset <- header[["Z offset"]]

    if (!is.quantized(data[["X"]], xscale, xoffset, sample = TRUE))
      warning("Detection of quantization errors for X", call. = FALSE)
    if (!is.quantized(data[["Y"]], yscale, yoffset, sample = TRUE))
      warning("Detection of quantization errors for Y", call. = FALSE)
    if (!is.quantized(data[["Z"]], zscale, zoffset, sample = TRUE))
      warning("Detection of quantization errors for Z", call. = FALSE)
  }

  header <- LASheader(header)

  if (is.na(crs))
    crs <- st_crs(header)

  if (is.null(index))
  {
    xrange = header[["Max X"]] - header[["Min X"]]
    yrange = header[["Max Y"]] - header[["Min Y"]]
    zrange = header[["Max Z"]] - header[["Min Z"]]
    area = xrange*yrange
    if (area > 0)
    {
      n = nrow(data)
      density = n/area
      zratio = min(zrange/xrange, zrange/yrange)
    }
    else
    {
      zratio = 0
      density = 0
    }

    if (zratio < 10/100)
      index <- LIDRALSINDEX
    else if ((zratio > 10/100 & density > 100) || density > 1000)
      index <- LIDRTLSINDEX
    else
      index <- LIDRALSINDEX
  }

  index$xprt <- NULL

  las        <- new("LAS")
  las@header <- header
  las@data   <- data
  las@crs    <- crs
  las@index  <- index

  if (isTRUE(sf::st_is_longlat(crs)))
    warning("Point-clouds with lon/lat coordinates are not well supported.", call. = FALSE)

  return(las)
}

#' Extract or Replace Parts of a LAS* Object
#'
#' Operators acting on \code{LAS*} objects. However, some have modified behaviors to prevent some
#' irrelevant modifications. Indeed, a \code{LAS*} object cannot contain anything, as the content
#' is restricted by the LAS specifications. If a user attempts to use one of these functions
#' inappropriately an informative error will be thrown.
#'
#' @param x A \code{LAS*} object
#' @param name A literal character string or a name (possibly backtick quoted).
#' @param value typically an array-like R object of a similar class as x.
#' @param i string, name of elements to extract or replace.
#' @param j Unused.
#'
#' @name Extract
#' @examples
#' LASfile <- system.file("extdata", "example.laz", package="rlas")
#' las = readLAS(LASfile)
#'
#' las$Intensity
#' las[["Z"]]
#' las[["Number of points by return"]]
#'
#' \dontrun{
#' las$Z = 2L
#' las[["Z"]] = 1:10
#' las$NewCol = 0
#' las[["NewCol"]] = 0
#' }
NULL

#' @export
#' @rdname Extract
setMethod("$", "LAS", function(x, name) { x[[name]] })

#' @export
#' @rdname Extract
setMethod("[[", c("LAS", "ANY", "missing"), function(x, i, j, ...) {

  if (is.character(i) && !i %in% names(x@data))
    return(x@header[[i]])

  return(x@data[[i]])
})

#' @export
#' @rdname Extract
setMethod("$<-", "LAS", function(x, name, value)
{
  x[[name]] <- value
  return(x)
})

#' @export
#' @rdname Extract
setMethod("[[<-", c("LAS", "ANY", "missing", "ANY"),  function(x, i, j, value)
{
  if (i %in% names(header(x)))
  {
    x@header[[i]] <- value
    return(x)
  }

  if (!i %in% names(x))
    stop("Addition of a new column using [[ is forbidden for LAS objects. See ?add_attribute", call. = FALSE)

  if (i %in% LASATTRIBUTES)
  {
    type1 <- storage.mode(x@data[[i]])
    type2 <- storage.mode(value)

    if (type1 != type2)
      stop(glue::glue("Trying to replace data of type {type1} by data of type {type2}: this action is not allowed"), call. = FALSE)
  }

  if (!i %in% names(x@data))
    stop("Addition of a new column using $ is forbidden for LAS objects. See ?add_attribute", call. = FALSE)

  if (i %in% LASATTRIBUTES)
  {
    type1 <- storage.mode(x@data[[i]])
    type2 <- storage.mode(value)

    if (type1 != type2)
      stop(glue::glue("Trying to replace data of type {type1} by data of type {type2}: this action is not allowed"), call. = FALSE)
  }

  for (coordinate_name in c("X", "Y", "Z"))
  {
    if (i == coordinate_name)
    {
      scale_string <- paste(i, "scale factor")
      offset_string <- paste(i, "offset")
      scale  <- x[[scale_string]]
      offset <- x[[offset_string]]
      valid_range <- storable_coordinate_range(scale, offset)
      value_range <- range(value)

      if (value_range[1] < valid_range[1] | value_range[2] > valid_range[2])
        stop(glue::glue("Trying to store values ranging in [{value_range[1]}, {value_range[2]}] but storable range is [{valid_range[1]}, {valid_range[2]}]"), call. = FALSE)

      if(!is.quantized(value, scale, offset))
        value <- quantize(value, scale, offset, FALSE)
    }
  }

  x@data[[i]] = value

  if (i %in% c("X", "Y", "Z"))
    x <- las_update(x)

  return(x)
})

#' @export
#' @rdname Extract
setMethod("[", c("LAS", "numeric"),  function(x, i)
{
  return(smart_subset(x, i))
})

#' @export
#' @rdname Extract
setMethod("[", c("LAS", "logical"),  function(x, i)
{
  return(smart_subset(x, i))
})

#' @export
#' @rdname Extract
setMethod("[", c("LAS", "sf"),  function(x, i)
{
  return(x[sf::st_geometry(i)])
})

#' @export
#' @rdname Extract
setMethod("[", c("LAS", "sfc"),  function(x, i)
{
  if (length(i) > 1)
    stop("A single geometry feature is supported for subsetting. See clip_roi() for more options", call. = FALSE)

  i <- i[[1]]

  if (!is(i, "POLYGON") && !is(i, "MULTIPOLYGON"))
    stop("Only POLYGONS are supported for subsetting. See clip_roi() for more options", call. = FALSE)

  return(clip_roi(x, i))
})

fix_name_convention <- function(names)
{
  input <- tolower(names)
  lasattr <- tolower(LASATTRIBUTES)
  idx <- match(input, lasattr)
  new <- LASATTRIBUTES[idx]
  changed <- new != names
  for (i in which(changed)) message(glue::glue("Attribute '{names[i]}' renamed '{new[i]}' to match with default attribute names."))
  return(new)
}

smart_subset <- function(las, i)
{
  data <- payload(las)

  # Start a copy by preserving addresses
  # data[, c("X", "Y", "Z")] makes deeps copies
  new_data <- data.frame(X = data[["X"]], Y = data[["Y"]], Z = data[["Z"]])
  data.table::setDT(new_data)

  # Subset the coodinate
  new_data <- new_data[i]

  # Iteratively add subseted elements in a way that preserve ALTREP
  attr <- names(las)
  attr <- attr[!attr %in% c("X", "Y", "Z")]
  for (name in attr) new_data[[name]] <- data[[name]][i]

  new_las <- LAS(new_data, las@header, st_crs(las), FALSE, las@index, no_attr_name_check = TRUE)
  new_las <- las_update(new_las)
  return(new_las)
}

#' @rdname aggregate
#' @export
polygon_metrics = function(las, func, geometry, ...)
{
  UseMethod("polygon_metrics", las)
}

#' @export
polygon_metrics.LAS = function(las, func, geometry, ...)
{
  stopifnot(is(geometry, "sf") | is(geometry, "sfc"))
  geometry <- sf::st_geometry(geometry)
  stopifnot(is(geometry, "sfc_POLYGON") | is(geometry, "sfc_MULTIPOLYGON"))
  M <- template_metrics(las, func, geometry, ...)
  M <- sf::st_set_geometry(M, geometry)
  return(M)
}

---
title: "LAS formal class"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{1. LAS formal class}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
knitr::knit_hooks$set(time_it = local({
  now <- NULL
  function(before, options) {
    if (before) {
      # record the current time before each chunk
      now <<- Sys.time()
    } else {
      # calculate the time difference after a chunk
      res <- difftime(Sys.time(), now)
      # return a character string to show the time
      #if (res > 0.1)
      #paste("<br/>========================<br/>Time for this code chunk ", options$label, " to run:", round(res,2), "<br/>========================<br/>")
    }
  }
}))
knitr::opts_chunk$set(time_it = TRUE)
#rgl::setupKnitr()
options(rmarkdown.html_vignette.check_title = FALSE)
library(lidR)
```

A LAS class is a representation in R of a las file that aims to respect as closely as possible the official [LAS specification](https://community.asprs.org/leadership-restricted/leadership-content/public-documents/standards) that describes the file format. "As closely as possible" means that, due to R internal limitations, it is not possible to represent a las file exactly as it should be represented. Additionally, some aspects of the las format specifications are not supported in lidR. Still, the contents of a `LAS` object must reflect the fact that it is a representation of a standardized format, so some restrictions are imposed on users.

## Build a LAS object reading a las file

The function `readLAS` reads one or several .las or .laz file(s) to build a LAS object.

```{r}
LASfile <- system.file("extdata", "example.laz", package="rlas")
las <- readLAS(LASfile)
print(las)
```

## Basic structure of a LAS object

A `LAS` object is composed of four slots: `@data`, `@header`, `@crs` and `@index`.

### @data: the point cloud

The slot `data` of a LAS object contains a `data.table` with the data read from .las or .laz file(s). The columns of the table are named after the [LAS specification](https://community.asprs.org/leadership-restricted/leadership-content/public-documents/standards) version 1.4. Each name is reserved and is associated with a given type:

- `X` `Y` `Z` (dbl)
- `Intensity` (int)
- `gpstime` (dbl)
- `ReturnNumber` (int)
- `NumberOfReturns` (int)
- `ScanDirectionFlag` (int)
- `EdgeOfFlightline`(int)
- `Classification` (int)
- `ScannerChannel` (int) (point format >= 6)
- `Synthetic_flag` (bool)
- `Withheld_flag` (bool)
- `Keypoint_flag` (bool)
- `Overlap_flag` (bool) (point format >= 6)
- `ScanAngleRank` (int) (point format < 6)
- `ScanAngle` (dbl) (point format >= 6)
- `UserData` (int)
- `PointSourceID` (int)
- `R` `G` `B` (int)
- `NIR` (int)

Here we can already see some deviations from the official las format specifications. For example, the attribute 'Classification' should be an `unsigned char` stored on 8 bits. However, the R language does not support this data type and consequently this attribute is stored in a 32-bit signed `int`. One can read the official las specifications to figure out the other deviations from the original file format induced by the fact that R only has 32-bit signed integers and 64-bit signed decimal numbers.

### @header: the header

A `LAS` object contains a slot `@header` that represents the header of the las file. The header is stored in a `LASheader`  object. A `LASheader` object contains two slots: `@PHB` for the public header block and `@VLR` for the variable length records. Both slots are lists labelled according to the las file format specification. See [public documentation of las file format](https://community.asprs.org/leadership-restricted/leadership-content/public-documents/standards) for more information about las headers. Users should never normally have to worry about the header as long as they use functions from lidR. Everything is managed internally to ensure that objects are valid. However, users still need to know that the contents of the header are important, especially when writing `LAS` objects into las or laz files.

```{r}
print(header(las))
```

### @crs: the CRS

The slot `@crs` contains a `crs` object from package `sf` and stores the coordinate reference system (CRS) of the las file. In the official las specifications the CRS is stored in the header. In a LAS object the CRS is stored in the header using the EPSG code of the CRS or a WKT string (depending how it has been recorded in the file), but it is also stored in the slot `@crs`. This is to ensure it meets R standards and is in accordance with other spatial data packages in the R ecosystem. Consequently, to get a valid LAS object properly written into a las file it is important to set the CRS using the function `st_crs()`. This function updates the header of the LAS object **and** the `crs`

```r
st_crs(las) <- 26917
st_crs(las)
```

According to LAS specification the CRS system can also be a WKT string when the WKT bit flag is set to `TRUE`.

```r
las@header@PHB[["Global Encoding"]][["WKT"]] = TRUE
```

```{r, echo=FALSE, eval=FALSE}
las@header@PHB[["Global Encoding"]][["WKT"]] = TRUE

st_crs(las) <- 26917

# Header has been updated but users do not need to take care of that
las@header@VLR[["WKT OGC CS"]][["WKT OGC COORDINATE SYSTEM"]]
```

### @index: the spatial indexing mode

Records what kind of spatial index will be used internally to perform computations that require spatial indexing. See `help("lidR-spatial-index")`.

## Allowed and non-allowed manipulation of a LAS object

R users who are used to manipulating spatial data are likely to be very familiar with the `sp` or `sf` packages and all the classes used to store spatial data, such as `SpatialPointsDataFrame`, `SpatialPolygonsDataFrame`, `sf` and so on. The data contained in these classes are freely modifiable by the user because they can be of any type. A `LAS` object is not freely modifiable because it is a strongly standardized representation of a las file.

For example, users cannot replace the `Classification` attribute with the value `0` because `0` is a decimal number in R and the 'Classification' attribute is an integer. The following throws an error:

```{r, error = TRUE, purl = FALSE}
las$Classification <- 0
```

In R `0L` is an integer and thus the following is allowed:

```{r}
las$Classification <- 0L
```

It would be possible to automatically cast the input into the correct type without throwing an error. But for the lidR package we chose to be very pedantic on this point to avoid any potential problems and because we would prefer users to be careful about the content of their data.

The addition of a new column is also restricted. For example, one may want to add an attribute `R` corresponding to the red channel. 
```{r, error = TRUE, purl = FALSE}
las$R <- 0
```

This is not allowed because a LAS object should always be valid. By allowing the user to add an R column the LAS object would no longer be valid for two reasons:

1. `R` is a reserved name of the core attributes and must be an integer. In the example above it is a decimal number.
2. A LAS file with RGB attributes is of type 3, 7 or 8. As a result the header must be updated, but in the previous example it is not.

In consequence, adding a column must be done via the functions `add_attribute`, `add_lasattribute`, `add_lasrgb`. This way users are forced to read the documentation of these two functions. And yet some restrictions are still in place. For example, the following is not allowed for the same reasons as above:

```{r, error = TRUE, purl = FALSE}
las <- add_attribute(las, 0, "R")
```

But anyway, R being R, there is no way to completely restrict editing of objects. Users can always by-pass the restrictions to make LAS objects that are not strictly valid:

```{r}
las@data$R <- 0
```

```{r, echo = FALSE}
las@data$R <- NULL
```

In conclusion, a LAS object is not actually immutable but at least there are some restrictions to ensure that the user is aware that not everything is authorized.

## Extra attributes and extra bytes in a LAS object

As we have seen, a LAS object contains a core of attributes associated with reserved names in accordance with the las specifications. It is possible, however, to add more attributes to a LAS object even if they are not part of the core attributes imposed by the las specifications.

### Extra attributes

Extra attributes are just like adding a column in a regular table in R. One can freely modify the data using the function `add_attribute`. It is thus possible to add an attribute to a LAS object. For example, it is possible to attribute an ID to each point and use this value in subsequent code:

```{r}
las  <- add_attribute(las, 1:30, "ID")
las2 <- filter_poi(las, ID > 15)
```

But it is important to understand that this attribute is invalid with respect to the las specifications. Thus it can be used at the R level but will not be written in a las file and thus will be lost at write time. Depending on the purpose of this attribute it may or may not be useful to be able to write this extra data. Most of the time the information is only useful at the R level but sometimes it might be appropriate to store the data in a file.

### Extra bytes attributes

The las specifications allow for storing extra attributes that are not part of the core attributes. The way to do this is more complex. Basically it is called extra bytes attributes and it implies modification of the LAS object header to indicate that the contents of the file contains more than the core attributes. This is abstracted with the function `add_lasattribute`.

```{r}
las  <- add_lasattribute(las, 1:30, "ID", "An ID for each point")
```

Using this function, the header is updated according to the las specification and thus the extra bytes attributes can be written in the file. lidR supports up to 10 extra bytes attributes. The extra bytes attributes are limited to being of type numeric. Indeed, the las specifications do not allow for storing extra bytes attributes of type string or type boolean. Thus the following fails:

```{r, error = TRUE, purl = FALSE}
abc  <- sample(letters, 30, replace = TRUE)
las  <- add_lasattribute(las, abc, "ID", "An ID for each point")
```


## Validation of LAS objects

It is common that users report bugs arising from the fact that a point cloud is invalid. This is why we introduced the function `las_check` to perform a deep inspection of LAS objects. This function checks if a LAS object is in accordance with the las specifications but also it checks for weird point clouds that could be valid with respect to the specifications but invalid for actual processing. For example, it often happens that a las file contains duplicated points for no valid reason. This may lead to trees being detected twice, to invalid metrics, or to errors in DTM generation, and so on...

```{r}
las_check(las)
```

## Display a LAS object

lidR provides a simple `plot` function to plot a LAS object in 3D. It is based on the `rgl` package. The `rgl` package is amazing but has some problems working with large point clouds. We are currently developing [our own viewer](https://github.com/r-lidar/lidRviewer) to overcome this issue. This viewer is fully compatible with `lidR` but still in heavy development.

```r
plot(las)
```

```{r, eval = FALSE, echo = FALSE, rgl=TRUE, dev='png'}
LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
las = readLAS(LASfile)
m = structure(c(0.921, -0.146, 0.362, 0, 0.386, 0.482, -0.787, 0, 
-0.06, 0.864, 0.5, 0, 0, 0, 0, 1), .Dim = c(4L, 4L))
plot(las)
rgl::view3d(fov = 50, userMatrix = m)
```

<img src="data:image/png;base64,[encoded data, 72280 characters]" />


The parameter `color` expects the name of the attribute you want to use to colorize the points. Default is `Z`

```r
plot(las, color = "Intensity")
```

If your file contains RGB data the string `"RGB"` is supported:

```r
plot(las, color ="RGB")
```

## Memory considerations

This section is of major importance because there are many instances where R is weak at memory management. 

Firstly, it is important to note that R only enables manipulation of 32-bit integers and 64-bit decimal numbers. But the las specification states, for example, that the intensity is stored on 16 bits (see previous sections). When read in R it must be converted to 32 bits and therefore will use twice as much memory than is needed. Worse, the return numbers are stored on 3 bits in las files but 32 bits in R, therefore using 11 times more memory than is required. Last but not least, flags are stored on 1 bit, whereas R uses 32 bits. This is 32 times more memory than is needed. As a consequence, a LAS object is 2 to 3 times larger than it needs to be.

Secondly, the way the point cloud is stored and the way R works implies that copies will be made of the point cloud either in the user's workspace or internally. Considering that point clouds can be huge it is important to be aware of this point.

### Deep copies

Let's assume we have loaded a large las file that uses 1 GB of R memory.

```r
las.original <- readLAS("big_file.las")
```

Suppose we now want to remove a few outliers above 50 m. One can write the following:

```r
las.denoised <- filter_poi(las.original, Z < 50)
```

And the user now has two objects:

- `las.original` of size 1 GB
- `las.denoised` that is also 1 GB, because we only removed a dozen or so points out of millions.

This uses 2 GB of memory. This is how R works. When a vector is subsetted it is necessarily copied. We talk about **deep copies**. In regular data processing it rarely matters and this behavior is barely noticeable. Indeed, it is rare that data uses a lot of memory. But LiDAR datasets are often massive, and this necessitates that users must carefully consider memory usage to avoid running out of RAM.

### Shallow copies

In the previous example we showed a deep copy. A deep copy means that the point cloud is actually copied into the memory. A deep copy occurs when the number of points of the output is different from the number of points of the input. But many functions return the same number of point as the input. In such cases only **shallow copies** are made. For example, when classifying points into ground and non-ground:

```r
las.classified <- classify_ground(las.original, csf())
```

In this case the vectors that store the X Y Z coordinates as well as those that store the Intensity, ReturnNumber, NumberOfReturn and other attributes were not modified by the function. Only the contents of the 'Classification' attribute were modified. In this case `las.classified` and `las.original`, even though they are two different objects, share the same memory for X Y Z, and so on, but the attributes 'Classification' are different. In conclusion:

- `las.original` is of size 1 GB
- `las.classified` is also 1 GB.

But both together they are not equal to 2 GB, but ~1.1 GB because they share the same memory. The content of the original LAS object was shallow copied. An understanding of the concepts of deep and shallow copies is important for optimizing your scripts. 

As we have seen, because of the way R is designed, lidR uses a large amount of memory anyway. To deal with this limitation `readLAS` has two optimizations: the parameter `select` and the parameter `filter`.

### Parameter `select`

To save memory only useful data can be loaded. `readLAS` can take an optional parameter `select` which enables the user to selectively load the data of interest. For example, one can load only the `X Y Z` fields. This selection is done at the C++ level while reading and is memory-optimized.

```r
las = readLAS("file", select = "xyz")
las = readLAS("file", select = "xyzi")
las = readLAS("file", select = "* -i -u") # Negation works too
```

### Parameter `filter`

While `select` enables the user to select "columns" (or attributes) while reading files, `filter` allows selection of "rows" (or points) while reading. Again, the selection is done at the C++ level and is memory-optimized so not a single bit is lost at the R level. Removing data at reading time that is superfluous for your purposes saves memory and decreases computation time.

```r
las = readLAS("file", filter = "-keep_first")
las = readLAS("file", select = "xyzi", filter = "-keep_first -drop_z_below 5 -drop_z_above 50")
```

---
title: "Create a function that can process a LAScatalog"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{5. Create a function to process a LAScatalog}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
options(rmarkdown.html_vignette.check_title = FALSE)
```

The following demonstrates how to write your own functions that are fully applicable on a broad collection of point clouds and based on the available `lidR` tools. We will create a simple `filter_noise` function. This example **should not be considered as the reference method for filtering noise**, but rather as a demonstration to help understand the logic behind the design of lidR, and as a full example of how to create a user-defined function that is fully operational.

## Design a noise filter 

A simple (too simple) way to detect outliers is to measure the 95th percentile of height in 10 x 10-m pixels (area-based approach) and then remove the points that are above the 95th percentile in each pixel plus, for example, 20%. This can easily be built in lidR using `pixel_metrics`, `merge_spatial` and `filter_poi`, and should work either on a normalized or a raw point cloud. Let's create a function method `filter_noise`:

```r
filter_noise = function(las, sensitivity)
{
  p95 <- pixel_metrics(las, ~quantile(Z, probs = 0.95), 10)
  las <- merge_spatial(las, p95, "p95")
  las <- filter_poi(las, Z < p95*sensitivity)
  las$p95 <- NULL
  return(las)
}
```

This function is fully functional on a point cloud loaded in memory

```r
las <- readLAS("file.las")
las <- filter_noise(las, sensitivity = 1.2)
writeLAS(las, "denoised-file.las")
```

## Extend the `filter_noise` function to a `LAScatalog`

Users can access the catalog processing engine with the function `catalog_apply` i.e. the engine used internally. It can be applied to any function over an entire collection. This function is complex and we created a simplified (but less versatile) version names `catalog_map` that suit for most cases. Here we will apply our custom `filter_noise` function with `catalog_map`. To use our function `filter_noise` on a `LAScatalog` we must create a compatible function (see documentation of `catalog_apply`):

```r
filter_noise = function(las, sensitivity)
{
  if (is(las, "LAS"))
  {
      p95 <- pixel_metrics(las, ~quantile(Z, probs = 0.95), 10)
      las <- merge_spatial(las, p95, "p95")
      las <- filter_poi(las, Z < p95*sensitivity)
      las$p95 <- NULL
      return(las)
  }
  
  if (is(las, "LAScatalog"))
  {
      res <- catalog_map(las, filter_noise, sensitivity = sensitivity)
      return(res)
  }
}
```

And it just works. This function `filter_noise` is now fully compatible with the catalog processing engine and supports all the options of the engine.

```r
myproject <- readLAScatalog("folder/to/lidar/data/")

opt_filter(myproject)       <- "-drop_z_below 0"
opt_chunk_buffer(myproject) <- 10
opt_chunk_size(myproject)   <- 0
opt_output_files(myproject) <- "folder/to/lidar/data/denoised/{ORIGINALFILENAME}_denoised"

output <- filter_noise(myproject, tolerance = 1.2)
```

## Finalize the functions

As is, the function `filter_noise` is not actually complete. Indeed the processing options were not checked. For example, this function should not allow the output to be returned into R otherwise the whole point cloud will be returned.

```r
filter_noise = function(las, sensitivity)
{
  if (is(las, "LAS"))
  {
      p95 <- pixel_metrics(las, ~quantile(Z, probs = 0.95), 10)
      las <- merge_spatial(las, p95, "p95")
      las <- filter_poi(las, Z < p95*sensitivity)
      las$p95 <- NULL
      return(las)
  }
  
  if (is(las, "LAScatalog"))
  {
     options <- list(
       need_output_file = TRUE,    # Throw an error if no output template is provided
       need_buffer = TRUE)         # Throw an error if buffer is 0
     res <- catalog_map(las, filter_noise, sensitivity = sensitivity, .options = options)
     return(res)
  }
}
```

Now you know how to build your custom functions that work either on a `LAS` or a `LAScatalog` object. Be careful, `catalog_map` is only a simplification of `catalog_apply` with restricted capabilities. Check out the documentation of `catalog_apply`.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/track_sensor.R
\name{track_sensor}
\alias{track_sensor}
\title{Reconstruct the trajectory of the LiDAR sensor using multiple returns}
\usage{
track_sensor(
  las,
  algorithm,
  extra_check = TRUE,
  thin_pulse_with_time = 0.001,
  multi_pulse = FALSE
)
}
\arguments{
\item{las}{An object of class \link[lidR:LAS-class]{LAS} or \link[lidR:LAScatalog-class]{LAScatalog}.}

\item{algorithm}{function. An algorithm to compute sensor tracking. \code{lidR} implements
\link{Roussel2020} and  \link{Gatziolis2019} (see respective documentation and examples).}

\item{extra_check}{boolean. Datasets are rarely perfectly populated, leading to unexpected errors.
Time-consuming checks of data integrity are performed. These checks can be skipped as they account
for an significant proportion of the computation time. See also section 'Tests of data integrity'.}

\item{thin_pulse_with_time}{numeric. In practice, it is not useful to compute the position using all
multiple returns. It is more computationally demanding but not necessarily more accurate. This keeps
only one pulse every x seconds. Set to 0 to use all multiple returns. Use 0 if the file has already
been read with \code{filter = "-thin_pulses_with_time 0.001"}.}

\item{multi_pulse}{logical. TRUE only for systems with multiple pulses. Pulse ID must be recorded
in the UserData attribute.}
}
\value{
An sf object with POINT Z geometries. Information about the time interval and the score of
the positioning (according to the method used) are also in the table of attributes.
}
\description{
Use multiple returns to estimate the positioning of the sensor by computing the intersection in
space of the line passing through the first and last returns. To work, this function requires a
dataset where the 'gpstime', 'ReturnNumber', 'NumberOfReturns' and 'PointSourceID' attributes are
properly populated, otherwise the output may be incorrect or weird. For LAScatalog processing
it is recommended to use large chunks and large buffers (e.g. a swath width). The point cloud must
not be normalized.
}
\section{Non-supported LAScatalog options}{

The option `select` is not supported and not respected  because it is internally known what is the
best to select\cr
The option `output_files` is not supported and not respected because the output must be post-processed
as a whole
}

\section{Test of data integrity}{

In theory, sensor tracking is a simple problem to solve as long as each pulse is properly
identified from a well-populated dataset. In practice, many problems may arise from datasets that are populated
incorrectly. Here is a list of problems that may happen. Those with a * denote problems already encountered and
internally checked to remove weird points:
\itemize{
\item 'gpstime' does not record the time at which pulses were emitted and thus pulses are not identifiable
\item *A pulse (two or more points that share the same gpstime) is made of points from different
flightlines (different PointSourceID). This is impossible and denotes an improperly populated PointSourceID
attribute.
\item 'ReturnNumber' and 'NumberOfReturns' are wrongly populated with either some ReturnNumber > NumberOfReturn
 or several first returns by pulses
}
For a given time interval, when weird points are not filtered, the position is not computed for this
interval.
}

\examples{
# A valid file properly populated
LASfile <- system.file("extdata", "Topography.laz", package="lidR")
las = readLAS(LASfile)
#plot(las)

# pmin = 15 because it is an extremely small file
# strongly decimated to reduce its size. There are
# actually few multiple returns
flightlines <- track_sensor(las, Roussel2020(pmin = 15))

plot(las@header)
plot(sf::st_geometry(flightlines), add = TRUE)

#plot(las) |> add_flightlines3d(flightlines, radius = 10)

\dontrun{
# With a LAScatalog "-drop_single" and "-thin_pulses_with_time"
# are used by default
ctg = readLAScatalog("folder/")
flightlines <- track_sensor(ctg,  Roussel2020(pmin = 15))
plot(flightlines)
}
}
\author{
Jean-Francois Bourdon & Jean-Romain Roussel
}
\concept{range}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/connected_components.R
\name{connected_components}
\alias{connected_components}
\title{Connected-Component Labeling}
\usage{
connected_components(las, res, min_pts, name = "clusterID", connectivity = 6)
}
\arguments{
\item{las}{A LAS object representing the point cloud data.}

\item{res}{Grid resolution. If two non-empty voxels are contiguous, they are considered part of
the same component.}

\item{min_pts}{Minimum number of points in a cluster. If a cluster contains fewer than \code{min_pts}
points, it is assigned an ID of 0, indicating that it does not belong to any group.}

\item{name}{A string specifying the name of the new attribute used to store the ID of each cluster.}

\item{connectivity}{integer. 6, 18 or 26 neighbors search.}
}
\value{
A LAS object with an additional attribute named as specified by \code{name}.
}
\description{
Assigns an ID to each cluster of connected components in a point cloud. The point cloud is subdivided
into a 3D grid, and a classical Connected-Component Labeling algorithm is applied in 3D.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods-LAS.R, R/methods-LAScatalog.R,
%   R/methods-LASheader.R
\name{Extract}
\alias{Extract}
\alias{$,LAS-method}
\alias{[[,LAS,ANY,missing-method}
\alias{$<-,LAS-method}
\alias{[[<-,LAS,ANY,missing-method}
\alias{[,LAS,numeric,ANY-method}
\alias{[,LAS,logical,ANY-method}
\alias{[,LAS,sf,ANY-method}
\alias{[,LAS,sfc,ANY-method}
\alias{$,LAScatalog-method}
\alias{[[,LAScatalog,ANY,missing-method}
\alias{[,LAScatalog,ANY,ANY-method}
\alias{[,LAScatalog,logical,ANY-method}
\alias{[,LAScatalog,sf,ANY-method}
\alias{[,LAScatalog,sfc,ANY-method}
\alias{[[<-,LAScatalog,ANY,ANY-method}
\alias{$<-,LAScatalog-method}
\alias{$,LASheader-method}
\alias{$<-,LASheader-method}
\alias{[[,LASheader,ANY,missing-method}
\alias{[[<-,LASheader,character,missing-method}
\title{Extract or Replace Parts of a LAS* Object}
\usage{
\S4method{$}{LAS}(x, name)

\S4method{[[}{LAS,ANY,missing}(x, i, j, ...)

\S4method{$}{LAS}(x, name) <- value

\S4method{[[}{LAS,ANY,missing}(x, i, j) <- value

\S4method{[}{LAS,numeric,ANY}(x, i)

\S4method{[}{LAS,logical,ANY}(x, i)

\S4method{[}{LAS,sf,ANY}(x, i)

\S4method{[}{LAS,sfc,ANY}(x, i)

\S4method{$}{LAScatalog}(x, name)

\S4method{[[}{LAScatalog,ANY,missing}(x, i, j, ...)

\S4method{[}{LAScatalog,ANY,ANY}(x, i, j, ..., drop = FALSE)

\S4method{[}{LAScatalog,logical,ANY}(x, i)

\S4method{[}{LAScatalog,sf,ANY}(x, i)

\S4method{[}{LAScatalog,sfc,ANY}(x, i)

\S4method{[[}{LAScatalog,ANY,ANY}(x, i, j) <- value

\S4method{$}{LAScatalog}(x, name) <- value

\S4method{$}{LASheader}(x, name)

\S4method{$}{LASheader}(x, name) <- value

\S4method{[[}{LASheader,ANY,missing}(x, i, j, ...)

\S4method{[[}{LASheader,character,missing}(x, i) <- value
}
\arguments{
\item{x}{A \code{LAS*} object}

\item{name}{A literal character string or a name (possibly backtick quoted).}

\item{i}{string, name of elements to extract or replace.}

\item{j}{Unused.}

\item{...}{Unused}

\item{value}{typically an array-like R object of a similar class as x.}

\item{drop}{Unused}
}
\description{
Operators acting on \code{LAS*} objects. However, some have modified behaviors to prevent some
irrelevant modifications. Indeed, a \code{LAS*} object cannot contain anything, as the content
is restricted by the LAS specifications. If a user attempts to use one of these functions
inappropriately an informative error will be thrown.
}
\examples{
LASfile <- system.file("extdata", "example.laz", package="rlas")
las = readLAS(LASfile)

las$Intensity
las[["Z"]]
las[["Number of points by return"]]

\dontrun{
las$Z = 2L
las[["Z"]] = 1:10
las$NewCol = 0
las[["NewCol"]] = 0
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algorithm-snag.R
\name{snag_wing2015}
\alias{snag_wing2015}
\alias{wing2015}
\title{Snags Segmentation Algorithm}
\usage{
wing2015(
  neigh_radii = c(1.5, 1, 2),
  low_int_thrsh = 50,
  uppr_int_thrsh = 170,
  pt_den_req = 3,
  BBPRthrsh_mat = NULL
)
}
\arguments{
\item{neigh_radii}{numeric. A vector of three radii used in quantifying local-area centered
neighborhoods. See Wing et al. (2015) reference page 171 and Figure 4. Defaults are 1.5,
1, and 2 for the sphere, small cylinder and large cylinder neighborhoods, respectively.}

\item{low_int_thrsh}{numeric. The lower intensity threshold filtering value. See Wing
et al. (2015) page 171. Default is 50.}

\item{uppr_int_thrsh}{numeric. The upper intensity threshold filtering value. See Wing
et al. (2015) page 171. Default is 170.}

\item{pt_den_req}{numeric. Point density requirement based on plot-level point density
defined classes. See Wing et al. (2015) page 172. Default is 3.}

\item{BBPRthrsh_mat}{matrix. A 3x4 matrix providing the four average BBPR (branch and bole
point ratio) values for each of the three neighborhoods (sphere, small cylinder and large
cylinder) to be used for conditional assessments and classification into the following four snag
classes: 1) general snag 2) small snag 3) live crown edge snag 4) high canopy
cover snag. See Wing et al. (2015) page 172 and Table 2. This matrix must be provided by
the user.}
}
\description{
This function is made to be used in \link{segment_snags}. It implements an algorithms for snags segmentation
based on Wing et al (2015) (see references). This is an automated filtering algorithm that utilizes
three dimensional neighborhood lidar point-based intensity and density statistics to remove lidar
points associated with live trees and retain lidar points associated with snags.
}
\details{
Note that this algorithm strictly performs a classification based on user input while
the original publication's methods also included a segmentation step and some pre-
(filtering for first and single returns only) and post-process (filtering for only the
snag classified points prior to segmentation) tasks which are now expected to be performed
by the user. Also, this implementation may have some differences compared with the original
method due to potential mis-interpretation of the Wing et al. manuscript, specifically
Table 2 where they present four groups of conditional assessments with their required
neighborhood point density and average BBPR values (BBPR = branch and bole point ratio;
PDR = point density requirement).\cr\cr
This algorithm attributes each point in the point cloud (\code{snagCls} column) into the
following five snag classes:
\itemize{
\item 0: live tree - not a snag\cr
\item 1: general snag - the broadest range of snag point situations\cr
\item 2: small snag - isolated snags with lower point densities\cr
\item 3: live crown edge snag - snags located directly adjacent or intermixing with live trees crowns \cr
\item 4: high canopy cover snag - snags protruding above the live canopy in dense conditions (e.g.,
canopy cover >= 55\%).
}
}
\examples{
LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
# Wing also included -keep_single
poi ="-keep_first -inside 481260 3812920 481310 3812960"
las <- readLAS(LASfile, select = "xyzi", filter = poi)

# For the Wing2015 method, supply a matrix of snag BranchBolePtRatio conditional
# assessment thresholds (see Wing et al. 2015, Table 2, pg. 172)
bbpr_thresholds <- matrix(c(0.80, 0.80, 0.70,
                          0.85, 0.85, 0.60,
                          0.80, 0.80, 0.60,
                          0.90, 0.90, 0.55),
                          nrow =3, ncol = 4)

# Run snag classification and assign classes to each point
las <- segment_snags(las, wing2015(neigh_radii = c(1.5, 1, 2), BBPRthrsh_mat = bbpr_thresholds))

# Plot it all, tree and snag points...
#plot(las, color="snagCls", colorPalette = rainbow(5))

# Filter and plot snag points only
snags <- filter_poi(las, snagCls > 0)
#plot(snags, color="snagCls", colorPalette = rainbow(5)[-1])

# Wing et al's (2015) methods ended with performing tree segmentation on the
# classified and filtered point cloud using the watershed method

}
\references{
Wing, Brian M.; Ritchie, Martin W.; Boston, Kevin; Cohen, Warren B.; Olsen, Michael J. 2015.
Individual snag detection using neighborhood attribute filtered airborne lidar data. Remote
Sensing of Environment. 163: 165-179 https://doi.org/10.1016/j.rse.2015.03.013
}
\author{
Implementation by Andrew Sánchez Meador & Jean-Romain Roussel
}
\concept{snags segmentation algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Class-LAS.R
\docType{class}
\name{LASheader-class}
\alias{LASheader-class}
\title{An S4 class to represent the header of .las or .laz files}
\description{
An S4 class to represent the header of .las or .laz files according to the
\href{https://community.asprs.org/leadership-restricted/leadership-content/public-documents/standards}{LAS file format specifications}.
A \code{LASheader} object contains a \code{list} in the slot \verb{@PHB} with
the data read from the Public Header Block, a \code{list} in the slot \verb{@VLR} with
the data read from the Variable Length Records and a \code{list} in the slot \code{EVLR} with the data read
from the Extended Variable Lenght Records.
}
\section{Slots}{

\describe{
\item{\code{PHB}}{list. Represents the Public Header Block}

\item{\code{VLR}}{list. Represents the Variable Length Records}

\item{\code{EVLR}}{list. Represents the Extended Variable Length Records}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge_spatial.R
\name{merge_spatial}
\alias{merge_spatial}
\title{Merge a point cloud with a source of spatial data}
\usage{
merge_spatial(las, source, attribute = NULL)
}
\arguments{
\item{las}{An object of class \code{LAS}}

\item{source}{An object of class \verb{SpatialPolygons*} or \code{sf} or \code{sfc} or \code{RasterLayer} or
\code{RasterStack} or \code{RasterBrick} or \code{stars}.}

\item{attribute}{character. The name of an attribute in the table of attributes or
the name of a new column in the LAS object. Not relevant for RGB colorization.}
}
\value{
a \code{LAS} object
}
\description{
Merge a point cloud with a source of spatial data. It adds an attribute along each point based on
a value found in the spatial data. Sources of spatial data can be a \verb{SpatialPolygons*}, an \code{sf}/\code{sfc},
a \verb{Raster*}, a \code{stars}, or a \code{SpatRaster}.\cr
\itemize{
\item{\verb{SpatialPolygons*}, \code{sf} and \code{sfc}: it checks if the points belongs within each polygon. If
the parameter \code{attribute} is the name of an attribute in the table of attributes it assigns
to the points the values of that attribute. Otherwise it classifies the points as boolean.
TRUE if the points are in a polygon, FALSE otherwise.}
\item{\code{RasterLayer}, single band \code{stars} or single layer \code{SpatRaster}: it attributes to each point
the value found in each pixel of the raster}.
\item{\code{RasterStack}, \code{RasterBrick}, multibands \code{stars} or multilayer \code{SpatRaster} must have 3
layers for RGB colors. It colorizes the point cloud with RGB values.}
}
}
\examples{
LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
shp     <- system.file("extdata", "lake_polygons_UTM17.shp", package = "lidR")

las   <- readLAS(LASfile, filter = "-keep_random_fraction 0.1")
lakes <- sf::st_read(shp, quiet = TRUE)

# The attribute "inlake" does not exist in the shapefile.
# Points are classified as TRUE if in a polygon
las    <- merge_spatial(las, lakes, "inlakes")     # New attribute 'inlakes' is added.
names(las)

forest <- filter_poi(las, inlakes == FALSE)
#plot(forest)

# The attribute "LAKENAME_1" exists in the shapefile.
# Points are classified with the values of the polygons
las <- merge_spatial(las, lakes, "LAKENAME_1")     # New column 'LAKENAME_1' is added.
names(las)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_area.R
\name{st_area}
\alias{st_area}
\alias{st_area.LAS}
\alias{st_area.LASheader}
\alias{st_area.LAScatalog}
\alias{area,LAS-method}
\alias{area,LASheader-method}
\alias{area,LAScatalog-method}
\title{Surface covered by a LAS* object}
\usage{
\method{st_area}{LAS}(x, ...)

\method{st_area}{LASheader}(x, ...)

\method{st_area}{LAScatalog}(x, ...)

\S4method{area}{LAS}(x, ...)

\S4method{area}{LASheader}(x, ...)

\S4method{area}{LAScatalog}(x, ...)
}
\arguments{
\item{x}{An object of class \code{LAS*}.}

\item{...}{unused.}
}
\value{
numeric. A number in the same units as the coordinate reference system.
}
\description{
Surface covered by a \verb{LAS*} object. The surface covered by a point cloud is mathematically 0.
To compute non zero values the function uses different strategies. The area is computed based on
the number of occupied cells, or on the area of the convex hull of the points depending on the density
and the size of the point cloud. The result is necessarily an approximation that depends on the method
used.\cr For a \code{LAScatalog} it is computed as the sum of the bounding boxes of the files. For
overlapping tiles the value may be larger than the total area covered because some regions are
sampled twice. For a \code{LASheader} it is computed with the bounding box. As a consequence, for the same file
\code{st_area} applied on a LASheader or on a LAS can return slightly different values. \code{st_area()}
extends \code{sf:st_area()}, \code{area()} extends \code{raster:area()}. \code{area()} is provided for backward
compatibility.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_crs.R
\name{st_crs}
\alias{st_crs}
\alias{st_crs.LAS}
\alias{st_crs.LAScatalog}
\alias{st_crs.LASheader}
\alias{st_crs.LAScluster}
\alias{st_crs<-}
\alias{st_crs<-.LAS}
\alias{st_crs<-.LASheader}
\alias{st_crs<-.LAScatalog}
\alias{projection}
\alias{projection<-}
\alias{crs,LASheader-method}
\alias{crs,LAS-method}
\alias{crs<-,LAS-method}
\alias{crs,LAScatalog-method}
\alias{crs,LAScluster-method}
\alias{crs<-,LAScatalog-method}
\alias{crs<-,LASheader-method}
\alias{epsg}
\alias{epsg<-}
\alias{epsg,LASheader-method}
\alias{epsg<-,LASheader-method}
\alias{epsg,LAS-method}
\alias{epsg<-,LAS-method}
\alias{wkt}
\alias{wkt<-}
\alias{wkt,LASheader-method}
\alias{wkt<-,LASheader-method}
\alias{wkt,LAS-method}
\alias{wkt<-,LAS-method}
\title{Get or set the projection of a LAS* object}
\usage{
\method{st_crs}{LAS}(x, ...)

\method{st_crs}{LAScatalog}(x, ...)

\method{st_crs}{LASheader}(x, ...)

\method{st_crs}{LAScluster}(x, ...)

\method{st_crs}{LAS}(x) <- value

\method{st_crs}{LASheader}(x) <- value

\method{st_crs}{LAScatalog}(x) <- value

projection(x, asText = TRUE)

projection(x) <- value

\S4method{crs}{LASheader}(x)

\S4method{crs}{LAS}(x)

\S4method{crs}{LAS}(x, ...) <- value

\S4method{crs}{LAScatalog}(x, asText = FALSE)

\S4method{crs}{LAScluster}(x, asText = FALSE)

\S4method{crs}{LAScatalog}(x, ...) <- value

\S4method{crs}{LASheader}(x, ...) <- value

epsg(object, ...)

epsg(object) <- value

\S4method{epsg}{LASheader}(object, ...)

\S4method{epsg}{LASheader}(object) <- value

\S4method{epsg}{LAS}(object)

\S4method{epsg}{LAS}(object) <- value

wkt(obj)

wkt(obj) <- value

\S4method{wkt}{LASheader}(obj)

\S4method{wkt}{LASheader}(obj) <- value

\S4method{wkt}{LAS}(obj)

\S4method{wkt}{LAS}(obj) <- value
}
\arguments{
\item{x, object, obj}{An object of class LAS*}

\item{...}{Unused.}

\item{value}{A \code{CRS} or a \code{crs} or a \code{proj4string} string or WKT string or an EPSG code.}

\item{asText}{logical. If TRUE, the projection is returned as text. Otherwise a CRS object is returned.}
}
\value{
A \code{st_crs()} returns a \code{sf::crs}. \code{projection()} and \code{crs()} return a \code{sp::CRS} and should
no longer be used.
}
\description{
Get or set the projection of a \verb{LAS*} object.  \code{st_crs()} extends \code{sf:st_crs()}, \code{projection()} and
\code{crs()} extend \code{raster:projection()} and \code{raster:crs()}. \code{projection()} and \code{crs()} are provided
for backward compatibility. For \code{epsg()} and \code{wkt()}, see details.
}
\details{
There are two ways to store the CRS of a point cloud in a LAS file:
\itemize{
\item Store an EPSG code (for LAS 1.0 to 1.3)
\item Store a WTK string (for LAS 1.4)
}

On the other hand, R spatial packages use a \code{crs} object to store the CRS. This is why the CRS is duplicated
in a LAS object. The information belongs within the header in a format that can be written in a
LAS file and in the slot \code{crs}, and also in a format that can be understood by other R packages.
\itemize{
\item \code{st_crs} return the CRS in \code{sf} format.
\item \verb{st_crs<-}: assigns a CRS from a \code{CRS} (\code{sp}), a \code{crs} (\code{sf}), a WKT
string, a proj4string or an epsg code. It updates the header of the LAS
object either with the EPSG code for LAS formats < 1.4 or with a WKT string
\item \code{epsg}: reads the epsg code from the header.
\item \code{wkt}: reads the WKT string from the header.
}
}
\examples{
LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
las <- readLAS(LASfile)

# Get the EPSG code stored in the header (returns 0 if not recorded)
epsg(las)

# Get the WKT string stored in the header (LAS >= 1.4)
wkt(las)

# Overwrite the CRS (but does not reproject)
st_crs(las) <- 26918
las

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Class-LAScatalog.R
\docType{class}
\name{LAScatalog-class}
\alias{LAScatalog-class}
\title{An S4 class to represent a collection of .las or .laz files}
\description{
A \code{LAScatalog} object is a representation of a collection of las/laz files. A \code{LAScatalog} is
a way to manage and batch process a lidar coverage. It allows the user to process a large area, or to
selectively clip data from a large area without loading all the data into computer memory.
A \code{LAScatalog} can be built with the function \link{readLAScatalog}.
}
\details{
A \code{LAScatalog} contains an \code{sf} object to store the geometry and metadata. It is extended with slots
that contain processing options. In \code{lidR}, each function that supports a \code{LAScatalog} as
input will respect these processing options. Internally, processing a catalog is almost always the
same and relies on a few steps:\cr
\enumerate{
\item Define chunks. A chunk is an arbitrarily-defined region of interest (ROI) of the
collection. Altogether, the chunks are a wall-to-wall set of ROIs that encompass the whole dataset.
\item Loop over each chunk (in parallel or not).
\item For each chunk, load the points inside the ROI into R, run some R functions,
return the expected output.
\item Merge the outputs of the different chunks once they are all processed to build a continuous
(wall-to-wall) output.
}

So basically, a \code{LAScatalog} is an object that allows for batch processing but with the specificity
that \code{lidR} does not loop through LAS or LAZ files, but loops seamlessly through chunks that do not
necessarily match with the file pattern. This way \code{lidR} can sequentially process tiny ROIs even if
each file may be individually too big to fit in memory. This is also why point cloud indexation
with \code{lax} files may significantly speed-up the processing.\cr\cr
It is important to note that catalogs with files that overlap each other are not natively supported
by \code{lidR}. When encountering such datasets the user should always filter any overlaps if
possible. This is possible if the overlapping points are flagged, for example in the
'withheld' attribute. Otherwise \code{lidR} will not be able to process the dataset correctly.
}
\section{Slots}{

\describe{
\item{\code{data}}{sf. An \code{sf} \code{data.frame} with the bounding box of each file as well as all the information
read from the header of each LAS/LAZ file.}

\item{\code{processing_options}}{list. A list that contains some settings describing how the collection will be
processed (see dedicated section).}

\item{\code{chunk_options}}{list. A list that contains some settings describing how the collection will be
sub-divided into chunks to be processed (see dedicated section).}

\item{\code{output_options}}{list. A list that contains some settings describing how the collection will return
the outputs (see dedicated section).}

\item{\code{input_options}}{list. A list of parameters to pass to \link{readLAS} (see dedicated section).}

\item{\code{index}}{list. See \link[=lidR-spatial-index]{spatial indexing}.}
}}

\section{Processing options}{

The slot \verb{@processing_options} contains a \code{list} of options that determine how chunks
(the sub-areas that are sequentially processed) are processed.
\itemize{
\item \strong{progress}: boolean. Display a progress bar and a chart of progress. Default is TRUE.
Progress estimation can be enhanced by installing the package \code{progress}. See \link{opt_progress}.
\item \strong{stop_early}: boolean. Stop the processing if an error occurs in a chunk. If \code{FALSE}
the process can run until the end, removing chunks that failed. Default is TRUE and the user should
have no reason to change this. See \link{opt_stop_early}.
\item \strong{wall.to.wall} logical. The catalog processing engine always guarantees to return a
continuous output without edge effects, assuming that the catalog is a wall-to-wall catalog. To do
so, some options are checked internally to guard against bad settings, such as \code{buffer = 0} for an
algorithm that requires a buffer. In rare cases it might be useful to disable these controls. If
\code{wall.to.wall = FALSE} controls are disabled and wall-to-wall outputs cannot be guaranteed.
See \link{opt_wall_to_wall}
}
}

\section{Chunk options}{

The slot \verb{@chunk_options} contains a \code{list} of options that determine how chunks
(the sub-areas that are sequentially processed) are made.
\itemize{
\item \strong{chunk_size}: numeric. The size of the chunks that will be sequentially processed.
A small size allows small amounts of data to be loaded at once, saving computer memory.
With big chunks the computation is usually faster but uses much more memory. If \code{chunk_size = 0} the
chunk pattern is build using the file pattern. The chunks are expecting to be wall-to-wall coverage,
which means that \code{chunk_size = 0} has meaning only if the files are not overlapping. Default is 0 i.e.
by default the processing engine respects the existing tiling pattern. See \link{opt_chunk_size}.
\item \strong{buffer}: numeric. Each chunk can be read with an extra buffer around it to ensure there are
no edge effects between two independent chunks and that the output is continuous. This is mandatory for
some algorithms. Default is 30. See \link{opt_chunk_buffer}.
\item \strong{alignment}: numeric. A vector of size 2 (x and y coordinates, respectively) to align the
chunk pattern. By default the alignment is made along (0,0), meaning that the edge of the first chunk
will belong on x = 0 and y = 0 and all the the other chunks will be multiples of the chunk size.
Not relevant if \code{chunk_size = 0}. See \link{opt_chunk_alignment}.
\item \strong{drop}: integers. A vector of integers that specify the IDs of the chunks that should not be
created. This is designed to enable users to restart a computation that failed without reprocessing
everything. See \link{opt_restart<-}. Technically, this option may be used for partial processing of
a collection, but it generally should not be. Partial processing is already a feature of the engine. See
\href{https://cran.r-project.org/package=lidR/vignettes/lidR-LAScatalog-engine.html#partial-processing}{this vignette}
}
}

\section{Output options}{

The slot \verb{@output_options} contains a \code{list} of options that determine how chunks
(the sub-areas that are sequentially processed) are written. By "written" we mean written to files
or written in R memory.
\itemize{
\item \strong{output_files}: string. If \code{output_files = ""} outputs are returned in R. Otherwise, if
\code{output_files} is a string the outputs will be written to files.
This is useful if the output is too big to be returned in R. A path to a filename template
without a file extension (the engine guesses it for you) is expected. When several files are going to be
written a single string is provided with a template that is automatically filled. For example,
the following file names are possible:
\preformatted{
"/home/user/als/normalized/file_{ID}_segmented"
"C:/user/document/als/zone52_{XLEFT}_{YBOTTOM}_confidential"
"C:/user/document/als/{ORIGINALFILNAME}_normalized"
}
This option will generate as many filenames as needed with custom names for each file. The allowed
templates are \code{{XLEFT}, {XRIGHT}, {YBOTTOM}, {YTOP}, {ID}, {XCENTER},
{YCENTER}, {ORIGINALFILENAME}}. See \link{opt_output_files}.
\item \strong{drivers}: list. This contains all the drivers required to seamlessly write \verb{Raster*},
\code{SpatRaster}, \code{stars}, \verb{Spatial*}, \code{sf}, and \code{LAS} objects. It is recommended that only advanced
users change this option. A dedicated page describes the drivers in \link{lidR-LAScatalog-drivers}.
\item \strong{merge}: boolean. Multiple objects are merged into a single object at the end of the processing.
See \link{opt_merge}.
}
}

\section{Input options}{

The slot \verb{@input_options} contains a \code{list} of options that are passed to the function
\link{readLAS}. Indeed, the \code{readLAS} function is not called directly by the user but by the
internal processing engine. Users can propagate these options through the \code{LAScatalog} settings.
\itemize{
\item \strong{select}: string. The option \code{select}. Usually this option is not respected because
each function knows which data must be loaded or not. This is documented in each function. See
\link{opt_select}.
\item \strong{filter}: string. The option \code{filter}. See \link{opt_filter}.
}
}

\examples{
\dontrun{
# Build a catalog
ctg <- readLAScatalog("filder/to/las/files/", filter =  "-keep_first")


# Summary gives a summary of how the catalog will be processed
summary(ctg)

# We can seamlessly use lidR functions
hmean <- pixel_metrics(ctg, ~~mean(Z), 20)
ttops <- tree_detection(ctg, lmf(5))

# For low memory config it is probably advisable not to load entire files
# and process chunks instead
opt_chunk_size(ctg) <- 500

# Sometimes the output is likely to be very large
# e.g. large coverage and small resolution
dtm <- rasterize_terrain(ctg, 1, tin())

# In that case it is advisable to write the output(s) to files
opt_output_files(ctg) <- "path/to/folder/DTM_chunk_{XLEFT}_{YBOTTOM}"

# Raster will be written to disk. The list of written files is returned
# or, in this specific case, a virtual raster mosaic.
dtm <- rasterize_terrain(ctg, 1, tin())

# When chunks are files the original names of the las files can be preserved
opt_chunk_size(ctg) <- 0
opt_output_files(ctg) <- "path/to/folder/DTM_{ORIGINALFILENAME}"
dtm <- rasterize_terrain(ctg, 1, tin())

# For some functions, files MUST be written to disk. Indeed, it is certain that R cannot
# handle the entire output.
opt_chunk_size(ctg) <- 0
opt_output_files(ctg) <- "path/to/folder/{ORIGINALFILENAME}_norm"
opt_laz_compression(ctg) <- TRUE
new_ctg <- normalize_height(ctg, tin())

# The user has access to the catalog engine through the functions catalog_apply
# and catalog_map
output <- catalog_apply(ctg, FUN, ...)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/connected_components.R
\name{knn_distance}
\alias{knn_distance}
\title{Computes the Distance to k-Nearest Neighbors}
\usage{
knn_distance(las, k = 10, name = "distance")
}
\arguments{
\item{las}{A LAS object representing the point cloud data.}

\item{k}{The number of nearest neighbors.}

\item{name}{A string specifying the name of the new attribute used to store the computed distances.}
}
\value{
A LAS object with an additional attribute named as specified by `name`.
}
\description{
Computes the average distance between each point and its k-nearest neighbors in a point cloud.
The results are stored in a new attribute.
}
\examples{
LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
las <- readLAS(LASfile, select = "xyz", filter = "-inside 481250 3812980 481300 3813030")
las = knn_distance(las)
#plot(las, color = "distance", breaks = "quantile", legend = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rasterize.R, R/rasterize_canopy.R,
%   R/rasterize_density.R, R/rasterize_terrain.R
\name{rasterize}
\alias{rasterize}
\alias{rasterize_canopy}
\alias{rasterize_density}
\alias{rasterize_terrain}
\title{Rasterize a point cloud}
\usage{
rasterize_canopy(las, res = 1, algorithm = p2r(), ...)

rasterize_density(las, res = 4, ...)

rasterize_terrain(
  las,
  res = 1,
  algorithm = tin(),
  use_class = c(2L, 9L),
  shape = "convex",
  ...
)
}
\arguments{
\item{las}{An object of class \link[lidR:LAS-class]{LAS} or \link[lidR:LAScatalog-class]{LAScatalog}.}

\item{res}{numeric. The size of a grid cell in point cloud coordinates units. Can also be
\code{RasterLayer} or a \code{stars} or a \code{SpatRaster} used as layout.}

\item{algorithm}{function. A function that implements an algorithm to compute a digital surface model
or a digital terrain model. \code{lidR} implements \link{p2r}, \link{dsmtin}, \link{pitfree}
for digital surface models, and \link{knnidw}, \link{tin}, and \link{kriging} for digital terrain
models (see respective documentation and examples).}

\item{...}{Use \code{pkg = "terra|raster|stars"} to get an output in \code{SpatRaster}, \code{RasterLayer}
or \code{stars} format. Default is \code{getOption("lidR.raster.default")}.}

\item{use_class}{integer vector. By default the terrain is computed by using ground points
(class 2) and water points (class 9).}

\item{shape}{By default the interpolation is made only within the \code{"convex"} hull of
the point cloud to get a DTM with the shape of the point cloud. This prevents meaningless
interpolations where there is no data. It can also be \code{"concave"} or \code{"bbox"}. It can also be an \code{sfc}
to define a polygon in which to perform the interpolation.}
}
\value{
\code{RasterLayer} or a \code{stars} or a \code{SpatRaster} depending on the settings.
}
\description{
Rasterize a point cloud in different ways to compute a DTM, a CHM or a density map. Most
raster products can be computed with \link{pixel_metrics} but some are more complex and require
dedicated and optimized functions. See Details and Examples.
}
\details{
\describe{
\item{\code{rasterize_terrain}}{Interpolates the ground points and creates a rasterized
digital terrain model. The algorithm uses the points classified as "ground" and "water"
(Classification = 2 and 9, respectively, according to
\href{https://community.asprs.org/leadership-restricted/leadership-content/public-documents/standards}{LAS file format specifications})
to compute the interpolation. How well the edges of the dataset are interpolated depends on the
interpolation method used. A buffer around the region of interest is always recommended to avoid
edge effects.}
\item{\code{rasterize_canopy}}{Creates a digital surface model (DSM) using several
possible algorithms. If the user provides a normalized point cloud, the output is indeed a canopy
height model (CHM).}
\item{\code{rasterize_density}}{Creates a map of the point density. If a "pulseID"
attribute is found, also returns a map of the pulse density.}
}
}
\section{Non-supported LAScatalog options}{

The option \code{select} is not supported and not respected in \verb{rasterize_*} because it is internally
known what is best to select.\cr
The option \code{chunk_buffer} is not supported and not respected in \code{rasterize_canopy} and
\code{rasterize_density} because it is not necessary.
}

\examples{

# =====================
# Digital Terrain Model
# =====================

LASfile <- system.file("extdata", "Topography.laz", package="lidR")
las = readLAS(LASfile, filter = "-inside 273450 5274350 273550 5274450")
#plot(las)

dtm1 = rasterize_terrain(las, algorithm = knnidw(k = 6L, p = 2))
dtm2 = rasterize_terrain(las, algorithm = tin())

\dontrun{
dtm3 = rasterize_terrain(las, algorithm = kriging(k = 10L))

plot(dtm1, col = gray(0:25/25))
plot(dtm2, col = gray(0:25/25))
plot(dtm3, col = gray(0:25/25))
plot_dtm3d(dtm1)
plot_dtm3d(dtm2)
plot_dtm3d(dtm3)
}

# =====================
# Digital Surface Model
# =====================

LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
las <- readLAS(LASfile, filter = "-inside 481280 3812940 481330 3812990")
col <- height.colors(15)

# Points-to-raster algorithm with a resolution of 1 meter
chm <- rasterize_canopy(las, res = 1, p2r())
plot(chm, col = col)

# Points-to-raster algorithm with a resolution of 0.5 meters replacing each
# point by a 20-cm radius circle of 8 points
chm <- rasterize_canopy(las, res = 0.5, p2r(0.2))
plot(chm, col = col)

# Basic triangulation and rasterization of first returns
chm <- rasterize_canopy(las, res = 0.5, dsmtin())
plot(chm, col = col)

# Khosravipour et al. pitfree algorithm
chm <- rasterize_canopy(las, res = 0.5, pitfree(c(0,2,5,10,15), c(0, 1.5)))
plot(chm, col = col)

# ====================
# Digital Density Map
# ====================

LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
las <- readLAS(LASfile,  filter = "-inside 684800 5017800 684900 5017900")

d <- rasterize_density(las, 5)
plot(d)

las <- retrieve_pulses(las)
d <- rasterize_density(las)
plot(d)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics_cloud.R, R/metrics_crowns.R,
%   R/metrics_hexagons.R, R/metrics_pixels.R, R/metrics_plot.R,
%   R/metrics_polygon.R, R/metrics_template.R, R/metrics_voxels.R
\name{aggregate}
\alias{aggregate}
\alias{cloud_metrics}
\alias{crown_metrics}
\alias{hexagon_metrics}
\alias{pixel_metrics}
\alias{plot_metrics}
\alias{polygon_metrics}
\alias{template_metrics}
\alias{voxel_metrics}
\title{Metric derivation at different levels of regularization}
\usage{
cloud_metrics(las, func, ...)

crown_metrics(
  las,
  func,
  geom = "point",
  concaveman = c(3, 0),
  attribute = "treeID",
  ...
)

hexagon_metrics(las, func, area = 400, ...)

pixel_metrics(las, func, res = 20, start = c(0, 0), ...)

plot_metrics(las, func, geometry, ..., radius)

polygon_metrics(las, func, geometry, ...)

template_metrics(las, func, template, filter = NULL, by_echo = "all", ...)

voxel_metrics(las, func, res = 1, ..., all_voxels = FALSE)
}
\arguments{
\item{las}{An object of class \link[lidR:LAS-class]{LAS} or \link[lidR:LAScatalog-class]{LAScatalog}.}

\item{func}{formula or expression. An expression to be applied to each element of the template (see
section "Parameter func").}

\item{...}{propagated to \code{template_metrics} i.e. \code{filter} and \code{by_echo}. \code{pixel_metrics()} also
supports \code{pkg = "terra|raster|stars"} to get an output in \code{SpatRaster}, \verb{Raster*}
or \code{stars} format. Default is \code{getOption("lidR.raster.default")}.}

\item{geom}{character. Geometry type of the output. Can be 'point', 'convex', 'concave' or 'bbox'.}

\item{concaveman}{numeric. Only if \code{type = "concave"}. Vector with the two parameters of the
function \link{concaveman}.}

\item{attribute}{character. The column name of the attribute containing tree IDs. Default is
\code{"treeID"}}

\item{area}{numeric. Area of the hexagons}

\item{res}{numeric. The resolution of the output. Can optionally be a \code{RasterLayer} or a \code{stars} or
a \code{SpatRaster}. In that case the raster is used as the template.}

\item{start}{vector of x and y coordinates for the reference raster. Default is (0,0) meaning that the
grid aligns on (0,0). Not consiered if \code{res} is a raster}

\item{geometry}{A spatial vector object. \code{sp} and \code{sf}' objects are supported. \code{plot_metrics()}
supports point and polygons but \code{polygon_metrics()} supports only polygons.}

\item{radius}{numeric. If the geometry is spatial points a radius must be defined. Support one
radius or a vector of radii for variable plot sizes.}

\item{template}{can be of many types and corresponds to the different levels of regularization.
\code{RasterLayer/stars/SpatRaster}, \code{sf/sfc} (polygons), \code{numeric}, \code{bbox}, \code{NULL}. The metrics are
computed for each element of the template. See examples.}

\item{filter}{formula of logical predicates. Enables the function to run only on points of interest
in an optimized way. See examples.}

\item{by_echo}{characters. The metrics are computed multiple times for different echo types. Can
be one or more of "all", "first", "intermediate", "lastofmany", "single", and "multiple". See examples.
Default is "all" meaning that it computes metrics with all points provided.}

\item{all_voxels}{boolean. By default the function returns only voxels that
contain 1 or more points. Empty voxels do not exist as the metrics are undefined.
If \code{all_voxels = TRUE} all the voxels are returned and metrics are NA for
voxels with 0 points.}
}
\value{
Depends on the function, the template and the number of metrics. Can be a \code{RasterLayer},
a \code{RasterBrick}, a \code{stars}, a \code{SpatRaster} a \code{sf/sfc}, a \code{list}, a \code{SpatialPolygonDataFrame}, or
a \code{data.table}. Functions are supposed to return an object that is best suited for storing the level
of regularization needed.
}
\description{
\code{template_metrics()} computes a series of user-defined descriptive statistics for a LiDAR dataset
within each element of a template. Depending on the template it can be for each pixel of a raster
(area-based approach), or each polygon, or each segmented tree, or on the whole point cloud. Other
functions are convenient and simplified wrappers around \code{template_metrics()} and are expected to be
the actual functions used. See Details and Examples.
}
\details{
\describe{
\item{\code{pixel_metrics}}{Area-based approach. Computes metrics in a square tessellation. The output is a
raster.}
\item{\code{hexagon_metrics}}{Computes metrics in an hexagon tessellation. The output is a \code{sf/sfc_POLYGON}}
\item{\code{plot_metrics}}{Computes metrics for each plot of a ground inventory by 1. clipping the plot
inventories with \link{clip_roi}, 2. computing the user's metrics for each plot with  \link{cloud_metrics}, and
3. combining spatial data and metrics into one data.frame ready for statistical modelling with
\code{cbind}. The output is of the class of the input.}
\item{\code{cloud_metrics}}{Computes a series of user-defined descriptive statistics for an entire point cloud.
The output is a \code{list}}
\item{\code{crown_metrics}}{Once the trees are segmented, i.e. attributes exist in the
point cloud that reference each tree, computes a set of user-defined descriptive statistics for
each individual tree. The output can be spatial points or spatial polygons (\code{sf/sfc_POINT} or \code{sf/sfc_POLYGON})}
\item{\code{voxel_metrics}}{Is a 3D version of \code{pixel_metrics}. It creates a 3D matrix of voxels with a given
resolution. It creates a voxel from the cloud of points if there is at least one point. The output is
a \code{data.frame}}
\item{\code{point_metrics}}{Is a bit more complex and is documented in \link{point_metrics}}
}
}
\section{Parameter \code{func}}{

The function to be applied to each cell is a classical function (see examples) that
returns a labelled list of metrics. For example, the following function \code{f} is correctly formed.
\preformatted{
f = function(x) {list(mean = mean(x), max = max(x))}
}
And could be applied either on the \code{Z} coordinates or on the intensities. These two
statements are valid:
\preformatted{
pixel_metrics(las, f(Z), res = 20)
voxel_metrics(las, f(Intensity), res = 2)
}
The following existing functions allow the user to
compute some predefined metrics: \link[=stdmetrics]{stdmetrics}
\link[=entropy]{entropy}, \link[=VCI]{VCI}, \link[=LAD]{LAD}. But usually users must write their own
functions to create metrics. \code{template_metrics} will dispatch the point cloud in the user's
function.
}

\examples{
LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
las <- readLAS(LASfile, filter = "-keep_random_fraction 0.5")
col <- sf::sf.colors(15)
fun1 <- ~list(maxz = max(Z))
fun2 <- ~list(q85 = quantile(Z, probs = 0.85))

set_lidr_threads(1) ; data.table::setDTthreads(1) # for cran only

# ================
# CLOUD METRICS
# ================

cloud_metrics(las, .stdmetrics_z)

# ================
# PIXEL METRICS
# ================

m <- pixel_metrics(las, fun1, 20)
#plot(m, col = col)

# ================
# PLOT METRICS
# ================

shpfile <- system.file("extdata", "efi_plot.shp", package="lidR")
inventory <- sf::st_read(shpfile, quiet = TRUE)
inventory # contains an ID and a Value Of Interest (VOI) per plot

m <- plot_metrics(las, fun2, inventory, radius = 11.28)
#plot(header(las))
#plot(m["q85"], pch = 19, cex = 3, add = TRUE)

\donttest{
# Works with polygons as well
inventory <- sf::st_buffer(inventory, 11.28)
#plot(header(las))
#plot(sf::st_geometry(inventory), add = TRUE)
m <- plot_metrics(las, .stdmetrics_z, inventory)
#plot(m["zq85"], pch = 19, cex = 3, add = TRUE)
}

# ================
# VOXEL METRICS
# ================

m <- voxel_metrics(las, length(Z), 8)
m <- voxel_metrics(las, mean(Intensity), 8)
#plot(m, color = "V1", colorPalette = heat.colors(50), trim = 60)
#plot(m, color = "V1", colorPalette = heat.colors(50), trim = 60, voxel = TRUE)

# ================
# CROWN METRICS
# ================

# Already tree-segmented point cloud
LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
trees <- readLAS(LASfile, filter = "-drop_z_below 0")

metrics <- crown_metrics(trees, .stdtreemetrics)
#plot(metrics["Z"], pch = 19)

metrics <- crown_metrics(trees, .stdtreemetrics, geom = "convex")
#plot(metrics["Z"])

metrics <- crown_metrics(trees, .stdtreemetrics, geom = "bbox")
#plot(metrics["Z"])

\donttest{
metrics <- crown_metrics(trees, .stdtreemetrics, geom = "concave")
#plot(metrics["Z"])
}
# ================
# ARGUMENT FILTER
# ================

# Compute using only some points: basic
first = filter_poi(las, ReturnNumber == 1)
metrics = pixel_metrics(first, mean(Z), 20)

# Compute using only some points: optimized
# faster and uses less memory. No intermediate object
metrics = pixel_metrics(las, mean(Z), 20, filter = ~ReturnNumber == 1)

# Compute using only some points: best
# ~50\% faster and uses ~10x less memory
las = readLAS(LASfile, filter = "-keep_first")
metrics = pixel_metrics(las, mean(Z), 20)

# ================
# ARGUMENT BY_ECHO
# ================

func = ~list(avgI = mean(Intensity))
echo = c("all", "first","multiple")

# func defines one metric but 3 are computed respectively for: (1) all echo types,
# (2) for first returns only and (3) for multiple returns only
metrics <- pixel_metrics(las, func, 20, by_echo = echo)
#plot(metrics, col = heat.colors(25))

cloud_metrics(las, func, by_echo = echo)

\dontrun{
# ================
# TEMPLATE METRICS
# ================

# a raster as template
template <- raster::raster(extent(las), nrow = 15, ncol = 15)
raster::crs(template) <- crs(las)
m <- template_metrics(las, fun1, template)
#plot(m, col = col)

# a sfc_POLYGON as template
sfc <- sf::st_as_sfc(st_bbox(las))
template <- sf::st_make_grid(sfc, cellsize = 20, square = FALSE)
m <- template_metrics(las, fun1, template)
#plot(m)

# a bbox as template
template <- st_bbox(las) + c(50,30,-50,-70)
plot(sf::st_as_sfc(st_bbox(las)), col = "gray")
plot(sf::st_as_sfc(template), col = "darkgreen", add = TRUE)
m <- template_metrics(las, fun2, template)
print(m)

# ================
# CUSTOM METRICS
# ================

# Define a function that computes custom metrics
# in an R&D perspective.
myMetrics = function(z, i) {
  metrics = list(
     zwimean = sum(z*i)/sum(i), # Mean elevation weighted by intensities
     zimean  = mean(z*i),       # Mean products of z by intensity
     zsqmean = sqrt(mean(z^2))) # Quadratic mean

   return(metrics)
}

# example with a stars template
template <- stars::st_as_stars(st_bbox(las), dx = 10, dy = 10)
m <- template_metrics(las, myMetrics(Z, Intensity), template)
#plot(m, col = col)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algorithm-dtm.R
\name{dtm_tin}
\alias{dtm_tin}
\alias{tin}
\title{Spatial Interpolation Algorithm}
\usage{
tin(..., extrapolate = knnidw(3, 1, 50))
}
\arguments{
\item{...}{unused}

\item{extrapolate}{There are usually a few points outside the convex hull, determined by the ground
points at the very edge of the dataset, that cannot be interpolated with a triangulation.
Extrapolation is done using \link{knnidw} by default.}
}
\description{
This function is made to be used in \link{rasterize_terrain} or \link{normalize_height}. It
implements an algorithm for spatial interpolation. Spatial interpolation is based on a Delaunay
triangulation, which performs a linear interpolation within each triangle. There are usually a
few points outside the convex hull, determined by the ground points at the very edge of the dataset,
that cannot be interpolated with a triangulation. Extrapolation can be performed with another algorithm.
}
\examples{
LASfile <- system.file("extdata", "Topography.laz", package="lidR")
las = readLAS(LASfile, filter = "-inside 273450 5274350 273550 5274450")

#plot(las)

dtm = rasterize_terrain(las, algorithm = tin())

#plot(dtm)
#plot_dtm3d(dtm)
}
\seealso{
Other dtm algorithms: 
\code{\link{dtm_idw}},
\code{\link{dtm_kriging}}
}
\concept{dtm algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algorithm-dsm.R
\name{dsm_tin}
\alias{dsm_tin}
\alias{dsmtin}
\title{Digital Surface Model Algorithm}
\usage{
dsmtin(max_edge = 0, highest = TRUE)
}
\arguments{
\item{max_edge}{numeric. Maximum edge length of a triangle in the Delaunay triangulation.
If a triangle has an edge length greater than this value it will be removed to trim dummy interpolation
on non-convex areas. If \code{max_edge = 0} no trimming is done (see examples).}

\item{highest}{bool. By default it keeps only the highest point per pixel before to triangulate to
decrease computation time. If highest = FALSE all first returns are used.}
}
\description{
This function is made to be used in \link{rasterize_canopy}. It implements an algorithm for digital
surface model computation using a Delaunay triangulation of first returns with a linear interpolation
within each triangle.
}
\examples{
LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
las <- readLAS(LASfile)
col <- height.colors(50)

# Basic triangulation and rasterization of first returns
chm <- rasterize_canopy(las, res = 1, dsmtin())
plot(chm, col = col)

\dontrun{
# Potentially complex concave subset of point cloud
x = c(481340, 481340, 481280, 481300, 481280, 481340)
y = c(3812940, 3813000, 3813000, 3812960, 3812940, 3812940)
las2 = clip_polygon(las,x,y)
plot(las2)

# Because the TIN interpolation is done within the convex hull of the point cloud
# dummy pixels are interpolated that are correct according to the interpolation method
# used, but meaningless in our CHM
chm <- rasterize_canopy(las2, res = 0.5, dsmtin())
plot(chm, col = col)

# Use 'max_edge' to trim dummy triangles
chm = rasterize_canopy(las2, res = 0.5, dsmtin(max_edge = 3))
plot(chm, col = col)
}
}
\seealso{
Other digital surface model algorithms: 
\code{\link{dsm_pitfree}},
\code{\link{dsm_point2raster}},
\code{\link{dsm_spikefree}}
}
\concept{digital surface model algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R, R/utils_colors.R
\name{plot}
\alias{plot}
\alias{plot,LAS,missing-method}
\alias{plot,LAScatalog,missing-method}
\alias{plot,LASheader,missing-method}
\alias{height.colors}
\alias{forest.colors}
\alias{random.colors}
\alias{pastel.colors}
\title{Plot a LAS* object}
\usage{
plot(x, y, ...)

\S4method{plot}{LAS,missing}(
  x,
  y,
  ...,
  color = "Z",
  pal = "auto",
  bg = "black",
  breaks = "pretty",
  nbreaks = "auto",
  backend = "rgl",
  clear_artifacts = TRUE,
  axis = FALSE,
  legend = FALSE,
  add = FALSE,
  voxel = FALSE,
  NAcol = "lightgray",
  mapview = FALSE
)

\S4method{plot}{LAScatalog,missing}(x, y, mapview = FALSE, chunk_pattern = FALSE, overlaps = FALSE, ...)

\S4method{plot}{LASheader,missing}(x, y, mapview = FALSE, ...)

height.colors(n)

forest.colors(n)

random.colors(n)

pastel.colors(n)
}
\arguments{
\item{x}{A \code{LAS*} object}

\item{y}{Unused (inherited from R base)}

\item{...}{Will be passed to \link[rgl:3dobjects]{points3d} (LAS) or \link[graphics:plot.default]{plot}
if \code{mapview = FALSE} or to `mapview()` if \code{mapview = TRUE} (LAScatalog).}

\item{color}{characters. The attribute used to color the point cloud. Default is Z coordinates. RGB
is an allowed string even if it refers to three attributes simultaneously.}

\item{pal}{palette function, similar to heat.colors, or palette values. Default is \code{"auto"}
providing an automatic coloring depending on the attribute \code{color}}

\item{bg}{The color for the background. Default is black.}

\item{breaks}{either a numeric vector with the actual breaks, or a name of a method accepted
by the style argument of \link[classInt:classIntervals]{classIntervals}}

\item{nbreaks}{Number of colors breaks.}

\item{backend}{character. Can be \code{"rgl"} or \code{"lidRviewer"}. If \code{"rgl"} is chosen
the display relies on the \code{rgl} package. If \code{"lidRviewer"} is chosen it relies on the
\code{lidRviewer} package, which is much more efficient and can handle million of points
using less memory. \code{lidRviewer} is not available on CRAN yet and should
be installed from github (see. \url{https://github.com/r-lidar/lidRviewer}).}

\item{clear_artifacts}{logical. It is a known and documented issue that the 3D visualisation with
\code{rgl} displays artifacts. The points look aligned and/or regularly spaced in some view angles.
This is because \code{rgl} computes with single precision \code{float}. To fix that the point
cloud is shifted to (0,0) to reduce the number of digits needed to represent its coordinates.
The drawback is that the point cloud is not plotted at its actual coordinates.}

\item{axis}{logical. Display axis on XYZ coordinates.}

\item{legend}{logical. Display a gradient colour legend.}

\item{add}{If \code{FALSE} normal behaviour otherwise must be the output of a prior plot function
to enable the alignment of a second point cloud.}

\item{voxel}{boolean or numeric. Displays voxels instead of points. Useful to render the output
of \link{voxelize_points}, for example. However it is computationally demanding to render and can
easily take 15 seconds for 10000 voxels. It should be reserved for small scenes. If boolean the voxel
resolution is guessed automatically. Otherwise users can provide the size of the voxels. To reduce the rendering time,
an internal optimization removes voxels that are not visible when surrounded by other voxels.}

\item{NAcol}{a color for NA values.}

\item{mapview}{logical. If \code{FALSE} the catalog is displayed in a regular plot from R base.
Since v4.0.4 `mapview = TRUE` is also possible with LAS objects.}

\item{chunk_pattern}{logical. Display the current chunk pattern used to process the catalog.}

\item{overlaps}{logical. Highlight the overlaps between files.}

\item{n}{The number of colors (> 1) to be in the palette}
}
\description{
Plot displays a 3D interactive windows based on rgl for \link{LAS} objects\cr\cr
Plot displays an interactive view for \link[=LAScatalog-class]{LAScatalog} objects with pan and
zoom capabilities based on `mapview()` from package `mapview`. If the coordinate reference
system (CRS) of the \code{LAScatalog} is non empty, the plot can be displayed on top of base maps
(satellite data, elevation, street, and so on).\cr\cr
Plot displays a \link[=LASheader-class]{LASheader} object exactly like it displays a LAScatalog
object.
}
\examples{
\dontrun{
LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
las <- readLAS(LASfile)

plot(las)
plot(las, color = "Intensity")
plot(las, color = "ScanAngleRank", pal = rainbow)

# If outliers break the color range, use the breaks parameter
las$Intensity[150] <- 1000L
plot(las, color = "Intensity")
plot(las, color = "Intensity", breaks = "quantile", nbreaks = 50)

plot(las, color = "Classification")

# This dataset is already tree segmented
plot(las, color = "treeID")
plot(las, color = "treeID", pal = random.colors)


# single file LAScatalog using data provided in lidR
ctg = readLAScatalog(LASfile)
plot(ctg)
plot(ctg, map = T, map.types = "Esri.WorldImagery")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_transform.R
\name{st_transform}
\alias{st_transform}
\alias{st_transform.LAS}
\title{Transform or convert coordinates of LAS objects}
\usage{
\method{st_transform}{LAS}(x, crs, ...)
}
\arguments{
\item{x, }{An object of class LAS}

\item{crs}{crs object from sf or CRS object from sp}

\item{...}{additional arguments \code{scale} and \code{xoffset}, \code{yoffset}, \code{zoffset}
for the output LAS objects. It is not  mandatory but recommended to consider providing such
information. Otherwise it will be guessed automatically which might not be the best choice.}
}
\value{
A LAS object
}
\description{
Transform or convert coordinates of LAS objects  \code{st_transform()} extends \code{sf::st_transform()}
}
\examples{
LASfile <- system.file("extdata", "example.laz", package="rlas")
las = readLAS(LASfile)
st_crs(las)$Name
st_bbox(las)
tlas <- sf::st_transform(las, sf::st_crs(26918))
st_crs(tlas)$Name
st_bbox(tlas)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algorithm-gnd.R
\name{gnd_csf}
\alias{gnd_csf}
\alias{csf}
\title{Ground Segmentation Algorithm}
\usage{
csf(
  sloop_smooth = FALSE,
  class_threshold = 0.5,
  cloth_resolution = 0.5,
  rigidness = 1L,
  iterations = 500L,
  time_step = 0.65
)
}
\arguments{
\item{sloop_smooth}{logical. When steep slopes exist, set this parameter to TRUE to reduce
errors during post-processing.}

\item{class_threshold}{scalar. The distance to the simulated cloth to classify a point cloud into ground
and non-ground. The default is 0.5.}

\item{cloth_resolution}{scalar. The distance between particles in the cloth. This is usually set to the
average distance of the points in the point cloud. The default value is 0.5.}

\item{rigidness}{integer. The rigidness of the cloth. 1 stands for very soft (to fit rugged
terrain), 2 stands for medium, and 3 stands for hard cloth (for flat terrain). The default is 1.}

\item{iterations}{integer. Maximum iterations for simulating cloth. The default value is 500. Usually,
there is no need to change this value.}

\item{time_step}{scalar. Time step when simulating the cloth under gravity. The default value
is 0.65. Usually, there is no need to change this value. It is suitable for most cases.}
}
\description{
This function is made to be used in \link{classify_ground}. It implements an algorithm for segmentation
of ground points base on a Cloth Simulation Filter. This method is a strict implementation of
the CSF algorithm made by Zhang et al. (2016) (see references) that relies on the authors' original
source code written and exposed to R via the the \code{RCSF} package.
}
\examples{
if (require(RCSF, quietly = TRUE))
{
LASfile <- system.file("extdata", "Topography.laz", package="lidR")
las <- readLAS(LASfile, select = "xyzrn")

mycsf <- csf(TRUE, 1, 1, time_step = 1)
las <- classify_ground(las, mycsf)
#plot(las, color = "Classification")
}
}
\references{
W. Zhang, J. Qi*, P. Wan, H. Wang, D. Xie, X. Wang, and G. Yan, “An Easy-to-Use Airborne LiDAR Data
Filtering Method Based on Cloth Simulation,” Remote Sens., vol. 8, no. 6, p. 501, 2016.
(http://www.mdpi.com/2072-4292/8/6/501/htm)
}
\seealso{
Other ground segmentation algorithms: 
\code{\link{gnd_mcc}},
\code{\link{gnd_pmf}},
\code{\link{ptd}()}
}
\concept{ground segmentation algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/engine.R, R/engine_apply.R, R/engine_chunks.R,
%   R/engine_crop.R, R/engine_merge.R, R/engine_write.R
\name{engine}
\alias{engine}
\alias{engine_apply}
\alias{engine_chunks}
\alias{engine_crop}
\alias{engine_merge}
\alias{engine_write}
\title{Functions for the LAScatalog processing engine not meant to be called directly by users}
\usage{
engine_apply(
  .CHUNKS,
  .FUN,
  .PROCESSOPT,
  .OUTPUTOPT,
  .GLOBALS = NULL,
  .AUTOREAD = FALSE,
  .AUTOCROP = FALSE,
  ...
)

engine_chunks(ctg, realignment = FALSE, plot = opt_progress(ctg))

engine_crop(x, bbox)

engine_merge(ctg, any_list, ...)

engine_write(x, path, drivers)
}
\arguments{
\item{.CHUNKS}{list. list of LAScluster}

\item{.FUN}{function. function that respects a template (see \link{catalog_apply})}

\item{.PROCESSOPT}{list. Processing option}

\item{.OUTPUTOPT}{list. Output option}

\item{.GLOBALS}{list. Force export of some object in workers}

\item{.AUTOREAD}{bool. Enable autoread}

\item{.AUTOCROP}{bool. Enable autocrop}

\item{...}{parameters of .FUN}

\item{ctg}{LAScatalog}

\item{realignment}{\code{FALSE} or \code{list(res = x, start = c(y,z))}. Sometimes the chunk must
be aligned with a raster, for example to ensure the continuity of the output. If the chunk size is
800 and the expected product is a raster with a resolution of 35, 800 and 35 are not compatible
and will create 2 different partial pixels on the edges. The realignment option forces the
chunk to fit the grid alignment.}

\item{plot}{logical. Displays the chunk pattern.}

\item{x}{LAS, Raster, stars, SpatRaster,sf, sfc, Spatial}

\item{bbox}{bbox}

\item{any_list}{list of LAS, Raster, stars, SpatRaster, sf, sfc, Spatial, data.frame}

\item{path}{strings}

\item{drivers}{list. Drivers of a LAScatalog}
}
\description{
Functions for the LAScatalog processing engine not meant to be called directly by users.
They are exported for debugging and to simplify export of internal functions when processing in
parallel
}
\seealso{
Other LAScatalog processing engine: 
\code{\link{engine_options}}

Other LAScatalog processing engine: 
\code{\link{engine_options}}
}
\concept{LAScatalog processing engine}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/catalog_boundaries.R
\name{catalog_boundaries}
\alias{catalog_boundaries}
\title{Computes the polygon that encloses the points}
\usage{
catalog_boundaries(ctg, ...)
}
\arguments{
\item{ctg}{A LAScatalog}

\item{...}{propagated to \link{st_concave_hull}}
}
\value{
A LAScatalog with true boundaries
}
\description{
Computes the polygon that encloses the points. It reads all the files one by one and computes a
concave hull using the \link{st_concave_hull} function. When all the hulls are computed it updates the
LAScatalog to set the true polygons instead of the bounding boxes.
}
\section{Non-supported LAScatalog options}{

The options `select`, `output files`, `chunk size`, `chunk buffer`, `chunk alignment` are not
supported and not respected in  `catalog_boundaries*` because the function must always process by
file, without buffer and knows which attributes to load.
}

\examples{
LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
ctg <- readLAScatalog(LASfile, filter = "-drop_z_below 0.5")
ctg2 <- catalog_boundaries(ctg, concavity = 1, length_threshold = 15)
plot(ctg)
plot(ctg2, add = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit_shapes.R
\name{fit_circle}
\alias{fit_circle}
\title{Fits 2D Circles on a Point Cloud}
\usage{
fit_circle(points, num_iterations = 100, inlier_threshold = 0.01)
}
\arguments{
\item{points}{A LAS object representing a clustered slice, or an nx3 matrix of point coordinates.}

\item{num_iterations}{The number of iterations for the RANSAC fitting algorithm.}

\item{inlier_threshold}{A threshold value; points are considered inliers if their residuals are
below this value.}
}
\value{
A list containing the circle parameters and additional information for determining whether
the data fit a circular shape:
\itemize{
\item \code{center_x}, \code{center_y}, \code{radius}: parameters of the fitted circle.
\item \code{z}: average elevation of the circle.
\item \code{rmse}: root mean square error between the circle and all points.
\item \code{covered_arc_degree}: angular sector covered by inlier points.
\item \code{percentage_inlier}: percentage of points that are inliers
\item \code{percentage_inside}: percentage of points inside the circle
\item \code{inliers}: IDs of the inlier points.
}
}
\description{
Fits a 2D horizontally-aligned circle to a set of 3D points. The method uses RANSAC-based fitting
for robust estimation. The function returns a list with the circle parameters and additional data
to help determine whether the points form a circular shape. While it is always possible to fit a
circle to any dataset, the additional information assists in evaluating if the data genuinely
represent a circular pattern. The root mean square error (RMSE) may not always be a reliable metric
for assessing the quality of the fit due to non-circular elements (such as branches) that can arbitrarily
increase the RMSE of an otherwise well-fitted circle, as shown in the example below. Therefore, the
function also returns the angular range of the data, which indicates the spread of the inlier points:
360 degrees suggests a full circle, 180 degrees indicates a half-circle, or a smaller range may
suggest a partial arc.
}
\examples{
LASfile <- system.file("extdata", "dbh.laz", package="lidR")
las <- readTLS(LASfile, select = "xyz")
xyz = sf::st_coordinates(las)
circle = fit_circle(xyz)
plot(xyz, asp = 1, pch = 19, cex = 0.1)
symbols(circle$center_x, circle$center_y, circles = circle$radius,
  add = TRUE, fg = "red", inches = FALSE)

trunk = las[circle$inliers]

# Fitting a half-circle
half = xyz[xyz[,1] > 101.45,]
circle = fit_circle(half)
plot(half, asp = 1, pch = 19, cex = 0.1)
symbols(circle$center_x, circle$center_y, circles = circle$radius,
  add = TRUE, fg = "red", inches = FALSE)
circle$covered_arc_degree
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algorithm-dsm.R
\name{dsm_spikefree}
\alias{dsm_spikefree}
\alias{spikefree}
\alias{lspikefree}
\title{Digital Surface Model Algorithm}
\usage{
spikefree(freeze_distance, height_buffer = 0.5)

lspikefree(slope = 1.96, intercept = 1.42, height_buffer = 0.5)
}
\arguments{
\item{freeze_distance}{freeze distance (see references). Recommended value: 3 times the pulse spacing or
a little higher.}

\item{height_buffer}{buffer distance (see references). Recommended value: 0.5 do not change.}

\item{slope, intercept}{default values selected according to Fisher F. J. (2024). It is recommended
to do not change.}
}
\description{
The spike-free algorithm developed by Khosravipour et al. (2016), which is based on the computation
of an incremental triangulation of all returns with triangle freezing criteria (see references).

The locally adaptive spikefree algorithm by Fischer et al. 2024 adjusts the freeze distance
automatically and locally depending on point density. Within the typical range of airborne laser
scan densities (>1 points/m2), the original 3 parameter formula can be simplified to a
linear relationship. This linear formula predicts freeze distance from local point spacing
(freeze distance ~ intercept + slope*point_spacing) and can be used to adapt the degree
of spike-removal to sampling density, so that data gaps are filled and canopy height estimates
are consistent across point densities. By default, the point spacing is computed at a resolution
of 5 m. It is recommended to apply this algorithm to first returns only and not at a point
spacing > 0.7 m.
}
\examples{
LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
poi = "-drop_z_below 0 -inside 481280 3812940 481330 3812990"
las <- readLAS(LASfile, filter = poi)
col <- height.colors(50)

# Khosravipour et al. spikefree algorithm
chm <- rasterize_canopy(las, res = 0.25, spikefree(1.5))
plot(chm, col = col)

# Locally adaptive spikefree algorithm
chm <- rasterize_canopy(las, res = 0.25, lspikefree())
plot(chm, col = col)
}
\references{
Khosravipour, Anahita & Skidmore, Andrew & Isenburg, Martin. (2016). Generating spike-free
digital surface models using LiDAR raw point clouds: A new approach for forestry applications.
International Journal of Applied Earth Observation and Geoinformation. 52. 104-114. 10.1016/j.jag.2016.06.005.

Fischer, F. J., Jackson, T., Vincent, G., & Jucker, T. (2024). Robust characterisation
of forest structure from airborne laser scanning—A systematic assessment and sample workflow for
ecologists. Methods in Ecology and Evolution, 15, 1873–1888. https://doi.org/10.1111/2041-210X.14416
}
\seealso{
Other digital surface model algorithms: 
\code{\link{dsm_pitfree}},
\code{\link{dsm_point2raster}},
\code{\link{dsm_tin}}
}
\concept{digital surface model algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filters.R
\name{filters}
\alias{filters}
\alias{filter_poi}
\alias{filter_first}
\alias{filter_firstlast}
\alias{filter_firstofmany}
\alias{filter_ground}
\alias{filter_last}
\alias{filter_nth}
\alias{filter_single}
\alias{filter_duplicates}
\alias{filter_duplicates.LAS}
\alias{filter_duplicates.LAScatalog}
\alias{remove_noise}
\alias{remove_ground}
\alias{remove_water}
\title{Filter points of interest}
\usage{
filter_poi(las, ...)

filter_first(las)

filter_firstlast(las)

filter_firstofmany(las)

filter_ground(las)

filter_last(las)

filter_nth(las, n)

filter_single(las)

filter_duplicates(las)

\method{filter_duplicates}{LAS}(las)

\method{filter_duplicates}{LAScatalog}(las)

remove_noise(las)

remove_ground(las)

remove_water(las)
}
\arguments{
\item{las}{An object of class \link[=LAS-class]{LAS}}

\item{\dots}{Logical predicates. Multiple conditions are combined with '&' or ','}

\item{n}{integer  ReturnNumber == n}
}
\value{
An object of class \link[=LAS-class]{LAS}
}
\description{
Filter points of interest (POI) from a LAS object where conditions are true.
}
\details{
\itemize{
\item{\code{filter_poi} Select points of interest based on custom logical criteria.}
\item{\code{filter_first} Select only the first returns.}
\item{\code{filter_firstlast} Select only the first and last returns.}
\item{\code{filter_ground} Select only the returns classified as ground according to LAS specification.}
\item{\code{filter_last} Select only the last returns i.e. the last returns and the single returns.}
\item{\code{filter_nth} Select the returns from their position in the return sequence.}
\item{\code{filter_firstofmany} Select only the first returns from pulses which returned multiple points.}
\item{\code{filter_single} Select only the returns that return only one point.}
\item{\code{filter_duplicates} \strong{Removes} the duplicated points (duplicated by XYZ)}
\item{\code{remove_noise} Removes the returns classified as noise according to LAS specification.}
\item{\code{remove_ground} Removes the returns classified as ground according to LAS specification.}
\item{\code{remove_water} Removes the returns classified as water according to LAS specification.}
}
}
\section{Non-supported LAScatalog options}{

The option \code{select} is not supported and not respected because it always preserves the file format
and all the attributes. \code{select = "*"} is imposed internally.
}

\examples{
LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
lidar = readLAS(LASfile)

# Select the first returns classified as ground
firstground = filter_poi(lidar, Classification == 2L & ReturnNumber == 1L)

# Multiple arguments are equivalent to &
firstground = filter_poi(lidar, Classification == 2L, ReturnNumber == 1L)

# Multiple criteria
first_or_ground = filter_poi(lidar, Classification == 2L | ReturnNumber == 1L)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algorithm-dec.R
\name{sample_maxima}
\alias{sample_maxima}
\alias{highest}
\alias{lowest}
\title{Point Cloud Decimation Algorithm}
\usage{
highest(res = 1)

lowest(res = 1)
}
\arguments{
\item{res}{numeric. The resolution of the grid used to filter the point cloud}
}
\description{
These functions are made to be used in \link{decimate_points}. They implement algorithms that
create a grid with a given resolution and filters the point cloud by selecting the highest/lowest
point within each cell.
}
\examples{
LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
las = readLAS(LASfile, select = "xyz")

# Select the highest point within each cell of an overlayed grid
thinned = decimate_points(las, highest(4))
#plot(thinned)

# Select the lowest point within each cell of an overlayed grid
thinned = decimate_points(las, lowest(4))
#plot(thinned)
}
\seealso{
Other point cloud decimation algorithms: 
\code{\link{sample_homogenize}},
\code{\link{sample_per_voxel}},
\code{\link{sample_random}}

Other point cloud decimation algorithms: 
\code{\link{sample_homogenize}},
\code{\link{sample_per_voxel}},
\code{\link{sample_random}}
}
\concept{point cloud decimation algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/locate_trees.R
\name{locate_trees}
\alias{locate_trees}
\title{Individual tree detection}
\usage{
locate_trees(las, algorithm, uniqueness = "incremental")
}
\arguments{
\item{las}{An object of class \code{LAS} or \code{LAScatalog}. Can also be a  raster from \code{raster}, \code{stars} or \code{terra}
representing a canopy height model, in which case it is processed like a regularly-spaced point cloud.}

\item{algorithm}{An algorithm for individual tree detection. lidR has: \link{lmf} and \link{manual}.
More experimental algorithms may be found in the package \href{https://github.com/Jean-Romain/lidRplugins}{lidRplugins}.}

\item{uniqueness}{character. A method to compute a unique ID. Can be 'incremental', 'gpstime' or
'bitmerge'. See section 'Uniqueness'. This feature must be considered as 'experimental'.}
}
\value{
\code{locate_trees} returns an sf object with POINT Z geometries. The table of attributes
contains a column \code{treeID} with an individual ID for each tree. The height of the trees (\code{Z}) are
also repeated in the table of attribute to be analysed as an attribute and not as a coordinate.
}
\description{
Individual tree detection function that find the position of the trees using several possible
algorithms.
}
\section{Uniqueness}{


By default the tree IDs are numbered from 1 to n, n being the number of trees found. The problem
with such incremental numbering is that, while it ensures a unique ID is assigned for each tree in
a given point-cloud, it also guarantees duplication of tree IDs in different tiles or chunks when
processing a \code{LAScatalog}. This is because each chunk/file is processed independently of the others
and potentially in parallel on different computers. Thus, the index always restarts at 1 on each
chunk/file. Worse, in a tree segmentation process, a tree that is located exactly between
2 chunks/files will have two different IDs for its two halves.

This is why we introduced some uniqueness strategies that are all imperfect and that should be seen
as experimental. Please report any troubleshooting. Using a uniqueness-safe strategy ensures that
trees from different files will not share the same IDs. It also ensures that two halves of a tree
on the edge of a processing chunk will be assigned the same ID.

\describe{
\item{incremental}{Number from 0 to n. This method \strong{does not} ensure uniqueness of the IDs. This
is the legacy method.}
\item{gpstime}{This method uses the gpstime of the highest point of a tree (apex) to create a
unique ID. This ID is not an integer but a 64-bit decimal number, which is suboptimal but at
least it is expected to be unique \strong{if the gpstime attribute is consistent across files}.
If inconsistencies with gpstime are reported (for example gpstime records the week time and was
reset to 0 in a coverage that takes more than a week to complete), there is a (low) probability of
getting ID attribution errors.}
\item{bitmerge}{This method uses the XY coordinates of the highest point (apex) of a tree to
create a single 64-bit number with a bitwise operation. First, XY coordinates are converted to
32-bit integers using the scales and offsets of the point cloud. For example, if the apex is at
(10.32, 25.64) with a scale factor of 0.01 and an offset of 0, the 32-bit integer coordinates are
X = 1032 and Y = 2564. Their binary representations are, respectively, (here displayed as 16 bits)
0000010000001000 and 0000101000000100. X is shifted by 32 bits and becomes a 64-bit integer. Y is kept
as-is and the binary representations are unionized into a 64-bit integer like (here displayed as 32 bit)
00000100000010000000101000000100 that is guaranteed to be unique. However R
does not support 64-bit integers. The previous steps are done at C++ level and the 64-bit binary
representation is reinterpreted into a 64-bit decimal number to be returned in R. The IDs thus generated
are somewhat weird. For example, the tree ID 00000100000010000000101000000100 which is 67635716 if
interpreted as an integer becomes 3.34164837074751323479078607289E-316 if interpreted as a decimal
number. This is far from optimal but at least it is guaranteed to be unique  \strong{if all files have
the same offsets and scale factors}.}
}

All the proposed options are suboptimal because they either do not guarantee uniqueness in all cases
(inconsistencies in the collection of files), or they imply that IDs are based on non-integers or
meaningless numbers. But at least it works and deals with some of the limitations of R.
}

\examples{
LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
las <- readLAS(LASfile, select = "xyz", filter = "-inside 481250 3812980 481300 3813030")

ttops <- locate_trees(las, lmf(ws = 5))

#plot(las) |> add_treetops3d(ttops)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/segment.R, R/segment_shapes.R,
%   R/segment_snags.R, R/segment_trees.R
\name{segment}
\alias{segment}
\alias{segment_shapes}
\alias{segment_snags}
\alias{segment_trees}
\title{Segment a point cloud}
\usage{
segment_shapes(las, algorithm, attribute = "Shape", filter = NULL)

segment_snags(las, algorithm, attribute = "snagCls")

segment_trees(las, algorithm, attribute = "treeID", uniqueness = "incremental")
}
\arguments{
\item{las}{An object of class \link[lidR:LAS-class]{LAS} or \link[lidR:LAScatalog-class]{LAScatalog}.}

\item{algorithm}{function. An algorithm for segmentation. For individual tree segmentation, lidR
has \link{dalponte2016}, \link{watershed}, \link{li2012}, and \link{silva2016}. More experimental
algorithms may be found in the package \href{https://github.com/Jean-Romain/lidRplugins}{lidRplugins}.
For snag segmentation, \code{lidR} has \link{wing2015}. For geometry segmentation, lidR has
\link{shp_plane}, \link{shp_hplane}, and \link{shp_line}.}

\item{attribute}{character. The returned LAS object as a new attribute (in a new column).
This parameter controls the name of the new attribute.}

\item{filter}{formula of logical predicates. Enables the function to run only on points of interest
in an optimized way. See the examples.}

\item{uniqueness}{character. A method to compute a unique ID. Can be 'incremental', 'gpstime' or
'bitmerge'. See section 'Uniqueness'. This feature must be considered as 'experimental'.}
}
\description{
Segment a point cloud using different methods. \verb{segment_*} functions add a new attribute to the
point cloud to label each point. They segment either individual trees, snags, or
geometrical features.
}
\details{
\describe{
\item{\code{segment_trees}}{Individual tree segmentation with several possible algorithms. The returned
point cloud has a new extra byte attribute named after the parameter \code{attribute} independently
of the algorithm used.}
\item{\code{segment_shapes}}{Computes, for each point, the eigenvalues of the covariance matrix of the
neighbouring points. The eigenvalues are later used either to segment linear/planar points or to
compute derived metrics. The points that meet a given criterion based on the eigenvalue are labelled
as approximately coplanar/colinear or any other shape supported.}
\item{\code{segment_snags}}{Snag segmentation using several possible algorithms. The function attributes
a number identifying a snag class (\code{snagCls} attribute) to each point of the point cloud.
The classification/segmentation is done at the point cloud level and currently only one algorithm is
implemented, which uses LiDAR intensity thresholds and specified neighbourhoods to differentiate
bole and branch from foliage points.}
}
}
\section{Non-supported LAScatalog options}{

The option \code{select} is not supported and not respected because it always preserves the file format
and all the attributes. \code{select = "*"} is imposed internally.
}

\section{Uniqueness}{


By default the tree IDs are numbered from 1 to n, n being the number of trees found. The problem
with such incremental numbering is that, while it ensures a unique ID is assigned for each tree in
a given point-cloud, it also guarantees duplication of tree IDs in different tiles or chunks when
processing a \code{LAScatalog}. This is because each chunk/file is processed independently of the others
and potentially in parallel on different computers. Thus, the index always restarts at 1 on each
chunk/file. Worse, in a tree segmentation process, a tree that is located exactly between
2 chunks/files will have two different IDs for its two halves.

This is why we introduced some uniqueness strategies that are all imperfect and that should be seen
as experimental. Please report any troubleshooting. Using a uniqueness-safe strategy ensures that
trees from different files will not share the same IDs. It also ensures that two halves of a tree
on the edge of a processing chunk will be assigned the same ID.

\describe{
\item{incremental}{Number from 0 to n. This method \strong{does not} ensure uniqueness of the IDs. This
is the legacy method.}
\item{gpstime}{This method uses the gpstime of the highest point of a tree (apex) to create a
unique ID. This ID is not an integer but a 64-bit decimal number, which is suboptimal but at
least it is expected to be unique \strong{if the gpstime attribute is consistent across files}.
If inconsistencies with gpstime are reported (for example gpstime records the week time and was
reset to 0 in a coverage that takes more than a week to complete), there is a (low) probability of
getting ID attribution errors.}
\item{bitmerge}{This method uses the XY coordinates of the highest point (apex) of a tree to
create a single 64-bit number with a bitwise operation. First, XY coordinates are converted to
32-bit integers using the scales and offsets of the point cloud. For example, if the apex is at
(10.32, 25.64) with a scale factor of 0.01 and an offset of 0, the 32-bit integer coordinates are
X = 1032 and Y = 2564. Their binary representations are, respectively, (here displayed as 16 bits)
0000010000001000 and 0000101000000100. X is shifted by 32 bits and becomes a 64-bit integer. Y is kept
as-is and the binary representations are unionized into a 64-bit integer like (here displayed as 32 bit)
00000100000010000000101000000100 that is guaranteed to be unique. However R
does not support 64-bit integers. The previous steps are done at C++ level and the 64-bit binary
representation is reinterpreted into a 64-bit decimal number to be returned in R. The IDs thus generated
are somewhat weird. For example, the tree ID 00000100000010000000101000000100 which is 67635716 if
interpreted as an integer becomes 3.34164837074751323479078607289E-316 if interpreted as a decimal
number. This is far from optimal but at least it is guaranteed to be unique  \strong{if all files have
the same offsets and scale factors}.}
}

All the proposed options are suboptimal because they either do not guarantee uniqueness in all cases
(inconsistencies in the collection of files), or they imply that IDs are based on non-integers or
meaningless numbers. But at least it works and deals with some of the limitations of R.
}

\examples{
# ==============
# Segment trees
# ==============

LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
las <- readLAS(LASfile, select = "xyz", filter = "-drop_z_below 0")

# Using Li et al. (2012)
las <- segment_trees(las, li2012(R = 3, speed_up = 5))
#plot(las, color = "treeID")

# ==============
# Segment shapes
# ==============

LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
las <- readLAS(LASfile, filter = "-keep_random_fraction 0.5")

# Use the eigenvalues to estimate if points are part of a local plan
las <- segment_shapes(las, shp_plane(k = 15), "Coplanar")
#plot(las, color = "Coplanar")

\dontrun{
# Drop ground point at runtime
las <- segment_shapes(las, shp_plane(k = 15), "Coplanar", filter = ~Classification != 2L)
#plot(las, color = "Coplanar")

# ==============
# Segment snags
# ==============

LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
las <- readLAS(LASfile, select = "xyzi", filter="-keep_first") # Wing also included -keep_single

# For the Wing2015 method, supply a matrix of snag BranchBolePtRatio conditional
# assessment thresholds (see Wing et al. 2015, Table 2, pg. 172)
bbpr_thresholds <- matrix(c(0.80, 0.80, 0.70,
                            0.85, 0.85, 0.60,
                            0.80, 0.80, 0.60,
                            0.90, 0.90, 0.55),
                            nrow =3, ncol = 4)

# Run snag classification and assign classes to each point
las <- segment_snags(las, wing2015(neigh_radii = c(1.5, 1, 2), BBPRthrsh_mat = bbpr_thresholds))

# Plot it all, tree and snag points...
plot(las, color="snagCls", colorPalette = rainbow(5))

# Filter and plot snag points only
snags <- filter_poi(las, snagCls > 0)
plot(snags, color="snagCls", colorPalette = rainbow(5)[-1])

# Wing et al's (2015) methods ended with performing tree segmentation on the
# classified and filtered point cloud using the watershed method
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algorithm-its.R
\name{its_dalponte2016}
\alias{its_dalponte2016}
\alias{dalponte2016}
\title{Individual Tree Segmentation Algorithm}
\usage{
dalponte2016(
  chm,
  treetops,
  th_tree = 2,
  th_seed = 0.45,
  th_cr = 0.55,
  max_cr = 10,
  ID = "treeID"
)
}
\arguments{
\item{chm}{`RasterLayer`, `SpatRaster` or `stars`. Canopy height model. Can be computed with \link{rasterize_canopy} or read from
an external file.}

\item{treetops}{`SpatialPoints*` or `sf/sfc_POINT` with 2D or 3D coordinates. Can be computed with
\link{locate_trees} or read from an external file}

\item{th_tree}{numeric. Threshold below which a pixel cannot be a tree. Default is 2.}

\item{th_seed}{numeric. Growing threshold 1. See reference in Dalponte et al. 2016. A pixel
is added to a region if its height is greater than the tree height multiplied by this value.
It should be between 0 and 1. Default is 0.45.}

\item{th_cr}{numeric. Growing threshold 2. See reference in Dalponte et al. 2016. A pixel
is added to a region if its height is greater than the current mean height of the region
multiplied by this value. It should be between 0 and 1. Default is 0.55.}

\item{max_cr}{numeric. Maximum value of the crown diameter of a detected tree (in pixels).
Default is 10.}

\item{ID}{character. If \code{treetops} contains an attribute with the ID for
each tree, the name of this attribute. This way, original IDs will be preserved.}
}
\description{
This function is made to be used in \link{segment_trees}. It implements an algorithm for tree
segmentation based on Dalponte and Coomes (2016) algorithm (see reference).
This is a seeds + growing region algorithm. This algorithm exists in the package \code{itcSegment}.
This version has been written from the paper in C++. Consequently it is hundreds to millions times
faster than the original version. Note that this algorithm strictly performs a segmentation, while the
original method as implemented in \code{itcSegment} and described in the manuscript also performs
pre- and post-processing tasks. Here these tasks are expected to be done by the user in separate functions.
}
\details{
Because this algorithm works on a CHM only there is no actual need for a point cloud. Sometimes the
user does not even have the point cloud that generated the CHM. \code{lidR} is a point cloud-oriented
library, which is why this algorithm must be used in \link{segment_trees} to merge the result with the point
cloud. However the user can use this as a stand-alone function like this:
\preformatted{
 chm <- raster("chm.tif")
 ttops <- locate_trees(chm, lmf(3))
 crowns <- dalponte2016(chm, ttops)()
}
}
\examples{
LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
poi <- "-drop_z_below 0 -inside 481280 3812940 481320 3812980"
las <- readLAS(LASfile, select = "xyz", filter = poi)
col <- pastel.colors(200)

chm <- rasterize_canopy(las, 0.5, p2r(0.3))
ker <- matrix(1,3,3)
chm <- terra::focal(chm, w = ker, fun = mean, na.rm = TRUE)

ttops <- locate_trees(chm, lmf(4, 2))
las   <- segment_trees(las, dalponte2016(chm, ttops))
#plot(las, color = "treeID", colorPalette = col)
}
\references{
Dalponte, M. and Coomes, D. A. (2016), Tree-centric mapping of forest carbon density from
airborne laser scanning and hyperspectral data. Methods Ecol Evol, 7: 1236–1245. doi:10.1111/2041-210X.12575.
}
\seealso{
Other individual tree segmentation algorithms: 
\code{\link{its_li2012}},
\code{\link{its_silva2016}},
\code{\link{its_watershed}}

Other raster based tree segmentation algorithms: 
\code{\link{its_silva2016}},
\code{\link{its_watershed}}
}
\concept{individual tree segmentation algorithms}
\concept{raster based tree segmentation algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/catalog_boundaries.R
\name{catalog_split_clusters}
\alias{catalog_split_clusters}
\title{Split a catalog into spatially clustered subsets}
\usage{
catalog_split_clusters(ctg, buffer = 1)
}
\arguments{
\item{ctg}{A LAScatalog}

\item{buffer}{Numeric. Buffer distance applied to tile geometry before clustering. Defaults to 1.}
}
\value{
A list of LAScatalog
}
\description{
This function groups elements of a catalog into spatial clusters based on
contiguity of the tiles. The original catalog is then split into `n` catalogs
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algorithm-its.R
\name{its_li2012}
\alias{its_li2012}
\alias{li2012}
\title{Individual Tree Segmentation Algorithm}
\usage{
li2012(dt1 = 1.5, dt2 = 2, R = 2, Zu = 15, hmin = 2, speed_up = 10)
}
\arguments{
\item{dt1}{numeric. Threshold number 1. See reference page 79 in Li et al. (2012). Default is 1.5.}

\item{dt2}{numeric. Threshold number 2. See reference page 79 in Li et al. (2012). Default is 2.}

\item{R}{numeric. Search radius. See page 79 in Li et al. (2012). Default is 2. If \code{R = 0}
all the points are automatically considered as local maxima and the search step is skipped (much
faster).}

\item{Zu}{numeric. If point elevation is greater than Zu, \code{dt2} is used, otherwise \code{dt1} is
used. See page 79 in Li et al. (2012). Default is 15.}

\item{hmin}{numeric. Minimum height of a detected tree. Default is 2.}

\item{speed_up}{numeric. Maximum radius of a crown. Any value greater than a crown is
good because this parameter does not affect the result. However, it greatly affects the
computation speed by restricting the number of comparisons to perform.
The lower the value, the faster the method. Default is 10.}
}
\description{
This functions is made to be used in \link{segment_trees}. It implements an algorithm for tree
segmentation based on Li et al. (2012) (see reference). This method is a growing region
method working at the point cloud level. It is an implementation by lidR authors, from the original
paper, as close as possible from the original description. However we added a parameter \code{hmin}
to prevent over-segmentation for objects that are too low. This algorithm is known to be slow because
it has an algorithmic complexity worst that O(n^2).
}
\examples{
LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
poi <- "-drop_z_below 0 -inside 481280 3812940 481320 3812980"
las <- readLAS(LASfile, select = "xyz", filter = poi)
col <- pastel.colors(200)

las <- segment_trees(las, li2012(dt1 = 1.4))
#plot(las, color = "treeID", colorPalette = col)
}
\references{
Li, W., Guo, Q., Jakubowski, M. K., & Kelly, M. (2012). A new method for segmenting individual
trees from the lidar point cloud. Photogrammetric Engineering & Remote Sensing, 78(1), 75-84.
}
\seealso{
Other individual tree segmentation algorithms: 
\code{\link{its_dalponte2016}},
\code{\link{its_silva2016}},
\code{\link{its_watershed}}
}
\concept{individual tree segmentation algorithms}
\concept{point-cloud based tree segmentation algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algorithm-dec.R
\name{sample_per_voxel}
\alias{sample_per_voxel}
\alias{random_per_voxel}
\alias{barycenter_per_voxel}
\alias{lowest_attribute_per_voxel}
\alias{highest_attribute_per_voxel}
\title{Point Cloud Decimation Algorithm}
\usage{
random_per_voxel(res = 1, n = 1)

barycenter_per_voxel(res = 1)

lowest_attribute_per_voxel(res, attribute = "Z")

highest_attribute_per_voxel(res, attribute = "Z")
}
\arguments{
\item{res}{numeric. The resolution of the voxel grid used to filter the point cloud}

\item{n}{integer. The number of points to select}

\item{attribute}{string name of an attribute (such as 'intensity')}
}
\description{
These functions are made to be used in \link{decimate_points}. They implements algorithm that
creates a 3D grid with a given resolution and filters the point cloud by selecting
points of interest within each voxel. `random_per_voxel()` sample random points. `barycenter_per_voxel()`
samples the point that is the closest to the barycenter of the points within a given voxel.
`[lowest|highest]_attribute_per_voxel()` sample respectively the point that have the highest/lowest
attribute (e.g. Intensity) per voxel.
}
\examples{
LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
las <- readLAS(LASfile, select = "xyz")
thinned <- decimate_points(las, random_per_voxel(8, 1))
#plot(thinned)
}
\seealso{
Other point cloud decimation algorithms: 
\code{\link{sample_homogenize}},
\code{\link{sample_maxima}},
\code{\link{sample_random}}
}
\concept{point cloud decimation algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_chm.R
\name{pitfill_stonge2008}
\alias{pitfill_stonge2008}
\title{Pits and spikes filling}
\usage{
pitfill_stonge2008(
  x,
  lap_size = 3L,
  thr_lap = 0.1,
  thr_spk = -0.1,
  med_size = 3L,
  dil_radius = 0L
)
}
\arguments{
\item{x}{raster. SpatRaster, RasterLayer, stars.}

\item{lap_size}{integer. Size of the Laplacian filter kernel (integer value, in pixels).}

\item{thr_lap}{numeric. Threshold Laplacian value for detecting a cavity (all values above this
value will be considered a cavity). A positive value.}

\item{thr_spk}{numeric. Threshold Laplacian value for detecting a spike (all values below this
value will be considered a spike). A negative value.}

\item{med_size}{integer. Size of the median filter kernel (integer value, in pixels).}

\item{dil_radius}{integer. Dilation radius (integer value, in pixels).}
}
\description{
Pits and spikes filling for raster. Typically used for post-processing CHM. This algorithm
is from St-Onge 2008 (see reference).
}
\examples{
LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
las <- readLAS(LASfile)
chm <- rasterize_canopy(las, 0.5, dsmtin())
sto <- pitfill_stonge2008(chm)

#terra::plot(c(chm, sto), col = lidR::height.colors(25))
}
\references{
St-Onge, B., 2008. Methods for improving the quality of a true orthomosaic of Vexcel UltraCam
images created using alidar digital surface model, Proceedings of the Silvilaser 2008, Edinburgh,
555-562. https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=81365288221f3ac34b51a82e2cfed8d58defb10e
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_hull.R
\name{st_hull}
\alias{st_hull}
\alias{st_concave_hull}
\alias{st_convex_hull.LAS}
\alias{concaveman}
\title{Concave and convex hulls for LAS objects}
\usage{
st_concave_hull(x, method = "concaveman", ...)

\method{st_convex_hull}{LAS}(x)

concaveman(x, y = NULL, concavity = 2, length_threshold = 0)
}
\arguments{
\item{x, y}{An object of class LAS or XY coordinates of points in case of \code{concaveman}. This can be
specified as two vectors x and y, a 2-column matrix x, a list with two components, etc.}

\item{method}{string. currently supports "concaveman".}

\item{...}{Propagate to the method.}

\item{concavity}{numeric a relative measure of concavity. 1 results in a relatively detailed shape,
Infinity results in a convex hull. You can use values lower than 1, but they can produce pretty crazy
shapes.}

\item{length_threshold}{numeric. When a segment length is below this threshold, it stops being
considered for further detailed processing. Higher values result in simpler shapes.}
}
\value{
A \code{sfc_POLYGON} from \code{sf} or a \code{data.frame} in the case of \code{concaveman}
}
\description{
Concave and convex hulls for LAS objects. \code{st_convex_hull} extends \code{sf::st_convex_hull} for LAS
objects. Both functions return a \code{sfc_POLYGON}. \code{concaveman} is very a fast 2D concave hull algorithm
for a set of points.
}
\details{
The concaveman algorithm is based on ideas from Park and Oh (2012). A first implementation in
JavaScript was proposed by Vladimir Agafonkin in \href{https://github.com/mapbox/concaveman}{mapbox}.
This implementation dramatically improved performance over the one stated in the paper
using a spatial index. The algorithm was then ported to R by Joël Gombin in the R package
\href{https://github.com/joelgombin/concaveman}{concaveman} that runs the JavaScript
implementation proposed by Vladimir Agafonkin. Later, a C++ version of Vladimir Agafonkin's
JavaScript implementation was proposed by Stanislaw Adaszewski in
\href{https://github.com/sadaszewski/concaveman-cpp}{concaveman-cpp}. This concaveman
function uses Stanislaw Adaszewski's C++ code making the concaveman algorithm an
order of magnitude (up to 50 times) faster than the Javascript version.
}
\examples{
x <- runif(35)
y <- runif(35)
hull <- concaveman(x,y)
plot(x,y, asp = 1)
lines(hull, lwd = 3, col = "red")

LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
las = readLAS(LASfile, filter = "-drop_z_below 1")
hull = st_concave_hull(las, length_threshold = 10)
plot(hull)
}
\references{
Park, J.-S & Oh, S.-J. (2013). A New Concave Hull Algorithm and Concaveness Measure
for n-dimensional Datasets. Journal of Information Science and Engineering. 29. 379-392.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Class-LAS.R, R/methods-LAS.R
\docType{class}
\name{LAS-class}
\alias{LAS-class}
\alias{LAS}
\title{An S4 class to represent a .las or .laz file}
\usage{
LAS(data, header = list(), crs = sf::NA_crs_, check = TRUE, index = NULL, ...)
}
\arguments{
\item{data}{a \link[data.table:data.table]{data.table} containing the data of a las or laz file.}

\item{header}{a \code{list} or a \link[=LASheader-class]{LASheader} containing the header of
a las or laz file.}

\item{crs}{crs object of class \link[sf:st_crs]{crs} from sf}

\item{check}{logical. Conformity tests while building the object.}

\item{index}{list with two elements \code{list(sensor = 0L, index = 0L)}.
See \link[=lidR-spatial-index]{spatial indexing}}

\item{...}{internal use}
}
\value{
An object of class \code{LAS}
}
\description{
Class LAS is the representation of a las/laz file according to the
\href{https://community.asprs.org/leadership-restricted/leadership-content/public-documents/standards}{LAS file format specifications}.
}
\details{
A \code{LAS} object contains a \code{data.table} with the data read from a \code{las/laz} file and
a \link[=LASheader-class]{LASheader} (see the ASPRS documentation for the LAS file format for more information).
Because las files are standardized the table of attributes read from the las/laz file is also standardized.
Columns are named:
\itemize{
\item{\code{X}, \code{Y}, \code{Z} (numeric)}
\item{\code{gpstime} (numeric)}
\item{\code{Intensity} (integer)}
\item{\code{ReturnNumber}, \code{NumberOfReturns} (integer)}
\item{\code{ScanDirectionFlag} (integer)}
\item{\code{EdgeOfFlightline} (integer)}
\item{\code{Classification} (integer)}
\item{\code{Synthetic_flag},\code{Keypoint_flag}, \code{Withheld_flag}  (logical)}
\item{\code{ScanAngleRank}/\code{ScanAngle}  (integer/numeric)}
\item{\code{UserData} (integer)}
\item{\code{PointSourceID} (integer)}
\item{\code{R},\code{G},\code{B}, \code{NIR} (integer)}
}
}
\section{Functions}{
\itemize{
\item \code{LAS()}: creates objects of class LAS. The original data is updated by reference to
quantize the coordinates according to the scale factor of the header if no header is provided.
In this case the scale factor is set to 0.001

}}
\section{Slots}{

\describe{
\item{\code{crs}}{Object of class \link[sf:st_crs]{crs} from sf.}

\item{\code{data}}{Object of class \link[data.table:data.table]{data.table}. Point cloud data according to the LAS file format.}

\item{\code{header}}{Object of class \link[=LASheader-class]{LASheader}. LAS file header according to the LAS file format.}

\item{\code{index}}{list. See \link[=lidR-spatial-index]{spatial indexing}.}
}}

\examples{
# Read a las/laz file
LASfile <- system.file("extdata", "example.laz", package="rlas")
las <- readLAS(LASfile)
las

# Creation of a LAS object out of external data
data <- data.frame(X = runif(100, 0, 100),
                   Y = runif(100, 0, 100),
                   Z = runif(100, 0, 20))

# 'data' has many decimal digits
data

# Create a default header and quantize *by reference*
# the coordinates to fit with offset and scale factors
cloud <- LAS(data)

# 'data' has been updated and coordinates were quantized
data
cloud

# Be careful when providing a header the function assumes that
# it corresponds to the data and won't quantize the coordinates
data <- data.frame(X = runif(100, 0, 100),
                   Y = runif(100, 0, 100),
                   Z = runif(100, 0, 20))
header <- header(las)

# This works but triggers warnings and creates an invalid LAS object
cloud <- LAS(data, header)

las_check(cloud)
}
\seealso{
\link{readLAS}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/normalize.R, R/normalize_height.R,
%   R/normalize_intensity.R
\name{normalize}
\alias{normalize}
\alias{normalize_height}
\alias{unnormalize_height}
\alias{-,LAS,ANY-method}
\alias{height_above_ground}
\alias{normalize_intensity}
\title{Normalize point cloud}
\usage{
normalize_height(las, algorithm, use_class = c(2L, 9L), dtm = NULL, ...)

unnormalize_height(las)

\S4method{-}{LAS,ANY}(e1, e2)

height_above_ground(las, ...)

normalize_intensity(las, algorithm)
}
\arguments{
\item{las}{An object of class \link[lidR:LAS-class]{LAS} or \link[lidR:LAScatalog-class]{LAScatalog}.}

\item{algorithm}{(1) An algorithm for spatial interpolation. \code{lidR} has \link{tin},
\link{kriging}, \link{knnidw} or a raster representing a digital terrain
model. (2) An algorithm for intensity normalization. \code{lidR} currently has \link{range_correction}.}

\item{use_class}{integer vector. By default the terrain is computed by using ground points
(class 2) and water points (class 9). Relevant only for a normalization without a raster DTM.}

\item{dtm}{raster. If \code{dtm} is provided, then the DTM is used in place of ground points. This is
different than providing a DTM in \code{algorithm}. If \code{algorithm = dtm} the dtm is subtracted naively.
If \code{algorithm = tin()} and \code{dtm = raster} the ground points are not used and the DTM is
interpolated as if it were made of regularly-spaced ground points.}

\item{...}{passed to \code{normalized_height()}}

\item{e1}{a LAS object}

\item{e2}{A raster representing a digital terrain model in format from \code{raster}, \code{stars} or \code{terra}..}
}
\description{
Normalize elevation or intensity values using multiple methods.
}
\details{
\describe{
\item{normalize_height}{Subtract digital terrain model (DTM) from a LiDAR point cloud to create a
dataset normalized with the ground at 0. The DTM can be a raster, but it can also be computed
on-the-fly. In this case the algorithm does not use rasterized data and each point is interpolated.
There is no inaccuracy due to the discretization of the terrain and the resolution of the terrain
is virtually infinite. A new attribute 'Zref' records the former elevation values, which enables
the use of \link{unnormalize_height} to restore original point elevations.}
\item{normalize_intensity}{Normalize intensity values using multiple methods. The attribute 'Intensity'
records the normalized intensity. An extra attribute named 'RawIntensity' records the original
intensities.}
\item{height_above_ground}{instead of normalizing the point cloud, records the height above ground (HAG)
in a new attribute named 'hag'}
}
}
\section{Non-supported LAScatalog options}{

The option \code{select} is not supported and not respected because it always preserves the file format
and all the attributes. \code{select = "*"} is imposed internally.
}

\examples{
LASfile <- system.file("extdata", "Topography.laz", package="lidR")
las <- readLAS(LASfile)

# ====================
# Normalize elevation
# ====================

# First option: use a raster as DTM
# --------------------------------------

dtm <- rasterize_terrain(las, 1, knnidw(k = 6L, p = 2))
nlas <- normalize_height(las, dtm)

# restore original elevations
las <- unnormalize_height(nlas)

# operator - can be used. This is equivalent to the previous
nlas <- las - dtm

# restore original elevations
las <- unnormalize_height(las)

# Second option: interpolate each point (no discretization)
# ---------------------------------------------------------

nlas <- normalize_height(las, tin())

# operator - can be used. This is equivalent to the previous
las <- unnormalize_height(nlas)
nlas <- las - tin()

\dontrun{
# All the following syntaxes are correct
las <- normalize_height(las, knnidw())
las <- normalize_height(las, knnidw(k = 8, p = 2))
las <- las - knnidw()
las <- las - knnidw(k = 8)
las <- normalize_height(las, kriging())
las <- las - kriging(k = 8)
}

# ====================
# Normalize intensity
# ====================

# pmin = 15 because it is an extremely small file
# strongly decimated to reduce its size. There are
# actually few multiple returns
sensor <- track_sensor(las, Roussel2020(pmin = 15))

# Here the effect is virtually null because the size of
# the sample is too small to notice any effect of range
las <- normalize_intensity(las, range_correction(sensor, Rs = 2000))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algorithm-dtm.R
\name{dtm_kriging}
\alias{dtm_kriging}
\alias{kriging}
\title{Spatial Interpolation Algorithm}
\usage{
kriging(model = gstat::vgm(0.59, "Sph", 874), k = 10L)
}
\arguments{
\item{model}{A variogram model computed with \code{vgm()} from package \code{gstat}. If NULL it performs an ordinary
or weighted least squares prediction.}

\item{k}{numeric. Number of k-nearest neighbours. Default 10.}
}
\description{
This function is made to be used in \link{rasterize_terrain} or \link{normalize_height}. It
implements an algorithm for spatial interpolation. Spatial interpolation is based on universal
kriging using the \code{krige()} function from \code{gstat}. This method combines the
KNN approach with the kriging approach. For each point of interest it kriges the terrain using
the k-nearest neighbour ground points. This method is more difficult to manipulate but it is also
the most advanced method for interpolating spatial data.
}
\examples{
\dontrun{
LASfile <- system.file("extdata", "Topography.laz", package="lidR")
las = readLAS(LASfile)

plot(las)

dtm = rasterize_terrain(las, algorithm = kriging())

plot(dtm)
plot_dtm3d(dtm)
}
}
\seealso{
Other dtm algorithms: 
\code{\link{dtm_idw}},
\code{\link{dtm_tin}}
}
\concept{dtm algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algorithm-dec.R
\name{sample_homogenize}
\alias{sample_homogenize}
\alias{homogenize}
\title{Point Cloud Decimation Algorithm}
\usage{
homogenize(density, res = 5, use_pulse = FALSE)
}
\arguments{
\item{density}{numeric. The desired output density.}

\item{res}{numeric. The resolution of the grid used to filter the point cloud}

\item{use_pulse}{logical. Decimate by removing random pulses instead of random points (requires running
\link{retrieve_pulses} first)}
}
\description{
This function is made to be used in \link{decimate_points}. It implements an algorithm that
creates a grid with a given resolution and filters the point cloud by randomly selecting some
points in each cell. It is designed to produce point clouds that have uniform densities throughout
the coverage area. For each cell, the proportion of points or pulses that will be retained is computed
using the actual local density and the desired density. If the desired density is greater than the actual
density it returns an unchanged set of points (it cannot increase the density). The cell size must be
large enough to compute a coherent local density. For example, in a 2 points/m^2 point cloud, 25 square
meters would be feasible; however 1 square meter cells would not be feasible because density does
not have meaning at this scale.
}
\examples{
LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
las = readLAS(LASfile, select = "xyz")

# Select points randomly to reach an homogeneous density of 1
thinned <- decimate_points(las, homogenize(1,5))
plot(rasterize_density(thinned, 10))
}
\seealso{
Other point cloud decimation algorithms: 
\code{\link{sample_maxima}},
\code{\link{sample_per_voxel}},
\code{\link{sample_random}}
}
\concept{point cloud decimation algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io_writeLAS.R
\name{writeLAS}
\alias{writeLAS}
\title{Write a .las or .laz file}
\usage{
writeLAS(las, file, index = FALSE)
}
\arguments{
\item{las}{an object of class LAS.}

\item{file}{character. A character string naming an output file.}

\item{index}{boolean. Also write a lax file to index the points in the files}
}
\value{
Nothing. This function is used for its side-effect of writing a file.
}
\description{
Write a \link[=LAS-class]{LAS} object into a binary .las or .laz file (compression
specified in filename)
}
\examples{
LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
las = readLAS(LASfile)
subset = clip_rectangle(las, 684850, 5017850, 684900, 5017900)
writeLAS(subset, tempfile(fileext = ".laz"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algorithm-out.R
\name{noise_ivf}
\alias{noise_ivf}
\alias{ivf}
\title{Noise Segmentation Algorithm}
\usage{
ivf(res = 5, n = 6)
}
\arguments{
\item{res}{numeric. Resolution of the voxels}

\item{n}{integer. The maximal number of 'other points' in the 27 voxels}
}
\description{
This function is made to be used in \link{classify_noise}. It implements an
algorithm for outliers (noise) segmentation based on isolated voxels filter (IVF).
The algorithm finds points that have only a few other points in their surrounding
3 x 3 x 3 = 27 voxels.
}
\examples{
LASfile <- system.file("extdata", "Topography.laz", package="lidR")
las <- readLAS(LASfile, filter = "-inside 273450 5274350 273550 5274450")

# Add some artificial outliers
set.seed(314)
id = round(runif(20, 0, npoints(las)))
set.seed(42)
err = runif(20, -50, 50)
las$Z[id] = las$Z[id] + err

las <- classify_noise(las, ivf(5,2))
}
\seealso{
Other noise segmentation algorithms: 
\code{\link{noise_sor}}
}
\concept{noise segmentation algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classify.R, R/classify_ground.R,
%   R/classify_noise.R, R/classify_poi.R
\name{classify}
\alias{classify}
\alias{classify_ground}
\alias{classify_noise}
\alias{classify_poi}
\title{Classify points}
\usage{
classify_ground(las, algorithm, last_returns = TRUE)

classify_noise(las, algorithm)

classify_poi(
  las,
  class,
  poi = NULL,
  roi = NULL,
  inverse_roi = FALSE,
  by_reference = FALSE
)
}
\arguments{
\item{las}{An object of class \link[lidR:LAS-class]{LAS} or \link[lidR:LAScatalog-class]{LAScatalog}.}

\item{algorithm}{An algorithm for classification. lidR has has: \link{sor}, \link{ivf} for noise
classification, and \link{pmf}, \link{csf}, \link{mcc} for ground classification (see respective
documentation).}

\item{last_returns}{logical. The algorithm will use only the last returns (including the first returns
in cases of a single return) to run the algorithm. If FALSE all the returns are used. If the attributes
\code{'ReturnNumber'} or \code{'NumberOfReturns'} are absent, \code{'last_returns'} is turned
to \code{FALSE} automatically.}

\item{class}{The ASPRS class to attribute to the points that meet the criterion.}

\item{poi}{a formula of logical predicates. The points that are \code{TRUE} will be classified \code{class}.}

\item{roi}{A \verb{SpatialPolygons*}, from \code{sp} or a \code{sf/sfc_POLYGON} from \code{sf}.
The points that are in the region of interest delimited by the polygon(s) are classified
\code{class}.}

\item{inverse_roi}{bool. Inverses the \code{roi}. The points that are outside the polygon(s)
are classified \code{class}.}

\item{by_reference}{bool. Updates the classification in place (LAS only).}
}
\description{
Classify points that meet some criterion and/or that belong in a region of interest. The
functions updates the attribute \code{Classification} of the LAS object according to
\href{https://community.asprs.org/leadership-restricted/leadership-content/public-documents/standards}{las specifications}
}
\details{
\describe{
\item{classify_noise}{Classify points as 'noise' (outliers) with several possible algorithms.
lidR has: \link{sor}, \link{ivf}. The points classified as 'noise' are assigned a value of 18.}
\item{classify_ground}{Classify points as 'ground' with several possible algorithms.
lidR has \link{ptd}, \link{pmf}, \link{csf} and \link{mcc}. The points classified as 'ground'
are assigned a value of 2. Using \code{ptd} is strongly recommended.}
\item{classify_poi}{Classify points that meet some logical criterion and/or that belong in a
region of interest with class of choice.}
}
}
\section{Non-supported LAScatalog options}{

The option \code{select} is not supported and not respected because it always preserves the file format
and all the attributes. \code{select = "*"} is imposed internally.
}

\examples{
# ===============
# Classify ground
# ===============

if (require(RCSF, quietly = TRUE))
{
LASfile <- system.file("extdata", "Topography.laz", package="lidR")
las <- readLAS(LASfile, select = "xyzrn", filter = "-inside 273450 5274350 273550 5274450")

# (Parameters chosen mainly for speed)
mycsf <- csf(TRUE, 1, 1, time_step = 1)
las <- classify_ground(las, mycsf)
#plot(las, color = "Classification")
}

# ===============
# Classify noise
# ===============

LASfile <- system.file("extdata", "Topography.laz", package="lidR")
las <- readLAS(LASfile, filter = "-inside 273450 5274350 273550 5274450")

# Add 20 artificial outliers
set.seed(314)
id = round(runif(20, 0, npoints(las)))
set.seed(42)
err = runif(20, -50, 50)
las$Z[id] = las$Z[id] + err

# Using IVF
las <- classify_noise(las, ivf(5,2))
#plot(las, color = "Classification")

# Remove outliers using filter_poi()
las_denoise <- filter_poi(las, Classification != LASNOISE)

# ===============
# Classify POI
# ===============

LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
shp <- system.file("extdata", "lake_polygons_UTM17.shp", package = "lidR")

las  <- readLAS(LASfile, filter = "-keep_random_fraction 0.1")
lake <- sf::st_read(shp, quiet = TRUE)

# Classifies the points that are NOT in the lake and that are NOT ground points as class 5
poi <- ~Classification != LASGROUND
las <- classify_poi(las, LASHIGHVEGETATION, poi = poi, roi = lake, inverse = TRUE)

# Classifies the points that are in the lake as class 9
las <- classify_poi(las, LASWATER, roi = lake, inverse = FALSE)

#plot(las, color = "Classification")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics_stdmetrics.R
\docType{data}
\name{stdmetrics}
\alias{stdmetrics}
\alias{stdmetrics_z}
\alias{stdmetrics_i}
\alias{stdmetrics_rn}
\alias{stdmetrics_pulse}
\alias{stdmetrics_ctrl}
\alias{stdtreemetrics}
\alias{stdshapemetrics}
\alias{.stdmetrics}
\alias{.stdmetrics_z}
\alias{.stdmetrics_i}
\alias{.stdmetrics_rn}
\alias{.stdmetrics_pulse}
\alias{.stdmetrics_ctrl}
\alias{.stdtreemetrics}
\alias{.stdshapemetrics}
\title{Predefined standard metrics functions}
\format{
An object of class \code{formula} of length 2.

An object of class \code{formula} of length 2.

An object of class \code{formula} of length 2.

An object of class \code{formula} of length 2.

An object of class \code{formula} of length 2.

An object of class \code{formula} of length 2.

An object of class \code{formula} of length 2.

An object of class \code{formula} of length 2.
}
\usage{
stdmetrics(x, y, z, i, rn, class, dz = 1, th = 2, zmin = 0)

stdmetrics_z(z, dz = 1, th = 2, zmin = 0)

stdmetrics_i(i, z = NULL, class = NULL, rn = NULL)

stdmetrics_rn(rn, class = NULL)

stdmetrics_pulse(pulseID, rn)

stdmetrics_ctrl(x, y, z)

stdtreemetrics(x, y, z)

stdshapemetrics(x, y, z)

.stdmetrics

.stdmetrics_z

.stdmetrics_i

.stdmetrics_rn

.stdmetrics_pulse

.stdmetrics_ctrl

.stdtreemetrics

.stdshapemetrics
}
\arguments{
\item{x, y, z, i}{Coordinates of the points, Intensity}

\item{rn, class}{ReturnNumber, Classification}

\item{dz}{numeric. Layer thickness  metric \link[=entropy]{entropy}}

\item{th}{numeric. Threshold for metrics pzabovex. Can be a vector to compute with several thresholds.}

\item{zmin}{numeric. Lower bound of the integral for zpcumx metrics.
See \href{https://github.com/r-lidar/lidR/wiki/stdmetrics}{wiki page} and Wood et al.
(2008) reference.}

\item{pulseID}{The number referencing each pulse}
}
\description{
Predefined metrics functions intended to me used in \verb{*_metrics} function such as
\link{pixel_metrics}, \link{cloud_metrics}, \link{crown_metrics}, \link{voxel_metrics} and
so on. Each function comes with a convenient shortcuts for lazy coding. The \code{lidR} package aims
to provide an easy way to compute user-defined metrics rather than to provide them. However, for
efficiency and to save time, sets of standard metrics have been predefined (see details). Every
function can be computed by every \verb{*_metrics} functions however \verb{stdmetrics*} are
more pixel-based metrics, \code{stdtreemetrics} are more tree-based metrics and \code{stdshapemetrics} are
more point-based metrics. For example the metric \code{zmean} computed by \code{stdmetrics_z} makes sense
when computed at the pixel level but brings no information at the voxel level.
}
\details{
The function names, their parameters and the output names of the metrics rely on a nomenclature
chosen for brevity:
\itemize{
\item{\code{z}: refers to the elevation}
\item{\code{i}: refers to the intensity}
\item{\code{rn}: refers to the return number}
\item{\code{q}: refers to quantile}
\item{\code{a}: refers to the ScanAngleRank or ScanAngle}
\item{\code{n}: refers to a number (a count)}
\item{\code{p}: refers to a percentage}
}
For example the metric named \code{zq60} refers to the elevation, quantile, 60 i.e. the 60th percentile
of elevations. The metric \code{pground} refers to a percentage. It is the percentage of points
classified as ground. The function \code{stdmetric_i} refers to metrics of intensity. A description of
each existing metric can be found on the \href{https://github.com/r-lidar/lidR/wiki/stdmetrics}{lidR wiki page}.\cr\cr
Some functions have optional parameters. If these parameters are not provided the function
computes only a subset of existing metrics. For example, \code{stdmetrics_i} requires the intensity
values, but if the elevation values are also provided it can compute additional metrics such as
cumulative intensity at a given percentile of height.\cr\cr
Each function has a convenient associated variable. It is the name of the function, with a
dot before the name. This enables the function to be used without writing parameters. The cost
of such a feature is inflexibility. It corresponds to a predefined behaviour (see examples)\cr
\describe{
\item{\code{stdmetrics}}{is a combination of \code{stdmetrics_ctrl} + \code{stdmetrics_z} +
\code{stdmetrics_i} +  \code{stdmetrics_rn}}
\item{\code{stdtreemetrics}}{is a special function that works with \link{crown_metrics}. Actually,
it won't fail with other functions but the output makes more sense if computed at the
individual tree level.}
\item{\code{stdshapemetrics}}{is a set of eigenvalue based feature described in Lucas et al, 2019
(see references).}
}
}
\examples{
LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
las <- readLAS(LASfile, select = "*", filter = "-keep_random_fraction 0.5")

# All the predefined metrics
m1 <- pixel_metrics(las, ~stdmetrics(X,Y,Z,Intensity,ReturnNumber,Classification,dz=1), res = 40)

# Convenient shortcut
m2 <- pixel_metrics(las, .stdmetrics, res = 40)

# Basic metrics from intensities
m3 <- pixel_metrics(las, ~stdmetrics_i(Intensity), res = 40)

# All the metrics from intensities
m4 <- pixel_metrics(las, ~stdmetrics_i(Intensity, Z, Classification, ReturnNumber), res = 40)

# Convenient shortcut for the previous example
m5 <- pixel_metrics(las, .stdmetrics_i, res = 40)

# Combine some predefined function with your own new metrics
# Here convenient shortcuts are no longer usable.
myMetrics = function(z, i, rn)
{
  first  <- rn == 1L
  zfirst <- z[first]
  nfirst <- length(zfirst)
  above2 <- sum(z > 2)

  x <- above2/nfirst*100

  # User's metrics
  metrics <- list(
     above2aboven1st = x,       # Num of returns above 2 divided by num of 1st returns
     zimean  = mean(z*i),       # Mean products of z by intensity
     zsqmean = sqrt(mean(z^2))  # Quadratic mean of z
   )

  # Combined with standard metrics
  return( c(metrics, stdmetrics_z(z)) )
}

m10 <- pixel_metrics(las, ~myMetrics(Z, Intensity, ReturnNumber), res = 40)

# Users can write their own convenient shorcuts like this:
.myMetrics = ~myMetrics(Z, Intensity, ReturnNumber)
m11 <- pixel_metrics(las, .myMetrics, res = 40)

}
\references{
M. Woods, K. Lim, and P. Treitz. Predicting forest stand variables from LiDAR data in
the Great Lakes – St. Lawrence forest of Ontario. The Forestry Chronicle. 84(6): 827-839.
https://doi.org/10.5558/tfc84827-6

Lucas, C., Bouten, W., Koma, Z., Kissling, W. D., & Seijmonsbergen, A. C. (2019). Identification
of Linear Vegetation Elements in a Rural Landscape Using LiDAR Point Clouds. Remote Sensing, 11(3), 292.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io_readLAS.R
\name{readLASheader}
\alias{readLASheader}
\title{Read a .las or .laz file header}
\usage{
readLASheader(file)
}
\arguments{
\item{file}{characters. Path to one file.}
}
\value{
A LASheader object
}
\description{
Reads a .las or .laz file header into an object of class \link[=LASheader-class]{LASheader}.
This function strictly reads the header while the function \link{readLAS} can alter the header to
fit the actual data loaded.
}
\examples{
LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
header = readLASheader(LASfile)

print(header)
plot(header)

\dontrun{
plot(header, mapview = TRUE)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/doc-lidR.R
\docType{package}
\name{lidR-package}
\alias{lidR}
\alias{lidR-package}
\title{lidR: airborne LiDAR for forestry applications}
\description{
lidR provides a set of tools to manipulate airborne LiDAR data in forestry contexts. The package
works with .las or .laz files. The toolbox includes algorithms for DSM, CHM, DTM, ABA,
normalisation, tree detection, tree segmentation, tree delineation, colourization, validation and
other tools, as well as a processing engine to process broad LiDAR coverage split into many files.
}
\details{
To learn more about lidR, start with the vignettes: browseVignettes(package = "lidR"). Users can also
find unofficial supplementary documentation in the \href{https://r-lidar.github.io/lidRbook/}{lidR book}.
To ask "how to" questions please ask on \href{https://gis.stackexchange.com/}{gis.stackexchange.com}
with the tag \code{lidr}.
}
\section{Package options}{

\describe{
\item{\code{lidR.progress}}{Several functions have a progress bar for long operations (but not all).
Should lengthy operations show a progress bar? Default: TRUE}
\item{\code{lidR.progress.delay}}{The progress bar appears only for long operations. After how many seconds
of computation does the progress bar appear? Default: 2}
\item{\code{lidR.raster.default}}{The functions that return a raster are raster agnostic meaning
that they work either with rasters from packages `raster`, `stars` or `terra`. By default they return
rasters from `stars`. Can be one of "raster", "stars" or "terra". Default: "terra"}
\item{\code{lidR.check.nested.parallelism}}{The catalog processing engine (\link{catalog_apply})
checks the parallel strategy chosen by the user and verify if C++ parallelization with
OpenMP should be disabled to avoid nested parallel loops. Default: TRUE. If FALSE
the catalog processing engine will not check for nested parallelism
and will respect the settings of \link{set_lidr_threads}.}
}
}

\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/r-lidar/lidR}
  \item Report bugs at \url{https://github.com/r-lidar/lidR/issues}
}

}
\author{
\strong{Maintainer}: Jean-Romain Roussel \email{info@r-lidar.com} [copyright holder]

Authors:
\itemize{
  \item David Auty (Reviews the documentation) [contributor]
}

Other contributors:
\itemize{
  \item Florian De Boissieu (Fixed bugs and improved catalog features) [contributor]
  \item Andrew Sánchez Meador (Implemented wing2015() for segment_snags()) [contributor]
  \item Bourdon Jean-François (Contributed to Roussel2020() for track_sensor()) [contributor]
  \item Gatziolis Demetrios (Implemented Gatziolis2019() for track_sensor()) [contributor]
  \item Leon Steinmeier (Contributed to parallelization management) [contributor]
  \item Stanislaw Adaszewski (Author of the C++ concaveman code) [copyright holder]
  \item Benoît St-Onge (Author of the 'chm_prep' function) [copyright holder]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decimate_points.R
\name{decimate_points}
\alias{decimate_points}
\title{Decimate a LAS object}
\usage{
decimate_points(las, algorithm)
}
\arguments{
\item{las}{An object of class \link[lidR:LAS-class]{LAS} or \link[lidR:LAScatalog-class]{LAScatalog}.}

\item{algorithm}{function. An algorithm of point decimation. \code{lidR} have: \link{random},
\link{homogenize}, \link{highest}, \link{lowest}, \link{random_per_voxel} and \link{barycenter_per_voxel}.}
}
\value{
If the input is a \code{LAS} object, returns a \code{LAS} object. If the input is a
\code{LAScatalog}, returns a \code{LAScatalog}.
}
\description{
Reduce the number of points using several possible algorithms.
}
\section{Non-supported LAScatalog options}{

The option `select` is not supported and not respected because it always preserves the file format
and all the attributes. `select = "*"` is imposed internally.\cr
The options `chunk buffer` is not supported and not respected because it is not needed.
}

\examples{
LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
las <- readLAS(LASfile, select = "xyz")

# Select points randomly to reach an overall density of 1
thinned1 <- decimate_points(las, random(1))
#plot(rasterize_density(las))
#plot(rasterize_density(thinned1))

# Select points randomly to reach an homogeneous density of 1
thinned2 <- decimate_points(las, homogenize(1,5))
#plot(rasterize_density(thinned2))

# Select the highest point within each pixel of an overlayed grid
thinned3 = decimate_points(las, highest(5))
#plot(thinned3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clip_roi.R
\name{clip}
\alias{clip}
\alias{clip_roi}
\alias{clip_rectangle}
\alias{clip_polygon}
\alias{clip_circle}
\alias{clip_transect}
\title{Clip points in regions of interest}
\usage{
clip_roi(las, geometry, ...)

clip_rectangle(las, xleft, ybottom, xright, ytop, ...)

clip_polygon(las, xpoly, ypoly, ...)

clip_circle(las, xcenter, ycenter, radius, ...)

clip_transect(las, p1, p2, width, xz = FALSE, ...)
}
\arguments{
\item{las}{An object of class \link[lidR:LAS-class]{LAS} or \link[lidR:LAScatalog-class]{LAScatalog}.}

\item{geometry}{a geometric object. spatial points, spatial polygons in sp or sf/sfc format, Extent,
bbox, 2x2 matrix}

\item{...}{in \code{clip_roi}: optional supplementary options (see supported geometries). Unused in
other functions}

\item{xleft, ybottom, xright, ytop}{numeric. coordinates of one or several rectangles.}

\item{xpoly, ypoly}{numeric. x coordinates of a polygon.}

\item{xcenter, ycenter}{numeric. x coordinates of on or several disc centres.}

\item{radius}{numeric. disc radius or radii.}

\item{p1, p2}{numeric vectors of length 2 that gives the coordinates of two points that define a
transect}

\item{width}{numeric. width of the transect.}

\item{xz}{bool. If \code{TRUE} the point cloud is reoriented to fit on XZ coordinates}
}
\value{
If the input is a LAS object: an object of class LAS, or a \code{list} of LAS objects if the
query implies several regions of interest.\cr\cr
If the input is a LAScatalog object: an object of class LAS, or a \code{list} of LAS
objects if the query implies several regions of interest, or a LAScatalog if the
queries are immediately written into files without loading anything in R.
}
\description{
Clip points within a given region of interest (ROI) from a point cloud (LAS object) or a collection
of files (LAScatalog object).
}
\section{Non-supported LAScatalog options}{

The option \verb{chunk size}, \code{buffer}, \verb{chunk alignment} and \code{select} are not supported by \verb{clip_*}
because they are meaningless in this context.
}

\examples{
LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")

# Load the file and clip the region of interest
las = readLAS(LASfile, select = "xyz", filter = "-keep_first")
subset1 = clip_rectangle(las, 684850, 5017850, 684900, 5017900)

# Do not load the file(s), extract only the region of interest
# from a bigger dataset
ctg = readLAScatalog(LASfile, progress = FALSE, filter = "-keep_first")
subset2 = clip_rectangle(ctg, 684850, 5017850, 684900, 5017900)

# Extract all the polygons from a shapefile
f <- system.file("extdata", "lake_polygons_UTM17.shp", package = "lidR")
lakes <- sf::st_read(f, quiet = TRUE)
subset3 <- clip_roi(las, lakes)

# Extract the polygons for a catalog, write them in files named
# after the lake names, do not load anything in R
opt_output_files(ctg) <- paste0(tempfile(), "_{LAKENAME_1}")
new_ctg = clip_roi(ctg, lakes)
plot(new_ctg)

# Extract a transect
p1 <- c(684800, y = 5017800)
p2 <- c(684900, y = 5017900)
tr1 <- clip_transect(las, p1, p2, width = 4)

\dontrun{
plot(subset1)
plot(subset2)
plot(subset3)

plot(tr1, axis = TRUE, clear_artifacts = FALSE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/retrieve_info.R
\name{retrieve_pulses}
\alias{retrieve_pulses}
\alias{retrieve_flightlines}
\alias{retrieve_scanlines}
\title{Retrieve individual pulses, flightlines or scanlines}
\usage{
retrieve_pulses(las)

retrieve_flightlines(las, dt = 30)

retrieve_scanlines(las)
}
\arguments{
\item{las}{A LAS object}

\item{dt}{numeric. The threshold time-lag used to retrieve flightlines}
}
\value{
An object of class \code{LAS}
}
\description{
Retrieve each individual pulse, individual flightline or individual scanline and assigns a number
to each point. The LAS object must be properly populated according to LAS specifications otherwise
users could find unexpected outputs.
}
\details{
\describe{
\item{\code{retrieve_pulses}}{Retrieves each individual pulse. It uses GPS time. An attribute
\code{pulseID} is added in the \code{LAS} object}
\item{\code{retrieve_scanlines}}{Retrieves each individual scanline. When data are sampled according to a
saw-tooth pattern (oscillating mirror), a scanline is one line, or row of data. The function relies
on the GPS field time to order the data. Then, the \code{ScanDirectionFlag} attribute is used to
retrieve each scanline. An attribute \code{scanlineID} is added in the \code{LAS} object}
\item{\code{retrieve_flightlines}}{Retrieves each individual flightline. It uses GPS time. In a
continuous dataset, once points are ordered by GPS time, the time between two consecutive points
does not exceed a few milliseconds. If the time between two consecutive points is too long it means
that the second point is from a different flightline. The default threshold is 30 seconds.
An attribute \code{flightlineID} is added in the \code{LAS} object.}
}
}
\examples{
LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
las <- readLAS(LASfile)

las <- retrieve_pulses(las)
las

las <- retrieve_flightlines(las)
#plot(las, color = "flightlineID")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics_point.R
\name{point_metrics}
\alias{point_metrics}
\alias{point_eigenvalues}
\title{Point-based metrics}
\usage{
point_metrics(las, func, k, r, xyz = FALSE, filter = NULL, ...)

point_eigenvalues(
  las,
  k,
  r,
  xyz = FALSE,
  metrics = FALSE,
  coeffs = FALSE,
  filter = NULL
)
}
\arguments{
\item{las}{An object of class LAS}

\item{func}{formula. An expression to be applied to each point neighbourhood (see also
\link{template_metrics}).}

\item{k, r}{integer and numeric respectively for k-nearest neighbours and radius of the neighborhood
sphere. If k is given and r is missing, computes with the knn, if r is given and k is missing
computes with a sphere neighborhood, if k and r are given computes with the knn and a limit on the
search distance.}

\item{xyz}{logical. Coordinates of each point are returned in addition to each metric. Otherwise an
ID referring to each point.}

\item{filter}{formula of logical predicates. Enables the function to run only on points of interest
in an optimized way. See examples.}

\item{...}{unused.}

\item{metrics}{logical. Compute metrics or not}

\item{coeffs}{logical. Principal component coefficients are returned}
}
\description{
Computes a series of user-defined descriptive statistics for a LiDAR dataset for each point based
on its k-nearest neighbours or its sphere neighbourhood.
}
\details{
When the neighbourhood is knn the user-defined function is fed with the current
processed point and its k-1 neighbours. The current point being considered as
the 1-neighbour with a distance 0 to the reference point. The points are ordered
by distance to the central point. When the neighbourhood is a sphere the processed
point is also included in the query but points are coming in a random order. \code{point_eigenmetrics}
computes the eigenvalues of the covariance matrix and computes associated metrics following
Lucas et al, 2019 (see references). It is equivalent to \code{point_metrics(las, .stdshapemetrics)}
but much faster because it is optimized and parallelized internally.
}
\section{Performances}{

It is important to bear in mind that this function is very fast for the feature it provides i.e.
mapping a user-defined function at the point level using optimized memory management. However, it
is still computationally demanding.\cr\cr
To help users to get an idea of how computationally demanding this function is, let's compare it to
\link{pixel_metrics}. Assuming we want to apply \code{mean(Z)} on a 1 km² tile with 1 point/m²
with a resolution of 20 m (400 m² cells), then the function \code{mean} is called roughly 2500
times (once  per cell). On the contrary, with \code{point_metrics}, \code{mean} is called 1000000
times (once per point). So the function is expected to be more than 400 times slower in this specific
case (but it does not provide the same feature).\cr\cr
This is why the user-defined function is expected to be well-optimized, otherwise it might drastically
slow down this already heavy computation. See examples.
}

\examples{
\dontrun{
LASfile <- system.file("extdata", "Topography.laz", package="lidR")

# Read only 0.5 points/m^2 for the purposes of this example
las = readLAS(LASfile, filter = "-thin_with_grid 2")

# Computes the eigenvalues of the covariance matrix of the neighbouring
# points and applies a test on these values. This function simulates the
# 'shp_plane()' algorithm from 'segment_shape()'
plane_metrics1 = function(x,y,z, th1 = 25, th2 = 6) {
  xyz <- cbind(x,y,z)
  cov_m <- cov(xyz)
  eigen_m <- eigen(cov_m)$value
  is_planar <- eigen_m[2] > (th1*eigen_m[3]) && (th2*eigen_m[2]) > eigen_m[1]
  return(list(planar = is_planar))
}

# Apply a user-defined function
M <- point_metrics(las, ~plane_metrics1(X,Y,Z), k = 25)
#> Computed in 6.3 seconds

# We can verify that it returns the same as 'shp_plane'
las <- segment_shapes(las, shp_plane(k = 25), "planar")
#> Computed in 0.1 seconds

all.equal(M$planar, las$planar)

# At this stage we can be clever and find that the bottleneck is
# the eigenvalue computation. Let's write a C++ version of it with
# Rcpp and RcppArmadillo
Rcpp::sourceCpp(code = "
#include <RcppArmadillo.h>
// [[Rcpp::depends(RcppArmadillo)]]

// [[Rcpp::export]]
SEXP eigen_values(arma::mat A) {
arma::mat coeff;
arma::mat score;
arma::vec latent;
arma::princomp(coeff, score, latent, A);
return(Rcpp::wrap(latent));
}")

plane_metrics2 = function(x,y,z, th1 = 25, th2 = 6) {
  xyz <- cbind(x,y,z)
  eigen_m <- eigen_values(xyz)
  is_planar <- eigen_m[2] > (th1*eigen_m[3]) && (th2*eigen_m[2]) > eigen_m[1]
  return(list(planar = is_planar))
}

M <- point_metrics(las, ~plane_metrics2(X,Y,Z), k = 25)
#> Computed in 0.5 seconds

all.equal(M$planar, las$planar)
# Here we can see that the optimized version is way better but is still 5-times slower
# because of the overhead of calling R functions and switching back and forth from R to C++.

M <- point_eigenvalues(las, k = 25)
is_planar = M$eigen_medium > (25*M$eigen_smallest) & (6*M$eigen_medium) > M$eigen_largest

# Use the filter argument to process only first returns
M1 <- point_metrics(las, ~plane_metrics2(X,Y,Z), k = 25, filter = ~ReturnNumber == 1)
dim(M1) # 13894 instead of 17182 previously.
}
}
\concept{metrics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/voxelize_points.R
\name{voxelize_points}
\alias{voxelize_points}
\title{Voxelize a point cloud}
\usage{
voxelize_points(las, res)
}
\arguments{
\item{las}{An object of class \link[lidR:LAS-class]{LAS} or \link[lidR:LAScatalog-class]{LAScatalog}.}

\item{res}{numeric. The resolution of the voxels. \code{res = 1} for a 1x1x1 cubic voxels. Optionally
\code{res = c(1,2)} for non-cubic voxels (1x1x2 cuboid voxel).}
}
\value{
If the input is a \code{LAS} object, returns a \code{LAS} object. If the input is a
\code{LAScatalog}, returns a \code{LAScatalog}.
}
\description{
Reduce the number of points by voxelizing the point cloud. If the Intensity is part of the attributes
it is preserved and aggregated as \code{mean(Intensity)}. Other attributes cannot be aggregated and
are lost.
}
\examples{
LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
las = readLAS(LASfile, select = "xyz")

las2 = voxelize_points(las, 5)
#plot(las2, voxel = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_constant.R
\docType{data}
\name{asprs}
\alias{asprs}
\alias{LASNONCLASSIFIED}
\alias{LASUNCLASSIFIED}
\alias{LASGROUND}
\alias{LASLOWVEGETATION}
\alias{LASMEDIUMVEGETATION}
\alias{LASHIGHVEGETATION}
\alias{LASBUILDING}
\alias{LASLOWPOINT}
\alias{LASKEYPOINT}
\alias{LASWATER}
\alias{LASRAIL}
\alias{LASROADSURFACE}
\alias{LASWIREGUARD}
\alias{LASWIRECONDUCTOR}
\alias{LASTRANSMISSIONTOWER}
\alias{LASBRIGDE}
\alias{LASNOISE}
\title{ASPRS LAS Classification}
\format{
An object of class \code{integer} of length 1.

An object of class \code{integer} of length 1.

An object of class \code{integer} of length 1.

An object of class \code{integer} of length 1.

An object of class \code{integer} of length 1.

An object of class \code{integer} of length 1.

An object of class \code{integer} of length 1.

An object of class \code{integer} of length 1.

An object of class \code{integer} of length 1.

An object of class \code{integer} of length 1.

An object of class \code{integer} of length 1.

An object of class \code{integer} of length 1.

An object of class \code{integer} of length 1.

An object of class \code{integer} of length 1.

An object of class \code{integer} of length 1.

An object of class \code{integer} of length 1.

An object of class \code{integer} of length 1.
}
\usage{
LASNONCLASSIFIED

LASUNCLASSIFIED

LASGROUND

LASLOWVEGETATION

LASMEDIUMVEGETATION

LASHIGHVEGETATION

LASBUILDING

LASLOWPOINT

LASKEYPOINT

LASWATER

LASRAIL

LASROADSURFACE

LASWIREGUARD

LASWIRECONDUCTOR

LASTRANSMISSIONTOWER

LASBRIGDE

LASNOISE
}
\description{
A set of global variables corresponding to the point classification defined by the ASPRS for the
LAS format. Instead of remembering the classification table of the specification it is possible
to use one of these global variables.
}
\examples{
LASfile <- system.file("extdata", "example.laz", package="rlas")
las = readLAS(LASfile)
las2 = filter_poi(las, Classification \%in\% c(LASGROUND, LASWATER))

print(LASGROUND)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_is.R
\name{is}
\alias{is}
\alias{is.empty,LAS-method}
\alias{is.overlapping}
\alias{is.indexed}
\alias{is.algorithm}
\alias{is.parallelised}
\title{A set of boolean tests on objects}
\usage{
\S4method{is.empty}{LAS}(x)

is.overlapping(catalog)

is.indexed(catalog)

is.algorithm(x)

is.parallelised(algorithm)
}
\arguments{
\item{x}{\code{LAS} object or any R object.}

\item{catalog}{A \code{LAScatalog} object.}

\item{algorithm}{An \code{algorithm} object.}
}
\value{
TRUE or FALSE
}
\description{
\code{is.empty} tests if a \code{LAS} object is a point cloud with 0 points.\cr
\code{is.overlapping} tests if a \code{LAScatalog} has overlapping tiles.\cr
\code{is.indexed} tests if the points of a \code{LAScatalog} are indexed with \code{.lax} files.\cr
\code{is.algorithm} tests if an object is an algorithm of the lidR package.\cr
\code{is.parallelised} tests if an algorithm of the lidR package is natively parallelised with OpenMP.
Returns TRUE if the algorithm is at least partially parallelised i.e. if some portion of the code is
computed in parallel.
}
\examples{
LASfile <- system.file("extdata", "example.laz", package="rlas")
las = readLAS(LASfile)
is.empty(las)

las = new("LAS")
is.empty(las)

f <- lmf(2)
is.parallelised(f)

g <- pitfree()
is.parallelised(g)

ctg <- readLAScatalog(LASfile)
is.indexed(ctg)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/smooth_height.R
\name{smooth_height}
\alias{smooth_height}
\alias{unsmooth_height}
\title{Smooth a point cloud}
\usage{
smooth_height(
  las,
  size,
  method = c("average", "gaussian"),
  shape = c("circle", "square"),
  sigma = size/6
)

unsmooth_height(las)
}
\arguments{
\item{las}{An object of class \code{LAS}}

\item{size}{numeric. The size of the windows used to smooth.}

\item{method}{character. Smoothing method. Can be 'average' or 'gaussian'.}

\item{shape}{character. The shape of the windows. Can be circle or square.}

\item{sigma}{numeric. The standard deviation of the gaussian if the method is gaussian.}
}
\value{
An object of the class \code{LAS}.
}
\description{
Point cloud-based smoothing algorithm. Two methods are available: average within a window and
Gaussian smooth within a window. The attribute \code{Z} of the returned LAS object is the smoothed Z.
A new attribute \code{Zraw} is added to store the original values and can be used to restore the
point cloud with \code{unsmooth_height}.
}
\details{
This method does not use raster-based methods to smooth the point cloud. This is a true point cloud
smoothing. It is not really useful by itself but may be interesting in combination with filters,
for example to develop new algorithms.
}
\examples{
LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
las <- readLAS(LASfile, select = "xyz")

las <- decimate_points(las, highest(1))
#plot(las)

las <- smooth_height(las, 5, "gaussian", "circle", sigma = 2)
#plot(las)

las <- unsmooth_height(las)
#plot(las)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algorithm-dec.R
\name{sample_random}
\alias{sample_random}
\alias{random}
\title{Point Cloud Decimation Algorithm}
\usage{
random(density, use_pulse = FALSE)
}
\arguments{
\item{density}{numeric. The desired output density.}

\item{use_pulse}{logical. Decimate by removing random pulses instead of random points (requires running
\link{retrieve_pulses} first)}
}
\description{
This function is made to be used in \link{decimate_points}. It implements an algorithm that
randomly removes points or pulses to reach the desired density over the area.
}
\examples{
LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
las = readLAS(LASfile, select = "xyz")

# Reach a pulse density of 1 on the overall dataset
thinned1 = decimate_points(las, random(1))
plot(rasterize_density(las))
plot(rasterize_density(thinned1))
}
\seealso{
Other point cloud decimation algorithms: 
\code{\link{sample_homogenize}},
\code{\link{sample_maxima}},
\code{\link{sample_per_voxel}}
}
\concept{point cloud decimation algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.s3.R
\name{plot.lasmetrics3d}
\alias{plot.lasmetrics3d}
\title{Plot voxelized LiDAR data}
\usage{
\method{plot}{lasmetrics3d}(x, y, ...)
}
\arguments{
\item{x}{An object of the class \code{lasmetrics3d}}

\item{y}{Unused (inherited from R base)}

\item{\dots}{Supplementary parameters for \link[lidR:plot]{plot}. The function internally uses the
same plot function than LAS objects.}
}
\description{
This function implements a 3D plot method for 'lasmetrics3d' objects
}
\examples{
\dontrun{
LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
lidar = readLAS(LASfile)

voxels = voxel_metrics(lidar, list(Imean = mean(Intensity)), res = 5)
plot(voxels, color = "Imean", colorPalette = heat.colors(50), trim=60)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algorithm-shp.R
\name{shape_detection}
\alias{shape_detection}
\alias{shp_plane}
\alias{shp_hplane}
\alias{shp_line}
\alias{shp_hline}
\alias{shp_vline}
\title{Algorithms for shape detection of the local point neighbourhood}
\usage{
shp_plane(th1 = 25, th2 = 6, k = 8)

shp_hplane(th1 = 25, th2 = 6, th3 = 0.98, k = 8)

shp_line(th1 = 10, k = 8)

shp_hline(th1 = 10, th2 = 0.02, k = 8)

shp_vline(th1 = 10, th2 = 0.98, k = 8)
}
\arguments{
\item{th1, th2, th3}{numeric. Threshold values (see details)}

\item{k}{integer. Number of neighbours used to estimate the neighborhood.}
}
\description{
These functions are made to be used in \link{segment_shapes}. They implement algorithms for local
neighbourhood shape estimation.
}
\details{
In the following, \eqn{a_1, a_2, a_3}{a1, a2, a3} denote the eigenvalues of the covariance matrix
of the neighbouring points in ascending order. \eqn{|Z_1|, |Z_2|, |Z_3|}{|Z1|, |Z2|, |Z3|} denote
the norm of the Z component of first, second and third axis of the decomposition.
\eqn{th_1, th_2, th_3}{th1, th2, th3}  denote a set of threshold values. Points are labelled
\code{TRUE} if they meet the following criteria. \code{FALSE} otherwise.\cr
\describe{
\item{shp_plane}{Detection of plans based on criteria defined by Limberger & Oliveira (2015) (see
references). A point is labelled TRUE if the neighborhood is approximately planar, that is:
\deqn{a_2 > (th_1 \times a_1) \& (th_2 \times a_2) > a_3}{a2 > (th1*a1) and (th2*a2) > a3}}
\item{shp_hplane}{The same as 'plane' but with an extra test on the orientation of the Z vector
of the principal components to test the horizontality of the surface.
\deqn{a_2 > (th_1 \times a_1) \& (th_2 \times a_2) > a_3 \& |Z_3| > th_3}{a2 > (th1*a1) and (th2*a2) > a3 and |Z3| > th3}
In theory  \eqn{|Z_3|}{|Z3|} should be exactly equal to 1. In practice 0.98 or 0.99 should be fine}
\item{shp_line}{Detection of lines inspired by the Limberger & Oliveira (2015) criterion. A point is
labelled TRUE if the neighbourhood is approximately linear, that is:
\deqn{th_1 \times a_2 < a_3 \& th_1 \times a_1 < a_3}{th1*a2 < a3 and th1*a1 < a3}}
\item{shp_hline}{Detection of horizontal lines inspired by the Limberger & Oliveira (2015) criterion.
A point is labelled TRUE if the neighbourhood is approximately linear and horizontal, that is:
\deqn{th_1 \times a_2 < a_3 \& th_1 \times a_1 < a_3 \& |Z_1| < th_2}{th1*a2 < a3 and th1*a1 < a3 and |Z1| < th2}
In theory \eqn{|Z_1|}{|Z1|} should be exactly equal to 0. In practice 0.02 or 0.01 should be fine}
\item{shp_vline}{Detection of vertical lines inspired by the Limberger & Oliveira (2015) criterion.
A point is labelled TRUE if the neighbourhood is approximately linear and vertical, that is:
\deqn{th_1 \times a_2 < a_3 \& th_1 \times a_1 < a_3 \& |Z_1| > th_2}{th1*a2 < a3 and th1*a1 < a3 and |Z1| > th2}
In theory \eqn{|Z_1|}{|Z1|} should be exactly equal to 1. In practice 0.98 or 0.99 should be fine}
}
}
\examples{
# Generating some data
n = 400
xplane = runif(n,0,6)
yplane = runif(n,0,6)
zplane = xplane + 0.8 * yplane + runif(n, 0, 0.1)
plane = data.frame(X = xplane, Y = yplane, Z = zplane)

xhplane = runif(n,5,15)
yhplane = runif(n,0,10)
zhplane = 5 + runif(n, 0, 0.)
hplane = data.frame(X = xhplane, Y = yhplane, Z = zhplane)

tline = 1:n
xline = 0.05*tline
yline = 0.01*tline
zline = 0.02*tline + runif(n, 0, 0.1)
line = data.frame(X = xline, Y = yline, Z = zline)

thline = 1:n
xhline = 0.05*thline + runif(n, 0, 0.05)
yhline = 10 - 0.01*thline + runif(n, 0, 0.05)
zhline = 3 + runif(n, 0, 0.05)
hline = data.frame(X = xhline, Y = yhline, Z = zhline)

tvline = 1:n
xvline = 5 + runif(n, 0, 0.05)
yvline = 5 + runif(n, 0, 0.05)
zvline = 0.02*tline
vline = data.frame(X = xvline, Y = yvline, Z = zvline)

las <- rbind(plane, line, hplane, hline, vline)
las <- LAS(las)

las <- segment_shapes(las, shp_plane(k = 20), "plane")
las <- segment_shapes(las, shp_hplane(k = 20), "hplane")
las <- segment_shapes(las, shp_line(k = 20), "line")
las <- segment_shapes(las, shp_hline(k = 20), "hline")
las <- segment_shapes(las, shp_vline(k = 20), "vline")

#plot(las)
#plot(las, color = "plane")
#plot(las, color = "hplane")
#plot(las, color = "line")
#plot(las, color = "hline")
#plot(las, color = "vline")
}
\references{
Limberger, F. A., & Oliveira, M. M. (2015). Real-time detection of planar regions in unorganized
point clouds. Pattern Recognition, 48(6), 2043–2053. https://doi.org/10.1016/j.patcog.2014.12.020\cr\cr
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algorithm-gnd.R
\name{ptd}
\alias{ptd}
\title{Progressive TIN Densification}
\usage{
ptd(res = 10, angle = 30, distance = 2, spacing = 0.25, verbose = FALSE)
}
\arguments{
\item{res}{Scalar. The resolution for locating seed points. It takes the lowest point per grid cell.
It should be larger than any above ground structure such as building or bridges. In forested lands
10 m is good but larger values are recommended if human made structure can be observed. 50 m is
good in urban scene}

\item{angle}{Scalar. The threshold for iteration angle, as illustrated in Figure 3 of Axelsson's paper.
Use smaller values (close to 20) for flat areas and larger values (close to 40) for mountainous
areas.}

\item{distance}{Scalar. The threshold for iteration distance, as described in Figure 3 of Axelsson's paper.}

\item{spacing}{Scalar. The spacing of ground points. TIN densification halts if triangles are
smaller than this limit avoiding useless over-densification. The default setting is generally suitable
for most scenarios.}

\item{verbose}{bool}
}
\description{
Classify ground points using a modified version of the Progressive TIN
densification method by Axelsson (2000) (see references). The approach involves iteratively
constructing a triangulated surface model (TIN) to classify ground points (see details).
}
\details{
The method begins by identifying local low points, assuming at least one ground-level point exists
within any X meters area. Initially, the triangles of the TIN model are mostly below the ground
surface, only touching lowest points at their vertices. As the iterations progress, additional
points are incorporated to more accurately follow the true ground surface.\cr\cr
The classification process is controlled by two main parameters:
\itemize{
\item Iteration angle: This controls the rate at which points are classified based on variations in
ground level. Use smaller values (e.g., 20) for flat terrain and larger values (e.g., 40) for
mountainous regions.
\item Iteration distance: This limits large upward jumps in the model, preventing the misclassification
of points in low vegetation or on small buildings.
}
The algorithm is robust to low noise points, meaning it can handle some level of noise without
needing explicit noise classification. However, it is preferable to have noise classified points
to guarantee optimal results.
\cr\cr
The densification process can also stop if the triangle edges become shorter than a specified length.
}
\examples{
LASfile <- system.file("extdata", "Topography.laz", package="lidR")
las <- readLAS(LASfile, select = "xyzrn")

myptd <- ptd(res = 20, distance = 1, angle = 20)
las <- classify_ground(las, myptd)
#plot(las, color = "Classification")
}
\references{
Axelsson, P. (2000). DEM Generation from Laser Scanner Data Using adaptive TIN Models. International
Archives of Photogrammetry and Remote Sensing, 33(B4), 110–117.
https://www.isprs.org/proceedings/xxxiii/congress/part4/111_xxxiii-part4.pdf
}
\seealso{
Other ground segmentation algorithms: 
\code{\link{gnd_csf}},
\code{\link{gnd_mcc}},
\code{\link{gnd_pmf}}
}
\concept{ground segmentation algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_threads.R
\name{set_lidr_threads}
\alias{set_lidr_threads}
\alias{get_lidr_threads}
\title{Set or get number of threads that lidR should use}
\usage{
set_lidr_threads(threads)

get_lidr_threads()
}
\arguments{
\item{threads}{Positive scalar. Default 0 means use all CPU available. Values > 1 mean
using n cores, values in ]0, 1[ mean using a fraction of the cores e.g. 0.5 = half.}
}
\description{
Set and get number of threads to be used in lidR functions that are parallelized with OpenMP.
0 means to utilize all CPU available. \code{get_lidr_threads()} returns the number
of threads that will be used. This affects \code{lidR} package but also the \code{data.table} package
by internally calling \link[data.table:openmp-utils]{setDTthreads} because several functions  of
lidR rely on \code{data.table} but it does not change R itself or other packages using OpenMP.
}
\seealso{
\link{lidR-parallelism}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io_readXLAS.R, R/io_readLAS.R
\name{readALS}
\alias{readALS}
\alias{readTLS}
\alias{readMSLAS}
\alias{readLAS}
\title{Read .las or .laz files}
\usage{
readALS(files, select = "*", filter = "")

readTLS(files, select = "*", filter = "", sort = TRUE)

readMSLAS(files1, files2, files3, select = "*", filter = "")

readLAS(files, select = "*", filter = "")
}
\arguments{
\item{files}{characters. Path(s) to one or several a file(s). Can also be a
\link[=LAScatalog-class]{LAScatalog} object.}

\item{select}{character. Read only attributes of interest to save memory (see details).}

\item{filter}{character. Read only points of interest to save memory (see details).}

\item{sort}{boolean. To optimize even more the computation speed the point cloud is spatially sorted}

\item{files1, files2, files3}{characters. Path(s) to one or several a file(s). Each argument being
one channel (see section 'Multispectral data'). `files2` and `files3` can be missing.}
}
\value{
A LAS object
}
\description{
Reads .las or .laz files into an object of class \link[=LAS-class]{LAS}. If several files are read at
once the returned LAS object is considered as one LAS file. The optional parameters enable the user
to save a substantial amount of memory by choosing to load only the attributes or points of interest.
LAS formats 1.0 to 1.4 are supported. Point Data Record Format 0 to 10 are supported.\cr\cr
\code{readLAS} is the original function and always works. Using one of the \code{readALS} or \code{readTLS} functions
adds information to the returned object to register a point-cloud type. Registering the correct point
type improves the performance of some functions by enabling users to select an appropriate spatial index.
See \link[=lidR-spatial-index]{spatial indexing}. Notice that by legacy and for backwards-compatibility reasons,
\code{readLAS()} and \code{readALS()} are equivalent because lidR was originally designed for ALS and thus the
original function \code{readLAS()} was (supposedly) used for ALS. Reading a TLS dataset with \code{readLAS()} instead
of \code{readTLS()} is perfectly valid but less powerful.
}
\details{
\strong{Select:} the 'select' argument specifies the data that will actually be loaded. For example,
'xyzia' means that the x, y, and z coordinates, the intensity and the scan angle will be loaded.
The supported entries are t - gpstime, a - scan angle, i - intensity, n - number of returns,
r - return number, c - classification, s - synthetic flag, k - keypoint flag, w - withheld flag,
o - overlap flag (format 6+), u - user data, p - point source ID, e - edge of flight line flag,
d - direction of scan flag, R - red channel of RGB color, G - green channel of RGB color,
B - blue channel of RGB color, N - near-infrared channel, C - scanner channel (format 6+),
W - Full waveform.
Also numbers from 1 to 9 for the extra bytes data numbers 1 to 9. 0 enables all extra bytes to be
loaded and '*' is the wildcard that enables everything to be loaded from the LAS file. \cr
Note that x, y, z are implicit and always loaded. 'xyzia' is equivalent to 'ia'.\cr\cr
\strong{Filter:} the 'filter' argument allows filtering of the point cloud while reading files.
This is much more efficient than \link{filter_poi} in many ways. If the desired filters are known
before reading the file, the internal filters should always be preferred. The available filters are
those from \code{LASlib} and can be found by running the following command: \code{readLAS(filter = "-help")}.
(see also \link[rlas:read.las]{rlas::read.las}). From \code{rlas} v1.3.6 the transformation commands
can also be passed via the argument filter.
}
\section{Multispectral data}{

Multispectral laser data are often stored in 3 different files. If this is the case
\code{readMSLAS} reads the .las or .laz files of each channel and merges them into
an object of class \link[=LAS-class]{LAS} and takes care of attributing an ID to each
channel. If the multisprectral point cloud is already stored in a single file
leave \code{file2} and \code{file3} missing.
}

\section{Full waveform}{

With most recent versions of the \code{rlas} package, full waveform (FWF) can be read and \code{lidR}
provides some compatible functions. However, the support of FWF is still a work-in-progress
in the \code{rlas} package. How it is read, interpreted and represented in R may change. Consequently,
tools provided by \code{lidR} may also change until the support of FWF becomes mature and
stable in \code{rlas}. See also \link[rlas:read.las]{rlas::read.las}.\cr\cr
Remember that FWF represents an insanely huge amount of data. It terms of memory it is like
having between 10 to 100 times more points. Consequently, loading FWF data in R should be
restricted to relatively small point clouds.
}

\examples{
LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
las = readLAS(LASfile)
las = readLAS(LASfile, select = "xyz")
las = readLAS(LASfile, select = "xyzi", filter = "-keep_first")
las = readLAS(LASfile, select = "xyziar", filter = "-keep_first -drop_z_below 0")

# Negation of attributes is also possible (all except intensity and angle)
las = readLAS(LASfile, select = "* -i -a")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_coordinates.R
\name{st_coordinates}
\alias{st_coordinates}
\alias{st_coordinates.LAS}
\alias{st_coordinates.LAScatalog}
\title{Coordinates of a LAS* object in a matrix form}
\usage{
\method{st_coordinates}{LAS}(x, z = TRUE, ...)

\method{st_coordinates}{LAScatalog}(x, ...)
}
\arguments{
\item{x}{A LAS* object}

\item{z}{bool. Return XY or XYZ matrix}

\item{...}{unused.}
}
\value{
matrix
}
\description{
Retrieve coordinates of a \verb{LAS*} object in matrix form. It creates a copy of the coordinates
because of the coercion from \code{data.frame} to \code{matrix}. This function inherits \code{sf::st_coordinates}
}
\examples{
LASfile <- system.file("extdata", "example.laz", package="rlas")
las <- readLAS(LASfile)
sf::st_coordinates(las)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/las_tools.R
\name{las_utilities}
\alias{las_utilities}
\alias{las_rescale}
\alias{las_reoffset}
\alias{las_quantize}
\alias{las_update}
\alias{quantize}
\alias{is.quantized}
\alias{count_not_quantized}
\alias{storable_coordinate_range}
\alias{header}
\alias{payload}
\alias{phb}
\alias{vlr}
\alias{evlr}
\title{LAS utilities}
\usage{
las_rescale(las, xscale, yscale, zscale)

las_reoffset(las, xoffset, yoffset, zoffset)

las_quantize(las, by_reference = TRUE)

las_update(las)

quantize(x, scale, offset, by_reference = TRUE, ...)

is.quantized(x, scale, offset, ...)

count_not_quantized(x, scale, offset)

storable_coordinate_range(scale, offset)

header(las)

payload(las)

phb(las)

vlr(las)

evlr(las)
}
\arguments{
\item{las}{An object of class LAS}

\item{xscale, yscale, zscale}{scalar. Can be missing if not relevant.}

\item{xoffset, yoffset, zoffset}{scalar. Can be missing if not relevant.}

\item{by_reference}{bool. Update the data in place without allocating new memory.}

\item{x}{numeric. Coordinates vector}

\item{scale, offset}{scalar. scale and offset}

\item{...}{Unused.}
}
\description{
Tools to manipulate LAS objects maintaining compliance with
\href{https://community.asprs.org/leadership-restricted/leadership-content/public-documents/standards}{ASPRS specification}
}
\details{
In the specification of the LAS format the coordinates are expected to be given
with a certain precision e.g. 0.01 for a millimeter precision (or millifeet), meaning
that a file records e.g. 123.46 not 123.45678. Also, coordinates are stored as
integers. This is made possible with a scale and offset factor. For example,
123.46 with an offset of 100 and a scale factor of 0.01 is actually stored as
(123.46 - 100)/0.01 = 2346. Storing 123.45678 with a scale factor of 0.01 and an offset
of 100 is invalid because it does not convert to an integer: (123.45678-100)/0.01
= 2345.678. Having an invalid LAS object may be critical in some lidR applications.
When writing into a LAS file, users will loose the extra precision without
warning and some algorithms in lidR use the integer conversion to make integer-based
computation and thus speed-up some algorithms and use less memory. Creation of an
invalid LAS object may cause problems and incorrect outputs.
}
\examples{
LASfile <- system.file("extdata", "example.laz", package="rlas")
las = readLAS(LASfile)

# Manual modification of the coordinates (e.g. rotation, re-alignment, ...)
las@data$X <- las@data$X + 2/3
las@data$Y <- las@data$Y - 5/3

# The point cloud is no longer valid
las_check(las)

# It is important to fix that
las_quantize(las)

# Now the file is almost valid
las_check(las)

# Update the object to set up-to-date header data
las <- las_update(las)
las_check(las)

# In practice the above code is not useful for regular users because the operators
# $<- already perform such operations on-the-fly. Thus the following
# syntax must be preferred and returns valid objects. Previous tools
# were only intended to be used in very specific cases.
las$X <- las$X + 2/3
las$Y <- las$Y - 5/3

# Rescale and reoffset recompute the coordinates with
# new scales and offsets according to LAS specification
las <- las_rescale(las, xscale = 0.01, yscale = 0.01)
las <- las_reoffset(las, xoffset = 300000, yoffset = 5248000)
}
\seealso{
Other las utilities: 
\code{\link{las_check}()}
}
\concept{las utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algorithm-its.R
\name{its_silva2016}
\alias{its_silva2016}
\alias{silva2016}
\title{Individual Tree Segmentation Algorithm}
\usage{
silva2016(chm, treetops, max_cr_factor = 0.6, exclusion = 0.3, ID = "treeID")
}
\arguments{
\item{chm}{`RasterLayer`, `SpatRaster` or `stars`. Canopy height model. Can be computed with \link{rasterize_canopy} or read from
an external file.}

\item{treetops}{`SpatialPoints*` or `sf/sfc_POINT` with 2D or 3D coordinates. Can be computed with
\link{locate_trees} or read from an external file}

\item{max_cr_factor}{numeric. Maximum value of a crown diameter given as a proportion of the
tree height. Default is 0.6, meaning 60\% of the tree height.}

\item{exclusion}{numeric. For each tree, pixels with an elevation lower than \code{exclusion}
multiplied by the tree height will be removed. Thus, this number belongs between 0 and 1.}

\item{ID}{character. If \code{treetops} contains an attribute with the ID for
each tree, the name of this attribute. This way, original IDs will be preserved.}
}
\description{
This functions is made to be used in \link{segment_trees}. It implements an algorithm for tree
segmentation based on Silva et al. (2016) (see reference). This is a simple method
based on seed + voronoi tesselation (equivalent to nearest neighbour). This algorithm is implemented
in the package \code{rLiDAR}. This version is not the version from \code{rLiDAR}. It is
code written from the original article by the lidR authors and is considerably (between 250
and 1000 times) faster.
}
\details{
Because this algorithm works on a CHM only there is no actual need for a point cloud. Sometimes the
user does not even have the point cloud that generated the CHM. \code{lidR} is a point cloud-oriented
library, which is why this algorithm must be used in \link{segment_trees} to merge the result into the point
cloud. However, the user can use this as a stand-alone function like this:
\preformatted{
 chm <- raster("chm.tif")
 ttops <- locate_trees(chm, lmf(3))
 crowns <- silva2016(chm, ttops)()
}
}
\examples{
LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
poi <- "-drop_z_below 0 -inside 481280 3812940 481320 3812980"
las <- readLAS(LASfile, select = "xyz", filter = poi)
col <- pastel.colors(200)

chm <- rasterize_canopy(las, res = 0.5, p2r(0.3))
ker <- matrix(1,3,3)
chm <- terra::focal(chm, w = ker, fun = mean, na.rm = TRUE)

ttops <- locate_trees(chm, lmf(4, 2))
las   <- segment_trees(las, silva2016(chm, ttops))
#plot(las, color = "treeID", colorPalette = col)
}
\references{
Silva, C. A., Hudak, A. T., Vierling, L. A., Loudermilk, E. L., O’Brien, J. J., Hiers,
J. K., Khosravipour, A. (2016). Imputation of Individual Longleaf Pine (Pinus palustris Mill.)
Tree Attributes from Field and LiDAR Data. Canadian Journal of Remote Sensing, 42(5), 554–573.
https://doi.org/10.1080/07038992.2016.1196582.
}
\seealso{
Other individual tree segmentation algorithms: 
\code{\link{its_dalponte2016}},
\code{\link{its_li2012}},
\code{\link{its_watershed}}

Other raster based tree segmentation algorithms: 
\code{\link{its_dalponte2016}},
\code{\link{its_watershed}}
}
\concept{individual tree segmentation algorithms}
\concept{raster based tree segmentation algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/catalog_retile.R
\name{catalog_retile}
\alias{catalog_retile}
\title{Retile a LAScatalog}
\usage{
catalog_retile(ctg)
}
\arguments{
\item{ctg}{A \link[=LAScatalog-class]{LAScatalog} object}
}
\value{
A new \code{LAScatalog} object
}
\description{
Splits or merges files to reshape the original files collection (.las or .laz) into smaller or larger
files. It also enables the addition or removal of a buffer around the tiles. Internally, the
function reads and writes the chunks defined by the internal processing options of a
\link[=LAScatalog-class]{LAScatalog}. Thus, the function is flexible and enables the user to retile
the dataset, retile while adding or removing a buffer (negative buffers are allowed), or optionally
to compress the data by retiling without changing the pattern but by changing
the format (las/laz). \strong{This function is does not load the point cloud into R memory} It streams
from input file(s) to output file(s) and can be applied to large point-cloud with low memory
computer.\cr\cr
Note that this function is not actually very useful because \code{lidR} manages everything
(clipping, processing, buffering, ...) internally using the proper options. Thus, retiling may be
useful for working in other software, for example, but not in \code{lidR}
}
\section{Non-supported LAScatalog options}{

The option \code{select} is not supported and not respected because it always preserves the file
format and all the attributes. \code{select = "*"} is imposed internally.
}

\examples{
\donttest{
LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
ctg = readLAScatalog(LASfile)
plot(ctg)

# Create a new set of 200 x 200 m.las files with first returns only
opt_chunk_buffer(ctg) <- 0
opt_chunk_size(ctg) <- 200
opt_filter(ctg) <- "-keep_first"
opt_chunk_alignment(ctg) <- c(275, 90)
opt_output_files(ctg) <- paste0(tempdir(), "/retile_{XLEFT}_{YBOTTOM}")

# preview the chunk pattern
plot(ctg, chunk = TRUE)

newctg = catalog_retile(ctg)

plot(newctg)

# Create a new set of 200 x 200 m.las files
# but extended with a 50 m buffer in the folder

opt_chunk_buffer(ctg) <- 25
opt_chunk_size(ctg) <- 200
opt_filter(ctg) <- ""
opt_chunk_alignment(ctg) <- c(275, 90)
opt_output_files(ctg) <- paste0(tempdir(), "/{XLEFT}_{YBOTTOM}_buffered")
newctg = catalog_retile(ctg)

plot(newctg)
}

\dontrun{
# Create a new set of compressed .laz file equivalent to the original, keeping only
# first returns above 2 m

opt_chunk_buffer(ctg) <- 0
opt_chunk_size(ctg) <- 0
opt_laz_compression(ctg) <- TRUE
opt_filter(ctg) <- "-keep_first -drop_z_below 2"
opt_output_files(ctg) <- paste0(tempdir(), "/{ORIGINALFILENAME}_first_2m")
newctg = catalog_retile(ctg)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.s3.R
\name{plot_3d}
\alias{plot_3d}
\alias{plot_dtm3d}
\alias{add_dtm3d}
\alias{add_treetops3d}
\alias{add_flightlines3d}
\alias{add_circle3d}
\title{Add a spatial object to a point cloud scene}
\usage{
plot_dtm3d(dtm, bg = "black", clear_artifacts = TRUE, ...)

add_dtm3d(x, dtm, ...)

add_treetops3d(x, ttops, z = "Z", ...)

add_flightlines3d(x, flightlines, z = "Z", ...)

add_circle3d(x, center_x, center_y, radius, height)
}
\arguments{
\item{dtm}{An object of the class `RasterLayer` or `stars` or `SpatRaster`}

\item{bg}{The color for the background. Default is black.}

\item{clear_artifacts}{logical. It is a known and documented issue that 3D visualisation with
\code{rgl} displays artifacts. The points and lines are inaccurately positioned in the space and thus
the rendering may look false or weird. This is because `rgl` computes with single precision `float`.
To fix this, the objects are shifted to (0,0) to reduce the number of digits needed to represent
their coordinates. The drawback is that the objects are not plotted at their actual coordinates.}

\item{\dots}{Supplementary parameters for \link[rgl]{surface3d} or
\link[rgl:spheres]{spheres3d}.}

\item{x}{The output of the function plot used with a LAS object.}

\item{ttops}{A `SpatialPointsDataFrame` or `sf/sfc` that contains tree tops coordinates.}

\item{z}{character. The name of the attribute that contains the height of the tree tops or of the
flightlines. Only for XY geometries Ignored if the input have XYZ geometries}

\item{flightlines}{A `SpatialPointsDataFrame` or `sf` that contains flightlines coordinates.}

\item{center_x, center_y, radius, height}{horizontal circle parameters}
}
\description{
Add a raster (`raster`, `stars` `terra`) object that represents a digital terrain model or a
`SpatialPointsDataFrame` or `sf` that represents tree tops to a point cloud scene. To add elements
to a scene with a point cloud plotted with the function plot from lidR, the functions `add_*`
take as first argument the output of the plot function (see examples), because the plot function
does not plot the actual coordinates of the point cloud, but offset values. See function
\link[=plot]{plot} and its argument `clear_artifacts` for more details. It works only
with `rgl` i.e. `backend = "rgl"` which is the default.
}
\examples{
\dontrun{
LASfile <- system.file("extdata", "Topography.laz", package="lidR")
las <- readLAS(LASfile)

dtm <- rasterize_terrain(las, algorithm = tin())
ttops <- locate_trees(las, lmf(ws = 5))

plot_dtm3d(dtm)

x <- plot(las)
add_dtm3d(x, dtm)
add_treetops3d(x, ttops)

plot(las) |> add_dtm3d(dtm) |> add_treetops3d(ttops)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algorithm-gnd.R
\name{gnd_pmf}
\alias{gnd_pmf}
\alias{pmf}
\alias{util_makeZhangParam}
\title{Ground Segmentation Algorithm}
\usage{
pmf(ws, th)

util_makeZhangParam(
  b = 2,
  dh0 = 0.5,
  dhmax = 3,
  s = 1,
  max_ws = 20,
  exp = FALSE
)
}
\arguments{
\item{ws}{numeric. Sequence of windows sizes to be used in filtering ground returns.
The values must be positive and in the same units as the point cloud (usually meters, occasionally
feet).}

\item{th}{numeric. Sequence of threshold heights above the parameterized ground surface to be
considered a ground return. The values must be positive and in the same units as the point cloud.}

\item{b}{numeric. This is the parameter \eqn{b} in Zhang et al. (2003) (eq. 4 and 5).}

\item{dh0}{numeric. This is \eqn{dh_0} in Zhang et al. (2003) (eq. 7).}

\item{dhmax}{numeric. This is \eqn{dh_{max}} in Zhang et al. (2003) (eq. 7).}

\item{s}{numeric. This is \eqn{s} in Zhang et al. (2003) (eq. 7).}

\item{max_ws}{numeric. Maximum window size to be used in filtering ground returns. This limits
the number of windows created.}

\item{exp}{logical. The window size can be increased linearly or exponentially (eq. 4 or 5).}
}
\description{
This function is made to be used in \link{classify_ground}. It implements an algorithm for segmentation
of ground points based on a progressive morphological filter. This method is an implementation of
the Zhang et al. (2003) algorithm (see reference). Note that this is not a strict implementation
of Zhang et al. This algorithm works at the point cloud level without any rasterization process.
The morphological operator is applied on the point cloud, not on a raster. Also, Zhang et al.
proposed some formulas (eq. 4, 5 and 7) to compute the sequence of windows sizes and thresholds.
Here, these parameters are free and specified by the user. The function \link{util_makeZhangParam}
enables computation of the parameters according to the original paper.
}
\details{
The progressive morphological filter allows for any sequence of parameters. The `util_makeZhangParam`
function enables computation of the sequences using equations (4),
 (5) and (7) from Zhang et al. (see reference and details).


In the original paper the windows size sequence is given by eq. 4 or 5:\cr\cr
\eqn{w_k = 2kb + 1} \cr\cr
or\cr\cr
\eqn{w_k = 2b^k + 1}\cr\cr
In the original paper the threshold sequence is given by eq. 7:\cr\cr
\eqn{th_k = s*(w_k - w_{k-1})*c + th_0}\cr\cr
Because the function \link{classify_ground} applies the morphological operation at the point
cloud level the parameter \eqn{c} is set to 1 and cannot be modified.
}
\examples{
LASfile <- system.file("extdata", "Topography.laz", package="lidR")
las <- readLAS(LASfile, select = "xyzrn", filter = "-inside 273450 5274350 273550 5274450")

ws <- seq(3,12, 3)
th <- seq(0.1, 1.5, length.out = length(ws))

las <- classify_ground(las, pmf(ws, th))
#plot(las, color = "Classification")
}
\references{
Zhang, K., Chen, S. C., Whitman, D., Shyu, M. L., Yan, J., & Zhang, C. (2003). A progressive
morphological filter for removing nonground measurements from airborne LIDAR data. IEEE
Transactions on Geoscience and Remote Sensing, 41(4 PART I), 872–882. http:#doi.org/10.1109/TGRS.2003.810682.
}
\seealso{
Other ground segmentation algorithms: 
\code{\link{gnd_csf}},
\code{\link{gnd_mcc}},
\code{\link{ptd}()}
}
\concept{ground segmentation algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/las_check.R
\name{las_check}
\alias{las_check}
\title{Inspect a LAS object}
\usage{
las_check(las, print = TRUE, ...)
}
\arguments{
\item{las}{An object of class \link[lidR:LAS-class]{LAS} or \link[lidR:LAScatalog-class]{LAScatalog}.}

\item{print}{logical. By default, prints a report and returns a \code{list} invisibly. If
\code{print = FALSE} the functions returns a \code{list} visibly and do not print the report.}

\item{...}{Use \code{deep = TRUE} on a LAScatalog only. Instead of a shallow inspection it reads
all the files and performs a deep inspection.}
}
\value{
A list with three elements named \code{message}, \code{warnings} and \code{errors}. This list is returned
invisibly if \code{print = TRUE}. If \code{deep = TRUE} a nested list is returned with one element
per file.
}
\description{
Performs a deep inspection of a LAS or LAScatalog object and prints a report.\cr\cr
For a LAS object it checks:
\itemize{
\item if the point cloud is valid according to las specification
\item if the header is valid according to las specification
\item if the point cloud is in accordance with the header
\item if the point cloud has duplicated points and degenerated ground points
\item if gpstime and pulses are consistent
\item it the coordinate reference sytem is correctly recorded
\item if some pre-processing, such as normalization or ground filtering, is already done.
\item and much more
}
For a LAScatalog object it checks:
\itemize{
\item if the headers are consistent across files
\item if the files are overlapping
\item if some pre-processing, such as normalization, is already done.
}
For the pre-processing tests the function only makes an estimation and may not be correct.
}
\examples{
LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
las <- readLAS(LASfile)
las_check(las)
}
\seealso{
Other las utilities: 
\code{\link{las_utilities}}
}
\concept{las utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algorithm-out.R
\name{noise_sor}
\alias{noise_sor}
\alias{sor}
\title{Noise Segmentation Algorithm}
\usage{
sor(k = 10, m = 3, quantile = FALSE)
}
\arguments{
\item{k}{numeric. The number of neighbours}

\item{m}{numeric. Multiplier. The maximum distance will be: \verb{avg distance + m * std deviation}.
If \code{quantile = TRUE}, \code{m} becomes the quantile threshold.}

\item{quantile}{boolean. Modification of the original SOR to use a quantile
threshold instead of a standard deviation multiplier. In this case the maximum
distance will be: \code{quantile(distances, probs = m)}}
}
\description{
This function is made to be used in \link{classify_noise}. It implements an
algorithm for outliers (noise) segmentation based on Statistical Outliers
Removal (SOR) methods first described in the PCL library
and also implemented in CloudCompare (see references).
For each point, it computes the mean distance to all its k-nearest neighbours.
The points that are farther than the average distance plus a number of times
(multiplier) the standard deviation are considered noise.
}
\examples{
LASfile <- system.file("extdata", "Topography.laz", package="lidR")
las <- readLAS(LASfile, filter = "-inside 273450 5274350 273550 5274450")

# Add some artificial outliers because the original
# dataset is 'clean'
set.seed(314)
id = round(runif(20, 0, npoints(las)))
set.seed(42)
err = runif(20, -50, 50)
las$Z[id] = las$Z[id] + err

las <- classify_noise(las, sor(15,7))
}
\references{
https://pointclouds.org/documentation/tutorials/statistical_outlier.html \cr
https://www.cloudcompare.org/doc/wiki/index.php?title=SOR_filter
}
\seealso{
Other noise segmentation algorithms: 
\code{\link{noise_ivf}}
}
\concept{noise segmentation algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/las_compression.R
\name{las_compression}
\alias{las_compression}
\alias{las_is_compressed}
\alias{las_size}
\title{Compression of the point cloud}
\usage{
las_is_compressed(las)

las_size(las)
}
\arguments{
\item{las}{A LAS object.}
}
\description{
Package rlas 1.6.0 supports compact representation of non populated attributes. For example \code{UserData}
is usually populated with zeros (not populated). Yet it takes 32 bits per point to store each 0.
With rlas 1.6.0 it can now use 644 bits no matter the number of points loaded if it is not
populated or populated with a unique value.
}
\details{
\code{las_is_compressed} test each attributes and returns a named vector with TRUE if the attribute is
compressed FALSE otherwise.\cr\cr
\code{las_size} returns the true size of a LAS object by considering the compression. \code{object.size} from
base R does not account for ALTREP and consequently cannot measure properly the size of a LAS object
}
\examples{
LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
las = readLAS(LASfile)
las_is_compressed(las)

format(object.size(las), units = "MB")
format(las_size(las), units = "MB")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics_stdmetrics.R
\name{nstdmetrics}
\alias{nstdmetrics}
\alias{rumple_index}
\alias{gap_fraction_profile}
\alias{LAD}
\alias{entropy}
\alias{VCI}
\title{Predefined non standard metrics}
\usage{
rumple_index(x, y = NULL, z = NULL, ...)

gap_fraction_profile(z, dz = 1, z0 = 2)

LAD(z, dz = 1, k = 0.5, z0 = 2)

entropy(z, by = 1, zmax = NULL)

VCI(z, zmax, by = 1)
}
\arguments{
\item{x}{A \code{RasterLayer}, a \code{stars} a \code{SpatRaster} or a vector of x coordinates.}

\item{y}{numeric. If \code{x} is a vector of coordinates: the associated y coordinates.}

\item{z}{vector of positive z coordinates}

\item{...}{unused}

\item{z0}{numeric. The bottom limit of the profile}

\item{k}{numeric. is the extinction coefficient}

\item{by, dz}{numeric. The thickness of the layers used (height bin)}

\item{zmax}{numeric. Maximum elevation for an entropy normalized to zmax.}
}
\value{
numeric. The computed Rumple index.

A data.frame containing the bin elevations (z) and the gap fraction for each bin (gf)

A number between 0 and 1

A number between 0 and 1
}
\description{
Functions and metrics from the literature. See details and references
}
\details{
\describe{
\item{rumple_index}{Computes the roughness of a surface as the ratio between its area and its
projected area on the ground. If the input is a gridded object (raster) the function computes the
surfaces using Jenness's algorithm (see references). If the input is a point cloud the function
uses a Delaunay triangulation of the points and computes the area of each triangle.}
\item{gap_fraction_profile}{Computes the gap fraction profile using the method of Bouvier et al.
(see reference). The function assesses the number of laser points that actually reached the layer
z+dz and those that passed through the layer [z, z+dz]. By definition the layer 0
will always return 0 because no returns pass through the ground. Therefore, the layer 0 is removed
from the returned results.}
\item{LAD}{Computes a leaf area density profile based on the method of Bouvier et al. (see reference)
The function assesses the number of laser points that actually reached the layer z+dz and those that
passed through the layer [z, z+dz] (see \link[=gap_fraction_profile]{gap_fraction_profile}).
Then it computes the log of this quantity and divides it by the extinction coefficient k as
described in Bouvier et al. By definition the layer 0 will always return infinity because no returns
 pass through the ground. Therefore, the layer 0 is removed from the returned results.}
 \item{entropy}{A normalized Shannon vertical complexity index. The Shannon diversity index is a
 measure for quantifying diversity and is based on the number and frequency of species present. This index,
developed by Shannon and Weaver for use in information theory, was successfully transferred
to the description of species diversity in biological systems (Shannon 1948). Here it is applied
to quantify the diversity and the evenness of an elevational distribution of las points. It
makes bins between 0 and the maximum elevation. If there are negative values the function
returns NA.}
\item{VCI}{Vertical Complexity Index. A fixed normalization of the entropy function from van Ewijk
et al. (2011) (see references)}
}
}
\examples{
x <- runif(20, 0, 100)
y <- runif(20, 0, 100)

if (require(geometry, quietly = TRUE))
{
# Perfectly flat surface, rumple_index = 1
z <- rep(10, 20)
rumple_index(x, y, z)

# Rough surface, rumple_index > 1
z <- runif(20, 0, 10)
rumple_index(x, y, z)

# Rougher surface, rumple_index increases
z <- runif(20, 0, 50)
rumple_index(x, y, z)

# Measure of roughness is scale-dependent
rumple_index(x, y, z)
rumple_index(x/10, y/10, z)
}
z <- c(rnorm(1e4, 25, 6), rgamma(1e3, 1, 8)*6, rgamma(5e2, 5,5)*10)
z <- z[z<45 & z>0]

hist(z, n=50)

gapFraction = gap_fraction_profile(z)

plot(gapFraction, type="l", xlab="Elevation", ylab="Gap fraction")
z <- c(rnorm(1e4, 25, 6), rgamma(1e3, 1, 8)*6, rgamma(5e2, 5,5)*10)
z <- z[z<45 & z>0]

lad <- LAD(z)

plot(lad, type="l", xlab="Elevation", ylab="Leaf area density")
z <- runif(10000, 0, 10)

# expected to be close to 1. The highest diversity is given for a uniform distribution
entropy(z, by = 1)

 z <- runif(10000, 9, 10)

# Must be 0. The lowest diversity is given for a unique possibility
entropy(z, by = 1)

z <- abs(rnorm(10000, 10, 1))

# expected to be between 0 and 1.
entropy(z, by = 1)
z <- runif(10000, 0, 10)

VCI(z, by = 1, zmax = 20)

z <- abs(rnorm(10000, 10, 1))

# expected to be closer to 0.
VCI(z, by = 1, zmax = 20)
}
\references{
Jenness, J. S. (2004). Calculating landscape surface area from digital elevation
models. Wildlife Society Bulletin, 32(3), 829–839.

Bouvier, M., Durrieu, S., Fournier, R. a, & Renaud, J. (2015).  Generalizing predictive
models of forest inventory attributes using an area-based approach with airborne las data. Remote
Sensing of Environment, 156, 322-334. http://doi.org/10.1016/j.rse.2014.10.004

Pretzsch, H. (2008). Description and Analysis of Stand Structures. Springer Berlin Heidelberg. http://doi.org/10.1007/978-3-540-88307-4 (pages 279-280)
Shannon, Claude E. (1948), "A mathematical theory of communication," Bell System Tech. Journal 27, 379-423, 623-656.

van Ewijk, K. Y., Treitz, P. M., & Scott, N. A. (2011). Characterizing Forest Succession in Central Ontario using LAS-derived Indices. Photogrammetric Engineering and Remote Sensing, 77(3), 261-269. Retrieved from <Go to ISI>://WOS:000288052100009
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algorithm-dsm.R
\name{dsm_pitfree}
\alias{dsm_pitfree}
\alias{pitfree}
\title{Digital Surface Model Algorithm}
\usage{
pitfree(
  thresholds = c(0, 2, 5, 10, 15),
  max_edge = c(0, 1),
  subcircle = 0,
  highest = TRUE
)
}
\arguments{
\item{thresholds}{numeric. Set of height thresholds according to the Khosravipour et al. (2014) algorithm
description (see references)}

\item{max_edge}{numeric. Maximum edge length of a triangle in the Delaunay triangulation.
If a triangle has an edge length greater than this value it will be removed. The first number is the value
for the classical triangulation (threshold = 0, see also \link{dsmtin}), the second number
is the value for the pit-free algorithm (for thresholds > 0). If \code{max_edge = 0} no trimming
is done (see examples).}

\item{subcircle}{numeric. radius of the circles. To obtain fewer empty pixels the algorithm
can replace each return with a circle composed of 8 points (see details).}

\item{highest}{bool. By default it keeps only the highest point per pixel before to triangulate to
decrease computation time. If highest = FALSE all first returns are used.}
}
\description{
This function is made to be used in \link{rasterize_canopy}. It implements the pit-free algorithm
developed by Khosravipour et al. (2014), which is based on the computation of a set of classical
triangulations at different heights (see references). The \code{subcircle} tweak replaces each
point with 8 points around the original one. This allows for virtual 'emulation' of the fact that
a lidar point is not a point as such, but more realistically a disc. This tweak densifies the point
cloud and the resulting canopy model is smoother and contains fewer 'pits' and empty pixels.
}
\examples{
LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
poi = "-drop_z_below 0 -inside 481280 3812940 481330 3812990"
las <- readLAS(LASfile, filter = poi)
col <- height.colors(50)

# Basic triangulation and rasterization of first returns
chm <- rasterize_canopy(las, res = 0.5, dsmtin())
plot(chm, col = col)

# Khosravipour et al. pitfree algorithm
chm <- rasterize_canopy(las, res = 0.5, pitfree(c(0,2,5,10,15), c(0, 1.5)))
plot(chm, col = col)

\dontrun{
# Potentially complex concave subset of point cloud
x = c(481340, 481340, 481280, 481300, 481280, 481340)
y = c(3812940, 3813000, 3813000, 3812960, 3812940, 3812940)
las2 = clip_polygon(las,x,y)
plot(las2)

# Because the TIN interpolation is done within the convex hull of the point cloud
# dummy pixels are interpolated that are correct according to the interpolation
# method used, but meaningless in our CHM
chm <- rasterize_canopy(las2, res = 0.5, pitfree(max_edge = c(0, 1.5)))
plot(chm, col = col)

chm = rasterize_canopy(las2, res = 0.5, pitfree(max_edge = c(3, 1.5)))
plot(chm, col = col)
}
}
\references{
Khosravipour, A., Skidmore, A. K., Isenburg, M., Wang, T., & Hussin, Y. A. (2014).
Generating pit-free canopy height models from airborne lidar. Photogrammetric Engineering &
Remote Sensing, 80(9), 863-872.
}
\seealso{
Other digital surface model algorithms: 
\code{\link{dsm_point2raster}},
\code{\link{dsm_spikefree}},
\code{\link{dsm_tin}}
}
\concept{digital surface model algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/catalog_intersect.R, R/catalog_select.R
\name{catalog_intersect}
\alias{catalog_intersect}
\alias{catalog_subset}
\alias{catalog_select}
\title{Subset a LAScatalog}
\usage{
catalog_intersect(
  ctg,
  y,
  ...,
  subset = c("subset", "flag_unprocessed", "flag_processed")
)

catalog_select(
  ctg,
  mapview = TRUE,
  subset = c("subset", "flag_unprocessed", "flag_processed")
)
}
\arguments{
\item{ctg}{A \link[=LAScatalog-class]{LAScatalog}}

\item{y}{`bbox`, `sf`, `sfc`, `Extent`, `Raster*`, `Spatial*` objects}

\item{...}{ignored}

\item{subset}{character. By default it subsets the collection It is also possible to flag
the files to maintain the collection as a whole but process only a subset of its content.
\code{flag_unprocessed} flags the files that will not be processed.
\code{flag_processed} flags the files that will be processed.}

\item{mapview}{logical. If \code{FALSE}, use R base plot instead of mapview (no pan, no zoom, see
also \link[=plot]{plot})}
}
\value{
A LAScatalog object
}
\description{
Subset a LAScatalog interactively using the mouse. Subset a LAScatalog with a spatial object to
keep only the tiles of interest. It can be used to select tiles of interest that encompass spatial
objects.
}
\examples{
\dontrun{
ctg = readLAScatalog("<Path to a folder containing a set of .las files>")
new_ctg = catalog_select(ctg)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods-LASheader.R
\name{as}
\alias{as}
\alias{as.list.LASheader}
\title{Transform to a list}
\usage{
\method{as.list}{LASheader}(x, ...)
}
\arguments{
\item{x}{A LASheader object}

\item{...}{unused}
}
\description{
Functions to construct, coerce and check for both kinds of R lists.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algorithm-dtm.R
\name{dtm_idw}
\alias{dtm_idw}
\alias{knnidw}
\title{Spatial Interpolation Algorithm}
\usage{
knnidw(k = 10, p = 2, rmax = 50)
}
\arguments{
\item{k}{integer. Number of k-nearest neighbours. Default 10.}

\item{p}{numeric. Power for inverse-distance weighting. Default 2.}

\item{rmax}{numeric. Maximum radius where to search for knn. Default 50.}
}
\description{
This function is made to be used in \link{rasterize_terrain} or \link{normalize_height}. It implements an algorithm
for spatial interpolation. Interpolation is done using a k-nearest neighbour (KNN) approach with
an inverse-distance weighting (IDW).
}
\examples{
LASfile <- system.file("extdata", "Topography.laz", package="lidR")
las = readLAS(LASfile)

#plot(las)

dtm = rasterize_terrain(las, algorithm = knnidw(k = 6L, p = 2))

#plot(dtm)
#plot_dtm3d(dtm)
}
\seealso{
Other dtm algorithms: 
\code{\link{dtm_kriging}},
\code{\link{dtm_tin}}
}
\concept{dtm algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algorithm-noi.R
\name{range_correction}
\alias{range_correction}
\alias{get_range}
\title{Intensity normalization algorithm}
\usage{
range_correction(sensor, Rs, f = 2.3, gpstime = "gpstime", elevation = "Z")

get_range(las, sensor, gpstime = "gpstime", elevation = "Z")
}
\arguments{
\item{sensor}{`SpatialPointsDataDrame` or `sf` object containing the coordinates of
the sensor at different timepoints t. The time and elevation are stored as attributes
(default names are 'gpstime' and 'Z'). Z can also come from the geometry if the input records XYZ
coordinates. It can be computed with \link{track_sensor}.}

\item{Rs}{numeric. Range of reference.}

\item{f}{numeric. Exponent. Usually between 2 and 3 in vegetation contexts.}

\item{gpstime, elevation}{character. The name of the attributes that store the gpstime of the
position and the elevation of the sensor respectively. If the input contains 3 coordinates points,
`elevation` is not considered.}

\item{las}{an object of class LAS. \code{get_range()} is a regular function documented here for
convenience.}
}
\description{
This function is made to be used in \link{normalize_intensity}. It corrects intensity with a
range correction according to the formula (see references):
\deqn{I_{norm} = I_{obs} \left(\frac{R}{Rs}\right)^f}{Inorm = Iobs * (R/Rs)^f}
To achieve the range correction the position of the sensor must be known at different discrete times.
Using the 'gpstime' of each point, the position of the sensor is interpolated from the reference
and a range correction is applied.
}
\examples{
# A valid file properly populated
LASfile <- system.file("extdata", "Topography.laz", package="lidR")
las <- readLAS(LASfile)

# pmin = 15 because it is an extremely tiny file
# strongly decimated to reduce its size. There are
# actually few multiple returns
sensor <- track_sensor(las, Roussel2020(pmin = 15))

# Here the effect is virtually null because the size of
# the sample is too small to notice any effect of range
las <- normalize_intensity(las, range_correction(sensor, Rs = 2000))

# This might be useful for some applications
R = get_range(las, sensor)
}
\references{
Gatziolis, D. (2011). Dynamic Range-based Intensity Normalization for Airborne, Discrete Return
Lidar Data of Forest Canopies. Photogrammetric Engineering & Remote Sensing, 77(3), 251–259.
https://doi.org/10.14358/pers.77.3.251
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fullwaveform.R
\name{interpret_waveform}
\alias{interpret_waveform}
\title{Convert full waveform data into a regular point cloud}
\usage{
interpret_waveform(las)
}
\arguments{
\item{las}{An object of class LAS with full waveform data}
}
\value{
An object of class LAS 1.2 format 0 with one point per records
}
\description{
Full waveform can be difficult to manipulate and visualize in R. This function converts
a LAS object with full waveform data into a regular point cloud. Each waveform record
becomes a point with XYZ coordinates and an amplitude (units: volts) and an ID that records
each original pulse. Notice that this has the effect of drastically inflating the size of the
object in memory, which is likely already very large
}
\section{Full waveform}{

With most recent versions of the \code{rlas} package, full waveform (FWF) can be read and \code{lidR}
provides some compatible functions. However, the support of FWF is still a work-in-progress
in the \code{rlas} package. How it is read, interpreted and represented in R may change. Consequently,
tools provided by \code{lidR} may also change until the support of FWF becomes mature and
stable in \code{rlas}. See also \link[rlas:read.las]{rlas::read.las}.\cr\cr
Remember that FWF represents an insanely huge amount of data. It terms of memory it is like
having between 10 to 100 times more points. Consequently, loading FWF data in R should be
restricted to relatively small point clouds.
}

\examples{
\dontrun{
LASfile <- system.file("extdata", "fwf.laz", package="rlas")
fwf <- readLAS(LASfile)
las <- interpret_waveform(fwf)
x <- plot(fwf, size = 3, pal = "red")
plot(las, color = "Amplitude", bg = "white", add = x, size = 2)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algorithm-dsm.R
\name{dsm_point2raster}
\alias{dsm_point2raster}
\alias{p2r}
\title{Digital Surface Model Algorithm}
\usage{
p2r(subcircle = 0, na.fill = NULL)
}
\arguments{
\item{subcircle}{numeric. Radius of the circles. To obtain fewer empty pixels the algorithm
can replace each return with a circle composed of 8 points (see details).}

\item{na.fill}{function. A function that implements an algorithm to compute spatial interpolation
to fill the empty pixel often left by points-to-raster methods. \code{lidR} has \link{knnidw},
\link{tin}, and \link{kriging} (see also \link{rasterize_terrain} for more details).}
}
\description{
This function is made to be used in \link{rasterize_canopy}. It implements an algorithm for digital
surface model computation based on a points-to-raster method: for each pixel of the output raster
the function attributes the height of the highest point found. The \code{subcircle} tweak replaces
each point with 8 points around the original one. This allows for virtual 'emulation' of the fact
that a lidar point is not a point as such, but more realistically a disc. This tweak densifies the
point cloud and the resulting canopy model is smoother and contains fewer 'pits' and empty pixels.
}
\examples{
LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
las <- readLAS(LASfile)
col <- height.colors(50)

# Points-to-raster algorithm with a resolution of 1 meter
chm <- rasterize_canopy(las, res = 1, p2r())
plot(chm, col = col)

# Points-to-raster algorithm with a resolution of 0.5 meters replacing each
# point by a 20 cm radius circle of 8 points
chm <- rasterize_canopy(las, res = 0.5, p2r(0.2))
plot(chm, col = col)

\dontrun{
chm <- rasterize_canopy(las, res = 0.5, p2r(0.2, na.fill = tin()))
plot(chm, col = col)
}
}
\seealso{
Other digital surface model algorithms: 
\code{\link{dsm_pitfree}},
\code{\link{dsm_spikefree}},
\code{\link{dsm_tin}}
}
\concept{digital surface model algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algorithm-trk.R
\name{track_sensor_gatziolis2019}
\alias{track_sensor_gatziolis2019}
\alias{Gatziolis2019}
\title{Sensor tracking algorithm}
\usage{
Gatziolis2019(SEGLENFactor = 1.0059, AngleFactor = 0.8824, deltaT = 0.5)
}
\arguments{
\item{SEGLENFactor}{scalar. Weighting factor for the distance b/w 1st and last pulse returns}

\item{AngleFactor}{scalar. Weighting factor for view angle of mother pulse of a return}

\item{deltaT}{scalar. TimeBlock duration (in seconds)}
}
\description{
This function is made to be used in \link{track_sensor}. It implements an algorithm from Gatziolis
and McGaughey 2019 (see reference) for sensor tracking using multiple returns to estimate the positioning
of the sensor by computing the intersection in space of the lines passing through the first and
last returns.
}
\details{
In the original paper, two steps are described: (1) closest point approach (CPA) and (2) cubic
spline fitting. Technically, the cubic spline fitting step is a post-processing step and is not
included in this algorithm.\cr\cr
The source code of the algorithm is a slight modification of the original source code provided
with the paper to fit with the lidR package.
}
\examples{
# A valid file properly populated
LASfile <- system.file("extdata", "Topography.laz", package="lidR")
las = readLAS(LASfile)
flightlines <- track_sensor(las, Gatziolis2019())

plot(las@header)
plot(flightlines, add = TRUE)
}
\references{
Gatziolis, D., & McGaughey, R. J. (2019). Reconstructing Aircraft Trajectories from
Multi-Return Airborne Laser-Scanning Data. Remote Sensing, 11(19), 2258.
}
\author{
Demetrios Gaziolis and Jean-Romain Roussel
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/knn.R
\name{knn}
\alias{knn}
\alias{knnx}
\title{Search Nearest Neighbors}
\usage{
knn(data, k = 10)

knnx(data, query, k = 10)
}
\arguments{
\item{data}{LAS object for input point cloud}

\item{k}{number of nearest neighbors to search.}

\item{query}{LAS object for query locations}
}
\value{
a list contains:
\enumerate{
\item **nn.index** an n x k matrix for the nearest neighbor indice.
\item **nn.dist** an n x k matrix for the nearest neighbor Euclidean distances.
}
}
\description{
Fast parallelized k-neareast neighbor searching algorithms for point cloud in LAS format
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/engine_options.R
\name{engine_options}
\alias{engine_options}
\alias{opt_chunk_buffer}
\alias{opt_chunk_buffer<-}
\alias{opt_chunk_size}
\alias{opt_chunk_size<-}
\alias{opt_chunk_alignment}
\alias{opt_chunk_alignment<-}
\alias{opt_restart<-}
\alias{opt_progress}
\alias{opt_progress<-}
\alias{opt_stop_early}
\alias{opt_stop_early<-}
\alias{opt_wall_to_wall}
\alias{opt_wall_to_wall<-}
\alias{opt_independent_files}
\alias{opt_independent_files<-}
\alias{opt_output_files}
\alias{opt_output_files<-}
\alias{opt_laz_compression}
\alias{opt_laz_compression<-}
\alias{opt_merge}
\alias{opt_merge<-}
\alias{opt_select}
\alias{opt_select<-}
\alias{opt_filter}
\alias{opt_filter<-}
\title{Get or set LAScatalog processing engine options}
\usage{
opt_chunk_buffer(ctg)

opt_chunk_buffer(ctg) <- value

opt_chunk_size(ctg)

opt_chunk_size(ctg) <- value

opt_chunk_alignment(ctg)

opt_chunk_alignment(ctg) <- value

opt_restart(ctg) <- value

opt_progress(ctg)

opt_progress(ctg) <- value

opt_stop_early(ctg)

opt_stop_early(ctg) <- value

opt_wall_to_wall(ctg)

opt_wall_to_wall(ctg) <- value

opt_independent_files(ctg)

opt_independent_files(ctg) <- value

opt_output_files(ctg)

opt_output_files(ctg) <- value

opt_laz_compression(ctg)

opt_laz_compression(ctg) <- value

opt_merge(ctg)

opt_merge(ctg) <- value

opt_select(ctg)

opt_select(ctg) <- value

opt_filter(ctg)

opt_filter(ctg) <- value
}
\arguments{
\item{ctg}{An object of class \link[=LAScatalog-class]{LAScatalog}}

\item{value}{An appropriate value depending on the expected input.}
}
\description{
The names of the options and their roles are documented in \link[=LAScatalog-class]{LAScatalog}.
The options are used by all the functions that support a \code{LAScatalog} as input. Most options are
easy to understand and to link to the documentation of \link[=LAScatalog-class]{LAScatalog} but some
need more details. See section 'Details'.
}
\details{
\itemize{
\item \strong{opt_restart()} automatically sets the chunk option named "drop" in such a way that
the engine will restart at a given chunk and skip all previous chunks. Useful for restarting after a crash.
\item \strong{opt_independent_file()} automatically sets the chunk size, chunk buffer and wall-to-wall options
to process files that are not spatially related to each other, such as plot inventories.
\item \strong{opt_laz_compression()} automatically modifies the drivers to write LAZ files instead of LAS files.
}
}
\examples{
LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
ctg = readLAScatalog(LASfile)

plot(ctg, chunk_pattern = TRUE)

opt_chunk_size(ctg) <- 150
plot(ctg, chunk_pattern = TRUE)

opt_chunk_buffer(ctg) <- 10
plot(ctg, chunk_pattern = TRUE)

opt_chunk_alignment(ctg) <- c(270,250)
plot(ctg, chunk_pattern = TRUE)

summary(ctg)

opt_output_files(ctg) <- "/path/to/folder/templated_filename_{XBOTTOM}_{ID}"
summary(ctg)
}
\seealso{
Other LAScatalog processing engine: 
\code{\link{engine}}
}
\concept{LAScatalog processing engine}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algorithm-its.R
\name{its_watershed}
\alias{its_watershed}
\alias{watershed}
\title{Individual Tree Segmentation Algorithm}
\usage{
watershed(chm, th_tree = 2, tol = 1, ext = 1)
}
\arguments{
\item{chm}{`RasterLayer`, `SpatRaster` or `stars`. Canopy height model. Can be computed with \link{rasterize_canopy} or read from
an external file.}

\item{th_tree}{numeric. Threshold below which a pixel cannot be a tree. Default is 2.}

\item{tol}{numeric. Tolerance see ?EBImage::watershed.}

\item{ext}{numeric. see ?EBImage::watershed.}
}
\description{
This function is made to be used in \link{segment_trees}. It implements an algorithm for tree
segmentation based on a watershed. It is based on the bioconductor package \code{EBIimage}. You
need to install this package to run this method (see its \href{https://github.com/aoles/EBImage}{github page}).
Internally, the function EBImage::watershed is called.
}
\details{
Because this algorithm works on a CHM only there is no actual need for a point cloud. Sometimes the
user does not even have the point cloud that generated the CHM. \code{lidR} is a point cloud-oriented
library, which is why this algorithm must be used in \link{segment_trees} to merge the result into the point
cloud. However, the user can use this as a stand-alone function like this:
\preformatted{
 chm <- raster("chm.tif")
 crowns <- watershed(chm)()
}
}
\examples{
\dontrun{
LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
poi <- "-drop_z_below 0 -inside 481280 3812940 481320 3812980"
las <- readLAS(LASfile, select = "xyz", filter = poi)
col <- pastel.colors(250)

# Using raster because focal does not exist in stars
chm <- rasterize_canopy(las, res = 0.5, p2r(0.3), pkg = "raster")
ker <- matrix(1,3,3)
chm <- raster::focal(chm, w = ker, fun = mean, na.rm = TRUE)
las <- segment_trees(las, watershed(chm))

plot(las, color = "treeID", colorPalette = col)
}
}
\seealso{
Other individual tree segmentation algorithms: 
\code{\link{its_dalponte2016}},
\code{\link{its_li2012}},
\code{\link{its_silva2016}}

Other raster based tree segmentation algorithms: 
\code{\link{its_dalponte2016}},
\code{\link{its_silva2016}}
}
\concept{individual tree segmentation algorithms}
\concept{raster based tree segmentation algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/doc-drivers.R
\name{lidR-LAScatalog-drivers}
\alias{lidR-LAScatalog-drivers}
\title{LAScatalog drivers}
\description{
This document explains how objects are written on disk when processing a LAScatalog. As mentioned
in \link{LAScatalog-class}, users can set a templated filename to store the outputs on disk instead
of in R memory. By defaut \code{LAS} objects are stored in .las files with \link{writeLAS},
raster objects are stored in .tif files using native write function for \code{raster}, \code{stars} or \code{terra},
\verb{Spatial*} objects are stored in .shp files with after coercion to \code{sf} with \link[sf:st_write]{st_write},
\code{sf} object are stored to \code{.gpkg} files with \link[sf:st_write]{st_write}.\code{data.frame} objects are
stored in .csv files with \link[data.table:fwrite]{fwrite}, and other objects are not supported.
However, users can modify all these default settings and even add new drivers.
This manual page explain how. One may also refer to some unofficial documentation
\href{https://github.com/r-lidar/lidR/wiki/Modify-the-LAScatalog-drivers}{here} or
\href{https://gis.stackexchange.com/questions/325367/configuring-lidr-catalog-to-overwrite-raster-output-with-different-extension}{here}.
}
\section{Generic form of a driver}{

A driver is stored in the  slot \verb{@output_options} of a \code{LAScatalog}. It is a list that contains:
\describe{
\item{write}{A function that receives an object and a path, and writes the object into a file using
the path. The function can also have extra options.}
\item{extension}{A string that gives the file extension.}
\item{object}{A string that gives the name of the argument used to pass the object to write in the
function used to write the object.}
\item{path}{A string that gives the name of the argument used to pass the path of the file to write
in the function used to write the object.}
\item{param}{A labelled list of extra parameters for the function used to write the object}
}
For example, the driver to write a  \verb{Raster*} is
\preformatted{
list(
 write = raster::writeRaster,
 extension = ".tif",
 object = "x",
 path = "filename",
 param = list(format = "GTiff"))
}
And the driver to write a \code{LAS} is
\preformatted{
list(
 write = lidR::writeLAS,
 extension = ".las",
 object = "las",
 path = "file",
 param = list())
}
}

\section{Modify a driver (1/2)}{

Users can modify the drivers to write different file types than the default. For example, to write in
shapefile instead of a GeoPackage, one must change the \code{sf} driver:
\preformatted{
ctg@output_options$drivers$sf$extension <- ".shp"
}
To write a \verb{Raster*} in .grd files instead of .tif files one must change the \code{Raster} driver:
\preformatted{
ctg@output_options$drivers$Raster$extension <- ".grd"
ctg@output_options$drivers$Raster$param$format <- "raster"
}
To write in .laz files instead of .las files one must change the \code{LAS} driver:
\preformatted{
ctg@output_options$drivers$LAS$extension <- ".laz"
}
}

\section{Add a new driver}{

The drivers allow \code{LAS}, \code{Spatial},\verb{sf,} \code{Raster}, \code{stars}, \code{SpatRaster} and \code{data.frame} objects
to be written. When using the engine (\link{catalog_apply}) to build new tools, users may need to
be able to write other objects such as a \code{list}. To do that users need to add a \code{list} element
into \verb{@output_options}:
\preformatted{
ctg@output_options$drivers$list = list(
 write = base::saveRDS,
 object = "object",
 path = "file",
 extension = ".rds",
 param = list(compress = TRUE))
}
The \code{LAScatalog} now has a new driver capable of writing a \code{list}.
}

\section{Modify a driver (2/2)}{

It is also possible to completely overwrite an existing driver. By default \code{sf} objects are written
into GeoPackage with \link[sf:st_write]{st_write}. \code{st_write} can also wite in GeoJSON and even in
SQLlite database objects. But it cannot add data into an existing SQLlite database. Let's create
our own driver for a \code{sf}. First we need a function able to write and append a \code{sf} into a \code{SQLlite}
database from the object and the path.
\preformatted{
dbWrite_sf = function(x, path, name)
{
 x <- sf::st_drop_geometry(x)
 con <- RSQLite::dbConnect(RSQLite::SQLite(), path)
 RSQLite::dbWriteTable(con, name, x, append = TRUE)
 RSQLite::dbDisconnect(con)
}}
Then we create the driver. User-defined drivers supersede default drivers:
\preformatted{
ctg@output_options$drivers$sf = list(
 write = dbWrite_sf,
 extension = ".sqlite",
 object = "x",
 path = "path",
 param = list(name = "layername"))
}
Then to be sure that we do not write several .sqlite files, we don't use templated filename.
\preformatted{
opt_output_files(ctg) <- paste0(tempdir(), "/mysqlitefile")}
And all the \code{sf} will be appended in a single database. To preserve the geometry one can
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_attribute.R
\name{add_attribute}
\alias{add_attribute}
\alias{add_lasattribute}
\alias{add_lasattribute_manual}
\alias{add_lasrgb}
\alias{add_lasnir}
\alias{remove_lasattribute}
\title{Add attributes into a LAS object}
\usage{
add_attribute(las, x, name)

add_lasattribute(las, x, name, desc)

add_lasattribute_manual(
  las,
  x,
  name,
  desc,
  type,
  offset = NULL,
  scale = NULL,
  NA_value = NULL
)

add_lasrgb(las, R, G, B)

add_lasnir(las, NIR)

remove_lasattribute(las, name)
}
\arguments{
\item{las}{An object of class \link[=LAS-class]{LAS}}

\item{x}{a vector that needs to be added in the LAS object. For \code{add_lasattribute*} it can
be missing (see details).}

\item{name}{character. The name of the extra bytes attribute to add in the file.}

\item{desc}{character. A short description of the extra bytes attribute to add in the file (32 characters).}

\item{type}{character. The data type of the extra bytes attribute. Can be "uchar", "char", "ushort",
"short", "uint", "int", "uint64", "int64", "float", "double".}

\item{scale, offset}{numeric. The scale and offset of the data. NULL if not relevant.}

\item{NA_value}{numeric or integer. NA is not a valid value in a las file. At time of writing it will
be replaced by this value that will be considered as NA. NULL if not relevant.}

\item{R, G, B, NIR}{integer. RGB and NIR values. Values are automatically scaled to 16 bits if they are
coded on 8 bits (0 to 255).}
}
\value{
An object of class \link[=LAS-class]{LAS}
}
\description{
A \link[=LAS-class]{LAS} object represents a las file in R. According to the
\href{https://community.asprs.org/leadership-restricted/leadership-content/public-documents/standards}{LAS specifications}
a las file contains a core of defined attributes, such as XYZ coordinates, intensity, return number,
and so on, for each point. It is possible to add supplementary attributes.
}
\details{
Users cannot assign names that are the same as the names of the core attributes. These functions are
dedicated to adding data that are not part of the LAS specification. For example, \code{add_lasattribute(las, x, "R")}
will fail because \code{R} is a name reserved for the red channel of a .las file that contains RGB
attributes. Use \code{add_lasrgb} instead.
\describe{
\item{\code{add_attribute}}{Simply adds a new column in the data but does not update the header. Thus the LAS
object is not strictly valid. These data will be temporarily usable at the R level but will not
be written in a las file with \link{writeLAS}.}

\item{ \code{add_lasattribute}}{Does the same as \code{add_attribute} but automatically updates the header of
the LAS object. Thus, the LAS object is valid and the new data is considered as "extra bytes". This new
data will be written in a las file with \link{writeLAS}.}

\item{\code{add_lasattribute_manual}}{Allows the user to manually write all the extra bytes metadata.
This function is reserved for experienced users with a good knowledge of the LAS specifications.
The function does not perform tests to check the validity of the information.
When using \code{add_lasattribute} and \code{add_lasattribute_manual}, \code{x} can only be of type numeric,
(\code{integer} or \code{double}). It cannot be of type \code{character} or \code{logical} as these are
not supported by the LAS specifications. The types that are supported in lidR are types 0 to 10
(Table 24 on page 25 of the specification). Types greater than 10 are not supported.}

\item{\code{add_lasrgb}}{Adds 3 columns named RGB and updates the point format of the LAS object
for a format that supports RGB attributes. If the RGB values are ranging from 0 to 255 they are
automatically scaled on 16 bits.}
}
}
\examples{
LASfile <- system.file("extdata", "example.laz", package="rlas")
las <- readLAS(LASfile, select = "xyz")

print(las)
print(header(las))

x <- 1:30

las <- add_attribute(las, x, "mydata")
print(las)        # The las object has a new attribute called "mydata"
print(header(las)) # But the header has not been updated. This new data will not be written

las <- add_lasattribute(las, x, "mydata2", "A new data")
print(las)        # The las object has a new attribute called "mydata2"
print(header(las)) # The header has been updated. This new data will be written

# Optionally if the data is already in the LAS object you can update the header skipping the
# parameter x
las <- add_attribute(las, x, "newattr")
las <- add_lasattribute(las, name = "newattr", desc = "Amplitude")
print(header(las))

# Remove an extra bytes attribute
las <- remove_lasattribute(las, "mydata2")
print(las)
print(header(las))

las <- remove_lasattribute(las, "mydata")
print(las)
print(header(las))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io_readLAScatalog.R
\name{readLAScatalog}
\alias{readLAScatalog}
\alias{readALScatalog}
\alias{readTLScatalog}
\alias{catalog}
\title{Create an object of class LAScatalog}
\usage{
readLAScatalog(
  folder,
  progress = TRUE,
  select = "*",
  filter = "",
  chunk_size = 0,
  chunk_buffer = 30,
  ...
)

readALScatalog(folder, ...)

readTLScatalog(folder, ...)

catalog(folder, ...)
}
\arguments{
\item{folder}{string. The path of a folder containing a set of las/laz files.
Can also be a vector of file paths.}

\item{progress, select, filter, chunk_size, chunk_buffer}{Easily accessible processing
options tuning. See \link{LAScatalog-class} and \link{engine_options}.}

\item{\dots}{Extra parameters to \link[base:list.files]{list.files}. Typically
\code{recursive = TRUE}. Propagates also to \code{readLAScatalog}}
}
\value{
A \code{LAScatalog} object
}
\description{
Create an object of class \link[=LAScatalog-class]{LAScatalog} from a folder
or a collection of filenames. A LAScatalog is a representation of a collection
of las/laz files. A computer cannot load all the data at once. A \code{LAScatalog}
is a simple way to manage all the files sequentially. Most functions from
\code{lidR} can be used seamlessly with a LAScatalog using the internal
\code{LAScatalog} processing engine. To take advantage of the \code{LAScatalog}
processing engine the user must first adjust some processing options using the
\link[=engine_options]{appropriate functions}. Careful reading of the
\link[=LAScatalog-class]{LAScatalog class documentation} is required to use the
\code{LAScatalog} class correctly.\cr\cr
\code{readLAScatalog} is the original function and always works. Using one of the \code{read*LAScatalog} functions
adds information to the returned object to register a point-cloud type. Registering the correct point
type \strong{may} improve the performance of some functions by enabling users to select an appropriate spatial index.
See \link[=lidR-spatial-index]{spatial indexing}. Notice that by legacy and for backwards-compatibility
reasons \code{readLAScatalog()} and \code{readALSLAScatalog()} are equivalent because lidR was originally designed
for ALS and thus the original function \code{readLAScatalog()} was (supposedly) used for ALS.
}
\examples{
# A single file LAScatalog using data provided with the package
LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
ctg = readLAScatalog(LASfile)
plot(ctg)

\dontrun{
ctg <- readLAScatalog("</path/to/folder/of/las/>")

# Internal engine will sequentially process chunks of size 500 x 500 m
opt_chunk_size(ctg) <- 500

# Internal engine will align the 500 x 500 m chunks on x = 250 and y = 300
opt_alignment(ctg) <- c(250, 300)

# Internal engine will not display a progress estimation
opt_progress(ctg) <- FALSE

# Internal engine will not return results into R.
# Instead it will write results in files.
# Files will be named e.g.
# filename_256000_1.ext
# filename_257000_2.ext
# filename_258000_3.ext
# ...
opt_output_files(ctg) <- "/path/filename_{XBOTTOM}_{ID}"

# More details in the documentation
help("LAScatalog-class", "lidR")
help("engine_options", "lidR")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/backward_compatibility.R
\name{old_spatial_packages}
\alias{old_spatial_packages}
\alias{as.spatial}
\alias{as.spatial.LAS}
\alias{as.spatial.LAScatalog}
\alias{tree_metrics}
\alias{grid_canopy}
\alias{grid_density}
\alias{grid_terrain}
\alias{grid_metrics}
\alias{find_trees}
\alias{delineate_crowns}
\title{Older R Spatial Packages}
\usage{
as.spatial(x)

\method{as.spatial}{LAS}(x)

\method{as.spatial}{LAScatalog}(x)

tree_metrics(las, func = ~list(Z = max(Z)), attribute = "treeID", ...)

grid_canopy(las, res, algorithm)

grid_density(las, res = 4)

grid_terrain(
  las,
  res = 1,
  algorithm,
  ...,
  keep_lowest = FALSE,
  full_raster = FALSE,
  use_class = c(2L, 9L),
  Wdegenerated = TRUE,
  is_concave = FALSE
)

grid_metrics(
  las,
  func,
  res = 20,
  start = c(0, 0),
  filter = NULL,
  by_echo = "all"
)

find_trees(las, algorithm, uniqueness = "incremental")

delineate_crowns(
  las,
  type = c("convex", "concave", "bbox"),
  concavity = 3,
  length_threshold = 0,
  func = NULL,
  attribute = "treeID"
)
}
\arguments{
\item{x, las}{an object of class LAS*}

\item{func}{see \link{template_metrics}}

\item{attribute, type}{see \link{crown_metrics}}

\item{...}{ignored}

\item{res, start}{see \link{pixel_metrics}}

\item{algorithm}{see \link{rasterize_canopy}, \link{rasterize_terrain}}

\item{full_raster, use_class, Wdegenerated, is_concave, keep_lowest}{see \link{rasterize_density}}

\item{filter, by_echo}{see \link{template_metrics}}

\item{uniqueness}{see \link{crown_metrics}}

\item{concavity, length_threshold}{see \link{concaveman}}
}
\description{
lidR 4.0.0 no longer uses the \code{sp} and \code{raster} packages. New functions are based on \code{sf}, \code{terra} and \code{stars}.
However, to maintain backward compatibility the old functions from v<4.0.0 were preserved.\cr\cr
\code{rgdal} and \code{rgeos} will be retired on Jan 1st 2024. The \code{raster} and \code{sp} packages are based on
\code{rgdal} and \code{rgeos}. \code{lidR} was based on \code{raster} and \code{sp} because it was created before the \code{sf}, \code{terra}
and \code{stars} packages. This means that sooner or later users and packages that are still based on
old R spatial packages will run into trouble. According to Edzer Pebesma, Roger Bivand: \cr\cr
\emph{R users who have been around a bit longer, in particular before packages like \code{sf} and \code{stars} were
developed, may be more familiar with older packages like \code{maptools}, \code{sp}, \code{rgeos}, and \code{rgdal}. A fair
question is whether they should migrate existing code and/or existing R packages depending on these
packages. The answer is: yes (see reference).}\cr\cr
The following functions are not formally deprecated but users should definitely move their workflow to modern
spatial packages. lidR will maintain the old functions as long as it does not generate issues
on CRAN. So, it might be until Jan 1st 2024 or later, who knows...
}
\references{
Edzer Pebesma, Roger Bivand Spatial Data Science with applications in R
https://keen-swartz-3146c4.netlify.app/older.html
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algorithm-itd.R
\name{itd_manual}
\alias{itd_manual}
\alias{manual}
\title{Individual Tree Detection Algorithm}
\usage{
manual(detected = NULL, radius = 0.5, color = "red", button = "middle", ...)
}
\arguments{
\item{detected}{\code{SpatialPoints* or }sf/sfc_POINT` with  2 or 3D points of already found tree tops
that need manual correction. Can be NULL}

\item{radius}{numeric. Radius of the spheres displayed on the point cloud (aesthetic purposes only).}

\item{color}{character. Colour of the spheres displayed on the point cloud (aesthetic purposes only).}

\item{button}{Which button to use for selection. One of "left", "middle", "right". lidR using left
for rotation and right for dragging using one of left or right will disable either rotation or dragging}

\item{...}{supplementary parameters to be passed to \link{plot}.}
}
\description{
This function is made to be used in \link{locate_trees}. It implements an algorithm for manual
tree detection. Users can pinpoint the tree top positions manually and interactively using the mouse.
This is only suitable for small-sized plots. First the point cloud is displayed, then the user is
invited to select a rectangular region of interest in the scene using the mouse button.
Within the selected region the highest point will be flagged as 'tree top' in the scene. Once all the trees
are labelled the user can exit the tool by selecting an empty region. Points can also be unflagged.
The goal of this tool is mainly for minor correction of automatically-detected tree outputs. \cr
\strong{This algorithm does not preserve tree IDs from \code{detected} and renumber all trees. It also looses
all attributes}
}
\examples{
\dontrun{
LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
las = readLAS(LASfile)

# Full manual tree detection
ttops = locate_trees(las, manual())

# Automatic detection with manual correction
ttops = locate_trees(las, lmf(5))
ttops = locate_trees(las, manual(ttops))
}
}
\seealso{
Other individual tree detection algorithms: 
\code{\link{itd_lmf}}
}
\concept{individual tree detection algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods-LASheader.R
\name{LASheader}
\alias{LASheader}
\title{Create a \code{LASheader} object}
\usage{
LASheader(data = list())
}
\arguments{
\item{data}{a list containing the data from the header of a LAS file. Can also be
a \code{data.frame} or \code{data.table}}
}
\value{
An object of class \code{LASheader}
}
\description{
Creates a  \code{LASheader} object either from a raw \code{list} containing all the
elements named according to the \code{rlas} package or creates a header from a \code{data.frame}
or \code{data.table} containing a point cloud. In the latter case it will generate a header
according to the data using \link[rlas:header_create]{rlas::header_create()}. It will
guess the LAS file format, the point data format, and initialize the scale factors and offsets,
but these may not suit a user's needs. Users are advised to
manually modify the results to fit their specific needs.
}
\examples{
data = data.frame(X = c(339002.889, 339002.983, 339002.918),
                  Y = c(5248000.515, 5248000.478, 5248000.318),
                  Z = c(975.589, 974.778, 974.471),
                  gpstime = c(269347.28141, 269347.28142, 269347.28143),
                  Intensity = c(82L, 54L, 27L),
                  ReturnNumber = c(1L, 1L, 2L),
                  NumberOfReturns = c(1L, 1L, 2L),
                  ScanDirectionFlag = c(1L, 1L, 1L),
                  EdgeOfFlightline = c(1L, 0L, 0L),
                  Classification = c(1L, 1L, 1L),
                  ScanAngleRank = c(-21L, -21L, -21L),
                  UserData = c(32L, 32L, 32L),
                  PointSourceID = c(17L, 17L, 17L))

header = LASheader(data)
header

# Record an EPSG code
epsg(header) <- 32618
header

las <- LAS(data, header)
las

# The function inferred a LAS 1.2 format 1 which is correct
# Upgrade to LAS 1.4 for the example
header@VLR <- list() # Erase VLR previously written
header@PHB[["Global Encoding"]][["WKT"]] <- TRUE
header@PHB[["Version Minor"]] <- 4L
header@PHB[["Header Size"]] <- 375L
header@PHB[["Offset to point data"]] <- 375L
wkt(header) <- sf::st_crs("EPSG:32618")$wkt
header
las1.4 <- LAS(data, header)
las1.4
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated.R
\name{deprecated}
\alias{deprecated}
\alias{readALSLAS}
\alias{readTLSLAS}
\alias{readUAVLAS}
\alias{readDAPLAS}
\alias{readALSLAScatalog}
\alias{readTLSLAScatalog}
\alias{readUAVLAScatalog}
\alias{readDAPLAScatalog}
\alias{filter_surfacepoints}
\alias{filter_surfacepoints.LAS}
\alias{filter_surfacepoints.LAScatalog}
\title{Deprecated functions in lidR}
\usage{
readALSLAS(files, select = "*", filter = "")

readTLSLAS(files, select = "*", filter = "", sort = TRUE)

readUAVLAS(files, select = "*", filter = "", sort = TRUE)

readDAPLAS(files, select = "*", filter = "", sort = TRUE)

readALSLAScatalog(folder, ...)

readTLSLAScatalog(folder, ...)

readUAVLAScatalog(folder, ...)

readDAPLAScatalog(folder, ...)

filter_surfacepoints(las, res)

\method{filter_surfacepoints}{LAS}(las, res)

\method{filter_surfacepoints}{LAScatalog}(las, res)
}
\arguments{
\item{files, select, filter, sort}{See the new functions that replace the old ones}

\item{folder, ...}{See the new functions that replace the old ones}

\item{las, res}{See the new functions that replace the old ones}
}
\description{
These functions are provided for compatibility with older versions of lidR but are deprecated.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/catalog_apply.R
\name{catalog_apply}
\alias{catalog_apply}
\alias{catalog_sapply}
\alias{catalog_map}
\title{LAScatalog processing engine}
\usage{
catalog_apply(ctg, FUN, ..., .options = NULL)

catalog_sapply(ctg, FUN, ..., .options = NULL)

catalog_map(ctg, FUN, ..., .options = NULL)
}
\arguments{
\item{ctg}{A \link[=LAScatalog-class]{LAScatalog} object.}

\item{FUN}{A user-defined function that respects a given template (see section function template)}

\item{...}{Optional arguments to FUN.}

\item{.options}{See dedicated section and examples.}
}
\description{
This function gives users access to the \link[=LAScatalog-class]{LAScatalog} processing engine.
It allows the application of a user-defined routine over a collection of LAS/LAZ files. The
LAScatalog processing engine is explained in the \link[=LAScatalog-class]{LAScatalog class}\cr\cr
\code{catalog_apply()} is the core of the lidR package. It drives every single function that can process a
\code{LAScatalog}. It is flexible and powerful but also complex. \code{catalog_map()} is a simplified version
of \code{catalog_apply()} that suits for 90\% of use cases.\cr\cr
\code{catalog_sapply()} is a previous attempt to provide simplified version of \code{catalog_apply()}. Use
\code{catalog_map()} instead.
}
\section{Edge artifacts}{


It is important to take precautions to avoid 'edge artifacts' when processing wall-to-wall
tiles. If the points from neighbouring tiles are not included during certain processes,
this could create 'edge artifacts' at the tile boundaries. For example, empty or incomplete
pixels in a rasterization process, or dummy elevations in a ground interpolation. The LAScatalog
processing engine provides internal tools to load buffered data 'on-the-fly'. \code{catalog_map()} takes
care of removing automatically the results computed in the buffered area to avoid unexpected output
with duplicated entries or conflict between values computed twice. It does that in predefined way
that may not suit all cases. \code{catalog_apply()} does not remove the buffer and leave users free
to handle this in a custom way. This is why \code{catalog_apply()} is more complex but gives more freedom
to build new applications.
}

\section{Buffered data}{


The LAS objects loaded in theses functions have a special attribute called 'buffer' that indicates,
for each point, if it comes from a buffered area or not. Points from non-buffered areas have a
'buffer' value of 0, while points from buffered areas have a 'buffer' value of 1, 2, 3 or 4, where
1 is the bottom buffer and 2, 3 and 4 are the left, top and right buffers, respectively. This allows
for filtering of buffer points if required.
}

\section{Function template}{


The parameter \code{FUN} of \code{catalog_apply} expects a function with a first argument that will be
supplied automatically by the \code{LAScatalog} processing engine. This first argument is a \code{LAScluster}.
A \code{LAScluster} is an internal undocumented class but the user needs to know only three things about
this class:
\itemize{
\item It represents a chunk of the file collection
\item The function \link{readLAS} can be used with a \code{LAScluster}
\item The function \link[raster:extent]{extent} or \link[sf:st_bbox]{st_bbox}
can be used with a \code{LAScluster} and they return the bounding box of the chunk without the buffer.
It must be used to clip the output and remove the buffered region (see examples).
}

A user-defined function must be templated like this:

\if{html}{\out{<div class="sourceCode">}}\preformatted{myfun <- function(chunk, ...) \{
   # Load the chunk + buffer
   las <- readLAS(chunk)
   if (is.empty(las)) return(NULL)

   # do something
   output <- do_something(las, ...)

   # remove the buffer of the output
   bbox <- bbox(chunk)
   output <- remove_buffer(output, bbox)
   return(output)
\}
}\if{html}{\out{</div>}}

The line \code{if (is.empty(las)) return(NULL)} is important because some clusters (chunks) may contain
0 points (we can't know this before reading the file). In this case an empty point cloud with 0 points
is returned by \code{readLAS()} and this may fail in subsequent code. Thus, exiting early from the user-defined
function by returning \code{NULL} indicates to the internal engine that the chunk was empty.

\code{catalog_map} is much simpler (but less versatile). It automatically takes care of reading the chunk
and checks if the point cloud is empty. It also automatically crop the buffer. The way it crops the
buffer suits for most cases but for some special cases it may be advised to handle this in a more
specific way i.e. using \code{catalog_apply()}. For \code{catalog_map()} the first argument is a \code{LAS} and the
template is:

\if{html}{\out{<div class="sourceCode">}}\preformatted{myfun <- function(las, ...) \{
   # do something
   output <- do_something(las, ...)
   return(output)
\}
}\if{html}{\out{</div>}}
}

\section{.options}{

Users may have noticed that some lidR functions throw an error when the processing options are
inappropriate. For example, some functions need a buffer and thus \code{buffer = 0} is forbidden.
Users can add the same constraints to protect against inappropriate options. The \code{.options}
argument is a \code{list} that allows users to tune the behaviour of the processing engine.
\itemize{
\item \code{drop_null = FALSE}: not intended to be used by regular users. The engine does not remove
NULL outputs
\item \code{need_buffer = TRUE}: the function complains if the buffer is 0.
\item \code{need_output_file = TRUE} the function complains if no output file template is provided.
\item \code{raster_alignment = ...} the function checks the alignment of the chunks. This option is
important if the output is a raster. See below for more details.
\item \code{automerge = TRUE} by default the engine returns a \code{list} with one item per chunk. If
\code{automerge = TRUE}, it tries to merge the outputs into a single object: a \verb{Raster*}, a
\verb{Spatial*}, a \verb{LAS*}, an \code{sf}, a \code{ stars} similarly to other functions of the package. This is a
fail-safe option so in the worst case, if the merging fails, the \code{list} is returned. This is
activated by \code{catalog_map} making it simpler.
\item \code{autoread = TRUE}. Introduced in v3.0.0 this option enables to get rid of the first steps of the
function i.e \code{readLAS()} and \verb{if (is.empty())}. In this case the function must take two
objects as input, first a \code{LAS} object and second a \code{bbox} from \code{sf}. This is
activated by \code{catalog_map} making it simpler.
\item \code{autocrop = TRUE}. Introduced in v4.0.0 this option enables to get rid of the buffer crop steps
of the function i.e \code{something <- remove_buffer(something, bbox)}. In this case the function must
take one \code{LAS} object as input. This is activated by \code{catalog_map} making it simpler.
}

When the function \code{FUN} returns a raster it is important to ensure that the chunks are aligned
with the raster to avoid edge artifacts. Indeed, if the edge of a chunk does not correspond to the edge
of the pixels, the output will not be strictly continuous and will have edge artifacts (that might
not be visible). Users can check this with the options \code{raster_alignment}, that can take the
resolution of the raster as input, as well as the starting point if needed. The following are accepted:

\if{html}{\out{<div class="sourceCode">}}\preformatted{# check if chunks are aligned with a raster of resolution 20
raster_alignment = 20
raster_alignment = list(res = 20)

# check if chunks are aligned with a raster of resolution 20
# that starts at (0,10)
raster_alignment = list(res = 20, start = c(0,10))
}\if{html}{\out{</div>}}
}

\examples{
# More examples might be avaible in the official lidR vignettes or
# on the github book <https://jean-romain.github.io/lidRbook/>

## =========================================================================
## Example 1: detect all the tree tops over an entire catalog
## (this is basically a reproduction of the existing function 'locate_trees')
## =========================================================================

# 1. Build the user-defined function that analyzes each chunk of the catalog.
# The function's first argument is a LAScluster object. The other arguments can be freely
# chosen by the users.
my_tree_detection_method <- function(chunk, ws)
{
  # The chunk argument is a LAScluster object. The users do not need to know how it works.
  # readLAS will load the region of interest (chunk) with a buffer around it, taking advantage of
  # point cloud indexation if possible. The filter and select options are propagated automatically
  las <- readLAS(chunk)
  if (is.empty(las)) return(NULL)

  # Find the tree tops using a user-developed method
  # (here simply a LMF for the example).
  ttops <- locate_trees(las, lmf(ws))

  # ttops is an sf object that contains the tree tops in our region of interest
  # plus the trees tops in the buffered area. We need to remove the buffer otherwise we will get
  # some trees more than once.
  bbox  <- st_bbox(chunk)
  ttops <- sf::st_crop(ttops, bbox)
  return(ttops)
}

# 2. Build a collection of file
# (here, a single file LAScatalog for the purposes of this simple example).
LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
ctg <- readLAScatalog(LASfile)
plot(ctg)

# 3. Set some processing options.
# For this small single file example, the chunk size is 100 m + 10 m of buffer
opt_chunk_buffer(ctg) <- 10
opt_chunk_size(ctg)   <- 100            # Small because this is a dummy example.
opt_chunk_alignment(ctg) <- c(-50, -35) # Align such as it creates 2 chunks only.
opt_select(ctg)       <- "xyz"          # Read only the coordinates.
opt_filter(ctg)       <- "-keep_first"  # Read only first returns.

# 4. Apply a user-defined function to take advantage of the internal engine
opt    <- list(need_buffer = TRUE,   # catalog_apply will throw an error if buffer = 0
               automerge   = TRUE)   # catalog_apply will merge the outputs into a single object
output <- catalog_apply(ctg, my_tree_detection_method, ws = 5, .options = opt)

plot(output)

\donttest{
## =========================================================================
## Example 1: simplified. There is nothing that requires special data
## manipulation in the previous example. Everything can be handled automatically
##=========================================================================

# 1. Build the user-defined function that analyzes a point cloud.
my_tree_detection_method <- function(las, ws)
{
  # Find the tree tops using a user-developed method
  # (here simply a LMF for the example).
  ttops <- locate_trees(las, lmf(ws))
  return(ttops)
}

# 2. Build a project
LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
ctg <- readLAScatalog(LASfile)
plot(ctg)

# 3. Set some processing options.
# For this dummy example, the chunk size is 100 m and the buffer is 10 m
opt_chunk_buffer(ctg) <- 10
opt_chunk_size(ctg)   <- 100            # small because this is a dummy example.
opt_chunk_alignment(ctg) <- c(-50, -35) # Align such as it creates 2 chunks only.
opt_select(ctg)       <- "xyz"          # Read only the coordinates.
opt_filter(ctg)       <- "-keep_first"  # Read only first returns.

# 4. Apply a user-defined function to take advantage of the internal engine
opt    <- list(need_buffer = TRUE)   # catalog_apply will throw an error if buffer = 0
output <- catalog_map(ctg, my_tree_detection_method, ws = 5, .options = opt)
}

\dontrun{
## ===================================================
## Example 2: compute a rumple index on surface points
## ===================================================

rumple_index_surface = function(las, res)
{
  las    <- filter_surfacepoints(las, 1)
  rumple <- pixel_metrics(las, ~rumple_index(X,Y,Z), res)
  return(rumple)
}

LASfile <- system.file("extdata", "Megaplot.laz", package="lidR")
ctg <- readLAScatalog(LASfile)

opt_chunk_buffer(ctg) <- 1
opt_chunk_size(ctg)   <- 140     # small because this is a dummy example.
opt_select(ctg)       <- "xyz"   # read only the coordinates.

opt    <- list(raster_alignment = 20) # catalog_apply will adjust the chunks if required
output <- catalog_map(ctg, rumple_index_surface, res = 20, .options = opt)

plot(output, col = height.colors(25))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/doc-spatialindex.R, R/utils_spatial_index.R
\name{lidR-spatial-index}
\alias{lidR-spatial-index}
\alias{sensor}
\alias{sensor<-}
\alias{index}
\alias{index<-}
\title{Spatial index}
\usage{
sensor(las, h = FALSE)

sensor(las) <- value

index(las, h = FALSE)

index(las) <- value
}
\arguments{
\item{las}{An object of class \link[lidR:LAS-class]{LAS} or \link[lidR:LAScatalog-class]{LAScatalog}.}

\item{h}{boolean. Human readable. Everything is stored as integers that are understood
internally. Use \code{h = TRUE} for user readable output.}

\item{value}{integer or character. A code for referring to a sensor type or a spatial
index type. Use one of \code{"unknown"}, \code{"als"}, \code{"tls"}, \code{"uav"}, \code{"dap"}, \code{"multispectral"}
for sensor type and one of \code{"auto"}, \code{"gridpartition"}, \code{"voxelpartition"}, \code{"quadtree"}, \code{"octree"}
for spatial index type.}
}
\description{
This document explains how to process point-clouds taking advantage of different spatial
indexes available in the lidR package. lidR can use several types of spatial indexes to
apply algorithms (that need a spatial indexing) as fast as possible. The choice of the spatial
index depends on the type of point-cloud that is processed and the algorithm that is performed.
lidR can use a grid partition, a voxel partition, a quadtree or an octree. See details.
}
\details{
From lidR (>= 3.1.0), a \link[=LAS-class]{LAS} object records the sensor used to sample
the point-cloud (ALS, TLS, UAV, DAP) as well as the spatial index that must be used
for processing the point cloud. This can be set manually by the user but the simplest is
to use one of the \link[=readLAS]{read*LAS()} functions. By default a point-cloud is associated
to a sensor and the best spatial index is chosen on-the-fly depending on the algorithm
applied. It is possible to force the use of a specific spatial index.

Information relative to the spatial indexing is stored in slot \verb{@index} that contains
a \code{list} with two elements:
\itemize{
\item \code{sensor}: an integer that records the sensor type
\item \code{index}: an integer that records the spatial index to be used
}

By default the spatial index code is 0 ("automatic") meaning that each function is free
to choose a different spatial index depending on the recorded sensor. If the code is not
0 then each function will be forced to used the spatial index that is imposed. This,
obviously, applies only to functions that use spatial indexing.

\link[=LAScatalog-class]{LAScatalog} objects also record such information that is automatically
propagated to the LAS objects when processing.

Note: before version 3.1.0, point-clouds were all considered as ALS because lidR was originally
designed for ALS. Consequently, for legacy and backwards-compatibility reasons, \code{readLAS()}
and \code{readALSLAS()} are actually equivalent. \code{readLAS()} tags the point cloud with "unknown"
sensor while \code{readALSLAS()} tags it with 'ALS'. Both behave the same and this is
especially true compared with versions < 3.1. As a consequence, using \code{readLAS()} provides
the same performance (no degradation) than in previous versions, while using one of the \code{read*LAS()}
functions may improve the performance.
}
\examples{
LASfile <- system.file("extdata", "example.laz", package="rlas")
las <- readLAS(LASfile)

# By default the sensor and spatial index codes are 0
sensor(las)
index(las)

# Codes are used internally and not intended to be known by users
# Use h option for human readable output
sensor(las, h = TRUE)
index(las, h = TRUE)

# Modification of the sensor enables users to select a better spatial index
# when processing the point-cloud.
sensor(las) <- "tls"
sensor(las, h = TRUE)
index(las, h = TRUE)

# Modification of the spatial index forces users to choose one of the available
# spatial indexes.
index(las) <- "quadtree"
sensor(las, h = TRUE)
index(las, h = TRUE)

# The simplest way to take advantage of appropriate spatial indexing is
# to use one of the read*LAS() functions.
las <- readTLSLAS(LASfile)
sensor(las, h = TRUE)
index(las, h = TRUE)

# But for some specific point-clouds / algorithms it might be advisable to force
# the use of a specific spatial index to perform the computation faster
index(las) <- "voxelpartition"
index(las, h = TRUE)

# With a LAScatalog, spatial indexing information is propagated to the
# different chunks
ctg = readTLSLAScatalog(LASfile)
index(ctg) <- "voxelpartition"
sensor(ctg, h = TRUE)
index(ctg, h = TRUE)

# ==================
# PERFORMANCE TESTS
# ==================

\dontrun{
# Performance tests on TLS
# ------------------------

# The package does not include TLS data
# so we can generate something that looks TLS-ish
# >>>>>>>>>>>
X <- runif(50, -25, 25)
Y <- runif(50, -25, 25)
X <- as.numeric(sapply(X, function(x) rnorm(2000, x, 2)))
Y <- as.numeric(sapply(Y, function(x) rnorm(2000, x, 2)))
Z <- abs(rnorm(length(Y), 10, 5))
veg <- data.frame(X,Y,Z)
X <- runif(5000, -30, 30)
Y <- runif(5000, -30, 30)
Z <- runif(5000, 0, 1)
ground <- data.frame(X,Y,Z)
X <- runif(30, -30, 30)
Y <- runif(30, -30, 30)
Z <- runif(30, 0, 30)
noise <- data.frame(X,Y,Z)
las <- LAS(rbind(ground, veg, noise))
# <<<<<<<<<<<<<

plot(las)

# If read with readALSLAS()
sensor(las) <- "als"
system.time(classify_noise(las, sor(20, 8)))
#> 1.5 sec

# If read with readTLSLAS()
sensor(las) <- "tls"
system.time(classify_noise(las, sor(20, 8)))
#> 0.6 sec

# Performance tests on ALS
# ------------------------

# The package does not include large ALS data
# so we can generate something that looks ALS-ish
# >>>>>>>>>>>
X <- runif(4e5, 0, 1000)
Y <- runif(4e5, 0, 1000)
Z <- 40*sin(0.01*X) + 50*cos(0.005*Y) + abs(rnorm(length(Y), 10, 5))
veg <- data.frame(X,Y,Z)
X <- runif(100, 0, 1000)
Y <- runif(100, 0, 1000)
Z <- 40*sin(0.01*X) + 50*cos(0.005*Y) + abs(rnorm(length(Y), 10, 5)) + runif(100, 30, 70)
noise <- data.frame(X,Y,Z)
las <- LAS(rbind(veg, noise))
# <<<<<<<<<<<<<

plot(las)

# If read with readALSLAS()
sensor(las) <- "als"
system.time(classify_noise(las, sor(15, 8)))
#> 3 sec

# If read with readTLSLAS()
sensor(las) <- "tls"
system.time(classify_noise(las, sor(15, 8)))
#> 4.3 sec
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugins.R
\name{plugins}
\alias{plugins}
\alias{plugin_dsm}
\alias{plugin_dtm}
\alias{plugin_gnd}
\alias{plugin_decimate}
\alias{plugin_shape}
\alias{plugin_snag}
\alias{plugin_track}
\alias{plugin_nintensity}
\alias{plugin_outliers}
\alias{plugin_itd}
\alias{plugin_its}
\title{Plugin system}
\usage{
plugin_dsm(f, omp = FALSE)

plugin_dtm(f, omp = FALSE)

plugin_gnd(f, omp = FALSE)

plugin_decimate(f, omp = FALSE)

plugin_shape(f, omp = FALSE)

plugin_snag(f, omp = FALSE)

plugin_track(f, omp = FALSE)

plugin_nintensity(f, omp = FALSE)

plugin_outliers(f, omp = FALSE)

plugin_itd(f, omp = FALSE, raster_based = FALSE)

plugin_its(f, omp = FALSE, raster_based = FALSE)
}
\arguments{
\item{f}{a function}

\item{omp}{logical is the function natively parallized with OpenMP}

\item{raster_based}{logical. For ITS and ITD algorithms, is the method raster-based or
or point-cloud-based?}
}
\description{
Tools to build plugin functions for lidR
}
\examples{
\dontrun{
mba <- function(n = 1, m = 1, h = 8, extend = TRUE) {
  f <- function(las, where) {
    res <- MBA::mba.points(las@data, where, n, m , h, extend)
    return(res$xyz.est[,3])
  }

  f <- plugin_dtm(f)
  return(f)
}

LASfile <- system.file("extdata", "Topography.laz", package="lidR")
las = readLAS(LASfile)

dtm = rasterize_terrain(las, algorithm = mba())
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algorithm-gnd.R
\name{gnd_mcc}
\alias{gnd_mcc}
\alias{mcc}
\title{Ground Segmentation Algorithm}
\usage{
mcc(s = 1.5, t = 0.3)
}
\arguments{
\item{s}{numeric. Scale parameter. The optimal scale parameter is a function of
1) the scale of the objects (e.g., trees) on the ground, and 2) the sampling
interval (post spacing) of the lidar data.}

\item{t}{numeric. Curvature threshold}
}
\description{
This function is made to be used in \link{classify_ground}. It implements an
algorithm for segmentation of ground points base on a Multiscale Curvature
Classification. This method is a strict implementation of the MCC algorithm
made by Evans and Hudak. (2007) (see references) that relies on the authors'
original source code written and exposed to R via the the \code{RMCC} package.
}
\details{
There are two parameters that the user must define, the scale (s) parameter and
the curvature threshold (t). The optimal scale parameter is a function of
1) the scale of the objects (e.g., trees) on the ground, and 2) the sampling
interval (post spacing) of the lidar data. Current lidar sensors are capable
of collecting high density data (e.g., 8 pulses/m2) that translate to a spatial
sampling frequency (post spacing) of 0.35 m/pulse (1/sqrt(8 pulses/m2) = 0.35 m/pulse),
which is small relative to the size of mature trees and will oversample larger
trees (i.e., nominally multiple returns/tree). Sparser lidar data (e.g., 0.25 pulses/m2)
translate to a post spacing of 2.0 m/pulse (1/sqrt(0.25 pulses/m2) = 2.0 m/pulse),
which would undersample trees and fail to sample some smaller trees (i.e.,
nominally <1 return/tree).\cr\cr
Therefore, a bit of trial and error is warranted to determine the best scale
and curvature parameters to use. Select a las tile containing a good variety
of canopy and terrain conditions, as it's likely the parameters that work best
there will be applicable to the rest of your project area tiles. As a starting
point: if the scale (post spacing) of the lidar survey is 1.5 m, then try 1.5.
Try varying it up or down by 0.5 m increments to see if it produces a more desirable
digital terrain model (DTM) interpolated from the classified ground returns in
the output file. Use units that match the units of the lidar data. For example,
if your lidar data were delivered in units of feet with a post spacing of 3 ft,
set the scale parameter to 3, then try varying it up or down by 1 ft increments
and compare the resulting interpolated DTMs. If the trees are large, then
it's likely that a scale parameter of 1 m (3 ft) will produce a cleaner DTM
than a scale parameter of 0.3 m (1 ft), even if the pulse density is 0.3 m
(1 ft). As for the curvature threshold, a good starting value to try might be
0.3 (if data are in meters; 1 if data are in feet), and then try varying this
up or down by 0.1 m increments (if data are in meters; 0.3 if data are in feet).
}
\examples{
if (require(RMCC, quietly = TRUE))
{
LASfile <- system.file("extdata", "Topography.laz", package="lidR")
las <- readLAS(LASfile, select = "xyzrn", filter = "-inside 273450 5274350 273550 5274450")

las <- classify_ground(las, mcc(1.5,0.3))
#plot(las, color = "Classification")
}
}
\references{
Evans, Jeffrey S.; Hudak, Andrew T. 2007. A multiscale curvature
algorithm for classifying discrete return LiDAR in forested environments.
IEEE Transactions on Geoscience and Remote Sensing. 45(4): 1029-1038.
}
\seealso{
Other ground segmentation algorithms: 
\code{\link{gnd_csf}},
\code{\link{gnd_pmf}},
\code{\link{ptd}()}
}
\concept{ground segmentation algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algorithm-itd.R
\name{itd_lmf}
\alias{itd_lmf}
\alias{lmf}
\title{Individual Tree Detection Algorithm}
\usage{
lmf(ws, hmin = 2, shape = c("circular", "square"), ws_args = "Z")
}
\arguments{
\item{ws}{numeric or function. Length or diameter of the moving window used to detect the local
maxima in the units of the input data (usually meters). If it is numeric a fixed window size is used.
If it is a function, the function determines the size of the window at any given location on the canopy.
By default function takes the height of a given pixel or point as its only argument and return the
desired size of the search window when centered on that pixel/point. This can be controled with
the `ws_args` parameter}

\item{hmin}{numeric. Minimum height of a tree. Threshold below which a pixel or a point
cannot be a local maxima. Default is 2.}

\item{shape}{character. Shape of the moving window used to find the local maxima. Can be "square"
or "circular".}

\item{ws_args}{list. Named list of argument for the function `ws` if `ws` is a function.}
}
\description{
This function is made to be used in \link{locate_trees}. It implements an algorithm for tree
detection based on a local maximum filter. The windows size can be fixed or variable and its
shape can be square or circular. The internal algorithm works either with a raster or a point cloud.
It is deeply inspired by Popescu & Wynne (2004) (see references).
}
\examples{
LASfile <- system.file("extdata", "MixedConifer.laz", package="lidR")
las <- readLAS(LASfile, select = "xyzi", filter = "-inside 481250 3812980 481300 3813050")

# =================
# point-cloud-based
# =================

# 5x5 m fixed window size
ttops <- locate_trees(las, lmf(5))

#plot(las) |> add_treetops3d(ttops)

# variable windows size
f <- function(x) { x * 0.07 + 3}
ttops <- locate_trees(las, lmf(f))

#plot(las) |> add_treetops3d(ttops)

# Very custom variable windows size
f <- function(x, y, z) { x * 0.07 + y * 0.01 + z}
ws_args <- list(x = "Z", y = "Intensity", z = 3)
ttops <- locate_trees(las, lmf(f, ws_args = ws_args))

# ============
# raster-based
# ============

chm <- rasterize_canopy(las, res = 1, p2r(0.15), pkg = "terra")
ttops <- locate_trees(chm, lmf(5))

plot(chm, col = height.colors(30))
plot(sf::st_geometry(ttops), add = TRUE, col = "black", cex = 0.5, pch = 3)

# variable window size
f <- function(x) { x * 0.07 + 3 }
ttops <- locate_trees(chm, lmf(f))

plot(chm, col = height.colors(30))
plot(sf::st_geometry(ttops), add = TRUE, col = "black", cex = 0.5, pch = 3)
}
\references{
Popescu, Sorin & Wynne, Randolph. (2004). Seeing the Trees in the Forest: Using Lidar and
Multispectral Data Fusion with Local Filtering and Variable Window Size for Estimating Tree Height.
Photogrammetric Engineering and Remote Sensing. 70. 589-604. 10.14358/PERS.70.5.589.
}
\seealso{
Other individual tree detection algorithms: 
\code{\link{itd_manual}}
}
\concept{individual tree detection algorithms}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R, R/utils_base.R
\name{print.LAS}
\alias{print.LAS}
\alias{print.LAScatalog}
\alias{print.lidRAlgorithm}
\alias{print.raster_template}
\alias{summary.LAS}
\alias{summary.LAScatalog}
\alias{tools}
\alias{dim.LAS}
\alias{dim.LAScatalog}
\alias{ncol.LAS}
\alias{nrow.LAScatalog}
\alias{names.LAS}
\alias{names.LASheader}
\alias{rbind.LAS}
\alias{npoints}
\alias{density}
\alias{density,LAS-method}
\alias{density,LASheader-method}
\alias{density,LAScatalog-method}
\title{Tools inherited from base R for LAS* objects}
\usage{
\method{print}{LAS}(x, ...)

\method{print}{LAScatalog}(x, ...)

\method{print}{lidRAlgorithm}(x, ...)

\method{print}{raster_template}(x, ...)

\method{summary}{LAS}(object, ...)

\method{summary}{LAScatalog}(object, ...)

\method{dim}{LAS}(x)

\method{dim}{LAScatalog}(x)

ncol.LAS(x)

nrow.LAScatalog(x)

\method{names}{LAS}(x)

\method{names}{LASheader}(x)

\method{rbind}{LAS}(...)

npoints(x, ...)

density(x, ...)

\S4method{density}{LAS}(x, ...)

\S4method{density}{LASheader}(x, ...)

\S4method{density}{LAScatalog}(x, ...)
}
\arguments{
\item{x}{a LAS* object}

\item{...}{LAS* objects if it is the sole argurment (e.g. in rbind())}

\item{object}{A \code{LAS*} object or other lidR related objects.}
}
\description{
Tools inherited from base R for LAS* objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_bbox.R
\name{st_bbox}
\alias{st_bbox}
\alias{st_bbox.LAS}
\alias{st_bbox.LASheader}
\alias{st_bbox.LAScatalog}
\alias{st_bbox.LAScluster}
\alias{ext,LAS-method}
\alias{ext,LASheader-method}
\alias{ext,LAScatalog-method}
\alias{ext,LAScluster-method}
\title{Bounding box of a LAS* object}
\usage{
\method{st_bbox}{LAS}(obj, ...)

\method{st_bbox}{LASheader}(obj, ...)

\method{st_bbox}{LAScatalog}(obj, ...)

\method{st_bbox}{LAScluster}(obj, ...)

\S4method{ext}{LAS}(x, ...)

\S4method{ext}{LASheader}(x, ...)

\S4method{ext}{LAScatalog}(x, ...)

\S4method{ext}{LAScluster}(x, ...)
}
\arguments{
\item{obj, x}{An object of class \code{LAS*}.}

\item{...}{unused}
}
\value{
A \code{bbox} from sf, or a \code{SpatExtent} from \code{terra}.
}
\description{
Bounding box of a \verb{LAS*} object. \code{st_bbox()} extends \code{sf}, and \code{ext()} extends \code{terra}. The values returned are similar to their
parent functions.
}
\examples{
f <- system.file("extdata", "example.las", package="rlas")
las <- readLAS(f)

st_bbox(las)
ext(las)
}
